//? cpmangle.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


//Description: see cpmangle.h

#include <cpmangle.h>

using namespace CpmStd;

using namespace CpmGeo;
using CpmRoot::R;
using CpmRoot::Z;
using CpmRoot::C;
using CpmRoot::Word;

namespace{
   const R fullAngle=cpmpi*2;
   const R iFullAngle=R(1.)/fullAngle;
   const R hour=fullAngle/24;
}

Angle Angle::deg1(1,DEG);
Angle Angle::deg90(90,DEG);
Angle Angle::deg180(180,DEG);
Angle Angle::deg270(270,DEG);
Angle Angle::deg360(360,DEG);

void Angle::normalizeAngle(R& a)
{
   if (a>-cpmpi && a<=cpmpi) return;
      // then no normalization needed. Introduced 2006-03-29
      // after it was observed that normalizing the zero-angle
      // caused difficulties
   a*=iFullAngle;
   a-=cpmfloor(a);
   if (a>0.5) a-=1;
   a*=fullAngle;
}

Angle::Angle(R angle, bool deg):a_(angle)
{
   if (deg) a_*=cpmdeg;
   normalize();
}

Angle::Angle(R angle, AngleUnit au):a_(angle)
{
   if (au==RAD){
      ;
   }
   else if (au==DEG){
      a_*=cpmdeg;
   }
   else if (au==CYCLE){
      a_*=fullAngle;
   }
   else if (au==HOUR){
      a_*=hour;
   }
   else{
      cpmerror("Angle(R, AngleUnit): unvalid AngleUnit");
   }
   normalize();
}

void Angle::setArc(R alpha)
{
   a_=alpha;
   normalize();
}

namespace{
void quantize(R& x, Z n)
{
   R y=x*n;
   y=cpmfloor(y+0.5);
   x=y/n;
}

}
Angle& Angle::operator +=(const R& s)
{
   a_+=s;
   normalize();
   return *this;
}

Angle& Angle::operator *=(const R& s)
{
   a_*=s;
   normalize();
   return *this;
}

Angle Angle::operator -(void)const
{
   Angle res;
   res.a_=-a_;
   res.normalize();
   return res;
}

Angle& Angle::operator +=(const Angle& s )
{
   a_+=(s.a_);
   normalize();
   return *this;
}

Angle& Angle::operator -=(const Angle& a1 )
{
   a_-=a1.a_;
   normalize();
   return *this;
}

Angle Angle::operator +(const Angle& a1)const
{
   return Angle(a_+a1.a_,RAD);
}

Angle Angle::operator -(const Angle& a1)const
{
   return Angle(a_-a1.a_,RAD);
}

R Angle::operator /(const Angle& a1)const
{
   R a1Inv=cpminv(a1.a_);
   return a_*a1Inv;
}

Angle Angle::con(void)const
{
   if (a_==cpmpi) return *this;
   else return Angle(-a_,RAD);
}

R Angle::cut(R c)const
{
   R lim=c*cpmdeg;
   R ar=a_+fullAngle;
   if (ar>lim) ar-=fullAngle;
   return ar;
}

bool Angle::prnOn(ostream& out)const
{
   return CpmRoot::write(a_,out);
}

bool Angle::scanFrom(istream& in)
{
   bool res=CpmRoot::read(a_,in);
   normalize();
   return res;
}

Angle CpmGeo::mean(const CpmArrays::V<Angle>& angles)
{
   Z i,n=angles.dim();
   Word loc("Angle CpmGeo::mean(const CpmArrays::V<Angle>& angles)");
   cpmassert(n>0,loc);
   C z,sum(0,0);
   for (i=1;i<=n;i++){
      z.polar(1.,angles[i].toR());
      sum+=z;
   }
   R r,phi;
   sum.toPolar(r,phi);
   return Angle(phi,RAD);
}

Angle CpmGeo::mean(const CpmArrays::V<Angle>& angles,
   const CpmArrays::V<R>& weights)
{
   Word loc("Angle CpmGeo::mean(...)");
   Z i,n=angles.dim(),n1=weights.dim();
   cpmassert(n>0,loc);
   cpmassert(n1>=n,loc);
   C z,sum(0,0);
   for (i=1;i<=n;i++){
      z.polar(1.,angles[i].toR());
      sum+=z*weights[i];
   }
   R r,phi;
   sum.toPolar(r,phi);
   return Angle(phi,RAD);
}

Z Angle::com(const Angle& a1)const
{
   if (a_<a1.a_) return 1;
   else if (a_>a1.a_) return -1;
   else return 0;
}

R Angle::dis(const Angle& phi)const
{
   Angle a1=*this - phi;
   Angle a2= phi - *this;
   R d1=cpmabs(a1.toR());
   R d2=cpmabs(a2.toR());
   return cpminf(d1,d2)*iFullAngle;
}

R Angle::toDeg(Z i)const // 'to degrees'
   {
      static R iDegree=R(180)/cpmpi;
      R b=a_*iDegree;
      if (i==1){
         if (b<0.) b+=360.;
      }
      return b;
   }

Word Angle::toWord(AngleUnit u, AngleRange r, AngleHex h, Z digits)const
{
   const R iDegree=R(180)/cpmpi;
   const R ihour=R(12)/cpmpi;
   R y=a_;
   if (r==POS && y<0) y+=fullAngle;
    Z flag=0;
   Word res, sep=":";
   const char* ff="%-#3.1f";
   const char* fff="%2d";
   if (digits==0)  ff="%-#2.0f";
   if (digits==1)  ff="%-#3.1f";
   if (digits==-1){ff="%-#3.1f"; flag=1;}
   if (digits==2)  ff="%-#4.2f";
   if (digits==3)  ff="%-#5.3f";
   if (digits==4)  ff="%-#6.4f";
   if (digits==5)  ff="%-#7.5f";
   if (digits==6)  ff="%-#8.6f";

   if (y<0){
      res="-";  // this is the first part of the result
                // further parts will be appended later
      y=-y;
   }
   else{
      res=" ";
   }

   if (u==CYCLE){
      y*=iFullAngle; // positive angle in full angles (turns)
   }
    else if (u==HOUR){
      y*=ihour; // positive angle in hours
   }
   else if (u==DEG){
      y*=iDegree; // positive angle in degrees
   }

   if (h==DEC){
      if (flag) quantize(y,2);
      res=res&CpmRoot::toWord(y,ff);
      return res;
   }

   Z ip=cpmtoz(y); // integer part (degrees or hours depending on u)
   res=res&CpmRoot::toWord(ip,fff);
   y-=ip;
   y*=60;  // y=minutes (time or angle depending on u)
   if (h==MIN){
      if (flag) quantize(y,2);
      res=res&sep&CpmRoot::toWord(y,ff);
      return res;
   }

   ip=cpmtoz(y); // integer part (minutes of time or angle)
   res=res&sep&CpmRoot::toWord(ip,fff);
   y-=ip;
   y*=60; // y=seconds (time or angle depending on u)
   if (h==SEC){
      if (flag) quantize(y,2);
      res=res&sep&CpmRoot::toWord(y,ff);
      return res;
   }
   return res;  // this should not be reached, for symmetry and for
      // avoiding a warning
}

Angle CpmGeo::RealToAngle(R x, AngleUnit u, AngleHex h)
{
   const R i60=1./60.;
   Z sign;
   Angle res;
   if (h==DEC){
      if (u==CYCLE) x*=fullAngle;
      if (u==HOUR) x*=hour;
      if (u==DEG) x*=cpmdeg;
      Angle::normalizeAngle(x);
      res.a_=x;
      return res;
   }
   if (x<0.){
      sign=-1; x=-x;
   }
   else{
      sign=1;
   } // now x is positive
   R ix=cpmfloor(x);
   R y=(x-ix)*100;
   if (h==MIN){
      x=ix+y*i60;
      if (u==CYCLE) x*=fullAngle;
      if (u==HOUR) x*=hour;
      if (u==DEG) x*=cpmdeg;
        x*=sign;
      Angle::normalizeAngle(x);
      res.a_=x;
      return res;
   }

   R iy=cpmfloor(y);
   R z=(y-iy)*100;
   if (h==SEC){
      x=ix+(iy+z*i60)*i60;
      if (u==CYCLE) x*=fullAngle;
      if (u==HOUR) x*=hour;
      if (u==DEG) x*=cpmdeg;
      x*=sign;
      Angle::normalizeAngle(x);
      res.a_=x;
      return res;
   }
   return res;
}

V<R> Angle::smoothAngleList(const V<Angle>& va)
{
   Z i,n=va.dim();
   V<R> res(n);
   for (i=1;i<=n;i++) res[i]=va[i].toR();
   for (i=2;i<=n;i++){
      R diff=res[i]-res[i-1];
      R dz=cpmabs(diff);
      if (dz<cpmpi) continue;
      R dp=cpmabs(diff+fullAngle);
      R dm=cpmabs(diff-fullAngle); //p,z,m stands for plus, zero, minus
      R dMin=cpminf(dp,dz,dm);
      if (dp==dMin) res[i]+=fullAngle;
      else if (dm==dMin) res[i]-=fullAngle;
      else ;
   }
   return res;
}

V<R> Angle::smoothList(const V<R>& vr)
{
   Z i,n=vr.dim();
   V<R> res=vr;
   for (i=2;i<=n;i++){
      R ri=res[i];
      R rim=res[i-1];
      R diff=ri-rim;
      R dz=cpmabs(diff);
      if (dz<cpmpi) continue;
      R dp=cpmabs(diff+fullAngle);
      R dm=cpmabs(diff-fullAngle); //p,z,m stands for plus, zero, minus
      R dMin=cpminf(dp,dz,dm);
      if (dp==dMin){
         res[i]+=fullAngle;
      }
      else if (dm==dMin){
         res[i]-=fullAngle;
      }
      else ;
   }
   return res;
}

void Angle::qun_(Z n)
{
   if (n<1) return; // do nothing for meaningless input
   R a=(a_+cpmpi)*iFullAngle*n; // ia a value between 0 and n
   a=cpmrnd(a);
   a*=(fullAngle/n);
   a_=a-cpmpi;
}
