//? cpmclinalg.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


////? cpmclinalg.h
////? C+- by Ulrich Mutze. Status of work 2022-03-22.
////? Copyright (c) 2022 Ulrich Mutze
////? contact: see contact-info at www.ulrichmutze.de
////?
////? This program is free software: you can redistribute it and/or
////? modify it under the terms of the GNU General Public License as
////? published by the Free Software Foundation, either version 3 of
////? the License, or (at your option) any later version.
////?
////?
////? This program is distributed in the hope that it will be useful,
////? but WITHOUT ANY WARRANTY; without even the implied warranty of
////? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
////? GNU General Public License <http://www.gnu.org/licenses/> for
////? more details.
#ifndef CPM_CLINALG_H_
#define CPM_CLINALG_H_
//   Description: classes C1, C2, C3, C4, CC2, CC3, CC4
//   Basis for effective geometry in 1,2,3,4-dimensional spaces.
#include <cpmlinalg.h>
#include <cpmperm.h>

namespace CpmLinAlg{

   using namespace CpmStd;
   using CpmRoot::operator""_R;
   using CpmRoot::operator""_Z;
   using CpmRoot::operator""_B;
   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::C;
   using CpmRoot::B;
   using CpmRoot::Word;
   using CpmArrays::V;
   using CpmArrays::T2;
   using CpmArrays::X2;
   using CpmArrays::Perm;

///////////////////////////// class C2 /////////////////////////////////
//class C4;
//class CC2;

class C2{  // tuples of two complex elements
   // In physics this is also the state space of the spin degree of
   // freedom for spin 1/2. Some of the functions to be defined here deal
   // with this situation.
public:
   C x1;
   C x2;

   void ini_(R theta, R phi){
      R ph=phi*0.5_R;
      R th=theta*0.5_R;
      x1=C::expi(-ph)*cpmcos(th);
      x2=C::expi(ph)*cpmsin(th);
   }
public:

   typedef C2 Type;

 // constructors
   C2():x1(0.),x2(0.){}
   explicit C2(C const& z):x1(z),x2(z){}
   C2(C const& a, C const& b):x1(a),x2(b){}
   explicit C2(V<C> const&);
      // Construction from dimension-independent lists.
      // No restiction for the argument.
   explicit C2(R theta, R phi=0){ini_(theta,phi);}
      // construction from the usual polar coordinates on the Bloch sphere
      // Gives always a unit vector. phi does not change the proportion
      // |x1| : |x2|. theta==0 => x2=0, ('north pole')
      // theta==pi => x1=0 ('south pole')
      // see function blcAng for getting the angles from a C2 object.

   C2(R theta, R phi, R chi){
      ini_(theta,phi);
      C omega=C::expi(chi);
      x1*=omega;
      x2*=omega;
   }

// access
   C& operator[](Z i){ if (i>=2) return x2; else return x1;}
   C const& operator[](Z i)const{ if (i>=2) return x2; else return x1;}
   void set(C t){ x1=t; x2=t;}
// mutating arithmetics (linear algebra)
   C2& operator +=(C2 const& a)
   { x1+=a.x1; x2+=a.x2; return *this;}
   C2& operator -=(C2 const& a)
   { x1-=a.x1; x2-=a.x2; return *this;}
   C2& operator *=(C const& r)
   { x1*=r; x2*=r; return *this;}
   C2& operator *=(R const& r)
   { x1*=r; x2*=r; return *this;}
   C2& neg(void)
   { x1=-x1; x2=-x2; return *this;  }
   C operator |(C2 const& a)const{ return ~x1*a.x1 + ~x2*a.x2;}
   C trc(C2 const& a)const{ return x1*a.x1 + x2*a.x2;}

   bool isVal()const{ return x1.isVal()&&x2.isVal();}
//   C2 ran(Z j=0)const{
//      if (j==0) return C2(x1.ran(),x2.ran());
//      else{
//         Z j2=j*2; return C2(x1.ran(j2),x2.ran(j2+1));
//      }
//   }
   C2 con()const{ return C2(x1.con(),x2.con());}
   C2 swap()const{ return C2(x2,x1);}
      //: swap
   C2& swap_(){ C temp=x1; x1=x2; x2=temp; return *this;}
      //: swap
      // obviously we have for the CC2 matrix sig3 and any C2 instance x:
      // sig3*x == R(r)*x ==> sig*x.swap() == R(-r)*x.swap() i.e. swap()
      // transforms a spin up state into a spin down state.
   C2 operator-()const{ return C2(-x1,-x2);}
   X2<R,C2 > polDec()const;
      //: polar decomposition
      // First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0)
   R normalize()
      { X2<R,C2> p=polDec(); *this=p.c2();  return p.c1();}
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R nor_(){ return normalize();}
      //: normalize

   R nor1_(){ R a=x1.arg(); C w=C::expi(-a); operator*=(w); return a;}
      //: normalize 1 (i.e. component 1)
      // normalizing means making real in this case

   V<C> toVC()const{ return V<C>{x1,x2};}

   R2 blcAng()const;
      //: bloch angles R2(theta,phi)

   C blc()const{ R2 ba=blcAng();
      return C::expi(ba[2])*(cpmpi-ba[1])/cpmpi;}
      // Bloch angles coded as a single C.
      // Useful for conversion to display color
      // Then theta modulates luminance: theta=0 (north pole) bright
      // theta=Pi (south pole) dark. Phi determines hue along
      // the color circle.

   R3 toR3()const
      //: to R3
      // returns (y1,y2,y3) with 0<=yi<=1;
      // Let C2n := { c \in C2 | c.abs()==1}
      // and R3n := { r \in R3 | r[1],r[2],r[3] \in Iv(0,1)}
      // then c |--> c.toR3() defines a surjective and continuous
      // mapping A: C2n --> R3n. As a surjective mapping it has a
      // right inverse.
      // The main reason for considering this function (and the next)
      // is that R3 can be naturally converted to display color.
      // See class CpmImaging::Color.
   {
      Word loc("C2::toR3()");
      R r1,r2,phi1,phi2;
      x1.pol_1_2(r1,phi1);
      x2.pol_1_2(r2,phi2);
      R phi3=C(r1,r2).arg();
      R y1=(cpmsin(phi1)+1_R)*0.5_R;
      R y2=(cpmsin(phi2)+1_R)*0.5_R;
      R y3=cpmsin(phi3);
      return R3(y1,y2,y3);
   }

   C2(R3 const& y)
      // constructor from y \in R3n (see previous function).
      // This function B: R3n --> C2n, y |--> C2n(y) is the
      // right inverse of A (as defined above) i.e.
      // A(B(y)==y
   {
      Word loc("C2(R3)");
      R r2=y[3];
      R r1=cpmsqrt(-r2*r2+1_R);
      R phi1=cpmasin(y[1]*2_R-1_R);
      R phi2=cpmasin(y[2]*2_R-1_R);
      x1=C::pol(r1,phi1);
      x2=C::pol(r2,phi2);
   }

   C2 operator*(R r)const{ return C2(x1*r,x2*r);}

   C sel(bool second)const{ return second ? x2 : x1;}

   static C2 a,b,c,d;

   static R blcTest(Z n=100);
      //: Bloch test
      // tests with random input whether
      // for any R2 r (sic!), we have z ~ z' for
      // z:=C2(r), r':= z.blcAng, z':=C2(r')
      // '~' means equality up to a common phase.
      // The accumulated error (which should be numerical
      // noise) is returned. Worked with result ~ 1e-13
      // for n=1000 (2007-12-05)

   static R codTest(Z n=1000);
      //: coding test
      // tests whether going from R3 to C2 and back to R3 gives
      // the identity (see C2(R3 const& y) for the context).

   CPM_IO
   CPM_TEST_X
   CPM_WORD
   CPM_ORDER
   CPM_NAM(C2)
   Z dim()const{ return 2;}
      //: dimension
};

inline C2 operator +(C2 const& x1, C2 const& x2)
{ C2 res=x1; return res+=x2;}

inline C2 operator -(C2 const& x1, C2 const& x2)
{ C2 res=x1; return res-=x2;}

inline C2 operator *(C2 const& x1, C const& s)
{ C2 res=x1; return res*=s;}

inline C2 operator *(C const& s, C2 const& x1)
{ C2 res=x1; return res*=s;}

///////////////////////////// class C3 /////////////////////////////////

class C3{ // tuples of 3 complex elements
public:
   typedef C3 Type;
   C x1;
   C x2;
   C x3;
public:
// constructors
   C3():x1(0.),x2(0.),x3(0.){}
   explicit C3(C const& z):x1(z),x2(z),x3(z){}
   C3(C const& a, C const& b, C const& c):x1(a),x2(b),x3(c){}
   C3(const V<C>&); // construction from dimension-independent lists
      // no restiction to the argument
   C3(R3 const& re, R3 const& im):x1(C(re.x1,im.x1)), x2(C(re.x2,im.x2)),
      x3(C(re.x3,im.x3)){}
      // construction from real part and imaginary part
// access
   C& operator[](Z i)
   { if (i>=3) return x3; else if (i==2) return x2; else return x1;}
   C const& operator[](Z i)const
   { if (i>=3) return x3; else if (i==2) return x2; else return x1;}
   void set(C t){ x1=t; x2=t; x3=t;}
// mutating arithmetics (linear algebra)
   C3& operator +=(C3 const& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; return *this;}
   C3& operator -=(C3 const& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; return *this;}
   C3& operator *=(C const& r)
   { x1*=r; x2*=r; x3*=r; return *this;}
   C3& operator *=(const R& r)
   { x1*=r; x2*=r; x3*=r; return *this;}
   C3& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; return *this;  }
   C3 operator-()const{ return C3(-x1,-x2,-x3);}
   C3 con()const{ return C3(~x1,~x2,~x3);}
   C trc(C3 const& a)const{ return x1*a.x1+x2*a.x2+x3*a.x3;}
   C operator |(C3 const& a)const
   { return ~x1*a.x1+~x2*a.x2+~x3*a.x3;}
   R abs(bool careful=true)const;
   R abs2()const{ return x1.abs2()+x2.abs2()+x3.abs2();}
   R dis(C3 const&)const;
   X2<R,C3 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0)
   R normalize()
   { X2<R,C3> p=polDec(); *this=p.c2();  return p.c1(); }
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R nor_(){ return normalize();}
   C3 ran(Z j=0)const{
      if (j==0) return C3(x1.ran(),x2.ran(),x3.ran());
      else{
         Z j3=j*3;
         return C3(x1.ran(j3),x2.ran(j3+1),x3.ran(j3+2));
      }
   }
   V<C> toVC()const{ return V<C>{x1,x2,x3};}
   CPM_IO
   CPM_WORD
   CPM_ORDER
   CPM_NAM(C3)
   Z dim()const{ return 3;}
};

inline C3 operator +(C3 const& x1, C3 const& x2)
{ C3 res=x1; return res+=x2;}

inline C3 operator -(C3 const& x1, C3 const& x2)
{ C3 res=x1; return res-=x2;}

inline C3 operator *(C3 const& x1, C const& s)
{ C3 res=x1; return res*=s;}

inline C3 operator *(C const& s, C3 const& x1)
{ C3 res=x1; return res*=s;}

///////////////////////////// class C4 /////////////////////////////////

class CC2;

class C4{  //  tuples of 4 complex elements
public:
   typedef C4 Type;
   C x1;
   C x2;
   C x3;
   C x4;
public:
// constructors
   C4():x1(0.),x2(0.),x3(0.),x4(0.){}
   explicit C4(C const& z):x1(z),x2(z),x3(z),x4(z){}
   C4(C const& a, C const& b, C const& c, C const& d):x1(a),x2(b),
      x3(c),x4(d){}
   explicit C4(const V<C>&); // construction from dimension-independent
   // lists
      // no restiction to the argument
   C4(C2 const& a, C2 const& b):x1(a.x1*b.x1),x2(a.x1*b.x2),x3(a.x2*b.x1),
      x4(a.x2*b.x2){}
      // Here we consider C4 as the tensor product C2 times C2 (not the
      // direct sum.
   C4(Word const& idle, C2 const& a, C2 const& b):x1(a.x1),x2(a.x2),
      x3(b.x1),x4(b.x2){}
      // Here we consider C4 as the direct sum of  C2 and C2
      // The dummy argument is to discern this (less important)
      // construct from the more canonical tensor product
   explicit C4(CC2 const& cc);
      // Here we set the 4 components of C4 from the 4 components
      // of CC2
// type conversion
   CC2 toCC2()const;
      // this is the inverse of the mapping defined by the previous
      // function
// access
   C& operator[](Z i)
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   C const& operator[](Z i)const
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   void set(C t){ x1=t; x2=t; x3=t; x4=t;}
// mutating arithmetics (linear algebra)
   C4& operator +=(C4 const& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; x4+=a.x4; return *this;}
   C4& operator -=(C4 const& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; x4-=a.x4; return *this;}
   C4& operator *=(C const& r)
   { x1*=r; x2*=r; x3*=r; x4*=r; return *this;}
   C4& operator *=(const R& r)
   { x1*=r; x2*=r; x3*=r; x4*=r; return *this;}
   C4& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; x4=-x4; return *this;  }
   C operator |(C4 const& a)const
   { return (~x1)*a.x1 + (~x2)*a.x2 + (~x3)*a.x3 + (~x4)*a.x4;}
   C trc(C4 const& a)const{ return x1*a.x1+x2*a.x2+x3*a.x3+x4*a.x4;}
   X2<R,C4 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0,0)
   R normalize(){ X2<R,C4> p=polDec(); *this=p.c2(); return p.c1(); }
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R nor_(){ return normalize();}
   R norm()const{ return abs(); }
   C4 rev()const{ return C4(x4,x3,x2,x1);}
      // multiplication with alpha_1 (topic: Dirac equation)
   C4 beta()const{ return C4(x1,x2,-x3,-x4);}
      // multiplication with  beta (topic: Dirac equation)
   C4 operator-()const{ return C4(-x1,-x2,-x3,-x4);}
   C4 operator*(R r)const{return C4(x1*r,x2*r,x3*r,x4*r);}

   V<C> toVC()const{ return V<C>{x1,x2,x3,x4};}
   CPM_IO
   CPM_TEST_ALL
   CPM_WORD
   CPM_ORDER
   CPM_NAM(C4)

   Z dim()const{ return 4;}
   Z b()const{ return 1;}
   Z e()const{ return 4;}
   bool isVal()const{
      return x1.isVal()&&x2.isVal()&&x3.isVal()&&x4.isVal();
   }
};

inline C4 operator +(C4 const& x1, C4 const& x2)
{ C4 res=x1; return res+=x2;}

inline C4 operator -(C4 const& x1, C4 const& x2)
{ C4 res=x1; return res-=x2;}

inline C4 operator *(C4 const& x1, C const& s)
{ C4 res=x1; return res*=s;}

inline C4 operator *(C const& s, C4 const& x1)
{ C4 res=x1; return res*=s;}

//////////// lean square matrices (2*2) - (4*4) /////////////////////////

// comment to the last function where matrices are acting on vectors
// from the right side. This kind of action on a vector space gives a
// representation of the multiplicative monoid of matrices if the product
// law for endomorphisms of linear spaces is defined as contravariant.
// In our specific notation here:
// x*(a*b)=(x*a)*b
// Proof: x*(a*b)=~(a*b)*x=((~b)*(~a))*x=(~b)*((~a)*x)=((~a)*x)*b=(x*a)*b

/// notice: operator ~ not yet analysed and tested for the complex case
// !!!!!!!!!

///////////////////////////// class CC2 ///////////2//////////////////////
class CCC2;
class CCCC2;
class CC2{  //  complex 2*2 matrices

public:
   friend class CCC2;
   friend class CCCC2;
   typedef CC2 Type;
   C2 x1; // this suggests that CC2 is the direct sum C2 + C2 which is a
      // 4-dimensional complex linear space. This can also viewed as
      // the tensor product of C2 times C2. Reason is
      // 2 + 2 = 2*2. Notice that n=2 is the only non-vanishing solution
      // of n + n = n*n.
   C2 x2;
      // The usual matrix element xij is written in terms of the
      // data elements x1, x2 as xi.xj which implies that x1 is the first
      // row of the matrix and x2 the second row. This agrees with
      // Mathematica's way of representing matrices as lists of lists.
   static CC2 zero;
   static CC2 unit;
   static CC2 sig0,sig1,sig2,sig3;
      // Pauli sigma matrices, also named sigma_x, sigma_y, sigma_z;
      // sig0 is the unit matrix.
   CPM_IO
   CPM_TEST_ALL
      // this interface was added 2016-05-22 to facilitate testing.
      // Would be natural to have it implemented for all classes defined
      // clinalg.
   CPM_MAT_ALG(C2,CC2)  // see cpmlinalg.h
   CPM_NAM(CC2)
// constructors
   CC2():x1(),x2(){}
   CC2( C2 const& a1, C2 const& a2):x1(a1),x2(a2){}
   CC2(C z11, C z12, C z21, C z22):x1(z11,z12),x2(z21,z22){}
   explicit CC2(C4 const& c4):x1(c4[1],c4[2]),x2(c4[3],c4[4]){}
   explicit CC2(V<C> const& c4):x1(c4[1],c4[2]),x2(c4[3],c4[4]){}
   CC2(R4 const& x):x1(C(x.x1+x.x4),C(x.x2,-x.x3)),x2(C(x.x2,x.x3),
      C(x.x1-x.x4)){}
      // hermitean matrix generated from a four-vector, which is the basis
      // of spinor calculus
// access
   C2& operator[](Z i){ if (i>=2) return x2; return x1;}
   C2 const& operator[](Z i)const{ if (i>=2) return x2; return x1;}
   C det()const{ return x1.x1*x2.x2-x1.x2*x2.x1;}
   C operator|(CC2 const& cc)const{ return (hc()*cc).trc();}
//   CC2 operator !()const;
      // inverse matrix can be expressed easily in this case
   CC2 her()const;
      //: hermitean (conjugate)
      // returns the hermitean conjugate
   CC2 hc()const{ return her();}
      //. hermitean conjugation
   CC2 trn()const;
      //: transpose

   CC2 com(CC2 const& cc)const{ return (*this)*cc-cc*(*this);}
      //: commutator

   C act(C z)const{ return (x1.x1*z+x1.x2)/(x2.x1*z+x2.x2);}
      // CC2 acts on C as a 'broken linear transformation'

   CC2 operator*(C z)const{ return CC2(x1*z,x2*z);}

   CC2 operator*(R r)const{ return CC2(x1*r,x2*r);}

   CC2& operator+=(CC2 const& cc2)
   {
      x1+=cc2.x1;
      x2+=cc2.x2;
      return *this;
   }

   C4 toC4()const{ return C4(x1.x1,x1.x2,x2.x1,x2.x2);}
      // one has: CC2 x => x.toC4().toCC2()==x
      // and:     C4 y => y.toCC2().toC4()==y

   V<C> toVC()const{ return V<C>{x1.x1,x1.x2,x2.x1,x2.x2};}
      //: to V<C>

   CC2 per(Perm const& p)const;
      //: permute
      // assuming that p.dim()==4 the permutation p mixes the 4 components
      // and takes these as the components of a new CC2-matrix which
      // gets returned.

   R4 toR4(bool careful=true)const
      // recovering the four-vector from the hermitean matrix.
      // If *this is known to be hermitean one may call the function
      // with careful==false.
   {
      R y0,y1,y2,y3;
      // *this is the matrix
      //  y0 + y3  y1 -iy2
      //  y1 +iy2  y0 - y3
      // and the simple task is to extract the yi.
      if (careful){
         CC2 h=(*this+her())*0.5;
         y0=(h.x1.x1+h.x2.x2)[1]*0.5;
         y3=(h.x1.x1-h.x2.x2)[1]*0.5;
         y1=(h.x2.x1+h.x1.x2)[1]*0.5;
         y2=(h.x2.x1-h.x1.x2)[2]*0.5;
      }
      else {
         y0=(x1.x1+x2.x2)[1]*0.5;
         y3=(x1.x1-x2.x2)[1]*0.5;
         y1=(x2.x1+x1.x2)[1]*0.5;
         y2=(x2.x1-x1.x2)[2]*0.5;
      }
      return R4(y0,y1,y2,y3);
   }

   C2 trc2(C2 const& a)const{
      C u1=a.x1.con();
      C u2=a.x2.con();
      return C2(u1*x1.x1+u2*x1.x2,u1*x2.x1+u2*x2.x2);
   }

   C2 trc1(C2 const& a)const{
      C u1=a.x1.con();
      C u2=a.x2.con();
      return C2(u1*x1.x1+u2*x2.x1,u1*x1.x2+u2*x2.x2);
   }

   C trc()const{ return x1.x1+x2.x2;}
      //: trace

   CC2& trcLess_()
   // changes *this into a version,the trace of which vanishes
   {
      C tr=trc();
      C trh=tr*0.5_R;
      x1.x1-=trh;
      x2.x2-=trh;
      return *this;
   }

   C app(C2 const& a, bool second)const
      //: apply
      // returns the first or the second component of (*this)*a
      // depending of the speakingly named second argument.
      // A boolean argument is chosen here since it is available in this
      // form most directly in function Qreg::app_
   {
      return
         second ? x2.x1*a.x1+x2.x2*a.x2 : x1.x1*a.x1+x1.x2*a.x2;
   }

   C biLinForm(C2 const& a, C2 const& b)const
      //: bi-linear form
   {
      return x1.x1*a.x1*b.x1 +
             x1.x2*a.x1*b.x2 +
             x2.x1*a.x2*b.x1 +
             x2.x2*a.x2*b.x2 ;
   }

   C operator()(bool i, bool j)const{
      if (i && j)  return x2.x2;
      if (i && !j) return x2.x1;
      if (!i && j) return x1.x2;
      if (!i && !j)return x1.x1;
      return C();
   }

   V< V<C> > toVVC()const{ return V< V<C> >{x1.toVC(),x2.toVC()};}

   CC2 operator-()const{ return CC2(-x1,-x2);}

   CC2 cc()const{ return CC2(x1.con(),x2.con());}
   //. complex conjugate

   CC2 map(F<R,C> const& f)const;
      // For a CC2-matrix mat the CC2 matrix mat.map(f) is just what in the
      // functional calculus of operators would be written as
      // f((mat+mat*)/2).

   CC2 exp(R const& t, R const& dt=0.001_R)const;
      // mat.exp(t,dt) is an approximate value of exp(t*mat) in which
      // dt is the step width of the underlying step algorithm.

   CC2 mapLean(F<R,C> const& f)const;
      // For a CC2-matrix mat the CC2 matrix mat.map(f) is just what in the
      // functional calculus of operators would be written as
      // f((mat+mat*)/2).
      // This is an implementation based on exact formulas which are
      // available for matrices as small as 2 by 2 (verified by means of
      // MATHEMATICA. In my program ~/e/cpm/sam1/sam it turned out to
      // take quite precisely the same time as the previous function
      // CC2 map(F<R,C> const& f)const; in the implementation based on
      // Eigen.

    X2<R2,CC2> specRep()const;
      // The first component of the result is the pair of real eigenvalues.
      // As for the previous function, from the matrix *this only the
      // hermitean part enters the algorithm. The second component is a
      // matrix which yields the eigenvectors row by row in the order
      // of the eigenvalues.

   R specRad()const{ return specRep().c1().pos().sup();}
      // spectral radius of the hermitean part of *this.
      // It equals the operator norm of this hermitean part.

   R opNorm()const;
      // operator norm. No hermiticity assumed.

   R norm()const{ return specRad();}
      // see pecRad()

   R hsn()const{ return cpmsqrt(abs2());}
      //. Hilbert Schmidt norm

   static CC2 dydPrd(C2 const& a, C2 const& b)
      //: dyadic product
      // building a matrix as a dyadic product of vectors
      // (tensor product)
      // A probably more canonical realization of this functionality
      // would be a constructor CC2(C2 const& a, C2 const& b).
      // However such constructor (with a different simple functionality)
      // was defined earlier to satisfy indispensible needs.
   {
      CC2 res;
      res.x1.x1=a.x1*~b.x1; // n o t  ~a.x1*b.x1 ...
         // In Dirac notation the components of x with respect to a basis
         // (e_i) are defined as x_i := <e_i|x>. If one would use <x|e_i>
         // instead, one had (alpha*x)_i = <alpha*x|e_i> =
         // alpha.con()<x|e_i> = alpha.con() x_i for any C alpha. This
         // would be an unacceptable deviation from established rules.
      res.x1.x2=a.x1*~b.x2;
      res.x2.x1=a.x2*~b.x1;
      res.x2.x2=a.x2*~b.x2;
      return res;
   }

   static R testPauli(){
   // see Mathematica script TwobytwoZc.nb, 2022-04-23
   // Not clear whether the C2 elements a,b,c,d are the only choices
   // that allow to represent the Pauli matrices by formulas of this
   // simplicity.
   // Test at 2022-04-23 gave result 0 and thus proved correctness.

      CC2 ab=dydPrd(C2::a,C2::b) + dydPrd(C2::b,C2::a);
      CC2 bc=dydPrd(C2::b,C2::c) + dydPrd(C2::c,C2::b);
      CC2 aabd=dydPrd(C2::a,C2::a) + dydPrd(C2::b,C2::d);
      CC2 aabb=dydPrd(C2::a,C2::a) + dydPrd(C2::b,C2::b);

      R err=0_R;
      err+=(ab - CC2::sig1).abs();
      err+=(bc - CC2::sig2).abs();
      err+=(aabd - CC2::sig3).abs();
      err+=(aabb - CC2::sig0).abs();

      return err;
   }

   static V<CC2> bas();
      //: basis
      // canonical basis
};

C2 operator*(const CC2&, const C2&);
CC2 operator*(const CC2& , const CC2&);
CC2 com(const CC2& , const CC2&); // commutator
CC2 comFac(const CC2& , const CC2&, const C& ); // commutator with factor
CC2 operator+(const CC2& , const CC2&);
CC2 operator~(const CC2&);
C2 operator*(const C2&, const CC2&);
//CC2 operator-(const CC2&, const CC2&);

///////////////////////////// class CC3 /////////////////////////////////

class CC3{  //  complex 3*3 matrices
public:
   typedef CC3 Type;
   C3 x1; // this suggests CC3 = C3 + C3 + C3 (where '+' means the
      // direct sum. This is a 9-dimensional complex linear space.
   C3 x2;
   C3 x3;
      // Notice that n=3 is the only non-trivial solution of
      // n + n + n = n*n
public:
   static CC3 zero;
   static CC3 unit;
// constructors
   CC3(){}
   CC3(C3 const& a1, C3 const& a2, C3 const& a3):x1(a1),x2(a2),x3(a3){}
// access
   C3& operator[](Z i)
   { if (i>=3) return x3; if (i==2) return x2; return x1;}

   C3 const& operator[](Z i)const
   { if (i>=3) return x3; if (i==2) return x2; return x1;}

   CC3 operator-()const{ return CC3(-x1,-x2,-x3);}

   CC3 cc()const{ return CC3(x1.con(),x2.con(),x3.con());}
   //. complex conjugate

   V< V<C> > toVVC()const{ return
      V< V<C> >{x1.toVC(),x2.toVC(),x3.toVC()};}
   R abs2()const{ return x1.abs2()+x2.abs2()+x3.abs2();}
   R abs()const{ return cpmsqrt(abs2());}
   R hsn()const{ return abs();}
      //. Hilbert Schmidt norm
   X2<R3,CC3> specRep()const;
      // The first component of the result is the pair of real eigenvalues.
      // As for the previous function, from the matrix *this only the
      // hermitean part enters the algorithm. The second component is a
      // matrix which yields the eigenvectors row by row in the order
      // of the eigenvalues.

   R specRad()const{ return specRep().c1().pos().sup();}
      // spectral radius of the hermitean part of *this.
      // It equals the operator norm of this hermitean part.

   R opNorm()const;
      // operator norm. No hermiticity assumed.

   R norm()const{ return specRad();}
      // see pecRad()

   CPM_IO
   CPM_MAT_ALG(C3,CC3)
};

C3 operator*(const CC3&, const C3&);
CC3 operator*(const CC3& , const CC3&);
CC3 operator+(const CC3& , const CC3&);
CC3 operator~(const CC3&);
C3 operator*(const C3&, const CC3&);
CC3 operator-(const CC3&, const CC3&);

///////////////////////////// class CC4 /////////////////////////////////

class CC4{  // complex 4*4 matrices
public:
   typedef CC4 Type;
   C4 x1;
   C4 x2;
   C4 x3;
   C4 x4;

   // this suggests CC4 = C4 + C4 + C4 + C4 (where '+' means the
   // direct sum. This is a 16-dimensional complex linear space.
   // Notice that n=4 is the only non-trivial solution of
   // n + n + n + n = n*n

// constructors

   CC4(){}
   CC4(C4 const& a1, C4 const& a2, C4 const& a3, C4 const& a4):
      x1(a1),x2(a2),x3(a3),x4(a4){}

   CC4(C4 const& a, C4 const& b):x1(a.x1*b),x2(a.x2*b),
      x3(a.x3*b),x4(a.x4*b){}
         // terse but correct coding of c_ij = a_i * b_j,
         // dyadic product |a><b*|

   CC4(CC2 const& a);
      // Consider the linear mapping L(a): CC2 --> CC2 , x |--> a.com(x)
      // The natural mapping T: CC2 --> C4 allows to form the linear
      // mapping T L(a) inv(T) : C4 --> C4. The matrix corresponding to
      // this mapping is the CC4 created by this constructor.

   CC4(CC2 const& a, CC2 const& b)
   {
      *this=CC4(a.toC4(),b.toC4());
   }

   CC4(CC2 const& a11, CC2 const& a12, CC2 const& a21, CC2 const& a22)
   {
      x1.x1=a11.x1.x1;
      x1.x2=a11.x1.x2;
      x2.x1=a11.x2.x1;
      x2.x2=a11.x2.x2;

      x3.x1=a12.x1.x1;
      x3.x2=a12.x1.x2;
      x4.x1=a12.x2.x1;
      x4.x2=a12.x2.x2;

      x1.x3=a21.x1.x1;
      x1.x4=a21.x1.x2;
      x2.x3=a21.x2.x1;
      x2.x4=a21.x2.x2;

      x3.x3=a22.x1.x1;
      x3.x4=a22.x1.x2;
      x4.x3=a22.x2.x1;
      x4.x4=a22.x2.x2;
   }

   explicit CC4(V<C> const& vc);
      // constructor from 16 complex numbers

// access
   C4& operator[](Z i)
      { if (i>=4) return x4; if (i==3) return x3; if (i==2) return x2;
       return x1;}

   C4 const& operator[](Z i)const
       { if (i>=4) return x4; if (i==3) return x3; if (i==2) return x2;
        return x1;}

   CC4 operator-()const{ return CC4(-x1,-x2,-x3,-x4);}

   CC4 operator*(R r)const{ return CC4(x1*r,x2*r,x3*r,x4*r);}

   CC4 operator*(C z)const{ return CC4(x1*z,x2*z,x3*z,x4*z);}

   R hsn()const;

   X2<R4,CC4> specRep()const;
      // The first component of the result is the pair of real eigenvalues.
      // As for the previous function, from the matrix *this only the
      // hermitean part enters the algorithm. The second component is a
      // matrix which yields the eigenvectors row by row in the order
      // of the eigenvalues.

   R specRad()const{ return specRep().c1().pos().sup();}
      // spectral radius of the hermitean part of *this.
      // It equals the operator norm of this hermitean part.

   R norm()const{ return specRad();}
      // see pecRad()

   R opNorm()const;
      // operator norm, no hermiticy assumed

   CC4 hc()const{ return ~CC4(x1.con(),x2.con(),x3.con(),x4.con());}
      //. hermitian conjugate

   CC4 cc()const{ return CC4(x1.con(),x2.con(),x3.con(),x4.con());}
   //. complex conjugate

   V< V<C> > toVVC()const{ return
      V< V<C> >{x1.toVC(),x2.toVC(),x3.toVC(),x4.toVC()};}

   V<C> toVC()const{ return x1.toVC()&x2.toVC()&x3.toVC()&x4.toVC();}

   CPM_IO
   CPM_TEST_ALL
   CPM_MAT_ALG(C4,CC4) // ~ is transposition
   Word nameOf()const{ return "CC4";}

};

C4 operator*(const CC4&, const C4&);
CC4 operator*(const CC4& , const CC4&);
CC4 operator+(const CC4& , const CC4&);
CC4 operator~(const CC4&);
C4 operator*(const C4&, const CC4&);
CC4 operator-(const CC4&, const CC4&);

///////////////////////////// class Gamma //////////////////////////////
// Presenting means for defining Dirax 4x4 matrices: alpha, beta, gamma
// by simple initializers
class Gamma{
   CC4 g_;
   Gamma(CC4 const& g):g_(g){}
public:
   Gamma(Z i)
   {
      if (i==0)
         g_=CC4(CC2::sig0, CC2::zero, CC2::zero, -CC2::sig0);
      else if (i==1)
         g_=CC4(CC2::zero, CC2::sig1, -CC2::sig1, CC2::zero);
      else if (i==2)
            g_=CC4(CC2::zero, CC2::sig2, -CC2::sig2, CC2::zero);
      else if (i==3)
            g_=CC4(CC2::zero, CC2::sig3, -CC2::sig3, CC2::zero);
      else if (i==5)
            g_=CC4(CC2::zero, CC2::sig0, CC2::sig0, CC2::zero);
      else
            cpmerror("unvalid i = "&cpm(i)&" in Gamma(Z i)");
   }

   static CC4 beta(){ return Gamma(0).rep(); }
   static CC4 alpha(Z i){ return Gamma(0).rep()*Gamma(i).rep();}

   Gamma(Word const& w)
   {
      if (w=="unit")
             g_=CC4(CC2::sig0,CC2::zero,CC2::zero,CC2::sig0);
      else if (w=="zero")
             g_=CC4(CC2::zero,CC2::zero,CC2::zero,CC2::zero);
      else if (w=="U_Weyl")
            g_=CC4(CC2::sig0,-CC2::sig0,CC2::sig0,CC2::sig0)*
               cpminv(cpmsqrt(R(2.)));
      else if (w=="U_Majorana")
            g_=CC4(CC2::sig0,CC2::sig2,CC2::sig2,-CC2::sig0)*
               cpminv(cpmsqrt(R(2.)));
            // from Itzykson Zuber: Quantum Field Theory p. 694
      else cpmerror("unvalid w in Gamma(Word w)");
   }

   Gamma(Z i, Word const& w)
   {
      if (w=="Dirac") g_=Gamma(i).rep();
      else if (w=="Weyl"){
         CC4 U = Gamma("U_Weyl").rep();
         g_ = U*Gamma(i).rep()*U.hc();
      }
      else if (w=="Majorana"){ // For i=0,1,2,3,5 the matrix
         // Gamma(i,"Majorana") is purely imaginary. Tested!
         CC4 U = Gamma("U_Majorana").rep();
         g_ = U*Gamma(i).rep()*U.hc();
      }
      else cpmerror("unvalid w in Gamma(Z i, Word w)");
   }

   static CC4 beta(Word const& w)
      { return Gamma(0,w).rep(); }

   static CC4 alpha(Z i,Word const& w)
      { return Gamma(0,w).rep()*Gamma(i,w).rep();}

   CC4 rep()const{ return g_;}
};

/// special topic: tensor products of C2: CCC2, CCCC2
/// for my spin interaction application I do not longer need CC4
/// and use CCCC2 instead. CC4 is needed for Dirac equation in
/// cpmstatedirac.h.

class CCC2{ // pairs of CC2
   CC2 x1,x2;
   // this suggests CC2 = CC2 + CC2 which is a linear complex space
   // of dimension 2*4 = 8.
   CCC2(CC2 const& y1,CC2 const& y2):x1(y1),x2(y2){}
public:
   friend class CCCC2;
   CCC2():x1(),x2(){}
   CCC2 operator + (CCC2 const& a)const{ return CCC2(x1+a.x1,x2+a.x2);}
   CCC2 operator - (CCC2 const& a)const{ return CCC2(x1-a.x1,x2-a.x2);}
   CC2 sel(bool second)const{ return second ? x2 : x1;}
      //: select
   CCC2 operator* (R r)const
   {
      return CCC2(x1*r,x2*r);
   }

   R abs2()const{ return x1.abs2() + x2.abs2();}
};

class CCCC2{ // pairs of CCC2, for implementation
   // of tensor products of sigma matrices
   CCC2 x1,x2;
   // This suggests CCCC2 = CCC2 + CCC2 which is a linear complex space
   // of dimension 2*8 = 16. This is the tensor product of two
   // linear complex spaces of dimension 4 e.g. of CC2 and CC2
   CCCC2(CCC2 const& y1,CCC2 const& y2):x1(y1),x2(y2){}
public:
   CCCC2():x1(),x2(){}

// constructor for the tensor product of a with b
   CCCC2(CC2 const& a, CC2 const& b):x1(),x2()
   {
      x1.x1.x1.x1 = a.x1.x1 * b.x1.x1;
      x1.x1.x1.x2 = a.x1.x1 * b.x1.x2;
      x1.x1.x2.x1 = a.x1.x1 * b.x2.x1;
      x1.x1.x2.x2 = a.x1.x1 * b.x2.x2;

      x1.x2.x1.x1 = a.x1.x2 * b.x1.x1;
      x1.x2.x1.x2 = a.x1.x2 * b.x1.x2;
      x1.x2.x2.x1 = a.x1.x2 * b.x2.x1;
      x1.x2.x2.x2 = a.x1.x2 * b.x2.x2;

      x2.x1.x1.x1 = a.x2.x1 * b.x1.x1;
      x2.x1.x1.x2 = a.x2.x1 * b.x1.x2;
      x2.x1.x2.x1 = a.x2.x1 * b.x2.x1;
      x2.x1.x2.x2 = a.x2.x1 * b.x2.x2;

      x2.x2.x1.x1 = a.x2.x2 * b.x1.x1;
      x2.x2.x1.x2 = a.x2.x2 * b.x1.x2;
      x2.x2.x2.x1 = a.x2.x2 * b.x2.x1;
      x2.x2.x2.x2 = a.x2.x2 * b.x2.x2;
   }

// self-evident
   CCCC2 operator + (CCCC2 const& a)const{ return CCCC2(x1+a.x1,x2+a.x2);}

   // self-evident
   CCCC2 operator - (CCCC2 const& a)const{ return CCCC2(x1-a.x1,x2-a.x2);}

   R abs2()const{return x1.abs2()+x2.abs2();}

// the isotropic interaction term between two spin 1/2
// magnetic moments (or Heisenberg exchange interaction
// between spins)
   static CCCC2 spinSpin(R q, bool full=true){
      CCCC2 resPrel=CCCC2(CC2::sig1*q,CC2::sig1)
              + CCCC2(CC2::sig2*q,CC2::sig2);
      if (full) return resPrel+CCCC2(CC2::sig3*q,CC2::sig3);
      else return resPrel;
   }

// more flexible form for the interaction term between two spin 1/2
// magnetic moments
   static CCCC2 spinSpinXYZ(R qXY, R qZ){
      CCCC2 res;
      if (qXY != 0_R) res = res + CCCC2(CC2::sig1*qXY,CC2::sig1)
              + CCCC2(CC2::sig2*qXY,CC2::sig2);
      if (qZ != 0_R) res = res + CCCC2(CC2::sig3*qZ,CC2::sig3);
      return res;
   }

// Returns *this(a,b) where *this is considered as a bi-linear form.
   CC2 operator()(C2 const& a, C2 const& b)const
   {
      CC2 res;
      res.x1.x1= x1.x1.x1.x1 * a.x1 * b.x1
               + x1.x1.x1.x2 * a.x1 * b.x2
               + x1.x1.x2.x1 * a.x2 * b.x1
               + x1.x1.x2.x2 * a.x2 * b.x2 ;

      res.x1.x2= x1.x2.x1.x1 * a.x1 * b.x1
               + x1.x2.x1.x2 * a.x1 * b.x2
               + x1.x2.x2.x1 * a.x2 * b.x1
               + x1.x2.x2.x2 * a.x2 * b.x2 ;

      res.x2.x1= x2.x1.x1.x1 * a.x1 * b.x1
               + x2.x1.x1.x2 * a.x1 * b.x2
               + x2.x1.x2.x1 * a.x2 * b.x1
               + x2.x1.x2.x2 * a.x2 * b.x2 ;

      res.x2.x2= x2.x2.x1.x1 * a.x1 * b.x1
               + x2.x2.x1.x2 * a.x1 * b.x2
               + x2.x2.x2.x1 * a.x2 * b.x1
               + x2.x2.x2.x2 * a.x2 * b.x2 ;

      return res;
   }

   CCC2 sel(bool second)const{ return second ? x2 : x1;}
      //: select

   CC2 operator * (CC2 const& a)const;
      // strangling the second and the fourth index of *this by the
      // 2*2 matrix a and thus getting a 2*2 matrix, which gets returned.

   C app(CC2 const& a, bool b1, bool b2)const;
      // returns the matrix element (i,j) of (*this)*a
      // where
      // i=1 for b1=false, i=2 for b1=true
      // j=1 for b2=false, j=2 for b2=true
      // Only those products and sums are computed which
      // make up the requested matrix element. Thus if not
      // all matrix elements of (*this)*a are needed, we do less
      // work than computing the whole matrix (*this)*a.
};

#undef CPM_MAT_ALG

} // namespace

#endif
