//? cpmcmatrix.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_CMATRIX_H_
#define CPM_CMATRIX_H_
/*

   Purpose: class of complex matrices together with singular value
      decomposition and eigenvalues/eigenvectors for square
      Hermitean matrices.

*/
#include <cpmcvector.h>
#include <cpmrmatrix.h>
#include <cpmf.h>

namespace CpmArrays{

   using CpmFunctions::F;
   using CpmLinAlg::R2;
#if defined(CPM_USE_EIGEN)
   using namespace Eigen;
   #include <eigen3/Eigen/Dense>
   #if defined(CPM_MP) || defined(CPM_QUAD)
      #include <boost/multiprecision/eigen.hpp>
   #endif
#endif

/*
Since 2020-04-17 Eigen cooperates with boost/multiprecision. However,
compilation works only with compileroption -fpermissive
*/

//////////////////////// class C_Matrix //////////////////////////////
//
// Matrices with matrix elements of type C. This class implements the
// major algorithms associated with matrices. The most complex (and
// useful) one is forming the pseudo-inverse !a of an arbitrary
// (m,n)-matrix a. Also to be noticed: ~a is the hermitian conjugate of a.
// Matrix algebra is expressed in terms of *,+,-, and even
// (a/b := a*!b, of course ). All natural ways to construct
// C_Matrices from C_Vectors are implemented. Further C_Matrix objects
// can be appended row-wise using operator & and column-wise using
// operator &&.

class C_Matrix{ // complex matrices, rich interface
   Vr<C_Vector> rep;
   C_Matrix(const V<C_Vector>& x):rep(x){}

public:
   static R etaInv;
      // fraction of the largest 'singular value' which is considered
      // 'intentionally different from zero'. If this is negative, a
      // smooth version of the normal pseudo-inverse is chosen.
   typedef C_Matrix Type;
   typedef C ScalarType;
   CPM_IO
   CPM_ORDER
   CPM_TEST
   CPM_DESCRIPTORS
   CPM_CONJUGATION
   CPM_SUM_M
   CPM_PRODUCT_C
   CPM_SCALAR_M
   CPM_DIFFERENCE_PLAIN
   CPM_DIVISION
   CPM_DOT_PRODUCT
   CPM_NORMALIZE
      // Inversion is pseudo-inversion in our case. This is always
      // defined and very useful. An internal test on correctness of
      // inversion (actually of singular value decomposition) will be
      // done if cpmverbose>=3

// constructors
   C_Matrix():rep(1,C_Vector(1)){}
      // creating a (1,1) zero-matrix

   explicit C_Matrix(Z n):rep(n,C_Vector(n)){}
      // creating a (n,n) zero-matrix

   C_Matrix(Z m, Z n):rep(m,C_Vector(n)){}
       // creating a (m,n) zero-matrix

   C_Matrix(Z m, Z n, C z):rep(m,C_Vector(n,z)){}
       // creating a (m,n) matrix with all matrix elements
       // equal to r

   explicit C_Matrix(const VV<C>&);
      // evident

   explicit C_Matrix(const V< V<C> >&);
      // evident

   explicit C_Matrix(C_Vector const& diag):
      rep(diag.dim(), C_Vector(diag.dim()))
   { for (Z i=1;i<=diag.dim();i++)  rep[i][i]=diag[i];}
      // creating a diagonal matrix from a list diag of diagonal elements

   explicit C_Matrix(R_Vector const& diag):
      rep(diag.dim(), C_Vector(diag.dim()))
   { for (Z i=1;i<=diag.dim();i++)  rep[i][i]=C(diag[i]);}

   C_Matrix(Z m, C_Vector const& row):rep(m,row){}
       // creating a (m,row.dim()) matrix made out of m row-vectors row

   C_Matrix(C_Vector const& col, Z n);
       // creating a (col.dim(),n) matrix made out of n column-vectors col
       //

   C_Matrix(C_Vector const& a, C_Vector const& b);
      // dyadic product of the two vectors a and b
      // Let mat be the constructed matrix, then mat[i][j]=a[i]*~b[j]

   C_Matrix(F<C_Vector,C_Vector> const& f , Z n);
      // Let f : C_Vector|n -> C_Vector|n be a linear function. With
      // respect to the canonical base in C_Vector|n this creates a
      // complex n x n - matrix which the present function
      // is designed to construct. Here, of course, C_Vector|n is the
      // subset of C_Vector of vectors with valid indexes 1,2,...,n.

#if defined(CPM_USE_EIGEN)
   explicit C_Matrix( Matrix<C,Dynamic,Dynamic> const& );

   Matrix<C,Dynamic,Dynamic> toEigen()const;

#endif

// conversions
   VV<C> toVV()const;
      // evident

   V< V<C> > toVVC()const;
      // evident
      // For V< V<C> > we have the most advanced graphical representation
      // function in cpmgraph.h/cpp

   V<C_Vector> toRep()const{ return rep;}
      // evident

// accessors
   Z dim()const{ return rep.dim();}
      //: dimension

   Z dim1()const{ return rep.dim();}
      // number of rows

   Z dim2()const{ return rep.fir().dim();}
      // number of columns

   IvZ dom1()const{ return rep.dom();}
      // domain of index 1

   IvZ dom2()const{ return rep.fir().dom();}
      // domain of index 2

   Z b()const{ return rep.b();}
      //. begin

   Z e()const{ return rep.e();}
      //. end

   C_Vector fir()const{ return rep.fir();}
      //: first

   C_Vector last()const{ return rep.last();}
      //: last

   Vr<C_Vector> const& c()const{ return rep;}
   Vr<C_Vector> & c(){ return rep;}

   C_Vector& operator[](Z i){ return rep[i];}

   const C_Vector& operator[](Z i)const{ return rep[i];}

   C operator()(Z i, Z j, Outside mode)const{
      Z ic=makeIndValNotMember(i,dom1(),mode);
      Z jc=makeIndValNotMember(j,dom2(),mode);
      return rep[ic][jc];
   }

   const C_Vector& cyc(Z i)const{ return rep.cyc(i);}
      //: cyclic

   const C_Vector& read(Z i, Outside mode)const{ return rep.cyc(i);}


   C_Vector& cyc(Z i){ return rep.cyc(i);}
      //: cyclic

   C_Vector& cui(Z i){ return rep.cui(i);}
      //. component (with) unchecked index
   const C_Vector& cui(Z i)const{ return rep.cui(i);}
      //. component (with) unchecked index

   C_Vector getCol(Z j)const;
      // returns the vector (rep[1][j],...rep[m][j])
      // (m=dim1())

   void setCol(Z j, const C_Vector& v);
      // replaces (rep[1][j],...rep[m][j]) by
      // v[1],...v[k],rep[k+1][j],..rep[m][j], k=v.dim()
      // (in most cases k==m, the the j-th column of rep
      // gets simply replaced by v

   C_Matrix trn()const;
      //: transpose
      // returns the transposed matrix

   C_Matrix cc(void)const;
      //. complex conjugate
      // returns the element-wise complex conjugate matrix

   C_Matrix hc(void)const{ return con();}
      //. hermitian conjugate

   C_Matrix permuteColumns(const V<Z>& per);

   C_Matrix permuteRows(const V<Z>& per)
   { return C_Matrix( rep.permute(per));}

   C_Matrix ins(Z i, const C_Vector& v)const
   { return rep.ins(i,v);}
      // Let b=a.ins(i,v);
      // Then we have
      //    b[j]=a[j]   for j<i
      //    b[j]=v      for j==i
      //    b[j]=a[j-1] for j>i
      // see also Vl<T>::ins(Z, const T&)

   C_Matrix ins(Z i, const C_Matrix& v)const
   {  return rep.ins(i,v.toRep());}
      // Let b=a.ins(i,v);
      // d=v.dim()
      // Then we have
      //    b[j]=a[j]   for j<i
      //    b[j+k]=v[k] for 1<=k<=d
      //    b[j]=a[j-d] for j>j+d
      // see also Vl<T>::ins(Z, const Vl<T>&)

   C_Matrix resize(Z newDim)const
   { return C_Matrix(rep.resize(newDim));}

   C_Matrix resize2(Z newDim2)const
   { return con().resize(newDim2).con();}
      // resizing the range of the second index
      // Could be made twice as fast by a independent implementation.
      // Here I prefer the principle to have each concept in one place.

   C_Matrix eliFirst(Z nDel=1)const
   { return C_Matrix(rep.eliFirst(nDel));}
      // returs a matrix with the first line eliminated from *this

   C_Matrix eliFirst2(Z nDel=1)const
   { return con().eliFirst(nDel).con();}
      // returs a matrix with the first column eliminated from *this

   C_Matrix eli1(IvZ const& iv)const
   { return C_Matrix(rep.eli(iv));}
      // Returns a matrix with the lines indexes by iv eliminated from
      // *this

   C_Matrix eli2(IvZ const& iv)const
   { return con().eli1(iv).con();}
     // Returs a matrix with the columns indexes by iv eliminated from
     // *this

   C_Matrix operator()(F<C,C> const& g)const;
      // returns a matrix obtained by applying function g
      // to each component of *this

   C_Matrix map(F<R,C> const& f)const;
      // returns a matrix obtained by applying function f
      // in the sense of functional operator calculusto
      // to the hermitean part of *this.

   C_Vector operator * (const C_Vector&)const;
      // natural action of a matrix on a vector

   C_Matrix operator * (R const& r)const{ return operator*(C(r));}

   R2 centerOfMass()const;
      // interpreting the matrix components as a mass distribution
      // we return the center of mass as a pair (i,j) of
      // (fractional) pixel coordinates of the center om mass.
      // Since the components of the matrix may be mixed
      // positiv and negativ it is more general to think about
      // a center of electric charge.

   C_Vector linForm()const;
      //: linear form
      // rep[1]&rep[2]&...&rep[n], n=rep.dim(). This allows
      // carrying over from C_Vector many useful descriptors:

   R_Matrix real()const;
      // real part of matrix

   R_Matrix toRealForm(bool test=false)const;
      // returns a real (2*d,2*d) matrix from which *this can
      // be reconstructed. If test==true this reconstruction gets
      // tested.

   static C_Matrix fromRealForm(R_Matrix const& r);
      // Returns a complex matrix reconstructed from its real form r

   C_Vector dig()const;
      //: diagonal
      // Returns the diagonal as a vector. No restrictions
      // concerning dim1() and dim2().

   C inf(Z i=0)const{ return linForm().inf(i);}
      // returns the 'ith-smallest' component. No ordering
      // for i=0. For i>0 it is the component indexed 1+i
      // of the increasingly ordered version of rep

   C sup(Z i=0)const{ return linForm().sup(i);}
      // returns the 'ith-largest' largest component.
      // For i=0 fast operation, no ordering
      // For i>0 it is the component indexed rep.dim()-i
      // of the increasingly ordered version of rep

   R supAbs()const;
      // returns sup{ rep[i].supAbs()} (i.e. the largest
      // absolute value of a component, sometimes called max norm)

   C sum()const{ return linForm().sum();}
      // returns the sum of all components

   C mean()const { return linForm().mean();}
      // arithmetic mean value of all components, fast

   C med()const { return linForm().med();}
      // median value of all components

    Iv infSup()const{ return real().linForm().infSup();}
    // returns the smallest and the largest component. Fast operation,
      // no ordering

   C3 infMeanSup()const{ return linForm().infMeanSup();}
   // returns the smallest component, the arithmetic mean value,
   // and the largest component of *this. Fast operation, no ordering

   C3 infMedSup()const{ return linForm().infMedSup();}
   // returns the smallest component, the median value,
   // and the largest component of *this. Achieved by ordering the
   // components. If one should also need the mean value, don't call
   // the previous function in addition, but simply the cheap function
   // mean()

 // concatenation functions as in C_Vector (with the natural replacement
// C_Vector |--> C_Matrix, R |--> C_Vector)

   C_Matrix& operator &=(const C_Matrix& x1)
   { rep&=x1.rep; return *this;}
   // mutating form of concatenation

   C_Matrix& operator &=(C_Vector r)
   { rep&=r; return *this;}

   C_Matrix append2(C_Matrix const& x1)const;
      // concatenation 'with respect to the second index'

   C_Matrix operator && (C_Matrix const& x1)const
   { return append2(x1);}
      // concatenation 'with respect to the second index'

   C_Matrix operator & (const C_Matrix& x1)const
   { return C_Matrix(rep&x1.rep);}
   // concatenation

   C_Matrix operator & (const C_Vector& r)const
   { return C_Matrix(rep&r);}

   C_Matrix prepend(const C_Vector& r)const
   { return C_Matrix(rep.prepend(r));}

// Connection with functions
   F<R2,C> toF(Iv const& iR=Iv(0,1),Iv const& jR=Iv(0,1) )const;
   //: to F
   // Returns a function defined on the rectangle iR x jR which is
   // an area analog to the discrete lattice of pixel centers.
   // Direction of the continuous axes follows the direction of
   // increasing matrix indexes.

// textfiles of third parties
   bool wrtFor(ostream& str)const;
      // write formatted
   bool readFor_(istream& str);
      // read formatted
   bool wrtForTrn(ostream& str)const;
      // write formatted transposed
   bool readForTrn_(istream& str);
   // read formatted transposed

   bool toGnuPlot(Word const& fileName)const;
   //: to gnu-plot
   // create an outputfile with name 'fileName' and write into it
   // the matrix elements of *this in a manner that the file is suitable
   // as a data input file for making gnuplot graphics with them.
   // Similar to function
   // CpmArrays::toGnuPlot(Word const& fn, V<V<R2> > const& x ).

   bool toGnuPlot(ostream& str)const;
   //: to gnu-plot
   // Works as previous function, except that it writes
   // to a given stream instead of creating a file.

   static void svdcmp(C_Matrix& a, R_Vector& w, C_Matrix& v);
   // singuar value decomposition
   // Implementation tool that now is made public in order to
   // enable a short program that can directly be compared with
   // a Ruby program.

   R opNorm()const;
   //: operator norm
   // The largest of the singular values of *this (source: Mathematica)
   // Notice that abs is the Hilbert Schmidt norm
};

inline bool C_Matrix::prnOn(ostream& out)const
{ return rep.prnOn(out);}

inline bool C_Matrix::scanFrom(istream& in)
{ return rep.scanFrom(in);}

inline Z C_Matrix::com(const C_Matrix& a)const
{ return rep.com(a.rep);}

inline C_Matrix& C_Matrix::operator+=(C_Matrix const& a)
{
   for (Z i=1;i<=dim(); ++i){ rep[i]+=(a.rep)[i]; }
   return *this;
   //cpmcerr<<"C_Matrix::operator|(C_Matrix const& a)const started"<<endl;
  // return (linForm()|a.linForm());
}

inline C_Matrix& C_Matrix::operator *=(const C& r)
{ for (Z i=1;i<=dim(); i++) rep[i]*=r;  return *this;}

inline C_Matrix C_Matrix::operator -()const
{ return C_Matrix(-rep);}

inline C_Matrix C_Matrix::ran(Z i)const
{ return C_Matrix(rep.ran(i));}

inline C_Matrix C_Matrix::test(Z i)const
{ return C_Matrix(rep.test(i));}

inline Word C_Matrix::nameOf()const
{ return Word("C_Matrix");}

inline Word C_Matrix::toWord()const
{ return rep.toWord();}

inline C C_Matrix::operator|(C_Matrix const& a)const
{
   //cpmcerr<<"C_Matrix::operator|(C_Matrix const& a)const started"<<endl;
   return (linForm()|a.linForm());
}

/////////////////////// class SVDC ///////////////////////////////////////
//
// Singular Value Decomposition and functional calculus for complex
// matices. Let a be an arbitrary matrix and f:C->C an arbitrary function.
// Then one defines
// f(a) by applying f to the singular values. This is an extension of the
// functional calculus based on a spectral representation of a matrix.
// In our setting this f(a) is obtained in the following way:
// F<R,R> f=...;
// C_Matrix a=...;
// SVD sa(a);  // does the work
// C_Matrix fa=sa(f);  // relatively easy job.
//
// If we have to vary f for fixed a, we have to do the pre-computations
// encapsulated in the SVD constructor only once. We would not get this
// benefit if we would implement the functional calculus by a member
// function of C_Matrix.

/*
Let A be a complex (m,n)-matrix. Let d=MIN(m,n). A singular value
decomposition of A is a representation
   A=iso1 * diag * (iso2~)
where
   iso1 is an isometric (m,d)-matrix
   diag is a diagonal   (d,d)-matrix
   iso2 is an isometric (n,d)-matrix.
(A (i,j)-matrix m is called isometric if m~ * m = (j,j)-unit matrix,
where m~ denotes the transposed version of m. If i=j, it also satisfies
m * m~ = unit matrix (not trivial !), then it is called orthogonal.)
The diagonal components of diag will be held as a vector w.
The singular value decomposition is not unique. If all singular
values are different (most natural case) the only degree of freedom is
in the order of the singular values and in 'phasefactors' +/-1.
Here the singular values are ordered in descending order of their
absolute value.
*/

class SVDC{
// singular value decomposition of complex (m,n)-matrices

   Z m,n,d;
      // as in previous explanation
   C_Matrix A;
      // as in previous explanation
   C_Matrix U;
      // iso1 of previous explanation
   C_Matrix V;
      // iso2
   C_Matrix Vt;
      // iso2~
   R_Vector w;
      // diag=C_Matrix(w)
public:
   explicit SVDC(const C_Matrix& a);
      // creates a instance of SVD which provides all quantities which
      // result from a singular value decomposition of a. If such an
      // instance is created once, several 'services' like finding f(a)
      // for arbitrary f:R->R can be provided without much computational
      // burden.
   C_Matrix operator()(const F<R,R>& f)const;
      // returns f(A)
// simple access functions
   C_Matrix getU()const{ return U;}
      // isometric leading factor of singular value decomosition
      // into three factors
   C_Matrix getW()const{ return C_Matrix(w);}
      // singular values as a diagonal matrix central factor of singular
      // value decomosition into three factors
   R_Vector get_w()const{ return w;}
      // list of singular values
   C_Matrix getV()const{ return V;}
      // isometric last 'factor' (actually Vt is the factor, Vt has not
      // to be isometric, however)
   C_Matrix getVt()const{ return Vt;}
      // last factor of singular value decomposition
   Z get_m()const{ return m;}
   Z get_n()const{ return n;}
   Z get_d()const{ return d;}
      // number of singular values
   R getError()const;
      // defined such that the result should deviate from 0 only by
      // numerical inaccuracies. Any deviation from the defining
      // properties of a singular value decomposition will be detected.
};

class SpcObj{
// Eigen values and eigenvectors of a Hermitean (n,n)-matrix

   Z n_;
      // 'dimension of the problem'
      // val_.dim() == vec_.dim1() == vec_.dim2() == mat_.dim1()
      // == mat_.dim2();
   C_Matrix mat_;
      // Matrix for which eigenvalues and eigenvectors are requested

   R_Vector val_;
      // The components of this vector are the eigenvalues

   C_Matrix vec_;
      // The rows of this matrix are the eigenvectors
public:

   explicit SpcObj(const C_Matrix& a, bool enfSym=false);
      // creates a instance of SpcObj which provides all quantities which
      // result from a spectral representation of a. If such an
      // instance is created once, several 'services' like finding f(a)
      // for arbitrary f:R->R can be provided without much computational
      // burden. If the last argument is true we work with the Hermitean
      // part (a + ~a)*0.5 instead of a.

   SpcObj(F<C_Vector,C_Vector> const& f, Z n)
   { *this=SpcObj(C_Matrix(f,n),true);}
      // Notice that a Hamiltonian is most naturally defined as a linear
      // function mapping states to states and the states (in 1-dimensional
      // discrete systems) correspond to C_Vectors. In order not to require
      // utmost care in handling the rims of the 'biotope' we enforce
      // 'hermitization' as atask of the present function.

   C_Matrix func(const F<R,R>& f)const;
      // returns f(mat_)
   C_Matrix func(const F<R,C>& f)const;
      // returns f(mat_)
   C_Matrix operator()(const F<R,R>& f)const{ return func(f);}
   C_Matrix operator()(const F<R,C>& f)const{ return func(f);}
// simple access functions
   C_Matrix mat()const{ return mat_;}
   C_Matrix vec()const{ return vec_;}
   R_Vector val()const{ return val_;}

   R operator()(Z i)const{ return val_[i];}
   C_Vector operator[](Z i)const{ return vec_[i];}

   Z dim()const{ return n_;}

   R getError()const;
      // defined such that the result should deviate from 0 only by
      // numerical inaccuracies. Any deviation from the defining
      // properties of a spectral representation will be detected.

   static R_Vector test(Z n, Z jSeed=0);
      // Creates the hermitean part mat of random n x n C_Matrix and calls
      // SpcObj(mat,false).getError(). jSeed is the random seed.
      // The result is a vector which holds three time spans and finally
      // the errorsum obtained by getError().
      // See cpmrmateig.h
      // comments to
      // R_Vector eigen(const R_Matrix& a, R_Matrix& eig, bool test=false,
      // Z iterMax=1000);
      // for a performance measurement. The most time consuming run is
      // documented in ~/e/cpm/quantumscattering/cpmcerr2020_04_19a.txt.
};

template<Z n>
class C_Mat{
// complex square matrices with fixed size
   C_Matrix m_;
public:
   typedef C_Mat<n> Type;
   typedef C ScalarType;
   CPM_IO
   CPM_ORDER
   CPM_TEST
   CPM_DESCRIPTORS
   CPM_CONJUGATION
   CPM_SUM_M
   CPM_PRODUCT_C
   CPM_SCALAR_M
   CPM_DIFFERENCE_PLAIN
   CPM_DIVISION
   CPM_DOT_PRODUCT
   CPM_NORMALIZE

   C_Mat():m_(n){}

   explicit C_Mat(C_Matrix const& a):m_(a){
      cpmassert(a.dim1()==n,"dim1 mismatch");
      cpmassert(a.dim2()==n,"dim2 mismatch");}

   // dyadic product
   C_Mat(C_Vec<n> const& v1, C_Vec<n> const& v2): m_(v1,v2){}

   C_Mat(Z i, Z j):m_(n){ m_[i][j]=C(1_R);}

   C_Mat(std::function<C_Vec<n>(C_Vec<n>)> f);

   C_Mat(V<C> const& v):m_(n){
      for(Z i=1,p=1;i<=n;++i){
         for(Z j=1;j<=n;++j,++p){
            m_[i][j]=v[p];
         }
      }
   }

   C& operator()(Z i,Z j){ return m_[i][j];}
   C const& operator()(Z i,Z j)const{ return m_[i][j];}

   C_Vec<n> operator*(C_Vec<n> const& a)const{
      C_Vector m=m_*a.c();
      return C_Vec<n>(m);
   }

   C_Mat<n> operator*(R const& r)const{
      C_Matrix m=m_*r;
      return C_Mat<n>(m);
   }

   //: tanspose
   C_Mat<n> trn()const{
      C_Matrix m=m_.trn();
      return C_Mat<n>(m);
   }

   //. complex conjugate (no transpose)
   C_Mat<n> cc()const{
      C_Matrix m=m_.cc();
      return C_Mat<n>(m);
   }

   //. hermitian conjugate
   C_Mat<n> hc()const{
      C_Matrix m=m_.con();
      return C_Mat<n>(m);
   }

   //: trace
   C trc()const{
      C sum;
      for (Z i=1;i<=n;++i) sum+=m_[i][i];
      return sum;
   }

   // operator norm
   R opNorm()const{ return m_.opNorm(); }

   V<C> toVC()const{
      V<C> v(n);
      for(Z i=1,p=1;i<=n;++i){
         for(Z j=1;j<=n;++j,++p){
            v[p]=m_[i][j];
         }
      }
      return v;
   }

   static V<C_Mat<n>> bas(){
      Z n2=n*n;
      V<C_Mat<n>> res(n2);
      for (Z i=1,p=1;i<=n;++i){
         for (Z j=1;j<=n; ++j,++p){
            res[p]=C_Mat<n>(i,j);
         }
      }
      return res;
   }

};

template<Z n>
C_Mat<n>::C_Mat(std::function<C_Vec<n>(C_Vec<n>)> f): m_(n){
   auto b=C_Vec<n>::bas();
   for (Z i=1;i<=n;++i){
      for (Z j=1;j<=n; ++j){
         m_[i][j]=(b[i]|f(b[j]));
      }
   }
}

template<Z n> inline C_Mat<n>& C_Mat<n>::operator+=(C_Mat<n> const& a)
   { m_+=a.m_; return *this;}

template<Z n> inline C_Mat<n>& C_Mat<n>::operator-=(C_Mat<n> const& a)
   { m_-=a.m_; return *this;}

template<Z n> inline C_Mat<n> C_Mat<n>::operator-(C_Mat<n> const& a)const
   { C_Matrix m=m_-a.m_; return C_Mat<n>(m);}

template<Z n> inline C_Mat<n> C_Mat<n>::operator-()const
   { C_Matrix m=-m_; return C_Mat<n>(m);}

template<Z n> inline C_Mat<n> C_Mat<n>::operator*(C_Mat<n> const& a)const
   { C_Matrix m=m_*a.m_; return C_Mat<n>(m);}

template<Z n> inline C_Mat<n>& C_Mat<n>::operator*=(C const& a)
   { m_*=a; return *this;}

template<Z n> inline C C_Mat<n>::operator|(C_Mat<n> const& a)const
   {  return (m_|a.m_); }

template<Z n> inline Word C_Mat<n>::nameOf()const
   { Word wi="C_Mat<"; Word nt=cpm(n); return wi&nt&">";}

template<Z n> inline Word C_Mat<n>::toWord()const
   { return m_.toWord();}

template<Z n> inline bool C_Mat<n>::prnOn(std::ostream& os)const
   { return m_.prnOn(os);}

template<Z n> inline bool C_Mat<n>::scanFrom(std::istream& is)
   { return m_.scanFrom(is);}

template<Z n> inline Z C_Mat<n>::com(C_Mat<n> const& a)const
   { return m_.com(a.m_);}

template<Z n> inline C_Mat<n> C_Mat<n>::inv()const
   { C_Matrix m=m_.inv(); return C_Mat<n>(m);}

template<Z n> inline C_Mat<n> C_Mat<n>::net(Z i)const
   { C_Matrix m=m_.net(i); return C_Mat<n>(m);}

template<Z n> inline C_Mat<n> C_Mat<n>::con()const
   { C_Matrix m=m_.con(); return C_Mat<n>(m);}

template<Z n> inline C_Mat<n> C_Mat<n>::test(Z i)const
   {  C z; C zi=z.test(i); C_Mat<n> res;
      for (Z j=1;j<=n;++j){
         for (Z k=1;k<=n;++k){
            res(j,k)=zi;
         }
      }
      return res;}

template<Z n> inline C_Mat<n> C_Mat<n>::ran(Z i)const
   { C_Matrix m=m_.ran(i); return C_Mat<n>(m);}


} // namespace CpmArrays

#endif
