//? cpmdigraph.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_DIGRAPH_H_
#define CPM_DIGRAPH_H_
/*
    Description: directed graphs originally introduced
      for organizing the communication between parallel processes.
      Later also used as base class for the graphics class
      Skeleton and for creating a graphical representation of
      header file hierarchies.

      Recent history: 2012-01-13 second argument in
      void DiGraph<>::setPlnPos_(Word meth="ordered", Z iRandIni=11);
      added.
      2012-01-19
      void DiGraph<>::setPlnPos_(Word meth="shellwise", ...);
      added.

*/
#include <cpms.h>
#include <cpmlinalg.h>

#include <cpmgraph.h>
   // probably graphs without visualization capabilitirs are too poor

namespace CpmGeo{

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::Word;
   using CpmArrays::S;
   using CpmArrays::Outside;
   using CpmArrays::V;
   using CpmArrays::X2;
   using CpmArrays::R_Vector;
   using CpmLinAlg::Z2;
   using CpmLinAlg::R2;
   using CpmFunctions::F;
#ifdef CPM_Fn
   using CpmFunctions::F1;
#endif

   using CpmGraphics::Graph;

   template<class X>
   class DiGraph{ // directed graph
      // directed graph. Has nothing to do with
      // CpmGraphics::Graph

      Z n_{0};
         // number of vertices;

      mutable V<B> act_;
         // an array which holds a boolean variable for each
         // vertex. Will be used for book-keeping during algorithm
         // runs. Also functions declared as const will be
         // allowed to change the values of act_ during run;
         // they will restore finally to correct initial values
         // so that constancy holds effectively.
         // This tool helped to increase speed of function steps
         // by a factor of 100 !

      void refresh()const{act_.set_(B(true));} //for (Z i=1;i<=n_;i++)
      // act_[i]=true;}
         // tool for un-doing the changes in act_

      S<Z> grow(const S<Z>& sv)const;
         // adds to a set sv of vertices all vertices which can be
         // reached by one step along an edge. If the argument sv is not
         // a subset of {1,...n_}, an error will occur during run.
         // Therefore the function is private and will only be used
         // under controlled conditions for implementing function steps.

      S<Z> growRes(const S<Z>& sv)const;
         // returns a a set of active particles which can be reached
         // by a single step from sv. The vertices j returned
         // by the functions get marked as in-active by
         // setting act_[j]=false.

      static Z cutNeg(Z n){ return n>0 ? n : 0;}
         // trivial tool for safely defining the constructor

      mutable Z pc_;
         // processing counter, not needed so far

   protected:

      V<X> x_;
         // Array of objects associated with the vertices
         // x.dim()>=n_ has to be assured, so that for each i
         // with 1<=i<=n_, an x[i] is defined.

      V<C> pp_;
         // planar positions, needed for visualization, added 2007-07

      V< S<Z> > neg_; // neighbors
         // neg_.dim() equals the number of vertices n_
         // neg_[i] is the set of vertices which can be reached via an
         // (directed) edge starting at i (in one step, to be sure). For
         // this to make sense, neg_[i].car() has to be <= neg_.dim() for
         // all admissible i's.
         // Further, S<Z> has to be a subset (not necessarily a proper
         // one) of {1,2,...,n_}

   public:
      typedef DiGraph Type;
      typedef DiGraph CloneBase;
      virtual DiGraph* clone()const{ return new DiGraph(*this);}
      CPM_IO_V
         // IO should be virtual in order to allow
         // additional attributes in derived classes
// all functions test the validity of their arguments and
// stop via cpmerror for unvalid ones.

      DiGraph(V<X> const& obj, Z connect=0):
         n_(obj.dim()),act_(n_,B(true)),x_(obj),pp_(n_),neg_(n_)
         // construction from a list of vertices.
         // Notice that is no problem if two vertices are equal as
         // objects in X. They are different vertices due to their
         // differing index.
         {
            if (connect>0){
               for (Z i=1;i<n_;i++) addEdge(i,i+1);
               if (connect>1) addEdge(n_,1); // was done always till
               // 2003-07-23
            }
         }

         void wrt()const{
            cpmdebug(x_);
            cpmdebug(pp_);
            cpmdebug(neg_);
            cpmdebug(n_);
            cpmdebug(act_);
         }

      DiGraph(Z n=1):n_(cutNeg(n)),act_(n_,B(true)),
         x_(n_),pp_(n_),neg_(n_){}
         // a graph with n vertices and no edges

      DiGraph(Z m, Z n);
         // creates a graph with m*n vertices with the edges
         // which are natural for my particle modeling application

      DiGraph<X>& operator&=(const DiGraph<X>& g);
         // adding DiGraph g to *this in a way that
         // no edges connecting these two objects are added.
         // Also it is irrelevant whether some x[i] equals
         // a g.x[j]. As vetexes of the new graph these
         // are considered different.

      void set_neg(V< S<Z> > const& negNew);
         //: set neighbors
         // setting the neighbors list from outside
         // only matching data from negNew get used.
         // data out of range are ignored

      V< S<Z> > const& get_neg()const{ return neg_;}
         //: get neighbors

      S<Z> const& get_neg(Z i)const{ return neg_.read(i);}
         //: get neighbors
         // if i is not the index of a valid vertex, the return value
         // is the void set

      V<X> const& get_x()const{ return x_;}
         //: get x

      X const& operator[](Z i)const{ return x_[i];}

      void addVertex(const X&);
         //: add vertex
         // adds a new vertex, no change of edges

      void addEdge(Z i, Z j);
         //: add edge
         // makes (i->j) an edge
         // does nothing if i,j are not admissible ,
         // also does nothing for i==j

      void addEdgeSym(Z i, Z j){ addEdge(i,j); addEdge(j,i);}
         //: add edge symetrically
         // makes (i->j) and (j->i) edges

      void remEdge(Z i, Z j);
         //: remove edge
         // 'removes' an edge (nothing is done if no edge is there
         // or i isn't a vertex, no stop in this case)

      void remEdge(Z i, const S<Z>& s);
         //: remove edge
         // remove a set of edges from vertex i

      void remEdgeSym(Z i, Z j){ remEdge(i,j); remEdge(j,i);}
         //: remove edge symmetrically
         // self-explanatory

      DiGraph<X> select(const V<B>& s)const;
         //:: select
         // returns the sub-DiGraph of *this which has as vertices
         // just those x[i] for which s[i]==true. The edges
         // of the new objects are just those
         // x[i]-->x[j] for s[i]==true and s[j]==true
         // j \in neg[i]. An application of this operation
         // is to select the part of *this that is within the
         // frame of some visualization tool.

      S<Z2> edges(Z i)const;
         //: edges
         // returns the set of edges starting at i in a natural
         // representation as pairs of integers

      S<Z2> edges()const;
         //:: edges
         // returns the set of all edges of *this in a natural
         // representation as a set of Z2's

      Z steps(Z i, Z j)const;
         //:: steps
         // returns the length (as number of steps) of the shortest path
         // from i to j. If there is no such path we return -1

      V< S<Z> > shells(Z i)const;
         //:: shells
         // retuns a vector res such that res[1] is the set of all
         // vertices which can be reached from i by one step
         // res[2] that can be reached by 2 steps and not by 1 step
         // res[3] those that can be reached by 3 steps but not by less
         // ...

      Z dim()const{ return n_;}
         //: dimension

      Z car()const{ return n_;}
         //: cardinality

// functions added 2007-07
      void setPlnPos_(Word meth="ordered", Z iRandIni=11);
         //: set planar positions
         // Tool for visualising the dependency graph of cpm header files.
         // "random" ,"ordered", and "shellwise" are the present options.
         // Only the first option is general in scope whereas the two
         // following ones contain references to the cpm header files.
         // For all options some randomization of positions is done.
         // So all cases the second argument which controls randomization
         // may useful in avoiding unhappy coincidences in graphics
         // created by function mark.
      void setPlnPos_(V<C> const& vc)
      { if (pp_.dim()==vc.dim()) pp_=vc;}
         //: set planar positions
         // from a list of positions
      void mark(Graph& g, bool autoScale=true)const;
         //: mark
         // If the second argument is true, the size of the
         // objectspace of g is adjusted to the size of
         // *this in planar position space (= complex plane).
      void rev_();
         //: revert
         // changes each edge i->j into an edge j->i
      void removeAllEdges_(){ neg_=V< S<Z> >(n_,S<Z>());}
         //:: remove all edges
      X2< S<Z>, S<Z> > sorSink()const;
         //: source sink
         // Let res be the return value.
         // Then res.c1()
         // is the set of all indexes of vertices which are
         // source points (i.e. have out-going edges only)
         // And res.c2()
         // is the set of all indexes of vertices which are
         // sink points (i.e. have in-coming edges only).
      Word toWord(Z i)const{ return x_(i).toWord();}
         //: to word
         // returns a Word characterizing x_[i]. Valid for all i
         //
      bool wrtFor(ostream& str , bool writeComments)const;
         //: write formatted
   };

   bool close(Z2 a, Z2 b);
   bool rimRelated(Z2 a, Z2 b, Z rim);
      // defined in cpmdigraph.cpp
   S<Z> shiftNegFunc(S<Z> const& s, Z const& n);
      // means for implementing operator &=

template<class X>
DiGraph<X>::DiGraph(Z m, Z n):
   n_(cutNeg(m*n)),neg_(n_),act_(n_,B(true)),x_(n_),pp_(n_)
{
   if (m<1 || n<1) cpmerror("DiGraph(Z,Z): unvalid arguments");
   V<Z2> ij(n_);
   for (Z k=0;k<n_;k++){
      Z i=k/n;
      Z j=k%n;
      ij[k+1]=Z2(i,j);
      x_[k+1]=R2(i,j);
   }
   for (Z k1=1;k1<=n_;k1++){
      for (Z k2=1;k2<=n_;k2++){
         if (close(ij[k1],ij[k2])||rimRelated(ij[k1],ij[k2],n))
            addEdge(k1,k2);
      }
   }
}

template<class X>
DiGraph<X>& DiGraph<X>::operator&=(const DiGraph<X>& g)
{
#ifdef CPM_Fn
   F< S<Z>, S<Z> > tr=F1<S<Z>,Z,S<Z> >(n_)(shiftNegFunc);
#else
   F< S<Z>, S<Z> > tr(bind(shiftNegFunc,_1,n_));
#endif

   V< S<Z> > gn=g.neg_(tr);
   neg_&=gn;
   x_&=g.x_;
   pp_&=g.pp_;
   n_+=g.n_;
   return *this;
}

template <class X>
void DiGraph<X>::set_neg(V< S<Z> > const& negNew)
{
   const Z mL=5;
   cpmmessage(mL,"DiGraph<X>::set_neg(V< S<Z> >) started");
   Z d1=n_;
   Z d2=negNew.dim();
   Z d=cpminf<Z>(d1,d2);
    for (Z i=1;i<=d;i++){
      S<Z> negni=negNew[i];
      if (negni.car()>0){
//         while (negni.last()>d1){
//            negni.remove(negni.last());
//         }
         S<Z> negniMem(negni);
         negni=S<Z>();
         for (auto const& x : negniMem){
            if (x<=d1) negni<<x;
         } // negni holds only those elements of negni which are not
         // greater than d1
      }
      neg_[i]=negni;
   }
   cpmmessage(mL,"DiGraph<X>::set_neg(V< S<Z> >) done");
}

template <class X>
void DiGraph<X>::addVertex(const X& v)
{
   x_&=v;
   n_+=1;
   B actAdd=B(true);
   act_&=actAdd;
   S<Z> negAdd;
   neg_&=negAdd;
}

template<class X>
void DiGraph<X>::addEdge(Z i, Z j)
{
   if(i<1) return;
   else if (j<1) return;
   else if (i> n_) return;
   else if (j> n_ ) return;
   else if (i==j) return;
   else neg_[i].add_(j);
}

template<class X>
S<Z> DiGraph<X>::grow(const S<Z>& sv)const
{
   S<Z> res(sv);
   for (Z i=1;i<=sv.car();i++){
      Z vi=sv[i];
      Z dvi=neg_[vi].car();
      for (Z j=1;j<=dvi;j++){
         res.add_(neg_[vi][j]);
      }
   }
   return res;
}

template<class X>
S<Z> DiGraph<X>::growRes(const S<Z>& sv)const
{
   S<Z> res;
   for (Z i=1;i<=sv.car();i++){
      pc_+=1;
      Z vi=sv[i];
      Z dvi=neg_[vi].car();
      for (Z j=1;j<=dvi;j++){
         Z vj=neg_[vi][j];
         if (act_[vj]){
            res.add_(vj);
            act_[vj]=B(false);
         }
      }
   }
   return res;
}

template<class X>
void DiGraph<X>::remEdge(Z i, Z j)
{
   if(i<1 || j<1 || i> n_ || j> n_) return;
      // nothing to do
   S<Z> remj;
   remj.add_(j); // direct construction interferes with the protected
      // constructor S<T>(Z) which here causes trouble due to T=Z
   neg_[i]=neg_[i]-remj;
}

template<class X>
void DiGraph<X>::remEdge(Z i, const S<Z>& s)
{
   if(i<1 || i> n_) return;
      // nothing to do
   neg_[i]=neg_[i]-s;
}

template<class X>
DiGraph<X> DiGraph<X>::select(const V<B>& s)const
{
   X2<V<X>, V<Z>> pvv=x_.select(s);
   V<X> xNew=pvv.c1();
   V<Z> ind=pvv.c2();
   DiGraph<X> res(xNew);
   Z n=res.dim(),i;
   for (i=1;i<=n;i++){
      S<Z> nii=neg_[ind[i]];
      nii.select(s);
      res.neg_[i]=nii;
   }
   return res;
}

template<class X>
S<Z2> DiGraph<X>::edges(Z i)const
{
   if(i<1 || i> n_)
      cpmerror("DiGraph::::edges(Z)const: unvalid argument");
   S<Z2> res;
   S<Z> v(neg_[i]);
   for (auto const& x : v){
      res.add_(Z2(i,x));
   }
//   for (Z j=1;j<=v.car();j++){
//      res.add_(Z2(i,v[j]));
//   }
   return res;
}

template<class X>
S<Z2> DiGraph<X>::edges()const
{
   S<Z2> res;
   for (Z i=1;i<=n_;i++) res=(res|edges(i));
   return res;
}

template<class X>
V< S<Z> > DiGraph<X>::shells(Z i)const
{
   if(i<1 || i> n_)
      cpmerror("DiGraph::::shells(Z)const: unvalid argument");
   S<Z> s=neg_[i];
   V<S<Z> > res(1);
   res[1]=s;
Label:
   s=growRes(s);
   if (s.car()==0){
      refresh();
      return res;
   }
   else{
      res&=s;
      goto Label;
   }
}

template<class X>
Z DiGraph<X>::steps(Z i, Z j)const
{
   if(i<1 || i> n_ || j<1 || j> n_)
      cpmerror("DiGraph::steps(Z,Z)const: unvalid argument");
   S<Z> s=neg_[i];
   Z count=1;
Label:
   if (s.hasElm(j)){
      refresh(); // this compensates our
         // breaking strict constancy
      return count;
   }
   s=growRes(s);
   if (s.car()==0){
      refresh();
      return -1;
   }
   else{
      count++;
      goto Label;
   }
}

template<class X>
bool DiGraph<X>::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("DiGraph",str);
   cpmp(n_);
   cpmp(neg_);
   cpmp(x_);
   return true;
}

template<class X>
bool DiGraph<X>::scanFrom(istream& str)
{
   cpms(n_);
   cpms(neg_);
   cpms(x_);
   return true;
}

template<class X>
void DiGraph<X>::rev_()
{
   S<Z2> ed=edges();
   removeAllEdges_();
 //  Z n=ed.car();
   for (auto const& x : ed){
 //  for (Z i=1;i<=n;++i){
   //   Z2 z2=ed[i];
      Z2 z2=x;
      addEdge(z2[2],z2[1]);
   }
}

template<class X>
void DiGraph<X>::setPlnPos_(Word meth, Z ir)
{
   Z i,j;
   if (meth=="random"){
      R eps=0.05;
      Iv ivx(eps,R(1)-eps);
      Iv ivy=ivx;
      for (i=1;i<=n_;++i){
         R xi=ivx.ran(ir++);
         R yi=ivy.ran(ir++);
         pp_[i]=C(xi,yi);
      }
   }
   else if (meth=="ordered"){
      Z i0=1;
      Word w0("cpmdefinitions.h");
      for (i=1;i<=n_;++i){
         if (x_[i]==w0){
            i0=i;
            break;
         }
      }
      pp_[i0]=C(0,0.5);
      R fracAmpl=0.25;
      V< S<Z> > vsz=shells(i0);
      Z nSh=vsz.dim();
      Iv iv(0,1);
      V<R> vr=iv.div(nSh+1);
      R xAmpl=fracAmpl/nSh;
      for (i=1;i<=nSh;++i){
         S<Z> si=vsz[i];
         Z sic=si.car();
         R xi=vr[i+1];
         V<R> ivd=iv.div(sic);
         R yAmpl=fracAmpl/sic;
         yAmpl=CpmRootX::inf<R>(yAmpl,xAmpl);
         for (j=1;j<=sic;++j){
            Z nij=si[j];
            pp_[nij]=C(xi+Iv(xAmpl).ran(ir++),
               ivd[j]+Iv(yAmpl).ran(ir++));
         }
      }
      C z0(0.5,0.5);
      R fac=0.8;
      Iv ivx(0.03,-0.03);
      Iv ivy(0.02,-0.02);
      for (j=1;j<=n_;++j){
         pp_[j]=z0+(pp_[j]-z0)*fac+C(ivx.ran(ir++),ivy.ran(ir++));
      }
   }
   else if (meth=="shellwise"){
      Word w0("cpmdefinitions.h");
      pp_.set_(C(-0.7,-0.7));
      Z i0=x_.locAsc(w0);
      pp_[i0]=C(0,0);
      V< S<Z> > vsz=shells(i0);
      Z nSh=vsz.dim();
      V<R> rSh=Iv(0.,1.).cen(nSh);
      Iv ar(0.,cpm2pi); // ar: angle range
      for (i=1;i<=nSh;++i){
         S<Z> si=vsz[i];
         Z sic=si.car();
         Iv ivi=Iv(-1.,1.);
         ivi*=cpminv(R(sic));
         R ri=rSh[i];
         V<R> phi=ar.cen(sic);
         for (j=1;j<=sic;++j){
            Z nij=si[j];
            pp_[nij]=ri*C::expi(phi[j]+ivi.ran(ir++));
         }
      }
   }
   else ;
}

namespace{
   Word transForm(Word const& fn)
      // ad hoc for processing cpm file names
   {
      return fn.cut(2).cut(-3);
   }
}

template<class X>
void DiGraph<X>::mark(Graph& g, bool autoScale)const
{
   using namespace CpmGraphics;
   using namespace CpmImaging;
   Z i,j;
   if (autoScale)g.setXY(pp_,1.1,1);
   g.setThcLin_(false);
   for (i=1;i<=n_;++i){
      S<Z2> edgi=edges(i);
      Z ci=edgi.car();
      for (j=1;j<=ci;++j){
         Z2 a=edgi[j];
         C c1=pp_[a[1]];
         C c2=pp_[a[2]];
         C cm=(c1+c2)*0.5;
         g.draw(c1,cm,RED);
         g.draw(cm,c2,GREEN);
      }
   }
   R h=R(g.textHeightAct())*g.dy();
   R rCircle=h*0.5;
   for (i=1;i<=n_;++i) g.fillCircle(pp_[i],rCircle,RED);
   R eps=h;
   cpmdebug(eps);
   for (i=1;i<=n_;++i){
      C ci=pp_[i];
      Word xi=transForm(x_[i]);
         // ad hoc for processing cpm file names
      R2 xFyF=g.get_xFyF(R2(ci[1]+eps,ci[2]));
      g.addText(xi,xFyF,BLACK,0);
   }
}

//template<class X>
//X2< S<Z>, S<Z> > DiGraph<X>::sorSink()const
//{
//  // S<Z2> ed=edges();
//   V<Z2> ed=edges().toV();
//   V<B> isSink(n_,B(true));
//   V<B> isSource(n_,B(true));
//   Z i;
//   for (i=ed.b();i<=ed.e();++i){
//      Z2 z2=ed[i];
//      Z i=z2[1];
//      Z j=z2[2];
//      // now we have an edge (i-->j)
//      // if isSink should finally contain all indexes from which
//      // no edge arises, it has to be disqualified now for index i:
//      // (since there is a edge arising from i)
//      isSink[i]=B(false);
//      // if isSource should finally contain all indexes to which
//      // no edge enters, it has to be disqualified now for index j:
//      // (since there is an edge entering j)
//      isSource[j]=B(false);
//   }
//   V<Z> list(n_);
//   for (i=1;i<=n_;++i) list[i]=i;
//   return X2< S<Z>, S<Z> >(S<Z>(list,isSource),S<Z>(list,isSink));
//}

template<class X>
X2< S<Z>, S<Z> > DiGraph<X>::sorSink()const
{
   S<Z2> ed=edges();
   S<Z> sources;
   S<Z> sinks;

//   V<B> isSink(n_,B(true));
//   V<B> isSource(n_,B(true));
//   Z i;
   for (auto const& x : ed){
//   for (i=ed.b();i<=ed.e();++i){
      Z2 z2=x;
      Z i=z2[1];
      Z j=z2[2];
      sources<<i;
      sinks<<j;
      // now we have an edge (i-->j)
      // if isSink should finally contain all indexes from which
      // no edge arises, it has to be disqualified now for index i:
      // (since there is a edge arising from i)
 //     isSink[i]=B(false);
      // if isSource should finally contain all indexes to which
      // no edge enters, it has to be disqualified now for index j:
      // (since there is an edge entering j)
//      isSource[j]=B(false);
   }
 //  V<Z> list(n_);
 //  for (i=1;i<=n_;++i) list[i]=i;
  // return X2< S<Z>, S<Z> >(S<Z>(list,isSource),S<Z>(list,isSink));
  return X2< S<Z>, S<Z> >(sources,sinks);
}

template<class X>
bool DiGraph<X>::wrtFor(ostream& str, bool writeComments)const
{
   X2< S<Z>, S<Z> > x2ss=sorSink();
   S<Z> source=x2ss.c1();
   S<Z> sink=x2ss.c2();
   Z j,k;
   Word tab0="";
   Word incr="  ";
   for (auto i=source.begin();i!=source.end();++i){
    //  Z vi=source[i];
      Z vi=*i;
      Word tab=tab0;
      str<<tab.str()<<toWord(vi).str()<<endl;
      S<Z> vsi=get_neg(vi);
      tab&=incr;
      for (auto j=vsi.begin();j!=vsi.end();++j){
         Z sj=*j;
         str<<tab.str()<<toWord(sj).str()<<endl;
         S<Z> vsj=get_neg(sj);
         tab&=incr;
         for (auto k=vsj.begin();k!=vsj.end();++k){
            Z sk=*k;
            str<<tab.str()<<toWord(sk).str()<<endl;
         }
      }
   }
   return true;
}

} // namespace

#endif
