//? cpmdistribution.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_DISTRIBUTION_H_
#define CPM_DISTRIBUTION_H_

/*
   Description: This class Distribution describes data that
      can be binned together and form a histogram. These are the
      discrete version of probability measures
*/

#include <cpminterfaces.h>
#include <cpmtypes.h>
#include <cpminterval.h>
#include <cpmv.h>
#include <cpmf.h>
#include <cpmlinalg.h>
#include <cpmgraph.h>
#include <vector>

namespace CpmArrays{

   using CpmRoot::R;

template <class X>
class Distribution{
   // general framework for distributions ( = histograms)
   // My present workhorse is DisR which is derived from Distribution<R>

protected: // efficient formation of derived classes is important here

   Z n{0};
      // number of bins

   V<Z> rep;
      // rep[i] is the number of X-instances which belong to the bin
      // indexed i (1<=i<=n)

   CpmFunctions::F<X,Z> f;
      // classifying function such that f(x) is the index of the bin
      // to which x belongs. Values 0 and n+1 indicate 'out of range'.
      // the second argument of function put decides whether
      // out-of-range x's influence rep or not.

public:
   Distribution(){}
      // notice that n, rep, and f get initialized anyway,
      // since there is a default constructor for their types
   Distribution(Z n_):
      n(n_),rep(n,0){}

   Distribution(Z n_, const CpmFunctions::F<X,Z>& f_):
      n(n_),rep(n,0),f(f_){}

   void put(const X& x, bool all=true);
      // putting x to the distribution. If all==true all x 'get
      // registered' even those out of range. If all==false
      // only those x for which f(x)>=1 && f(x)<=n count.

   void put(const std::vector<X>& v, bool all=true)
      { for (Z i=0;i<(Z)v.size();i++) put(v[i],all);}
      // adding the components of v to the distribution
      // we use std::vector here since many sources of vectors
      // are such that the length is not known in advance. Here
      // std::vector is efficient because of it's efficient
      // push_back function

   void put(const V<X>& v, bool all=true)
      { for (Z i=1;i<=v.dim();i++) put(v[i],all);}
      // adding the components of v to the distribution

   virtual void show(CpmGraphics::Graph& g, bool autoX=false)const;
      // graphical representation of i |--> rep[i].card().
      // g: Screen area in whitch the histogram is shown
      // argument autoX has no effect here (will become
      // meaningful in DisR)

   Z sum()const{ Z s=0; for(Z i=1;i<=n;i++) s+=rep[i]; return s;}
      // sum of the bins

   Z singular()const;
      // Let k := card{ i | i \in {1,...n} and rep[i]>0}
      // We return k if k<2 and -1 else
      // Thus the return value 1 says that the distribution is 'singular'
      // in the sense of being concentrated to a single point.
      // This is an exceptional situation in cases where one wants to
      // extract a parameter from fitting the distribution to a
      // certain family of curves (see function DisR::tBoltzman).
      // The computation is very fast for normal distributions
      // which have not nearly all bins zero.
};

template <class X>
void Distribution<X>::show(CpmGraphics::Graph& g, bool autoX)const
{
   autoX; // not used
   R_Vector x(n);
   for (Z i=1;i<=n;i++) x[i]=rep[i];
   g.show(x);
}

template <class X>
void Distribution<X>::put(const X& x, bool all)
{
   Z i=f(x);
   if (all){
      if (i<=1) rep[1]++;
      else if (i>=n) rep[n]++;
      else rep[i]++;
   }
   else{
      if (i>=1 && i<=n) rep[i]++;
   }
}

template <class X>
Z Distribution<X>::singular()const
{
   Z count=0;
   for (Z i=1;i<=n;i++){
      Z val=rep[i];
      if (val>0) count++; // val>=0 anyway
      if (count>1) return -1; // then there are two points
   }
   // if we come here there, count==0 or count==1
   return count;
}

//////////////////////// class DisR ////////////////////////////////////

class DisR: public Distribution<R>{ // distribution of reals
   CpmGeo::Iv range;
      // The interval, the subdivision of which defines
      // binning of R values
   void adjust();
      // sets the dependent data according to
      // the independent ones. In our case range and n are
      // independent and f is dependent
public:
   typedef DisR Type;
   typedef Distribution<R> Base;

   static Z verbose;
      // enables visualization of the success of
      // tBoltzmann (presently for verbose>0)

   DisR():Base(){}

   DisR(CpmGeo::Iv iv, Z n_):Base(n_),range(iv){adjust();}

   DisR(R rMin, R rMax, Z n_):Base(n_),range(rMin,rMax){adjust();}

explicit DisR(const std::vector<R>& v, Z n=16, R fac1=1, R fac2=1,
      bool all=true);
      // here range is determined by the smallest and the largest
      // component of v, vMin, vMax as the interval
      // [fac1*vMin, fac2*vMax].

explicit DisR(const std::vector<Z>& v, Z n=16, R fac1=1, R fac2=1,
      bool all=true);
      // here range is determined by the smallest and the largest
      // component of v, vMin, vMax as the interval
      // [fac1*vMin, fac2*vMax].

explicit DisR(const V<R>& v, Z n=16, R fac1=1, R fac2=1,
      bool all=true);
      // here range is determined by the smallest and the largest
      // component of v, vMin, vMax as the interval
      // [fac1*vMin, fac2*vMax].

   CPM_IO
   virtual DisR* clone()const{ return new DisR(*this);}

   void show(CpmGraphics::Graph& g, bool autoX=false)const;
      // representation in a bar diagram, where the range of the axes is
      // given by g.getIvX() and g.getIvY().

   CpmLinAlg::R2 tBoltzmann(R tShow=0, Word title="")const;
      // Extracting a 'temperature' out of the distribution.
      // Let res be the return value. Then res.x1 is this
      // 'temperature' and res.x2 is a misfit indicator.
      // It is >=0 where 0 indicates perfect fit. Values
      // arround 1 say essentially that the fit is
      // of no value, values >1 indicate that the distribution
      // has no similarity with one that would arrise from
      // a Boltzmann distribution belonging to some (even fuzzy)
      // value of a temperature.
      // More precisely: We try to find c and xm such that
      // sum over i of (res[i]-c*exp(-x[i]/xm))^2, x:=range.centers(n)
      // is as small as possible.
      // For each xm, we determine the Gauss least square solution
      // c(xm) for c, and then we minimize
      // error(xm):=sum over i of (res[i]-c(xm)*exp(-x[i]/xm))^2
      // to get the value res.x1=xm.
      // res.x2 is sqrt(error(xm)/(sum over i of (res[i])^2))
      // which is a dimension-less misfit measure.
};

// Since the introduction of namespaces in C++ it is much more
// adequate than before to define functions which do not belong to
// a class. In the following function, the concept of a distribution
// is only an auxiliar one and so it looks not natural to make it
// a member function (static or not) of DisR.

CpmLinAlg::R2 disBoltzmann(std::vector<R>& v, R frac=0.9, Z nBins=32);
   // similar to DisR::tBoltzmann() but the distribution is formed out
   // of v such that a fraction (1-frac) of the largest v-values is
   // omitted (to achieve this, v gets ordered internally with
   // using std::sort, so v gets changed by call, so it is not of
   // type const &) in forming the distribution which finally gets
   // analyzed.

} // namespace

#endif
