//? cpmfa.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_FA_H_
#define CPM_FA_H_
//////////////////////////////////////////////////////////////////////////
//
// Purpose: Define classes describing function-like objects with
//   arithmetic operations
//
//////////////////////////////// class FUOBJ<> ///////////////////////////
#include <cpmfo.h>
#include <cpmtypes.h>

namespace CpmFunctions{

  // using std::ostream;
  // using std::istream;
   using namespace CpmStd;
   using CpmRoot::Z;

   namespace auxiliary { // is within a Cpm... namespace !

///////////////////// tools for defining class Fa<X,Y> ///////////////////

// Advanced arithmetics of functions based on handles to FncObjs

// Arithmetic operations: Here it is assumed that Y is a ring i.e. that
// Y+Y, Y-Y, Y*Y, -Y are defined. Also division Y/Y is assumed to be
// defined. Also template functions/operators will be implemented which
// do not depend only on X and Y.

// class SUMFUO<>

template <class X, class Y>
class SUMFUO: public BINOPFUO<X,Y>{
   typedef SUMFUO<X,Y> Type;
   CPM_INVAR(SUMFUO)
public:
   Y operator()(const X& x)const{ return this->f1(x)+this->f2(x);}
   SUMFUO( const F<X,Y> & f1In, const F<X,Y>& f2In)
   :BINOPFUO<X,Y>(f1In,f2In){}
};

// class DIFFFUO<>

template <class X, class Y>
class DIFFFUO: public BINOPFUO<X,Y>{

public:
   Y operator()(const X& x)const{ return this->f1(x)-this->f2(x);}
   DIFFFUO( const F<X,Y> & f1In, const F<X,Y> & f2In)
   :BINOPFUO<X,Y>(f1In,f2In){}
};

// class NEGFUO<>

template <class X, class Y>
class NEGFUO: public MONOPFUO<X,Y>{

public:
   Y operator()(X const& x)const{ return -this->f1(x);}
   NEGFUO(const F<X,Y> & f1In):MONOPFUO<X,Y>(f1In){}
};

// class PRODFUO<>

template <class X, class Y>
class PRODFUO: public BINOPFUO<X,Y>{

public:
   Y operator()(const X& x)const{ return this->f1(x)*this->f2(x);}
   PRODFUO( const F<X,Y> & f1In, const F<X,Y> & f2In)
   :BINOPFUO<X,Y>(f1In,f2In){}
};

// class DIVIFUO<>

template <class X, class Y>
class DIVIFUO: public BINOPFUO<X,Y>{

public:
   Y operator()(const X& x)const{ return  this->f1(x)/this->f2(x);}
   DIVIFUO( const F<X,Y> & f1In, const F<X,Y> & f2In)
   :BINOPFUO<X,Y>(f1In,f2In){}
};

template <class X, class Y>
class INVFUO : public FncObj<X,Y>{
   const F<X,Y> f;
public:
   INVFUO(const F<X,Y>& f_):f(f_){}
   Y operator()(const X& x)const
   { return CpmRoot::invT<Y>(f(x));}
};

// treating operations which need more than 2 template arguments

template <class X1, class X2, class X3>
class TENSFUO: public FncObj< cpmX2<X1,X2>, X3 >{
   // TENS stands for Tensor Product

   const F<X1,X3>  f1;
   const F<X2,X3>  f2;

public:
   X3 operator()(const cpmX2<X1,X2>& x)const
      { return f1(x.get1())*f2(x.get2());}
   TENSFUO( const F<X1,X3> & f1In, const F<X2,X3> & f2In):
      f1(f1In),f2(f2In){}
};

template <class X1, class X2, class X12, class Y1, class Y2, class Y12>
class GENPRODFUO: public FncObj< X12,Y12 >{
   // GENPROD stands for general product. We assume that X12 is a class
   // which defines functions get1() and get2() such that for each x12 in
   // X12 the values of x12.get1() and x12.get2() allow unique automatic
   // conversion to X1 and X2 respectively. Further, we assume that for
   // each y1 in Y1 and y2 in Y2 the value of Y12(y1,y2) is an element in
   // Y12. Notice that the normal tensor product
   // TENSFUO<X1,X2,X3>(f1,f2)
   // would be obtained as
   // GENPRODFUO<X1,X2,cpmX2<X1,X2>,Y3,Y3,Triv<Y3> > (f1,f2)
   // where
   //    template <class X>
   //    class Triv: public X{
   //    public:
   //       Triv(const X& x):X(x){}
   //       Triv(const X& x1, const X& x2):X(x1*x2){}
   //   };
   //

   const F<X1,Y1>  f1;
   const F<X2,Y2>  f2;

public:
   Y12 operator()(const X12& x)const{ return Y12(f1(x.c1()),f2(x.c2()));}
   GENPRODFUO( const F<X1,Y1> & f1In, const F<X2,Y2>& f2In):
      f1(f1In),f2(f2In){}
};

template <class X, class Y, class S>
   // assumption S*Y defined, in typical situations, Y is a linear space
   // over S

class MULTFUO: public FncObj<X,Y>{

   const F<X,Y>  f1;
   const S s;

public:

   Y operator()(const X& x)const{ return s*(f1(x));}

   MULTFUO(const S& sIn, const F<X,Y> & f1In):f1(f1In),s(sIn){}

};

template <class X, class Y, class S>
   // assumption Y*S defined, in typical situations, Y is a linear space
   // over S

class MULTFUOR: public FncObj<X,Y>{

   const F<X,Y>  f1;
   const S s;

public:

   Y operator()(const X& x)const{ return (f1(x))*s;}

   MULTFUOR(const S& sIn, const F<X,Y> & f1In):f1(f1In),s(sIn){}

};

template <class X, class Y>
   // assumption Y+Y defined

class Y_SUMFUO: public FncObj<X,Y>{

   const F<X,Y>  f1;
   const Y y;

public:

   Y operator()(const X& x)const{ return (f1(x))+y;}

   Y_SUMFUO(const F<X,Y> & f1In, const Y& yIn):f1(f1In),y(yIn){}

};

template <class X, class Y>
   // assumption Y-Y defined

class Y_DIFFFUO: public FncObj<X,Y>{

   const F<X,Y> f1;
   const Y y;

public:

   Y operator()(const X& x)const{ return f1(x)-y;}

   Y_DIFFFUO(const F<X,Y> & f1In, const Y& yIn):f1(f1In),y(yIn){}

};

template <class X, class Y>
   // assumption Y*Y defined

class Y_PRODFUO: public FncObj<X,Y>{

   const F<X,Y>  f1;
   const Y y;

public:

   Y operator()(const X& x)const{ return f1(x)*y;}

   Y_PRODFUO(const F<X,Y> & f1In, const Y& yIn):f1(f1In),y(yIn){}

};

template <class X, class Y>

class Y_DIVIFUO: public FncObj<X,Y>{

protected:

   const F<X,Y>  f1;
   const Y y;

public:

   Y operator()(const X& x)const{ return f1(x)/y;}

   Y_DIVIFUO(const F<X,Y> & f1In,const Y& yIn):f1(f1In),y(yIn){}

};

} // auxiliary

/////////////////////////////////// class Fa<X,Y> ////////////////////////

// Arithmetics and some additional functionality for Fo<X,Y>

// Now arithmetics is introduced. Just as in class Va, we assume that
// -Y, Y+Y, Y-Y, Y*Y, Y/Y , Y < Y, Y > Y, are defined
// Note 02-05-24:
// Notice that e.g. addition of functions is not implemented by really
// 'adding something together' but arranging pointers to the code
// of all functions in the sum in a manner that when the sum function is
// asked to be evaluated for some value of the argument, all functions
// needed in the sum are ready for evaluation and addition of the
// output-values. This is sometimes called 'deferred evaluation'.
// I found that code of type:
// Fa<X,Y> f;
// Z i,n=3300;
// V< Fa<X,Y> > vf(n);
// for (i=1;i<=n;i++) vf[i]=...;
// for (i=1;i<=n;i++) f+=vf[i];
// caused stack overflow in the last line

template <class X, class Y>
class Fa: public Fo<X,Y> { // version of Fo with arithmetics operations

public:

   typedef Fa<X,Y> Type;
   typedef Y ScalarType;

   Fa(void):Fo<X,Y>(){}
      // default constructor

   Fa(const F<X,Y>& g):Fo<X,Y>(g){}
      // downcast constructor

   Fa(const Fo<X,Y>& g):Fo<X,Y>(g){}
      // downcast constructor

   Fa(const Fa<X,Y>& g):Fo<X,Y>(g){}
      // copy constructor

   Fa(FncObj<X,Y>* fop):Fo<X,Y>(fop){}
      // constructor from pointers to FncObj

// Construction from function pointers

   Fa( Y (*f)(const X& )):Fo<X,Y>(f){}
   Fa( Y (*f)(X)):Fo<X,Y>(f){}

   Fa( std::function<Y(X)> f):Fo<X,Y>(f){}

// Construtor for constant function

   explicit Fa(const Y& y1):Fo<X,Y>(y1){}

   virtual F<X,Y>* clone()const{ return new Fa(*this);}

   virtual Word nameOf()const
   {
      Word wi="Fa<";
      Word wx=CpmRoot::Name<X>()(X());
      Word wy=CpmRoot::Name<Y>()(Y());
      return wi&wx&","&wy&">";
   }

   CPM_SUM_C
   CPM_PRODUCT_C
   CPM_DIFFERENCE
   CPM_DIVISION
   CPM_SCALAR_C
   CPM_IO_V

   void scl_(X const& fac, X const& x0=X());
      //: scale
      // changes the function f into x|-->f(x0+(x-x0)*fac)
      // assumes that X defines addition, subtraction, and multiplication,
      // typically that X is a ring. Similar to function stretch_ in
      // class R_Func. For X=R this operation applied with fac>1
      // lets features of function *this (such as a peak width) shrink and
      // fac<1 lets them grow.
};

namespace{// anonymus
   template <class X, class Y>
   Y scaleFunc(X const& x, X const& x0, X const& fac, F<X,Y> const& f)
   { return f(x0+(x-x0)*fac);}
}// anonymus

template <class X, class Y>
void Fa<X,Y>::scl_(X const& fac, X const& x0)
// elegant implementation, avoids using operator new directly.
{
#ifdef CPM_Fn
   *this=F3<X,X,X,F<X,Y>,Y>(x0,fac,*this)(scaleFunc);
#else
   *this=F<X,Y>(bind(scaleFunc<X,Y>,_1,x0,fac,*this));
#endif
}

// functions(operators) outside this class. Here additional template
// arguments are allowed, which creates the flexibility needed for
// treating composition and tensor product.

template <class X1, class X2, class X3>

Fa<cpmX2<X1,X2>,X3> operator || (const Fa<X1,X3>& f13,
         const Fa<X2,X3>& f23)
   // Tensor product of functions: Assumption is that X3*X3 is defined.
   // This creates functions of two variables out of two functions of one
   // variable. Symbol || should remind to the old symbol )( for the
   // 'dyadic product' which is essentially the same construction
{
   using namespace auxiliary;
   return Fa<cpmX2<X1,X2>,X3>(new TENSFUO<X1,X2,X3>(f13,f23));
}

template <class X, class Y, class S>
   // assumption S*Y defined, in typical situations,
   // Y is a linear space over S

Fa<X,Y> mult(const S& sIn, const Fa<X,Y>& fIn)
   // Function values are multiplied by s, operator * instead of
   // mult results in compiler errors caused by mistaking this as
   // another operator *
{
   using namespace auxiliary;
   return Fa<X,Y>(new MULTFUO<X,Y,S>(sIn,fIn));
}

template <class X, class Y, class S>
   // assumption Y*S defined, in typical situations,
   // Y is a linear space over S
Fa<X,Y> multR(const S& sIn, const Fa<X,Y>& fIn)
   // Function values are multiplied by s, operator * instead of
   // mult results in compiler errors caused by mistaking this as
   // another operator *
{
   using namespace auxiliary;
   return Fa<X,Y>(new MULTFUOR<X,Y,S>(sIn,fIn));
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::neg(void)const
{
   using namespace auxiliary;
   return Fa<X,Y>(new NEGFUO<X,Y>(*this));
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::inv(void)const
{
   return Fa<X,Y>(new auxiliary::INVFUO<X,Y>(*this));
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::net(CpmRoot::Z i)const
{
   Y y0;
   Y yn=CpmRoot::netT<Y>(y0,i);
   return Fa<X,Y>(yn);
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::operator +(const Fa<X,Y>& f )const
{
   using namespace auxiliary;
   return Fa<X,Y>(new SUMFUO<X,Y>(*this,f));
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::operator *(const Fa<X,Y>& f )const
{
   using namespace auxiliary;
   return Fa<X,Y>(new PRODFUO<X,Y>(*this,f));
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::operator *(const Y& y )const
{
   using namespace auxiliary;
   return Fa<X,Y>(new Y_PRODFUO<X,Y>(*this,y));
}

template <class X, class Y>
Fa<X,Y> Fa<X,Y>::operator +(const Y& y )const
{
   using namespace auxiliary;
   return Fa<X,Y>(new Y_SUMFUO<X,Y>(*this,y));
}

template <class X, class Y>
Fa<X,Y> operator *( const Y& y, const Fa<X,Y>& f1 )
{
//   using namespace auxiliary;
   return (f1*y);
}

// I/O interface trivial. No assumtions on X,Y

template <class X, class Y>
bool Fa<X,Y>::prnOn(ostream& out)const
{
   return CpmRoot::writeTitle(" a function",out);
}

template <class X, class Y>
bool Fa<X,Y>::scanFrom(istream& in)
{
   in;
  // return (in!=0);
   if(!in) return false;
   else return true;
}

} // namespace

#endif
