//? cpmfftmore.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_FFTMORE_H
#define CPM_FFTMORE_H
/*

   Description: Fourier transform imbedded into a suitable
      class frame. We use different classes for space domain
      and frequency domain since the property of being real-valued
      has very different relevance in both cases that should
      be reflected in the definition of the member functions.
     
*/

#include <cpmfft.h>
#include <cpmva.h>

namespace CpmFourier{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::VVV;
   using CpmArrays::Va;
   using CpmArrays::VVa;
   using CpmArrays::VVVa;

//////////////////// classes SD1 and FD1 ///////////////////////////////
class FD1;
class SD1{ // space domain 1D
   Va<C> rep;
   SD1(V<C> const& v):rep(v){} // fast but not save
      // thus private
   friend class FD1;
   typedef SD1 Type;
   typedef C ScalarType;
public:
   CPM_LIN // linear algebra, scalar product, and
      // conjugation. Function con() and fft() 
      // differ in SDi and FDi (i=1,2,3); all other
      // functions are the same. So the present implementation
      // which does not derive FDi and SDi from a common base class Di
      // has a lot of code doublication but a clean user interface.
      // Notice the following dilemma: The base class would return
      // its own instances (typed Di) as return values of arithmetical
      // operations.
      // To interprete these as either FDi or SDi (in order to know what
      // kind of conjugation or Fourier transform would have to be
      // chosen) has then to be done by calling some constructor or other
      // conversion operator. This would be tedious and unnatural for the
      // user. As observed many times: inheritance  and arithmetics
      // don't fit together very well. Arithmetics should not be defined
      // before the last level of inheritance  is reached.
   explicit SD1(Z n=2);
      // all components are C(0,0); even if n is not a power of
      // 2, the array length will be of this kind. 
   SD1(V<C> const& v, bool careful);
      // using v as the internal value array rep. It will be taken as it
      // is for careful==false, otherwise it will be extended by zeros up
      // to the next power-of-two length if needed
   SD1(Z n, C z, Z& seed);
      // random generator. The internal value array of length
      // powOf2(n) will be filled by number z.ran(k++)
      // with k starting with k=seed. After call seed will have
      // a value that was not used so far.
   FD1 fft()const;
      // discrete fourier transform from space domain to frequency
      // domain
   SD1 re()const{ return (*this+con())*C(0.5);}
   SD1 im()const{ return (*this-con())*C(0.5);}
      // For any SD1 x we have x=x.re()+x.im(), 
      // (in the complex analogy z=u+iv, im() corresponds
      // to iv and not to v alone)
      // we have x.re().con()=x.re() and
      // x.im().con()=-x.im()
      // Notice that we have no equality operation defined in the
      // class since this would be useless due to numerical errors
      // 'a=b' above means (a-b).abs() not significantly above the
      // numerical noise level.
   V<C> toV()const{ return rep;}
      // retreiving the internal value array. This allows a simple
      // V<C> x handled by the presnt class. E.g.
      // x=SD1(x).fft().toV() changes x into its fourier transform.
   Z dim()const{ return rep.dim();}
   C& operator()(Z i){return rep[i];} 
   C const& operator()(Z i)const{return rep[i];} 
};

class FD1{ // frequency domain 1D
   Va<C> rep;
   FD1(V<C> const& v):rep(v){} 
   friend class SD1;
   typedef FD1 Type;
   typedef C ScalarType;
public:
   CPM_LIN
      // conjugation in frequency domain is conceptionally
      // defined as the conjugation in space domain carried
      // over to frequency domain via fourier transform.
      // As an implementation this would be rather inefficient;
      // fortunately there is the explicit representation
      // as CK as explained in cpmfft.h, equation (*)

   explicit FD1(Z n=2);
   FD1(V<C> const&, bool careful);
   FD1(Z n, C z, Z& seed);
   SD1 fft()const;
      // discrete Fourier transform from frequency domain to 
      // space domain
   FD1 re()const{ return (*this+con())*C(0.5);}
   FD1 im()const{ return (*this-con())*C(0.5);}
   V<C> toV()const{ return rep;}
   Z dim()const{ return rep.dim();}
   C& operator()(Z i){return rep[i];} 
   C const& operator()(Z i)const{return rep[i];}
};

//////////////////// classes SD2 and FD2 ///////////////////////////////
// See classes SD1 and FD1 for explanations to the various functions 
class FD2;
class SD2{ // space domain 2D
   VVa<C> rep;
   SD2(VV<C> const& v):rep(v){} // fast but not save
      // thus private
   friend class FD2;
   typedef SD2 Type;
   typedef C ScalarType;
public:
   CPM_LIN
   explicit SD2(Z n1=2,Z n2=2);
   SD2(VV<C> const&, bool careful);
   SD2(Z n1,Z n2,C z,Z& seed);
      // random generator
   FD2 fft()const;//{ return FD2(CpmFourier::fft(rep,1));}
   SD2 re()const{ return (*this+con())*C(0.5);}
   SD2 im()const{ return (*this-con())*C(0.5);}
   VV<C> toVV()const{ return rep;}
   Z dim1()const{ return rep.dim1();}
   Z dim2()const{ return rep.dim2();}
   C& operator()(Z i,Z j){return rep[i][j];} 
   C const& operator()(Z i,Z j)const{return rep[i][j];}
};

class FD2{ // frequency domain 2D
   VVa<C> rep;
   FD2(VV<C> const& v):rep(v){} 
   friend class SD2;
   typedef FD2 Type;
   typedef C ScalarType;
public:
   CPM_LIN
   explicit FD2(Z n1=2,Z n2=2);
   FD2(VV<C> const&, bool careful);
   FD2(Z n1,Z n2,C z,Z& seed);
   SD2 fft()const;
   FD2 re()const{ return (*this+con())*C(0.5);}
   FD2 im()const{ return (*this-con())*C(0.5);}
   VV<C> toVV()const{ return rep;}
   Z dim1()const{ return rep.dim1();}
   Z dim2()const{ return rep.dim2();}
   C& operator()(Z i,Z j){return rep[i][j];} 
   C const& operator()(Z i,Z j)const{return rep[i][j];}
};

//////////////////// classes SD3 and FD3 ///////////////////////////////
// See classes SD1 and FD1 for explanations to the various functions 
class FD3;
class SD3{ // space domain 3D
   VVVa<C> rep;
   SD3(VVV<C> const& v):rep(v){} 
   friend class FD3;
   typedef SD3 Type;
   typedef C ScalarType;
public:
   CPM_LIN
   explicit SD3(Z n1=2,Z n2=2,Z n3=2);
   SD3(VVV<C> const&, bool careful);
   SD3(Z n1,Z n2,Z n3,C z,Z& seed);
   FD3 fft()const;
   SD3 re()const{ return (*this+con())*C(0.5);}
   SD3 im()const{ return (*this-con())*C(0.5);}
   VVV<C> toVVV()const{ return rep;}
   Z dim1()const{ return rep.dim1();}
   Z dim2()const{ return rep.dim2();}
   Z dim3()const{ return rep.dim3();}
   C& operator()(Z i,Z j,Z k){return rep[i][j][k];} 
   C const& operator()(Z i,Z j,Z k)const{return rep[i][j][k];}
};

class FD3{ // frequency domain 3D
   VVVa<C> rep;
   FD3(VVV<C> const& v):rep(v){} 
   friend class SD3;
   typedef FD3 Type;
   typedef C ScalarType;
public:
   CPM_LIN
   explicit FD3(Z n1=2,Z n2=2, Z n3=2);
   FD3(VVV<C> const&, bool careful);
   FD3(Z n1,Z n2,Z n3,C z,Z& seed);
   SD3 fft()const;
   FD3 re()const{ return (*this+con())*C(0.5);}
   FD3 im()const{ return (*this-con())*C(0.5);}
   VVV<C> toVVV()const{ return rep;}
   Z dim1()const{ return rep.dim1();}
   Z dim2()const{ return rep.dim2();}
   Z dim3()const{ return rep.dim3();}
};

// Functions that should return only numerical noise
// if everything if implementation is OK
// Actual tests not yet systematic and covering
// see implementation for scope of tests
R fftTest1();
R fftTest2();
R fftTest3();

} // namespace

#endif

