//? cpmfieldoperator.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmoperator.h>
#include <cpmfieldoperator.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::State;
using CpmQuantumMechanics::FieldState;
using CpmQuantumMechanics::Operator;
using CpmQuantumMechanics::FieldOperator;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::TrjCon;
using CpmQuantumMechanics::TrjDMI_FilOpr;
#ifdef CPM_Fn
using CpmFunctions::F1;
using CpmFunctions::F2;
#endif
using CpmFunctions::R_Func;
using CpmFunctions::R2_Func;

using namespace CpmStd;

//////////////////////// class FieldOperator //////////////////////////

namespace {
   FieldState cook1Func(X2<FieldState,R> const& fs, Observable const& ob)
   { return fs.c1().cook1(ob);}

   FieldState cook2Func(X2<FieldState,R> const& fs, Observable const& ob)
   { return fs.c1().cook2(ob);}

   FieldState manParHamFunc(X2<FieldState,R> const& fs,
      R2_Func const& pot1, R2_Func const& pot2)
   { return fs.c1().manyParHam(pot1,pot2,fs.c2());}
      // calls FieldState::manyParHam(...) see cpmfieldstate.h

   FieldState manParHamFunc_t(X2<FieldState,R> const& fs,
      R2_Func const& pot1, R_Func const& tf1,
      R2_Func const& pot2, R_Func const& tf2)
   { return fs.c1().manyParHam_t(pot1,tf1,pot2,tf2,fs.c2());}
      // calls FieldState::manyParHam_t(...) see cpmfieldstate.h

   FieldState manParHamFunc_w(X2<FieldState,R> const& fs,
      R2_Func const& pot, R2_Func const& well, R_Func const& tf)
   { return fs.c1().manyParHam_w(pot,well,tf,fs.c2());}
      // calls FieldState::manyParHam_w(...) see cpmfieldstate.h

   FieldState hubFunc(X2<FieldState,R> const& fs,
      R const& a, R const& b){ return fs.c1().hub(a,b,fs.c2());}
}

FieldOperator::FieldOperator(Z d, F<FieldState,FieldState> const& f):
d_(d)
{
#ifdef CPM_Fn
   f_=F1<XFR,F<FieldState,FieldState>,FieldState>(f)(conInTime);
#else
   f_=F<XFR,FieldState>(bind(conInTime,_1,f));
#endif
}

FieldOperator FieldOperator::operator*(FieldOperator const& op)const
{
   return
#ifdef CPM_Fn
   FieldOperator(d_,
      F2<XFR,FieldOperator,FieldOperator,FieldState>(*this,op)(mul2));
#else
   FieldOperator(d_,F<XFR,FieldState>(bind(mul2,_1,*this,op)));
#endif
}

FieldOperator FieldOperator::operator+(FieldOperator const& op)const
{
   return
#ifdef CPM_Fn
   FieldOperator(d_,
      F2<XFR,FieldOperator,FieldOperator,FieldState>(*this,op)(add2));
#else
   FieldOperator(d_,F<XFR,FieldState>(bind(add2,_1,*this,op)));
#endif
}

FieldOperator FieldOperator::operator-(FieldOperator const& op)const
{
   return
#ifdef CPM_Fn
   FieldOperator(d_,
      F2<XFR,FieldOperator,FieldOperator,FieldState>(*this,op)(dif2));
#else
   FieldOperator(d_,F<XFR,FieldState>(bind(dif2,_1,*this,op)));
#endif
}

FieldOperator FieldOperator::operator*(C z)const
{
   return
#ifdef CPM_Fn
   FieldOperator(d_,F2<XFR,FieldOperator,C,FieldState>(*this,z)(mulC));
#else
   FieldOperator(d_,F<XFR,FieldState>(bind(mulC,_1,*this,z)));
#endif
}

FieldOperator FieldOperator::operator*(R r)const
{
   return
#ifdef CPM_Fn
   FieldOperator(d_,F2<XFR,FieldOperator,R,FieldState>(*this,r)(mulR));
#else
   FieldOperator(d_,F<XFR,FieldState>(bind(mulR,_1,*this,r)));
#endif
}

FieldOperator& FieldOperator::operator+=(FieldOperator const& fop)
{
   return *this=operator+(fop);
}

FieldOperator& FieldOperator::operator-=(FieldOperator const& fop)
{
   return *this=operator-(fop);
}

FieldOperator& FieldOperator::operator*=(C z)
{
   return *this=operator*(z);
}

FieldOperator& FieldOperator::operator*=(R r)
{
   return *this=operator*(r);
}

FieldOperator FieldOperator::cook1(Observable const& ob)
{
   return
#ifdef CPM_Fn
   FieldOperator(ob.dim(),F1<XFR,Observable,FieldState>(ob)(cook1Func));
#else
   FieldOperator(ob.dim(),F<XFR,FieldState>(bind(cook1Func,_1,ob)));
#endif
}

FieldOperator FieldOperator::cook2(Observable const& ob)
{
   return
#ifdef CPM_Fn
   FieldOperator(ob.dim(),F1<XFR,Observable,FieldState>(ob)(cook2Func));
#else
   FieldOperator(ob.dim(),F<XFR,FieldState>(bind(cook2Func,_1,ob)));
#endif
}

FieldOperator FieldOperator::manParHam(R2_Func const& pot1,
   R2_Func const& pot2, Z d)
{
   return
#ifdef CPM_Fn
   FieldOperator(d,F2<XFR,R2_Func,R2_Func,FieldState>
      (pot1,pot2)(manParHamFunc));
#else
   FieldOperator(d,F<XFR,FieldState>(bind(manParHamFunc,_1,pot1,pot2)));
#endif
   // thus finally uses FieldState::manParHam(...), see function
   // manParHamFunc above.
}

FieldOperator FieldOperator::manParHam_t(R2_Func const& pot1,
   R_Func const& tf1, R2_Func const& pot2, R_Func const& tf2, Z d)
{
   return
#ifdef CPM_Fn
   FieldOperator(d,F4<XFR,R2_Func,R_Func,R2_Func,R_Func,FieldState>
      (pot1,tf1,pot2,tf2)(manParHamFunc_t));
#else
   FieldOperator(d,F<XFR,FieldState>(bind(manParHamFunc_t,_1,
      pot1,tf1,pot2,tf2)));
#endif
   // thus finally uses FieldState::manParHam(...), see function
   // manParHamFunc above.
}

FieldOperator FieldOperator::manParHam_w(R2_Func const& pot,
      R2_Func const& well, R_Func const& tf, Z d)
{
   return
#ifdef CPM_Fn
   FieldOperator(d,F3<XFR,R2_Func,R2_Func,R_Func,FieldState>
      (pot,well,tf)(manParHamFunc_w));
#else
   FieldOperator(d,F<XFR,FieldState>(bind(manParHamFunc_w,_1,
      pot,well,tf)));
#endif
   // thus finally uses FieldState::manParHam(...), see function
   // manParHamFunc above.
}

FieldOperator FieldOperator::hub(R a, R b, Z d)
{
   return
#ifdef CPM_Fn
   FieldOperator(d,F2<XFR,R,R,FieldState>(a,b)(hubFunc));
#else
   FieldOperator(d,F<XFR,FieldState>(bind(hubFunc,_1,a,b)));
#endif
   // thus finally uses FieldState::hub(...), see function
   // hubFunc above.
}

FieldOperator FieldOperator::cook1(Operator const& op)
{
   return FieldOperator();
}

FieldOperator FieldOperator::cook2(Operator const& op)
{
   return FieldOperator();
}

//////////////////// class TrjDMI_FilOpr ///////////////////////////////

void TrjDMI_FilOpr::showAsArea(Graph& g)const
{
   Z mL=3;
   static Word loc("TrjDMI_FilOpr::showAsArea(Graph)");
   CPM_MA
   x_.mark(g,tc_.intMet_,tc_.gam_);
   g.vis();
   CPM_MZ
}
