//? cpmfieldoperator.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_FIELDOPERATOR_H_
#define CPM_FIELDOPERATOR_H_
/*
   Description: objects for describing quantum mechanical operators
      as functions from FieldSpace to FieldSpace
*/

#include <cpmfieldstate.h>
#include <cpmoperator.h>

namespace CpmQuantumMechanics{

   using namespace CpmArrays;
   using namespace CpmFunctions;
   using namespace CpmGraphics;
   using CpmAlgorithms::PeriodicAlarm;

////////////////////// class FieldOperator ///////////////////////////////

class FieldOperator: public QMBase{
   // field operators as functions FieldState x R -->  FieldState
   typedef FieldOperator Type;
   typedef X2<FieldState,R> XFR;
      // pair of a quantum state and a point in time
   Z d_;
      // number of discrete points on which the one particle
      // states are defined as complex-valued functions.
   F< XFR ,FieldState> f_;
      // we prefer not to rely on Fa<> for implementing arithmetics since
      // we don't need all the stuff implemented there. The direct and
      // lean implementation of arithmetics in cpmfieldobservable.cpp
      // is considered instructive.

   FieldOperator(Z d, F<XFR,FieldState> const& f):d_(d),f_(f){}

   static FieldState conInTime(XFR const& xsr,
      F<FieldState,FieldState> const& f){ return f(xsr.c1());}
      //: constant in time
      // auxiliar device for implementing the constructor
      // FieldOperator(Z,F<FieldState,FieldState>);

   static FieldState mul2(XFR const& xsr, FieldOperator const& op1,
      FieldOperator const& op2)
   {  return op1.f_(XFR(op2.f_(xsr),xsr.c2()));}

   static FieldState add2(XFR const& xsr, FieldOperator const& op1,
      FieldOperator const& op2)
   { return op1.f_(xsr)+op2.f_(xsr);}

   static FieldState dif2(XFR const& xsr, FieldOperator const& op1,
      FieldOperator const& op2)
   { return op1.f_(xsr)-op2.f_(xsr);}

   static FieldState mulC(XFR const& xsr, FieldOperator const& op,
      C const& z){ return op.f_(xsr)*z;}

   static FieldState mulR(XFR const& xsr, FieldOperator const& op,
      R const&r){  return op.f_(xsr)*r;}

public:

   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)

   FieldOperator(){}

   FieldOperator(R2_Func const& pot,Z d);
      // many-particle Hamiltonian with pot as time-dependent
      // pair potential

   FieldOperator(Z d, F<FieldState,FieldState> const& f);
      // constructor from function

   explicit FieldOperator(Operator const& op);

   FieldOperator operator*(FieldOperator const& fop)const;

   FieldState actOn(FieldState const& st, R t)const
      {
       //  cpmcerr<<"FieldOperator::actOn started"<<endl;
         auto temp=XFR(st,t);
      //   cpmcerr<<"temp done for t ="<<t<<endl;
        // cout<<"temp done for t ="<<t<<endl;
        /* Graph gd;
         st.mark(gd);
         cpmwait(15,2);
         temp.c1().mark(gd);
         cpmwait(20,2);*/
         FieldState res=f_(temp);
       //  cpmcerr<<"res done"<<endl;
       //  cpmcerr<<"FieldOperator::actOn done"<<endl;
         return res;
      }
      // Operator acts on state at time t

   FieldState operator*(XFR const& xsr)const{ return f_(xsr);}
      // another form of actOn;

// generators for creation and destruction operators as static members

   static FieldOperator crt(Z i);
   static FieldOperator des(Z i);
   static FieldOperator crt(State const& st);
   static FieldOperator des(State const& st);
   static FieldOperator psi(Z i);
   static FieldOperator phi(Z i);
   static FieldOperator psi(State const& st);
   static FieldOperator phi(State const& st);
   static FieldOperator cook1(Observable const& ob);
   static FieldOperator cook2(Observable const& ob);

   static FieldOperator cook1(Operator const& op);
      // trivial implementation so far
   static FieldOperator cook2(Operator const& op);
       // trivial implementation so far

   static FieldOperator manParHam(R2_Func const& pot1,
      R2_Func const& pot2, Z d);
   // Many particle Hamiltonian for one type of particles.
   // pot1 is a potential in one particle space
   // and pot2 a pair potential which, together with d
   // and the data from QMBase define a Hamiltonian.

   static FieldOperator manParHam_t(R2_Func const& pot1,
   R_Func const& tf1, R2_Func const& pot2, R_Func const& tf2, Z d);
   // Many particle time-dependend Hamiltonian for one type of particles.
   // pot1*tf1(t) is a potential in one particle space valid at time t,
   // and pot2*tf2(t) a pair potential at time t which, together with d
   // and the data from QMBase define a Hamiltonian.

   static FieldOperator manParHam_w(R2_Func const& pot,
      R2_Func const& well, R_Func const& tf, Z d);
   // Many particle time-dependend Hamiltonian for one type of particles.
   // pot + well*tf1(t) is a potential in one particle space valid at time
   // t,
   // which, together with d and the data from QMBase define a Hamiltonian.
   // (this one has no particle to particle interaction)

   static FieldOperator hub(R a, R b, Z d);
   // Hubbard model Hamiltonian
};

/////////////////////// class TrjDMI_FilOpr //////////////////////////////
// trajectory for time-dependent Hamiltonians.
// similar to TrjTD but replaces Operators by
// FieldOperators

class TrjDMI_FilOpr: public TrjDMI<FieldState>{//trajectory method: DMI
   PeriodicAlarm pal_;
   FieldOperator op_;
   FieldState op(FieldState const& x, R t)const
      { //cpmcerr<<"op started"<<endl;
         auto res=op_.actOn(x,t);
        // cpmcerr<<"op done"<<endl;
         return res;
      }
public:
   typedef TrjDMI_FilOpr Type;
   virtual TrjBase* clone()const{ return new TrjDMI_FilOpr(*this);}
   TrjDMI_FilOpr(){}
   TrjDMI_FilOpr(FieldOperator const& h, FieldState const& x,
      TrjCon const& tc):
         TrjDMI<FieldState>(x,tc),op_(h){ cpmcerr<<
            "instance of TrjDMI_FilOpr constructed"<<endl;
         }
      // creating a trajectory from hamiltonian and initial state
   virtual void showAsArea(Graph& g)const;
   virtual Word text()const{ return Word("TrjDMI_FilOpr");}
};

} // namespace

#endif
