//? cpmfieldstate.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_FIELDSTATE_H_
#define CPM_FIELDSTATE_H_
/*
   Description: Development and test of quantized fields
      in one spatial dimension
*/

#include <cpmx.h>
#include <cpmoperator.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;
using CpmArrays::X3;

namespace CpmQuantumMechanics{

///////////////////////////////// class Ky ///////////////////////////////

class Ky{ // index set for an orthonormal base of FieldSpace
   // auxiliary class needed as an index set for an orthonormal
   // base of FieldSpace.
   // The name Key would be inconvenient to use
   // since there is already a class CpmRootX::Key in use.
   // I prefer short names anyway.
   // Actually the instances of Ky are multi-indexes i.e.
   // lists of integers. List length is s_ and list values are
   // not negative in our present case, due to their interpretation
   // as occupation numbers.
// independent;
   V<Z> v_;
      // Z-valued array, the components play the role of occupation
      // numbers at lattice points.
// dependent
   Z s_;
      // Number of lattice points, s_==v_.dim()
// private function
   R sigma(Z i)const;
      // function needed to deal with fermions;
   void ini_(){ s_=v_.dim();}
// static members
   static Z lim_;
      // limit to the occupation numbers, 1 for fermions,
      // large for bosons
   static Z eps_;
      // +1 for fermions , -1 for bosons, set depending on
      // lim_=1 (for fermions) or larger (for bosons)
// static member functions
   static void setEps(void){ if (lim_==1) eps_=1; else eps_=-1;}
      // setting eps_ consistently

   typedef Ky Type;

public:
   CPM_IO
   CPM_ORDER
   CPM_NAM(Ky)
  // explicit Ky(Z s=0):s_(CpmRoot::posVal(s)),v_(s_,Z(0),LEAN){}
      // Basic constructor
      // no exception for negative s, absolute value used!
   explicit Ky(V<Z> const& v):v_(v){ini_();}
   explicit Ky(Z s=0):v_(CpmRoot::posVal(s)){ini_();}
      // Basic constructor

   Z dim()const{ return s_;}
     //: dimension

   Z sup()const;
      //: supremum
      // returns the value of the largest component for s_>0
      // and -1 else

   Z inf()const;
      //: infimum
      // returns the value of the smallest component for s_>0
      // and -1 else

   void sup_(Ky const& k);
      //: supremum
      // changes *this into the component-wise maximum
      // of *this and k.
      // Error if both have not the the same dimension

// basic tools for defining creation operators
// (see code defining crt(Z), crt(State)
   bool up_(Z i);
   R up(Z i)const;
// basic tools for defining destruction operators
// (see code defining des(Z), des(State)
   bool down_(Z i);
   R down(Z i)const;
 //  Z operator[](Z i)const{ return v_.cui(i);} // needed for crt and des
   Z operator[](Z i)const{ return v_[i];} // needed for crt and des
      // in up_,up,down_,down, and operator[] it is now assumed that the
      // index is in proper range. The destruction and creation operators
      // use this functions in a manner which takes care.

// setting the boson-fermion alternative
   static void setLim(Z i){ lim_=CpmRoot::posVal(i); setEps();}
   static Z getLim(){ return lim_;}
   static Z getEps(){ return eps_;}
   static void setBosons(Z lim=1024){ lim_=lim; setEps();}
   static void setFermions(){ lim_=1;setEps();}
};

class FieldOperator;

/////////////////////////// class FieldState /////////////////////////////

class FieldState:public QMBase{
   // state space in occupation number representation
   // General considerations on finite-dimensional Hilbert spaces:
   // Let an orthonormal basis of Hilbert space H be represented as a
   // class, say B. This means that there is an interpretation according
   // to which each instance of B is an element of an orthonormal basis of
   // H and vice versa. Then each element psi \in H determines a
   // function B --> C, b |--> <b|psi> ( where <|> denotes the scalar
   // product in H and C are the complex numbers) and this mapping
   // H --> (B --> C) is linear and an isomorphism of Hilbert spaces if
   // the class B --> C is made a Hilbert space in a natural manner.
   // The class B-->C is naturally represented as CpmArrays::M<B,C> in C+-
   // or as std::map<B,C> in the C++ standard library (former STL).
   // If the cardinality |B|=:d is large, this is much more economic
   // than the class of all psi \in V<C> with psi.dim()=d.
   // In the present case we deal with a Hilbert space of states for which
   // the dimension d is written as a product of two natural numbers
   // which are the data members s_ and lim_ of the present class. For
   // convenience we refer to these numbers as s and l in the present
   // discussion. There are two natural interpretations of this space:
   //
   // 1. Tensor product of s single-particle spaces, each of dimension l.
   // Here we  consider the single-particle spaces as attached to s space
   // points, which may (or may not) form a linear lattice, the biotope of
   // the system. As mentioned already, these 'localized' state spaces all
   // have the same dimension l, and we consider them as identical to
   // some l-dimensional Hilbert space H_0 which has an orthonormal
   // base B_0.
   // The physical interpretation is that at each of the s space points
   // (sites) x_i, i=1,...,s,  there is a particle p_i (or other kind of
   // quantum system), which is a fermion for l=1 and a boson else. This
   // is not important for the state space as a Hilbert space but only for
   // the definition of operators.
   // A basis of H is given by the set of lists (b_1,...b_s) where b_i
   // \in B_0. Put differently, a natural basis of H has the form (B_0)^s
   // so that the dimension of H is d = l^s and thus grows exponentially
   // with s. If, however, l=1 we have d=1 for all s so that this case is
   // not interesting for fermions, so lets consider bosons only.
   // For bosons it is natural to consider for each site i the 'local'
   // creation and destruction operators c_i and d_i satisfying
   // [d_i,c_i]=1. Since operators referring to different sites commute
   // we have the general canonical relations [d_i,c_j]=\delta_{i,j},
   // [d_i,d_j]=0, [c_i,c_j]=0 . Since operators referring directly to
   // different sites never anti-commute, a correspondingly trivial
   // realization of canonical anti-commutation relations is not possible.
   // In the bosonic case, it is a consitent physical picture that at each
   // site there is placed a boson (oscillator). The state of the whole
   // system is, however, not given by assigning to each site a local
   // state. This would only lead to pure product states (nuclear states)
   // in the tensor product state. The generic state is a linear
   // combination of nuclear states. Notice that this is exactly what
   // holds for the state space of s distinguishable bosons. If these
   // bosons have a two-dimensional state space (qu-bits), e. g.
   // neutral atoms in electronic ground-state with respect to the
   // normal fine structure but showing a two-level hyperfine splitting.
   // For the states of this bosonic system there is also a representation
   // in terms of fermions via the Jordan-Wigner transformation, which we
   // do not consider here.

   // 2. Fock space over a s-dimensional single-particle space.
   // This works for bosons and fermions in nice parallelity and is the
   // case to which the creation and destruction operators defined here
   // directly refer. It is instructive to look at the single particle
   // space as the space of complex functions on B_0 = {x_1,x_2,...,x_s}.
   // For each i \in B_0 there is a creation operator c_i and its
   // Hermitean conjugate d_i (destruction operator) and a
   // orthonormal base of H is of the form
   // (n_1! n_2!...n_s!)^{-1/2} {c_1}^{n_1} {c_2}^{n_2}...{c_s}^{n_s} |0>
   // =: |n_1,n_2,...,n_s> where n_i \in {0,...s-1}, notice 0!=1
   // see e.g. Messiah I, p. 453, (XII.68) For fermions n_i \in {0,1}
   // and all pre-factors are 1.
   // On the other hand, simply declaring the set B :=
   // { (n_1,n_2,...,n_s)| n_i \in {0,...s-1}}
   // as an orthonormal base can be implemented by considering the
   // Hilberspace M<B,C> as mentioned above. In this 'Hilbertspace in
   // occupation number representation' creation and destruction operators
   // can be defined by the formulas
   // d_i |n_1,...n_i,...,n_s> = {n_i}^{1/2} |n_1,...,n_i - 1,...,n_s>
   // c_i |n_1,...n_i,...,n_s> = {n_i + 1}^{1/2} |n_1,...,n_i + 1,...,n_s>
   // (see Schweber, p. 130 (47), p. 131 (51))
   // for bosons and
   // d_i |n_1,...n_i,...,n_s> = (-1)^{q_i} n_i |n_1,...,n_i - 1,...,n_s>
   // c_i |n_1,...n_i,...,n_s> =(-1)^{q_i}(1-n_i)|n_1,...,n_i + 1,...,n_s>
   // where q_i := n_1 + n_2 + ... + n_(i-1)
   // (see Schweber, p. 138 (87), p. 139 (89),(90))
   // for fermions
   // As an effect of this definition, we have
   // |n_1,n_2,...,n_s> =
   // (n_1!n_2!...n_s!)^{-1/2} {c_1}^n_1 {c_2}^n_2...{c_s}^n_s |0,0,...,0>

   Z dL_;
      // 'local dimension' 4 for Dirac particle, 2 for Pauli particle
      // 1 for Schroedinger particle

   Z s_;
      // Dimension of single-particle space from which FieldState is the
      // Fock space.
   Z lim_;
      // limit for occupation number, 1 for fermions.
      // Should be large enough not to have an influence
      // on results if >1 (then we have bosons). So the type
      // of statistics is a state attribute here.
   bool pbc_;
      // periodic boundary condition: normally true
      // false for reflecting boundary conditions.
      // Influences function b_1(Z&)

   typedef CpmArrays::M<Ky,C> Mkc;
      // M<X,Y> has a data member p_ of type std::map<X,Y>

   // associative array either from C+- or from the standard library
   Mkc m_;
   FieldState(Z s, Mkc const& m, Z lim, bool pbc=true):
      s_(s>=0 ? s : -s ), lim_(lim),pbc_(pbc),m_(m){}

   FieldState mod(Mkc const& m)const{return FieldState(s_,m,lim_,pbc_);}
      // returns a modified version of *this

   typedef FieldState Type;
   void b_1(Z& i)const
      // changes i into the periodic or reflected position
   {
      if(i<1){ if (pbc_) i+=s_; else i=-i;}
      else if(i>s_){ if (pbc_) i-=s_; else i=-i+2*(s_-1);}
      else return;
   }
public:
   CPM_IO
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
   CPM_NAM(FieldState)

   FieldState():s_(1),lim_(1),pbc_(true){}

   FieldState(Z n):s_(n),lim_(1),pbc_(true){}

   FieldState(Z s, Z kind, Z lim, bool pbc=true);
      // Creating the normalized ground state on a lattice of s points
      // (Fock vacuum state)
      // for kind==1 and the zero-vector on the same lattice else

   FieldState zero()const{ return FieldState(s_,0,lim_,pbc_);}
      //: zero
      // zero-vector with the same general data as *this

   FieldState vac()const{ return FieldState(s_,1,lim_,pbc_);}
      //: vacuum
      // Fock vacuum with the same general data as *this

   FieldState(Ky const& k, C z, Z lim, bool pbc=true):
      s_(k.dim()),lim_(lim),pbc_(pbc){ m_[k]=z; }

   FieldState(State const& st,Z lim, bool pbc=true);
      // takes the lattice size from n=st.dim(). Creates a
      // FieldState which is obtained by letting the
      // operator sum from i=1 to i=n of crt(i)*st[i] act on
      // the ground state FieldState(n,1). The actual implementation
      // is more direct and more efficient.

   FieldState(V<State> const& vs,Z lim, bool pbc=true);
      // Assumes that the components of vs are States on the same
      // lattice.
      // Let m=vs.dim(). Then the generated space is a m-particle
      // state.

   void set_(Ky const& k, C z){ m_[k]=z;}
      // notice that mutating functions end in _ by convention

   Z size()const{ return s_;}
      // Dimension of the one-particle space
      // over which FieldState happens to be the Fock space.
   Z dim()const{ return s_;}
      // Dimension of the one-particle space
      // over which FieldState happens to be the Fock space.

   Z locDim()const{ return dL_;}
      //: local dimension

  // Bio bio()const;
      // biotope

   C get(Ky const& k)const;
      // gets the value of the wave function at 'occupation index'
      // i.e. at a list of discrete positions given by k.

   C operator[](Ky const& k)const{ return get(k);}
      // notice that return m_[k] would not define a constant operator
      // since it sets m_[k]=0 in cases (i.e. for values of k) in which
      // m_[k] was not given a value earlier.

   C operator[](Z i)const;
      // returns a a sum of all those values m_[k] for which k[i]>0.
      // This is OK for rough visualization of the state. If one would
      // use the probability density instead of the amplitude density
      // ('the wave function') this would under the name marginal
      // distribution of the probability density of all particles.
      // Proper index range is 0,1,...s_-1

   State toState(bool ignorePhases)const;
      // If ignorePhases==0, this returns a State s such that
      // s.dim()=size() and s[i]=(*this)[i] for all i \in {0,...s_-1}
      // with operator[] just introduced. If ignorePhases!=0 the
      // resulting State res describes the 'marginal probability density'
      // i.e. res[i] is the probability to find a particle at place i
      // irrespective of other particles being arround elsewhere.

   State den()const;
      //: density
      // probability density as a distribution over the s_ discrete
      // points.
      // The square roout could be interpreted as a real state that
      // is conveniently to display.
      // Meaning has to be checked, clarify connection to Bethe-Salpeter
      // amplitudes.

   Observable toObservable(bool ignorePhases)const;

   Type crt(Z i)const;
      // crt =creator = creation operator
      // 'creates an oscillator quantum at lattice position i.
      // Notice that for i out of range (with respect to the state
      // to which the operator is being applied) effectively a lattice
      // point within the proper range is used (considering the
      // index range a torus)

   Type des(Z i)const;
      // destructor (not in the sense of C++) = destruction operator.
      // The basic property of crt(i) and des(j) is:
      //    psi.crt(i).des(j)-psi.des(j).crt(i)==psi*krn(i,j)
      // which corresponds to
      //    [a(i),a*(j)]=[des(i),crt(j)]=krn(i,j)*1
      // taking into account the order of application given by
      // postfix notation.
      // Notice that for i out of range (with respect to the state
      // to which the operator is being applied) effectively a lattice
      // point within the proper range is used (considering the
      // index range a torus).
      // Old explanation which holds for bosons only,
      // see des(State const& st) for the inclusion of fermions.

   Type crt(State const& st)const;
      // The natural situation is st.dim()==size().
      // equals sum of crt(i)*st(i)

   Type des(State const& st)const;
      // The natural situation is st.dim()==size().
      // equals sum of des(i)*st(i).con()

      // For R eps=Ky::getEps();
      // and State s1,s2; FieldState fs; s1.dim()==s2.dim()==fs.size();
      // fs.crt(s2).des(s1)+fs.des(s1).crt(s2)*eps-fs*(s1|s2)==0

    Type phi(Z i)const{  return (crt(i)+des(i))*facPhi;}
       // The couple (phi,psi) is the natural self-adjoint
       // derivate of the mutually adjoint couple (crt,des)
       // of operators.
       // [phi(i),psi(j)]=C(0,krn(i,j))*1

   Type psi(Z i)const{ return (crt(i)-des(i))*facPsi;}
      // see comment to previous function phi

   Type phi(State const& st)const{  return (crt(st)+des(st))*facPhi;}
      // Going from from crt(i),des(j) to crt(s1), des(s2) is
      // analogous to going from psi(i),phi(j) to psi(s1),phi(s2)

   Type psi(State const& st)const{ return (crt(st)-des(st))*facPsi;}
      // For R eps=Ky::getEps();
      // and State s1,s2; FieldState fs; s1.dim()==s2.dim()==fs.size();
      // C sp12=((s1|s2)-(s1|s2).con()*eps)*C(0,0.5);
      // fs.psi(s2).phi(s1)+fs.phi(s1).psi(s2)*eps-fs*sp12==0;

   Type cook1(Observable const& ob)const;
      // Lets the observable ob (assuming ob.dim()==size()) act
      // on *this by the Fock-Cook construction. Since the matrix
      // elements of ob get automatically periodically continued,
      // it is not necessary that ob.dim()==size() holds.

   Type cook2(Observable const& ob)const;
      // Lets the observable ob (assuming ob.dim()==size()) act
      // on *this by the Fock-Cook construction. Since the matrix
      // elements of ob get automatically periodically continued, it is
      // not necessary that ob.dim()==size() holds.

   Type manyParHam(R2_Func const& pot1, R2_Func const& pot2, R t)const;
      //: 'many particle Hamiltonian'
      // returns a FieldState which results from application of
      // a many-body hamiltonian with position dependent single particle
      // potential pot1 and pair potential pot2. Both potentials are
      // functions of two variables (instances of R2_Func), where the
      // second argument is time. The first argument is an absolute
      // particle position in case of pot1 and a relative position of
      // a particle pair in the case of pot2.
      // Positions are always given by QMBase::posNor(d) and thus are in
      // the range [-1/2,+1/2]

   Type manyParHam_t(
      R2_Func const& pot1, R_Func const& tf1,
      R2_Func const& pot2, R_Func const& tf2, R t)const;

   Type manyParHam_w(
      R2_Func const& pot, R2_Func const& well, R_Func const& tf, R t)const;
      // No inter-particle interaction. External potential at time t
      // is pot+well*tf(t). So, the well changes height with time and so
      // allows to separate a biotope in two parts in a way that the
      // particles in the two parts don't influence each other in case of
      // a sufficiently high and thick well.

   Type hub(R a, R b, R t)const;
        // Hamiltonian of the Hubbard model

   R nor_(R r=1);
      // returns the norm of *this and changes *this into
      // a FieldState of norm r by multiplying with a real factor.
      // If *this was the zero FieldState, it will be turned into
      // a 'standard State' of norm r

   R sup()const;
      // max of absolute value of components

   Type reduce(Z d=32)const;
      // returns a state res that no occupation number higher
      // than d occurs

   void show(Frame& f, Z mult_i=8, Z mult_j=4,
      Z iShift=0, Z jShift=0, bool ignorePhases=false)const;
      //: show
      // This is the comment to the obsolete version. A valid
      // implementation of graphical representation is still missing.
      // notice that the state space has a very high dimension even
      // if s_ and Ky::lim_ have moderate values.
      // one component of rep will be represented as a rectangle of
      // mult_i*mult_j pixels. The whole image will be shifted by
      // (iShift,jShift) with respect to the left upper corner
      // of the graphics window f.
      // Actually the components of rep are not converted to a screen
      // color directly but before get renormalized by a real factor
      // chosen such that the dominant component of rep gets an absolute
      // value of 1. If the true values of the components of a state s
      // are to be shown use
      // Graph g;
      // g.show(s.toRealForm())
      // If ignorePhases differs from 0, what is represented is the
      // probability density instead of the complex-valued wave function.

      void mark(Graph& f, Z ipolMeth=1, R gamma=0.5, R secFac=1.1)const;
         //: mark
         // See State2::mark in cpmstate2.h for the meaning of the
         // arguments. Shows the two-particle component of Fock-state
         // *this by finally calling
         // State2::mark(f,ipolMeth,gamma,secFac).

   X3<Ky,Z,R> dgn()const;
      // diagnosis for state *this. The first component of the result
      // gives the maximum occupation number found at each lattice point,
      // the second component gives the number of non-zero components,
      // and the third component gives the norm of the state.

   Type makeReal()const;
      // returns a FieldState with all components of *this
      // replaced by their absolute value

   static R test(Z com);
      //: test
      // test of the most important relations.
      // returns the sum of errors (also the unavoidable
      // small numeric ones, so that only large values
      // indicate malfunction.

   static R graphicsTest(Z com);
      //:: graphics test
      // test of graphic representation modalities.
   };

} // namespace

#endif
