//? cpmfl.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_FL_H_
#define CPM_FL_H_
/*
Purpose: Define class F<X,Y> describing 'functions' from
   X to Y. Later a derived class Fa<X,Y> will define arithmetic
   operations. This description also has in mind this class.
   The classes have default constructor, copy constructor, and assignment
   operator. Thus these 'functions' behave like the most conventional
   arithmetic quantities (they have value semantics, moreover they
   satisfy the strict value interface, see preamble of cpmv.h)

   Example:

      F<R,R> f1(cos);
      F<R,R> f2(sin);
      F<R,R> f3=f1&f2;
      F<R,R> f4=f1+f2;
      R x=1.2345;
      R eps1=f3(x)-sin(cos(x));
      R eps2=f4(x)-sin(x)-cos(x);

   will result in eps1=eps2=0. I.e. the concatenation of functions sin()
   and cos() can be created as a F object simply by a binary operator
   acting on the F objects corresponding to sin() and cos() (notice the
   order convention for & which is not intuitive to everybody).
   F<>-instances are allowed as arguments and return values of functions:

      F<C,C> power3(const F<C,C>& g)
      {
         return g*g*g;
      }

   creates the third power of a complex function, i.e.
      power3(g)(x)=g(x)*g(x)*g(x).

   In a nutshell, these new functions behave just like normal variables.
   They can also serve as data members in class definitions. This is
   important in physics and engineering applications where e.g.
   electrical fields now can be directly represented as functions instead
   of discrete value samples. They also can form components of vectors.

   For instance, a family of functions consisting of the first 101
   Chebyshev polynomials
         T0(x),....T100(x)
   can be defined by verbally coding the (numerically not optimal)
   definition

         Tn(x):=cos(n*arccos(x)):
   Z n=100;
   IvZ iv(0,n);
   V<Fa<R,R> > Chebyshev(iv);
   Fa<R,R> ACos(acos);
   Fa<R,R> Cos(cos);
   for (Z i=0; i<=n; i++) Chebyshev[i]=(ACos*i)&Cos ;

   The value of the 27-th Chebyshev polynomial for x=0.5 then will be
   written as Chebyshev[27](0.5)

   In defining new classes, data members of type F<X,Y> are allowed
   and for the classes defined this way the default copy constructor
   and default assignment operator are OK.

   Compared to all other classes that I dealt with, F<X,Y> shows
   the particularity that it is not possible to implement
   CPM_IO and CPM_ORDER (see cpminterfaces.h) in a useful manner.
   Later Fr<X,Y> tries to do the best one can in this respect.

   Remarks on 'functions as values':

   Consider

      R sum=0;
      for (Z i=1;i<=1000;++i) sum+=(result of some complicated
         computation with i);
      R x=sum;

   and the similiar construct

      F<R,R> f;
      for (Z i=1;i<=1000;++i) f+=(Sin*i); // (*)
      F<R,R> g=f;

   There is a remarkable difference between the two:
   Whenever x will be used in subsequent parts of the program it is
   simply a number and the potential pain of getting to it is forgotten.
   On the contrary, when using g by evaluating it for specific
   values of the argument, e.g.
      R y=g(137)-g(1./137);
   each evaluation has to go through the huge expression which g is
   defined in (*) to be. What was achieved by defining g compactly
   was to compose the code defining the Sin*i into the code defining g.
   But this code was composed verbatim without any intent (and chance)
   of simplification or compactification. So the normal idea  associated
   with evaluation is misleading in this case.
   In a time-stepping simulation one could be tempted
   to replace in each step a data member F<X,X> f of the evolving system
   by f&g with some transformation g characteristic for the step under
   consideration. This would make f more expensive to compute with each
   step (actually we would run into stack overflow sometime).
   Here the right way is to parameterize the mappings and express their
   composition in terms of a computation done with these parameters.
   Sophus Lie has forseen the importance of this kind of computational
   treatment of the combination of transformations. With group theory,
   Lie groups of transformations, and their linear representations, there
   has a huge body of knowledge grown on this topic.
*/

#include <cpmuc.h>
   // for reference counting handle template CpmArrays::P<X>
#include <cpmword.h>
#include <functional>

//#undef CPM_Fn
//#define CPM_Fn CPM_Fn has to be set in cpmdefinitions.
// The next version of C+- will very probably have the form which
// corresponds here to #undef CPM_Fn.
// No longer so sure. The simple basic run of the my pala program
// first ran 1% faster with the old method. But a new run did fiffer only
// by 0.2 s of 95 s. So,there seems to be no significant speed difference.

namespace CpmFunctions{

   using CpmRoot::Z; // for order relation
   using CpmRoot::Word;
   using CpmArrays::P;
  // using std::function;

/////////////////////////////// class FncObj<> /////////////////////////

template <class X, class Y>
class  FncObj{ // function objects
   // FncObj stands for function object
   // A class which is derived from some FncObj<X,Y> is said to be a
   // function class and all instances of function classes are called
   // function objects. A function object can do everything (and more)
   // than a function can do. Unlike a function, a function class can be
   // defined within a function bloc. Let G be a function class and g an
   // object belonging to it. The value of g at x can be written as
   // g(x) (see. operator()). If in deriving G from
   // FncObj<X,Y>, no further non-static data members are added (then
   // G is said to be a pure function class), G has just one instance
   // g; this is selected by a statement as simple as
   //      G g;
   // Despite this close relationship between function objects and their
   // classes, there is an important syntactic difference between these
   // entities. As already mentioned, function values are accessed via
   // instances, e.g. Y y=g(x). Function arithmetics can be implemented
   // via template classes, among the arguments of which are
   // function classes. We will use such templates only as a transitory
   // step in implementing value-like function objects in F<X,Y>,
   // Fa<X,Y>.

   // Notice that FncObj<X,Y> is an abstract class. This is a usefull
   // prevention against invalid attempts to re-define operator() (e.g.
   // by forgetting the final 'const' attribute.
   //
   // Example: function class describing the sin function:
   //
   //    class Sin: public FncObj<R,R>{
   //       public:
   //       R operator()(const R& x)const{ return sin(x);}
   //    };
   // classes derived from FncObj may implement a more flexible and
   // effective evaluation scheme than mere function: certain
   // preparatory actions (e.g. computation of auxiliary quantities that
   // are held as data members) can be done by constructor code, so that
   // evaluation of operator()(X const&) gets simpler.

   typedef FncObj<X,Y> Type;
   CPM_INVAR(FncObj)
      // makes sure that no client class will use FncObj in a way that
      // makes use of copy or assignment

public:

   FncObj(void){}

   virtual ~FncObj(void){}

   virtual   Y operator()(X const& x)const=0;

};

namespace aux{
// prevention against polluting namespace CpmFunctions
// The classes to be defined in this namespace aux are not intended
// for direct usage. They will be used internally for building
// classes which have a more expressive interface: the
// classes F<>,F1<>, .... F6<>
// and
// F1_1<>, F2_1<>, F2_2<>, F3_1<>, F3_2<>, F3_3<>, F4_1<>, F4_2<>,
// F4_3<>, F4_4<>

///////////////// ParFncObj ////////////////////////////////////////////

template <class X, class Y>
class ParFncObj : public FncObj<X,Y>{

   Y (* const fp)(X);
public:
   Y operator()(X const& x)const{ return (*fp)(x);}
   ParFncObj( Y (*g)(X)):fp(g){}
};

///////////////// Par0FncObj ///////////////////////////////////////////
template <class X, class Y>
class Par0FncObj : public FncObj<X,Y>{

   Y (* const fp)(X const&);

public:
   Y operator()(X const& x)const{ return (*fp)(x);}
   Par0FncObj( Y (*g)(X const&)):fp(g){}
};

///////////////// StdFunctionObj //////////////////////////////////////////
template <class X, class Y>
// Addition 2020-02-13. Means to use also std::function<Y(X)> objects
// as a source for F<X,Y>-objects. Requires C++11 or higher.
class StdFunctionObj : public FncObj<X,Y>{

   const std::function<Y(X)> f_;

public:
   Y operator()(X const& x)const{ return f_(x);}
   StdFunctionObj( std::function<Y(X)> f):f_(f){}
};

///////////////// ParcFncObj ///////////////////////////////////////////
// Addition 2010-01-13. Means to use also functions with
// constant return value as a source for F<X,Y>-objects.
template <class X, class Y>
class ParcFncObj : public FncObj<X,Y>{

   const Y (* const fp)(X const&); // This is the prototype of
   // many mathematical functions in mpreal have (unfortunately
   // up to an additional default argument)

public:
   Y operator()(X const& x)const{ return Y((*fp)(x));}
   ParcFncObj( const Y (*g)(X const&)):fp(g){}
};

//////////////////// Par1FncObj ///////////////////////////////////////////
template <class X, class P1, class Y>
class Par1FncObj : public FncObj<X,Y>{

   const P1 p_;
   Y (* const f_)(X const&, P1 const& );

public:
   Y operator()(X const& x)const{ return (*f_)(x,p_);}
 //  P1 const& getP(void)const{return p_;} // experiment
   Par1FncObj( Y (*g)(X const&, P1 const&), P1 const& p):p_(p),f_(g){}
};

/////////////// Par1_1FncObj ///////////////////////////////////////////
template <class X, class P1, class Y>
class Par1_1FncObj : public FncObj<X,Y>{

   const X x_;
   Y (* const f_)(X const&, P1 const& );

public:
   Y operator()(P1 const& p)const{ return (*f_)(x_,p);}
   Par1_1FncObj( Y (*g)(X const&, P1 const&), X const& x):x_(x),f_(g){}
};

#ifdef CPM_Fn
/////////////// Par2FncObj ///////////////////////////////////////////
// See F2, F2_1, F2_2 for the reasons to introduce these class templates

template <class X, class P1, class P2, class Y>
class Par2FncObj : public FncObj<X,Y>{

   const P1 p1_;
   const P2 p2_;
   Y (* const f_)(X const&, P1 const&, P2 const& );

public:
   Y operator()(X const& x)const{ return (*f_)(x,p1_,p2_);}
   Par2FncObj(Y (*g)(X const&, P1 const&, P2 const&), P1 const& p1,
      P2 const& p2)
   :p1_(p1),p2_(p2),f_(g){}
};

/////////////// Par2_1FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class Y>
class Par2_1FncObj : public FncObj<X,Y>{

   const X x_;
   const P2 p2_;
   Y (* const f_)(X const&, P1 const&, P2 const& );

public:
   Y operator()(P1 const& p1)const{ return (*f_)(x_,p1,p2_);}
   Par2_1FncObj(Y (*g)(X const&, P1 const&, P2 const&), X const& x,
      P2 const& p2)
   :x_(x),p2_(p2),f_(g){}
};

/////////////// Par2_2FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class Y>
class Par2_2FncObj : public FncObj<X,Y>{

   const X x_;
   const P1 p1_;
   Y (* const f_)(X const&, P1 const&, P2 const& );

public:
   Y operator()(P2 const& p2)const{ return (*f_)(x_,p1_,p2);}
   Par2_2FncObj(Y (*g)(X const&, P1 const&, P2 const&), X const& x,
      P1 const& p1)
   :x_(x),p1_(p1),f_(g){}
};

/////////////// Par3FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class Par3FncObj : public FncObj<X,Y>{

   const P1 p1_;
   const P2 p2_;
   const P3 p3_;
   Y (* const f_)(X const&,P1 const&,P2 const&,P3 const&);

public:
   Y operator()(X const& x)const{ return (*f_)(x,p1_,p2_,p3_);}
   Par3FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&),
   P1 const& p1, P2 const& p2, P3 const& p3)
   :p1_(p1),p2_(p2),p3_(p3),f_(g){}
};

/////////////// Par3_1FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class Par3_1FncObj : public FncObj<X,Y>{

   const X x_;
   const P2 p2_;
   const P3 p3_;
   Y (* const f_)(X const&,P1 const&,P2 const&,P3 const&);

public:
   Y operator()(P1 const& p1)const{ return (*f_)(x_,p1,p2_,p3_);}
   Par3_1FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&),
   X const& x, P2 const& p2, P3 const& p3)
   :x_(x),p2_(p2),p3_(p3),f_(g){}
};

/////////////// Par3_2FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class Par3_2FncObj : public FncObj<X,Y>{

   const X x_;
   const P1 p1_;
   const P3 p3_;
   Y (* const f_)(X const&,P1 const&,P2 const&,P3 const&);

public:
   Y operator()(P2 const& p2)const{ return (*f_)(x_,p1_,p2,p3_);}
   Par3_2FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&),
   X const& x, P1 const& p1, P3 const& p3)
   :x_(x),p1_(p1),p3_(p3),f_(g){}
};

/////////////// Par3_3FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class Par3_3FncObj : public FncObj<X,Y>{

   const X x_;
   const P1 p1_;
   const P2 p2_;
   Y (* const f_)(X const&,P1 const&,P2 const&,P3 const&);

public:
   Y operator()(P3 const& p3)const{ return (*f_)(x_,p1_,p2_,p3);}
   Par3_3FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&),
   X const& x, P1 const& p1, P2 const& p2)
   :x_(x),p1_(p1),p2_(p2),f_(g){}
};

/////////////// Par4FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class Par4FncObj : public FncObj<X,Y>{

   const P1 p1_;
   const P2 p2_;
   const P3 p3_;
   const P4 p4_;
   Y (* const f_)(X const&, P1 const&, P2 const& ,P3 const& ,P4 const&);

public:
   Y operator()(X const& x)const{ return (*f_)(x,p1_,p2_,p3_,p4_);}
   Par4FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&,P4 const&),
   P1 const& p1,P2 const& p2,P3 const& p3,P4 const& p4)
   :p1_(p1),p2_(p2),p3_(p3),p4_(p4),f_(g){}
};

/////////////// Par4_1FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class Par4_1FncObj : public FncObj<X,Y>{

   const X x_;
   const P2 p2_;
   const P3 p3_;
   const P4 p4_;
   Y (* const f_)(X const&, P1 const&, P2 const& ,P3 const& ,P4 const&);

public:
   Y operator()(P1 const& p1)const{ return (*f_)(x_,p1,p2_,p3_,p4_);}
   Par4_1FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&,P4 const&),
   X const& x,P2 const& p2,P3 const& p3,P4 const& p4)
   :x_(x),p2_(p2),p3_(p3),p4_(p4),f_(g){}
};

/////////////// Par4_2FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class Par4_2FncObj : public FncObj<X,Y>{

   const X x_;
   const P1 p1_;
   const P3 p3_;
   const P4 p4_;
   Y (* const f_)(X const&, P1 const&, P2 const& ,P3 const& ,P4 const&);

public:
   Y operator()(P2 const& p2)const{ return (*f_)(x_,p1_,p2,p3_,p4_);}
   Par4_2FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&,P4 const&),
   X const& x,P1 const& p1,P3 const& p3,P4 const& p4)
   :x_(x),p1_(p1),p3_(p3),p4_(p4),f_(g){}
};

/////////////// Par4_3FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class Par4_3FncObj : public FncObj<X,Y>{

   const X x_;
   const P1 p1_;
   const P2 p2_;
   const P4 p4_;
   Y (* const f_)(X const&, P1 const&, P2 const& ,P3 const& ,P4 const&);

public:
   Y operator()(P3 const& p3)const{ return (*f_)(x_,p1_,p2_,p3,p4_);}
   Par4_3FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&,P4 const&),
   X const& x,P1 const& p1,P2 const& p2,P4 const& p4)
   :x_(x),p1_(p1),p2_(p2),p4_(p4),f_(g){}
};

/////////////// Par4_4FncObj ///////////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class Par4_4FncObj : public FncObj<X,Y>{

   const X x_;
   const P1 p1_;
   const P2 p2_;
   const P3 p3_;
   Y (* const f_)(X const&, P1 const&, P2 const& ,P3 const& ,P4 const&);

public:
   Y operator()(P4 const& p4)const{ return (*f_)(x_,p1_,p2_,p3_,p4);}
   Par4_4FncObj(Y (*g)(X const&,P1 const&,P2 const&,P3 const&,P4 const&),
   X const& x,P1 const& p1,P2 const& p2,P3 const& p3)
   :x_(x),p1_(p1),p2_(p2),p3_(p3),f_(g){}
};

#endif

template <class X1, class X2, class X3> class comp;

} // aux

/////////////////////////// class F<X,Y> ///////////////////////////////

// Defining a template class implementing the value interface out of the
// 'poor' class FncObj<X,Y> via the handle template CpmArrays::P<>

// No assumption on X
// Y has to define Y()

template <class X, class Y>
class F: public P< FncObj<X,Y> >{ // functions as a class
   // F stands for function

   typedef F<X,Y> Type;

   typedef P< FncObj<X,Y> > base;

   static Y f_const(X const& x, const Y& y0){x; return y0;}
   static Y f_default(X const& x){ x; return Y();}

public:
   virtual F<X,Y>* clone()const{ return new F(*this);}
   F<X,Y> toClnBase()const{ return *this;}
   CPM_ORDER
      // order is here a simple bookkeeping device
      // (based on position in memory space not on
      // function values, so no order relations in X
      // or Y are needed. This is sufficient for the set
      // template  S< F<X,Y> > to work.
      // Operations that deal with order and equality in Y
      // will be introduced in Fo and Fr.
   CPM_IO_V

      // neeeded for using V< <F<X,Y> >

// destructor, and assignment inherited from P<...>
// constructors
   F(FncObj<X,Y>* fop):base(fop){}
      // constructor from pointers to FncObj's. All deeper constructions
      // utilize this way. F<X,Y>(new DC(...)) where DC is a class
      // derived from FncObj<X,Y> and (...) are suitable constructor
      // arguments, is thus a correct construction of a F<X,Y> object.
      // The class DC should be immutable, i.e. all its
      // data are declared const.
      // That this is also directly possible for classes derived from
      // F<X,Y> it is essential to have this as automatic conversion
      // i.e. without an 'explicit' in front

   F(base const& g):base(g){}
      // upcast constructor
      // conversion needed that F<X,Y> is just a new name for
      // P< FncObj<X,Y> >

   F(void):base(new aux::Par0FncObj<X,Y>(f_default)){}
      // default constructor

   explicit F(Y const& y1):base(
      new aux::Par1FncObj<X,Y,Y>(f_const,y1)){}
      // constructor for constant function

// Construction from function pointers

   explicit F( Y (*f)(X const& )):
   base(new aux::Par0FncObj<X,Y>(f)){}

   explicit F( const Y (*f)(X const& )):
   base(new aux::ParcFncObj<X,Y>(f)){}

   explicit F( Y (*f)(X)):
   base(new aux::ParFncObj<X,Y>(f)){}

   explicit F( std::function<Y(X)> f):
   base(new aux::StdFunctionObj<X,Y>(f)){}
      // This C++11-facility allows the definition of F<X,Y> instances
      // from functional expressions in many variables in a convenient
      // manner: consider
      // Y f(X1 const& x1, X2 const& x2, X3 const& x3)
      // we would like to define a function X3 --> Y, x3 |--> f(a,b,x3)
      // for some a in X1 and b in X2.
      // We set:
      // #include <functional>
      // using std::bind;
      // using std::function;
      // using namespace std::placeholders;
      // function<Y(X3)> g(bind(f,a,b,_1))
      // and F<X3,Y> f3(g) or less verbose F<X3,Y> f3(bind(f,a,b,_1)).
      // In expression bind(f,a,b,_1) the placeholder is _1 and not _3
      // since it becomes the single first argument of function f3.
      // Making use of this method, we never need the classes
      // F<>,F1<>, .... F6<>
      // and
      // F1_1<>, F2_1<>, F2_2<>, F3_1<>, F3_2<>, F3_3<>, F4_1<>, F4_2<>,
      // F4_3<>, F4_4<>. Actually, the macro CPM_Fn allows to purge those
      // from the active code. With my present system the larger generality
      // of the method based on std::function and std::bind results in
      // slightly longer compilation time (10%) and longer run time (1%-2%)
      // a single test with program pala.

   //F<X,Y>& operator=( F<X,Y> const& f){ return *this=f;}
      // should not be needed
   F<X,Y>& operator=( Y (*f)(X) ){ return *this=F<X,Y>(f);}
      // added 2005-04-16. Now we can write also
      // F<R,R> sinCpm; sinCpm=sin;
      // instead of F<R,R> sinCpm(sin);
      // The = syntax is particularly convenient in arrays:
      // V< F<R,R> > fList(3);
      // fList[1]=sin; fList[2]=cos; fList[3]=exp;
      // Beware of F<R,R> sinCpm=sin; which can't be valid in one's
      // right mind
   F<X,Y>& operator=( Y (*f)(X const&)){ return *this=F<X,Y>(f);}
      // see previous function

// enabling 'function value taking' via '()' for the underlying FncObj
   virtual Y operator()(X const& x)const
   {
      return (base::operator()())(x);
      // this complicated-looking syntax is the price for using
      // operator()(void) for 'de-referencing' in 'pointer'-class P<>,
      // since the function to be defined itself is operator(), a fully
      // scope resolved name is to be used for an expression which in a
      // different scope could be simply (*this)()(x)
   }

   Y evl(X const& x)const{ return (this->base::operator()())(x);}
      // evl stands for evaluation

// turning *this into a constant-valued approximation (the most simple
// easy-to-evaluate approximating function)

   void makeConstant(X const& x0){ Y y0=(*this)(x0);*this=F<X,Y>(y0);}
      // the modified object is a constant-valued function taking
      // everywhere the value y0, which the un-modified one takes at
      // x=x0.

   void makeConstant_(X const& x0){ makeConstant(x0);}
      // proper naming for a mutating function.

// concatenation (composition) of functions.

   template< class T>
   F<X,T> operator&(F<Y,T> const& g)const
      // x |--> y=(*this)(x) |--> t=g(y)=g((*this)(x))
      // thus (f&g)(x)=g(f(x))
      // notice that f & g : x |--> f(x) |--> g(f(x))
      // first action by f, second action by g thus n o t
      // (f&g)(x) = f(g(x))    wrong  !!!!
      // but
      // (f&g)(x) = g(f(x))     right !!!!
      // If one would write x.f for f(x) one had
      // x.(f&g) = x.f.g which looks much better
      // In the literature one finds f;g for f&g
   { return F<X,T>(new aux::comp<X,Y,T>(*this,g));}

   F<X,Y>& operator&=(F<Y,Y> const& g)
   { *this=operator&(g); return *this;}
      // mutating form of &, but type should not change, thus
      // T == Y needed

   template< class T>
   F<T,Y> circ(F<T,X> const& g)const
      // t|-->x=g(t)|-->y=(*this)(x)=(*this)(g(t))
      // thus (f.circ(g))(x)=f(g(x))
      // \circ is the LATEX name for the usual function
      // composition symbol
   { return F<T,Y>(new aux::comp<T,X,Y>(g,*this));}

   F<X,Y>& circ_(F<X,X> const& g){ *this=circ(g); return *this;}
      // mutating form of circ, but type should not change, thus
      // T == X needed

   void leftCombine(F<X,X> const& fL){ *this=fL&(*this);}
      // *this is changed into fL & *this

   void rightCombine(F<Y,Y> const& fR){ *this=operator&(fR);}
      // *this is changed into *this & fR

   virtual Word nameOf()const
   {
      Word wi="F<";
      Word wx=CpmRoot::Name<X>()(X());
      Word wy=CpmRoot::Name<Y>()(Y());
      return wi&wx&","&wy&">";
   }
};

template <class X, class Y>
bool F<X,Y>::prnOn(std::ostream&)const{return false;}

template <class X, class Y>
bool F<X,Y>::scanFrom(std::istream&){return false;}

template <class X, class Y>
Z F<X,Y>::com(F<X,Y> const& f)const
{
   return base::com(f);
}

namespace aux{

template <class X1, class X2, class X3>
class comp: public FncObj<X1,X3>{
   const F<X1,X2> f1;
   const F<X2,X3> f2;
public:
   comp(const F<X1,X2>& f1_, const F<X2,X3>& f2_):f1(f1_),f2(f2_){}
   X3 operator ()(const X1& x1)const{ return f2(f1(x1));}
};

} // aux

#ifdef CPM_Fn
////////////////////// class F1<> //////////////////////////////////////

template <class X, class P1, class Y>
class F1{ // functions with one parameter
    const P1 p1_;
    typedef F1<X,P1,Y> Type;
    CPM_INVAR(F1)
public:
explicit F1( P1 const& p):p1_(p){}
    F<X,Y> operator()( Y (*f)(X const&, P1 const&) )const;
};

// We explain the usage of F2<X,P1,P2,Y> which is representative
// for F1,F2,F3,..., F6 (see cpmf.h for F5 and F6).
// F2<X,P1,P2,Y> is a tool for constructing instances of
// F<X,Y> in a flexible manner from classical C-functions
// of prototype Y (*f)(X const&, P1 const&, P2 const&). The
// syntax is as follows:
// Assume definitions
// P1 p1(...);
// P2 p2(...);
// Y fc(X const& x, P1 const& p1, P2 const& p2){....}
// Then
// F<X,Y> f=F2<X,P1,P2,Y>(p1,p2)(fc);
// ( or F<X,Y> f(F2<X,P1,P2,Y>(p1,p2)(fc));
// defines a function f for which
// f(x)==fc(x,p1,p2)
// Thus f is 'the function fc with the parameters of types
// P1 and P2 given fixed values'.
// If the fixed values p1, p2 of these parameters are to be used for
// several classical functions of identical prototype,
// say fc, gc, hc, the the following syntax is economic:
// F2<X,P1,P2,Y> f2(p1,p2);
// F<X,Y> f=f2(fc);
// F<X,Y> g=f2(gc);
// F<X,Y> h=f2(hc);

// Addition 2011-03-10
// Tools to let any of the parameter play the role of the function
// argument:
// Consider the function fc as defined above. Sometimes one needs to
// consider it as, say, a function of the parameter p1 for fixed values
// of parameters x and p2. In such a case it is inconvenient to be forced
// to define a new function which has p1 in the first argument slot such
// as by the definition
//    Y fc1(P1 const& p1, X const& x, P2 const& p2)
//    {
//       return fc(x,p1,p2);
//    }
// To avoid this need we define a class template F2_1 which considers the
// first parameter slot as providing the function argument. Of course, the
// truly first slot is taken as a parameter slot in turn. Further, we
// define a class template F2_2 which takes the second parameter slot as
// the argument slot, and the other slots as parameter slots.
// This should be usable as a pattern to explain also the class templates
// F3_1, F3_2, F3_3, F4_1, F4_2, F4_3, F4_4.
// We don't define F5_{1,...,5} and F6_{1,...,6} since already 4 parameters
// is more than I ever needed in a context in which the Fn_m were useful.

// Addition 2015-06-11
// Having in mind a general function
// Y f(X1 const& x1, ... Xn const& xn)
// one could be interested in partitioning the index set {1,...,n} in an
// arbitrary way into two disjoint sets S1 and S2 and take the xi indexed
// by the second set as parameters and those indexed by the first set as
// arguments of a 'function of many variables'. This turned out to be very
// and even after replacing my caresian products by std::tupel I found no
// solution. The problem is that in these heterotypic arrays one has no
// for-loops over all components. Only costant expressions are allowed
// to access the components. If all types X1,...Xn had a common base class
// one could use pointers for the loops and one would have no problem.
// Taking this difficulties into account, one can appreciate the fact that
// the special case card(S1)=1 which we solved  by means of the functions
// Fn_m can be treated in a clear and elementary manner.

template <class X, class P1, class Y>
F<X,Y> F1<X,P1,Y>::operator()( Y (*f)(X const&, P1 const&) )const
{
   return F<X,Y>(new aux::Par1FncObj<X,P1,Y>(f,p1_));
}

////////////////////// class F1_1<> ////////////////////////////////////

template <class X, class P1, class Y>
class F1_1{
// as F1, but the first parameter is the function argument
    const X x_;
    typedef F1_1<X,P1,Y> Type;
    CPM_INVAR(F1_1)
public:
explicit F1_1( X const& x):x_(x){}
    F<P1,Y> operator()( Y (*f)(X const&, P1 const&) )const;
};

template <class X, class P1, class Y>
F<P1,Y> F1_1<X,P1,Y>::operator()( Y (*f)(X const&, P1 const&) )const
{
   return F<P1,Y>(new aux::Par1_1FncObj<X,P1,Y>(f,x_));
}

////////////////////// class F2<> //////////////////////////////////////

template <class X, class P1, class P2, class Y>
class F2{ // functions with two parameters
    const P1 p1_;
    const P2 p2_;
    typedef F2<X,P1,P2,Y> Type;
    CPM_INVAR(F2)
public:
    F2( P1 const& p1, P2 const& p2):p1_(p1),p2_(p2){}
    F<X,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&))const;
};

template <class X, class P1, class P2, class Y>
F<X,Y> F2<X,P1,P2,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&) )const
{
   using namespace aux;
   return F<X,Y>(new Par2FncObj<X,P1,P2,Y>(f,p1_,p2_));
}

////////////////////// class F2_1<> //////////////////////////////////////

template <class X, class P1, class P2, class Y>
class F2_1{
// as F2, but the first parameter is the function argument
    const X x_;
    const P2 p2_;
    typedef F2_1<X,P1,P2,Y> Type;
    CPM_INVAR(F2_1)
public:
    F2_1( X const& x, P2 const& p2):x_(x),p2_(p2){}
    F<P1,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&))const;
};

template <class X, class P1, class P2, class Y>
F<P1,Y> F2_1<X,P1,P2,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&) )const
{
   using namespace aux;
   return F<P1,Y>(new Par2_1FncObj<X,P1,P2,Y>(f,x_,p2_));
}

////////////////////// class F2_2<> //////////////////////////////////////

template <class X, class P1, class P2, class Y>
class F2_2{
// as F2, but the second parameter is the function argument
    const X x_;
    const P1 p1_;
    typedef F2_2<X,P1,P2,Y> Type;
    CPM_INVAR(F2_2)
public:
    F2_2( X const& x, P1 const& p1):x_(x),p1_(p1){}
    F<P2,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&))const;
};

template <class X, class P1, class P2, class Y>
F<P2,Y> F2_2<X,P1,P2,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&) )const
{
   using namespace aux;
   return F<P2,Y>(new Par2_2FncObj<X,P1,P2,Y>(f,x_,p1_));
}

////////////////////// class F3<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class F3{ // functions with three parameters
    const P1 p1_;
    const P2 p2_;
    const P3 p3_;
    typedef F3<X,P1,P2,P3,Y> Type;
    CPM_INVAR(F3)
public:
    F3( P1 const& p1, P2 const& p2, P3 const& p3):
    p1_(p1),p2_(p2),p3_(p3){}
    F<X,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&,
       P3 const&) )const;
};

template <class X, class P1, class P2, class P3, class Y>
F<X,Y> F3<X,P1,P2,P3,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&) )const
{
   using namespace aux;
   return F<X,Y>(new Par3FncObj<X,P1,P2,P3,Y>(f,p1_,p2_,p3_));
}
////////////////////// class F3_1<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class F3_1{
// as F3, but the first parameter is the function argument
    const X x_;
    const P2 p2_;
    const P3 p3_;
    typedef F3_1<X,P1,P2,P3,Y> Type;
    CPM_INVAR(F3_1)
public:
    F3_1( X const& x, P2 const& p2, P3 const& p3):
    x_(x),p2_(p2),p3_(p3){}
    F<P1,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&,
       P3 const&) )const;
};

template <class X, class P1, class P2, class P3, class Y>
F<P1,Y> F3_1<X,P1,P2,P3,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&) )const
{
   using namespace aux;
   return F<P1,Y>(new Par3_1FncObj<X,P1,P2,P3,Y>(f,x_,p2_,p3_));
}

////////////////////// class F3_2<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class F3_2{
// as F3, but the second parameter is the function argument
    const X x_;
    const P1 p1_;
    const P3 p3_;
    typedef F3_2<X,P1,P2,P3,Y> Type;
    CPM_INVAR(F3_2)
public:
    F3_2( X const& x, P1 const& p1, P3 const& p3):
    x_(x),p1_(p1),p3_(p3){}
    F<P2,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&,
       P3 const&) )const;
};

template <class X, class P1, class P2, class P3, class Y>
F<P2,Y> F3_2<X,P1,P2,P3,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&) )const
{
   using namespace aux;
   return F<P2,Y>(new Par3_2FncObj<X,P1,P2,P3,Y>(f,x_,p1_,p3_));
}

////////////////////// class F3_3<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class Y>
class F3_3{
// as F3, but the third parameter is the function argument
    const X x_;
    const P1 p1_;
    const P2 p2_;
    typedef F3_3<X,P1,P2,P3,Y> Type;
    CPM_INVAR(F3_3)
public:
    F3_3( X const& x, P1 const& p1, P2 const& p2):
    x_(x),p1_(p1),p2_(p2){}
    F<P3,Y> operator()(Y (*f)(X const&, P1 const&, P2 const&,
       P3 const&) )const;
};

template <class X, class P1, class P2, class P3, class Y>
F<P3,Y> F3_3<X,P1,P2,P3,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&) )const
{
   using namespace aux;
   return F<P3,Y>(new Par3_3FncObj<X,P1,P2,P3,Y>(f,x_,p1_,p2_));
}

////////////////////// class F4<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class F4{ // functions with four parameters
    const P1 p1_;
    const P2 p2_;
    const P3 p3_;
    const P4 p4_;
    typedef F4<X,P1,P2,P3,P4,Y> Type;
    CPM_INVAR(F4)
public:
    F4( P1 const& p1, P2 const& p2, P3 const& p3, P4 const& p4):
    p1_(p1),p2_(p2),p3_(p3),p4_(p4){}
    F<X,Y> operator()(Y (*f)(X const&, P1 const&,
       P2 const&, P3 const&, P4 const&))const;
};

template <class X, class P1, class P2, class P3, class P4,
     class Y>

F<X,Y> F4<X,P1,P2,P3,P4,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&, P4 const&))const
{
   using namespace aux;
   return
   F<X,Y>(new Par4FncObj<X,P1,P2,P3,P4,Y>(f,p1_,p2_,p3_,p4_));
}

////////////////////// class F4_1<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class F4_1{
// as F4, but the first parameter is the function argument
    const X x_;
    const P2 p2_;
    const P3 p3_;
    const P4 p4_;
    typedef F4_1<X,P1,P2,P3,P4,Y> Type;
    CPM_INVAR(F4_1)
public:
    F4_1( X const& x, P2 const& p2, P3 const& p3, P4 const& p4):
    x_(x),p2_(p2),p3_(p3),p4_(p4){}
    F<P1,Y> operator()(Y (*f)(X const&, P1 const&,
       P2 const&, P3 const&, P4 const&))const;
};

template <class X, class P1, class P2, class P3, class P4,
     class Y>

F<P1,Y> F4_1<X,P1,P2,P3,P4,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&, P4 const&))const
{
   using namespace aux;
   return
   F<P1,Y>(new Par4_1FncObj<X,P1,P2,P3,P4,Y>(f,x_,p2_,p3_,p4_));
}

////////////////////// class F4_2<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class F4_2{
// as F4, but the second parameter is the function argument
    const X x_;
    const P1 p1_;
    const P3 p3_;
    const P4 p4_;
    typedef F4_2<X,P1,P2,P3,P4,Y> Type;
    CPM_INVAR(F4_2)
public:
    F4_2( X const& x, P1 const& p1, P3 const& p3, P4 const& p4):
    x_(x),p1_(p1),p3_(p3),p4_(p4){}
    F<P2,Y> operator()(Y (*f)(X const&, P1 const&,
       P2 const&, P3 const&, P4 const&))const;
};

template <class X, class P1, class P2, class P3, class P4,
     class Y>

F<P2,Y> F4_2<X,P1,P2,P3,P4,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&, P4 const&))const
{
   using namespace aux;
   return
   F<P2,Y>(new Par4_2FncObj<X,P1,P2,P3,P4,Y>(f,x_,p1_,p3_,p4_));
}

////////////////////// class F4_3<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class F4_3{
// as F4, but the third parameter is the function argument
    const X x_;
    const P1 p1_;
    const P2 p2_;
    const P4 p4_;
    typedef F4_3<X,P1,P2,P3,P4,Y> Type;
    CPM_INVAR(F4_3)
public:
    F4_3( X const& x, P1 const& p1, P2 const& p2, P4 const& p4):
    x_(x),p1_(p1),p2_(p2),p4_(p4){}
    F<P3,Y> operator()(Y (*f)(X const&, P1 const&,
       P2 const&, P3 const&, P4 const&))const;
};

template <class X, class P1, class P2, class P3, class P4,
     class Y>

F<P3,Y> F4_3<X,P1,P2,P3,P4,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&, P4 const&))const
{
   using namespace aux;
   return
   F<P3,Y>(new Par4_3FncObj<X,P1,P2,P3,P4,Y>(f,x_,p1_,p2_,p4_));
}

////////////////////// class F4_4<> //////////////////////////////////////

template <class X, class P1, class P2, class P3, class P4, class Y>
class F4_4{
// as F4, but the fourth parameter is the function argument
    const X x_;
    const P1 p1_;
    const P2 p2_;
    const P3 p3_;
    typedef F4_4<X,P1,P2,P3,P4,Y> Type;
    CPM_INVAR(F4_4)
public:
    F4_4( X const& x, P1 const& p1, P2 const& p2, P3 const& p3):
    x_(x),p1_(p1),p2_(p2),p3_(p3){}
    F<P4,Y> operator()(Y (*f)(X const&, P1 const&,
       P2 const&, P3 const&, P4 const&))const;
};

template <class X, class P1, class P2, class P3, class P4,
     class Y>

F<P4,Y> F4_4<X,P1,P2,P3,P4,Y>::operator()(Y (*f)(X const&, P1 const&,
   P2 const&, P3 const&, P4 const&))const
{
   using namespace aux;
   return
   F<P4,Y>(new Par4_4FncObj<X,P1,P2,P3,P4,Y>(f,x_,p1_,p2_,p3_));
}

#endif // CPM_Fn

} // namespace

#endif
