//? cpmframe.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_FRAME_H_
#define CPM_FRAME_H_
/*
   Purpose: Declaration of class CpmGraphics::Frame

   Recent history: 2012-01-17 initializing of ft_ done such that
   ordering according to ascending text height is done, and first
   valid index is set to 1
*/

#include <cpmimagingtools.h>
#include <cpmimg24.h>
#include <cpmfontgeneration.h>

namespace CpmGraphics{

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::posVal;
   using CpmLinAlg::Z2;
   using CpmLinAlg::R2;
   using CpmLinAlg::R4;
   using CpmArrays::V;
   using CpmArrays::X4;
   using CpmImaging::Rect;
   using CpmImaging::Color;
   using CpmImaging::WRITECOLOR;
   using CpmImaging::WHITE;
   using CpmImaging::LIGHTBLUE;

   typedef CpmArrays::V<R> Rn;
   typedef CpmArrays::V<R2> Rn2;

/*
The geometry of the virtual screen:
We assume a rectangular screen and associate with it an orthogonal
system of coordinates which we refer to as screen coordinates.
We adopt a horizontal x-axis and a vertical y-axis.
The origin is assumed to be the left lower screen corner and the scale on
the axes is chosen such that the right upper screen corner has coordinates
x=1 and y=1.

A rectangular subset of the screen (a frame) is characterized
by the coordinates (xL,yL) and (xU,yU) of the left-lower corner and
of the right-upper corner of the frame in this system of coordinates.

The screen is subdivided into pixels, which are addressable by ordered
pairs of integers, say (i,j). Unfortunately, the indexing scheme adopted
for (nearly ?) all conventional graphics functions is different from any
scheme to be used in mathematics: neither (i,j) is a discretized version
of (x,y) nor (i,j) is the matrix index of a screen element, if the screen
is considered a large rectangular matrix. Instead we have

      i_Graphics = Min (xPixel_-1, (int)(xPixel_*x))
      j_Graphics = Min(yPixel_-1,(int)(yPixel_*(1-y))

(the Min prescription comes into action only to prevent pixel indexes
xPixel_ and yPixel_, which do not occur since enumeration starts with 0
and xPixel_ and yPixel_ are the horizonal and vertical pixel numbers
respectively.). (i_Graphics=0,j_Graphics=0) is the UPPER-LEFT corner of
the screen and i_Graphics varies in x-direction (along a pixel line)
and j_Graphics varies in y-direction (along a pixel column)

Corresponding matrix indexes are:

      i_Matrix=j_Graphics+1      1 <= i_Matrix <= yPixel_
      j_Matrix=i_Graphics+1      1 <= j_Matrix <= xPixel_

Some of the member functions of Frame will use matrix indexing.

With respect to the frame, we introduce a second system of coordinates
according to the same rules used for the x,y-system, but with the screen
replaced by the frame. Let us denote these coordinates xF and yF
and refer to the coordinate system as frame system. If the frame is
considered a matrix, the matrix indexes corresponding to a pixel are
denoted (iF,jF)

new nomenclature:
symbol for matrix indexes:                      ij
symbol for graphics indexes:                    ab
symbol for continuous Cartesian coordinates:    xy
prefixes S for screen(or, better, application window), F for frame,
G for graph so: SabToSxy means: from graphics indexes referring to
the screen to (x,y) coordinates of the screen system

*/

/////////////////////////// class FrmText ////////////////////////////////
class Frame;

typedef CpmArrays::X4<R2,Word,Color,R> TextCom;
      //: text component
      // holds write-position, text, write color, font height.
      // Position is in Frame specific x,y coordinates ranging
      // from 0 to 1. Origin is lower-left. Font height is a
      // tendative pixel height from which the actual pixel
      // height gets chosen by a selection table that takes
      // int account the available fonts.
      // write-position is the position of the upper-left
      // corner of the Word's writing box.
      // Thus x=0 and y=1 marks is the first writing position if
      // a text is to be written in the usual way from left to right and
      // from top to bottom. This selection allows for no margin.

class FrmText{ //Frame text
// independent
   V<TextCom> tcs_;
      // text components, see typedef above
   typedef FrmText Type;
public:
   CPM_IO
   CPM_ORDER
   Word nameOf()const{ return Word("FrmText");}
   FrmText(V<Word> const& vw, Frame const& fr,
      V<Color> const& c, R down, R toRight=-0.15);
      // tries to place a system of text lines in the pixel
      // rectangle given by fr. For down = 0 a text for which
      // is space enough to have freedom of placement, the position will
      // be chosen in the center of availabe height. If one whishes to
      // have the position different, the argument 'down' should be set.
      // For instance, I had to use down=0.66 for placing a
      // 3-line text at the bottom of a graphics. down = -0.2 moved it
      // just below the ceiling. 2022-09-28 the last argument 'toRight'
      // for horizontal shift within the frame was added.
      // An example of usage:
      // Graph gr;
      // FrmText legend(V<Word>{"potential","wave function:"
      // ,"real part","imaginary part","absolute value"},gr,
      // V<Color>{BLUE,DARKGRAY,GREEN,RED,BLACK},0.55,-0.15);
      // gr.setText(legend);
   FrmText(V<TextCom> const& tcs= V<TextCom>(1)); // 1 was 0
   FrmText& operator+=(TextCom const& tp);
   FrmText operator+(TextCom const& tp)const;
   const TextCom& operator[](Z i)const{ return tcs_[i];}
   Z dim()const{ return tcs_.dim();}
   Z b()const{ return tcs_.b();}
   Z e()const{ return tcs_.e();}
   bool noText()const
   {
      return tcs_.dim()==1 && tcs_[1].c2().dim()==0;
   }
};

////////////////////////// class Frame ///////////////////////////////////

class Frame {
// rectangular subsets of the screen with pixel structure
   // class contains no pointers, thus no = operator is needed
   // Class R2 describes pairs of continuous coordinates
   // Class Z2 describes discrete pixel positions.
   // A frame is a rectangular subset of the screen which is made up
   // of integer pixels. Thus the x-coordinates of the corners are
   // multiples of Dx=1./xPixel_ and the y-coordinates are multiples of
   // Dy=1./yPixel_.

friend class Frames;
friend class FrmText;

// independent data:

protected: // should be available in class Graph

   static V<CpmFonts::Font> ft_;
      // List of fonts (actually 'the same font' but different sizes).
      // Will be read from file by calling iniStatics().

   static Z fontSelector(R h);
      // A text with proposed text height h will be written using
      // ft_[fontSelector(h)]

   static Z xPixel_;
      // maximum number of columns of bitmaps

   static Z yPixel_;
      // maximum number of rows of bitmaps

   static xyxy scrRec;
      //: screen rectangle
      // virtual screen rectangle, described by the new type for
      // rectangular subsets of Z2. Internal data are all Sab.

   static bool xyPixelInitialized;
      // true if initialized

   static bool fontInitialized;
      // true if initialized

   static void iniStatics(void);
      // Static member function which initializes the data members which
      // characterize virtual screen. This function, together with the
      // similar public function
      //       static void iniFonts(void);
      // has to be called as a first action in each constructor body.

   static Color defaultColorActive;
      // this color will be chosen for colorActive of a frame in the
      // absence of explicit specification

   static R xPixelInv, yPixelInv;
      // 1/xPixel and 1/yPixel for efficiency

   static R hFrac;
      // Determines the default text hight as a fraction of total
      // frame height. See functions textHeight() and textHeightAct()
      // for how this determination actually works taking into account
      // that fonts have to be taken from the font list ft_.

   static R yText;
      // proposal for Fy of the line on which a single line
      // of text will be written on default

   static Z imageCounter;
      // useful in mode CPM_NOGRAPHICS in which screen-graphics
      // will be replaced by ppm-images

// independent data

   Color colorActive;
      // If a frame is created by any constructor except the default
      // constructor it will be painted first by a color.
      // This color is just colorActive.

   R2 cL; // cL[1], cL[2] are the coordinates of the
      // lower left corner of the frame in the screen system of
      // coordinates. In comments, cL[1] and cL[2] are sometimes
      // referred to as xL, yL. Notice that frame corners
      // coincide with pixel corners (not pixel centers)

   R2 cU; // upper right corner. See cL

   mutable Img24 img;
      // memory for RGB data, added 2000-9-18
      // Whether needed or not, it is always present with a size
      // determined by the pixel count of *this.

     FrmText text_;
      // notice that this will automatically initialized
      // such that text_.dim()==1, text_[1].c2()==""

// dependent data:

   Z2 pL;
      // lower left pixel of the Frame in the screen
      // graphics pixel numbering scheme.
      // putPel(pL[1],pL[2],c) puts the color c to the
      // lower left corner of the Frame *this on screen
   Z2 pU;
      // upper right pixel of the Frame in the screen
      // graphics pixel numbering scheme.
      // putPel(pU[1],pU[1],c) puts the color c to the
      // upper right corner of the Frame *this on screen

      // Given the discrete values, the continuous cL, cU are given as
      // follows:
      // cL[1]=pL[1]*xPixelInv // should be 0 for pL[1]=0
      // cU[1]=(pU[1]+1)*xPixelInv // should be 1 for pU[1]=xPix-1
      // cL[2]=1-(pL[2]+1)*yPixelInv // should be 0 for pL[2]=yPixel-1
      // cU[2]=1-pU[2]*yPixelInv // should be 1 for pU[2]=0
         // OK for these settings

   Z nCol; // Number of pixel columns of the frame:
      // nCol=pU.x1-pL.x1+1

   Z nRow;  // Number of pixel rows of the frame:
      // nRow=pL.x2-pU.x2+1

   R nColInv;      // inverse of nCol

   R nRowInv;      // inverse of nRow

   B markMode; // mark mode, if this is true, the
      // visualization function vis() has no effect on the
      // screen image. This helps to use many of the show
      // functions in a mere marking mode, just as the more
      // recent mark functions.

 // methods:

   void iniFrame(R x, R y, R dx, R dy, Color const& color);
      // tool for defining constructors without code repetition

   void updateRowCol();
   // updates nRow, nCol, nRowInv, nColInv if cL or/and cU have been
   // changed.

   void updateFromDiscrete(void);
   // updates cL,cU,pU, nColInv, nRowInv, img if
   // nCol or/and nRow and pL have already their desired values.
   // img content will be cleared.

public:

   typedef Frame Type;
   CPM_ORDER
   CPM_IO_V  // virtual version of the I/O basic functions
      // The I/O functions
      // bool prnOn(ostream&)const
      // and
      // bool scanFrom(istream&)
      // need explanation for that class which is more strongly
      // related to the computer hardware than most of my classes.
      // The intention is this: Let a program running on one computer
      // (under Windows) a program which instantiates some Frame fr;
      // and writes the status of fr on a textfile
      // ofstream out("fr_status.txt");
      // fr.prnOn(out); // or out<<fr;
      // Then a program running on any other Windows computer using
      // an arbitrary graphics resolution should get in a program
      // saying
      // Frame fr2;
      // ifstream in2("fr_status.txt");
      // fr2.scanFrom(in2); // or in2>>fr2;
      // a graphics frame which in proportion to the whole screen has the
      // same extension in both computers. <It is not the intention of
      // this I/O operations to save the content of video memory.>
      // 2009-10-02: data element img treated as all other data elements.

   static void iniFonts(void);
      // Initializes the fonts ft_. Needs to be public since it is called
      // in main(...) of cpmapplication.cpp.
      // Reads font list from a file. Font list gets ordered according
      // to increasing text height and first valid index is forced to
      // be 1.

   void static setStaticSize(Z width, Z height);
      // the static members xPixel_, yPixel_, xPixelInv, yPixelInv,
      // scrRec get set to values determined by the arguments.
      // By the recent code revision class Frame became nearly
      // independent in functionality of the graphics hardware.
      // Only the display mode makes use of it.
      // For producing high resolution graphics,
      // (i.e with much more pixels than my screen has) it becomes
      // possible with this function to create larger frames.
      // For such large images to make sense one has to extend
      // lines to lines with thickness, since one pixel width
      // becomes too thin for some applications in say 4000x6000 pixel
      // images. The benefit of this is that aliasing does not matter
      // if pixels are so small compared to total size.

   static Z getScreenWidth(){return xPixel_;}
      //:: get screen width
      // in pixels

   static Z getScreenHeight(){return yPixel_;}
      //:: get screen height
      // in pixels

   static void setDefaultColorActive(Color const& c)
      {defaultColorActive=c;}
      // allows to influence the primary color of frames which were
      // created without a color setting

   static void set_hFrac(R x){ hFrac=x;}
      //: set hFrac

   static R get_hFrac(){ return hFrac;}
      //: get hFrac

   static void set_yText(R x){ yText=x;}
      //: set yText

   static R get_yText(){ return yText;}
      //: get yText

   static xyxy getScrRec(){ return scrRec;}
      // getting the screen rectangle in 'modern' form

   static V<CpmFonts::Font> getFont(){ return ft_;}
      //: get font

   virtual Word nameOf()const{ return Word("Frame");}
      //: name of

   xyxy getFrmRec()const{ return xyxy(xy(pL.x1,pU.x2),xy(pU.x1,pL.x2));}
      // getting the frame rectangle in 'modern' form. We don't use a
      // data member to hold this object since I have too many
      // initilization and adaption methods to be sure that this would be
      // initialized correctly in all these methods.
      // Notice that in the graphics indexing scheme it is natural to
      // describe a rectangle by left-upper corner and right lower
      // corner.
      // This is the way followed in the definition of IvZ2.
      // By contrast, pL and pU are left-lower and right upper. This
      // causes the deviation from the probably expected formula
      // return IvZ2(Z2(pL.x1,pL.x2),Z2(pU.x1,pU.x2));

   static R2 SabToSxy(Z2 p)
   {   return R2((R(p.x1)+0.5)*xPixelInv,R(1.)-(R(p.x2)+0.5)*yPixelInv);}
      // for given graphics indexes p=(a,b) (thus Sab)
      // of a screen pixel, the center of the screen pixel is
      // returned, expressed as screen coordinates (x,y) (thus Sxy)

   static Z2 SxyToSab(R2 c);
      // given coordinates of a point in the screen (x,y)-sytem
      // the graphics indexes are given of the screen pixel to which
      // the point belongs

   static R2 cToR2(C const& z){ return R2(z.real(),z.imag());}
      //: C to R2

   static C r2ToC(R2 const& p){ return C(p.x1,p.x2);}
      //: R2 to C

   R getAspRat()const{ return R(nCol)*nRowInv;}
      //: get aspect ratio
      // aspect ratio of the frame assuming square pixels.
      // Landscape format gives values larger than 1
      // Portrait < 1;
      // image width / image height

   R getAspRatInv()const{ return R(nRow)*nColInv;}
      //: get aspect ratio's inverse
      // inverse of aspect ratio

   static R screenAspRat(){ return R(xPixel_)*yPixelInv;}
      //:: screen aspect ratio

   static R screenAspRatInv(){ return R(yPixel_)*xPixelInv;}
      //:: screen aspect ratio's inverse

   void setAspRatHoldWidth(R asp);
      // change the image height till the aspect ratio is as given
      // by the argument or the frame reaches the border of the
      // screen. The lower left corner is held fixed.
      // Image content will be cleared.

   void setAspRatHoldHeight(R asp);
      // change the image width till the aspect ratio is as given
      // by the argument or the frame reaches the border of the
      // screen. The lower left corner is held fixed.
      // Image content will be cleared.

   void setAspRat(R asp, bool holdHeight=true)
   { if (holdHeight) setAspRatHoldHeight(asp);
      else setAspRatHoldWidth(asp);}
      //: set aspect ratio
      // for convenience

   void setThcLin_(bool b=true){ img.setThickLines(b);}
      //: set thick lines
      // if this is set to true, all calls to function line(R,R,R,R,rgb)
      // will create thick lines.

   void setMarkMode(bool mode){ markMode=B(mode);}
      //: set mark mode
      // If markMode is true, the visualization function vis()
      // has no effect on the  screen image. This helps to use many of the
      // show functions in a mere marking mode, just as the more
      // recent mark functions.

   bool getMarkMode()const{ return markMode;}
      //: get mark mode

   bool getThcLin()const{ return img.getThickLines();}
      //: get thick lines
      // returns the status of line thicking

   void setNewLin_(bool b=true){ img.setNewLines(b);}
      //: set new lines
      // if this is set to true, all calls to function line(R,R,R,R,rgb)
      // will call lineThinNew(R,R,R,R,rgb).

   bool getNewLin()const{ return img.getNewLines();}
      //: get new lines
      // returns the status of line definition style

   void cut(R asp, Z yPixel);
      // reduces (or enlarges) *this to a maximum subset with
      // aspect ratio asp and an height in pixels less or equal to
      // yPixel.
      // If there is room for a frame height of yPixel, then
      // the subframe is placed approximately at a central
      // position.

   explicit Frame(const Rect& rec);
      // constructor from a rectangle in the whole screen

   Frame(const Frame& fr, const Rect& rec);
      // constructor from a rectangle of the rectangle given by fr

   Rect getRectangle(void)const;
      // Returns the rectangle from which the Frame 'could have been
      // created'.

   Frame(R x, R y, R dx, R dy, Color const& color=defaultColorActive);
      // Creates a frame, the lower corner of which has approximately
      // lower left corner (x,y) and upper right corner (x+dx,y+dy).
      // The precise corners are made to agree with the pixel pattern.
      // The argument color determines the data member colorActive
      // in which the frame will be shown on screen.
      // If one want to create a frame that is not visible on screen just
      // by call of the constuctor (e.g. since the aspect ration is only
      // preliminary and has to be reset by calling function
      //       setAspRatHoldHeight(R))
      // one may proceed as follows:
      // R x,y,dx,dy = ....
      // Color cF=...
      // Frame f0;
      // f0.setxy(x,y,dx,dy);
      // f0.setAspRatHoldHeight(1);
      // f0.paint(cF);
      // f0.vis(); // only now, a visible action happens

   Frame(void);
      // creates the full frame as given by xPixel_ and yPixel_
      // with colorActive=defaultColorActive but no painting activity
      // invoked

   explicit Frame(Color const& c);
      // same as Frame(void) but colorActive=c and painting invoked

   explicit Frame(Z code);
      // Constructor for convenient placing of a frame
      // The screen is divided in quadrants denoted in matrix-style
      // 11 12
      // 21 22
      // Posible position codes are 11, 12, 21, 22, 1112 (for upper
      // half part) 2122 (lower half part), 1121 (left half part)
      // 1222 (right part)

   explicit Frame(Img24 const& im, R x=0, R y=0, bool showIt=true);
      // creates a frame which has lower-left corner (x,y)
      // ( where 0<=x<=1, 0<=y<=1) and has the pixel content of im.
      // Notice that the input image may have been obtained via the
      // getImg()-method from another Frame (or Graph).
      // If the pixel content is large, not the whole image may be
      // shown on screen upon calling vis().
      // For showIt==false, no screen display action is caused by
      // calling this constructor.
      // Did not show the image im till 2011-07-26.

   Frame(Frame& fr, Img24 const& im, R2 const& uppLeftCor=R2())
   {
      *this=fr;
      Img24 imt=im;
      Z ix0=ix(uppLeftCor.x1);
      Z jy0=jy(uppLeftCor.x2);
      img.insert(xy(ix0,jy0),im);
   }

   Z2 discrete(R2& pLower, const R2& p, R2& pUpper);
      // consider a point, which in the screen system has
      // coordinates (p.x, p.y). 0<=p.x<=1, 0<=p.y<=1.
      // The screen is partitioned into pixels in a disjoint and
      // complete manner ( a normal pixel is considerd to own its left
      // and its bottom boundary; only the pixel of the upper row and
      // the rightmost column are considered to own all its boundary)
      // Then screen point belongs to just one pixel. The return value
      // is the double index of this pixel according to the graphics
      // pixel numbering scheme. pLower is the lower-left corner of this
      // pixel and pUpper is the upper-right corner of this pixel.

   R2 getLowerLeftCorner(void)const{ return cL;}
      // Put
      // R2 c=getLowerLeftCorner;
      // Then (c.x,c.y) is the lower left frame corner in the screen
      // system of reference

   R2 getUpperRightCorner(void)const{ return cU;}
      // see getLowerLeftCorner()

   Z dim1()const{ return nCol;}
   Z dim2()const{ return nRow;}
      // uniform over Viewport, Img24, Frame
   Z dim_x()const{ return nCol;}
   Z dim_y()const{ return nRow;}
   Z w()const{ return nCol;}
   Z h()const{ return nRow;}

   Img24 getImg()const{ return img;}
      //: get image
      // This is an important method since many of the C+- image
      // creation methods primarily create a graphical image
      // particularly employing class Graph which is derived from
      // Frame. Notice that displaying this image shows it always
      // matching the left upper corner of the viewport independent of
      // the position of the Frame *this. This could be improved.

   void repImg(Img24 const& im, xy const& corner=xy())const
   { img.insert(corner,im);}
      //: replace image
      // See Img24::insert for the details

   void fitImg(Img24 const& im, bool conAspRat=true)const;
      //: fit image
      // places a suitable interpolated version of im in the imgage area
      // of *this. If conAspRat (= conserve aspect ratio) is true,
      // the aspect ratio of im is kept fix in interpolation so that
      // the image frame of *this may not be filled completely. For
      // conAspRat = false im will exactly be interpolated to the size
      // of the original img.

   Z getCol(void)const{ return nCol;}
      // get number of columns (nCol, name as in image classes)

   Z getLin(void)const{ return nRow;}
      // get number of lines (nRow, name as in image classes)

   Z2 get_pU(void)const;
      // get upper right pixel pU.

   Z2 get_pL(void)const;
      // get lower left pixel pL.

   Z a0()const{ return pL.x1;}
      // difference Sa-Fa

   Z b0()const{ return pU.x2;}
       // difference Sb-Fb

   Z2 a0b0()const{ return Z2(pL.x1,pU.x2);}

   xy cor()const{ return xy(pL.x1,pU.x2);}
      // upper-left corner

   R2 shf()const{ return R2(-pL.x1,-pU.x2);}
      //: shift
      // combines a0(), and b0() in a way that they may
      // directly be used to shift real coordinates in the way
      // one needs it to connect the graphical primitives of class
      // Frame to those of class Img24, see function fill(TriAng)

// getting x,y,dx,dy having the same meaning as the arguments in
// setxy(R,R,R,R)

   R getx(void)const{ return cL.x1;}
      // (x,y) is the lower left corner

   R gety(void)const{ return cL.x2;}

   R getdx(void)const{ return cU.x1-cL.x1;}
      // (x+dx,y+dy) is the upper right corner

   R getdy(void)const{ return cU.x2-cL.x2;}

   R4 getxy()const{ return R4(cL.x1,cL.x2,cU.x1-cL.x1,cU.x2-cL.x2);}

   void setxy(R x, R y, R dx, R dy);
      // Helps to consistently set cL, cU. (x,y) is the intended
      // left-lower corner is the screen system; dx and dy are
      // width and height of the frame respectively in units in
      // which hight and width of the screen both are 1.
      // The precise values actually taken by the frame may differ
      // slightly from these prescriptions since the final frame will
      // always fit to the screen pixel raster. Clearly, the frame
      // never will be set larger than allowed by xPixel_ and yPixel_.

   void set_xy_(R4 const& xy){ setxy(xy[1],xy[2],xy[3],xy[4]);}
      //: set xy
      // same as previous function with a single argument

   void scl_(R fac)
   { R4 x=getxy(); set_xy_(R4(x[1],x[2],x[3]*fac,x[4]*fac));}
      //: scale
      // Location of the lower left corner remains fixed and the
      // with and the height, both get multiplied by fac.

   Frame& operator+=(R2 const& shift)
      // Changes frame into a shifted copy.
      // shift[1] refers to the x-axis, the full width of the master
      // window (also referred to as 'screen' or 'application window',
      // discussed at the very beginning of this file) is 1.
      // Correspondingly for shift[2] and y-axis, and full hight.
      // If *this is shifted as to leave the master window, it is cut, so
      // that only the part within the master window survives.
      // Notice that the data for frame store and viewport should coincde
      // in order to yield a clear situation.
   {
      R dx=getdx(), dy=getdy();
      R2 cLn=cL+shift;
      setxy(cLn.x1,cLn.x2,dx,dy);
      return *this;
   }

   Frame& operator*=(R fac)
      // Changes frame into a scaled copy.
      // The center remains fixed, behavior when reaching
      // the boundary of the master window is the same as for
      // operator +=(R2).
   {
      R fac_h=fac*0.5;
      R dxh=getdx()*fac_h, dyh=getdy()*fac_h;
      R2 cC=(cL+cU)*0.5;
      R xLn=cC.x1-dxh, yLn=cC.x2-dyh;
         // gives lower left corner
      setxy(xLn,yLn,dxh*2,dyh*2);
      return *this;
   }

   void setpLpU(Z2 pLIn, Z2 pUIn);
      // Helps to consistently set cL, cU. Corners of the frame are
      // set in terms of screen pixel numbers.
      // Needed to define a precise partition of a frame into subframes
      // without any overlap or gaps (96-11-28)

   void setColorActive(Color const& c){ colorActive=c;}

   void setColor(Color const& c){ colorActive=c; img.setColor(c);}
      // for uniformity with related classes. Controls
      // the color to be used by function clear
   Color getColorActive(void)const{ return colorActive;}

   void paint(Color const& c);
      // paints *this with color. The painted screen will become visible
      // only after calling vis().

   void clear();
      // paints *this with colorActive and clears text
      // The cleared screen will become visible
      // only after calling vis().

   void fullClear();
      // paints with screen area and image area with black
      // always

   void clr_(Color c=LIGHTBLUE){clearText();paint(c); vis();}
      //: clear
      // C+- conformant name of mutating function clear

   void clr(Color c=LIGHTBLUE){ clr_(c);}
      // clear
      // old name, use clr_ instead

   Word vis(bool file=false, bool lineArtProcessing=false)const;
      //: visualize
      // Displays the Frame together with the text content given by data
      // member text_. If the first argument has value true also a copy
      // on file as a ppm-image will be made. The name of the image
      // file is frame_001.ppm, ... frame_999.ppm
      // where the name and the number are assigned automatically.
      // If the second argument has value true, a transformation of the
      // image is done before display and image filing which is useful
      // for frame contents which are B/W line art. The pixel structure
      // which in my graphics system deviates sometimes by +/- one pixel
      // from what could be considered nice and natural should become
      // more regular by this processing.
      // Dependent on the return value of Message::appRunId(), in addition
      // to the frame number, a program run ID will appended to the name.
      // The return value is the name of the file which was written, or
      // the void Word() if no image file was written.

   void imgTest()const{ img.vis();}
      //: image test

   Word write(const Word& file)const;
      //:: write
      // writing the pictorial content of *this int a image file
      // file&".ppm" in the current output directory
      // returned is the file name with the extension actually appended

//   Word wrt(Word const& file)const
//   { if (file.isVoid()) return Word(); else return write(file);}
      //: write
      // nearly as write but C+- compliant name. Does not write
      // if the file name is void

   Word wrt(Word const& fileName, R const& tWait=0.)const
   //: write
   // flexible filing and viewing facility. If tWait > 0 visualization
   // is enabled. Filing is disabled if either fileName has length zero or
   // tWait < 0. The reason for this double role of tWait is that this
   // item in typical C+- programs will be controled by an ini-file
   // wheras the filename will typically be fixed in code.
   {
      if (tWait>0) vis(false,false);
      Word res;
      if (!fileName.isVoid() && !(tWait<0)) res=write(fileName);
      cpmwait(tWait,2);
      return res;
   }

   Word write(void)const;
      //:: write
      // the image file is that of img.write()

// transformation of xF,yF coordinates into graphics pixel indices

   Z ix(R xF)const;
   Z jy(R yF)const;
// transformation between frame related continuous and discrete
// coordinates (xF,yF) and (iF,jF)

   Z2 getiFjF(const R2& xFyF)const;
      // given coordinates of a point in the frame system, the
      // matrix indexes of the pixel to which the point belongs
      // is given

   xy FxyToFab( R2 p)const
      // frame x,y coordinates to frame graphics indexes
      // which are directly related to the pixels of img
   {
      return xy(Iv(0,1).quant(p.x1,w(),true)-1,
              Iv(0,1).quant(R(1)-p.x2,h(),true)-1);
   }

   R2 getxFyF(const Z2& iFjF)const
      // for given matrix indexes of a pixel within the frame,
      // the center of this pixel is returned, expressed as
      // coordinated in the frame system
{ return R2((R(iFjF.x2)-0.5)*nColInv,R(1)-(R(iFjF.x1)-0.5)*nRowInv);}

   void addColors(){ img.addCol();}
      // graphical primitives add their colors to background
      // (transparent colors, watercolor, Aquarellfarben)

   void setColors(){ img.setCol();}
      // graphical primitives replace background colors by their
      // own ones. // (body colors, Deckfarben)

   void mixColors(R w){ img.mixCol(w);}

   void stereo();
      // anaglyphic representation

   bool getAntiAliasLines()const
      { return img.getAntiAliasLines();}

   void setAntiAliasLines(bool aal=true, R gamma=-1)
      {
         img.setAntiAliasLines(aal);
         if (aal==true){ // only then influence gamma
            if (gamma<=0) img.setGamma(Viewport::getGamma());
            else img.setGamma(gamma);
         }
      }
      // this allows creation of better line graphics
      // and increases the value of class VecGraph
      // working of antiAliasLines strongly depends on gamma

   R getGamma()const{ return img.getGamma();}
   void setGamma(R g){ img.setGamma(g);}

   FrmText getText()const{ return text_;}
   void setText(FrmText const& text){ text_=text;}
   void addText(FrmText const& text);
      // adds text components if their locations are not yet
      // set. Otherwise components of text_ get replaced
      // by components of text.

   void setText(Word const& text, R x=0.05, R y=0,
      Color c=WRITECOLOR, R h=0);
      // Notice that the left lower corner of the Frame *this is
      // (x=0,y=0) and the right upper corner is (x=1,y=1).
      // Setting y=0, or using the default y which happenes to
      // be y=0 too, enforces an automatic setting for y which
      // should be OK for a single line diagram title.
      // For y>0 and x>=0 (x,y) is the position of the left - upper
      // corner of the writing box. Thus (x,y) = (0,1) writes in contact
      // to the upper rim of the frame, which is okay if there is no
      // space to waste.
      // h<=0 and y<=0 determines the values to be used from static
      // data hFrac and yText.

   void setText(Word const& text, R2 xy,
      Color c=WRITECOLOR, R h=0)
   { setText(text,xy.x1,xy.x2,c,h);}

   R2 plcText(Word const& text, R2 const& leftUppCor,
      Color c=WRITECOLOR, Z i=2);
      //: place text
      // Notice that the left lower corner of the Frame *this is
      // (x=0,y=0) and the right upper corner is (x=1,y=1).
      // leftUppCor is the left upper corner of the rectangular writing
      // box, and the return value gives the right lower corner of it.
      // These positions are derived from the actual font metric
      // of the font Frame::ft_[i]. Thus presently i \in {1,2,3,4,5,6}

   void setText(V<Word> const& lines, V<Color> const& c, R down, R toRight=-0.1);
      //: set text
      // setting a multi-line text that fits the frame
      // See comment to function FrmText(V<Word>,Frame,V<Color>,R)
      // for the effect of argument down.
      // Notice that in order not to destroy legend texts of auto-scaled
      // graphics one should first set the text and than call a function
      // mark or show (which adds its legend to the yet existing text_).

   void setText(V<Word> const& lines, R down=0){
      Z nL=lines.dim(); V<Color> c(nL,WRITECOLOR); setText(lines,c,down);
   }
      //: set text
      // setting a multi-line text that fits the frame
      // See comment to function FrmText(V<Word>,Frame,V<Color>,R)
      // for the effect of argument down.

   void addText(const Word& text, R x=0.05, R y=0,
       Color c=WRITECOLOR, R h=0);
      //: add text
      // adds text components if their locations are not yet
      // set. Otherwise components of text_ get replaced
      // by components of text.

   void addText(Word const& text, R2 xy,
      Color c=WRITECOLOR, R h=0)
   { addText(text,xy.x1,xy.x2,c,h);}

   void clearText(){ text_=FrmText();}
      // clear
   bool noText()const{ return text_.noText();}
      // returns true if text_ has no content to display

   R textHeight()const;
      //:: text Height
      // returns a proposal for font height (in pixels) to be used in a
      // frame for text that is ment primarily as a carrier of
      // information. Tries to achieve that smaller frames use
      // the smaller fonts. Presently ~ inf(nCol,nRow) with
      // a factor of proportionaliy hFrac is a static member of
      // Frame and thus gets set in cpmframe.cpp, presently to
      // hFrac=0.0225. This function takes not into account the
      // limitations resulting from the fact that we have only
      // the font list ft_ at our disposal.

   Z textHeightAct()const{ return ft_[fontSelector(textHeight())].h();}
      //:: text height actual
      // returns the actual font height (in pixels) resulting from
      // having to choose from font list ft_.

// 'modern' form of the functions with arguments of type Sab. If these
// functions would not influence the data element img, they could be
// declared static.
// Their visible action refers to the application window and not to the
// Frame. The signature of these functions is the same as that of
// corresponding functions of Viewport and very similar to that of
// corresponding functions of Img24. In an attempt to unify the
// services of these two classes I sacrificed the variability of line
// style and line width.

   // pixel commands

   void putPixel(Z a, Z b, Color const& c)const
      // puts color c in a meaningful way to screen pixel (a,b)
      // where a,b is understood in the graphics numbering scheme
      // of the screen, thus Sab
{
    img.putPel(a-a0(),b-b0(),c);
}
    void put(Z2 const& Fij, Color const& c)const
      // puts color c in a meaningful way to screen pixel (i,j)
      // where i,j is understood as matrix indexes of the frame (frame =
      // matrix of pixels). Thus i=1,j=1 corresponds to the upper-left
      // corner of the frame
   { img.putPel(Fij.x2-1,Fij.x1-1,c);}

   void putPel(R a, R b, rgb c)const
      // action img according
   {
      img.putPel(a-a0(),b-b0(),c);
   }

   void setPel(R a, R b, rgb c)const
      // ensures color setting mode
   {
      img.setCol();
      img.putPel(a-a0(),b-b0(),c);
   }

   rgb getPel(R a, R b)const
      // action on img
   {
      return img.getPel(a-a0(),b-b0());
   }

  // line commands

   void line(R a1, R b1, R a2, R b2, rgb c)const;
      // draws a line from (a1,b1) to (a2,b2) (points included),
      // properly clipped to the frame.
      // If thcLin_=true a thick line is drawn

 // area commands

   void fillRectangle(R a1, R b1, R a2, R b2, rgb c)const
      // 1 refers the the upper left corner, 2 to the lower right
      // corner. The color is a solid fill color
   {
      img.fillRectangle(a1-a0(),b1-b0(),a2-a0(),b2-b0(),c);
   }

   void fill(TriAng const& tri, rgb c)const
      //: fill
      // filling a triangle with color c. No restriction to tri.
      // Will be clipped to the imaging window.
   {
      img.fill(tri.shf(shf()),c);
   }

   void fillEllipse(R a1, R b1, R a2, R b2, rgb c)const
      // 1 refers the the upper left corner, 2 to the lower right
      // corner of a minimal rectangle that contains the ellipse.
      // notice that circles are special cases which are not very useful
      // since different scales on the axes make them ellipses anyway.
   {
      img.fillEllipse(a1-a0(),b1-b0(),a2-a0(),b2-b0(),c);
   }

   void fillDisk(R ac, R bc, R r, rgb c)const
      // 1 refers the the upper left corner, 2 to the lower right
      // corner of a minimal rectangle that contains the ellipse.
      // notice that circles are special cases which are not very useful
      // since different scales on the axes make them ellipses anyway.
   {
      img.fillDisk(ac-a0(),bc-b0(),r,c);
   }


// Addition 2001-4-26:
// pixel commands referring to frame indexing (as opposed to screen
// indexing. A pixel put loop over the frame is most conveniently written
// as (notice the start with 1)
//    for (Z x=1;x<=dim_x();x++){
//       for (Z y=1;y<=dim_y();y++){
//          Color c=... x...y...;
//          put_xy(x,y,c);
//       }
//    }
// see implementation of void Graph::show(const F<C,C>&,R)
// for actual usage (addition 2001-4-26)
//
   void put_xy(Z x, Z y, Color const& c)const
   {
      putPel(x+pL.x1-1,pL.x2-y+1,c);
   }

   rgb get_xy(Z x, Z y)const{ return getPel(x+pL.x1-1,pL.x2-y+1);}

// consider
// Img24 img=...;
// Frame fr;
// img.vis(fr.context(),10,10);
// This will show the image on a pixel by pixel basis with an upper left
// corner placed at graphical position 10,10 of the screen (not the
// window of frame).

   xyxy restrict(xyxy const& rec)const;
      // returns the part of the screen rectangle rec, that belongs to
      // *this in a way that it is directly a subrectangle of the pixel
      // matrix of img.
      // We look at the result as generated in two steps:
      // First *this and rec are considered as rectangles within the
      // screen pixel matrix. We form the set section and obtain a new
      // screen rectangle (which may be void). Then we express the
      // coordinates of this rectangle as relative to the upper left
      // corner of *this and obtain what we want.

   R markVVC(V<V<C>> const& mat, bool matrixstyle=false, Iv range=Iv(),
      Z ipolMethod=0, bool faithful=true, bool periodic=true,
      R secFac=-1)const;

// Till 2020-02-02 this was void Graph::mark(V<V<C>>,Iv,Z,bool,bool,R,bool)
// In an attempt to modify the code (by providing a useful return value,
// not perfectly clear whether this was the only change) I realized that
// it did not use any functionality of Graph which was not yet there in the
// base class Frame.
// Using the name mark also in the base class would be 'function
// overloading accross scopes' which to my surprise is not allowed in C++.
// To make the function easy to use also within Graph I gave it an modified
// name.
//
// An efficient way of Frame to graphically represent a rectangular
// field of data.
// Interpolating representation of a matrix 'mat' of complex numbers.
// The indexing of the matrix is as it arrose in 2-dimensional
// quantum mechanical applications: mat[i][j] is such that the first
// index i grows from 0 to mat.dim()-1 along the graphical x-axis
// (i.e. horizontally from left to right). And the second indey j
// grows from 0 to mat[0].dim()-1 along the y-axis on screen
// (i.e. vertically from bottom to top, sic!). This correspondence
// between matrix indexes and graphical position holds for the
// default value matrixstyle=false. For matrixstyle=true, mat will be
// represented as a mat.dim() times mat[0].dim() matrix with the
// first matrix index growing vertically downwards and the second
// matrix index growing horizontally to the right. Especially
// mat[0][0] is shown at the upper left screen corner.
// The parameter 'range' has the effect that the complex number z gets
// represented by the color Color(z,range), see cpmimagingtools.h class
// Color. If range is the void interval (default), the actual values
// of mat will be used to define a suitable range for representation.
// Notice that Color::gamma has an influence on Color(z,range) and
// thus on the functionality of the present function.
// Parameter 'ipolMethod' has to be either 0 or 1. 0 for pixelation,
// 1 for bilinear interpolation.
// Interpolation at the boundaries is controlled by parameter
// 'periodic'. If it has value true, we use periodic continuation of the
// data in mat which is the correct method for my quantum mechanics
// application. If it has value false the boundary interpolation data
// are taken from a  continuation 'as constant'.
// This is OK for arbitrary content of mat.
// The interpolation methods here are appropriate only if mat has
// not (much) more pixels than *this. This is the common situation
// in modelling, where mat is hard to compute and therefore has typically
// not more than 100 x 100 pixels and should be displayed on screen
// on a larger window for convenient visual presentation.
// If the argument 'faithful' has the value 'true', the actually painted
// area has the aspect ratio as prescribed by mat, and fills the
// frame of *this maximally while respecting this boundary condition.
// For value false, the whole frame of *this will be used for the
// representation of the data; this means that the aspect ratio
// gets set from *this.
// Addition 2007-09-26: 'first run auto adjustment' controlled
// by the newly introduced last argument 'secFac'.
// For secFac>0 the range will set in the first call to
//       secFac*Max{ mat[i][j].abs() | i,j}
// and then used in further calls.
// The return value is secFac*Max{ mat[i][j].abs() | i,j}.
// Addition 2021-05-29: The index ranges of mat may also start with 1
// instead of 0 (as assumed in the previous description). Setting the index
// range to one which starts with 0 is done internally anyway.
};

// tool for autoscaling

Z lad_4(R a, R b, Z n, R2& x);
   //: ladder
   // Changes the fourth argument, as the name indicates.
   // Assure a<b. The return-value m and xl=x.x1, xu=x.x2 are determined
   // such that
   // (1) The numbers xl+k*(xu-xl)/m, k=0,1,2,...m are simple numbers in
   // decimal scientific notation.
   // (2) xl<=a, b<=xu, both inequalities not too far from equality.
   // (3) m not too far from input-value n.
   // So, the function creates a suitable axis division for data ranging
   // from a to b.

}  // namespace

#endif
