//? cpmframes.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmframes.h>
#include <cpmvectormore.h>

using namespace CpmArrays;
using namespace CpmGraphics;

namespace{
   I_Vector chain(Z start, Z end, Z n)
   // a integer-valued chain starting with start ending at end
   // having n members and being equidistant as good as
   // integer-valuedness allows
{
   Word loc("I_Vector chain(Z start, Z end, Z n)");
   I_Vector res(n);
   cpmassert(n>1 && end > start,loc);
   R step=R(end-start)/R(n-1);
   res[1]=start;
   res[n]=end;
   for (Z i=2;i<n;i++) res[i]=cpmrnd(R(start)+step*R(i-1));
   return res;
}

   I_Vector chain2(Z start, Z end, const Vo<R>& div)
   // a integer-valued chain starting with start ending at end
   // and growing according to the sequence div which normally will
   // start at 0 and will end with 1. If not, start and end may not be
   // reached.
   // 0<=div[i]<=1 assumed. Order not needed since made internally.
{
   Word loc("I_Vector chain2(...)");
   Z d=div.dim();
   cpmassert(d>1,loc);
   cpmassert( end > start,loc);
   V<R> divO=div.increasingCopy();
   cpmassert(divO[1]>=0.,loc);
   cpmassert(divO[d]<=1.,loc);
   V<R> divD(d);
   R dx=R(end-start);
   Z i,val;
   for (i=1;i<=d;i++) divD[i]=R(start)+dx*divO[i];
   I_Vector res(d);
   for (i=1;i<=d;i++){
      val=cpmrnd(divD[i]);
      if (val<start) val=start;
      if (val>end) val=end;
      res[i]=val;
   }
   return res;
}

}

/////////////// implementation of class Frames ///////////////////////////

Frames::Frames(const Frame& f, Z m, Z n, R fill):fr_(f),vv_(m,n,fr_)
   // The frame f will be subdivided into a (m,n)-matrix
   // of non-overlapping subframes
{
   fr_.vis(); // experiment
   Z2 fpL=fr_.pL;
   Z2 fpU=fr_.pU;
   R cFill=R(1)-fill;
   Z bx=cpmrnd((cFill*fr_.nCol)/(2.*n));
   Z by=cpmrnd((cFill*fr_.nRow)/(2.*m));
   I_Vector horiz=chain(fpL.x1,fpU.x1,n+1);
   I_Vector vert=chain(fpU.x2,fpL.x2,m+1);

   Z i,j;
   Z2 pl,pu;

   for (i=1;i<=m;i++){
      pu.x2=vert[i]+by;
      pl.x2=vert[i+1]-1-by;
      for (j=1;j<=n;j++){
         pl.x1=horiz[j]+bx;
         pu.x1=horiz[j+1]-1-bx;
         (vv_[i][j]).setpLpU(pl,pu);
      }
   }
}

Frames::Frames
(const Frame& f, const Vo<R>& horizDiv, const Vo<R>& vertDiv, R fill):
fr_(f),vv_(vertDiv.dim()-1,horizDiv.dim()-1)
   // The frame f will be subdivided into a matrix
   // of non-overlapping subframes where the subdivision is given by the
   // real lists horizDiv, vertDiv
   // 0<= horizDiv[i]<=1; 0<= vertDiv[i]<=1;
   // List need not be ordered since done internally
{
   fr_.vis(); // experiment
   Z2 fpL=fr_.pL;
   Z2 fpU=fr_.pU;
   Z m=vertDiv.dim()-1;
   Z n=horizDiv.dim()-1;
   R cFill=R(1)-fill;
   Z bx=cpmrnd((cFill*fr_.nCol)/(2.*n));
   Z by=cpmrnd((cFill*fr_.nRow)/(2.*m));
   I_Vector horiz=chain2(fpL.x1,fpU.x1,horizDiv);
   I_Vector vert=chain2(fpU.x2,fpL.x2,vertDiv);
   Z i,j;
   Z2 pl,pu;

   for (i=1;i<=m;i++){
      pu.x2=vert[i]+by;
      pl.x2=vert[i+1]-1-by;
      for (j=1;j<=n;j++){
         pl.x1=horiz[j]+bx;
         pu.x1=horiz[j+1]-1-bx;
         (vv_[i][j]).setpLpU(pl,pu);
      }
   }
}

Frame& Frames::operator()(Z i)
{
   Z2 mtl=linearToMatrix(i,vv_.dim1(),vv_.dim2());
   return vv_[mtl.x1][mtl.x2];
}

Frame const& Frames::operator()(Z i)const
{
   Z2 mtl=linearToMatrix(i,vv_.dim1(),vv_.dim2());
   return vv_[mtl.x1][mtl.x2];
}

Frames::Frames(Frame const& f, Z n, R fill):fr_(f)
{
   Z2 vzn=CpmAlgorithms::genSqrRoot(n);
   *this=Frames(f,vzn.x1,vzn.x2,fill);
}

void Frames::vis()const
{
   Z m=vv_.dim1();
   Z n=vv_.dim2();
   Word idle;
   for (Z i=1; i<=m; ++i){
      for (Z j=1; j<=n; ++j){
         idle=vv_[i][j].vis();
      }
   }
}

Frame Frames::toFrm()const
{
   Frame res(fr_);
   Z m=vv_.dim1();
   Z n=vv_.dim2();
   for (Z i=1; i<=m; ++i){
      for (Z j=1; j<=n; ++j){
         res.repImg(vv_[i][j].getImg(),vv_[i][j].cor());
      }
   }
   return res;
}

bool  Frames::prnOn( ostream& str) const
{
   cpmwat;
   cpmp(fr_);
   cpmp(vv_);
   return true;
}

bool  Frames::scanFrom (istream& str)
{
   cpms(fr_);
   cpms(vv_);
   return true;
}

Z Frames::com(Frames const& s)const
{
   if (fr_<s.fr_) return 1;
   if (fr_>s.fr_) return -1;
   if (vv_<s.vv_) return 1;
   if (vv_>s.vv_) return -1;
   return 0;
}


