//? cpmgreg.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


// Inserting diagnostic messages in function bodies of this file may causes
// problems.
// Functions may be called in a situation where the log file streams are
// not yet created.

#include <time.h>
#include <iomanip>

#if !defined(WIN32)
   #include <stdlib.h>
#endif

#include <cpmgreg.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::Word;
using CpmRoot::toWord;
using CpmRoot::toDouble;

using CpmTime::Greg;
using CpmTime::TimeStyle;

using namespace CpmStd;

R Greg::mjd_(Z y, Z m, Z d)
{
   if (m<=2){
      m+=12;
      y-=1;
   }
   R a=10000.*y+100.*m+d;
   Z b;
   if (a<=15821004.1){
      b=-2+(y+4716)/4-1179;
   }
   else{
      b=y/400-y/100+y/4;
   }
   a=365.0*y-679004.0;
   R res=a+b+floor(30.6001*(m+1))+d;
   return res;
}

Z Greg::trunc(R x)
{
   return CpmRoot::toZ(x,true);
}

void Greg::caldat(Z mjd, Z& yy, Z& mm, Z& dd)
{
   Z b,d,f;
   R jd,jd0,c,e;
   jd=2400000.5+mjd;
   jd0=trunc(jd+0.5);
   if (jd0<2299161.0){
      b=0;
      c=jd0+1524.0;
   }
   else{
      b=trunc((jd0-1867216.25)/36524.25);
      c=jd0+(b-trunc(b/4))+1525.0;
   }
   d=trunc((c-122.1)/365.25);
   e=365.0*d+trunc(d/4);
   f=trunc((c-e)/30.6001);
   dd=trunc(c-e+0.5)-trunc(30.6001*f);
   mm=f-1-12*trunc(f/14);
   yy=d-4715-trunc((7+mm)/10);
}

Z Greg::yearOffset=2000;

const R Greg::mjdOffset=2400000.5;

const R Greg::hour=1./24;

const R Greg::minute=Greg::hour/60;

const R Greg::second=Greg::minute/60;

void Greg::update(void)
{
   mjd=mjd_(y,m,d)+R(h)*hour+R(min)*minute+s*second;
}

void Greg::makeDate(void)
{
   Z mjdInt=cpmtoz(mjd);
   R rest=mjd-mjdInt;
   caldat(mjdInt,y,m,d); // now y,m,d are valid
   R h1=rest*24;
   h=trunc(h1);
   R min1=(h1-h)*60;
   min=trunc(min1);
   s=(min1-min)*60;
}

Greg::Greg(void)
{
   y=2000;
   m=1;
   d=1;
   h=0;
   min=0;
   s=0.;
   mjd=51544;
}

Greg::Greg(Z yy, Z mm, Z dd):y(yy),m(mm),d(dd),h(0),min(0),s(0)
{
   if (m<1 || m>12) cpmerror("Greg","month must be 1,...,12");
   update();
   if (d<1 || d>28) makeDate();
}

Greg& Greg::nextMonth_()
{
   m++;
   if (m>12){
      y++;
      m=1;
   }
   update();
   if (d>28) makeDate(); // this makes a valid date even when called
   // e.g. for March 31, where April 31 would not be a correct
   // date.
   return *this;
}

bool Greg::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(y);
   cpmp(m);
   cpmp(d);
   cpmp(h);
   cpmp(min);
   cpmp(s);
   return true;
}

bool Greg::scanFrom(istream& str)
{
   cpms(y);
   cpms(m);
   cpms(d);
   cpms(h);
   cpms(min);
   cpms(s);
   update();
   return true;
}

Word Greg::getDate(bool verbose)const
{
   using std::setw;
   using std::setfill;
   ostringstream ost;
   if (verbose) ost<<"yyyy-mm-ddThh:mm:ss ";
   ost<<setw(4)<<setfill('0')<<y<<"-"
      <<setw(2)<<setfill('0')<<m<<"-"
      <<setw(2)<<setfill('0')<<d<<"T"
      <<setw(2)<<setfill('0')<<h<<":"
      <<setw(2)<<setfill('0')<<min<<":"
      <<setw(2)<<setfill('0')<<toDouble(s)<<"Z";
   return Word(ost.str());
}

Word Greg::getCode(Z n)const
{
   using std::setw;
   using std::setfill;
   ostringstream ost;
   ost<<setw(4)<<setfill('0')<<y
      <<setw(2)<<setfill('0')<<m
      <<setw(2)<<setfill('0')<<d
      <<setw(2)<<setfill('0')<<h
      <<setw(2)<<setfill('0')<<min
      <<setw(2)<<setfill('0')<<toDouble(s);
   return Word(ost.str()).tail(n);
}

bool Greg::readFormatted(istream& inStream)
{
   cout<<endl<<
   "Entering a point in time in terms of the Gregorian calendar";
   R d,h;
   cout<<"\n Enter the date in the format yy.mmdd or yyyy.mmdd:  ";
   inStream>>d;
   cout<<"\n Enter the time in the format hh.mmss:  "; inStream>>h;
   setDate(d);
   setWatch(h);
   update();
   return true;
}

bool Greg::writeFormatted(ostream& str)const
{
   Word yS=Word::write(y,4);
   Word mS=Word::write(m,2);
   Word dS=Word::write(d,2);
   Word GregS=yS&"-"&mS&"-"&dS&" ";
   Word hS=Word::write(h,2);
   Word minS=Word::write(min,2);
   Word sS=toWord(s,"%2.1f");
   Word watchS=hS&":"&minS&":"&sS&" Z";
   Word res=GregS&watchS;
   str<<res;
   //return (str!=0);
   if (!str) return false;
   else return true;
}

R CpmTime::operator - (const Greg& g1, const Greg& g2)
{
   return g1.mjd-g2.mjd;
}

Greg& Greg::operator +=(R t)
{
   mjd+=t;
   makeDate();
   return *this;
}

Greg& Greg::operator -=(R t)
{
   mjd-=t;
   makeDate();
   return *this;
}

void Greg::setDate(R date, bool yShort)
{
   const R eps=0.0001;
   Z sig=(date >=0. ? 1 : -1);
   date*=sig; // now date is positive
   y=cpmtoz(date);
   date-=y; // fractional part of the absolute value of date
   y*=sig; // now the year carries its sign
   if (yShort){
      if (y<=99) y+=yearOffset;
   }
   date*=100;
   m=cpmtoz(date+eps);
   date-=m;
   date*=100;
   d=cpmtoz(date+eps);
   h=0;
   min=0;
   s=0.;
   update(); // added 99-11-27
}

void Greg::setWatch(R watch)
{
   const R eps=0.0001;
   if (watch<0) cpmerror("Greg::setWatch(R_): argument should be >=0");
   h=cpmtoz(watch+eps);
   watch-=h; // fractional part of the absolute value of watch
   watch*=100;
   min=cpmtoz(watch+eps);
   watch-=min;
   watch*=100;
   s=watch;
   update(); // added 99-11-27
}

Word Greg::getWeekDay(void)const
{
   Z weekDay=cpmtoz(getMJD());
   weekDay=weekDay%7;
   Word res;
   switch (weekDay){
      case 0: res="Wendsday";break;
      case 1: res="Thursday";break;
      case 2: res="Friday";break;
      case 3: res="Saturday";break;
      case 4: res="Sunday";break;
      case 5: res="Monday";break;
      case 6: res="Tuesday";break;
   }
   return res;
}

void Greg::now( void)
{
   static bool firstrun=true;
   if (firstrun){
   #if defined(WIN32)
      _putenv("TZ=GMT0");
      _tzset();
   #else
      char tz[7]={'T','Z','=','G','M','T','0'};
      putenv(tz);
      tzset(); // needs stdlib.h
   #endif
      firstrun=false;
   }
   time_t t;
   time(&t);
   tm* all;
   all=gmtime(&t);
   // due to putenv() this gives the normal conversion
      // without a time zone shift
   s=all->tm_sec;
   min=all->tm_min;
   h=all->tm_hour;
   d=all->tm_mday;
   m=1+all->tm_mon;
   y=all->tm_year; // comes out as 99 for 1999, and 100 for 2000

   if (y>=90 /* && y<=99 */) y+=1900; // Windows NT gives 100 in 2000
      // this means that the rule to add 1900 is not changed
    update();
    if (cpmverbose>=4){
        cout<<endl<<"s="<<s;
        cout<<endl<<"min="<<min;
        cout<<endl<<"h="<<h;
        cout<<endl<<"d="<<d;
        cout<<endl<<"m="<<m;
        cout<<endl<<"y="<<y;
        cout<<endl<<"mjd="<<mjd;
    }
}

R Greg::toMJDTD(const TimeStyle& ts)const
{

   R shift=ts.lambdaOfTimeMeridian.toDeg(0);
    shift/=360.; // now it is in days
    R res=mjd;
    res+=shift; // now res can be compared with MJD of UT or TD
        // i.e. there is no longer a time zone contribution
    return res;
}

void Greg::setMJD(R mjdI, const TimeStyle& ts)
{
    mjd=mjdI;

   R shift=ts.lambdaOfTimeMeridian.toDeg(0);
    shift/=360.; // now shift is in days
    mjd-=shift; // now referring to the local meridian
    makeDate();
}

Word Greg::getCodeWord(Z n, Z p)
{
   now();
   R f=pow(10.,p);
   Z cod=cpmtoz(mjd*f);
   Word res=cpmwrite(cod);
   return res.tail(n);
}
