//? cpmgreg.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_GREG_H_
#define CPM_GREG_H_

/*

    Functions for handling Gregorian calendar dates

*/

#include <cpmangle.h>

namespace CpmTime{ // Quantities related to calendar, date, time, clock

using namespace CpmStd;
//using CpmSystem::R_;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::Word;
using CpmGeo::Angle;

////////////////////////// enum TimeScale ////////////////////////////////

enum TimeScale {TD, UT}; // presenly not used in this general-purpose
   // (non-astronomical) version of Greg in order not to need the more
   // astronomical files.

   // TD = Temps dynamique, UT = universal time

/////////////////////////////// struct TimeStyle /////////////////////////

struct TimeStyle{ // how to interprete time with respect to the globe
   TimeScale tsc;
   Angle lambdaOfTimeMeridian;
   TimeStyle(void):tsc(UT),lambdaOfTimeMeridian(Angle()){}
   TimeStyle(TimeScale s, const Angle& a):tsc(s),
      lambdaOfTimeMeridian(a){}
};

class Greg;
R operator -(const Greg& d1, const Greg& d2);

////////////////////////////// class Greg ////////////////////////////////

class Greg{  // time, date, and Gregorian Calendar
   // An instance of class Greg determines a point in time
   // coded according to the Gregorgian calendar (and according to the
   // Julian calendar for dates prior to the introduction of the
   // Gregorian calendar). No reference to a particular time scale (UT
   // or TD) or to a time zone (e.g. MEZ or MESZ ) is implied

private:

   static Z yearOffset;
      // The 'century' which is understood for 2 decimal year input
      // e.g. 2000 for dates in the 21th century
      // May be changed via setCentury(). Presently initialized as 2000.

   static const R mjdOffset;
      // JD=MJD+mjdOffset, mjdOffset=2400000.5

protected:

// primary (independent) information
   Z y;         // year
   Z m;         // month
   Z d;         // day
   Z h;         // hour
   Z min;       // minute
   R s;         // second

// secondary (dependent) information (function of the primary
// information)
   R mjd;         // analog to the MJD without reference to a concrete
               // time scale such as UT or TD
   void update(void);
      // updates dependent information (i.e. mjd)

   void makeDate(void);
      // adjusts the rest of data members to present value of mjd

public:

   typedef Greg Type;
   CPM_IO
   CPM_NAM_V(Greg)
   virtual Greg* clone()const{ return new Greg(*this);}

// parts of the day (the unit) which are often needed

   static const R hour;
   static const R minute;
   static const R second;

   static R mjd_(Z y, Z m, Z d);
      // Modified from Montenbruck, Pfleger: Astronomy on the Personal
      // Computer,
      // 4th edition, p. 15
      // Notice that mjd_==0 <==> y=1858 m=11 d=17

   static Z trunc(R x);
      // See O. Montenbruck, Grundlagen der Ephemeridenrechnung S. 49

   static void caldat(Z mjd, Z& yy, Z& mm, Z& dd);
      // From Montenbruck, Pfleger: Astronomie mit dem Personal Computer
      // p. 13
      // Translated from PASCAL to C

   void setCentury( Z century)
      // setCentury(20) sets the value to be added to a short
      // year, such as 95, to 1900, thus resulting in 1995.
      // Notice that this coincides with the conventional enumeration
      // of centuries.
      // Century should be positive, we are not interested in using
      // Gregorian Calendar in times much prior to its introduction !
      {
         yearOffset=(century-1)*100;
      }

   Greg(void);
      // Default constructor, creates 2000-1-1 0:0:0

   Greg(Z year, Z mm, Z dd);
      // mm has to be in the range [1,12]
      // dd is arbitrary (e.g. 94-12-31 = 95-1-0 = 95-2-(-31) )
      // if year<99 it will be understood as year+yearOffset.

   explicit Greg(R date){ setDate(date,false);}
     // date is assumed to be of the form year.mmdd
     // where year is the integer year number e.g. 2006 or
     // 1905 (no shortened numbers like 50 for 1950
     // or 6 for 2006 !) Only the first 4 decimal places get
     // used (no fractions of a day).

   Greg(R date, R watch){ setDate(date,false);setWatch(watch);}
      // Creates a Greg instance from date=YYYY.MMDD and
      // watch=HH.MMSSsss... (see setWatch() for details).

// getting calendar data, always clear from code

   Z getYear(void)const{ return y;}

   Z getMonth(void)const{ return m;}

   Z getDay(void)const{ return d;}

   Z getHour(void)const{ return h;}

   Z getMinute(void)const{ return min;}

   R getSecond(void)const{ return s;}

   R getMJD(void)const{ return mjd;}

   R getJD(void)const{ return mjd+mjdOffset;}

   Word getWeekDay(void)const;

   R yearsSince1900(void)const{ return (mjd-33281.)/R(365.2422)+50.;}
   R yearsSince2000(void)const{ return (mjd-51544.)/365.2422;}
   R years(void)const{ return yearsSince2000()+2000;}
      // Convenient unit for time axes which go over longer spans of time
      // and which are related to history.

   Word getDate(bool verbose=true)const;
      // "yyyy-mm-ddThh:mm:ssZ" ( ISO6801 time stamp )
      // a compact representation of a date in which the newline commands
      // of printOut() would cause trouble

   Word getCode(Z n=14)const;
      // returns a characterization of a point in time which
      // is compact enough to be used as an appendix to file names.
      // It consists of the n trailing characters of the string
      // yyyymmddhhmmss. So the default value for n means that
      // a condensed form of the date will be given. For
      // n=8 one gets a characterisation of the time within the
      // present day.

   Word getCodeWord(Z n, Z p=5);
      // returns a sequence of n digits which is determined by the
      // instance of call and may be used as random signature
      // method: multiply mjd by 10^p, form the integer part
      // and take the last n digits of the result. Thus, the larger
      // p, the faster the result changes with time. The frequency
      // of change is 10^p per day. Thus p=5 means 10^5 changes
      // in 86400 s, i.e. 1.157 changes per second. This is a
      // reasonable time resolution for differentiating between
      // susequent runs of a program.

// setting Greg according to numbered days input (Julian date and
//   modified Julian date

   void setMJD(R mjd){ this->mjd=mjd; makeDate();}
      // adjusts *this according to MJD input.
      // In constructor form, this function would enable automatic
      // conversion from Rd to Greg and finally would lead to
      // Greg + Greg being defined (since Greg + R is defined). This
      // is not desirable, only Greg - Greg is OK.

   void setMJD(R mjd, const TimeStyle& ts);
      // adjusts *this according to MJD input. The correspondence
      // between what the function does and the values
      // UT or TD of the component ts.tsc is logical if we
      // consider mjd as referring to TD. Then ts.tsc=UT means that
      // a deltaT subtraction has to be done to get a result in
      // UT or the civil time of a time zone specified by
      // ts.lambdaOfTimeMeridian

   void setJD(R jd){ mjd=jd-mjdOffset; makeDate();}
      // adjusts *this according to JD input.

   R toMJDTD(const TimeStyle& ts)const;
      // returns MJD (as TD) from the Greg instance *this where ts
      // tells how to interprete *this
      // (e.g. as civil time of a particular time zone)

   void now(void);
      // sets *this according to the computer's idea of Greenwich
      // mean time (Z, GMT0) at the moment of function call. Notice that
      // shifting this point in time according to needs can easily be done
      // by the arithmetics of Greg, so that no shift argument is needed
      // for this function.
      // e.g.
      // TimeStyle ts=...;
      // Greg g;
      // g.now();
      // g+=10*Greg::minute;
      // R t=g.toMJDTD(ts);
      // Earth e;
      // e.update(t);
      // will represent the state of the earth 10 minutes after the
      // function call if ts properly reflects the characteristics of the
      // computer clock

friend R operator -(const Greg& d1, const Greg& d2);
      // time span in days, notice that a corresponding + operator
        // makes no sense

   Greg& operator +=(R t);
      // shifting a point in time by t days

   Greg& nextMonth_();
      // shifting a point in time to the same day, minute, second
      // but one month in calander later (shift may be 28 days,
      // 30 days or 31 days according to the calender month we are
      // in). Name ends in '_' since this is a mutating function.

   Greg& nextDay_(){ mjd+=1.; makeDate(); return *this;}
      // shifting a point in time to by one day
      // in). Name ends in '_' since this is a mutating function.

   Greg& operator -=(R t);
      // shifting a point in time by -t days

   Greg operator +(R t){ Greg res=*this; return res+=t;}

   Greg operator -(R t){ Greg res=*this; return res-=t;}

   void setDate(R date, bool yShort=true);
      // Sets the values of y, m, d from date=Year.MMDD .
      // If yShort=true and Year<99 the integer number
      // yearOffset will be added to y.

   void setWatch(R watch);
      // Sets the values of h, min, s from watch=HH.MMSSsss...
      // where SS.sss... is a number of seconds possibly together with
      // decimal fractions of a second.

   bool readFormatted(istream& inStream);
   bool writeFormatted(ostream& outStream)const;
};

} // namespace

#endif
