//? cpminterval.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpminterval.h>
#include <cpmalgorithms.h>
#include <cpmrfunc.h>

using CpmGeo::Iv;
using CpmGeo::DiIv;
using CpmGeo::Iv2;

using CpmArrays::V;
using CpmArrays::Vo;
using CpmArrays::T2;
using CpmArrays::X2;
using CpmArrays::IvZ;
using CpmFunctions::F;
#ifdef CPM_Fn
using CpmFunctions::F2;
#else
using std::bind;
using std::placeholders::_1;
#endif
using CpmRoot::R;
using CpmRoot::Word;
using CpmRoot::Z;
using CpmRoot::B;
using CpmLinAlg::R2;
using CpmLinAlg::Z2;
using namespace std;


/////////////////////////// class Iv ///////////////////////////////////

#define CPM_SHC\
   center=(left+right)*R(0.5);\
   halfWidth=(right-left)*R(0.5);\
   absInv=(halfWidth>0 ? R(0.5)/halfWidth : R(0));

Iv::Iv(IvZ const& ivz, R const& d)
{
   cpmassert(d>0.,"Iv(IvZ,R)");
   cpmassert(ivz.car()>0.,"Iv(IvZ,R)");
   R dh=d*R(0.5);
   left=d*ivz.inf()-dh;
   right=d*ivz.sup()+dh;
   CPM_SHC
}

void Iv::update(){
   CpmRootX::order<R>(left,right);
   CPM_SHC
}

void Iv::updateLight(){
   CPM_SHC
}

#undef CPM_SHC

Iv::Iv(const V<R>& vr)
{
   T2<R> p=Vo<R>(vr).infSup();
   left=p.c1();
   right=p.c2();
   updateLight();
}

Iv Iv::operator|(const Iv& iv)const
{
   R amin=CpmRootX::inf<R>(left,iv.left);
   R amax=CpmRootX::sup<R>(right,iv.right);
   return Iv(amin,amax,true); // fast return
}

Iv Iv::operator&(const Iv& iv)const
{
   if (right<=iv.left) return Iv(); // open intervals disjoint
   if (left>=iv.right) return Iv(); // open intervals disjoint
   R amin=CpmRootX::sup<R>(left,iv.left);
   R amax=CpmRootX::inf<R>(right,iv.right);
   return Iv(amin,amax,true); // fast return
}

bool Iv::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Iv",str);
   cpmp(left);
   cpmp(right);
   return true;
}

bool Iv::scanFrom(istream& str)
{
   cpms(left);
   cpms(right);
   update();
   return true;
}

Z Iv::com(Iv const& iv)const
{
   if (left<iv.left) return 1;
   if (left>iv.left) return -1;
   if (right<iv.right) return 1;
   if (right>iv.right) return -1;
   return 0;
}

V<R> Iv::cen(Z n)const
{
   if (n<=0) cpmerror("Iv::cen(Z): n<=0");
   V<R> res(n);
   R d=(right-left)/n; // pixel width
   R xi=left+d*0.5;  // left+pixel half-width
   for (Z i=res.b();i<=res.e();i++){
      res.cui(i)=xi;
      xi+=d;
   }
   return res;
}

R Iv::cen(Z i, Z n)const
{
   cpmassert(n!=0,"Iv::cen(Z,Z)");
   R d=(right-left)/n;
   R x0=left-0.5*d;
   return x0+d*i;
}

R Iv::dis(Iv const& iv)const
{
   R d1=CpmRoot::dis(left,iv.left);
   R d2=CpmRoot::dis(right,iv.right);
   return cpmsup(d1,d2);
}

V<R> Iv::div(Z n)const
{
   if (n<0) n=-n;
   V<R> res(n);
   if (n==0) return res;
   if (n==1){
      res[1]=center;
      return res;
   }
   res[1]=left;
   res[n]=right;
   if (n>2){
      R d=(right-left)/(n-1);
      R ri=left;
      for (Z i=2;i<n;i++){
         ri+=d;
         res[i]=ri;
      }
   }
   return res;
}

V<Iv> Iv::divide(Z n)const
{
   if (n<=0) cpmerror("Iv::subIntervals(Z): n<=0");
   V<Iv> res(n);
   R d=abs()/n; // pixel width
   R bl=left;
   R br=bl+d;
   for (Z i=1;i<=n;i++){
      res[i]=Iv(bl,br,true);
      bl=br;
      br+=d;
   }
   return res;
}

const Iv& Iv::stretch(R k, Z s)
{
   R newRad=halfWidth*cpmabs(k);
   if (s==0){
      left=center-newRad;
      right=center+newRad;
   }
   else if(s==-1){
      right=left+newRad*2;
   }
   else if (s==1){
      left=right-newRad*2;
   }
   else{
      cpmerror("Iv::stretch: bad value of second argument");
   }
   update();
   return *this;
}

R Iv::containsFuzzy(R x, R frac)const
{
   cpmassert(frac>=1. && abs()>0.,"Iv::containsFuzzy(R,R)");
   if( contains(x)) return 1.;
   Iv iv=(*this)*frac;
   if (!iv.contains(x)) return 0.;
   R c=mean();
   R rx=cpmabs(x-c);
   R r1=radius();
   R r2=iv.radius();
   return (r2-rx)/(r2-r1);
}

R Iv::periodic(R xx)const
{
   R av=abs();
   if (av==0.){
      cpmerror("Iv::periodic(R) called for left==right");
   }
   R y=(xx-left)*absInv;
   return y-cpmfloor(y);
}

R Iv::per(R x)const
{
   if (x>=left && x<=right) return x;
   R av=abs();
   if (av==0.){
      cpmerror("Iv::per(R) called for left==right");
   }
   if (x>right){
      R res=x;
      while (res>right) res-=av;
      return res;
   }
   else { // then x<left
      R res=x;
      while (res<left) res+=av;
      return res;
   }
}

R Iv::ramp(R xx)const
{
   if (xx<left) return 0.;
   else if (xx>right) return 1;
   else {
      if (left==right) return 0.5;
      else return (xx-left)*absInv;
   }
}

R Iv::win(R x, R eps)const
{
   using CpmFunctions::hillFunc;
   if(x<=left || x>=right) return 0.;
   R p=(x-left)*absInv;
   R epsh=cpmabs(eps)*0.5;
   if (p>=epsh && p<=R(1.)-epsh) return 1.;
   R epshi=cpminv(epsh);
   if (p<epsh){
      R q=p*epshi;
      return R(1.)-hillFunc(q);
   }
   else{
      R q=(p-1.)*epshi+1.; // = (p-(1-epsh))/epsh
      return hillFunc(q);
   }
}

Z Iv::qun(R xx, Z n, bool restrict)const
{
   cpmassert(n>0,"Iv::qun(R,Z,bool)");
   if (restrict==false){
      if (xx==right) return n;
      else return 1+cpmtoz(R(n)*(xx-left)*absInv);
   } // this was the only functionality till 2003-02-06
   else{ // new
      if (xx<=left) return 1;
      else if (xx>=right) return n;
      else{
         Z p=cpmtoz(R(n)*(xx-left)*absInv)+1;
         if (p<1) return 1;
         if (p>n) return n;  // 2015-01-19: p=n+1 happened. To prevent this
            // we have now this two conditions.
         return p;
      }
   }
}

R Iv::qunInv(Z n, IvZ const& iv)const
      //: quantize inverse
      // returns y which is related to *this just as n is
      // related to iv.
{
   if (iv.isVoid()) return center;
   Z rp=iv.relPos(n);
   if (rp==-1) return left;
   else if (rp==1) return right;
   else{ // now n belongs to
      Z dn=n-iv.fir(); // varies from 0 to |iv|-1
      Z ni=iv.car();
      R dx=abs()/ni;
      return (R(dn)+0.5)*dx;
   }
}

X2<R2,Z2> Iv::intPolInfo(R xx, Z n)const
// only rather simple computations in each control path
// So, less complicated than it looks
{
   if (xx<=left) return X2<R2,Z2>(R2(0.5,0.5),Z2(1,1));
   if (xx>=right) return X2<R2,Z2>(R2(0.5,0.5),Z2(n,n));
   R r=R(n)*(xx-left)*absInv; // takes value in [0,n)
   R rL=cpmfloor(r); // takes values in {0,....n-1}
   Z iA=1+cpmtoz(rL); // takes values in {1,...,n}
   R qA=r-rL; // takes values [0,1). If this has value 0.5,
      // the interval indexed iA has dominant (actually
      // determining) influence, since we then are at the
      // center of this interval. For qA>0.5 the intervals
      // under consideration are indexed iA and iA+1
      // else iA and iA-1. The weight of this additional
      // index is |qA-0.5| and that of iA is 1-|aA-0.5|
   R wAdd=cpmabs(qA-0.5);
   R wA=R(1)-wAdd;
   Z iAdd;
   if (qA<=0.5){
      iAdd=iA>1 ? iA-1 : iA;
      return X2<R2,Z2>(R2(wAdd,wA),Z2(iAdd,iA));
   }
   else{
      iAdd=iA < n ? iA+1 : iA;
      return X2<R2,Z2>(R2(wA,wAdd),Z2(iA,iAdd));
   }
}

V<B> Iv::bisection(R xx, Z n)const
{
   Z mL=3;
   Word loc("Iv::bisection(R,Z)");
   cpmmessage(mL,loc&" started");
   cpmassert(n>0,loc);
   if (xx<left) xx=left;
   else if (xx>right) xx=right;
   else ;
   V<B> res(n,B(false));
   R separator=center;
   R d=radius();
   for (Z i=1;i<=n;i++){
      d*=0.5;
      if (xx<=separator){
         separator-=d;
      }
      else{
         separator+=d;
         res[i]=true;
      }
   }
   cpmmessage(mL,loc&" done");
   return res;
}

R Iv::bisectionInv(const V<B>& vb)const
{
   Z mL=3;
   Word loc("Iv::bisectionInv(V<bool>)");
   cpmmessage(mL,loc&" started");
   Z n=vb.dim();
   cpmassert(n>0,loc);
   R res=left;
   R d=abs();
   for (Z i=1;i<=n;i++){
      d*=0.5;
      if (vb[i]) res+=d;
   }
   cpmmessage(mL,loc&" done");
   return res;
}

namespace {
   R fAux2(const R& q, const R& b, const Z& n)
      // always defined unless ::pow(q,n-2) overflows
   {
      return q!=1. ? b-(R(1)-cpmpow(q,R(n-1)))/(R(1)-q): b-R(n)+R(1);
   }
}

V<R> Iv::divGeoProg(R dt,Z n)const
{
   Z mL=4;
   Word loc("V<R> Iv::divGeoProg(R dt,Z n)");
   cpmmessage(mL,loc&" started");
   cpmassert(n>2,loc);
   cpmassert(dt>0,loc);
   R av=abs();
   cpmassert(dt<av,loc);
   V<R> res(n);
   R b=av/dt;
#ifdef CPM_Fn
   F<R,R> f=F2<R,R,Z,R>(b,n)(fAux2);
#else
   F<R,R> f(bind(fAux2,_1,b,n));
#endif
   R qPrev=0.;
   R fPrev=f(qPrev);
   cpmassert(fPrev>0,loc);
   // lean and fast root finding since the function is
   // known to fall with increasing speed
   R qAct=0.1;
   R fAct=f(qAct);
   R incrFactor=2;
   R acc=1e-6;
   while (fAct>0){
      qPrev=qAct;
      qAct*=incrFactor;
      fAct=f(qAct);
   }
   R q=CpmAlgorithms::rootByBisection(f,qPrev,qAct,acc);
   res[1]=inf();
   res[n]=sup();
   R step=dt;
   res[2]=res[1]+step;
   if (n==3){
      return res;
   }
   for (Z i=3;i<=n;i++){
      step*=q;
      res[i]=res[i-1]+step;
   }
   cpmmessage(mL,loc&" done");
   return res;
}

namespace{
   V<R> equalize(const V<R>& v)
   {
      static R i3=1./3.;
      Z nSup=v.dim()-2;
      V<R> res=v;
      for (Z i=3;i<=nSup;i++){
         res[i]=(v[i-1]+v[i]+v[i+1])*i3;
      }
      return res;
   }
}

V<R> Iv::divSmooth(R dti,R dtf,Z n)const
{
   Z mL=4;
   Word loc("V<R> Iv::divGeoProg(R dti,dtf,Z n)");
   cpmmessage(mL,loc&" started");
   cpmassert(n>3,loc);
   cpmassert(dti>0,loc);
   cpmassert(dtf>0,loc);
   R av=abs();
   R gap=av-dti-dtf;
   cpmassert(gap>0,loc);
   V<R> res(n);
   res[1]=inf();
   res[2]=res[1]+dti;
   res[n]=sup();
   res[n-1]=res[n]-dtf;
   Z i;
   if (n==4){
      return res;
   }
   else{
      Z p=n-4; // these divide gap into p+1 intervals (notice that the
      // end points of gap are not among the p points!
      // of length gap/(p+1)
      R gapd=gap/(p+1);
      for (i=1;i<=p;i++) res[i+2]=res[i+1]+gapd;
   }
   Z nSteps=n;
   for (i=1;i<=nSteps;i++)  res=equalize(res);
   cpmmessage(mL,loc&" done");
   return res;
}

R Iv::ran(Z j)const
{
   Z mL=3;
   Word loc("Iv::ran(Z)");
   CPM_MA
   R xx=CpmRoot::randomR(j);
   R res = left+xx*(right-left);
   CPM_MZ
   return res;
}

Iv& Iv::lorentz_(R chi)
{ // see Domenico Castrigiano, (8) Corollary
   R cc=cpmcosh(chi);
   R ss=cpmsinh(chi);
   R sL=cpmabs(left*ss);
   R sR=cpmabs(right*ss);
   left*=cc;
   right*=cc;
   left-=sL;
   right+=sR;
   update();
   return *this;
}


////////// functions concerning proportions ////////////////////////////

R CpmGeo::proportion(R x1, R x2, R xd)
{
   R nominator=x2-xd;
   cpmassert(nominator!=0,"CpmGeo::proportion");
  // return (xd-x2)/nominator;
   return (xd-x1)/nominator; // correction 2005-06-22
}

R CpmGeo::pointFromProportion(R x1, R x2, R prp)
{
   R nominator=prp+1;
   cpmassert(nominator!=0,"CpmGeo::pointFromProportion");
   return (x1+prp*x2)/nominator;
}

////////////////////// class DiIv //////////////////////////////////////

void DiIv::update()
{
   CpmRoot::Word loc("DiIv::update()");
   range=x1-x0;
   cpmassert(range!=0,loc);
   rangeInv=R(1)/range;
}

DiIv::DiIv(Iv iv):x0(iv.inf()),x1(iv.sup())
{
   update();
}

DiIv::DiIv(R a, R b):x0(a),x1(b)
{
   update();
}

DiIv::DiIv(R2 const& r2):x0(r2[1]),x1(r2[2])
{
   update();
}

bool DiIv::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("DiIv",str);
   cpmp(x0);
   cpmp(x1);
   return true;
}

bool DiIv::scanFrom(istream& str)
{
   cpms(x0);
   cpms(x1);
   update();
   return true;
}

/////////////////////////// class Iv2 ////////////////////////////////////

namespace{
   R f1(R2 const& x){ return x[1];}
   R f2(R2 const& x){ return x[2];}
} // namespace

Iv2::Iv2(V<R2> const& vr):
iv1_(vr(F<R2,R>(f1))),iv2_(vr(F<R2,R>(f2)))
{}

bool Iv2::contains(R2 const& p)const
{
   bool b1 = iv1_.contains(p[1]);
   bool b2 = iv2_.contains(p[2]);
   return b1&&b2;
}

bool Iv2::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Iv2",str);
   cpmp(iv1_);
   cpmp(iv2_);
   return true;
}

bool Iv2::scanFrom(istream& str)
{
   cpms(iv1_);
   cpms(iv2_);
   return true;
}

Z Iv2::com(Iv2 const& iv)const
{
   if (iv1_<iv.iv1_) return 1;
   if (iv1_>iv.iv1_) return -1;
   if (iv2_<iv.iv2_) return 1;
   if (iv2_>iv.iv2_) return -1;
   return 0;
}
