//? cpmm.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_M_H_
#define CPM_M_H_
/*

   Description: Defines a template class M<X,Y> of mappings
      X-->Y comparable to std::map<X,Y>
*/

#include <map>
#include <cpmv.h>
#include <cpmsr.h>

////////////////////// class M<X,Y> ////////////////////////////////////

namespace CpmArrays{

   using namespace CpmStd;
   using namespace CpmRoot;
   using CpmFunctions::F;
   using CpmArrays::S;

template <class X, class Y>
   class M{ // map, associative array, dictionary, hash
   // it is assumed that X<X and X>X is defined

   Y y0_{};
   typedef M<X,Y> Type;

public:
   std::map<X,Y> p_;
 //  Z locate(X const& x)const{ return rep_.locate(aux::PairO1<X,Y>(x));}

   auto begin(){return p_.begin();}
   auto end(){return p_.end();}
   auto cbegin(){return p_.cbegin();}
   auto cend(){return p_.cend();}
   CPM_IO
   CPM_ORDER
   M():p_{}{}
      // constructor for the void set of X,Y pairs.
      // Non-trivial instances can be obtained by applying
      // mutating methods, such as function set(...) on the
      // default instance
   M<X,Y> test(Z i)const;
   M<X,Y> ran(Z i=0)const;
   R dis(M<X,Y> const& m)const;
   R absSqr()const;
   R abs()const;

    X const& inv(Y const& y, bool& found)const;
      //:inverse
      // Returns a value x in X such that (*this)[x]==y if such an x
      // exists. Otherwise we return the dfault element from X
      // an set 'found' equal to false.


   M(S<X> const& sx, F<X,Y> const& f):p_{},y0_{}
      // Setting the values by an algorithm.
   {
      S<X> s{sx};
      for (auto const& x:s){
         p_[x]=f(x);
      }
   }

   template <class T>
   M<X,T> operator&(M<Y,T> const& m)const;
   // concatenation corresponds to composition of functions
/*
   M(S<X> const& sx, Y const& y);
      // Setting a single value y for all elements of sx.

   void merge(V<Type> const& mapList, bool reverse=false);
      // combining *this and a list of maps into a single one.
      // This constructs a map that is formed out of all
      // (x,y)-pairs of *this and the components of mapList.
      // If for a x, there are more pairs (x,y), (x,y'), ...
      // The pair coming from mapList[i] with highest i
      // will actually be taken (if we would take more than one
      // we would not get a map). Thus we work through the
      // mapList in ascending order of their indexes and use it
      // in an overwriting mode. If the second argument differs
      // from the default value, we use reversed order so that
      // mapList[i] has highest priority for smallest i.
*/
   Y operator()(X const& x)const{ return (*this)[x];}
      // in analogy to F<X,Y>; returns the 'value of *this at x'

 //  S<Y> operator()(S<X> const& sx)const;
      // The common extension of functions from single argument values
      // to sets of argument values

   // cardinality access: number of (x,y) pairs stored in *this
   Z car()const{ return p_.size();}
      //: cardinality
   Z dim()const{ return p_.size();}
      //: dimension
      // for uniformity with other array types
   Z size()const{ return p_.size();}
      // for uniformity with STL code
   bool isVoid(void)const{ return p_.size()==0;}
      //: is void

   M<X,Y> select(F<X,bool> const& sel)const;
      // returns the M<X,Y> objects consisting of all those
      // (x,y)-pairs of *this for which sel(x)==true

   Y& operator[](X const& x){ return p_[x];}
      // setting values in array-style (which also is the STL-style)
      // for M<X,Y> f, X x, Y y the statement
      // f[x]=y has the same effect as f.set(x,y)

   Y const&  operator[](X const& x)const{
      auto q=p_.find(x);
      return q!=p_.end() ? q->second : y0_;}
      // getting to the values as references.
      // Let Z g(const& Y), M<X,Y> f, X x, Y y . Then
      // g(f[x]) does the same as
      // Y y1; f.get(x,y1); g(y1);
      // So, operator[] provides slightly more convenient grammar than
      // get(...) in this case.
      // This is the most efficient non-mutating access function

   bool defined(X const& x)const{ return p_.find(x) != p_.end();}
      // f.defined(x) iff before a statement f[x]=... or f.set(x,...)
      // happened.

   void set(X const& x, Y const& y){(*this)[x]=y;}
      //: set
      //  M<X,Y> f, X x, Y y ==>
      // ( f.set(x,y) ==> (y==f(x) evaluates to true)). Interesting:
      // it is possible to express this mixture of logic notation
      // and C++ much clearer in pure C++:
      // template <class X, class Y>
      // bool setTest(X x, Y y, M<X,Y> f){ f.set(x,y); return y==f(x);}
      // returns true for all arguments.

   bool get(X const& x, Y& y)const;
      //: get
      // template <class X, class Y>
      // bool getTest(X x, M<X,Y> f){ Y y; f.get(x,y); return y==f(x);}
      // returns true for all arguments.

   V<Y> const& toV()const;

   S<X> dom()const{
      S<X> res;
      for (auto const& [x,y]:p_){
         res.add_(x);
      }
      return res;
   }

   S<Y> rng()const{
      S<Y> res;
      for (auto const& [x,y]:p_){
         res.add_(y);
      }
      return res;
   }


   Word toWord()const;
      //: to word

   Word nameOf()const;
      //: name of
};


///////////////////////// Implementation ///////////////////////////////

//{
//   Vr<T> res(Base::dom());
//   T tempX;
//   if (j==0){
//      for (Z i=Base::b(); i<=Base::e(); i++){
//         tempX=(*this)[i];
//         res[i]=Root<T>(tempX).ran(j);
//      }
//   }
//   else{
//      Z j0=1000; // this is done to avoid an obvious correlation
//         // among the coordinates of x.ran(i)
//         // and x.ran(i+1)
//      Z jIncr=Root<Z>(j0).ran(j);
//      Z jc=j;
//      for (Z i=Base::b(); i<=Base::e(); i++){
//         tempX=(*this)[i];
//         jc+=jIncr;
//         res[i]=Root<T>(tempX).ran(jc);
//      }
//   }
//   return res;
//}

template <class X, class Y>
M<X,Y> M<X,Y>::test(Z i)const
{  Sr<X> s;
   s=s.test(i);
   Y y0;
   Y yi=Root<Y>(y0).test(i);
   M<X,Y> res;
   for (auto const& x : s){
     res[x]=Root<Y>(yi).ran(++i);
   }
   return res;
}

template <class X, class Y>
M<X,Y> M<X,Y>::ran(Z i)const
{
   M<X,Y> res;
   if (i==0){
      for (auto const& [x,y] : p_ ){
         res[x]=Root<Y>(y).ran(i);
      }
   }
   else{
      Z i0{1000};
      Z ic=i;
      Z iIncr=Root<Z>(i0).ran(i);
      for (auto const& [x,y] : p_ ){
         ic+=iIncr;
         res[x]=Root<Y>(y).ran(ic);
      }
   }
   return res;
}

template <class X, class Y>
R M<X,Y>::dis(M<X,Y> const& m)const
{
   R res{0.};
   for (auto const& [x,y] : p_){
      Y y2=m[x];
      R d=CpmRoot::Root<Y>(y).dis(y2);
      if (d>res) res=d;
   }
   return res;
}

template <class X, class Y>
R M<X,Y>::absSqr()const
{
   R res{0.};
   for (auto const& [x,y] : p_){
      res+=CpmRoot::Root<Y>(y).absSqr();
   }
   return res;
}

template <class X, class Y>
R M<X,Y>::abs()const
{
   return cpmsqrt(absSqr());
}

template <class X, class Y>
V<Y> const& M<X,Y>::toV()const
{
   Z n=size(),i=0;
   V<Y> res(n);
   for (auto const& [key, value] : p_){
      i++;
      res[i]=value;
   }
   res.sort_();
   return res;
}

template <class X, class Y>
template <class T>
M<X,T> M<X,Y>::operator&(M<Y,T> const& m)const
{
   M<X,T> res;
      for (auto const& [x,y] : p_){
      bool bi=m.defined(y);
      if (bi) res.set(x,m[y]);
   }
   return res;
}

template <class X, class Y>
bool M<X,Y>::get(X const& x, Y& y)const
{
   auto q=p_.find(x);
   if (q==p_.end()){
      return false;
   }
   y=q->second;
   return true;
}

template <class X, class Y>
X const& M<X,Y>::inv(Y const& y, bool& found)const
{
  X res{};
  found=false;
  for (auto const& [key, value] : p_){
      if (value==y){ res=key; found=true; break;}
  }
  return res;
}

template <class X, class Y>
M<X,Y>  M<X,Y>::select(F<X,bool> const& sel)const
{
   M<X,Y> res;
   for (auto const& [key,value] : p_){
      X2<X,Y> xy(key,value);
      X x=xy.c1();
      if (sel(x)) res[x]=xy.c2();
   }
   return res;
}

template <class X, class Y>
Word M<X,Y>::nameOf()const
{
   Word wi="M<";
   Word wx=CpmRoot::Name<X>()(X());
   Word wy=CpmRoot::Name<Y>()(Y());
   return wi&wx&","&wy&">";
}

template <class X, class Y>
Word M<X,Y>::toWord()const
{
   Word res="{ ";
   bool first=true;
   for (const auto& [key, value] : p_){
      Word wx=CpmRoot::toWord<X>(key);
      Word wy=CpmRoot::toWord<Y>(value);
      Word wb="("&wx&", "&wy&")"; // b for begining
      Word wf=", "&wb; // f for following
      first ? res&=wb : res&=wf;
      first=false;
   }
   res&=" }";
   return res;
}

template <class X, class Y>
bool M<X,Y>::prnOn(ostream& str)const
{
   Z mL=3;
   static Word loc("M<>::prnOn(ostream&)");
   CPM_MA
   Z n=size();
   cpmwat;
   cpmp(n);
   if (n==0){
      cpmmessage(mL, "no values there");
      goto label;
   }
   for (const auto& [key, value] : p_) {
      X2<X,Y> val(key,value);
      if (!val.prnOn(str)){
         CPM_MZ
         return false;
      }
    }
label:
   CPM_MZ
   return true;
}

template <class X, class Y>
bool M<X,Y>::scanFrom(istream& str)
{
   Z mL=3;
   Z mL2=5;
   static Word loc("M<>::scanFrom(istream&)");
   CPM_MA
   Z n;
   if (!CpmRoot::scanT<Z>(n,str)){
      cpmwarning(loc&": can't read dimension");
      CPM_MZ
      return false;
   }
   else if (n<0){
      cpmwarning(loc&": negative dimension: n="&cpm(n));
      CPM_MZ
      return false;
   }
   else if (n>dimMax)
      cpmwarning(loc&": dimension read as "&cpm(n)&
      ": probably too large");
   else if (n==0){
      cpmwarning(loc&" dimension read as 0");
      if (cpmverbose>mL2){ // copying the stream beyond the dubious point
         // can be useful
         char c;
         while (str.get(c)) cpmdata.put(c);
      }
   }
   else{
      cpmmessage(mL,loc&": dimension read as "&cpm(n));
   }
   M<X,Y> res;
   X2<X,Y> val;
   //for (const auto& [key, value] : p_) {
   for (Z i=1;i<=n;i++){
      if (!val.scanFrom(str)){
         cpmwarning(loc&": reading failed for i="&cpm(i)&" of "&cpm(n));
         CPM_MZ
         return false;
      }
      res[val.c1()]=val.c2();
   }
   *this=res;
   CPM_MZ
   return true;
}

template <class X, class Y >
Z M<X,Y>::com(M<X,Y> const& m)const
{
   Z d1=size(), d2=m.size();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{ // then d1==d2
      for (auto const& [key, value] : p_){
         if (value < m[key])   return 1;
         if (value > m[key])   return -1;
      }
      return 0;
   }
}

/******************************
template <class X, class Y>
template<class T>
M<X,T> M<X,Y>::operator()(F<Y,T> const& g)const
{
   M<X,T> res();
   for (Z i=b();i<=e();++i){
      X2<X,Y> xyi=xy(i);
      res.set(xyi.c1(),g(xyi.c2()));
   }
   return res;
}

template <class X, class Y>
template <class T>
M<X,T> M<X,Y>::operator&(M<Y,T> const& m)const
{
   M<X,T> res;
   for (Z i=b();i<=e();++i){
      X2<X,Y> xyi=xy(i);
      bool bi=m.defined(xyi.c2());
      if (bi) res.set(xyi.c1(),m[xyi.c2()]);
   }
   return res;
}

template <class X, class Y>
Word M<X,Y>::nameOf()const
{
   Word wi="M<";
   Word wx=CpmRoot::Name<X>()(X());
   Word wy=CpmRoot::Name<Y>()(Y());
   return wi&wx&","&wy&">";
}

template <class X, class Y>
Word M<X,Y>::toWord()const
{
   Word res="{ ";
   for (Z i=b();i<=e();++i){
      X2<X,Y> xyi=xy(i);
      Word wxi=CpmRoot::toWord<X>(xyi.c1());
      Word wyi=CpmRoot::toWord<Y>(xyi.c2());
      res&=" ( "&wxi&" , "&wyi&" ) ";
      if (i!=e()) res&=" , ";
   }
   res&=" }";
   return res;
}

template <class X, class Y>
M<X,Y>::M(S<X> const& sx, F<X,Y> const& f):y0_(),rep_()
{
   for (Z i=sx.b();i<=sx.e();++i){
      X xi=sx[i];
      set(xi,f(xi));
   }
}

template <class X, class Y>
M<X,Y>::M(S<X> const& sx, Y const& y):y0_(),rep_()
{
   for (Z i=sx.b();i<=sx.e();++i){
      X xi=sx[i];
      set(xi,y);
   }
}

template <class X, class Y>
Y M<X,Y>::operator()(X const& x)const
{
   Z i=locate(x);
   if (i==0) return Y();
   else return rep_[i].c2();
}

template <class X, class Y>
Y& M<X,Y>::operator[](X const& x)
{
   aux::PairO1<X,Y> fx(x);
   rep_<<fx;
   Z i=rep_.locate(fx);
   return rep_.ref(i).c2();
}

template <class X, class Y>
Y const& M<X,Y>::operator[](X const& x)const
{
   aux::PairO1<X,Y> fx(x);
   Z i=rep_.locate(fx);
   if (i<=0){
      return y0_;
   }
   else{
      return rep_[i].c2();
   }
}

template <class X, class Y>
bool M<X,Y>::get(X const& x, Y& y)const
{
   Z i=locate(x);
   if (i==0){
      return false;
   }
   else{
      y=rep_[i].c2();
      return true;
   }
}

template <class X, class Y>
S<Y> M<X,Y>::operator()(S<X> const& sx)const
{
   Z n=sx.car();
   V<Y> val(n);
   for (Z i=1;i<=n;i++){
      val[i]=y((sx[i]));
   }
   return S<Y>(val);
}

template <class X, class Y>
void M<X,Y>::merge(V<Type> const& mapList, bool reversed)
{
   Z ni,i,j,d=mapList.dim();
   if (!reversed){
      for (i=1;i<=d;i++){
         ni=mapList[i].car();
         for (j=1;j<=ni;j++){
            X x=mapList[i].x(j);
            Y y=mapList[i].y(j);
            set(x,y);
         }
      }
   }
   else{
      for (i=d;i>=1;i--){
         ni=mapList[i].car();
         for (j=1;j<=ni;j++){
            X x=mapList[i].x(j);
            Y y=mapList[i].y(j);
            set(x,y);
         }
      }
   }
}

template <class X, class Y>
S<X> M<X,Y>::dom()const
{
   S<X> res;
   for (Z i=1;i<=dim();++i) res.add(x(i));
   return res;
}

template <class X, class Y>
S<Y> M<X,Y>::ran()const
{
   S<Y> res;
   for (Z i=1;i<=dim();++i) res.add(y(i));
   return res;
}

template <class X, class Y>
M<X,Y>  M<X,Y>::select(F<X,bool> const& sel)const
{
   M<X,Y> res;
   for (Z i=1;i<=dim();++i){
      X2<X,Y> xyi=xy(i);
      X xi=xyi.c1();
      if (sel(xi)) res[xi]=xyi.c2();
   }
   return res;
}

template <class X, class Y>
bool M<X,Y>::prnOn(ostream& str)const
{
   Z mL=3;
   static Word loc("M<>::prnOn(ostream&)");
   CPM_MA
   Z n=car();
   cpmwat;
   cpmp(n);
   if (n==0){
      cpmmessage(mL, "no values there");
      goto label;
   }
   for (Z i=1;i<=n;i++){
      if (CpmRoot::wrtTit){
         Word wi("// i=");
         wi&=cpm(i);
         bool bi=wi.prnOn(str);
         cpmassert(bi==true,loc);
      }
      X2<X,Y> val(x(i),y(i));
      if (!val.prnOn(str)){
         CPM_MZ
         return false;
      }
   }
label:
   cpmwet;
   CPM_MZ
   return true;
}

template <class X, class Y>
bool M<X,Y>::scanFrom(istream& str)
{
   Z mL=3;
   Z mL2=5;
   static Word loc("M<>::scanFrom(istream&)");
   CPM_MA
   Z n;
   if (!CpmRoot::scanT<Z>(n,str)){
      cpmwarning(loc&": can't read dimension");
      CPM_MZ
      return false;
   }
   else if (n<0){
      cpmwarning(loc&": negative dimension: n="&cpm(n));
      CPM_MZ
      return false;
   }
   else if (n>dimMax)
      cpmwarning(loc&": dimension read as "&cpm(n)&
      ": probably too large");
   else if (n==0){
      cpmwarning(loc&" dimension read as 0");
      if (cpmverbose>mL2){ // copying the stream beyond the dubious point
         // can be useful
         char c;
         while (str.get(c)) cpmdata.put(c);
      }
   }
   else{
      cpmmessage(mL,loc&": dimension read as "&cpm(n));
   }
   M<X,Y> res;
   X2<X,Y> val;
   for (Z i=1;i<=n;i++){
      if (!val.scanFrom(str)){
         cpmwarning(loc&": reading failed for i="&cpm(i)&" of "&cpm(n));
         CPM_MZ
         return false;
      }
      res.set(val.c1(),val.c2());
   }
   *this=res;
   CPM_MZ
   return true;
}

template <class X, class Y >
Z M<X,Y>::com(M<X,Y> const& s)const
{
   Z d1=dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{ // then d1==d2
      for (Z i=1;i<=d1;i++){
         if (xy(i)<s.xy(i))   return 1;
         if (xy(i)>s.xy(i))   return -1;
      }
      return 0;
   }
}
******************************************/
} // CpmArrays

#endif
