//? cpmmini.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_MINI_H_
#define CPM_MINI_H_

/*
Purpose: Minimum search for a real-valued function of a real vector
   argument in 'modern' form, where the function is
   represented as an instance of F<R_Vector,R> and is a data member of
   the class Mini. The class thus is no longer a template.
*/

#include <cpmrmatrix.h>
#include <cpmfa.h>

///////////////////////// class Mini /////////////////////////////////////

namespace CpmAlgorithms{

   using namespace CpmStd;

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::B;
   using CpmFunctions::F;
   using CpmFunctions::Fa;
   using CpmArrays::R_Vector;
   using CpmArrays::X2;
   using CpmArrays::R_Matrix;

class Mini{   // finds minimum of a function
   // ff based on Powell's method, modified from Press et al.

   typedef Mini Type;
public:
   virtual Mini* clone()const{ return new Mini(*this);}

   Mini(R fracTol=1e-3);
      // fracTol=ftol is the the fractional tolerance in the function
      // value such that failure to decrease by more than this ammount on
      // one iteration signals doneness.

   explicit Mini(F<R_Vector,R> const& f, R fracTol=1e-3);
      // sets the function to be minimized and the fractional tolerance
      // (see previous function)

   void setAcc(R fracTol);
      // see first function

   virtual void setF(const F<R_Vector,R>& f);
      // setting the 'function' ff to be minimized.

   R_Vector getPoint(const R_Vector& xStartInput);
      // computing the minimum and returning the resulting x-value.
      // The starting point has to be provided.
      // The vectors x which are taken into account for minimization
      // are those for which dim(x)=dim(xStartInput)
      // An assert condition assures that the program gets interupted
      // if this function is called and data member ff was not yet
      // initialized by either a appropriate constructor call or by
      // calling setF

   virtual  R getValue(void)const;
      // getting the value of the minimized function at the minimum
      // position. Returns reasonable values only after getPoint was
      // called before. An assert condition assures that the program
      // gets interupted if this is not the case.

   X2<R_Vector,R> operator()(const R_Vector& x)
      // 'modern' interface to the main functionality of the class
   {  R_Vector xm=getPoint(x); R y=getValue();
      return X2<R_Vector,R>(xm,y);}

   Z getIter(void)const;
      // getting the number of iterations needed. Returns reasonable
      // values only after getMin was called before. An assert condition
      // assures that the program gets interupted if this is not the
      // case.

   Z nEval()const{ return nEval_;}

   Z getIterBrent(void)const;
      // getting the maximum number of iterations needed in brent().
      // Returns reasonable values only after getMin was called before.
      // An assert condition assures that the program gets interupted if
      // this is not the case.

   bool getSuc()const{ return suc_;}

   Z verbose{0};
      // for controlling messages in a flexible way without using
      // cpmverbose

   static Z iterMaxPowell;
      // 30 seems to be a good value; should be open to changes, thus
      // public

   static Z iterMaxBrent;
      // 30 seems to be a good value; should be open to changes, thus
      // public

   void show()const;
      // writes all relevant method parameters on file for diagnostic
      // purposes

protected:
   Fa<R_Vector,R> ff_;
       //    function object to be minimized

private:
   B suc_;
      // true if minimum sucessfully found, initialized as
      // false which is OK
   R minimize(R_Vector& x);
   R ftol, yMin;
   R_Vector xStart, xMin;
   Z done;
      // zero if no minimization was yet done

   Z ff_Defined;
      // zero if ff_ not yet explicitely initializes
   Z iter;
   Z iterBrent;
   Z callsBrent;
   R_Vector pPowell;
   R_Vector xiPowell;
   R f1Powell(R x){ nEval_++; return ff_(pPowell+x*xiPowell);}

// auxiliar functions modified from Press et al.
   bool brak(R& ax, R& bx, R& cx, R& fa, R& fb, R& fc);
   bool testBrak(R& ax, R& bx, R& cx, R& fa, R& fb, R& fc);
   R brent(R ax, R bx, R cx, R& xmin);
   bool linMinPowell(R_Vector& p, R_Vector& xi, R& fret);
       // sets pPowell and xiPowell
   bool powell(R_Vector& x, R_Matrix xi, R& fret);
      // core of the method
   Z nEval_;
      // number of evaluations of ff_ during the
      // whole minimization process

// method constants from Press et al.
   static R GOLD;
   static R GLIMIT;
   static R TINY;
   static R EPS;
   static R TOL;
   static R CGOLD;
   static R ZEPS;
};

class Maxi:public Mini{ // finds maximum of a function
   // which gets entered either via constructor or via
   // setF
   typedef Mini Base;
   typedef Maxi Type;
public:
   virtual Mini* clone()const{ return new Maxi(*this);}
      // this allows instances of Maxi and of Mini to be
      // components of the polymorphic array class Vp<Mini>
   Maxi(R fracTol=1e-3):Base(fracTol){}
   Maxi(const F<R_Vector,R>& f, R fracTol):Base(f,fracTol){ff_*=-1;}
      // f is a function for which the maximum will be searched
      // Internally this is done by searching the minimum of -f,
      // as is clear from the code
   R getValue(void)const{ return -Base::getValue();}
      // getting the value of function f at point of maximum
   void setF(const F<R_Vector,R>& f){ Base::setF(f); ff_*=-1;}
      // setting the function to be maximized.
};

} // namespace

#endif
