//? cpmnumbers.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_NUMBERS_H_
#define CPM_NUMBERS_H_
/*
   Description: The C+- class system needs some basic types from C++.
      These are in the first place: (long) int, unsigned (long) int,
      unsigned char, bool, string, and (long) double.
      With the availability of convenient and efficient C++ types for
      floating point numbers of 'multiple precision' a new degree of
      freedom can be added to the C+- class system. We therefore add
      the multiple precision class mpfr::mpreal (by Pavel Holoborodko)
      which is based on the multiple precision libraries gmp and mpfr.
      This class can be used if the macro CPM_MP is defined in the header
      file cpmdefinitions.h.

      Introduces a common infrastructure for some basic types.
      There are two topics of interest here:
      1. It is a common situation that a class is to be defined in a way
      that some (or all) its data members have types which belong to these
      basic types. Then the methods should be in place which allow to
      carry over some basic infra structure from the members to the
      class to be defined. Most frequently used in present C+- code is
      the implementation of stream I/O ('Marshalling') according to the
      following example:

      class X{
         B1 x1; // B1,B2,...,Bn basic types
         B2 x2;
         ...
         Bn xn;
         typedef X Type;
      public:
         CPM_IO
            // declaration macro for Input/Output
         Word nameOf()const{ return "X";}
            //: name of
            ...
      };

      bool X::prnOn(str)const
      {
         cpmwat; // uses the nameOf - function to write an
            // automatic 'title'
         cpmp(x1); // 'p' for 'print'
         cpmp(x2);
         ...
         cpmp(xn);
         return true;
      }

      bool X::scanFrom(str)
      {
         cpms(x1); // 's' for 'scan'
         cpms(x2);long
         ...
         cpms(xn);
         return true;
      }

      The declaration macro CPM_IO and the implementation macros
      cpmwat, cpmp, cpms are defined in cpminterfaces.h.
      The main achievement is the recursive nature of the scheme:
      If X is the type of a member of a class Y the same mechanism
      works:

      class Y{
         X x1;
         ...
         typedef Y Type;
      public:
         CPM_IO
         Word nameOf()const{ return "Y";}
         ...
      };

      bool Y::prnOn(str)const
      {
         cpmwat; // uses the nameOf - function to write an
            // automatic 'title'
         cpmp(x1); // 'p' for 'print'. According to the
            // definition of X this makes sense as a part of
            // the implementation of Y's stream interface.
            ...
         return true;
      }

      bool Y::scanFrom(str)
      {
         cpms(x1); // 's' for 'scan'
           ...
         return true;
      }

      The details of the scheme will be given later in this file.

      2. Functions which enable unbiased testing of mathematical
      properties especially of class templates.

         ... not yet finished

      Z, N, R, L, bool, string (summerized as bb-types:
      basic built-in types). Thus no longer restricted to numbers.
      Provides most of the content of CpmRoot, cpmword adds the
      rest.

      We try to avoid the unsigned number types and thus don't introduce
      Nh, Nd as in earlier versions of this class library.
      (See. Bjarne Stroustroup: The C++ Programming Language,
      Third Edition, Addison-Wesley p. 70-71 last and first few lines.
      This book will referred to as BS3, the second edition as BS2.
      At present we have a fourth edition BS4.
      Certainly it is highly desirably to have the basic number
      types defined as classes. These seems however be possible only by
      reducing speed by factor 8, or so, in numerics intensive
      programs. This is not considered acceptable here.
      More recent experiments (in 2010) gave a speed loss of only 20% upon
      speed-optimized compilation. So the previous judgement seems no
      longer to have a very firm basis.

      We do not yet make use of the class CpmRoot::Word in this file.
*/

// making available the basic C++ standard library facilities
#include <string>
#include <iostream>
#include <fstream>  // for ifstream, ofstream
#include <sstream>
#include <cmath>
   // all math functions in namespace std ???
   // They can be also used as if they were in the global namespace.
   // This is important that we don't need a directive
   // 'using namespace std' in this file (as I had prior to 2016-08-29
   // and got a clash with a probably newly introduced function
   // std::ignore.
#include <limits>
#include <functional>

#include <cpmbasictypes.h>
   // Defines integer types Z, N , L.
   // Includes cpmdefinitions.h from which defines (or does not
   // define) the macro CPM_MP which controls the usage of
   // 'multiple precision' floating-point numbers.

#include <cpmbasicinterfaces.h>
   // defines macros CPM_IO and CPM_ORDER

#if defined(CPM_MP)
   #if defined(CPM_USE_EIGEN)
      #include<boost/multiprecision/eigen.hpp>
   #endif
   #include <boost/multiprecision/gmp.hpp>
   #include <boost/math/special_functions/round.hpp>
   #include <boost/math/special_functions.hpp>
   #include <boost/math/special_functions/bessel.hpp>
   #include <boost/math/constants/constants.hpp>
#elif defined(CPM_QUAD)
   #include <boost/multiprecision/float128.hpp>
#else
   #include <boost/math/special_functions/round.hpp>
   #include <boost/math/special_functions.hpp>
   #include <boost/math/special_functions/bessel.hpp>
   #include <boost/math/constants/constants.hpp>
#endif

namespace CpmStd{
// This is the part of the standard library which we use frequently.
// With C++11 we replace F1,F2,... with the replace of std::bind and
// namespace std::placeholders.
   using std::string;
   using std::ostream;
   using std::istream;
   using std::iostream;
   using std::ofstream;
   using std::ifstream;
   using std::fstream;
   using std::ostringstream;
   using std::istringstream;
   using std::stringstream;
   using std::endl;
   using std::cout;
   using std::cin;
   using std::bind;
   using namespace std::placeholders;
 } // CpmStd

namespace CpmRoot{
   using namespace CpmStd;

// Defining type R, the real type with which present C+- code uses
// in all places where floating point values are needed. No longer
// I use floating point types of multiple and fixed precision (such
// as R together with Rh or R_) in parallel.
#if defined(CPM_MP) // MP stands here for 'multiple precision'.
   // a typical line to appear in cpmdefinitions.h is
   // #define CPM_MP 32
   // where the number gives the number of decimal digits.
   // Here we use the tools from boost/multiprecision. These interface
   // well with boost/math but show strange behavior with respect to
   // C++ conversion operators and create a lot of warnings which at
   // the end of building are reported as errors by the Code::blocks
   // IDE. Nevertheless the result of building is valid and works.
   // Standard functions such as sqrt are also in this namespace.
   // No longer any warnings with g++-10.

    using namespace boost::multiprecision;
    using namespace boost::math;

    typedef number<gmp_float<CPM_MP>, et_off > R;
      //typedef number<gmp_float<CPM_MP>, et_on > R;
      // Notice the et_off-switch which disables expression-templates.
      // If these would left active compilation of nearly all of my C+-
      // code basis would fail. Since the Eigen library is said to build
      // heavly on expression templates disabling expression templates
      // probably compromises the efficiency of Eigen. Should be clarified!
      // Notice here numerical precision is set at compile time and can
      // not be changed by reading input from cpmconfig.ini.
#else
   // Then we don't need the multiple precision libraries and get
   // more speed. This was the only mode of C+- till March 2009. We have
   // the choice between double (64 bit) and long double (80 bit on most
   // systems) as a definition of type R.
   #if defined(CPM_FLOAT) //
      typedef float R;
   #elif defined(CPM_LONG) // then: typedef long int Z
      typedef long double R;
   #elif defined(CPM_QUAD)
      using namespace boost::multiprecision;
      using namespace boost::math;
      typedef number<float128_backend, et_off>  R;
   #else
      typedef double R;
   #endif
  // using namespace std; // experiment
#endif

// Remark on using CPM_LONG:
// C+- often overloads functions on the basis of argument type R or Z.
// As an example consider the following constructors in cpmviewport.h:
// CpmGraphics::rgb(Z r_, Z g_, Z b_, bool norm=true);
// CpmGraphics::rgb(R r_, R g_, R b_, bool norm=true);
// For R=double one can distinguish the two by calling e.g.
// rgb c1(255.,128.,10.); for the R-version, and
// rgb c2(0,0,255); for the Z-version.
// For R=long double the code will become ambiguous since interpreting
// 255 as long int is not considered better than interpreting it as long
// double. Better one avoids numerical literals as function arguments
// and writes
// R r=255, g=128, b=10; Z z0=0,z1=255;
// rgb c1(r,g,b); rgb c2(z0,z0,z1);
// This gives all quantities their matching type and is as readable as
// rgb c1(R(255),R(128),R(10)); rgb c2(Z(0),Z(0),Z(255));

// Two implementation tools:

inline R disDefFun(R a, R b, R d)
   //: distance defining function
   // Function for implementing function dis.
   // Needs also be known to cpminterfaces.h.
   // Continuous function, but not reasonable from a geometrical
   // point of view, since the distance dis(a,b):=disDefFun(|a|,|b|,|a-b|)
   // between points based naturally on this function is not translation
   // invariant.
{
   R s=a+b;
   if (s==0.) return R(0.);
   R d1=d/s;
   return (d < d1 ? d  : d1);
}

template <class T>
T posVal(T t){ return t<T() ? -t : t;}
   //: positive value
   // added 2005-09-10. Similar to abs but not necessarily
   // R-valued. Not intended to be used for aggregated data types

// Defining the basic R-related functions as class-less
// functions in namespace CpmRoot.

void prec(Z p);
   // Setting precision; a typical numerical error is ~ 10^(-p).
   // Thus p is a number of decimal digits.
   // Always defined, but active only if
   //    defined(CPM_MP)

Z getPrec(void);
   // Returns the active value of precision.
   // If !defined(CPM_MP) the corresponding value for type
   // double or long double (i.e. the actual type of R) is returned.
R inv(R const& r);
   // returns r==0. ? 0 : 1./r with warnings in the case r==0.
R randomR(Z j=0);
   // Returns uniformely distributed random values in [0,1) if
   // no argument (or argument 0) is used. For j != 0 the return
   // value is a 'chaotic' function of j, so that for any sequence
   // of argument values we get a random sequence. Since the value
   // is determined by j, one gets reproducible results which sometimes
   // is useful. In March 2013 I tested {0,1}-valued sequences of length
   // 1000000 with the NIST Random Number Test Suite and found all tests
   // passed convincingly. I used the NIST suite as translated to
   // Mathematica by Ilja Gerhardt in 2010 and I translated randomR also
   // to Mathematica.
R test(R const& r, Z i);
   // Implementation tool for class Test<>, see this file.
R ran(R const& r, Z i=0);
   // Implementation tool for class Ran<>, see this file.
   // The return value x satisfies -|r| <= x < |r|
Z hash(R const& r);
   // Implementation tool for class Hash<>, see this file.
R dis(R const& r1, R const& r2);
   // Implementation tool for class Dis<>, see this file.
Z getPrec(R const& r);
   // Returns the number of decimal digits used to hold the value of r.
bool isVal(R const& r);
   // Returns 'false' if r is not a regular number (infinite or NaN).
Z toZ(R const& r, bool toZero=false);
   // For positive r we return floor(r) as a Z instead of a R.
   // For negative r, the second argument determines whether the next
   // integer in minus-direction (toZero=false) or in plus-direction
   // (toZero=true) is returned. So, for toZero==false, the result is
   // floor(r) also for negative r.
// functions with inline implementation
inline Z sgn(R const& r)
{
   if (r>0.) return 1;
   if (r<0.) return -1;
   return 0;
}

inline R hypot(R const& x, R const& y)
// hypotenuse, a numerically careful version of sqrt(x*x + y*y).
{
   R tiny=1.e-12;
   R xa=posVal<R>(x),ya=posVal<R>(y),xya=xa+ya;
   if (xya==R(0.)) return R(0.);
   xya=(xya<tiny ? tiny : xya);
   R xyaInv=CpmRoot::inv(xya);
   R xr=xa*xyaInv, yr=ya*xyaInv;
   return R(xya*sqrt(xr*xr+yr*yr));
}

inline R con(R const& r){ return r;}
   // Implementation tool for class Conj<>, see this file.

inline R net(R const& r, Z i){ if (i==1) return R(1.); else return R(0.);}
   // Implementation tool for class Neutrals<>, see this file.

inline Z rnd(R const& r)
   // Returns the nearest integer value (rounding to integer).
{
#if defined(CPM_MP)
      return lround(r);
#else
   return (Z)floor(r+0.5);
#endif
}

//: division (of integers with remainder)
inline std::pair<Z,Z> div(Z x, Z a)
{
   auto q=std::div(x,a);
   return  std::pair<Z,Z>(q.quot, q.rem);
}

//: division (of unsigned integers with remainder)
inline std::pair<N,N> div(N x, N a)
{
   N quot=x/a;
   N rem=x%a;
   return  std::pair<N,N>(quot,rem);
}

//. remainder
// a assumed to be positive. This function matches the behavior of Mod in
// Wolfram Language.
inline Z modWL(Z x, Z a)
{
   if (x>=0_Z){
      return x%a;
   }else{
      Z q=x/a;
      Z qL=q-1_Z;
      return x-qL*a;
   }
}

inline bool isZero(R const& r){ return r==R(0.);}
   // Returns 'true' if r==0, and 'false' else.

inline double toDouble(R const& r){
   // to double, helps to program uniformly formatted screen output
   return static_cast<double>(r);
}

inline R absSqr(R const& r){ return r*r;}
   //: absolute (value) squared
   // Returns the square of |r| (which happens to be the square of r)

inline R arg(R const& x, R const& y)
   //: argument
   // Returns the polar angle (in (-Pi,Pi]) of point (x,y).
   { return atan2(y,x);}

// Split implementation for circle related matters.
#if !defined(CPM_MP)
   const R Pi=asin(0.5)*6.0;
      // Number pi=3.14159..., pi = 30*deg * 6, sin(30*deg) = 0.5.
      // Faster convergence expected then for Pi = atan(1.)*4.
      // This is certainly not an important issue in this place.
   const R Pi2=Pi*2;
      // 2 * pi
   const R AngDeg= Pi/180.;
      //: angular degree, pi/180
#endif

   R operator""_R(long double r);
   R operator""_R(unsigned long long int r);

} // CpmRoot closed, will be re-opened soon

   

// Convenient abbreviations for constants and service functions.
// See also corresponding abbreviations
// cpmnam, cpmtow in cpmword.h and
// cpmswp, cpmord, cpminf, cpmsup, cpmtin, cpmhug in cpmtypes.h.

#if defined CPM_MP
   #define cpmpi     R(boost::math::constants::pi<CpmRoot::R>())
   #define cpmpiInv  R(R(1.)/boost::math::constants::pi<CpmRoot::R>())
   #define cpm2pi    R(boost::math::constants::pi<CpmRoot::R>()*R(2.))
   #define cpmdeg    R(boost::math::constants::pi<CpmRoot::R>()/R(180.))
   #define cpmrho    CpmRoot::hypot
#else
   #define cpmpi     CpmRoot::Pi
   #define cpmpiInv  1./CpmRoot::Pi
   #define cpm2pi    CpmRoot::Pi2
   #define cpmdeg    CpmRoot::AngDeg
   #define cpmrho    CpmRoot::hypot
#endif

// easy access to C+- utility functions
// Notice that we are here not in CpmRoot.
#define cpmrnd    CpmRoot::rnd
#define cpmtoz    CpmRoot::toZ
#define cpmtod    CpmRoot::toDouble
#define cpmpos    CpmRoot::posVal
#define cpmprn    CpmRoot::prnOn
#define cpmscn    CpmRoot::scanFrom
#define cpmsgn    CpmRoot::sgn
#define cpmcom    CpmRoot::com
#define cpmcon    CpmRoot::con
#define cpmdis    CpmRoot::dis
#define cpmran    CpmRoot::ran
#define cpmtes    CpmRoot::test
#define cpmhas    CpmRoot::hash
#define cpmnet    CpmRoot::net
#define cpmiva    CpmRoot::isVal
#define cpminv    CpmRoot::inv
#define cpmab2    CpmRoot::absSqr
#define cpmarg    CpmRoot::arg
#define cpmprec    CpmRoot::prec
#define cpmgetprec CpmRoot::getPrec

// Easy access to mathematical functions of a real argument.
// CPM_ is either std or boost::multiprecision depending on CPM_MP.
// Since the function-names begin with 'cpm' it should not harm that they
// are defined here in the global namespace.
// Since these function names are defined already in namespace std it would
// be not convenient to define them first in CpmRoot and that use, for
// instance,
// #define cpmsin    CpmRoot::sin
// Instead, we define a function with name cpmsin.

#if defined(CPM_MP)||defined(CPM_QUAD)
   inline CpmRoot::R cpmabs(CpmRoot::R const& x)
      { return boost::multiprecision::abs(x);}
   inline CpmRoot::R cpmfloor(CpmRoot::R const& x)
      { return boost::multiprecision::floor(x);}
   inline CpmRoot::R cpmceil(CpmRoot::R const& x)
      { return boost::multiprecision::ceil(x);}
   inline CpmRoot::R cpmsin(CpmRoot::R const& x)
      { return boost::multiprecision::sin(x);}
   inline CpmRoot::R cpmasin(CpmRoot::R const& x)
      { return boost::multiprecision::asin(x);}
   inline CpmRoot::R cpmsinh(CpmRoot::R const& x)
      { return boost::multiprecision::sinh(x);}
   inline CpmRoot::R cpmcos(CpmRoot::R const& x)
      { return boost::multiprecision::cos(x);}
   inline CpmRoot::R cpmacos(CpmRoot::R const& x)
      { return boost::multiprecision::acos(x);}
   inline CpmRoot::R cpmcosh(CpmRoot::R const& x)
      { return boost::multiprecision::cosh(x);}
   inline CpmRoot::R cpmtan(CpmRoot::R const& x)
      { return boost::multiprecision::tan(x);}
   inline CpmRoot::R cpmtanh(CpmRoot::R const& x)
      { return boost::multiprecision::tanh(x);}
   inline CpmRoot::R cpmatan(CpmRoot::R const& x)
      { return atan(x);}
   inline CpmRoot::R cpmatan2(CpmRoot::R const& y, CpmRoot::R const& x)
      { return boost::multiprecision::atan2(y,x);}
   inline CpmRoot::R cpmexp(CpmRoot::R const& x)
      { return boost::multiprecision::exp(x);}
   inline CpmRoot::R cpmsqrt(CpmRoot::R const& x)
      { return boost::multiprecision::sqrt(x);}
   inline CpmRoot::R cpmlog(CpmRoot::R const& x)
      { return boost::multiprecision::log(x);}
   inline CpmRoot::R cpmlog10(CpmRoot::R const& x)
      { return boost::multiprecision::log10(x);}
   inline CpmRoot::R cpmpow(CpmRoot::R const& a, CpmRoot::R const& b)
      { return boost::multiprecision::pow(a,b);}
#if !defined(CPM_QUAD)
   inline CpmRoot::R cpmbessel(CpmRoot::Z const& n, CpmRoot::R const& x)
      { return boost::math::cyl_bessel_j(n,x);}
#else
    inline CpmRoot::R cpmbessel(CpmRoot::Z const& n, CpmRoot::R const& x){
   //    cpmwarning("cpmbessel(Z,R): not implemented, 0 returned ");
      using namespace CpmRoot;
      return 0_R;
   }
#endif

#else // double or float
   inline CpmRoot::R cpmabs(CpmRoot::R const& x)
      { return std::abs(x);}
   inline CpmRoot::R cpmfloor(CpmRoot::R const& x)
      { return std::floor(x);}
   inline CpmRoot::R cpmceil(CpmRoot::R const& x)
      { return std::ceil(x);}
   inline CpmRoot::R cpmsin(CpmRoot::R const& x)
      { return std::sin(x);}
   inline CpmRoot::R cpmasin(CpmRoot::R const& x)
      { return std::asin(x);}
   inline CpmRoot::R cpmsinh(CpmRoot::R const& x)
      { return std::sinh(x);}
   inline CpmRoot::R cpmcos(CpmRoot::R const& x)
      { return std::cos(x);}
   inline CpmRoot::R cpmacos(CpmRoot::R const& x)
      { return std::acos(x);}
   inline CpmRoot::R cpmcosh(CpmRoot::R const& x)
      { return std::cosh(x);}
   inline CpmRoot::R cpmtan(CpmRoot::R const& x)
      { return std::tan(x);}
   inline CpmRoot::R cpmtanh(CpmRoot::R const& x)
      { return std::tanh(x);}
   inline CpmRoot::R cpmatan(CpmRoot::R const& x)
      { return std::atan(x);}
   inline CpmRoot::R cpmatan2(CpmRoot::R const& y, CpmRoot::R const& x)
      { return std::atan2(y,x);}
   inline CpmRoot::R cpmexp(CpmRoot::R const& x)
      { return std::exp(x);}
   inline CpmRoot::R cpmsqrt(CpmRoot::R const& x)
      { return std::sqrt(x);}
   inline CpmRoot::R cpmlog(CpmRoot::R const& x)
      { return std::log(x);}
   inline CpmRoot::R cpmlog10(CpmRoot::R const& x)
      { return std::log10(x);}
   inline CpmRoot::R cpmpow(CpmRoot::R const& a, CpmRoot::R const& b)
      { return std::pow(a,b);}
   inline CpmRoot::R cpmbessel(CpmRoot::Z const& n, CpmRoot::R const& x)
      { return std::cyl_bessel_j(n,x);}
#endif

// namespace re-opened
namespace CpmRoot{

// Numerical precision matters.
   extern Z numPrc;
      //: numerical precision
      // value for decimals used for the internal representation
      // of real numbers of type R.
// Formatting matters.

   extern Z wrtPrc;
      //: writing precision
      // value for decimals written to strings for real
      // numbers of type R. Notice that all data communicated
      // between processes in my MPI-based parallel computing
      // functions use writing on streams; so numerical errors in
      // communication are ignorable only if this number is large
      // e.g. 16. Now these functions set wrtPrc high and reset
      // it afterwards to a normal value (automatically!)

   bool writeTitle(const string& , ostream&);
      // write title
      // Uses the string argument to write a title which is precceced
      // by a comment line indicator. Is under the control of wrtTit.

   bool startsComment(char c);
      // starts comment
      // Returns true if c is a character which gives a line the
      // meaning of a comment if it is its first non-whitespace
      // character.
      // Presently startsComment(c)==true for
      // c \in { ';','/','#','*'}

   bool eatComments(istream& in);
      // eat comments
      // Returns false if no character not belonging to a comment
      // was found. If return is true, the stream is now in a
      // state that the next reading access will find something
      // not belonging to a comment.

// Unified interface to types Z,N,R,L,bool,string,
// which we refer to as bb-types (basic built-in types). Notice that the
// types Z, N, and L are defined via typedef and thus are no classes.
// R may be a class - depending on the macro CPM_MP.
// The types bool and string from standard C++ are useful as they stand.
// Nevertheless they will be wrapped into classes - CpmRootX::B in
// cpmtypes.h and class CpmRoot::Word in file cpmword.h.
// --------

   bool strVal(istream& str);
      //: stream validity
      // returns stream status

   bool strVal(ostream& str);
      //: stream validity
      // returns stream status

// Topic 1: reading instances of b-types from streams that may contain
// comments.

   template <class T>
   bool read(T& t, istream& in)
   //: read
   // reads a T from a istream; in doing so it looks out for characters
   // having the meaning of 'comment indicators'. After having identified
   // such a character, the reading action will jump to the beginning of
   // the next line (thus not reading the rest of the previous line,
   // which is supposed to be a comment not containing data for reading).
   // Of course, a comment indicator there, will result in skipping that
   // line too. See function startsComment for the characters that are
   // considered comment indicators.
   // If no T can be read the function returns
   // false otherwise true. Providing a program with numerical data from
   // a text file created by the programmer is useful only if the file is
   // structured by comments explaining the meaning of the data. Thus
   // reading from files containing comment lines is crucial.
   // On ENUMERATIONS:
   // Not only quantities of type Z but also enumerations will be
   // read by this function. Then the syntax of writing and reading
   // should be as follows:
/*
   enum Corners{LL,UL,LR,UR};
   Corners c;
   write((Z)c,cout);
   ...
   Z temp;
   read(temp,cin);
   Corners c2=(Corners)temp;
*/
   {
      if  (!eatComments(in)) return false;
      if (in>>t) return true; else return false;
   }

   template <>
   bool read(L& x, istream& in);

   bool readLine(string& str, istream&);
   // reads a non-comment line from a istream into str. Trailing
   // whitespace such as the most annoying such species CR (carriage
   // return) will not find a way to spoil str. Leading whitespace
   // will not be ignored here since it is essential for copying
   // texts in a readable form by reading and printing lines

// Topic 2: writing instances of b-types to streams
// in a similar syntax, to make large prnOn and scanFrom functions of
// classes more alike

   template <class T>
   bool write(T const& t, ostream& out)
   {
      out<<t<<endl; return strVal(out);
   }

   template <>
   bool write(L const& t, ostream& out);

   template <>
   bool write(R const& t, ostream& out);
      // special treatment to avoid problems with small
      // numbers (which should not exist)

// Miscellanea

   bool isVal(L x);
   bool isVal(Z x);
   bool isVal(N x);
      // No modes of invalidity of types L, bool, or string are
      // known to me. The real types are already treated.

   L getByte(Z z, Z n);
      // z is an (typically 32 bit)-integer and n varies over 1,2,3,4
      // returned is the first, second, third, forth byte (counted from
      // left to right)

   string skipLeadingWhitespace(string const& str);
      // clear from name

   string skipTrailingWhitespace(string const& str);
      // clear from name

//////////////// interface service classes ///////////////////////////////
// Unified interface to the bb-types
//    Z, N, R, L, bool, string
// and most of the other C+- classes.
// This uniform interface is given by the interface service class
// templates
//    IO<T>, Comp<T>, Inv<T>, AbsSqr<T>, Abs<T>, Dis<T>,
//    Test<T>, Ran<T>, Hash<T>, Conj<T>, Neutrals<T>,
//    ToWord<T>, Name<T> (these two to be added later in cpmword.h)
// For a class T to implement this interface means to define the following
// member functions ( here an T as argument may also mean T const&):
//    bool prnOn(ostream& str)const;         // print on
//    bool scanFrom(istream& str);           // scan from
//    Z com(T)const;                         // compare
//    T inv()const;                          // inverse
//    R absSqr()const;                       // absolute (value) squared
//    R abs()const;                          // absolute value
//    R dis(T)const;                         // distance value
//    T test(Z)const;                        // test value
//    T ran(Z)const;                         // random value
//    Z hash()const;                         // hash value
//    T con()const;                          // conjugate
//    Z net(Z)const;                         // neutrals ( = 0 and 1)
//    Word nameOf()const;                  // class name (see cpmword.h)
//    Word toWord()const;                  // string representation of
                                           // the value  (see cpmword.h)
// For the non-class b-types, the interface service class templates are
// subsequently defined as specializations.

///////////////////////// class IO<> /////////////////////////////////////
// Here we describe for the important example of stream IO
// (marshalling) the method by which C+- succeeds to treat
// the non-class b-types and C+- classes on equal
// footing when these are used as template arguments of
// C+- array classes.
// For a C+- class T it is normal to define the member functions
// (1) bool T::prnOn(ostream& str)const;
// (2) bool T::scanFrom(istream& str);
// Their declaration (together with the definition of some
// related non-member functions) is provided by using the
// macro CPM_IO after a statement 'typedef T Type;'
// For a C+- class template, e.g. CpmArrays::V<T> it is  n o t
// natural to assume that T defines the member functions
// (1) and (2) since T could be Z or R and thus not a class at all.
// Instead, all C+- class templates which implement
// interaction with streams assume that the template argument T
// defines functions
// (3) bool CpmRoot::IO<T>().o(T const& t, ostream& str)const;
// (4) bool CmRoot::IO<T>().i(T& t, istream& str)const;
// There are two possibilities for these functions to get defined:
// (i) It could be that functions (1),(2) are defined for T
//    e.g. since T is a C+- class implementing CPM_IO.
// Then the following un-specialized version of the class template IO<T>
// does the job.
// If, however, T is not a C+- class and cannot be modified by
// adding member functions (1),(2) then
// (ii) a specialization of the class template
//    IO has to make IO<T> defined: Such specializations
//    are given in the following for the  bb-types.
//    Partial specializations (in which T is a class template)
//    can be defined where needed. An example is the definition
//    of IO< std::vector<T> > in cpmv.h.
// Finally there are class-less function templates (near the end of the
// file)
// bool prnT(T const& t, ostream& str)
// bool scanT(T& t, istream& str)
// which will be called in the covenient read and write macros
// #define cpms(X) if (!CpmRoot::scanT((X),str)) return false
// #define cpmp(X) if (!CpmRoot::prnT((X),str)) return false
// fom file cpminterfaces.h. These functions can also be used in
// template code instead of the more clumpsy functions (3),(4).
// It should be noted that none of the interface service classes
// IO, Comp, Inv, AbsSqr, Abs, Dis, Test, Ran,
// Hash, Conj, Neutrals contains its own data, so that no transport
// of data is involved in their use. Notice that there are two more
// such service classes: Name and ToWord in cpmword.h.
// On Name the C+- type naming system is based, and ToName provides
// convenient short output string representation of values.
// The send and receive functions for MPI usage are based on the
// present IO-class (see macros CPM_MPI_O and #define CPM_MPI_I
// in cpminterfaces.h).
// For all such service classes except IO there is only one member
// function (not two, namely i and o, in IO) and this is chosen to
// be operator().

template<class T>
class IO{ // input output class template
public:
   IO(){}
   bool o(T const& t, ostream& str)const
      { return t.prnOn(str);}
   bool i(T& t, istream& str)const
      { return t.scanFrom(str);}
};

template<>
class IO<R>{ // specialization of IO
public:
   IO(){}
   bool o(R const& t, ostream& str)const
      { return CpmRoot::write(t,str);}
   bool i(R& t, istream& str)const
      { return CpmRoot::read(t,str);}
};

template<>
class IO<Z>{ // specialization of IO
public:
   IO(){}
   bool o(Z const& t, ostream& str)const
      { return CpmRoot::write(t,str);}
   bool i(Z& t, istream& str)const
      { return CpmRoot::read(t,str);}
};

template<>
class IO<N>{ // specialization of IO
public:
   IO(){}
   bool o(N const& t, ostream& str)const
      { return CpmRoot::write(t,str);}
   bool i(N& t, istream& str)const
      { return CpmRoot::read(t,str);}
};

template<>
class IO<L>{ // specialization of IO
public:
   IO(){}
   bool o(L const& t, ostream& str)const
      { return CpmRoot::write(t,str);}
   bool i(L& t, istream& str)const
      { return CpmRoot::read(t,str);}
};

template<>
class IO<bool>{ // specialization of IO
public:
   IO(){}
   bool o(bool const& t, ostream& str)const
      { return CpmRoot::write(t,str);}
   bool i(bool& t, istream& str)const
      { return CpmRoot::read(t,str);}
};

template<>
class IO<string>{ // specialization of IO
public:
   IO(){}
   bool o(string const& t, ostream& str)const
      { return CpmRoot::write(t,str);}
   bool i(string& t, istream& str)const
      { return CpmRoot::read(t,str);}
};

///////////////////////// class Comp<> ///////////////////////////////////
// class is named Comp instead of Com in order to avoid confusion
// with CpmMPI::Com
// Ruby style comparison
// com = 'compared' corresponds to the negative of
// Ruby's intelligent operator <=> (only the symbol '<=>' together
// with the ordered list -1,0,1 suggests Ruby's choice; if one is not
// bound by a symbol it is more natural to associate a<b (odered) with
// a.com(b)==1 and a>b (converse order) with a.com(b)==-1.
// Thus our normation:
// a.com(b) ==  0  for  a == b
// a.com(b) ==  1  for  a <  b
// a.com(b) == -1  for  a >  b

template<class T>
class Comp{ // compare
public:
   Comp(){}
   Z operator()(T const& x, T const& y){ return x.com(y);}
};

// Specializations building on > and < :
#define CPM_SC\
 if (x<y) return 1; if (x>y) return -1; return 0

template<>
class Comp<R>{
public:
   Comp(){}
   Z operator()(R const& x, R const& y){CPM_SC;}
};

template<>
class Comp<Z>{
public:
   Comp(){}
   Z operator()(Z const& x, Z const& y){CPM_SC;}
};

template<>
class Comp<N>{
public:
   Comp(){}
   Z operator()(N const& x, N const& y){CPM_SC;}
};

template<>
class Comp<L>{
public:
   Comp(){}
   Z operator()(L const& x, L const& y){CPM_SC;}
};

template<>
class Comp<bool>{
public:
   Comp(){}
   Z operator()(bool const& x, bool const& y){CPM_SC;}
};

template<>
class Comp<string>{
public:
   Comp(){}
   Z operator()(string const& x, string const& y){CPM_SC;}
};

#undef CPM_SC

///////////////////////// class Inv<> ////////////////////////////////////

template<class T>
class Inv{ // inverse
public:
   Inv(){}
   T operator()(T const& t)const{ return t.inv();}
};

template<>
class Inv<R>{ // inverse
public:
   Inv(){}
   R operator()(R const& t)const{ return cpminv(t);}
};

template<>
class Inv<Z>{ // inverse
public:
   Inv(){}
   Z operator()(Z const& t)const{ t; return Z(0);}
};

template<>
class Inv<N>{ // inverse
public:
   Inv(){}
   N operator()(N const& t)const{ t; return N(0);}
};

template<>
class Inv<L>{ // inverse, defined in an arbitrary manner
public:
   Inv(){}
   L operator()(L const& t)const{ return t;}
};

template<>
class Inv<bool>{ // inverse
public:
   Inv(){}
   bool operator()(bool const& t)const{ return !t;}
};

template<>
class Inv<string>{ // reverse string
public:
   Inv(){}
   string operator()(string const& t)const
   {
      string::const_reverse_iterator p1=t.rbegin(), p2=t.rend();
      return string(p1,p2);
   }
};

//////////////////////// class Test<> /////////////////////////////////
// The intented usage is as follows:
//    using CpmRoot;
//    T t;
//    Z complexity=3; // for example
//    T tc=Test<T>()(t,complexity);
//    T t1=Ran<T>()(tc,1);
//    T t2=Ran<T>()(tc,2);
//    T t3=Ran<T>()(tc,3);
// which is an unbiased way to create an arbitrarily large set of T-values.

template<class T>
class Test{ // test value
public:
   Test(){}
   T operator()(T const& t, Z complexity)const
   { return t.test(complexity);}
};

template<>
class Test<R>{
public:
   Test(){}
   R operator()(R const& x, Z complexity)const
   { return cpmtes(x,complexity);}
};

template<>
class Test<Z>{
public:
   Test(){}
   Z operator()(Z const& x, Z complexity)const
   { /*x;*/ return complexity;}
};

template<>
class Test<N>{
public:
   Test(){}
   L operator()(N const& x, Z complexity)const
   { x; Z cp=posVal(complexity); return N(cp);}
};

template<>
class Test<L>{
public:
   Test(){}
   L operator()(L const& x, Z complexity)const
   { x; Z cp=posVal(complexity); return L(cp%256);}
};

template<>
class Test<bool>{
public:
   Test(){}
   bool operator()(bool const& x, Z complexity)const
   { x; Z cp=posVal(complexity); return cp%2==1;}
};

template<>
class Test<string>{
public:
   Test(){}
   string operator()(string const& x, Z complexity)const
   {
      x;
      if (complexity<=1)
         return "az";
      else if (complexity==2)
         return "abcdefghijklmnopqrstuvwxyz";
      else if (complexity==3)
         return "abcdefghijklmnopqrstuvwxyz0123456789";
      else return
         "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
   }
};

/////////////////////// class Ran<> ///////////////////////////////////

Z randomZ(Z n, Z j=0);
   // returns uniformly distributed integers in the natural
   // index range [1,n]. Thus stops with error if n<1.
   // Function IvZ::ran allows to set lower limit and upper limit
   // arbitrarly.

template<class T>
class Ran{ // random value
public:
   Ran(){}
   T operator()(T const& t, Z j=0)const
   { return t.ran(j);}
};

template<>
class Ran<R>{
public:
   Ran(){}
   R operator()(R const& x, Z j=0)const
   {
     return cpmran(x,j);
   }
};

template<>
class Ran<Z>{
public:
   Ran(){}
   Z operator()(Z const& i, Z j=0)const
   {
      R y=randomR(j);
      Z ia=(i>=0 ? i : -i);
      if (ia==0) ia=1;
      // without this modification we would get 0,0,0,... for i=0. This
      // 'random' sequence will probably never be needed in a serious
      // context.
      // More probably i itself will result from a random process and
      // nevertheless ran(i,i2) will be expected to be random in i2.
      // Thus we increase ia in order to get the random sequence with
      // values in {-1,0,1}.
      Z n=ia*2+1;
      y*=n; // belongs to [0,2|i|+1)
      Z res=cpmtoz(y);
      if (res==n) res--; // should not happen
      return res-ia;
   }
};

template<>
class Ran<N>{
public:
   Ran(){}
   N operator()(N const& i, Z j=0)const
   {
      R y=randomR(j);
      Z ip=(i==0 ? 1 : i);
      // without this modification we would get 0,0,0,... for i=0. This
      // 'random' sequence will probably never be needed in a serious
      // context.
      // More probably i itself will result from a random process and
      // nevertheless ran(i,i2) will be expected to be random in i2.
      // Thus we increase ip in order to get the random sequence with
      // values in {0,1}.
      y*=ip; // belongs to [0,ip)
      Z res=cpmtoz(y);
      if (res==ip) res--; // should not happen
      return N(res); // belongs to {0,1} for i==0 and to {0,...i-1} else.
   }
};

template<>
class Ran<L>{
public:
   Ran(){}
   L operator()(L const& x, Z j=0)const
   {
      x;
      R y=randomR(j);
      y*=256; // belongs to [0,256)
      Z zy=cpmtoz(y); // belongs to {0,1,..., 255}=
      if (zy==256) zy--; // should not happen
      return (L)zy;
   }
};

template<>
class Ran<bool>{
public:
   Ran(){}
   bool operator()(bool const& b, Z j=0)const
   {
      R y=randomR(j);
      if (y<=0.5) return b; else return !b;
   }
};

// Here, all random strings have only printable ascii characters. See the
// definition of myChars for details.
template<>
class Ran<string>{
public:
   Ran(){}
   string operator()(string const& a, Z i=0)const
   {
      string res=a;
      Z l=i;
      string::const_iterator j;
      string::iterator k;
      const char myChars[] {'a','b','c','d','e','f','g','h','i','j','k',
      'l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
      'A','B','C','D','E','F','G','H','I','J','K',
      'L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
      '0','1','2','3','4','5','6','7','8','9','_','.'};
      const int nc=64;
      for (j=a.begin(),k=res.begin();j!=a.end(),k!=res.end();++j,++k){
         int z = *j;
         int r = z%nc + 1;
         int ii = i;
         if (ii!=0){
            l++;
            ii+=l;
         }
         int q = randomZ(nc,ii) + r;
         *k = myChars[q%nc];
      }
      return res;
   }
};

///////////////////// class Hash<> ////////////////////////////////////
// If for T t1,t2 one has hash(t1)==hash(t2) there
// should be a overwhelming probability that t1 is equal
// to t2 (unless t2 is obtained from t1 by only a minor
// modification, or by a modification which is aware
// of the algorithm for hash).

template<class T>
class Hash{ // hash value
public:
   Hash(){}
   Z operator()(T const& t)const
   { return t.hash();}
};

template<>
class Hash<R>{
public:
   Hash(){}
   Z operator()(R const& x)const
   {
      return cpmhas(x);
   }
};

template<>
class Hash<Z>{
public:
   Hash(){}
   Z operator()(Z const& x)const
   {  return x-x%137+(x/13)*7;}
};

template<>
class Hash<N>{
public:
   Hash(){}
   N operator()(N const& x)const
   {  return x+x%137+(x/13)*7;}
};

template<>
class Hash<L>{
public:
   Hash(){}
   Z operator()(L const& x)const
   {
      Z y=(Z)x+117;
      return Hash<Z>()(y);
   }
};

template<>
class Hash<bool>{
public:
   Hash(){}
   Z operator()(bool const& x)const
   { return x ? 137 : 13;}
};

template<>
class Hash<string>{
public:
   Hash(){}
   Z operator()(string const& x)const
   {
      Z sum=137;
      string::const_iterator i;
      for (i=x.begin();i!=x.end();++i) sum+=(Z)(*i);
      return Hash<Z>()(sum);
   }
};

//////////////////// class Conj<> ////////////////////////////////////////
// This describes a concept that does not convincingly arrises
// from the types under present consideration. It will
// represent complex conjugation of complex numbers, transposition
// of real matrices, and Hermitean conjugation of complex
// matrices.

template<class T>
class Conj{ // conjugation
public:
   Conj(){}
   T operator()(T const& t)const
   { return t.con();}
};

template<>
class Conj<R>{
public:
   Conj(){}
   R operator()(R const& t)const
   { return cpmcon(t);}
};

template<>
class Conj<Z>{
public:
   Conj(){}
   Z operator()(Z const& t)const
   { return t;}
};

template<>
class Conj<N>{
public:
   Conj(){}
   N operator()(N const& t)const
   { return t;}
};

template<>
class Conj<L>{
public:
   Conj(){}
   L operator()(L const& t)const
   { return t;}
};

template<>
class Conj<bool>{
public:
   Conj(){}
   bool operator()(bool const& t)const
   { return t;}
};

template<>
class Conj<string>{
public:
   Conj(){}
   string operator()(string const& t)const
   { return t;}
};

//////////////////// class Neutrals<> ////////////////////////////////////
// for T t; the quantity Neutrals<T>(t,0) is the
// instance of T that is zero or comes as close as possible
// to realising the concept of a zero of type T.
// Correspondingly Neutrals<T>(t,1) with zero
// replaced by unity. The name comes from the fact that
// zero is a neutral element for addition and
// unity is a neutral element for multiplication.
// This renders the concept ambiguous for T==string since
// here we do refer only to addition (concatenation)
// but not to multiplication.

template<class T>
class Neutrals{ // neutral elements
public:
   Neutrals(){}
   T operator()(T const& t, Z i)const
   { return t.net(i);}
};

template<>
class Neutrals<R>{
public:
   Neutrals(){}
   R operator()(R const& t, Z i)const
   { return cpmnet(t,i);}
};

template<>
class Neutrals<Z>{
public:
   Neutrals(){}
   Z operator()(Z const& t, Z i)const
   { t; return i==1 ? 1 : 0;}
};

template<>
class Neutrals<N>{
public:
   Neutrals(){}
   N operator()(N const& t, Z i)const
   { t; return i==1 ? N(1) : N(0);}
};

template<>
class Neutrals<L>{
public:
   Neutrals(){}
   L operator()(L const& t, Z i)const
   { t; return i==1 ? L(1) : L(0);}
};

template<>
class Neutrals<bool>{
public:
   Neutrals(){}
   bool operator()(bool const& t, Z i)const
   { t; return i==1 ? true : false;}
};

template<>
class Neutrals<string>{
public:
   Neutrals(){}
   string operator()(string const& t, Z i)const
   { t; return i==1 ? "1" : "";}
};

// Now some classes will be defined which use in their interface
// a common real type, which is chosen as R. This type appears here
// as a return value of operator() of classes
// AbsSqr<>, Abs<>, and Dis<>. The definition of class Root<> in
// cpmword.h puts this role into perspective.
////////////////// class AbsSqr<> ////////////////////////////////////////

template<class T>
class AbsSqr{ // absolute (value) squared
public:
   AbsSqr(){}
   R operator()(T const& t)const{ return t.absSqr();}
};

template<>
class AbsSqr<R>{ // absolute (value) squared
public:
   AbsSqr(){}
   R operator()(R const& x)const{ return cpmab2(x);}
};

template<>
class AbsSqr<Z>{ // absolute (value) squared
public:
   AbsSqr(){}
   R operator()(Z const& x)const{ return R(x)*R(x);}
};

template<>
class AbsSqr<N>{ // absolute (value) squared
public:
   AbsSqr(){}
   R operator()(N const& x)const{ return R(x)*R(x);}
};

template<>
class AbsSqr<L>{ // absolute (value) squared
public:
   AbsSqr(){}
   R operator()(L const& x)const{
      Z y = static_cast<Z>(x);
      R r = y;
      return r * r;
   }
};

template<>
class AbsSqr<bool>{ // absolute (value) squared
public:
   AbsSqr(){}
   R operator()(bool const& x)const
      { return x==false ? R(0) : R(1);}
};

template<>
class AbsSqr<string>{
   // here string is interpreted as an array of L's
public:
   AbsSqr(){}
   R operator()(string const& x)const
   { // string components are cased to type L
      R res=0;
      string::const_iterator i;
      for (i=x.begin();i!=x.end();++i) res+=AbsSqr<L>()((L)(*i));
      return res;
   }
};

////////////////// class Abs<> ////////////////////////////////////////

template<class T>
class Abs{ // absolute value
public:
   Abs(){}
   R operator()(T const& t)const{ return t.abs();}
};

template<>
class Abs<R>{ // absolute value
public:
   Abs(){}
   R operator()(R const& x)const{ return cpmabs(x);}
};

template<>
class Abs<Z>{ // absolute value
public:
   Abs(){}
   R operator()(Z const& x)const{ return x<0 ? R(-x) : R(x) ;}
};

template<>
class Abs<N>{ // absolute value
public:
   Abs(){}
   R operator()(N const& x)const{ return R(x);}
};

template<>
class Abs<L>{ // absolute value
public:
   Abs(){}
   R operator()(L const& x)const{ return R(x);}
};

template<>
class Abs<bool>{ // absolute value
public:
   Abs(){}
   R operator()(bool const& x)const{ return x==false ? R(0) : R(1) ;}
};

template<>
class Abs<string>{
   // here string is interpreted as an array of L's
public:
   Abs(){}
   R operator()(string const& x)const
   { return sqrt(AbsSqr<string>()(x)) ;}
};

////////////////////// class Dis<> ////////////////////////////////////
// For T t1,t2 the real number Dis<T>()(t1,t2) belongs to [0,1]
// and tries to assess whether t1 and t2 are significantly different
// or differ only by numerical noise. Thus for discrete types we
// set 0 for t1==t2 and 1 else. For 'continuous types' we use
// the natural construct dis(x,y)=Min( |x-y|, |x-y|/(|x|+|y|) )
// = distFunc(|x|,|y|,|x-y|)
// based on the concepts of absolute value and difference.

template<class T>
class Dis{
public:
   Dis(){}
   R operator()(T const& x, T const& y){ return x.dis(y);}
};

template<>
class Dis<R>{
public:
   Dis(){}
   R operator()(R const& x, R const& y)
   { return cpmdis(x,y);}
};

template<>
class Dis<Z>{
public:
   Dis(){}
   R operator()(Z const& x, Z const& y)
   { return x!=y ?  1. : 0.;}
};

template<>
class Dis<N>{
public:
   Dis(){}
   R operator()(N const& x, N const& y)
   { return x!=y ?  1. : 0.;}
};

template<>
class Dis<L>{
public:
   Dis(){}
   R operator()(L const& x, L const& y)
   { return x!=y ?  1. : 0.;}
};

template<>
class Dis<bool>{
public:
   Dis(){}
   R operator()(bool const& x, bool const& y)
   { return x!=y ?  1. : 0.;}
};

template<>
class Dis<string>{
public:
   Dis(){}
   R operator()(string const& x, string const& y)
   { return x!=y ?  1. : 0.;}
};

bool apprInt(R x, R tol);
   // returns true if the distance between x and the nearest
   // integer is <= tol (if tol<0, this is never the case)

////////////////////////// class Conv ////////////////////////////////////
template<class Ti, class Tf>
class Conv{
// conversion
// Allows to define conversion between basic types in a manner
// which differes from that provided by the compiler.
   Ti x_;
public:
   Conv(Ti const& t):x_(t){}
   Tf operator()(void){ return (Tf)x_;}
};

template<>
class Conv<R,L>{
   // The image classes Image<T> and MultiImage<T> have to convert
   // between T and R in some functions. By far the most important
   // case is T=R so that no real conversion is involved. However
   // also T = L occurs. This case is handled here.
   R x_;
public:
   Conv(R const& t):x_(t){}
   L operator()(void){ return (L)cpmrnd(x_);}
};

template<>
class Conv<L,R>{
   L x_;
public:
   Conv(L const& t):x_(t){}
   R operator()(void){ return (R)(Z)x_;}
};

/////////////////// class-less function templates ////////////////////////
// This is a convenient form for using the service functions:
// The final T in the name is to suggest "template" and avoids name
// clashes with functions already defined for T = R with names inv, abs,
// hash, ran, ...
// See the comments on template <class T> class Root at the end of this
// file. Calling these functions is possible without mentioning the type
// of the argument, e.g. prnT(t,cout) which makes them usable in macros
// such as cpmp(X) and cpms(X).

// implementation of CPM_IO
template<class T>
bool prnT(T const& t, ostream& str)
   // print on
{ return CpmRoot::IO<T>().o(t,str);}

template<class T>
bool scanT(T& t, istream& str)
   // scan from
{ return CpmRoot::IO<T>().i(t,str);}

// implementation of CPM_ORDER
template<class T>
Z comT(T const& t1, T const& t2){ return Comp<T>()(t1,t2);}
   //: compare

// implementation of CPM_TEST_ALL
template<class T>
T invT(T const& t){ return Inv<T>()(t);}
   //: inverse

template<class T>
R absSqrT(T const& x){ return AbsSqr<T>()(x);}
   // absolute (value) squared

template<class T>
R absT(T const& x){ return Abs<T>()(x);}
   // absolute value

template<class T>
R disT(T const& t1, T const& t2){ return Dis<T>()(t1,t2);}
   // distance value

template<class T>
T testT(T const& t, Z complexity){ return Test<T>()(t,complexity);}
   // test value

template<class T>
T ranT(T const& t, Z j=0){ return Ran<T>()(t,j);}
   // random value

template<class T>
Z hashT(T const& t){ return Hash<T>()(t);}
   //: hash value

template<class T>
T conT(T const& t){ return Conj<T>()(t);}
   //: conjugate

template<class T>
T netT(T const& t, Z i){ return Neutrals<T>()(t,i);}
   // neutrals
} // CpmRoot

// A probably more convenient collection of these service functions
// is given by the class template CpmRoot::Root<> defined in cpmword.h.
// The class Root<T> has for each of the previous class-less functions
// a member function of the same functionality. For instance:
// T t1 = ... ;
// T t2 = ... ;
// R d = dis(t1,t2); ( = cpmdis(t1,t2), see the defines above )
// could also be written
// R d = Root<T>(t1).dis(t2);
// Here the Root-template needs the template argument T, which for the
// function template is optional (we could also write
// R d = disT<T>(t1,t2);).
// This explicit indication of the template argument helps more the reader
// than the compiler (who should know what to do anyway).
// Notice that definition of the I/O macro cpmp(X) and cpms(X) is
// easily based on functions prnOnT and scanFromT. As defined in
// cpminterfaces.h:
// #define cpmp(X) if (!prnOnT((X),str)) return false
// Probably also
// Root<typeid((X))>((X)).prnOn(str) would work. But I want to avoid
// the overhead incured by using typeid.

#endif
