//? cpmobservable.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmv.h>
#include <cpmobservable.h>
#include <cpmrmateig.h>
#include <cpmcmatrix.h>
#include <cpmconstphys.h>
#include <cpmangle.h>
#include <cpmframes.h>
#include <cpmmultiimage.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::State;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::SpectralObject;
using CpmQuantumMechanics::SpectralMeasure;

using namespace CpmRoot;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmGraphics;

using CpmRootX::Record;
using CpmRootX::krn;
using CpmGeo::Angle;
using CpmLinAlg::R3;
using CpmLinAlg::Z2;
using CpmImaging::Color;
using CpmGraphics::Frames;

////////////////////// class SpectralObject ////////////////////////////

SpectralObject::SpectralObject(Z d):d_(d),spectrum(d),eigenVectors(d)
{
   for (Z i=1;i<=d_;i++){
      spectrum[i]=C(i);
      eigenVectors[i]=State(d_,i);
   }
}

SpectralObject::SpectralObject(const Word& w, Z d):
   d_(d),spectrum(d_),eigenVectors(d_)
{
   Z mL=2;
   Word loc( "SpectralObject(" );
   loc&=( w&", "&cpm(d_)&")" );
   CPM_MA
   cpmassert(d_>0,loc);
   Z i;
   for (i=1;i<=d_;i++) eigenVectors[i]=State(d_,i);
      // normalized states

   if (w=="space frequency"||w=="f"){
      V<R> fr=frequencies(d_);
      for (i=1;i<=d_;++i){
         spectrum[i]=C(fr[i]);
         eigenVectors[i]=eigenVectors[i].frmMomSpc();
      }
      normalize_();
   }
   else if (w=="momentum"||w=="p"){
      V<R> fr=frequencies(d_);
      for (i=1;i<=d_;++i){
         spectrum[i]=C(fr[i]*hPlanck);
         eigenVectors[i]=eigenVectors[i].frmMomSpc();
      }
      normalize_();
   }
   else if (w=="wave vector"||w=="k"){
      V<R> fr=frequencies(d_);
      for (i=1;i<=d_;++i){
         spectrum[i]=C(fr[i]*cpm2pi);
         eigenVectors[i]=eigenVectors[i].frmMomSpc();
      }
      normalize_();
   }
   else if (w=="kinetic energy"||w=="eKin"){
      V<R> fr=frequencies(d_);
      R fac=R(1)/(mass*2);
      for (i=1;i<=d_;++i){
         R pi=fr[i]*hPlanck; // fr[i]=1/lambda
         spectrum[i]=C(pi*pi*fac);
         eigenVectors[i]=eigenVectors[i].frmMomSpc();
      }
      normalize_();
   }
   else if (w=="relativistic kinetic energy"||w=="eKinRel"){
      R e0=mass*c0*c0;
      V<R> fr=frequencies(d_);
      for (i=1;i<=d_;++i){
         R si=fr[i]*hPlanck/c0; // pi/c
         R vali=cpmsqrt(e0*e0-si*si)-e0;
         spectrum[i]=C(vali);
         eigenVectors[i]=eigenVectors[i].frmMomSpc();
      }
      normalize_();
   }
   else if (w=="zero"||w=="0"){
      for (i=1;i<=d_;i++) spectrum[i]=C(0);
   }
   else if (w=="unit"||w=="1"){
      for (i=1;i<=d_;i++) spectrum[i]=C(1);
   }
   else if (w=="position"||w=="x"){
      V<R> pos=positions(d_);
      for (i=1;i<=d_;i++){
         spectrum[i]=C(pos[i]);
      }
   }
   else if (w=="boost"||w=="n"){
      V<R> pos=positions(d_);
      for (i=1;i<=d_;i++){
         spectrum[i]=C(pos[i]*mass);
      }
   }
   else{ // undefined
      cpmerror(loc&": unvalid value "&w&" for w");
   }
   if (cpmverbose>mL){
      ostringstream ost;
      ost<<"spectrum.first()="<<spectrum.fir();
      ost<<"spectrum.last()="<<spectrum.last();
      cpmmessage(ost);
   }
   CPM_MZ
}

SpectralObject::SpectralObject(const V<R>& s, const V<State>& e):
         d_(s.dim()),spectrum(d_),eigenVectors(e)
{
   cpmassert(d_==e.dim(),"SpectralObject(...)");
   for (Z i=1;i<=d_;i++) spectrum[i]=C(s[i],0);
}

State SpectralObject::operator *(State const& a)const
{
   State res=eigenVectors[1]*(spectrum[1]*(eigenVectors[1]|a));
   for (Z i=2;i<=d_;i++){
      C ci=spectrum[i]*(eigenVectors[i]|a);
      res+=eigenVectors[i]*ci;
   }
   return res;
}

static void display(const State& s, Frame& f, Z iOffset, Z mult, Z d)
{
   Z i,j;
   Z d_=d*mult;
   Z i1=1+iOffset;
   Z i2=mult+iOffset;
   for (i=i1;i<=i2;i++) for (j=0;j<d_;j++)
      f.put(Z2(i,j),Color(C(s[j/mult])));
}


void SpectralObject::show(Frame& f, Z mult, Z is, Z js)const
{
   Z mL=2;
   Word loc("SpectralObject::show(Frame,Z,Z,Z)");
   CPM_MA
   Z i,j;
   Z d=dim(),pix=mult;
   Z shiftH=2,gap=2;
   Z dTot=d+shiftH+gap;
   if (mult<0){
      Z nRow=f.getLin()-is;
      Z nCol=f.getCol()-js;
      Z nOK=cpminf(nCol,nRow);
      pix=nOK/(dTot);
      if (pix<1) pix=1;
   }
   State sp(d);
   for (i=1;i<=d;i++){
      sp[i]=spectrum[i];
   }
   Z pixH=pix<3 ? 3 : pix;
   sp.show(f,pix,pixH,0,shiftH);
   Observable aux(d);
   for (i=1;i<=d;i++){
      for (j=1;j<=d;j++){
         C val=eigenVectors[i][j];
         aux.set(i,j,val);
      }
   }
   aux.show(f,pix,0,pixH+shiftH+gap);
   CPM_MZ
}

void SpectralObject::mark(Graph& gr, Z ipolMeth, R gamma)const
{
   Z d=dim();
   VVV<R> vvvr(3,d,d+3); // spectal values get three pixel wide B/W stripe
   R_Vector gs=getSpectrum(gamma);
   for (Z i=1;i<=d;++i){
      R gsi=gs[i]*255_R;
      vvvr[1][i][1]=gsi;
      vvvr[2][i][1]=gsi;
      vvvr[3][i][1]=gsi;
      vvvr[1][i][2]=gsi;
      vvvr[2][i][2]=gsi;
      vvvr[3][i][2]=gsi;
      vvvr[1][i][3]=gsi;
      vvvr[2][i][3]=gsi;
      vvvr[3][i][3]=gsi;
      for (Z j=1;j<=d;++j){
         State si=eigenVectors[i];
         C val=si[j];
         Color col(val);
         vvvr[1][i][j+3]=col.Red();
         vvvr[2][i][j+3]=col.Green();
         vvvr[3][i][j+3]=col.Blue();
            // notice that the channel index is the first one
      }
   }
   CpmImaging::MultiImage<R> mi(vvvr);
   mi.mark(gr,ipolMeth);
}

 void SpectralObject::showSpectrum(Graph& gr)const
 {
   Z mL=2;
   Word loc("SpectralObject::showSpectrum(Graph)");
   CPM_MA
   Z d=dim();
   R_Vector sp(d);
   for (Z i=1;i<=d;i++){
      sp[i]=spectrum[i][1];
   }
   gr.show(sp,RED,1);
   CPM_MZ
 }

 void SpectralObject::markSpectrum(Graph& gr)const
 {
   Z mL=2;
   Word loc("SpectralObject::showSpectrum(Graph)");
   CPM_MA
   Z d=dim();
   R_Vector sp(d);
   for (Z i=1;i<=d;i++){
      sp[i]=spectrum[i][1];
   }
   gr.mark(sp,RED,1);
   CPM_MZ
 }

R SpectralObject::test()const
{
   Z mL=2;
   Word loc("SpectralObject::test()");
   CPM_MA
   R_Matrix sp(d_);
   for (Z i=1;i<=d_;i++){
      for (Z j=1;j<=d_;j++){
         sp[i][j]=(eigenVectors[i]|eigenVectors[j]).abs();
      }
   }
   R_Matrix u=sp.net(1);
   R res=(sp-u).abs();
   cpmcerr<<loc<<" res="<<res<<endl;
   CPM_MZ
   return res;
}

bool SpectralObject::writeSpectrum(ostream& str)const
{
   CpmRoot::writeTitle("Eigenvalues of a SpectralObject:",str);
   cpmp(spectrum);
   return true;
}

bool SpectralObject::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("a SpectralObject:",str);
   cpmp(d_);
   cpmp(spectrum);
   cpmp(eigenVectors);
   return true;
}

bool SpectralObject::scanFrom(istream& str)
{
   cpms(d_);
   cpms(spectrum);
   cpms(eigenVectors);
   return true;
}

X2<R,State> SpectralObject::reduction(State const& a, R p)const
{
   V<C> w(d_);
   Z i;
   for (i=0;i<d_;i++){
      w[i+1]=(eigenVectors[i]|a);
   }
   C zRand;
   R r1=1.;
   R phi=CpmRoot::ran(r1); // not yet an arc, in inteval (-1,+1)
   phi*=cpmpi; // arc phi
   zRand.polar(1.,phi); // random selection of a complex number of
      // absolute value 1 ('random phase')
   Vo<R> sig(d_);
      // to be used for a test on probabilistic correctness
      // not yet implemented !!!!!!!!!!!!!!!!!!!!!
   V<Z> accum(d_,0);
   Z nTest=20000;
   R nTestInv=1./nTest;
   Z iRes;
   for (i=1;i<=nTest;i++){
      cpmprogress("Test-loop",R(i)*nTestInv);
      phi=CpmRoot::ran(r1,i); // not yet an arc, in inteval
      // (-1,+1)
      phi*=cpmpi; // arc phi
      zRand.polar(r1,phi);
      for (Z j=1;j<=d_;j++){
         R valSig=(w[j]*zRand)[1];
         valSig=cpmabs(valSig);
         sig[j]=valSig;
      }
      iRes=sig.indSup();
      accum[iRes]+=1;
   }
   R_Vector res(d_);
   for (i=1;i<=d_;i++){
      R valW=w[i].abs();
      valW*=valW;
      res[i]=accum[i];
   }
   CpmGraphics::Graph g;
   g.show(res);
   iRes--;
   return X2<R,State>(spectrum[iRes][1],eigenVectors[iRes]);
}

State SpectralObject::dePhase(const State& psi, R delta, R p,
   Z iter)const
// most advanced version that gives the orthodox probabilities
// for a stochatically varying phase delta
{
   const R inv2pi=R(0.5)/cpmpi;
   bool useDegrees=false;
   CpmGeo::Angle phi(delta,useDegrees);
   R xi=(cpmpi+phi.toR())*inv2pi;
      // quantity in range (0,1]
// quantities which describe the relation of state psi to observable
// *this
   V<C> c(d_);
   V<R> w(d_);
   V<R> a(d_);
   Z i,j;
   for (i=1;i<=d_;i++){
      C psi_i=(eigenVectors[i]|psi);
      c[i]=psi_i;
      w[i]=psi_i.absSqr();
      a[i]=spectrum[i].real();
   }
// random decision for an eigenvalue of *this to be used for
// demodulation
   R sum=w[0];
   Z i0=0;
   Z dLim=d_-1;
   while(sum<xi && i0<dLim){
      i0++;
      sum+=w[i0];
   }
      // a[i0] is the selected eigenvalue
// de-phasing and averaging; makes the wave packet collapse
   for (i=1;i<=d_;i++){
      C z=c[i];
      C ciSum=z;
      C ph=C::expi(p*(a[i]-a[i0]));
      for (j=1;j<=iter;j++){
         z*=ph;
         ciSum+=z;
      }
      c[i]=ciSum;
   }
   State res(d_);
   for (i=1;i<=d_;i++) res+=eigenVectors[i]*c[i];
   res.normalize_();
   return res;
}

State SpectralObject::dePhase(const State& psi,R delta,
   const V<SpectralObject>& cwp,const V<R>& p)
{
   const R inv2pi=R(0.5)/cpmpi;
   Z iter=8;
      // corresponds to n (or n-1 ?) in the paper
   bool useDegrees=false;
   CpmGeo::Angle phi(delta,useDegrees);
   R xi=(cpmpi+phi.toR())*inv2pi;
      // quantity in range (0,1]
// quantities which describe the relation of state psi to observable
// *this
   Z dc=cwp.dim();
   V<Z> selInd(dc);
   Z d=cwp[1].dim();
   State res=psi;
   Z h,i,j;
   for (h=1;h<=dc;h++){
      if (h>0){
         Z d_=cwp[h].dim();
         if (d!=d_)
            cpmerror("SpectralObject::dePhase(...): dimension mismatch");\

      }
      V<C> c(d);
      V<R> w(d);
      V<R> a(d);
      for (i=1;i<=d;i++){
         C psi_i=(cwp[h][i]|res);
         c[i]=psi_i;
         w[i]=psi_i.absSqr();
         a[i]=cwp[h](i).real();
      }
// random decision for an eigenvalue of *this to be used for
// demodulation
      R sum=w[0];
      Z i0=0;
      Z dLim=d-1;
      while(sum<xi && i0<dLim){
         i0++;
         sum+=w[i0];
      }
      selInd[h]=i0;

// a[i0] is the selected eigenvalue
// de-phasing and averaging makes the wave packet collapse
      for (i=1;i<=d;i++){
         C z=c[i];
         C ciSum=z;
         C ph=C::expi((a[i]-a[i0])*R(p[h]));
         for (j=1;j<=iter;j++){
            z*=ph;
            ciSum+=z;
         }
         c[i]=ciSum;
      }
      for (i=1;i<=d;i++) res+=cwp[h][i]*c[i];
   }
   res.normalize_();
   return res;
}

State SpectralObject::represent(const State& psi)const
       // represents psi as a 'eigenfunction expansion with
       // respect to *this'
{
   State res(d_);
   for (Z i=0;i<d_;i++) res[i]=(eigenVectors[i]|psi);
   res.normalize_();
   return res;
}

namespace{

C f_exp(C const& z, R const& t){ return exp(C(0,t)*z) ;}

}

void  SpectralObject::expi(R t)
{
#ifdef CPM_Fn
   F<C,C> f=F1<C,R,C>(t)(f_exp);
#else
   F<C,C> f(bind(f_exp,_1,t));
#endif
   this->map(f);
}

namespace{
   R makeReal(C const& z){return z.real();}
}

SpectralObject SpectralObject::order()const
         // returns a new SpectralObject which results form *this
         // by common re-ordering of eigenvalues and eigenvectors
         // such that the real parts of the eigenvalues are in increasing
         // order with the index
{
   F<C,R> transf(makeReal);
      // going for the real part
   V<R> spr=spectrum(transf);
      // making an array that holds the real parts
   V<R> spr2(spr);
   Vo<R> sp(spr2);
      // making an auxiliary object that
      // allows to create an ordering permutation
   V<Z> per=sp.permutationForIncreasingOrder();
      // making the ordering permutation
// rest is simple
   V<C> spectrumTemp(spectrum);
   V<State> eigenVectorsTemp(eigenVectors);
   spectrumTemp=spectrumTemp.permute(per);
   eigenVectorsTemp=eigenVectorsTemp.permute(per);
   V<C> spectrumRes=spectrumTemp;
   V<State> eigenVectorsRes=eigenVectorsTemp;
   return SpectralObject(spectrumRes,eigenVectorsRes);
}

//////////////////////// class Observable ////////////////////////////
/**********
Z Observable::adjust(Z i)const
{
   Z d=rep.dim();
   while(i>d) i-=d;
   while(i<1) i+=d;
   return i;
}
*************/

Observable::Observable(const SpectralObject& p):
   rep(p.dim(),C_Vector(p.dim()))
{
   Z i,j,k,d=p.dim();
   for (i=1;i<=d;i++){
      for (j=1;j<=d;j++){
         C sum;
         for (k=1;k<=d;k++){
            sum+=p(k)*p[k][i].con()*p[k][j];
         }
         set(i,j,sum);
      }
   }
}

Observable::Observable(const Word& w, Z d)
{
   Z mL=2;
   Word loc( "Observable(" );
   loc&=( w&", "&cpm(d)&")" );
   CPM_MA
   if (w=="zero"||w=="0"){
      *this=Observable(0.,d);
   }
   else if (w=="unit"||w=="1"){
      *this=Observable(1.,d);
   }
   else if (w=="momentumF" || w=="pF"){
      Observable aux(0.,d);
      R diff=(size*2)/R(d);
      C fac=-ihDirac*(R(1)/diff);
      for (Z i=1;i<=d;++i){
         aux.set(i-1,i,fac);
         aux.set(i+1,i,-fac);
      }
      *this=aux;
   }
   else if (w=="kinetic energy F" || w=="eKinF"){
      Observable aux(0.,d);
      R diff=size/d;
      R fac1=hDirac/diff;
      R fac=-fac1*fac1/(mass*2);
      C co1(fac);
      C co2=co1*(-2);
      for (Z i=1;i<=d;++i){
         aux.set(i-1,i,co1);
         aux.setLean(i,i,co2);
         aux.set(i+1,i,co1);
      }
      // notice that this defines a symmetric operator
      // as it stands
      *this=aux;
   }
   else if (w=="eKinFd"){
      Observable aux(0.,d);
      R diff=size/d;
      R fac1=hDirac/diff;
      R fac=-fac1*fac1/(mass*2);
      C co1(fac);
      C co2=co1*(-2);
      for (Z i=1;i<=d;++i){
         if (i-1>=0) aux.setLean(i-1,i,co1);
         aux.setLean(i,i,co2);
         if (i+1<d) aux.setLean(i+1,i,co1);
      }
      // notice that this defines a symmetric operator
      // as it stands
      *this=aux;
   }
   else{
      SpectralObject so(w,d);
      *this=Observable(so);
   }
   CPM_MZ
}

Observable::Observable(const Word& w, Z d, const R_Func& f)
{
   Z mL=2;
   Word loc("Observable(const Word& w, Z d, const R_Func& f)");
   CPM_MA
   cpmmessage(loc&" started");
   SpectralObject so(w,d);
   so.map_(f);
   cpmmessage(loc&" done");
   *this=Observable(so);
   CPM_MZ
}

Observable::Observable(const Word& w):rep(2,C_Vector(2))
// initialization thus as a 2x2 zero matrix
// See L.I.Schiff: quantum mechanics, 3rd edition, p. 206,
// eq. (27.30)
{
   const R fac=R(0.25)/cpmpi;
   const C re=C(fac,0);
   const C im=C(0,fac);

   if (w=="sx"){
      rep[1][2]=re;
      rep[2][1]=re;
   }
   else if (w=="sy"){
      rep[1][2]=-im;
      rep[2][1]=im;
   }
   else if (w=="sz"){
      rep[1][1]=re;
      rep[2][2]=-re;
   }
   if (w=="sigma_x"){
      rep[1][2]=C(1.);
      rep[2][1]=C(1.);
   }
   else if (w=="sigma_y"){
      rep[1][2]=-C::I;
      rep[2][1]=C::I;
   }
   else if (w=="sigma_z"){
      rep[1][1]=C(1.);
      rep[2][2]=C(-1.);
   }
   else{ // undefined
      cpmerror("Observable(Word w): unvalid value for w");
   }
}

Observable::Observable(const Angle& phi, const R3& n)
// Landau Lifschitz III, (58.1-2)
{
   Angle phiHalf(phi.toR()*R(0.5));
   R c=cpmcos(phiHalf);
   C s=cpmsin(phiHalf)*C::I;
   R3 nn=n;
   nn.normalize();
   *this=Observable(Word("unit"),2)*c+Observable("sigma_x")*s*nn[1]+
      Observable("sigma_y")*s*nn[2]+Observable("sigma_z")*s*nn[3];
}

Observable::Observable(Word const& w, State const& st):
rep(st.dim(), C_Vector(st.dim()))
{
   Z mL=3;
   Word loc("Observable(Word, State)");
   CPM_MA
   Z i,j,d=rep.dim();
   if (w=="potential"){
      for (i=1;i<=d;++i){
         rep.cui(i).cui(i)=st(i);
      }
   }
   else if (w=="projector"){
      State x=st;
      C xi,xj;
      x.normalize_();
      for (i=1;i<=d;i++){
         xi=x.rep_.cui(i);
         for (j=1;j<=d;j++){
            xj=x.rep_.cui(j);
            rep.cui(i).cui(j)=xi*xj.con();
         }
      }
   }
   else{
      cpmerror(loc&":"&w&" not allowed as first argument");
   }
   CPM_MZ
}

Observable Observable::ran(Z j)const
{
   Z j0=j;
   Z d=dim();
   C z(1,1);
   Observable res(d);
   for (Z i=1;i<=d;i++){
      for (Z j=1;j<=d;j++){
         j0++;
         res.rep.cui(i).cui(j)=z.ran(j0);
      }
   }
   return res.symmetricPart();
}


Observable Observable::pot(R_Func const& f, Z d, R lambda)
{
   V<R> x0=posNor(d);
   Observable res(d);
   for (Z i=1;i<=d;i++){
      res.rep.cui(i).cui(i)=C(f(x0[i])*lambda);
   }
   return res;
}

Observable Observable::pot2(R_Func const& f, Z d, R lambda)
{
   V<R> x0=posNor(d);
   Observable res(d);
   for (Z i=1;i<=d;i++){
      for (Z j=1;j<=d;j++){
         R dij=x0[i]-x0[j];
         if (dij>0.5) dij-=1;
         else if (dij<-0.5) dij+=1;
         else ;
         res.rep.cui(i).cui(j)=C(f(dij)*lambda);
      }
   }
   return res;
}

// operations which transform rows into columns

Observable Observable::transpose(void)const
// fast implementation without index bounds check
{
   Z d=dim();
   Observable res(d);
   for (Z i=1;i<=d;i++){
      for (Z j=1;j<=d;j++){
         res.rep.cui(i).cui(j)=
         rep.cui(j).cui(i);
      }
   }
   return res;
}

Observable Observable::con(void)const
// fast implementation without index bounds check
{
   Z d=dim();
   Observable res(d);
   for (Z i=1;i<=dim();i++){
      for (Z j=1;j<=dim();j++){
         res.rep.cui(i).cui(j)=
         rep.cui(j).cui(i).con();
      }
   }
   return res;
}


Observable Observable::prodStar(const Observable& y)const
{

   Z d=dim();
   cpmassert(d==y.dim(),"CpmQuantumMechanics::prodStar(...)");
   Observable res(d);
   for (Z i=1;i<=d;i++){
      for (Z j=1;j<=d;j++){
         res.rep.cui(i).cui(j)=
         rep.cui(i).c().fAcc2(
            QMBase::prodFuncConj2,QMBase::accFunc,y.rep.cui(j));
      }
   }
   return res;
}

// linear algebra of observables


Observable& Observable::operator +=(const Observable& x)
{
   Z d=dim();
   cpmassert(d==x.dim(),"");
   for (Z i=1;i<=d;i++){
      rep.cui(i).c().ff_(sumFunc,x.rep.cui(i));
   }
   return *this;
}

Observable Observable::operator -(void)const
{
   Z d=dim();
   Observable res(d);
   for (Z i=1;i<=d;i++){
      res.rep.cui(i).c()=rep.cui(i).c().fa(negFunc);
   }
   return res;
}


Observable& Observable::operator -=(const Observable& x)
{
   Z d=dim();
   cpmassert(d==x.dim(),"");
   for (Z i=1;i<=d;i++){
      rep.cui(i).c().ff_(diffFunc,x.rep.cui(i));
   }
   return *this;
}

Observable& Observable::operator *=(C z)
{
   Z d=dim();
   for (Z i=1;i<=d;i++){
      rep.cui(i).c().fc_(prodFunc,z);
   }
   return *this;
}

Observable& Observable::operator *=(R r)
{
   Z d=dim();
   for (Z i=1;i<=d;i++){
      rep.cui(i).c().fc_(prodFuncReal,C(r));
   }
   return *this;
}

Observable Observable::operator +(const Observable& y)const
{
   Z d=dim();
   cpmassert(d==y.dim(),"operator +");
   Observable res(*this);
   return res+=y;
}

Observable Observable::operator -(const Observable& y)const
{
   Z d=dim();
   cpmassert(d==y.dim(),"operator -");
   Observable res(*this);
   return res-=y;
}

Observable Observable::operator *(C z)const
{
   Observable res(*this);
   return res*=z;
}

Observable Observable::operator *(R r)const
{
   Observable res(*this);
   return res*=r;
}

// Hilbert-Schmidt structure

C Observable::operator |(const Observable& x2)const
{
   C res;
   Z d=dim();
   cpmassert(d==x2.dim(),"");
   for (Z i=1;i<=d;i++){
      res+=rep.cui(i).c().fAcc2(
         QMBase::prodFuncConj1,QMBase::accFunc,x2.rep.cui(i));
   }
   return res;
}

State Observable::operator *(const State& a)const
{
   Z mL=4;
   Word loc("State Observable::operator *(const State& a)const");
   cpmmessage(mL,loc&" started");
   Z d=dim();
   cpmassert(d==a.dim(),loc);
   State res(d);
   for (Z i=1;i<=d;i++){
      res.rep_.cui(i)=
         rep.cui(i).c().fAcc2(prodFunc,accFunc,a.rep_);
   }
   cpmmessage(mL,loc&" done");
   return res;
}

/*****************************
State Observable::operator *(State const& a)const
{
   Z mL=4;
   Word loc("State Observable::operator *(const State& a)const");
   cpmmessage(mL,loc&" started");
   Z d=dim(),i,j;
   cpmassert(d==a.dim(),loc);
   State res(d);
   for (i=1;i<=d;i++){
      C ci;
      for (j=1;j<=d;j++) ci+= rep[i][j]*a[j];
      res[i]=ci;
   }
   cpmmessage(mL,loc&" done");
   return res;
}
***************************************/

namespace{
State complexFormOf(R_Vector const& x)
{
   Z i,d2=x.dim();
   Z d=d2/2;
   cpmassert(d*2==d2,"");
   State res(d);
   for (i=1;i<=d;i++) res[i]=C(x[i],x[i+d]);
   return res;
}

State stateFormOf(C_Vector const& x)
{
   Z i,d=x.dim();
   State res(d);
   for (i=1;i<=d;i++) res[i]=x[i];
   return res;
}
}

static bool C_dependent(const State& x, const State& y, R eps)
{
   State y1=y*C(0,1);
   if ((x-y1).norm()<eps ) return true;
   y1=y*C(0,-1);
   if ((x-y1).norm()<eps ) return true;
   return false;
}

R Observable::autoStep(State const& x1)const
{
   R n1=x1.norm();
   State x2=(*this)*x1;
   R n2=x2.norm()*hDiracInv;
   return step(n1,n2);
}

SpectralObject Observable::spectralRepresentation(bool enfSym,
   bool test, Z iterMax, R eps, R eps1, R eps2)const
// now the order of eigenvalues is in increasing order. I.e. the
// groundstate is number one!
{
   Z mL=1;
   Word loc("Observable::spectralRepresentation(bool,bool,Z,R,R,R)");
   CPM_MA
   if (QMBase::getVerbose()>1) cpmmessage(loc&": started");
   C_Matrix cRep(rep);
   SpcObj spo(cRep,enfSym); // implementation depends on CPM_USE_EIGEN
   R_Vector spec=spo.val();
   Z d=spo.dim();
   V<C> complexSpec(d);
   V<State> complexEig(d);
   for (Z i=1;i<=d;i++){
      complexSpec[i]=C(spo(i));
      complexEig[i]=stateFormOf(spo[i]); // cEig[i] is a C_Vector
   }
 //  CPM_MZ
 //  return SpectralObject(complexSpec, complexEig);
   SpectralObject res(complexSpec, complexEig);
   if (test){
      R errSpo=res.test();
      cout<<"errSpo = "<<errSpo<<endl;
   }
   CPM_MZ
   return res;
}

R_Matrix Observable::toRealForm()const
      // returns a real (2*d,2*d) matrix
{
   Z i,j,i2,j2,d=dim();
   Z d2=d*2;
   R_Matrix res(d2,d2);
   C val;
   R valr,vali;
   for (i=1;i<=d;i++){
      i2=i+d;
      for (j=1;j<=d;j++){
         j2=j+d;
         val=rep.cui(i).cui(j);
         valr=val[1];
         vali=val[2];
         res[i][j]=valr;
         res[i2][j2]=valr;
         res[i2][j]=vali;
         res[i][j2]=-vali;
      }
   }
   return res;
}

void Observable::fromRealForm(const R_Matrix& r)
      // changes *this into the one resuling from the real matrix r
{
   Z i,j,i2,d2=r.dim1();
   cpmassert(r.dim2()==d2,"");
   Z d=d2/2;
   cpmassert(2*d==d2,"");
   rep=C_Matrix(d, C_Vector(d));
   C val;
   R valr,vali;
   for (i=1;i<=d;i++){
      i2=i+d;
      for (j=1;j<=d;j++){
         valr=r[i][j];
         vali=r[i2][j];
         val=C(valr,vali);
         rep.cui(i).cui(j)=val;
      }
   }
}

void Observable::show(Frame& f, Z mult, Z is, Z js, bool markOnly)const
{
   Z mL=1;
   Word loc("Observable::show(Frame,Z,Z,Z,bool)");
   CPM_MA
   Z d=dim(),pix=mult;
   if (mult<0){
      Z nRow=f.getLin()-is;
      Z nCol=f.getCol()-js;
      Z nOK=cpminf(nCol,nRow);
      pix=nOK/d;
      if (pix<1) pix=1;
   }

   R valMax=sup();
   if (valMax==0) valMax=1.; // zero observable should be representable
   R valMaxInv=R(1)/valMax;
   Z i,j;
   Z d_=d*pix;
   for (i=1;i<=d_;i++){
      Z iEff=i/pix;
      for (j=1;j<=d_;j++){
         C val=rep.cyc(iEff).cyc(j/pix);
         val*=valMaxInv;
         f.put(Z2(i+1+is,j+1+js),Color(val));
            // put replaced by putSave 2004-09-21
      }
   }
   if (!markOnly) f.vis(); // added 2005-04-15
      // argument markOnly added 2014-03-30
   CPM_MZ
}

void Observable::showImg(Graph& f, Z ipolMeth)const
{
   Z d=dim();
   VVV<R> vvvr(3,d,d);
   for (Z i=1;i<=d;++i){
      for (Z j=1;j<=d;++j){
         C val=rep.cui(i).cui(j);
         Color col(val);
         vvvr[1][i][j]=col.Red();
         vvvr[2][i][j]=col.Green();
         vvvr[3][i][j]=col.Blue();
            // notice that the channel index is the first one
      }
   }
   CpmImaging::MultiImage<R> mi(vvvr);
   mi.mark(f,ipolMeth); // nice service of class MultiImage
   // iPolMeth==2 not yet working
}

void Observable::mark(Graph& gr, Z ipolMeth)const
{
   Z d=dim();
   VVV<R> vvvr(3,d,d);
   for (Z i=1;i<=d;++i){
      for (Z j=1;j<=d;++j){
         C val=rep.cui(i).cui(j);
         Color col(val);
         vvvr[1][i][j]=col.Red();
         vvvr[2][i][j]=col.Green();
         vvvr[3][i][j]=col.Blue();
            // notice that the channel index is the first one
      }
   }
   CpmImaging::MultiImage<R> mi(vvvr);
   mi.mark(gr,ipolMeth); // nice service of class MultiImage
   // iPolMeth==2 not yet working
}

R Observable::sup()const
{
   R res=-1;
   Z d=dim();
   for (Z i=1;i<=d;i++) for (Z j=1;j<=d;j++){
      R r=rep.cui(i).cui(j).abs();
      if (r>res) res=r;
   }
   return res;
}

Observable Observable::cayley()const
{
   Z d=dim();
   Observable res(Word("unit"),d); // unit matrix
   res=(res+(*this))*!(res-(*this))*C(0,1);
      // although I expected this to be selfadjoint anyway this is not the
      // case
   return res.symmetricPart();
}

// trying to improve unitarity: does not work so far
static Observable iter(const Observable& x)
{ return (x.con()+!x)*C(0.5);}

Observable Observable::unitaryVersion()const
      // if *this is nearly unitary, the return value should be so
      // to much better approximation. Does not work that way, however
{
   Observable res=iter(*this);
   res=iter(res);
   return (*this+res)*C(0.5);
}

Observable Observable::cut()const
{
   Observable res(*this);
   Z iMax=res.dim()-1;
   Z iMin=0;
   C z;
   res.set(iMin,iMax,z);
   res.set(iMax,iMin,z);
   return res;
}

bool Observable::prnOn(ostream& out)const
{
   Z d=dim();
   out<<endl<<"// an Observable:";
   CpmRoot::write(d,out);
   for (Z i=1;i<=d;i++)
      for (Z j=1;j<=d;j++)
         getLean(i,j).prnOn(out);
   return true;
}

bool Observable::scanFrom(istream& in)
{
   Z d=dim();
   CpmRoot::read(d,in);
   rep=C_Matrix(d,C_Vector(d));
   C val;
   for (Z i=1;i<=d;i++){
      for (Z j=1;j<=d;j++){
         val.scanFrom(in);
         setLean(i,j,val);
      }
   }
   return true;
}

Observable Observable::operator ||(const Observable& s2)const
{
   Z d1=dim();
   Z d2=s2.dim();
   Observable res(d1*d2);
   Z i1,i2,j1,j2,k,l;
   k=0;
   for (i1=1;i1<=d1;i1++){
      for (i2=1;i2<=d2;i2++){
         l=1;
         for (j1=1;j1<=d1;j1++){
            for (j2=1;j2<=d2;j2++){
               C val=getLean(i1,j1)*s2.getLean(i2,j2);
               res.setLean(k,l,val);
               l++;
            }
         }
         k++;
      }
   }
   return res;
}

Observable Observable::prdFock( )const
{
   Z d=dim();
   Observable unit(Word("unit"),d);
   Observable ob1=(*this)|| unit;
   Observable ob2=unit||(*this);
   return ob1+ob2;
}

Observable Observable::prdFock2( )const
{
   return (*this)||(*this);
}

R Observable::normIter(R eps)const
{
   static Word loc("Observable::normIter(R)");
   Z mL=3;
   CPM_MA
   static const Z iterMax=100;
   Z n=dim();
   State st0(n);
   State st=st0.ran();
   R res=-1e20;
   bool work=true;
   Z iter=0;
   while (work){
      iter++;
      st=(*this)*st;
      R resNew=st.normalize_();
      R diff=dis(res,resNew);
      if (diff<eps){
         work=false;
         res=(res+resNew)*0.5;
      }
      else if (iter>iterMax){
         work=false;
         cpmerror(loc&": stopped after "&cpm(iter)&" iterations");
         res=resNew;
      }
      else{
         res=resNew;
      }
   }
   CPM_MZ
   return res;
}

State Observable::expi(R t, State const& s, R eps)
      // A.expi(t,s) is exp(i*t*A)*s according to the Chebyshev
      // single step method advocated by Hans De Raedt
{

   R nor1=normIter();
   R security=1.2;
   nor1*=security;
   R nor1Inv=cpminv(nor1);
   // assumes that *this is symmetric
   Observable skew=(*this)*(C::I*nor1Inv);
   // now skew is skew-symmetric and has norm<1
   Z kMax=cpmtoz(cpmexp(1_R)*nor1*0.5_R)+1_Z;
   R z=t*nor1;
   State sum=s*besselJ(0_Z,z);
   State sp=s; // previous
   State sa=skew*s; // actual
   for (Z k=1;k<=kMax;++k){
      sum+=sa*(besselJ(k,z)*2_R);
      State temp=sa;
      sa=(skew*temp)*2_R+sp;
      sp=temp;
   }
   return sum;
}

////////////////////// class SpectralMeasure //////////////////////

SpectralMeasure::SpectralMeasure(const SpectralObject& spo):d(spo.dim())
{
   static const R epsRel=1e-6;
      // discrimination accuracy for spectral values

   if (QMBase::getVerbose()>1)
      cpmmessage("SpectralMeasure(const SpectralObject&) started");

   cpmassert(d==spo.dim(),"");
   V<R> v(d);
   V<Observable> vobs(d);
   Word w("projector");
   Z i;
   for (i=1;i<=d;i++) v[i]=spo(i)[1]; // spectrum is real anyway
   for (i=1;i<=d;i++) vobs[i]=Observable(w,spo[i]); // spectrum is real
   // anyway
   Observable sumO(d);
   for (i=1;i<=d;i++) sumO+=vobs[i];
   Observable unit(Word("unit"),d);
   cpmcerr<<endl<<"SpectralMeasure(const Observable&) test: err= "
      <<(sumO-unit).norm();
   R eps=(v[1]-v[d])*epsRel;
   V<R> sp(d);
   Observable zero(d);
   V<Observable> pr(d,zero);
   sp[1]=v[1];
   pr[1]=Observable(w,spo[1]);
   Z iAct=1;
   cpmassert(d>1,"");
   for (i=2;i<=d;i++){
      if (cpmabs(v[i]-v[iAct])<eps){
         pr[iAct]+=Observable(w,spo[i]);
      }
      else{
         iAct++;
         sp[iAct]=v[i];
         pr[iAct]=Observable(w,spo[i]);
      }
   }
   ds=iAct+1;
   spectrum=V<R>(ds);
   projectors=V<Observable>(ds);
   for (i=1;i<=ds;i++){
      spectrum[i]=sp[i];
      projectors[i]=pr[i];
   }
   if (QMBase::getVerbose()>1)
      cpmmessage("SpectralMeasure(const SpectralObject&) done");
}

SpectralMeasure::SpectralMeasure(const Observable& obs):d(obs.dim())
{
   static const R epsRel=1e-6;
      // discrimination accuracy for spectral values

   if (QMBase::getVerbose()>1)
      cpmmessage("SpectralMeasure(const Observable&) started");

   SpectralObject spo=obs.spectralRepresentation(false,false);
   cpmassert(d==spo.dim(),"");
   V<R> v(d);
   V<Observable> vobs(d);
   Z i;
   Word w("projector");
   for (i=1;i<=d;i++) v[i]=spo(i)[1]; // spectrum is real anyway
   for (i=1;i<=d;i++) vobs[i]=Observable(w,spo[i]); // spectrum is real
   // anyway
   Observable sumO(d);
   for (i=1;i<=d;i++) sumO+=vobs[i];
   Observable unit(Word("unit"),d);
   cpmcerr<<endl<<"SpectralMeasure(const Observable&) test: err= "
      <<(sumO-unit).norm();
   R eps=(v[1]-v[d])*epsRel;
   V<R> sp(d);
   Observable zero(d);
   V<Observable> pr(d,zero);
   sp[1]=v[1];
   pr[1]=Observable(w,spo[1]);
   Z iAct=1;
   cpmassert(d>1,"");
   for (i=2;i<=d;i++){
      if (cpmabs(v[i]-v[iAct])<eps){
         pr[iAct]+=Observable(w,spo[i]);
      }
      else{
         iAct++;
         sp[iAct]=v[i];
         pr[iAct]=Observable(w,spo[i]);
      }
   }
   ds=iAct+1;
   spectrum=V<R>(ds);
   projectors=V<Observable>(ds);
   for (i=1;i<=ds;i++){
      spectrum[i]=sp[i];
      projectors[i]=pr[i];
   }
   if (QMBase::getVerbose()>1)
      cpmmessage("SpectralMeasure(Observable&) done");
}

X2<R,State> SpectralMeasure::reduction(const State& a)const
   // performs the 'collapse of the wave packet' of the state
   // given by the argument and the observable of spectral
   // representation *this. Returned is the 'measured value' and the
   // state after measurement
{

   if (QMBase::getVerbose()>2)
      cpmmessage("SpectralMeasure::reduction(...) started");

   State psi=a;
   psi.normalize_();
   V<R> w(ds);
   Z i;
   for (i=1;i<=ds;i++){
      C cVal=(a|(projectors[i]*a)); // is real and positive anyway
      R rVal=cVal.abs(); // makes this sure even in the presence
         // of numerical errors
      w[i]=rVal;
   }
   R accu=0;
   Vo<R> ws(ds+1); // class provides function locate
   for (i=1;i<=ds;i++){
      ws[i]=accu;
      accu+=w[i-1];
      if (accu>1.) accu=1; // should not ocurr
   }
   ws[ds+1]=1.;
      // now the interval [0,1] is subdivided into ds intervals of length
      // w[i]
   R wr=CpmRoot::ran(R(0.5))+R(0.5); // random number ranges from 0
   // to 1
   Z iRes=ws.locate(wr); // random selection of a interval
      // with probability equal to its lenght
   iRes--;
   R value=spectrum[iRes];
   State st=projectors[iRes]*psi;
   st.normalize_();
   if (QMBase::getVerbose()>2)
      cpmmessage("SpectralMeasure::reduction(...) done");
   return X2<R,State>(value,st);
}

State SpectralMeasure::operator *(const State& a)const
{
   State res=(projectors[0]*a)*spectrum[0];
   for (Z i=1;i<ds;i++){
      res+=(projectors[i]*a)*spectrum[i];
   }
   return res;
}

bool SpectralMeasure::prnOn(ostream& out)const
{
   cpmmessage("SpectralMeasure::prnOn(...) started");
   using CpmRoot::write;
   out<<endl<<"// a SpectralMeasure:";
   write(d,out);
   write(ds,out);
   V<R> aux1(spectrum);
   aux1.prnOn(out);
   for (Z j=0;j<ds;j++) projectors[j].prnOn(out);
   cpmmessage("SpectralMeasure::prnOn(...) done");
   return true;
}

bool SpectralMeasure::scanFrom(istream& in)
{
   cpmmessage("SpectralMeasure::scanFrom(...) started");
   using CpmRoot::read;
   read(d,in);
   read(ds,in);
   V<R> aux1;
   aux1.scanFrom(in);
   spectrum=aux1;
   projectors=V<Observable>(ds);
   for (Z j=1;j<=ds;j++) projectors[j].scanFrom(in);
      // it is much faster to scan the components directly into
      // projectors, as done here, instead of scanning first into aux2
      // and then copying into projectors. Brings several hundered
      // percent in performance !!!!!
   cpmmessage("SpectralMeasure::scanFrom(...) done");
   return true;
}

R SpectralMeasure::test()const
{
   Observable sum(d); // default constructor makes zero observable
   for (Z i=1;i<=ds;i++) sum+=projectors[i];
   Observable unit(Word("unit"),d);
   R err1=(sum-unit).norm();
   return err1;
}

//////////////////////// class State ////////////////////////////////////

C fExp(R const& t, R const& a){ return C::expi(a*t);}

State State::shift(R sf)const
// sign changed after test: 2004-10-13
{
   Z d=dim();
//   cpmassert(dL_==1,"State::shift(R)");
   bool b=CpmRootX::isPowOfTwo(d);
   if (b){
      SpectralObject so("p",d);
#ifdef CPM_Fn
      F<R,C> fsh=F1<R,R,C>(sf*hDiracInv)(fExp);
#else
      F<R,C> fsh(bind(fExp,_1,sf*hDiracInv));
#endif
      so.map(fsh);
      return so*(*this);
   }
   else{
      Z sd=cpmrnd((sf/size)*d);
      State res(d);
      for (Z i=1;i<=d;++i) res[i]=(*this)[i-sd];
         // safe indexing does periodic continuation
      return res;
   }
}
