//? cpmobservable.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_OBSERVABLE_H_
#define CPM_OBSERVABLE_H_
/*
   Description: objects for describing quantum mechanical observables
      as Hilbert space operators
*/
#include <cpmv.h>
#include <cpmf.h>
#include <cpmrmatrix.h>
#include <cpmcmatrix.h>
#include <cpmangle.h>
#include <cpmp.h>
#include <cpmalgorithms.h>
#include <cpmrecord.h>
#include <cpmclinalg.h>
#include <cpmstate.h>

namespace CpmQuantumMechanics{

   using namespace CpmArrays;
   using namespace CpmFunctions;
   using namespace CpmGraphics;

   using CpmRootX::Record;
   using CpmGeo::Angle;
   using CpmLinAlg::R3;
   using CpmLinAlg::CC4;
   using CpmLinAlg::CC3;
   using CpmLinAlg::CC2;
   using CpmAlgorithms::CyclicAlarm;

///////////////////// class SpectralObject ///////////////////////////

// combines a set of (complex) spectral values and a set of orthonormal
// set of vectors into a single object which according to the spectral
// theorem results from spectral analysis of a normal operator.
// Here we know only how to generate such an object from a self-adjoint
// operator.
// The most important constructive method is function
// SpectralObject
//    Observable::spectralRepresentation(bool,bool,Z,R,R,R)const;
// for creation from an Observable (which is basicly a complex matrix).
// There are, however, elementary observables for which it is
// simpler to define them as SpectralObjects (see constructor
// SpectralObject(const Word& w, Z d_);

   class Observable;

class SpectralObject:public QMBase{//eigenvectors/values in combination
      Z d_;
      V<C> spectrum;
      V<State> eigenVectors;
         // complete set of orthonormal vectors
   public:
      typedef SpectralObject Type;
      CPM_IO
      CPM_NAM(SpectralObject)
      friend class Observable;
      Z dim()const{ return d_;}
      const C& operator()(Z i)const{ return spectrum[i];}
         // eigenvalues by (), notice that i has to be in
         // {1,...,d_}.
      const State& operator[](Z i)const{ return eigenVectors[i];}
         // eigenvectors by [], notice that i has to be in
         // {1,...,d_}, the order is such that the ground-state of
         // the harmonic oscillator is obtained for i=d_
         // is this OK ?

      SpectralObject():d_(0){}

      explicit SpectralObject(Z d);
         // has the 'canonical vectors' as eigenvectors
         // and the canonical count index 0,1,.... d-1 as eigenvectors

      SpectralObject(const Word& w, Z d);
         // for w="zero" or "0":  zero operator
         // for w="unit" or "1":  unit operator
         // for w="space frequency" or "f": space frequency operator
         // for w="wave vector" or "k": wave vector operator
         // for w="momentum" or "p": momentum operator
         // for w="kinetic energy" or "eKin": operator of kinetic energy
         // for w="relativistic kinetic energy" or "eKinRel":
         //    operator of relativistic kinetic energy T=p^2/(2*mass)
         //    for which we give here the nice implicit equation
         //    T=p^2/(2*mass + T/c^2) which is an easy consequence
         //    of (E/c)^2 - p^2 = (mass*c)^2 and T=E-mass*c^2.
         // for w="position" or "x": position operator
         // for w="boost" or "n": non-relativistic boost operator

         // if w="f" or "k" or "p" or "eKin" or one of the full word
         // equivalents, d_ has to be a power of two. This is since the
         // construction then is based on passage to momentum space by
         // discrete (fast) Fourier transform.
   protected:
      SpectralObject(const V<C>& s, const V<State>& e):
         d_(s.dim()),spectrum(s),eigenVectors(e)
         {cpmassert(d_==e.dim(),"SpectralObject(...)");}
         // unsafe construction which only tests
         // fit of the dimensions but not that e is orthonormal

      SpectralObject(const V<R>& s, const V<State>& e);
         // unsafe construction which only tests
         // fit of the dimensions but not that e is orthonormal

   public:
      R_Vector getSpectrum()const
      {
         R_Vector res(d_);
         for (Z i=1;i<=d_;i++) res[i]=spectrum[i][1];
         return res;
      }

      R_Vector getSpectrum(R gamma)const
         // Convenient method for showing thr spectrum of density operators
         // Here the spectral values are positive and add up to 1.
         // For critical inspection of small spectral values use
         // gamma << 1 and for critical expection of the largest values
         // values (by suppressing the small ones) use gamma > 1.
      {
         R_Vector res(d_);
         for (Z i=1;i<=d_;i++) res[i]=spectrum[i].abs();
         R supInv=cpminv(res.sup());
         for (Z i=1;i<=d_;i++) res[i]=cpmpow(res[i]*supInv,gamma);
         return res;
      }

      R supEigenVectors()const
      {
         R_Matrix rs(d_,d_);
         for (Z i=1;i<=d_;i++)
            for (Z j=1;j<=d_;j++)
               rs[i][j]=eigenVectors[i][j].abs();
         return rs.sup();
      }

      R spectralRadius()const
      {
         R_Vector rs(d_);
         for (Z i=1;i<=d_;i++) rs[i]=spectrum[i].abs();
         return rs.sup();
      }

      void setSpectrum_(C z){ for (Z i=1;i<=d_;i++) spectrum[i]=z;}

      void show(Frame& f, Z mult=-1, Z iShift=0, Z jShift=0)const;

      void mark(Graph& gr, Z ipolMethod=0, R gamma=0.5_R)const;
         // modern implementation, writes on image memory.
         // Visible only upon gr.vis().

      void showSpectrum(Graph& gr)const;
         // real part of eigenvalues, as autoscaled graph

      void markSpectrum(Graph& gr)const;
         // real part of eigenvalues, as autoscaled graph
         // visible only upon gr.vis()

      bool writeSpectrum(ostream& str)const;
         // writes the eigenvalues to str, returns true
         // if writing was successful

      void map(const F<C,C>& f)
         // f acts on the spectrum
      {for (Z i=1;i<=d_;i++) spectrum[i]=f(spectrum[i]);}

      void map(const F<R,C>& f)
         // f acts on the spectrum
      {for (Z i=1;i<=d_;i++) spectrum[i]=f(spectrum[i][1]);}

      void expi(R t);
         // changes *this into exp(i*t*(*this))

      void map_(const R_Func& f)
         // real function f acts on the real part of the spectrum
      {
         for (Z i=1;i<=d_;i++)
            spectrum[i]=C(f(spectrum[i][1]),spectrum[i][2]);
      }

      R test()const;
         // returns a error sum which is large if the spectral data
         // are not correct

      void normalize_()
         { for (Z i=1;i<=d_;i++) eigenVectors[i].normalize_();}
      // normalizes all eigenvectors

      State operator *(State const&)const;
      // SpectralObject acts on state just as the underlying observable

      X2<R,State> reduction(State const& psi, R p=1.)const;
      // performs the 'collapse of the wave packet' of the state
      // given by the argument and the observable of spectral
      // representation *this. Returned is the 'measured value' and the
      // state after measurement. The parameter gives the probability
      // for actually doing the reduction upon function call.

      State dePhase(const State& psi, R delta=0., R p=0.25,
         Z iter=8)const;
      // dephasing of a state due to interaction with spectral object
      // *this. delta is the 'decision variable'
      // See U. Mutze: Modeling the `Collapse of the Wave Function' as a
      // Deterministic Process. Developing further this collapse model
      // could benefit from using my new approach to computational
      // quantum mechanics as mentioned in U.Mutze: Predicting Classical
      // Motion Directly from the Action Principle III and coded in
      // cpmpsi.h, cpmpsi.cpp (in d:\cpm\cpm1). Since my present form of
      // the dephasing mechanism yet relies on spectral decomposition and
      // not entierly on applying operators of the form exp(itA) on
      // states, the class Psi has first to be equiped with spectral
      // theory or the latter has to be replaced by a more operational
      // alternative.

      State represent(const State& psi)const;
       // represents psi as a 'eigenfunction expansion with
       // respect to *this'

      static State dePhase(const State&,R,const V<SpectralObject>&,
         const V<R>&);
      // dephasing of a state due to interaction with a family of
      // spectral objects. The result has a cleaner meaning if the
      // spectral objects in the family commute with each other.

      SpectralObject order()const;
         // returns a new SpectralObject which results form *this
         // by common re-ordering of eigenvalues and eigenvectors
         // such that the real parts of the eigenvalues are in increasing
         // order with the index. Thus for the harmonic oscillator
         // the ground state corresponds to index 0.
         // Quite elegant implementation!
   };


////////////////////// class Observable /////////////////////////

class Observable: public QMBase{ // observable as matrix
   // not assumed selfadjoint in order
   // to include unitary operators, and perhaps others

   C_Matrix rep;

   Observable transpose(void)const;
      // is not a canonical operation in Hilbert space, only con() is
  // Z adjust(Z i)const;
      // returns a 'version of i' in the range 1<=i<=dim();
public:
   typedef Observable Type;
   CPM_IO
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
      // Scalar Product and norm of Hilbert-Schmidt operators
   CPM_NAM(Observable)
   Observable (C_Matrix const& cm):rep(cm){}
   Observable(V< V<C> > const& rep_):rep(rep_){}
      // was private till 2006-07-12
   explicit Observable(CC4 const& cc):rep(cc.toVVC()){}
      // complex 4 by 4 matrices occur in Dirac theory
   explicit Observable(CC3 const& cc):rep(cc.toVVC()){}
      // complex 2 by 2 matrices occur in relativistic quantum theory
   explicit Observable(CC2 const& cc):rep(cc.toVVC()){}
      // complex 2 by 2 matrices occur in relativistic quantum theory
   Observable operator *(Observable const& m)const
      { return Observable(rep*m.rep);}
      // Now the main work is done in C_Matrix
   Observable prodStar(Observable const&)const;
      // returns the value of x*y.con()
      // This is the fastest product operation, since it has not to
      // perform a transposition or conjugation in order to write the new
      // matrix elements as binary expressions of component vectors.
      // Had we available column vectors as representors of matrixes,
      // then the combination x.con()*y would be the one most
      // economically expressible in terms of binary expressions of
      // component vectors Comparison of prodStar and operator * gave 30%
      // slower execution for prodStar after optimized compilation! Not
      // clear why that way, where an opposit result was expected. We
      // learn at least that it is not worth the effort to replace
      // products of hermitean operators by prodStar systematically.

   Observable(){}

   explicit Observable(Z d):rep(d, C_Vector(d)){}
      // Thus d is the dimension of the underlying State
      // created is the zero operator

   Observable( R r, Z d):rep(d, C_Vector(d))
      { for (Z i=1;i<=d;++i) rep[i][i]=C(r);}
      // creating a real multiple of the unit operator.

   Observable( C z, Z d):rep(d, C_Vector(d))
      { for (Z i=1;i<=d;++i) rep[i][i]=z;}
      // creating a complex multiple of the unit operator.

   explicit Observable(const SpectralObject&);
      // forming the observable out of its spectral representation

   Observable(Word const& w, State const& s);
      // for w="potential" we create an observable obs such that
      // obs(i,j)=s(i)*krn(i,j)
      // s also provided the dimension
      // for w="projector" we create the one-dimensional projection
      // operator corresponding to the argument state

   Observable(const Word& w, Z d);
      // An appended F in names refers to 'finite difference'
      // (not 'Fourier'!)
      // for w="zero" or w="0",  zero operator = Observable(0.,d)
      // for w="unit" or w="1", unit operator = Observable(1.,d)
      // for w="position" or w="x", position operator
      // for w="space frequency" or w="f", space frequency operator F
      // for w="wave vector" or w="k",'wave vector' operator, K=2*pi*F
      //    equals the
      //    momentum operator in any system of units in which h_quer=1.
      // for w="momentum" or w="p", momentum operator P=hQuer*K=hPlanck*F
      // for w="momentumF" or w="pF", momentum operator as a finite
      // difference which approximates the derivative in position space.
      // pF and eKinF are the only arguments which are not based on
      // spectral representation and thus are not allowed in
      // as arguments in SpectralObject(Word,Z).
      // It turns out that major shifts generated by this kind of
      // momentum operator are accompanied by major distortions for all
      // but the smoothest wave functions.

      // for w="kinetic energy" or w="eKin", kinetic energy via
      // Fourier-defined P*P/(2*mass), where P is the Fourier-defined
      // momentum operator

      // for w="kinetic energy F" or w="eKinF", kinetic energy via
      // discrete Laplacian = Pr*Pl/(2*mass), where Pr is the right-
      // difference operator and Pl the left-difference operator.

      // for w="eKinFd", kinetic energy via just as "eKinF" but
      // not with periodic boundary conditions as there but with
      // wave function vanishing outside the box. This corresponds to
      // the default continuation of arrays, thus 'd'.

      // for w="boost" or w="n" generator of boosts N=mass*X
      // (non-relativisic, see Habilschrift (2.11); see (2.25) for the
      // relativistic generalization)

      // For s:=QMBase::size we consider the interval [-s/2,s/2] and
      // divide it into d equal subintervals (of length  delta=s/d). The
      // coordinates of the centers of these subintervals are the d
      // eigenvalues of our position operator. These thus range from
      // -s/2+delta/2 to s/2-delta/2 with a spacing equal to delta. This
      // implies that our position operator X is the natural
      // discretization of  t h e  position operator for the interval
      // [-s/2, s/2] in physical position space.
      //
      // The space frequency operator F, as defined here for w=="f",
      // has a equidistant lattice of d eigenvalues with spacing f0=1/s.
      // The maximum eigenvalue is fc=d/(2*s)=(d/2)*f0,a smallest
      // non-zero,
      //
      // non-negative eigenvalue f0, followed by 0, -f0,..., -fc+f0
      // Then we have exp(i*a*f0)=1 for a=2Pi*s and thus exp(i*a*F)=1.
      // Since s is the periodicity of the base interval of our
      // 'periodic position space' this corresponds to a shift by s or
      // -s.
      // Thus for the spatial translation operator T:
      // T(-s)=exp(i*(2Pi*s)*F) and hence in general T(b)=exp(-i*2Pi*b*F).
      //
      // The normal quantum mechanical relation between spatial shift and
      // Momentum operator is T(b)=exp(-ib*P/h_quer) (e.g. J.J.Sakurai:
      // Modern Quantum Mechanics, p. 67). Therefore, P/h_quer=2Pi*F
      // P/h = F.
      // We therefore define the momentum operator P as h*F and the wave
      // number operator K as 2Pi*F. Thus K = 2Pi*F = (2Pi/h)*P , hence
      // P = hQuer*K as it has to be. Therefore, in any system of units
      // hQuer=1, we have K=P. The space frequency operator is considered
      // here only since it is closely related to the implementation
      // of these operators in terms of the discrete (fast)
      // Fourier transform. Thus for the definition of either of F,P,and
      // K. d has to be a power of two.
      // We expect [X,K] = i, the Heisenberg commutation
      // relation, to hold 'as good as finite dimensional matrices
      // can do it'. Actually they do very well. It is very instructive
      // to inspect A = [K,X]*i which would be the identity operator
      // if the Heisenberg commutation relation would hold exactly.
      // Actually most  of the eigenvalues of are numerically
      // indistiquishable from 1. But there are rather symmetrical
      // strong deviations at both ends of the spectrum.
      // When looking to the color coded matrix
      // elements (e.g. by the program CpmVQM2::App2 for builtX==4)
      // one would hardly see the close relationship of this matrix to
      // unity matrix. Notice also that this matrix puts a hard case
      // for spectral representation due to the obvious degeneration of
      // the spectrum.

   Observable(const Word& w, Z d, const R_Func& f);
      // meaning of w, d, s as above (except the case "d")
      // but directly creates a function of the operator. This avoids
      // double application of spectral decomposition.

   Observable(const Word& w, Z d, const F<R,C>& f);
      // Just as the last function, but the function is allowed to be
      // complex-valued in order to create e.g. unitary operators

// operators related to spin 1/2, thus d=2
   explicit Observable(const Word& w);
      // w="sx", "sy", "sz". Gives spin-1/2 operators for h=1
      // (not hQuer=1!).
      // w="sigma_x", "sigma_y", "sigma_z" Pauli matrices ( e.g. Landau,
      // Lifschitz, III (55.7)) sx=sigma_x/(4*Pi), ...

   Observable(const Angle& phi, const R3& n);
      // unitary representation operator belonging to the
      // spatial rotation described by angle phi and axis n
      // n has not to be normalized but only the unit vector counts.

   C_Matrix const& c()const{ return rep;}
   C_Matrix c(){ return rep;}

   C operator()(Z i, Z j)const{ return rep.cyc(i).cyc(j);}
      // does not assume 1<=i,j<=dim()

   C getLean(Z i, Z j)const
      { return rep.cui(i).cui(j);}
   // fast read access to matrix elements without index range check

   void setLean(Z i, Z j, const C& z)
      { rep.cui(i).cui(j)=z;}
   // fast write access to matrix elements without index range check

   void set(Z i, Z j, const C& z){ rep.cyc(i).cyc(j)=z;}

 //  void set(Z i, const State& s){ rep.cyc(i)=s.rep_;}
      // a opservable can be considered as an array of rows, these rows
      // being instances of State. This function places s at row position
      // (starting with 0 !) i

   Observable symmetricPart(void)const
   { return (*this + (*this).con())*C(0.5);}

   Observable unitaryVersion()const;
      // if *this is nearly unitary, the return value will be so
      // to much better approximation (not certain !!!!)

   Observable cut()const;

   Observable ran(Z j=0)const;

   Z dim(void)const{return rep.dim();}

   R autoStep(State const& x)const;

   State operator *(State const& x)const;
      // Observable acts on state x

   State operator()(State const& x)const{ return (*this)*x;}

   SpectralObject spectralRepresentation(bool enfSym=false,
      bool test=false, Z iterMax=128, R eps=0,
      R eps1=1e-7, R eps2=1e-6)const;
      // returns a complete list of eigenvectors (second component of
      // return value together with a list of the corresponding
      // eigenvalues.
      // The algorithm works properly (and gives meaningfull results)
      // only if the observable *this is self-adjoint.
      // If the argument enfSym (i.e.'enforce symmetry') is true only
      // the self-adjoint part of it will be used for calculating
      // the spectral representation. If one can be sure that *this is
      // self-adjoint, one may use the argument.
      // The iterMax parameter limits the allowed iterations in a
      // subroutine which diagonalizes three-band matrices. Reaching the
      // limit stops the program with an error.
      // eps and n control phase adjustment of the eigenvectors to make
      // the visual representation of the eigenvectors regular
      // (see State::adjPhase(...))
      // The two last argument control the transition from the real form
      // to the complex one and should normally not need to be set
      // different from their default value. Only for consciouse
      // experimentation.
      // In the present implementation (which is based on Eigen), the last
      // four arguments are not active.

   R_Matrix toRealForm()const;
      // returns a (2*d,2*d) matrix

   void fromRealForm(const R_Matrix& r);
      // changes *this into the one resulting from the real matrix r

   Observable operator!()const
      // pseudoinverse of *this based on corresponding function of real
      // form
   {
      Observable res;
      res.fromRealForm(!toRealForm());
      return res;
   }

   Observable cayley()const;
      // returns the Cayley transform of *this

   State expi(R t, State const& s, R eps=1e-6);
      // A.expi(t,s) is exp(i*t*A)*s according to the Chebyshev
      // single step method advocated by Hans De Raedt

   CC4 toCC4()const{ CC4 res;
      for (Z i=1;i<=4;++i){
      for (Z j=1;j<=4;++j) {res[i][j]=(*this)(i,j);}
      }
      return res;
   }

   CC3 toCC3()const{ CC3 res;
      for (Z i=1;i<=3;++i){
      for (Z j=1;j<=3;++j) {res[i][j]=(*this)(i,j);}
      }
      return res;
   }

   CC2 toCC2()const{ CC2 res;
      for (Z i=1;i<=2;++i){
      for (Z j=1;j<=2;++j) {res[i][j]=(*this)(i,j);}
      }
      return res;
   }

   void show(Frame& f, Z mult=-1, Z iShift=0, Z jShift=0,
             bool markOnly=false)const;
      // one matrix element will be represented as a square of
      // mult*mult pixels. The whole image will be shifted by
      // (iShift,jShift) with respect to the left upper corner
      // of the graphics window f. ( i: vertical, j: horizontal
      // as in matrices)
      // If mult<0, then the pixel-size will be chosen such that
      // the frame f get filled as much as possible

   void showImg(Graph& f, Z ipolMeth=1)const;
      // show by means of a C+- image class
      // similar to present method but makes automaticly
      // best use of size of f and interpolates bilinearly
      // for ipolMeth=1 (for 0 we use simple pixelation)
      // for ipolMeth=2 averaging interpolator (not yet
      // correctly working, but implementation OK in structure.

    void mark(Graph& gr, Z ipolMethod=0)const;
         // modern implementation, writes on image memory.
         // Visible only upon gr.vis().

   R sup()const;
      // maximum absolut value of a matrix element of *this

   R normIter(R eps=1e-3)const;
      // Norm of the observable, computed by the power iteration method.
      // eps is a parameter which controls finishing the iteration.

   R expVal(const State& a)
      { State a1=a; a1.normalize_(); return (a1|((*this)*a1))[1];}
      // expectation value determined by state and observable
      // a needs not to be normalized.

   Type operator ||(const Type&)const;
      // tensor product; member function is preferred due to more
      // transparent behavior in the presence of namespaces

   Type prdFock()const;
      // Returns the operator (*this)||unit + unit||(*this)

   Type prdFock2()const;
      // Returns the operator (*this)||(*this)

   static Type pot(R_Func const& f, Z d, R lambda=1.);
      // Defines an observable obs such that
      // obs(i,j)=f(positions(i)/size)*krn(i,j)*lambda.
      // So the function f is (as always in a similar context)
      // assumed to be meaningfully defined on Iv(-1/2,+1/2)
      // Since this observable is diagonal in position space,
      // it also could be obtained as a function of the position
      // operator

    static Type pot2(R_Func const& f, Z d, R lambda=1.);
      // This is a quite different concept of a potential
      // compared to the previous one. It is designed to
      // be a valid input to
      //    FieldState::cook2(Observable const&)
      // and
      //    FieldOpertor::cook2(Observable const&)
      // that then describes pair interaction by the same
      // potential which as pot(f,d,lambda) would describe
      // interaction with a fixed force center at the origin.
      // Notice that then the same function f had to be used
      // in pot(f,...) and potential(f,....) but that in the
      // latter case arguments of f in Iv(-1,1) will occur.
};

///////////////////// class SpectralMeasure ///////////////////////////

// combines a set of real spectral values and a set of orthonormal
// set of projection operators into a single object which according to
// the spectral theorem results from spectral analysis of a self-adjoint
// operator.

   class SpectralMeasure{ // eigenvalues and spectral projectors
      Z d;
         // dim of statespace
      Z ds;
         // dim of spectrum
      V<R> spectrum;
      V<Observable> projectors;
   public:
      typedef SpectralMeasure Type;
      CPM_IO
      CPM_NAM(Type)
      Z dim()const{ return d;}
      Z dimSpc()const{ return ds;}
      const R& operator()(Z i)const{ return spectrum[i];}
         // eigenvalues by ()
      const Observable& operator[](Z i)const{ return projectors[i];}
         // eigenvectors by []

      SpectralMeasure():d(0),ds(0),spectrum(ds),projectors(ds){}

      explicit SpectralMeasure(const Observable&);
         // generation from a Observable

      explicit SpectralMeasure(const SpectralObject&);
         // and from a SpectralObject

      void map(const R_Func& f)
      {for (Z i=1;i<=ds;i++) spectrum[i]=f(spectrum[i]);}

      R test()const;
         // returns a error sum which is large if the spectral data
         // are not correct

      State operator *(const State&)const;
      // SpectralMeasure acts on state just as the underlying observable

      X2<R,State> reduction(const State&)const;
      // performs the 'collapse of the wave packet' of the state
      // given by the argument and the observable of spectral
      // representation *this. Returned is the 'measured value' and the
      // state after measurement
   };

} // namespace
#endif
