//? cpmp.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_P_H_
#define CPM_P_H_
/*

Purpose: defining classes of smart pointers, and polymorphic arrays.
classes Pp<>, Po<>, Vp<>.
class P<> is defined in cpmuc.h

Notice usage of the new macro definition (2017-05-23)
 #define const& const&
 from cpmbasictypes.h

*/

#include <cpmv.h>

namespace CpmArrays{

///////////////////////  class Pp< >  //////////////////////////////////
//
// class of non-constant smart pointers implementing 'copy on write' (see
// cpmv.h) and polymorphism (indicated by the 'p'). Let us first discuss
// the 'mono-morphic' aspects, i.e. we discuss the properties of Pp<T>
// for arbitrary but fixed T and not for a variety of T's:
//
// Let T be any non-abstact class that defines a member function
//
//       Tb* T::clone(void)const{ return new T(*this);}  (ii)
// (Notice that T has to define a copy constructor for this to work)
//
// where Tb is either T itself or a base class of T which declares
//
//       virtual *Tb Tb::clone(void)const{ return new Tb(*this);}
// or
//       virtual *Tb Tb::clone(void)const=0;
//
// Then, let us say that T is cloneable, with clone base Tb. Then the
// main property of the Pb<> template is:
//
//    cloneable T ==> value class Pb<T>
//
// If the source code of a non-abstract T is open to modifications, it
// can be made cloneable by inserting the definition
//    T* T::clone()const{ return new T(*this);}
// (lets ignore the complication that T may already use the name
//  clone in a conflicting manner).
//
// Properties of Pp<T> as a 'polymorphic container':
// ------------------------------------------------
//
// We consider a finite set L of cloneable classes which all have the
// same clone base T0. Let T1 be one of those and let L(T1) denote the
// classes in L which are derived from T1. Then Pp<T1> is a polymorphic
// container for L(T1) in the following sense:
// Let pp be an instance of Pp<T1>, created e.g. as
//    Pp<T1> pp;
// and let T2 be a class belonging to L(T1), and let t2 be an instance of
// T2.
//
// Then t2 can be put into the container pp by the
// statement
//
//       pp=t2; (or also pp.set(t2);) (**)
//
// Now pp carries just the information that was carried before (and is
// carried still) by t2. Warning: using to the right of = a ternary
// expression (... ? ... : ...) may destroy polymorphism. This is no real
// problem since the ternary can always replaced by if (...) ... else ...
// .
// If t2 gets changed or destroyed (e.g. as a local variable going out of
// scope) pp, will conserve the information and can return it in two
// ways.
// 1. Reconstruction as object: T2 t2_=pp(T2()) will create an object t2_
//    which is equivalent to t2.
// 2. Reconstruction by action: Let
//
//       virtual void T1::show(Media&)const;
//
// be a function which creates a detailed textual or graphical
// description of the calling instance by interacting with a class Media
// that provides the basis for that. Then 'Media m; pp().show(m);' will
// leave m in a state that carries the same record that we would get from
// 'Media m; t2.show(m)'. Technically, pp() is a quantity that is typed
// T1&; the object it refers to is of type T2 however.
// Let T3 be another class derived from T1, then pp, after pp.set(t3)
// behaves as T3 object. Thus pp behaves changing its type depending on
// initialization.
// Pp<> is closely related to the surrogate classes as discussed in
// Andrew Koenig: Ruminations on C++, AT&T 1997 Chapter 5. As stated
// there, the main benefit of such a 'polymorphic class' is that its
// instances can be put into a strictly typed container. This will be
// carried out in defining the extremely useful class Vp<>.
//
// Note that that in line (**) a constuction
// Vp<T1> pp=t2
// for a not yet existing pp does not work. The safe mode of operation is:
// 1. create the container; 2. fill the container.

template <class T>

class  Pp: public P<T>{ // polymorphic smart pointers
   typedef P<T> Base;
public:
    Pp(T* p=0):P<T>(p){}
      // carrying over the constructor from P. Notice that the
      // primary act of construction was to initialize p_. This can be by
      // T* p_=new Td(....) where Td is some derived class (or T itself)

    explicit Pp(T const& t):P<T>(static_cast<T*>(t.clone())){}
    // 2004-09-27: finally a pointer-less interface of Pp<> enabled!
    // The queer casting is needed in the general situation that
    // T0 and T1 are different (in the description above). This general
    // situation is probably not really important (I had it
    // unintentionally for T0=RigidObject and T1=ExtSys, but
    // simplified it to T0=T1=ExtSys without encountering problems)
    // Notice, however, that t.clone() involves call to the copy
    // constructor of T which might be expensive.

    Pp(const P<T>& x):P<T>(x){}
      // down-cast constructor

   T& operator()(void){ return *Base::p_;}
      // Pp<T> pp;
      // T t;
      // pp=t;
      // makes t.aVirtualFunction(...);
      // and
      // pp().aVirtualFunction(...);
      // calling the same function with the same data.

   T const& operator()(void)const{ return *Base::p_;}
      // not clear why the MS-compiler cannot infer this definition from
      // the base class.
      // Pp<T> pp;
      // T t=pp()
      // is the usual way to retreive the value of pp. If the value is
      // known to belong to a derived class Td, use the next function:
      //    Td t=pp(Td()).
      // If we don't want to retreive t but only want to call a
      // (potentially virtual) member function, one always proceeds as in
      // the following example:
      // Word w = pp().nameOf(); // see also comment to previous function

   template <class X>
      X operator()(X const& x)const
      // retreiving an element from the container as a copy of a
      // prescribed type.
      // The value of x has no influence, only the type matters.
      // T has to be a clone base of X.
      { X* px=static_cast<X*>(Base::p_->clone()); return X(*px);}

   template <class X>
   bool get(X& x)const
      // retreiving the element from the container as a reference
      // to the prescribed type. The value of x at input has no
      // influence, only the type matters. T has to be a clone base of X.
   { X* px=static_cast<X*>(Base::p_->clone()); x=X(*px); return true;}

   void set(T const& x){ cow(); Base::p_=static_cast<T*>(x.clone());}
      // loading an object of type T or of a derived type into Pp<T>
      // no longer needed since assignment works now

   Pp<T>& operator=(T const& x){ cow(); return *this=Pp<T>(x);}
      // 2004-09-27 together with a redefinition of
      // Vp<>::operator[]() this allows the intialization
      // of components of polymorphic arrays by assignment
      // so that the less idiomatic set(i,T $)-functions
      // should no longer be needed. The problems with the old
      // attempts to set values of components via assignment are
      // now understood (and eliminated)
   Word nameOf()const
   {
      return Word("Pp< "&CpmRoot::Name<T>()(T())&" >");
   }

protected:
   void cow(void);
      // 'copy on write'. The first '*this-changing' statement in the
      // body of a non-constant member function has to be cow();
};

template <class T>
void  Pp<T>::cow(void)
{
   if (Base::u_.makeOnly()){
      if (Base::p_!=0) Base::p_=static_cast<T*>(Base::p_->clone());
         // let Tb be the class closest to the root of the class
         // hierarchy where a virtual ... clone(void)const is defined,
         // then present clone is Tb* T::clone()const. Notice that
         // modern compilers sometimes accept T* T::clone()const.
         // Microsoft Visual C++ 5.0 is not of this kind,
         // so in defining clone() in a class, we have always to
         // remember the base. This is only a minor inconvenience.
      Base::u_.startNew_();
   }
}

/////////////////////////// class Po< > /////////////////////////////////

template <class T>
class  Po: public  Pp<T>{  // Adding order operations to Pp<T>
     // Allows to form, for instance, std::vector<Po<T> >
     // and thus allows to combine STL functionality with polymorphism
     // and banning of pointers
   typedef Pp<T> Base;
public:
   Po(T* p=0): Pp<T>(p){}
   Po(const Pp<T>& x): Pp<T>(x){}
   bool operator  <(const  Po& x)const{ return *Base::p_ <  *(x.p_);}
   bool operator ==(const  Po& x)const{ return *Base::p_ == *(x.p_);}
   bool operator  >(const  Po& x)const{ return *Base::p_ >  *(x.p_);}
   bool operator !=(const  Po& x)const{ return *Base::p_ != *(x.p_);}
   bool operator <=(const  Po& x)const{ return *Base::p_ <= *(x.p_);}
   bool operator >=(const  Po& x)const{ return *Base::p_ >= *(x.p_);}
};

///////////////////////////  class Vp< >  ///////////////////////////////
// This template behaves just as Pp<T> except that access is indexed
// which allows to store a number of instances and to retreive them:
// Let T,T1,T2,T3 as in the explanation of Pp<>. Then a typical usage is
//
// T1 t1(...);
// T2 t2(...);
//
// V<T3> t3(2);
// t3[1]=T3(...);
// t3[2]=T3(...);
//
// Vp<T1> vp(4);
// vp[1]=t1;
// vp[2]=t2;
// vp[3]=t3[1];
// vp[4]=t3[2];

// A detailed description (see Pp<>) of the the vp-components can be
// created like
// that:
//
// Z nItems=4;
// V<Media> m(nItems);
// (for Z i=1;i<=nItems;i++) vp(i).show(m[i]);
//
// Thus, despite their differing types, the t1, t2, t3[1], t3[2] get
// processed in a single loop; this capability is indispensible for
// polymorphism to be useful in non-trivial situations. The most
// expressive application is to use Vp<T1>-typed data members in classes:
//
// class ParticleMix{ // assume that Particle is clone base for both
//    // ParticleA and ParticleB
//    Vp<Particle> p;
// public:
//    ParticleMix(Z nA, Z nB):p(nA+nB)
//       // creates a mix of nA particles of type A and
//       // and nB particles of type B
//    {
//       Z i;
//       V<ParticleA> pA(nA); for (i=1;i<=nA;i++) p[i]=pA[i];
//       V<ParticleB> pB(nB); for (i=1;i<=nB;i++) p[i+nA]=pB[i];
//    }
//    void moveIndependently(R timeStep)
//    {
//       for (Z i=1;i<=p.dim();i++) p(i).moveIndependently(timeStep);
//             // applies the law of motion for A-particles if p(i) is
//             // a A-particle, and for B-particles correspondingly
//    }
// };
//
// Whereas in a 'stand alone'-control structure like the 'Media-loop'
// given above it would not be much more complicated to use a T1 pointer
// for implementing the loop, in the class example we gain more: Since
// Vp<Particle> is a value class, the compiler will properly and
// automatically define copy constructor, assignment operator, and
// destructor for us, wheras using pointers would leave this burden on
// us. Programmers who did this pointer implementation several times
// successfully might begin to love the procedure. I consider it wasted
// time and energy, and a lost opportunity to deal with complexity in a
// state-of-the-art manner.

template <class T>
class  Vp{ // polymorphic V template

   CpmArrays::V< Pp<T> > rep;

   explicit Vp(V< Pp<T> > const& vp):rep(vp){}

public:

    Vp(){}
      // notice that  both Pp<T> and V define a default constructor

   explicit Vp(Z n):rep(CpmArrays::V<Pp<T> >(n,Pp<T>())){}
      // No pre-setting of values occurs. If needed, this has to be
      // done by function void set(T const& x).

 //  Vp(Z n, Pp<T> const& p):rep(CpmArrays::V<Pp<T> >(n,p)){}
      // 'Making a series of copies of a single polymorphic object'.

 //  Pp<T> at(Z i)const{ return rep(i);}

   Pp<T>& operator[](Z i){ return rep[i];}
      // We use [] for access to the true components, which
      // are instances of the true value class Pp<T>.
      // Typical usage:
      // Vp<T> vp(2); T x=..., y=...; vp[1]=x; vp[2]=y;
   Pp<T> const& operator[](Z i)const{ return rep[i];}

   T& operator()(Z i){ return rep[i]();}
   T const& operator()(Z i)const{ return rep[i]();}
      // We use () for access to the T& that will be in
      // most cases the objects needed for further
      // processes, as for usage as function arguments.
      // Typical usage:
      // Vp<T> vp=...; Word w1=vp(1).nameOf(), w2=vp(2).nameOf();

    //  void operator<<(Pp<T> const& p){rep<<p;}

   template <class X>
   X operator()(Z i, const X& x)const
      // retreiving the i-th component from the container as a copy of a
      // prescribed type. The value of x has no influence, only the type
      // matters. T has to be a clone base of X.
      { X* px=static_cast<X*>(rep[i]().clone()); return X(*px);}

   template <class X>
   bool get(Z i, X& x)const
      // retreiving the i-th component from the container as a referemnce
      // to the prescribed type. The value of x at input has no
      // influence, only the type matters. T has to be a clone base of X.
   {
      if (i<1 || i>rep.dim()) return false;
      X* px=static_cast<X*>(rep[i]().clone()); x=X(*px);
      return true;
   }

   void set(Z i, const T& x){ rep[i]=x;}
      // no longer essential since the defining statement
      // is natural enough not to need a wrapper

   void set(T const& x){ for (Z i=1;i<=rep.dim();i++) rep[i]=x;}

   Z dim()const{ return rep.dim();}
   Z b()const{ return rep.b();}
   Z e()const{ return rep.e();}
   bool valInd(Z i)const{ return rep.valInd(i);}

   Vp<T> rev()const
      //: reversed
   {
      return Vp<T>(rep.rev());
   }

   Word nameOf()const{
      Word nt=CpmRoot::Name<T>()(T());
      Word wi="Vp<";
      return wi&nt&">";
   }
};

} // namespace

#endif
