//? cpmpsi.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmpsi.h>

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;
using namespace CpmFunctions;

using CpmWaves::Psi;
using CpmWaves::HarmOsc;

///////////////// class Psi /////////////////////////////////

Psi::Psi(Z d0, Iv iv0, R m0, R k0, const R_Func& V0):
   d(d0),iv(iv0),m(m0),k(k0),method(2),V(V0),psi(d),psiDot(d),x(d)
{
   dx=iv.abs()/R(d);
   R x0=iv.inf()+dx*R(0.5);
   for (Z i=1;i<=d;i++){
      x[i]=x0;
      x0+=dx;
   }
}

C_Vector Psi::dot(const C_Vector& psi0)
{
   R fac=-hq2/(dx*dx*m*2);
   C_Vector res(d);
   cpmassert(d>1,"Psi::dot");
   for (Z i=1;i<=d;i++){
      C val=fac*(psi0.cyc(i-1)-2*psi0[i]+psi0.cyc(i+1));
      val+=V(x[i])*psi0[i]*k; // up to here this is the Hamiltonian
      val*=ihqi; // times 1/(i*hQuer)
      res[i]=val;
   }
   return res;
}

void Psi::setState(const R_Func& psiRe, const R_Func& psiIm)
{
   for (Z i=1;i<=d;i++){
      C val(psiRe(x[i]),psiIm(x[i]));
      psi[i]=val;
   }
   psiDot=dot(psi);
   normalize();
}

void Psi::step(R dt)
// intentionally coded rather compactly (following a proposal
// of Thomas Dera)
{
   if (method==0){ // Euler method
      psi+=psiDot*C(dt);
      psiDot=dot(psi);
   }
   else if (method==1){ // explicit midpoint method for velocity
      R dth=dt*0.5;
      C_Vector v=dot(psi+psiDot*C(dth));
      psi+=v*C(dt);
      psiDot=v*C(2.)-psiDot;
   }
   else { // explicit midpoint method for acceleration
      R dth=dt*0.5;
      psi+=psiDot*C(dth);
      C_Vector a=dot(dot(psi));
         // acceleration for midpoint configuration
      psi+=(psiDot+=a*C(dt))*C(dth);
   }
}

R_Matrix Psi::rep()const
      // graphical object representing the amplitude distribution
{
   const Z mesLim=3;
   cpmmessage(mesLim,"Psi::rep() started");
   R_Matrix res(4,d);
   for (Z i=1;i<=d;i++){
      res[1][i]=x[i];
      res[2][i]=psi[i][1];
      res[3][i]=psi[i][2];
      res[4][i]=psi[i].abs();
   }
   cpmmessage(mesLim,"Psi::rep() done");
   return res;
}

R_Matrix Psi::repReal(R n)const
//not OK since creating internal compiler error
{
 //  const R lambdaFactor=6.123;
   R_Vector re=psi.real();
   R_Vector im=psi.imag();
   R_Func reF(x,re);
   R_Func imF(x,im);
   Z dn=cpmrnd(n*R(d));
   R_Matrix res(2,dn);
   R xStart=x[1];
   R xEnd=x[d];
   R ds=(xEnd-xStart)/(dn-1);
//   R fac=(dn-1.)/(d-1.);
   R s=xStart;
//   R lambda=lambdaFactor*ds;
//   R omega=cpm2pi/lambda;
   for (Z i=1;i<=dn;i++){
      res[1][i]=s;
   //   R zre=reF(s);
    //  R zim=imF(s);
//      res[2][i]=(C(zre,zim)*CpmRoot::expI(omega*s))[1];
      s+=ds;
   }
   return res;
}

R Psi::normSquare()const
{
   return psi.absSqr();
}

R Psi::normalize()
{
   R pn=norm();
   R pnInv=R(1)/pn;
   psi*=C(pnInv);
   psiDot*=C(pnInv);
   return pn;
}

C Psi::energy()const
{
   return (psi|psiDot)*ihq;
}

R Psi::naturalTime()const
{
   R p2=psi.absSqr();
   R pd2=psiDot.absSqr();
   R tSquare=p2/pd2;
   return cpmsqrt(tSquare);
}

R Psi::naturalTimeStep()const
{
   const R fac=0.5;
   return fac*naturalTime()/R(d);
}

void Psi::boost(R v)
{
   C fac=v*m*ihqi;
   for (Z i=1;i<=d;i++){
      psi[i]*=exp(x[i]*fac);
   }
   psiDot=dot(psi);
}

void Psi::dephase(R dPhi)
      // apply to the psi-value at each point a stochastically
      // chosed factor exp(i*dt*phiDot*random) where random
      // varies in [0,1]
{
   Z j=0;
   for (Z i=1;i<=d;i++){
      psi[i]*=C::expi(dPhi*randomR(j++));
   }
}

bool Psi::prnOn(ostream& str)const
{
   cpmp(psi);
   cpmp(psiDot);
   return true;
}

bool Psi::scanFrom(istream& str)
{
   cpms(psi);
   cpms(psiDot);
   return true;
}

//////////////////// class HarmOsc ////////////////////////////

namespace {

   Z fac2(Z n) // (2^n)*(n!)
   {
      if (n==0) return 1;
      else return 2*n*fac2(n-1);
   }

   R hermite(R x, Z n)
   {
      if (n==0) return 1;
      else if (n==1) return x*2;
      else return x*hermite(x,n-1)*2-R(2*(n-1))*hermite(x,n-2);
   }

   R u(const R& x, const Z& n, const R& alpha)
   {
      R xi=alpha*x;
      return hermite(xi,n)*cpmexp(R(-0.5)*xi*xi);
   }

   R kOsc(R m, R f)
   {
      R omega=cpm2pi*f;
      return omega*omega*m;
   }

   R alphaOsc(R m, R f)
   {
      R k=kOsc(m,f);
      R alphaSquare=cpmsqrt(m*k)/CpmWaves::hq; // Schiff p.67 (13.3)
      return cpmsqrt(alphaSquare);
   }

   Iv ivOsc(R m, R f, R n)
   {
      const R saftyFac=1.4;
      using CpmWaves::hq;
      R e=cpmpi*(n+0.5)*hq*f*4;
      R k=kOsc(m,f);
      R x0=cpmsqrt(e/k);
      x0*=saftyFac;
      return Iv(-x0,x0);
   }
}

HarmOsc::HarmOsc(Z d0, R m0, R f0, R n0):
Psi(d0,ivOsc(m0,f0,n0),m0,kOsc(m0,f0),R(0.5)*R_Func::x2),f(f0)
{
   alpha=alphaOsc(m,f);
}

R HarmOsc::N_n(Z n)const
{
   R resSquare=R(alpha)/(cpmsqrt(cpmpi)*R(fac2(n)));
   return cpmsqrt(resSquare);
}

void HarmOsc::setQuantumNumber(Z n)
{
   R_Func re=eigenFunction(n);
   setState(re);
}

R_Func HarmOsc::eigenFunction(Z n)const
{
#ifdef CPM_Fn
   return F2<R,Z,R,R>(n,alpha)(u);
#else
   return F<R,R>(bind(u,_1,n,alpha));
#endif
}
