//? cpmpsi.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_PSI_H_
#define CPM_PSI_H_

/*
   Description: quantum wave functions in one dimension.
   Time development according to the explicit midpoint method.
*/

#include <cpmrfunc.h>
#include <cpmrmatrix.h>
#include <cpmcvector.h>
#include <cpmconstphys.h>

namespace CpmWaves{

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::B;
using CpmArrays::C_Vector;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmFunctions::R_Func;
using CpmFunctions::F;
using CpmGeo::Iv;

const R hq=CpmPhysics::h__PHYS;
   // hQuer
const R hq2=hq*hq;
   // hQuer^2
const R hqi=R(1)/hq;
   // hQuerInv
const C ihq=C(0,hq);
   // i*hQuer
const C ihqi=C(0,-hqi);
   // 1./(i*hQuer) = (i*hQuer)Inv

class Psi{ // 1-dimensional discrete wave function
   // evolving according to Schroedinger's equation given by particle mass
   // and a potential function

protected:

   Z d{0};
      // number of space points that make up the space
      // Thus d=psi.dim();
   Iv iv;
      // available space as an interval
   R m{0.};
      // mass
   R k{0.};
   Z method{0};
   R_Func V;
      // potential
   C_Vector dot(const C_Vector&);
      // creator of the time derivative, up to a factor ihqi
      // the Hamiltonian determined by m and V
      // H psi=-hq2*psi''/2m + k*V
   C_Vector psi;
   C_Vector psiDot;
   R dx{0.};
   R_Vector x;
      // dependent
public:
   typedef Psi Type;
   CPM_IO

   Psi(){}
   Psi(Z d0, Iv iv0, R m0, R k0, const R_Func& V0);
   void setState(const R_Func& psiRe, const R_Func& psiIm=R_Func());
   void step(R dt);
   void dephase(R dPhi);
      // apply to the psi-value at each point a stochastically
      // chosed factor exp(i*dPhi*random) where random
      // varies in [0,1]
   Z dim()const{ return d;}
   R get_dx()const{ return dx;}
   Iv getSpace()const{ return iv;}
   R_Matrix rep()const;
      // graphical object representing the amplitude distribution
   R_Matrix repReal(R n=3)const;
      // real representation according to electrical engineer's
      // interpretation of complex numbers. The argument n
      // governs the resolution

   void setMethod(Z method0){ method=method0;}
   R norm()const{ return cpmsqrt(normSquare());}
   R normSquare()const;
   R normalize();
      // makes a unit-vector
   C energy()const;
      // expectation value of the Hamiltonian. Could
      // get a small imaginary component due to numerical
      // in-accuracies, so we return a complex number

   R naturalTime()const;
      // a span of time associated with psi and psiDot
      // not taking account discretization

   R naturalTimeStep()const;
      // returns a time span which is a save time step
      // for the non-Euler time stepping methods.
      // Using only half this time step is even more save
      // and using twice the step seems to work also.
      // Taking 4 times this as a time step is asking for trouble.
      // Closely related to Courant's criterion

   void boost(R v);
      // active transformation in which an coordinate system in
      // relative motion plays the same role, that in the previous
      // function was played by a shifted system
      // See Fonda, Ghirardi p. 89, (2.111)
      // changes *this into the 'boosted state'
};

class HarmOsc: public Psi{ // 1-dimensional harmonic oscillator

   R f{0.};
      // frequency of the oscillator
   R alpha{0.}; // see Leonhard I. Schiff pp 70,...

public:
   HarmOsc():Psi(){}
   HarmOsc(Z d0, R m0, R f0, R n0);
      // d0  number of space points that make up the space
      // m0 mass of the oscillating particle
      // f0 frequency of oscillation
      // n0 determines the space interval iv in a way that
      //    a state with quantum number n0 is vitually unaffected
      //    by the limitation of space
   R N_n(Z n)const; // normalization factor of eigen functions.
      // Nice recursive definition, not needed since states
      // get normalized anyway.
      // Schiff p.71 (13.17)
   void setQuantumNumber(Z n);
   R getAlpha()const{ return alpha;}
   R_Func eigenFunction(Z n)const;
      // is chosen real
};

} // namespace

#endif
