//? cpmqm.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_QM_H_
#define CPM_QM_H_
/*
   Description: An attempt to code quantum theory more clearly
*/

#include <cpmv.h>
#include <cpmc.h>
#include <cpmgraph.h>
#include <cpmrecordhandler.h>
#include <cpmclinalg.h>
#include <map>

using CpmRoot::B;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::Word;

using CpmRootX::RecordHandler;

using CpmGeo::Iv;

using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::X2;
using CpmArrays::X3;
using CpmArrays::R_Vector;
using CpmArrays::Outside;

using CpmFunctions::F;

using CpmGraphics::Graph;
using CpmGraphics::Img24;
using CpmGraphics::ColRef;

using CpmLinAlg::R2;
using CpmLinAlg::R3;
using CpmLinAlg::C2;
using CpmLinAlg::C4;
using CpmLinAlg::CC2;
using CpmLinAlg::CC4;
using CpmLinAlg::CCCC2;
using CpmLinAlg::Z4;

using CpmImaging::Color;

using std::map;
using std::vector;
using std::endl;

// List of classes to be introduced (in this order!) should help not to
// get lost:

// Lat1, DynDat, RunCon, DisCon,
// SysLat1, TrjLat1, IntActDat, SysLat2, SysLat3,
// TrjLat2, TrjLat3,
// Sep2, Sep3, Fun3, Ia, SpinN
// Ky. QregPar, Qreg, TrjQreg

// interfaces as declaration macros

#define CPM_QM_PURE_VECTOR_SPACE_INTERFACE(X,S)\
public:\
   X operator - (void)const;\
   X& operator *=(S);\
   X& operator +=(X const&);\
   X& operator -=(X const&);\
   X operator *(S)const;\
   X operator +(X const&)const;\
   X operator -(X const&)const;

#define CPM_QM_PURE_SCALAR_INTERFACE(X,S)\
public:\
    X& operator *=(S);\
    X operator *(S)const;

#define CPM_QM_SCALAR_PRODUCT_INTERFACE(X,S)\
public:\
   X con(void)const;\
   S operator |(X const&)const;\
   R normSqr(void)const{ return (operator|(*this)).real();}\
   R norm(void)const{ return cpmsqrt(normSqr());}\
   R abs(void)const{ return norm();}\
   R absSqr(void)const{ return normSqr();}

namespace CpmQM{

// Within this namespace we alway use units in which the numerical values
// of hDirac and c are 1. hPlanck thus has value 2*pi (=CpmRoot::Pi2).

typedef X2< V< V<R2> >, V<Word> > x2vv;

/////////////////////////// class Lat1 ///////////////////////////////////

class Lat1{ // lattice in one dimension
   // Many quantities of interest depend on a mass, which, however, is
   // not part of the concept. However, there is a natural maximum mass
   // associated with the lattice: mMax()

// independent quantities:
   Iv iv_;
      // biotope as an interval
   Z d_;
      // number of points in the biotope
// dependent quantities:
   R size_;
   R ds_;
   R m_;
      // maximum mass
   R p_;
      // maximum momentum

   void ini_()
   {
      cpmassert(d_>0,"Lat1::ini_()");
      size_=iv_.abs();
      ds_=size_/d_;
      cpmassert(ds_>0,"Lat1::ini_()");
      m_=R(0.5)/ds_;
      p_=cpm2pi*m_;
   }
   typedef Lat1 Type;
public:
   CPM_IO
   CPM_ORDER
   CPM_NAM(Lat1)
   Lat1(Iv const& iv, Z d):iv_(iv),d_(cpmpos(d)){ini_();}
   Lat1(R size, Z d):iv_(size*0.5),d_(cpmpos(d)){ini_();}

   Iv iv()const{ return iv_;}
   Z d()const{ return d_;}
   R size()const{ return size_;}
   R ds()const{ return ds_;}
   R m()const{ return m_;}
   R fir()const{ return iv_.inf()+ds_*0.5;}
      //: first
      // returns the first of the discrete positions associated
      // with *this
   R last()const{ return iv_.sup()-ds_*0.5;}
      //: last
      // returns the last of the discrete positions associated
      // with *this

   void next_1(R& x)const{ x+=ds_;}
      //: next
      // shifts the argument x by ds and thus provides a convenient
      // method for iterating over the discrete positions associated
      // with *this

   R pos(Z i)const{ return iv_.inf()+ds_*(i+0.5); }
      // for i=0,1,...d_-1, x runs over the discrete positions associated
      // with *this

   R frqMax()const{ return m_;}
      //: frequency maximum
      // returns the value of the spatial Nyquist frequency fc=d/(2*size)

   R pMax()const{ return p_;}
      //: p maximum
      // pMax=hPlanck*fc

   R mMax()const{ return m_;}
      //: m maximum
      // mMax()==frqMax(), due to hDirac=c=1, masses and spacial
      // frequencies have the same dimension.

   R vMax(R mass)const{ return p_/mass;}
      //: v maximum
      // Velocity which gives a particle of mass m just the maximum
      // momentum pMax().
      // Notice that a boost operation is essentially a shift operation
      // in momentum space so that state components that get shifted
      // (formally) to values larger than the upper end of the momentum
      // range will enter the momentum range at the other end.
      // Thus applying a boost with vMax to a Gaussian wave packet (at
      // rest) will result in a wave packet with momentum components
      // arround fc and -fc. Under time evolution this packet will split
      // into two beams; for boosts with velocities close to vMax, these
      // two beams are also obvious but their relative strength varies
      // with v. Notice that for mass -> 0 we have vMax -> infinity,
      // so that vMax <c (=1) is not a natural requirement here.
      // Even for the largest allowed mass, vMax(mMax)=2*pi, so that
      // velocities much larger than 1 (=c) are possible.

   R eKinMax(R mass)const{ R dsInv=R(1)/ds_; return dsInv*dsInv*2/mass;}
      //: e kinetic maximum
      // eKinMax=(2/m)*(hDirac/ds)^2, ds=(size/d); maximum kinetic energy
      // in free motion if the Hamiltonian is realized by the wellknown
      // differencing scheme for the second derivative.
      // If one writes this as p^2/(2*m), we get p=2*hDirac/ds
      // This p, which also is a kind of maximal momentum turns out to be
      // smaller than pMax as defined previously. It is smaller in the way
      // 2 is smaller than pi. This is the least-surprising direction ---
      // the opposit would say that the fastest possible motion would
      // have momenta exceeding pMax. For the Hamilton operator defined
      // in momentum space (by discrete Fourier transformation) no such
      // discrepency would occur.

   R tRef(R mass)const{ return size_/vMax(mass);}
      //: t reference
      // time needed for traversing the distance size when travelling at
      // speed vMax(m)

   R dtLim(R mass)const{ return ds_/vMax(mass);}
      //: dt limit
      // time needed for traversing the distance ds() when travelling
      // at speed vMax(m)

   R lamOsc(R mass)const
      //: lambda oscillator
   {
      R x=cpm2pi*d_/(size_*size_);
      return x*x*0.5/mass;
   }
      // (hPlanck*d/(size^2))^2 / (2*mass)
      // Consider the Hamiltonian of a harmonic oscillator:
      // Observable p("p",d), x("x",d);
      // Observable h=p*p/(2*mass)+x*x*lambda;
      // Then, for lambda=QMBase::lambdaOsc(d), a state of maximum
      // momentum will just have the kinetic energy lambda*(size/2)^2.
      // A fictitious state having this momentum and being localized near
      // x=0, would oscillate with an amplitude size*0.5.
      // Let us write  mass = xi*mMax, then one gets
      // lambdaOsc(Z d) = ((2*Pi)^2 * c * hDirac)/(xi * size^3),
      // which is independent of d. This is not too surprising since d is
      // inherent in our parameterization of the  mass as xi*mMax.
   };

//////////////////////////// class DynDat ////////////////////////////////

class DynDat{ // dynamical data
   // Data for defining dynamics, i.e. finally a free and an interacting
   // Hamiltonian.
   R m_;
      //: m
      // mass
   R q_;
      //: q
      // charge, influencing interaction between
      // different particles
// data for defining the binding potential pot(R)
   R lam_;
      //: lambda
      // binding strength
   R x0_;
      //: x0
      // binding position
   R p_;
      //: p
      // power of binding potential.  The particle
      // at place x feels the potential
      // lambda*|x-x0|^p
// nature of boundary conditions
   Z out_; // to be interpreted as of type CpmArrays::Outside
      //: out

   static R beta_;
      //: beta
      // strength of a homogeneous field
      // Corresponds in a sense to the inter-particle
      // coupling constant alpha_ to be introduced in
      // class IntActDat.
   typedef DynDat Type;
public:
   CPM_IO
   CPM_NAM(DynDat)
   DynDat(R m, R q, R lam, R x0, R p, Z out=0 ):
    m_(cpmabs(m)),q_(q),lam_(lam),x0_(x0),p_(p),out_(out){}
   // m_(cpmabs(m)),q_(q),lam_(lam),x0_(x0),p_(p),out_(2){}
      // constructor
   R m()const{ return m_;}
      //: m
   R q()const{ return q_;}
      //: q
   R lam()const{ return lam_;}
      //: lambda
      // value of the coupling constant, see function pot()
   R x0()const{ return x0_;}
      //: x0
   void shf_(R s){ x0_+=s;}
      //: shift
   R pot(R x)const{ return lam_*cpmpow(cpmabs(x-x0_),p_)+beta_*q_*x;}
      //: potential
   R force(R x)const{ return -( x>=x0_ ? R(1) : R(-1))*lam_*p_*
      cpmpow(cpmabs(x-x0_),p_-1)-beta_*q_;}
      // force
      // force(x)=-dpot(x)/dx
   void setOut_(Outside out){ out_= (Z)out;}
      //: set out
      // setting the boundary boundary conditions
   void setLam_(R x){ lam_=x;}
      //: set lambda
   void mulLam_(R x){ lam_*=x;}
      //: multiply lambda
   void set_q_(R x){ q_=x;}
      //: set q
   void mul_q_(R x){ q_*=x;}
      //: multiply q
   Outside out()const{ return (Outside)out_;}
      //: outside
   static R& beta(){ return beta_;}
      //: beta
      // Access to beta_:
      // reading: R x=DynDat::beta();
      // writing: R y=...; DynDat::beta()=y;
      // Is this mechanism realy preferable to having
      // beta_ public?
};

// two structs for compact input from an ini-file
////////////////////////////// class RunCon //////////////////////////////
   struct RunCon{ // run control
      R dt;
      Z nIter;
      R tWaitIter;
      R gammaIter;
      Z nSteps;
      Z viewPeriod;
      Z docuPeriod;
      Z imgPeriod;
      R tWaitIni;
      R tWaitView;
      R tWaitDocu;
      R tWaitDocu2;
      R tWaitDocu3;
      R tWaitRes;
      Z sel;
      B reversed;
      B fromDisk;
      V<Z> imgSteps;
   public:
      RunCon(RecordHandler const& rch, Word const& sec);
   };

////////////////////////////// class DisCon //////////////////////////////
   struct DisCon{ // display control
      Z ipolMet;
      R gamma;
      R secFac;
      B autoScl;
      B showCor;
         //: show correlations
      B showCls;
         //: show classical
         // show also corresponding data for classical particles
      B areaSty;
         //: area style (prefer representation as a color
         // image ove line graphics
      B wrtOnShow;
         //: write on show
   public:
      DisCon():ipolMet(-1),gamma(1),secFac(-1),autoScl(true){}
      DisCon(RecordHandler const& rch, Word const& sec)
      {
         cpmrh(ipolMet);
         cpmrh(gamma);
         cpmrh(secFac);
         cpmrh(autoScl);
         cpmrhf(showCor);
         cpmrhf(showCls);
         cpmrhf(areaSty);
         cpmrhf(wrtOnShow);
      }
   };

//////////////////////////// class SysLat1 ///////////////////////////////

   class SysLat1{ // quantum system on a one-dimensional lattice
      // The data which define a Hamiltonian are not part of the
      // class data; they enter as an argument of function dot_.
      Lat1 lat_;
         //: lattice
         // lattice in space
      V<C> rep_;
         //: representation
         // list of wave-function values on the
         // lattice positions
      void ini_()
         //: initializer
         // completes the construction of the zero-vector
      {
         Z d=lat_.d();
         rep_=V<C>(d);
      }
      void ini_(F<R,C> const& f);
         //: initializer
         // completes the construction of a vector getting
         // values from f
      void ini_(F<R,R> const& f);
         //: initializer
         // completes the construction of a vector getting
         // values from f
      typedef SysLat1 Type;
   public:
      CPM_IO
      CPM_NAM(SysLat1)
      CPM_QM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
      CPM_QM_SCALAR_PRODUCT_INTERFACE(Type,C)
      CPM_QM_PURE_SCALAR_INTERFACE(Type,R)

      SysLat1(Lat1 const& lat):lat_(lat){ini_();}
         // Constructs the zero vector.

      SysLat1(Lat1 const& lat, F<R,R> const& f):lat_(lat) {ini_(f);}
         // Constructs a unit vector such that rep_[i] is proportional
         // to f(lat_.pos()[i]) (with an i-independent factor).

      SysLat1(Lat1 const& lat, F<R,C> const& f):lat_(lat){ini_(f);}
         // Constructs a unit vector such that rep_[i] is proportional
         // to f(lat_.pos()[i]) (with an i-independent factor).

      SysLat1(Iv const& iv, V<C> const& v):lat_(iv,v.dim()),rep_(v){}
         // construction from biotope and wave function
         // no normalization!

      SysLat1(Lat1 const& lat, DynDat const& dd, Z n=0);
         // creates the n-th eigenstate (n=0 corresponds to
         // the ground state) of the Hamiltonian determined by
         // lat and dd.

      Lat1 lat()const{ return lat_;}
         //: lattice

      V<C> rep()const{ return rep_;}
         //: representation

      Z dim()const{ return lat_.d();}
         //: dimension

      C const& cui(Z i)const{ return rep_.cui(i);}
         //: cui (component for unchecked index)
         //       -             -         -
      C& cui(Z i){ return rep_.cui(i);}
         //: cui

      R dtLim(DynDat const& dd)const{ return lat_.dtLim(dd.m());}
         //: dt limit

      R nor_();
         //: normalize

      R s(R x0=0)const
         //: s (shift relative to x0)
         // expectation value of the particle position
         // relative to point x0. Thus <x-x0>, where
         // x is the particle position operator.
      {
         R res=0;
         for (Z i=rep_.b();i<=rep_.e();++i)
            res+=rep_.cui(i).absSqr()*(lat_.pos(i)-x0);
         return res;
      }

      SysLat1& makeReal_(R p=2);
      // returns a SysLat1 with all components of *this
      // replaced by their absolute value to power p

      R enrHomFil(DynDat const& dd)const;
         //: energy (in) homogeneous field

      void dot_(DynDat const& dd);
         //: dot
         // changes *this into its time derivative under the
         // assumption that dynamical data dd are valid

      C eTot(DynDat const& dd)const
         {
            SysLat1 v=*this;
            v.dot_(dd);
            return C::I*(*this|v);
         }

      void mark(Graph& g, DisCon const& dc)const;
         //: mark

      R dtSup(DynDat const& dd, Z p=100, R tWait=-1)const;
      //: dt supremum
      // upper limit for the time step according to
      // the new insights from the preparation of the
      // publication. Tries to estimate the largest
      // energy eigenvalue (better: absolute value of this)
      // if tWait>=0 it shows the process in graphics,
      // each of the intermediate states will be shown on
      // screen and tWait is the inspection time per screen
      // image.

      static void stepGam(R& t, SysLat1& x, SysLat1& v, R dt,
         DynDat const& dd)
         //: step gamma
         // the basic dynamical algorithms with all input in the form
         // of arguments (no internal state)
         //  eq. 8 in my paper
         // 'The direct midpoint method as a  quantum mechanical
         // integrator II'
         // Here it turned out that there are many advantages of such a
         // 'stateless' representation when one wants to use them in
         // various contexts (as in my case where different versions of
         // dynamics are to be compared.
      {
         R tau=dt*0.5;
         t+=tau;
         x+=v*tau;
         SysLat1 vmp(x);
         vmp.dot_(dd);
         v=vmp*2-v;
         t+=tau;
         x+=v*tau;
      }

      static void stepGam2(R& t, SysLat1& x, SysLat1& v, R dt,
         DynDat const& dd)
         //: step gamma 2
         // the basic dynamical algorithms with all input in the form
         // of arguments (no internal state)
         // norm conserving method
         // That two successive steps of stepGam_ are mor
         // natural than a single one was found 2007-11-25.
         // Now the curves of energy and norm have no longer
         // the alternating behavior which they had before.
         // Also the stability limit (according to my preliminary
         // accessment is now the same.)
      {
         R tau=dt*0.5;
         stepGam(t,x,v,tau,dd);
         stepGam(t,x,v,tau,dd);
      }

      static void step(R& t, SysLat1& x, SysLat1& v, R dt,
         DynDat const& dd)
         //: step
      {
          stepGam2(t,x,v,dt,dd);
      }
   };

////////////////////////////// class TrjLat1 /////////////////////////////

   class TrjLat1{ // trajectories of Lat1 systems
      // trajectories of a quantum system on a one-dimensional
      // lattice
      DynDat dd_;
      SysLat1 x_;
      SysLat1 v_;
      R t_;
      typedef TrjLat1 Type;
   public:
      CPM_IO
      CPM_NAM(TrjLat1)
      TrjLat1(DynDat const& dd, SysLat1 const& x, R t=0):dd_(dd),x_(x),
      v_(x_),t_(t){ v_.dot_(dd_);}
         // version in which the velocity is set from the state

      TrjLat1(DynDat const& dd, SysLat1 const& x,
         SysLat1 const& v, R t=0):dd_(dd),x_(x),v_(v),t_(t){}
         // version which also allows to set the velocity

      void step_(R dt){ SysLat1::step(t_,x_,v_,dt,dd_);}

      R t()const{ return t_;}
      R n()const{ return x_.abs();}
      C eTot()const{ return C::I*(x_|v_);}
      C hInt()const;
      C eSep()const{ return eTot()-hInt();}
      R e0()const{ return eSep()[1];}
      R eInt()const{ return hInt()[1];}
      R s()const{ return x_.s(dd_.x0());}
      SysLat1 x()const{ return x_;}
      SysLat1 v()const{ return v_;}

      void mark(Graph& g, DisCon const& dc)const{ x_.mark(g,dc);}

      V< V<R2> > run_(Graph& g, RunCon& rc, DisCon const& dc);

   };

//////////////////////////// class IntActDat /////////////////////////////

   class IntActDat{ // interaction data
      R alpha_;
      R mu_;
      R eps_;
      R p_;
      typedef IntActDat Type;
   public:
      CPM_IO
      CPM_NAM(IntActDat)
      IntActDat(RecordHandler const& rch, Word const& sec)
      {
         cpmrh(alpha_);
         cpmrh(mu_);
         cpmrh(eps_);
         cpmrh(p_);
      }

      IntActDat(R alpha, R mu, R eps, R p):alpha_(alpha),
      mu_(mu),eps_(eps),p_(p){}

      R pot(R x)const
         //: potential
      {
         if (alpha_==0.) return 0;
         else{
            R r=cpmabs(x);
            return alpha_*cpmexp(-mu_*r)*cpmpow(eps_+r,-(p_+1));
         }
      }

      R fi(R x)const
         //: interaction force
      {
         if (alpha_==0.) return 0;
         else{
            R r=cpmabs(x);
            R y1=alpha_*cpmexp(-mu_*r);
            R re=eps_+r;
            R sig=( x>=0 ? 1 : -1);
            return sig*y1*(mu_*cpmpow(re,-p_-1)+(p_+1)*cpmpow(re,-p_-2));
         }
      }

      void setAlp_(R a){ alpha_=a;}
         //: set alpha

      void free_(){alpha_=0;}
         //: free
         // setting all coupling constants equal to zero
         // switches interaction off

      R dis(R x, R y)const{ return cpmabs(x-y);}
         //: distance

      R& alpha(){ return alpha_;}
      R const& alpha()const{ return alpha_;}
   };

//////////////////////////// class SysLat2 ///////////////////////////////

   class SysLat2{ // quantum system on a two-dimensional lattice
      // The data which define a Hamiltonian are not part of the
      // class data; they enter as an argument of function dot_.
      Lat1 lat1_,lat2_;
         //: lattice
         // lattice in space
      V< V<C> > rep_;
         //: representation
         // list of wave-function values on the
         // lattice positions
       typedef SysLat2 Type;
   public:
      CPM_IO
      CPM_NAM(SysLat2)
      CPM_QM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
      CPM_QM_SCALAR_PRODUCT_INTERFACE(Type,C)
      CPM_QM_PURE_SCALAR_INTERFACE(Type,R)

      SysLat2(Lat1 const& lat1, Lat1 const& lat2, V< V<C> > const& rep):
      lat1_(lat1),lat2_(lat2),rep_(rep)
      {
         Word loc("SysLat2(Lat1,Lat1,V<V<C> >");
         cpmassert(lat1_.d()==rep_.dim(),loc); // lat1_d()>0 always
         cpmassert(rep_.fir().dim()==lat2_.d(),loc);
      }

      SysLat2(SysLat1 const& s1, SysLat1 const& s2);
         // not normalized unless s1 and s2 are

      SysLat1 red1(SysLat1 const& s1)const;
         //: reduce 1
         // reduce first index via s1

      SysLat1 red2(SysLat1 const& s2)const;
         //: reduce 2
         // reduce second index via s2

      Z dim1()const{ return lat1_.d();}
         //: dimension 1

      Z dim2()const{ return lat2_.d();}
         //: dimension 2

      V<C>& cui(Z i){ return rep_.cui(i);}
      V<C> const& cui(Z i)const{ return rep_.cui(i);}

      R nor_();
         //: normalize

      void dot_(DynDat const& dd1, DynDat const& dd2,
         IntActDat const& dd12);
         //: dot
         // changes *this into its time derivative under the
         // assumption that dynamical data dd1, dd2, dd12
         // are in action

      C eTot(DynDat const& dd1, DynDat const& dd2,
         IntActDat const& dd12)const
         {
            SysLat2 v=*this;
            v.dot_(dd1,dd2,dd12);
            return C::I*(*this|v);
         }

      R_Vector stt(DynDat const& dd1, DynDat const& dd2)const;
         //: statistics

      void mark(Graph& g, Z ipolMeth=0, R gamma=0.5, R secFac=1.2)const;
       //: mark
      // showing the wave function in a way that the x-axis refers to
      // particle 1 and the y-axis ( oriented vertically upward!)
      // If the last argument is >0 it triggers an automatic
      // bright-level adjustment only in the first call and then
      // holds this level so that changes in brightness of individual
      // pixels can't occur as the result of a new level adjustment.

      R dtSup(DynDat const& dd1, DynDat const& dd2,
         IntActDat const& dd12, Z p=100, R tWait=0, R gamma=0.1)const;
      //: dt supremum
      // upper limit for the time step according to
      // the new insights from the preparation of the
      // publication. Tries to estimate the largest
      // energy eigenvalue (better: absolute value of this)
      // if tWait>0 it shows the process in graphics,
      // each of the intermediate states will be shown on
      // screen and tWait is the inspection time per screen
      // image.
   };

//////////////////////////// class SysLat3 ///////////////////////////////

   class SysLat3{ // quantum system on a three-dimensional lattice
      // The data which define a Hamiltonian are not part of the
      // class data; they enter as an argument of function dot_.
      Lat1 lat1_,lat2_,lat3_;
         //: lattice
         // lattice in space
      V< V< V<C> > >rep_;
         //: representation
         // list of wave-function values on the lattice positions
       typedef SysLat3 Type;
   public:
      CPM_IO
      CPM_NAM(SysLat3)
      CPM_QM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
      CPM_QM_SCALAR_PRODUCT_INTERFACE(Type,C)
      CPM_QM_PURE_SCALAR_INTERFACE(Type,R)

      SysLat3(SysLat1 const& s1, SysLat1 const& s2, SysLat1 const& s3);
         // not normalized unless s1 and s2 are

      SysLat2 red3(SysLat1 const& s3)const;
         //: reduce 3
         // reduce last index via s3
         // results in a two dimensional system

      SysLat2 red(Z i)const{ return SysLat2(lat2_,lat3_,rep_[i]);}
         // The wave function on the 3-dimensional lattice
         // can be represented by its values on 2-dimensional
         // sub-lattices (slices). Due to the strorage organization
         // of rep_ it is most efficient to let the slices be states
         // of particles 2 and 3 (thus 'slicing particle 1').

      SysLat1 red12(SysLat1 const& s1, SysLat1 const& s2)const;
      SysLat1 red23(SysLat1 const& s2, SysLat1 const& s3)const;
      SysLat1 red31(SysLat1 const& s3, SysLat1 const& s1)const;


      SysLat1 red12(SysLat2 const& s12)const;
      SysLat1 red23(SysLat2 const& s23)const;
      SysLat1 red31(SysLat2 const& s31)const;

      Z dim1()const{ return lat1_.d();}
         //: dimension 1

      Z dim2()const{ return lat2_.d();}
         //: dimension 2

      Z dim3()const{ return lat3_.d();}
         //: dimension 2

      R nor_();
         //: normalize

      void dot_(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat const& dd12);
         //: dot
         // changes *this into its time derivative under the
         // assumption that dynamical data dd1, dd2, dd12
         // are in action

      C eTot(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat const& dd12)const
         {
            SysLat3 v=*this;
            v.dot_(dd1,dd2,dd3,dd12);
            return C::I*(*this|v);
         }

      R_Vector stt(DynDat const& dd1, DynDat const& dd2,
         DynDat const& dd3)const;
         //: statistics
         // returns the expectation values <s1>, <s2>, <s3>,
         // <s1*s2>, <s1*s3>, <s2*s3>, <s1*s2*s3>, where si is
         // the position operator of particle i relative to the
         // equilibrium position of this particle.

      void mark(Graph& g, DisCon const& dc)const;
         //: mark
         // showing the wave function in a way that the x-axis refers to
         // particle 1 and the y-axis ( oriented vertically upward!)
         // If the last argument is >0 it triggers an automatic
         // bright-level adjustment only in the first call and then
         // holds this level so that changes in brightness of individual
         // pixels can't occur as the result of a new level adjustment.

      vector< vector<C> > transform(Graph& g, DisCon const& dc)const;
         // returns a version of rep_ which consists of
         // dim1()+dim2()+dim3() C-arrays which allow to build a
         // trajectory by concatenating the state data that the present
         // function provides. Since fast concatenation is essential
         // in this context, one uses std::vector as the data structure
         // of the return value.

      R dtSup(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat const& dd12, Z p=100, R tWait=0, R gamma=0.1)const;
         //: dt supremum
         // upper limit for the time step according to
         // the new insights from the preparation of the
         // publication. Tries to estimate the largest
         // energy eigenvalue (better: absolute value of this)
         // if tWait>0 it shows the process in graphics,
         // each of the intermediate states will be shown on
         // screen and tWait is the inspection time per screen
         // image.
   };

////////////////////////////// class TrjLat2 /////////////////////////////

   class TrjLat2{ // trajectories of Lat2 systems
      // trajectories of a quantum system on a two-dimensional
      // lattice
   protected:
      DynDat dd1_,dd2_;
      IntActDat dd12_;
      SysLat2 x_;
      SysLat2 v_;

      R t_;
      typedef TrjLat2 Type;
   public:
      CPM_IO
      CPM_NAM(TrjLat2)

      TrjLat2(DynDat const& dd1, DynDat const& dd2, IntActDat dd12,
         SysLat2 const& x, R t=0): dd1_(dd1),dd2_(dd2),dd12_(dd12),x_(x),
         v_(x_),t_(t){ v_.dot_(dd1_,dd2_,dd12_);}

      TrjLat2(DynDat const& dd1, DynDat const& dd2, IntActDat dd12,
         SysLat1 const& x1, SysLat1 const& x2, R t=0): dd1_(dd1),
         dd2_(dd2),dd12_(dd12),x_(x1,x2), v_(x_),t_(t)
      { v_.dot_(dd1_,dd2_,dd12_);}

      TrjLat2(DynDat const& dd1, DynDat const& dd2, IntActDat dd12,
         SysLat2 const& x, SysLat2 const& v, R t=0):
         dd1_(dd1),dd2_(dd2),dd12_(dd12),x_(x),v_(v),t_(t){}
         // version which also allows to set the velocity

      R t()const{ return t_;}
      R n()const{ return x_.abs();}
      C xv()const{ return (x_|v_);}
      C eTot()const{ return C::I*(x_|v_);}
      C hInt()const;
         // irrelevant for dynamics
      SysLat2 x()const{ return x_;}
      SysLat2 v()const{ return v_;}
      DynDat dd1()const{ return dd1_;}
      DynDat dd2()const{ return dd2_;}
      IntActDat dd12()const{ return dd12_;}

      R_Vector stt()const{ return x_.stt(dd1_,dd2_);}

      R dtSup( Z p=100, R tWait=0, R gamma=0.1)const
      { return x_.dtSup(dd1_,dd2_,dd12_,p,tWait,gamma);}

      void stepGamTil_(R dt) // direct midpoint integrator
         //: step gamma tilde
      {
         R tau=dt*0.5;
         x_+=v_*tau;
         SysLat2 acc(x_);
         acc.dot_(dd1_,dd2_,dd12_);
         acc.dot_(dd1_,dd2_,dd12_);
         v_+=acc*dt;
         x_+=v_*tau;
         t_+=dt;
      }

      void stepGam_(R dt)
         //: step gamma
      {
         R tau=dt*0.5;
         x_+=v_*tau;
         SysLat2 vm(x_);
         vm.dot_(dd1_,dd2_,dd12_);
         v_=vm*2-v_;
         x_+=v_*tau;
         t_+=dt;
      }

      virtual void step_(R dt)
      {
         bool dmi=true; // direct midpoint integrator
            // false only for experimentation, TrjLat3
            // always works with direct midpoint method

         if (dmi){
            stepGamTil_(dt);
         }
         else{
            R tau=dt*0.5;
            stepGam_(tau);
            stepGam_(tau);
         }
      }

      virtual void mark(Graph& g, DisCon const& dc)const
      { x_.mark(g,dc.ipolMet,dc.gamma,dc.secFac);}

      V< V<R2> > run_(Graph& g, RunCon& rc, DisCon const& dc);
      //x2vv run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run
         // Since the first two arguments are non-constant references
         // the C+- conformal name would be run_1_2_(...)
   };

////////////////////////////// class TrjLat3 /////////////////////////////

   class TrjLat3{ // trajectories of Lat3 systems
   // trajectories of a quantum system on a
   // three-dimensional lattice
   protected:
      DynDat dd1_,dd2_,dd3_;
      IntActDat dd12_;
      SysLat3 x_;
      SysLat3 v_;

      R t_;
      typedef TrjLat3 Type;
   public:
      CPM_IO
      CPM_NAM(TrjLat3)

      TrjLat3(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat dd12, SysLat3 const& x, R t=0):
         dd1_(dd1),dd2_(dd2),dd3_(dd3),
         dd12_(dd12),x_(x),v_(x_),t_(t){ v_.dot_(dd1_,dd2_,dd3_,dd12_);}

      TrjLat3(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat dd12, SysLat3 const& x, SysLat3 const& v, R t=0):
         dd1_(dd1),dd2_(dd2),dd3_(dd3),dd12_(dd12),x_(x),v_(v),t_(t){}
         // version which also allows to set the velocity

      R t()const{ return t_;}
      R n()const{ return x_.abs();}
      C xv()const{ return (x_|v_);}
      C eTot()const{ return C::I*(x_|v_);}
      C hInt()const;
      SysLat3 x()const{ return x_;}
      SysLat3 v()const{ return v_;}
      DynDat dd1()const{ return dd1_;}
      DynDat dd2()const{ return dd2_;}
      DynDat dd3()const{ return dd3_;}
      IntActDat dd12()const{ return dd12_;}

      R_Vector stt()const{ return x_.stt(dd1_,dd2_,dd3_);}
         //: statistics

      R dtSup( Z p=100, R tWait=0, R gamma=0.1)const
      { return x_.dtSup(dd1_,dd2_,dd3_,dd12_,p,tWait,gamma);}

      virtual void step_(R dt)
         // direct midpoint integrator
      {
         R tau=dt*0.5;
         x_+=v_*tau;
         SysLat3 acc(x_);
         acc.dot_(dd1_,dd2_,dd3_,dd12_);
         acc.dot_(dd1_,dd2_,dd3_,dd12_);
         v_+=acc*dt;
         x_+=v_*tau;
         t_+=dt;
      }

      virtual void mark(Graph& g, DisCon const& dc)const
      { x_.mark(g,dc);}

      x2vv run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run
         // Since the first two arguments are non-constant references
         // the C+- conformal name would be run_1_2_(...)

      x2vv run_(Graph& g, RecordHandler const& rch);
         // testing the idea whether one gets flexability in this way
         // One gets! I should have chosen this approach from the
         // beginning. Nevertheless I make no use of it here for
         // staying compatible with previous results.
         // But for later use: Functions which directly control
         // a program run should get all their data from one argument
         // of type RecordHandler. In this way the implementation of new
         // control modalities does not imply the need for new arguments
         // or the definition of new aggregated data types.
   };

////////////////////////////// class Sep2 /////////////////////////////
// separated dynamics works and satisfies energy conservation law with
// high accuracy: 2007-11-27

   class Sep2{ // quantum system made of two separated subsystems
      DynDat dd1_,dd2_;
      IntActDat dd12_;
      SysLat1 x1_, x2_;
      SysLat1 v1_, v2_;
      R t_;
      R s1_,s2_,s3_;
      typedef Sep2 Type;
   public:
      CPM_IO
      CPM_NAM(Sep2)

      static X2<SysLat1, SysLat1> dot(SysLat1 const& x1,
         SysLat1 const& x2, DynDat const& dd1, DynDat const& dd2,
         IntActDat const& dd12);
      // returns the velocities determined by the arguments

      Sep2(DynDat const& dd1, DynDat const& dd2, IntActDat const& dd12,
         SysLat1 const& x1, SysLat1 const& x2, R t=0):
         dd1_(dd1),dd2_(dd2),dd12_(dd12),
         x1_(x1),x2_(x2),v1_(x1_),v2_(x2_),t_(t)
        // {dot_();}
        { X2<SysLat1,SysLat1> vv=dot(x1_,x2_,dd1_,dd2_,dd12_);
         v1_=vv.c1(); v2_=vv.c2();}

      SysLat1 x1()const{ return x1_;}
      SysLat1 x2()const{ return x2_;}
      SysLat1 v1()const{ return v1_;}
      SysLat1 v2()const{ return v2_;}
      R t()const{ return t_;}
      DynDat dd1()const{ return dd1_;}
      DynDat dd2()const{ return dd2_;}
      IntActDat dd12()const{ return dd12_;}

      static SysLat1 vInt(SysLat1 const& y1, SysLat1 const& y2,
         DynDat const& dyn1, DynDat const& dyn2, IntActDat const& ia12);
         //: v interaction
         // primary interaction descriptor. Changing this function carries
         // over to all other interaction related functions of Sep2 and
         // also
         // further to Sep3

      static R vIntScl(SysLat1 const& y1, SysLat1 const& y2,
         DynDat const& dyn1, DynDat const& dyn2, IntActDat const& ia12)
         //: v interaction scalar
      { return ((y2|vInt(y1,y2,dyn1,dyn2,ia12))/C::Ibar)[1];}

      R n()const{ return x1_.abs()*x2_.abs();}

      C hInt()const;
         //: h interaction (h for hamiltonian)

      C eSepBad()const{ return C::I*((x1_|v1_)+(x2_|v2_));}
      C eSep()const
         // heureka! here was the error: C::I*((x1_|v1_) is not only
         // (x1_| H_1 x_1) but it contains the contribution from the
         // other particle. Correction 2007-11-27
         // eSep should not contain the contribution of the homogeneous
         // field since this is included in hInt();
      {
         SysLat1 v1(x1_),v2(x2_);
         R betaMem=DynDat::beta();
         DynDat::beta()=0;
         v1.dot_(dd1_); v2.dot_(dd2_); // v1!=v1_; v2!=v2_;
         DynDat::beta()=betaMem;
         return C::I*((x1_|v1)+(x2_|v2));
      }
      C eTot()const{ return eSep()+hInt();}
         // this is a (approximately) conserved quantity
      R e0()const{ return eSep()[1];}
      R eInt()const{ return hInt()[1];}
      R_Vector stt()const
          //: statistics
      {
         R z1=x1_.s(dd1_.x0()), z2=x2_.s(dd2_.x0());
         return R_Vector{z1,z2,z1*z2};
      }

      void step_(R dt);
         //: step
         // according to (8) of my paper 'The direct midpoint method
         // as a quantum mechanical integrator II.
         // works since 2007-11-21
      void stepGam_(R dt);
      void stepGamTil_(R dt); // not useful
      void mark(Graph& g, DisCon const& dc)const;
         //: mark
      V< V<R2> > run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run
    };

/////////////////////// class CM3 ////////////////////////////////////////
// classical model 3
// Only those functions are implemented which are needed
// in Sep3::run_
   class CM3{ // classical model of 3 particles
      R x1_,x2_,x3_;
      R v1_,v2_,v3_;
   public:
      CM3(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3):
      x1_(dd1.x0()),x2_(dd2.x0()),x3_(dd3.x0()),
         v1_(0), v2_(0), v3_(0){}
         // construction of particles at rest, at places given by
         // the equilibrium points determined by dd1,dd2,dd3.

      static void step(R dt, R& x1, R& x2, R& x3, R& v1, R& v2, R& v3,
         DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat const& dd12);
         //: step
         // dynamical law as a stateless function

      R x1()const{ return x1_;}
         //: x1
      R x2()const{ return x2_;}
         //: x2
      R x3()const{ return x3_;}
         //: x3

      void step_(R dt, DynDat const& dd1, DynDat const& dd2,
         DynDat const& dd3, IntActDat const& dd12)
         //: step
         // dynamical law as a state-changing member function,
         // direct midpoint method
      {
         step(dt,x1_,x2_,x3_,v1_,v2_,v3_,dd1,dd2,dd3,dd12);
      }
   };

////////////////////////////// class Sep3 /////////////////////////////

   class Sep3{ // quantum system made of three separated subsystems
      // no extension to a trajectory class needed for defining  the
      // function run_(...) and thus to define trajectories of the system.
      DynDat dd1_,dd2_,dd3_;
      IntActDat dd12_;
      SysLat1 x1_, x2_, x3_;
      SysLat1 v1_, v2_, v3_;
      R t_;
      CM3 cs_; // classical state
      typedef Sep3 Type;
   public:
      CPM_IO
      CPM_NAM(Sep3)

      static X3<SysLat1, SysLat1, SysLat1> dot(SysLat1 const& x1,
         SysLat1 const& x2, SysLat1 const& x3,
         DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat const& dd12);
      // returns the velocities determined by the arguments

      Sep3(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat dd12,
         SysLat1 const& x1, SysLat1 const& x2, SysLat1 const& x3,
         R t=0): dd1_(dd1),dd2_(dd2),dd3_(dd3),dd12_(dd12),
         x1_(x1),x2_(x2),x3_(x3),v1_(x1_),v2_(x2_),v3_(x3_),t_(t),
         cs_(dd1_,dd2_,dd3_)
         {
            X3<SysLat1,SysLat1,SysLat1> vv=
               dot(x1_,x2_,x3_,dd1_,dd2_,dd3_,dd12_);
            v1_=vv.c1(); v2_=vv.c2(); v3_=vv.c3();
         }

      TrjLat3 toTrlLat3()const
      //: to TrjLat3
      // usefull for comparison with exact dynamics
      {
         return TrjLat3(dd1_,dd2_,dd3_,dd12_, SysLat3(x1_,x2_,x3_),t_);
      }

      R dtSup( Z p=100, R tWait=0, R gamma=0.1)const
      //: dt supremum
      // Obtains an upper limit for a stable time step based on the
      // corresponding Hamiltonian for which there is a good theory
      // (which for separated dynamics is missing)
      {
         return toTrlLat3().dtSup(p,tWait,gamma);
      }

      SysLat1 x1()const{ return x1_;}
      SysLat1 x2()const{ return x2_;}
      SysLat1 x3()const{ return x3_;}

      SysLat1 v1()const{ return v1_;}
      SysLat1 v2()const{ return v2_;}
      SysLat1 v3()const{ return v3_;}

      R t()const{ return t_;}

      DynDat dd1()const{ return dd1_;}
      DynDat dd2()const{ return dd2_;}
      DynDat dd3()const{ return dd3_;}

      IntActDat dd12()const{ return dd12_;}
      R n()const{ return x1_.abs()*x2_.abs()*x3_.abs();}

      C hInt()const;
         //: h interaction (h for hamiltonian)

      C eSepBad()const{ return C::I*((x1_|v1_)+(x2_|v2_)+(x3_|v3_));}
      C eSep()const
      {
         SysLat1 v1(x1_),v2(x2_),v3(x3_);
         R betaMem=DynDat::beta();
         DynDat::beta()=0;
         v1.dot_(dd1_); v2.dot_(dd2_); v3.dot_(dd3_);
         DynDat::beta()=betaMem;
         return C::I*((x1_|v1)+(x2_|v2)+(x3_|v3));
      }

      C eTot()const{ return eSep()+hInt();}
      R e0()const{ return eSep()[1];}
      R eInt()const{ return hInt()[1];}
      R_Vector stt()const
          //: statistics
          // expectation values of the quantum particle positions
          // relative to the equilibrium points of the particles.
          // No reference to the classical particles.
      {
         R z1=x1_.s(dd1_.x0()), z2=x2_.s(dd2_.x0()), z3=x3_.s(dd3_.x0());
         return R_Vector{z1,z2,z3,z1*z2,z1*z3,z2*z3,z1*z2*z3,z1+z2+z3};
      }
      R_Vector sttCls()const
         //: statistics classical
      {
         R z1=cs_.x1()-dd1_.x0(), z2=cs_.x2()-dd2_.x0(),
         z3=cs_.x3()-dd3_.x0();
         return R_Vector{z1,z2,z3,z1*z2,z1*z3,z2*z3,z1*z2*z3,z1+z2+z3};
      }
      void step_(R dt);
         //: step
         // implemented as a double step of stepGam_
      void stepGam_(R dt);
         //: step gamma
         // integrator from equation (8) of my paper 'The direct midpoint
         // method as a quantum mechanical integrator II.
         // step_ and stepGam_ don't propagate cs_. This reflects the
         // status of cs_ as an experimental device for a first comparison
         // between quantum motion and classical motion.
         // However, function run_(...) does propagate cs_.
      void mark(Graph& g, DisCon const& dc)const;
         //: mark
         // presently the vertical lines indicate the equilibrium position
         // of the particle. It would be useful also to mark the
         // expected position of the quantum particle and the position
         // of the classical particle.
      x2vv run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run
         // The return values are formed only from output values of
         // function stt() and thus don't refer to classical particles.
         // Since the first two arguments are non-constant references
         // the C+- conformal name would be run_1_2_(...)
      x2vv run_(Graph& g, RecordHandler const& rc);
    };

////////////////////////////// class Fun3 /////////////////////////////

   class Fun3{ // fundamental version of Sep3
      // Dynamics derived from TrjLat3 by means of destruction operators.
      // Sep3 is much more efficient.
      // Fun3 makes no use of the velocities and they thus are not among
      // the data members. This lets the method mis the stability of the
      // serious integrators employed in Sep3 and TrjLat3. It is only
      // by enforced re-normalization that the method allows trajectories
      // long enough for a meaningful comparison with the serious methods.
      // quantum system made of three separated subsystems
      // no extension to a trajectory class needed for defining  the
      // function run_(...) and thus to define trajectories of the system.

      DynDat dd1_,dd2_,dd3_;
      IntActDat dd12_;
      SysLat1 x1_, x2_, x3_;
      R t_;
      CM3 cs_; // classical state
         // not essential but convenient in order to let Fun3 replace
         // Sep3 in all places

      typedef Fun3 Type;
      static bool simplify_;
         // method parameter in function step_
      static R epsilon_;
         // method parameter in function step_
   public:
      CPM_IO
      CPM_NAM(Fun3)


      Fun3(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat dd12,
         SysLat1 const& x1, SysLat1 const& x2, SysLat1 const& x3,
         R t=0): dd1_(dd1),dd2_(dd2),dd3_(dd3),dd12_(dd12),
         x1_(x1),x2_(x2),x3_(x3),t_(t),
         cs_(dd1_,dd2_,dd3_){}

      SysLat1 x1()const{ return x1_;}
      SysLat1 x2()const{ return x2_;}
      SysLat1 x3()const{ return x3_;}

      R t()const{ return t_;}

      DynDat dd1()const{ return dd1_;}
      DynDat dd2()const{ return dd2_;}
      DynDat dd3()const{ return dd3_;}

      IntActDat dd12()const{ return dd12_;}
      R n()const{ return x1_.abs()*x2_.abs()*x3_.abs();}


      C eTot()const
      { return SysLat3(x1_,x2_,x3_).eTot(dd1_,dd2_,dd3_,dd12_); }
      C eSep()const
      {
         IntActDat dd12f(dd12_);
         dd12f.free_();
         return SysLat3(x1_,x2_,x3_).eTot(dd1_,dd2_,dd3_,dd12f);
      }
      C hInt()const{ return eTot()-eSep();}
         //: h interaction (h for hamiltonian)

      R e0()const{ return eSep()[1];}
      R eInt()const{ return hInt()[1];}
      R_Vector stt()const
          //: statistics
          // expectation values of the quantum particle positions
          // relative to the equilibrium points of the particles.
          // No reference to the classical particles.
      {
         R z1=x1_.s(dd1_.x0()), z2=x2_.s(dd2_.x0()), z3=x3_.s(dd3_.x0());
         return R_Vector{z1,z2,z3,z1*z2,z1*z3,z2*z3,z1*z2*z3,z1+z2+z3};
      }
      R_Vector sttCls()const
         //: statistics classical
      {
         R z1=cs_.x1()-dd1_.x0(), z2=cs_.x2()-dd2_.x0(),
         z3=cs_.x3()-dd3_.x0();
         return R_Vector{z1,z2,z3,z1*z2,z1*z3,z2*z3,z1*z2*z3,z1+z2+z3};
      }
      void step_(R dt);
         //: step
         // implemented as a double step of stepGam_

      void mark(Graph& g, DisCon const& dc)const;
         //: mark
         // presently the vertical lines indicate the equilibrium position
         // of the particle. It would be useful also to mark the
         // expected position of the quantum particle and the position
         // of the classical particle.
      x2vv run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run
         // The return values are formed only from output values of
         // function stt() and thus don't refer to classical particles.
         // Since the first two arguments are non-constant references
         // the C+- conformal name would be run_1_2_(...)

      static bool& sim(){ return simplify_;}
      static R& eps(){ return epsilon_;}
    };


/////////////////////////// class Ia /////////////////////////////////////

class Ia{ // interaction descriptor for spin one half
   // Interaction descriptor for spin 1/2
      CC2 h1_;
         // one-particle interaction with external field
      CCCC2 h12_;
         // pair interaction
   public:
      Ia(){}

      Ia(R3 const& h, R q, bool full=true):
         h1_(CC2::sig1*h[1]+CC2::sig2*h[2]+CC2::sig3*h[3]),
         h12_(CCCC2::spinSpin(q,full)){}
         // basic constructor. h homogeneous field
         // q coupling constant. For full==false we have
         // only XY-coupling, else rotation invariant coupling.

      Z dim()const{ return 2;}
         //: dimension
      C2 operator()(C2 const& x)const{ return h1_*x;}
      C2 operator()(C2 const& x, C2 const& y)const
      { return h12_(x,y).trc1(x);}

      C avr(C2 const& x, C2 const& y)const
         //: average
      { return (h12_(x,y).trc1(x)|y);}

      CC2 const& h1()const{ return h1_;}
      CCCC2 const& h12()const{ return h12_;}

   };

//////////////////////////// class SpinN ////////////////////////////////

   class SpinN{ // spin lattice quantum system
      Z m_,n_;
      Ia ia_; // contains the coupling constants
      V<C2> vx_;
      V<C2> vv_;
      R t_;
      R eps_;
      R dt_,dt0Inv_;
      Z sc_; // step counter
      R rf_; // reverse factor ( has the same dynamical effect as
         // reversal of dt_ . But running verbally back in time
         // conflicts with my program control, especially when
         // automatic time step adjustment is being used.
      typedef SpinN Type;
      void ini_(){
         C et=eTot();
         cpmdebug(et);
         dt_=cpminv(et.abs())*cpmsqrt(R(m_*n_));
         dt0Inv_=R(1)/dt_;
         t_=0;eps_=0;sc_=0;
      }
      void dot_();

   public:
      CPM_NAM(SpinN)

      SpinN(Z m, Z n, Ia const& ia): m_(m), n_(n),ia_(ia),
         vx_(m_*n_,C2(C(1,1),C(-1,-1))), rf_(1)
            // rf_ needs to be defined when dot_ gets called
      {

         for (Z i=vx_.b();i<=vx_.e();++i) vx_[i]=vx_[i].ran(i);
         dot_();
         ini_();
      }

      SpinN(Z m, V<C2> const& vx, Ia const& ia): m_(m),
         n_(vx.dim()/m_),ia_(ia), vx_(vx), rf_(1)
      {
         Z mL=1;
         Word loc("SpinN(Z,V<C2>,Ia)");
         CPM_MA
         dot_();
         ini_();
         CPM_MZ
      }

      SpinN(Img24 const& im, Ia const& ia):m_(im.h()),
         n_(im.w()),ia_(ia),vx_(m_*n_), rf_(1)
      {
         Z k=1;
         VV<ColRef> vvc=im.toVVColRef();
         for (Z i=1;i<=vvc.dim1();++i){
            for (Z j=1;j<=vvc.dim2();++j){
               vx_[k++]=C2(Color(vvc[i][j]).toR3());
            }
         }
         dot_();
         ini_();
      }

      Z w()const{return n_;}
      Z h()const{return m_;}
      Z hw()const{return m_*n_;}

      Ia const& getIa()const{ return ia_;}
         //: get Ia
      R getEps()const{ return eps_;}
         //: get epsilon
      void setEps_(R eps){ eps_=eps;}
         //: set epsilon
      R get_dt()const{ return dt_;}
         //: get dt
      void set_dt_(R dt){dt_=dt;}
         //: set dt
      void set_t_(R t){t_=t;}
         //: set t
      R t()const{ return t_;}
         //: t
      R dtMean()const{ return t_/sc_;}
         //: dt mean
      void mod_dt_(R fac){dt_*=fac; cpmdebug(dt_);}
         //: modify dt

      Ia ia()const{ return ia_;}

      Z parNum()const{ return vx_.dim();}
         //: particle number

      Z parDim()const{ return 2;}
         //: particle dimension

      Z4 neg(Z i)const;
         //: neighbors

      R n()const
         //: norm
      {
         R r=1;
         for (Z i=vx_.b();i<=vx_.e();++i) r*=vx_.cui(i).abs();
         return r;
      }

      C hInt()const;
         //: h interaction (h for hamiltonian)

      R hInt_i(Z i)const; // only needed for eTot_i
      R eSep_i(Z i)const; // only needed for eTot_i
      R eTot_i(Z i)const; // only needed for temperature,
         // not yet working

      R theta_i(Z i)const{ return vx_[i].blcAng()[1];}
      R phi_i(Z i)const{ return vx_[i].blcAng()[2];}
      R2 blc_i(Z i)const{ return vx_[i].blcAng();}

      R sig3()const
      {
         R res=0;
         for (Z i=vx_.b();i<=vx_.e();++i){
            C2 ci=vx_.cui(i);
            res+=(ci|CC2::sig3*ci)[1];
         }
         return res/(m_*n_);
      }

      R3 sig(Z i)const;
         //: sigma
         // expectation values of the three Pauli matrices
         // at chain position i

      V<C2> col()const;
         //: color

      R2 blcAng()const
         //: Bloch angles
         // returns mean values of Bloch angles over the lattice
      {
         R r1=0, r2=0;
         for (Z i=vx_.b();i<=vx_.e();++i){
            R2 ri=vx_.cui(i).blcAng();
            r1+=ri[1];
            r2+=ri[2];
         }
         R npInv=1./(m_*n_);
         return R2(r1*npInv,r2*npInv);
      }

      C eSep()const;
      C eTot()const{ return eSep()+hInt();}
         //: e total
      R_Vector stt()const{ return R_Vector();}

      void step_();
         //: step
         // implemented as a double step of stepGam_(h)
         // where h=dt_*0.5

      void stepGam_(R dt);
         //: step gamma
         // integrator from equation (8) of my paper 'The direct midpoint
         // method as a quantum mechanical integrator II.
         // Here the argument is active, not dt_

      void rev_()
         //: reversal
      {
         for (Z i=vv_.b();i<=vv_.e();++i) vv_[i]*=-1;
         t_=0;
         rf_*=-1;
      }

      void revDir_(){ dt_*=-1;}
         //: reverse directly

      void mark(Graph& g, DisCon const& dc)const;
         //: mark
         // The connection between C2 and color is given by function

      void run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run

      void runSim_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run simply

      void runRevTest_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run reversion test
         // This is actually the work hourse for the qid publication
     };

/////////////////////////// class Ky /////////////////////////////////////

    class Ky{ // lean class of keys, for forming map<Ky,C>
       std::vector<B> v_;
    public:
       explicit Ky(Z d=0): v_(cpmpos(d),B{false}){}
         //: Ky
       B operator<(Ky const& k)const{ return B{v_ < k.v_};}
         //: <
       void set_(Z i, bool b){ v_[i]=B{b};}
         //: set
       bool get(Z i)const{ return v_[i];}
         //: get
       bool next_(){ return CpmAlgorithms::next_1(v_);}
         //: next
       void ins_(Z i, bool b){ v_.insert(v_.begin()+i,B{b});}
         //: insert
    };

    typedef std::map<Ky,C> Mkc; // data type for wave functions

////////////////////////// struct QrecPar ////////////////////////////////
    struct QregPar{ // parameters needed in TrjQreg
       R h,q;
       bool full;
       QregPar(R hi, R qi, bool fi):h(hi),q(qi),full(fi){}
    };

////////////////////////////// class Qreg ////////////////////////////////

    class Qreg{ // array of qubits
      // q-register, system made out of a finite number of
      // q-bits
      Z n_; // number of q-bits available
      Mkc m_;
      Qreg(Z n, Mkc const& m):n_(n),m_(m){}
      typedef C ScalarType;

    public:
      CPM_QM_SCALAR_PRODUCT_INTERFACE(Qreg,C)
      CPM_NORMALIZE
      CPM_NAM("Qreg")
      explicit Qreg(Z n, Z kind=0);
         //: Qreg
      // Creating the normalized ground state on a lattice of s points
      // (Fock vacuum state) for kind==1 and the zero-vector on the same
      // lattice else

      explicit Qreg(V<C2> const& v);
         //: Qreg

      Z dim()const{ return n_;}
         //: dimension

      Z parNum()const{ return n_;}
         //: particle number

      V<C2> den()const;
         //: density

      V<C2> col()const;
         //: color

      R3 sig(Z i)const;
         //: sigma
         // expectation values of the three Pauli matrices
         // at chain position i

      C get(Ky const& k)const
          //: get
      {
        Mkc::const_iterator i;
        i=m_.find(k);
        return i!=m_.end() ? i->second : C();
      }

      void set_(Ky const& k, C z){ m_[k]=z;}
         //: set
         // notice that mutating functions end in _ by convention

      void add_(Ky const& k, C z){ m_[k]+=z;}
         //: add
      void sub_(Ky const& k, C z){ m_[k]-=z;}
         //: subtract
      void mul_(Ky const& k, R r){ m_[k]*=r;}
         //: multiply
      void mul_(Ky const& k, C z){ m_[k]*=z;}
         //: multiply
      Qreg operator+(Qreg const& q)const;
         //: +
      Qreg& operator+=(Qreg const& q);
         //: +=
      Qreg& operator-=(Qreg const& q);
         //: -=

      Qreg operator*(R r)const;
         //: *
      Qreg& operator*=(R r);
         //: *=

      Qreg& operator*=(C z);
         //: *=

      Qreg app(CC2 const& a, Z i)const;
         //: apply
         // changes *this into the state resulting from it by
         // application of a 'single particle operator' to the
         // subsystem (not subspace!) i

      Qreg app(CCCC2 const& b, Z i, Z j)const;
         //: apply
         // changes *this into the state resulting from it by
         // application of a 'particle pair operator' b
         // on the subsystem {i,j}

      Qreg ham(CC2 const& h0, CCCC2 const& hi)const;
         // : Hamiltonian
         // exact spin chain Hamiltonian
         // returns a Qreg state which results from application of
         // the spin chain Hamiltonian to Qreg state *this.
         // Explanation of parameters ....

      C hSep(CC2 const& h0)const
          //: h separated
       { return (*this|ham(h0,CCCC2()));}

      R dtSup(CC2 const& h0, CCCC2 const& hi, Z p=100, R tWait=-1)const;
         //: dt supremum

      R mag()const
         //: magnetization
         // expectation value of sigma 3
      { return (operator|(ham(CC2::sig3,CCCC2())))[1];}

      R_Vector stt()const;
          //: statistics

      void mark(Graph& g, DisCon const& dc)const;
         //: mark
    };

////////////////////////// class TrjQreg ////////////////////////////////

   class TrjQreg{ // Trajectory of a Qreg system
      QregPar p_;
      CC2 h0_;
      CCCC2 hi_;
      Qreg x_;
      Qreg v_;
      R t_;
    public:
      CPM_NAM(TrjQreg)
      TrjQreg(QregPar const& p, Qreg const& x):p_(p),h0_(CC2::sig3*p_.h),
         hi_(CCCC2::spinSpin(p_.q, p_.full)),x_(x),
         v_(x_.ham(h0_,hi_)),t_(0){v_*=C::Ibar;}
         // version in which the velocity is set from the state
      void step_(R dt);
      R t()const{ return t_;}
      R n()const{ return x_.abs();}
      C eTot()const{ return C::I*(x_|v_);}
      C hInt()const{ return eTot()-eSep();}
      C eSep()const{ return x_.hSep(h0_);}
      R e0()const{ return eSep()[1];}
      R eInt()const{ return hInt()[1];}
      Qreg x()const{ return x_;}
      Qreg v()const{ return v_;}
      R dtSup(Z p=100, R tWait=-1)const
         { return x_.dtSup(h0_,hi_,p,tWait);}
         //: dt supremum
      R_Vector stt()const{ return x_.stt();}
          //: statistics
      V< V<R2> > run_(Graph& g, RunCon& rc, DisCon const& dc);
         //: run
      void mark(Graph& g, DisCon const& dc)const{ x_.mark(g,dc);}
   };

   void testEC(Z d1, Z d2, Z d3);
      // experimental test of a property that should ensure
      // energy conservation. Three particle states

   void testEC4(Z d1, Z d2, Z d3, Z d4);
      // experimental test of a property that should ensure
      // energy conservation. Four-particle states.
      // This should be general enough to be convincing
      // for arbitrary particle number.

} // namespace

#endif
