//? cpmqmx.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


/*
   cpmqmx.cpp
   History: see cpmqmx.h

*/

#include <cpmqmx.h>
#include <cpmvectormore.h>
#include <cpmobservable.h>
#include <cpmdistribution.h>
#include <cpmframes.h>

using CpmQM::RunConx;
using CpmQM::DisConx;

using CpmQM::SpinNx;
using CpmQM::Iax;

using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::VVV;
using CpmArrays::VVVV;
using CpmArrays::R_Vector;
using CpmArrays::DisR;
using CpmLinAlg::CC2;
using CpmLinAlg::CCCC2;
using CpmGraphics::Img24;
using CpmImaging::Color;
using CpmImaging::ColRef;
using CpmImaging::WHITE;
using CpmImaging::BLACK;
using CpmImaging::LIGHTGREEN;
using CpmImaging::LIGHTCYAN;
using CpmImaging::LIGHTBLUE;
using CpmImaging::FAINTYELLOW;

namespace{ // anonymous
   // color related stuff
   const R xg=-0.5;
   const R yg=0.5*sqrt(3.);
   const R2 er(1,0);
   const R2 eg(xg,yg);
   const R2 eb(xg,-yg);
   C transform(R3 const& r){
      R xr=(r[1]+1);
      R xg=(r[2]+1);
      R xb=(r[3]+1);
      R2 res=er*xr+eg*xg+eb*xb;
      return C(res[1],res[2]);
   }
}

void lineMark_1(Graph& g, V<C2> const& vx, DisConx const& dc, Z met)
{
   using CpmImaging::Color;
   static Z jShift=1;
   Z gw=g.w();
   Z gh=g.h();
   Z np=vx.dim();
   Z pix=gh/np;
   Z mult_i=pix;
   Z mult_j=pix;
   Z js=pix,is=pix;
   Z i,j;
   Z d_=np*mult_i;
   jShift+=js;
   Z iState=1;
   Z iCounter=0;
   Z iEff=is;
   Color c;

   for (i=1;i<=d_;i++){
      if (iCounter==0){
         C2 ci=vx.cui(iState);
         if (dc.colMet==1){ // Bloch angles based coloring
            C val=ci.blc();
            c=Color(val);
   // if this is true we have rigid (reflective) boundary conditions,
   // periodic otherwiseal);
         }
         else if (dc.colMet==2){
            C val=ci[1];
            c=Color(val);
         }
         else if (dc.colMet==3){
            C val=ci[2];
            c=Color(val);
         }
         else if (dc.colMet==4){
            C val=ci[1]*ci[2];
            c=Color(val);
         }
         else{ // new natural method
            R r1=((ci|CC2::sig1*ci)[1]+1.)*0.5;
            R r2=((ci|CC2::sig2*ci)[1]+1.)*0.5;
            R r3=((ci|CC2::sig3*ci)[1]+1.)*0.5;
            c=Color(R3(r1,r2,r3));
         }
      }
      iCounter++;
      if (iCounter==mult_i){
         iCounter=0; // triggers new evaluation of c in the next loop
         iState++;
      }
      for (j=0;j<mult_j;j++){ g.put(Z2(iEff,(j+jShift)%gw),c);}
      iEff++;
   }
}

// helper function for runX_
R2 fp(R2 const& x, R const& fac){
   R x1=x[1];
   R x2=x[2];
   return R2(x1,fac*x2);
}

/////////////////////////// class SpinNx  /////////////////////////////////

V<C2> SpinNx::dot() // was a static function until May 1, 2022
{
   Word loc("SpinNx::dot()");
   Z mL=2;
   CPM_MA
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   V<C2> res(np_);
   for (Z i=1;i<=np_;++i){
      C2 ci=vx_.cui(i);
      C2 ri=ia_(ci)+ia_(vx_(i-1,os),ci)+ia_(vx_(i+1,os),ci);
      res.cui(i)=ri*C::Ibar;
   }
   CPM_MZ
   return res;
}

C SpinNx::hInt()const
{
   C res;
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   for (Z i=1;i<=np_;++i){
      C2 ci=vx_.cui(i);
      res+= (ia_.avr(vx_(i-1,os),ci)
             + ia_.avr(vx_(i+1,os),ci));
   }
   return res*0.5_R;
}

C SpinNx::hInt_i(Z i)const
{
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   C2 ci=vx_.cui(i);
   C z=ia_.avr(vx_(i-1,os),ci) + ia_.avr(vx_(i+1,os),ci);
   return z*0.5_R;
}

C SpinNx::test2()const
{
   Word loc("SpinNx::test2()");
   Z mL=2;
   CPM_MA
   C res1=hInt();
   C res2;
   for (Z i=1;i<=np_;++i){
      res2+=hInt_i(i);
   }
   CPM_MZ
   return res1-res2;
}

C SpinNx::eSep_i(Z i)const
{
   C2 ci=vx_.cui(i);
   return (ci|ia_(ci));
}

C SpinNx::eTot_i(Z i)const
{
   return eSep_i(i)+hInt_i(i);
}

C SpinNx::eSep()const
{
   C res;
   for (Z i=1;i<=np_;++i){
      C2 ci=vx_.cui(i);
      res += (ci|ia_(ci));
   }
   return res;
}

R3 SpinNx::sig(Z i)const
{
   C2 ci=vx_.cui(i);
   C v1=(ci|CC2::sig1*ci);
   C v2=(ci|CC2::sig2*ci);
   C v3=(ci|CC2::sig3*ci);
   return R3(v1[1],v2[1],v3[1]);
}

V<C2> SpinNx::col()const
{
   if (np_==0) return V<C2>(0);
   V<C2> res(np_);
   for (Z i=1;i<=np_;++i){
      R3 r3=sig(i);
      C z=transform(r3);
      res[i]=C2(z,z);
   }
   return res;
}

void SpinNx::stepLean_()
// step of the DALF and the ADALF method
{
   Z i;
   R tau=rc_.dt*0.25_R; // stepGam_() will be applied twice
   R tau2=tau*2_R; // stepGam_() will be applied twice

   V<C2> vv1; // needed only if rc_.adalf

   t_+=tau;

   for (i=1;i<=np_;++i) vx_.cui(i)+=vv_.cui(i)*tau;

   V<C2> vv=dot();
   for (i=1;i<=np_;++i) vv_.cui(i)+=(vv.cui(i)-vv_.cui(i))*2_R;

   if (rc_.adalf) vv1=vv;

   for (i=1;i<=np_;++i) vx_.cui(i)+=vv_.cui(i)*tau2;
   t_+=tau2;

   vv=dot();
   for (i=1;i<=np_;++i) vv_.cui(i)+=(vv.cui(i)-vv_.cui(i))*2_R;

   for (i=1;i<=np_;++i) vx_.cui(i)+=vv_.cui(i)*tau;

   if (rc_.adalf){
      for (i=1;i<=np_;++i) vv_.cui(i)=(vv_.cui(i)+vv1.cui(i))*0.5_R;
   }

   renormalize_(); // depends on rc_.eps if this parameter is > 0

   t_+=tau; // now t_ got increased by 4*tau = rc_.dt

}

void SpinNx::step_()
{
   Word loc("SpinNx::step_()");
   Z mL=3;
   CPM_MA
   if (!rc_.autoStep){
         stepLean_();
   }
   else{ // in this case we use automatic time step adaption during run
         // unlike using adj_dt_() in the constructor. Since in this
         // case the norm deviation controls the time step it is essential
         // not to set it to zero via rc_.enfNor.
      rc_.enfNor=B(false);
      const R facRed=0.75;
      const R facInc=1.1;
      const R epsLow=0.75; // not yet working well
      V<C2> xMem=vx_;
      V<C2> vMem=vv_;
      Z iter=0;
label:
      stepLean_();
      // now we test whether the step was sound or whether we have to
      // redo it.
      Z i;
      R sup=0.;
      for (i=vx_.b();i<=vx_.e();++i){ // this is vx_ after stepGam2_
         R ni=vx_.cui(i).abs();
         R si=cpmabs(ni-1.);
         if (si>sup) sup=si;
      }
      if (sup>10*rc_.eps){ // do the step again with smaller dt
         iter++;
         t_-=rc_.dt;
         rc_.dt*=facRed;
         vx_=xMem;
         vv_=vMem;
         renormalize_();
         goto label;
      }
      else if (sup<epsLow*rc_.eps){ rc_.dt*=facInc; }
      else ;
   }
   sc_++;
   CPM_MZ
}

void SpinNx::mark(Graph& g, Z met)const
{
   lineMark_1(g,vx_,dc_,met);
}

V<V<R2>> SpinNx::run_(Graph& g)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinNx::run_(Graph,DisConx)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   R n0=n();
   R n0Inv=cpminv(n0);
   C e0=eTot(); // energy has a small imaginary part in the present context
   C es=eSep(); // energies may have a small imaginary part in the present
      // context. Energy in the external field.
   C eint=hInt(); // contribution of spin-spin-interaction to total
      // energy
   R sig30=sig3();
   R2 blc0=blcAng();
   R theta0=blc0[1];
   R phi0=blc0[2];
   R eTotR=e0[1];
   R eTotAbs=e0.abs();

   cpmvalue("eTotR = ",eTotR,3);
   R eInt=hInt()[1];
   cpmvalues2("eTot,eInt",eTotR,eInt,4);
   R e0Inv=cpminv(e0.abs()); // e0.abs() == 0  does not harm.
   R e01=e0[1]; // real and imaginary part
   R e02=e0[2];
   R es1=es[1]; // real and imaginary part
   R es2=es[2];
   R eint1=eint[1]; // real and imaginary part
   R eint2=eint[2];
   Z nSteps=rc_.nSteps;
   R tTot=R(nSteps)*rc_.dt;
   R tTotInv=1_R/tTot;
   R dt0Inv=cpminv(rc_.dt);
   R tWaitDocuNormError=rc_.tWaitDocuNormError;
   R tWaitDocuEnergies=rc_.tWaitDocuEnergies;
   R tWaitDocuEnergyError=rc_.tWaitDocuEnergyError;
   R tWaitBlochAngles=rc_.tWaitBlochAngles;
   R tWaitSigma3=rc_.tWaitSigma3;
   CpmAlgorithms::PeriodicAlarm view(R(rc_.viewPeriod)*rc_.dt);
   CpmAlgorithms::PeriodicAlarm docu(R(rc_.docuPeriod)*rc_.dt);
   if (rc_.tWaitIni>0){
      mark(g);
      g.setText("         initial state");
      g.vis();
      cpmwait(rc_.tWaitIni,2);
   }
   g.clr_();
   R r0=0_R;
   vector<R> tv(1,r0), nv(1,r0), e1v(1,r0), e2v(1,r0),
      de1v(1,r0), de2v(1,r0),
      es1v(1,r0), es2v(1,r0),eint1v(1,r0), eint2v(1,r0),
      thetav(1,theta0), phiv(1,phi0), sig3v(1,r0);
   bool dtVar=rc_.autoStep;
   R ti=0;
   while (ti<tTot){
      step_();
      ti=t();
      if (dtVar){
         cpmvalue("dtRel",rc_.dt*dt0Inv,3);
      }
      cpmprogress("run",tTotInv*ti);
      if (docu(ti)){
         R ni=n();
         R sig3i=sig3();
         C ei=eTot();
         C esi=eSep();
         C einti=hInt();
         R2 blc=blcAng();
         R thetai=blc[1];
         R phii=blc[2];

         tv.push_back(ti);
            // points in time
         nv.push_back((ni-n0)*n0Inv);
            //

         e1v.push_back(ei[1]);
         e2v.push_back(ei[2]);
         es1v.push_back(esi[1]);
         es2v.push_back(esi[2]);
         eint1v.push_back(einti[1]);
         eint2v.push_back(einti[2]);

         de1v.push_back((ei[1]-e01)*e0Inv);
         de2v.push_back((ei[2]-e02)*e0Inv);

         thetav.push_back(thetai);
         phiv.push_back(phii);
         sig3v.push_back(sig3i-sig30);
      }
      if (view(ti)||ti>=tTot){
         mark(g);
         if (ti>=tTot) g.vis(true); // always conserve the last frame
         else g.vis(dc_.wrtOnShow);
         cpmwait(rc_.tWaitView);
      }
   }
   Z dv=tv.size();
   V<R2> nv2(dv);
   V<R2> e1v2(dv);
   V<R2> e2v2(dv);
   V<R2> es1v2(dv);
   V<R2> es2v2(dv);
   V<R2> eint1v2(dv);
   V<R2> eint2v2(dv);
   V<R2> de1v2(dv);
   V<R2> de2v2(dv);
   V<R2> thetav2(dv);
   V<R2> phiv2(dv);
   V<R2> sig3v2(dv);
   for (Z i=1;i<=dv;++i){
      Z j=i-1;
      R ti=tv[i-1];
      nv2[i]=R2(ti,nv[j]);
      e1v2[i]=R2(ti,e1v[j]);
      e2v2[i]=R2(ti,e2v[j]);
      es1v2[i]=R2(ti,es1v[j]);
      es2v2[i]=R2(ti,es2v[j]);
      eint1v2[i]=R2(ti,eint1v[j]);
      eint2v2[i]=R2(ti,eint2v[j]);
      de1v2[i]=R2(ti,de1v[j]);
      de2v2[i]=R2(ti,de2v[j]);

      thetav2[i]=R2(ti,thetav[j]);
      phiv2[i]=R2(ti,phiv[j]);
      sig3v2[i]=R2(ti,sig3v[j]);
   }

   V<V<R2>> doc1{nv2};
   V<V<R2>> doc2{e1v2,e2v2,es1v2,es2v2,eint1v2,eint2v2};
   V<V<R2>> doc3{thetav2,phiv2};
   V<V<R2>> doc4{sig3v2};
   V<V<R2>> doc5{de1v2,de2v2};

   Z dwl=4; // docu write limit
   bool fileIt=false;
   //Color cbg = FAINTYELLOW;
   Color cbg = LIGHTGREEN;
   if (tWaitDocuNormError>=0){
      Graph g1; // it is neither important nor always useful
      // if diagnostic graphics is of the same size as
      // the evolving image. Hence we use a new frame here.
      g1.clr_();
      fileIt=tWaitDocuNormError>=dwl;
      g1.setColorActive(cbg);
      g1.setText("SpinNx: evolution of norm error");
      g1.mark(doc1);
      g1.vis(fileIt);
      cpmwait(tWaitDocuNormError,2);
   }
   if (tWaitDocuEnergies>=0){
      Graph g2;
      g2.clr_();
      fileIt=tWaitDocuEnergies>=dwl;
      g2.setWithOrigin(0);
      g2.setColorActive(cbg);
      g2.setText("SpinNx: evolution of eTot, eSep and hInt");
      g2.mark(doc2);
      g2.vis(fileIt);
      cpmwait(tWaitDocuEnergies,2);
   }
   if (tWaitBlochAngles>=0){
      Graph g3;
      g3.clr_();
      fileIt=tWaitBlochAngles>=dwl;
      g3.setColorActive(cbg);
      g3.setText("SpinNx: evolution of mean Bloch angles");
      g3.mark(doc3);
      g3.vis(fileIt);
      cpmwait(tWaitBlochAngles,2);
   }
   if (tWaitSigma3>=0){
      Graph g4;
      g4.clr_();
      fileIt=tWaitSigma3>=dwl;
      g4.setColorActive(cbg);
      g4.setText("SpinNx: evolution of sig3");
      g4.mark(doc4);
      g4.vis(fileIt);
      cpmwait(tWaitSigma3,2);
   }
   if (tWaitDocuEnergyError>=0){
      Graph g5;
      g5.clr_();
      fileIt=tWaitDocuEnergyError>=dwl;
      g5.setWithOrigin(0);
      g5.setColorActive(cbg);
      g5.setText("SpinNx: evolution of eTot error");
      g5.mark(doc5);
      g5.vis(fileIt);
      cpmwait(tWaitDocuEnergyError,2);
   }
   CPM_MZ
   return V<V<R2>>{nv2,e1v2,e2v2,thetav2,phiv2,sig3v2};
}

void SpinNx::runX_(Graph& g/*, DisConx const& dc*/)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinN::runX_(Graph,DisConx)");
   Z mL=1;
   CPM_MA
   SpinNx stateMem(*this);
   V<V<R2>> res1=run_(g/*,dc*/);

   RunConx rc2=stateMem.rc_;
   Z fac=2;
   R facInv=cpminv(R(fac));
   rc2.dt*=facInv;
   rc2.nSteps*=fac;
   rc2.docuPeriod*=fac;
   R order=2.;
   R collapsFactor=cpmpow(R(fac),order);
   SpinNx state2(stateMem);
   state2.rc_=rc2;
   g.clr_();
   V<V<R2>> res2=state2.run_(g/*,dc*/);

#ifdef CPM_Fn
   F<R2,R2> ff=F1<R2,R,R2>(collapsFactor)(fp);
#else
   F<R2,R2> ff(bind(fp,_1,collapsFactor));
#endif

   V<R2> no1=res1[1];
   V<R2> no2=res2[1];
   no2=no2(ff);

   V<R2> en1=res1[2];
   V<R2> en2=res2[2];
   en2=en2(ff);

   Graph g1;
   g1.clr_();
   g1.setText("SpinNx: evolution of norm error for dt and "
      &cpm(fac)&" dt");
   g1.mark(V<V<R2>>{no1,no2});
   g1.vis(true);
   cpmwait(rc_.tWaitNorm2,2);

   Graph g2;
   g2.clr_();
   g2.setText("SpinNx: evolution of energy error for dt and "
      &cpm(fac)&" dt");
   g2.mark(V<V<R2>>{en1,en2});
   g2.vis(true);
   cpmwait(rc_.tWaitEnergy2,2);
   CPM_MZ
}

R abs2(V<C2> const& v){
   R sum=0_R;
   for (Z i=v.b();i<=v.e();++i) sum+=v[i].abs2();
   return sum;
}

void SpinNx::adj_dt_()
{
   Word loc("SpinNx::adj_dt_()");
   Z mL=1;
   CPM_MA
   R huge=1.e12_R;
   R fac=cpminv(4_R);
      // 10 instead of 20 was my best guess from previous applications.
      // For the present application the stable timestep seems to
      // be a growing function of the time span tTot. Propably
      // adaptive time stepping is what is needed
   setEps_(rc_.eps); // in most cases eps<0 most useful
   SpinNx s0=*this;
   Graph gr;
   gr.setAutoScale(true);
   R tTot=rc_.dt*rc_.nSteps;
   R tTotInv=cpminv(tTot);
   Z nIni=100_Z;
   R dt=tTot/nIni;
   Z verbMem=cpmverbose;
   cpmverbose=3;
   cpmvalue("dt",get_dt(),2);
   cpmverbose=verbMem;
label:
   *this=s0; // start always from the same initial state
 //  CpmAlgorithms::PeriodicAlarm view(R(rc_.viewPeriod)*dt);
   set_dt_(dt); // give the present value of dt to *this
   verbMem=cpmverbose;
   cpmverbose=3;
   cpmvalue("dt",get_dt(),2);
   cpmverbose=verbMem;
   vector<R> hv;
   while (t_<tTot){
      stepLean_();
      cpmprogress("run",tTotInv*t_,3);
      R a2=abs2(vx_)+abs2(vv_);
      hv.push_back(a2);
      if ( a2 > huge){
         gr.clr_();
         R_Vector hr(hv);
         gr.mark(hr); 
         gr.vis();
         cpmwait(4.,2);
         dt*=0.5_R;
         goto label;
      }
   }
   dt*=fac; // last safety
   *this=s0;
      // has valid vx_ and vv_
   set_dt_(dt);
   verbMem=cpmverbose;
   cpmverbose=3;
   rc_.nSteps=tTot/rc_.dt;
   cpmvalues2("dtFinal, nStepsFinal",rc_.dt,rc_.nSteps,2);
   rc_.dtAdjMethod=0;
   cpmverbose=verbMem;
   CPM_MZ
}

