//? cpmqmx.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_QMX_H_
#define CPM_QMX_H_

//   Description: An attempt to reconstruct the treatment of spin chains
//      as given as an example in my article on separated quantum mechanics
//      from 2008.

#include <cpmv.h>
#include <cpmx.h>
#include <cpmc.h>
#include <cpmgraph.h>
#include <cpmrecordhandler.h>
#include <cpmalgorithms.h>
#include <cpmclinalg.h>
#include <map>

using CpmRoot::B;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::Word;
using CpmRoot::randomR;


using CpmRootX::RecordHandler;

using CpmGeo::Iv;

using CpmArrays::V;
using CpmArrays::Va;
using CpmArrays::X2;
using CpmArrays::X3;
using CpmArrays::T3;
using CpmArrays::T4;
using CpmArrays::R_Vector;
using CpmArrays::Outside;

using CpmFunctions::F;

using CpmGraphics::Graph;

using CpmLinAlg::R2;
using CpmLinAlg::R3;
using CpmLinAlg::C2;
using CpmLinAlg::C4;
using CpmLinAlg::CC2;
using CpmLinAlg::CC4;
using CpmLinAlg::CCCC2;

using std::map;
using std::vector;
using  namespace CpmRoot;

namespace CpmQM{

// Within this namespace we alway use units in which the numerical values
// of hDirac and c are 1. hPlanck thus has value 2*pi (=CpmRoot::Pi2).
// Three structs for compact input from an ini-file

////////////////////////////// class SysConx //////////////////////////////

   struct SysConx{ // sys control
      Z np;
      R h;
      R qXY;
      R qZ;
      R dotFac;
      R fracRim;
      R phi0;
      R theta0;
      R flipProbability;
      B random;

   public:
      SysConx(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(np);
         cpmrhf(h);
         cpmrhf(qXY);
         cpmrhf(qZ);
         cpmrhf(dotFac);
         cpmrhf(fracRim);
         cpmrhf(phi0);
         cpmrhf(theta0);
         cpmrhf(flipProbability);
         cpmrhf(random);
      }
   };

////////////////////////////// class RunConx //////////////////////////////
// A collection of control parameters which makes up a part of the state of
// instances of class SpinNx.
   struct RunConx{ // run control
      R dt;
         // time step
      Z nSteps;
         // Function run_(Graph, DisConx) will run from t=0 till
         // t=dt*nSteps where dt is the value of rc_.dt at the point of
         // invocation of SpinNx::run(Graph,DisConx).
      B twoRuns;
         // If true, we run runX_ instead of run_. This allows to verify
         // the order of the integration method.
      Z dtAdjMethod;
         // control of manipulation that dt is subjected to when inserted
         // into the constructor of class SpinNx. These manipulations
         // determine the value of dt which is active in the first
         // invocation of function stepLean_()-
         // Presently the valid values are 0, 1, 2, else
         // 0: no change
         // 1: multiplication with a factor which accounts for particle
         //    number and total energy.
         // 2: invoce adj_dt_()
      Z nIter;
      R eps; // if this is > 0, automatic time step control is in action
         // and eps controls the sensitivity to norm deviations.
         // The auto step mode does not depend on enfNor and only slightly
         // on adalf. If the state descriptor SpinNx.vx_ gets a complicated
         // structure during evolution, the auto-selected time step may
         // become rather short, so that it may become very time-consuming
         // to do the step number set in advance.
      B autoStep;
      B adalf;
         // sets the ADALF mode instead of the DALF mode of the basic
         // integrator stepLean_()
      B rigid;
      B enfNor;
         // enforce normalization in the basic integrator stepLean_()

      Z viewPeriod;
         // Let n=viewPeriod. The each nth step will be shown on screen.
      Z docuPeriod;
         // Let n=docuPeriod. The each nth step will enter the
         // documentation graphics.
      R tWaitIni;
         // time for inspection of the image made by
         // the first invocation of SpinNx::mark(Graph,DisConx)
      R tWaitView;
         // time for inspection of the image made by
         // SpinNx::mark(Graph,DisConx)
      // tWaitXXX >= 0 activates documentation graphics concerning XXX
      // and gives time for viewing for this time in seconds.
      R tWaitDocuNormError;
      R tWaitDocuEnergies;
      R tWaitDocuEnergyError;
      R tWaitBlochAngles;
      R tWaitSigma3;
      R tWaitNorm2;
      R tWaitEnergy2;

    public:
      RunConx(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(dt);
         cpmrhf(nSteps);
         cpmrhf(twoRuns);
         cpmrhf(dtAdjMethod);
         cpmrhf(nIter);
         cpmrhf(eps);
         cpmrhf(autoStep);
         cpmrhf(adalf);
         cpmrhf(rigid);
         cpmrhf(enfNor);
         cpmrhf(viewPeriod);
         cpmrhf(docuPeriod);
         cpmrhf(tWaitIni);
         cpmrhf(tWaitView);
         cpmrhf(tWaitDocuNormError);
         cpmrhf(tWaitDocuEnergies);
         cpmrhf(tWaitDocuEnergyError);
         cpmrhf(tWaitBlochAngles);
         cpmrhf(tWaitSigma3);
         cpmrhf(tWaitNorm2);
         cpmrhf(tWaitEnergy2);
      }
   };

////////////////////////////// class DisConx //////////////////////////////
   struct DisConx{ // display control
      Z colMet;
      R gamma;
      B wrtOnShow;
         //: write on show
   public:
      DisConx():colMet(1_Z),gamma(1_R),wrtOnShow(0_B){}
      DisConx(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(colMet); // colorizing Method
         cpmrhf(gamma);
         cpmrhf(wrtOnShow);
      }
   };

/////////////////////////// class Iax /////////////////////////////////////

class Iax{ // interaction descriptor for spin one half
   // Interaction descriptor for spin 1/2
      CC2 h1_;
         // interaction with external field
      CCCC2 h12_;
         // pair interaction

   public:
      Iax(R h, R qXY, R qZ): h1_(CC2::sig3*h),
      h12_(CCCC2::spinSpinXYZ(qXY, qZ)){}

      Z dim()const{ return 2;}
         //: dimension
      C2 operator()(C2 const& x)const{ return h1_*x;}
      C2 operator()(C2 const& x, C2 const& y)const
      { return h12_(x,y).trc1(x);}

      C avr(C2 const& x, C2 const& y)const
         //: average
      { return (y|operator()(x,y));}

   };

//////////////////////////// class SpinNx ////////////////////////////////


   class SpinNx{ // spin chain quantum system
      Z np_;
      Iax ia_;     
      V<C2> vx_;
      V<C2> vv_;
      R t_{0.};
      Z sc_{0}; // step counter
      RunConx rc_;
      DisConx dc_;
      typedef SpinNx Type;

   public:
      CPM_NAM(SpinNx)

      V<C2> dot();

      static R nvx(V<C2> const& vx){
         Word loc("SpinNx::nvx(V<C2>)");
         Z mL=3;
         CPM_MA
         R r=1_R;
         for (Z i=vx.b();i<=vx.e();++i){ //r*=vx.cui(i).abs();
            R ri=vx.cui(i).abs();
            if (ri==0.) cpmwarning("ri=0, i="&cpm(i));
            r*=ri;
         }
         CPM_MZ
         return r;
      }

      static V<C2> iniState(SysConx const& sc){
       Word loc("SpinNx::iniState(SysConx const&)");
         Z mL=1;
         CPM_MA
         if (sc.random){
            C z0(1.,1.);
            V<C2> res(sc.np);
            for (Z i=res.b();i<=res.e();++i){
               C2 ci=C2(z0.ran(),z0.ran());
               ci.nor_();
               res.cui(i)=ci;
            }
            CPM_MZ
            return res;
         }
         else{
            V<C2> res(sc.np,C2(sc.theta0,sc.phi0));
            Z dRim=cpmrnd(sc.np * sc.fracRim);
            R flip=sc.flipProbability;
            for (Z i=1;i<=sc.np;++i){
               if (randomR()<flip && i>dRim && i+dRim<=sc.np){
                  res[i].swap_();
               }
            }
            CPM_MZ
            return res;
         }
      }

      SpinNx(V<C2> const& vx, R h, R qXY, R qZ, RunConx const& rc,
         DisConx const& dc):
         np_(vx.dim()), ia_(h,qXY,qZ), vx_(vx), rc_(rc), dc_(dc)
         
      {
         Word loc("SpinNx::SpinNx(...)");
         Z mL=1;
         CPM_MA
         for (Z i=1;i<=np_;++i){
            vx_.cui(i).nor_(); // ensuring that the vx_.cui(i) are
               // unit vectors. Separated quantum dynamics is such that
               // they stay unit vectors (up to discretization errors).
         }
         vv_=dot();
      // prior to the first call of stepLean_() methods for changing
      // rc_.dt may be invoked:
         Z dtAdjMethod=rc_.dtAdjMethod;
         if ( dtAdjMethod==0){
            cpmvalue("dt",rc_.dt,2);
         }
            // use rc_.dt as set in the  argument rc
            // of the constructor. The stability of the runs
            // strongly depends on rc_.enfNor and rc_.adalf
         else if (dtAdjMethod==1){
            R tTot=rc_.dt*rc_.nSteps;
            rc_.dt*=cpmsqrt(np_)*cpminv(eTot().abs());
            rc_.nSteps=tTot*cpminv(rc_.dt);
            cpmvalue("dt",rc_.dt,2);
            rc_.dtAdjMethod=0_Z;
         }
            // should take into account particle number and energy in the
            // way that worked in the image dynamics application.
         else if (dtAdjMethod==2){
            adj_dt_();
            cpmvalue("dt",rc_.dt,2);
         }
            // finds and sets the value of rc_.dt which allows to do
            // rc.nIter steps of size rc.dt (not rc_.dt!) without running
            // into norm explosion.
         else cpmerror("SpinNx","bad value of dtAdjMethod");
         CPM_MZ
      }

      R getEps()const{ return rc_.eps;}
      void setEps_(R eps){ rc_.eps=eps;}

      R get_dt()const{ return rc_.dt;}

      void set_dt_(R dt){ rc_.dt=dt;}

      R t()const{ return t_;}
         //: t

      Iax ia()const{ return ia_;}

      Z parDim()const{ return 2;}
         //: particle dimension

      R n()const
         //: norm
         // Notice that vx_ is the tensor product of the states
         // vx_.cui(i). This implies the product structure of the norm.
      {
         R r=1_R;
         for (Z i=vx_.b();i<=vx_.e();++i){ //r*=vx.cui(i).abs();
            R ri=vx_.cui(i).abs();
            if (ri==0.) cpmwarning("ri=0, i="&cpm(i));
            r*=ri;
         }
         return r;
      }

      C sclPrd(V<C2> const& vx)const
         //: scalar product
      {
         C z=C::one;
         for (Z i=1;i<=np_;++i) z*=(vx_.cui(i)|vx.cui(i));
         return z;
      }

      void renormalize_()
         // rescale the vectors vx_.cui(i) to unit vectors and change the
         // velocities vv_.cui(i) accordingly
      {
         if (rc_.eps < 0_R) return; // do nothing
         for (Z i=1;i<=np_;++i){
            R ni=vx_.cui(i).abs();
            if (cpmabs(ni-1_R) > rc_.eps ){
               // thus for rc_.eps == 0 we enforce full normalization
               R niInv=1./ni;
               vx_.cui(i)*=niInv;
               vv_.cui(i)*=niInv;
            }
         }
      }

      C hInt()const;
         //: h interaction (h for Hamiltonian)
         // More precisely: The expectation value of the particle-particle
         // interaction Hamiltonian in the state vx_

      // Energy contributions of the individual particles to the energy
      // expectation value. For reasons of efficiency this will not be used
      // to built up the total energy.
      C hInt_i(Z i)const;
         // gives the interaction of particle i with its direct neighbors.
         // hInt() = sum from i=1 to i= np_ of hInt_i(i).No factor 0.5 !

      C eSep_i(Z i)const;
         // gives the interaction of particle i with external field
      C eTot_i(Z i)const;
         // total energy of particle i

      C test2()const;
         // if hInt() and hInt_i cooperate correctly the return value
         // differs from 0 by numerical noise.

      // State descriptors of the individual particles.
      R theta_i(Z i)const{ return vx_.cui(i).blcAng()[1];}
      R phi_i(Z i)const{ return vx_.cui(i).blcAng()[2];}
      R2 blc_i(Z i)const{ return vx_.cui(i).blcAng();}

      // mean value of Pauli matrix sigma 3 over the chain
      R sig3()const
      {
         R res=0;
         for (Z i=1;i<=np_;++i){
            C2 ci=vx_.cui(i);
            res+=(ci|CC2::sig3*ci)[1];
         }
         return res/np_;
      }

      R3 sig(Z i)const;
         //: sigma
         // expectation values of the three Pauli matrices
         // at chain position i

      V<C2> col()const;
         //: color

      R2 blcAng()const
         //: Bloch angles
         // returns mean values of Bloch angles over the chain
      {
         R r1=0, r2=0;
         for (Z i=1;i<=np_;++i){
            R2 ri=vx_.cui(i).blcAng();
            r1+=ri[1];
            r2+=ri[2];
         }
         R npInv=1./np_;
         return R2(r1*npInv,r2*npInv);
      }

      C eSep()const;
         // total interaction energy with external field

     C eTot()const{ return eSep()+hInt();}
         // total interaction energy with external field and neighbor
         // particles

      void step_();
         //: step
         // implemented by means of stepLean

      void stepLean_();
         //: lean step

       void mark(Graph& g, Z met=1)const;
         //: mark

      V<V<R2>> run_(Graph& g );
         //: run

      void runX_(Graph& g);
         //: run extended

      void adj_dt_();
         //: adjust dt
         // Assumes that dt_ as set by the constructor. Successively halves
         // the steps till the rc_.nIter steps run fine. For security the
         // step is then halved again to give it the final value.

  }; // class SpinNx end

} // namespace

#endif
