//? cpmr2func.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_R2FUNC_H_
#define CPM_R2FUNC_H_
/*

Description:

   Specialicing the 'function template' Fa<X,Y> to X=RR (RR=C) and Y=R
   'real functions of two variables'. Beyond mere specialization,
   interpolation and visalization is treated

*/

#include <cpmrfunc.h>
#include <cpmc.h>

namespace CpmFunctions{

   using CpmRoot::C;
   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::L;

   typedef C RR;

   // C seems to be the most general and most comfortable description of
   // the Euclidian plane. Notice that angles and rotations can easily be
   // handled via the complex member functions.

//////////////////////////////////// class R2_Func ///////////////////////
class R2_Func;
F<RR, CpmArrays::V<R> > toFRRVR(const CpmArrays::VVV<L>& image,
   Z method=1, const Iv& ivx=Iv(0,1), const Iv& ivy=Iv(0,1));

CpmArrays::V<R2_Func> toVR2_Func(const CpmArrays::VVV<L>& image,
   Z method=1, const Iv& ivx=Iv(0,1), const Iv& ivy=Iv(0,1) );

class R2_Func: public Fa<RR,R>{
//R-valued functions on RR with arithmetics
   // R-valued functions on RR with arithmetics and trivially defined
   // order properties, and construction from discrete data via
   // interpolation.

public:

// carrying over the constructors from Fa<R,R>

   R2_Func(const F<RR,R>& g):Fa<RR,R>(g){}
      // upcast constructor

   R2_Func(const Fa<RR,R>& g):Fa<RR,R>(g){}
      // upcast constructor

   R2_Func(const R2_Func& g):Fa<RR,R>(g){}
      // copy constructor

   explicit R2_Func(const R_Func& g);
      // constructor for a function of two variables
      // that is constant with respect to the second
      // argument

   R2_Func(void):Fa<RR,R>(){}
      // default constructor

   explicit   R2_Func( R (*f)(const RR&)):Fa<RR,R>(f){};
      // constructor from native function pointer

   explicit   R2_Func( R (*f)(RR)):Fa<RR,R>(f){};
      // constructor from native function pointer

   explicit   R2_Func( const R& r):Fa<RR,R>(r){};
      // constructor for constant function

// Construction from function pointers

explicit   R2_Func( R (*f)(const R&, const R& ));

explicit   R2_Func( R (*f)(R,R));

// construction from two R_Functions; meaning is tensor product
   R2_Func(const R_Func& f1, const R_Func& f2);

// construction from the data structures used to define images

explicit R2_Func(const CpmArrays::VV<L>& grayImage,
   Z method=1, const Iv& ivx=Iv(0,1), const Iv& ivy=Iv(0,1) );

//friend F<RR, CpmArrays::V<R> > toFRRVR(const CpmArrays::VVV<L>& image,
//   Z method=1, const Iv& ivx=Iv(0,1), const Iv& ivy=Iv(0,1) );
   // returns a vector-valued function

friend F<RR, CpmArrays::V<R> > toFRRVR(const CpmArrays::VVV<L>& image,
   Z method, const Iv& ivx, const Iv& ivy );

friend CpmArrays::V<R2_Func> toVR2_Func(const CpmArrays::VVV<L>& image,
   Z method, const Iv& ivx, const Iv& ivy );
   // returns a vector of real-valued functions

// construction from real matrices

explicit R2_Func(const CpmArrays::VV<R>& matrix,
   Z method=1, const Iv& ivx=Iv(0,1), const Iv& ivy=Iv(0,1),
   bool periodic=true);
// The resulting function is periodic with the rectangle ivx times ivy
// in RR being the repeating cell for periodic==true. If
// periodic==false, the function is set to 0 outside the rectangle
// ivx times ivy.
// method=0 means pixel-wise representation
// method=1 is a fast simplex interpolation.
// Other values give error.
// The method VV<R> CpmGraphics::Img24::toVVR(R,R,R,Iv) generates a
// suitable real-valued image matrix from an instance of the
// basic C+- image class Img24
// Let matrix be of frame type (m,n). The relation of the function values
// to position is given by the matrix indexing scheme. This means that
// matrix[1][1] is the function value at the upper-left corner,
// matrix[m][n] at the lower-right corner. The matrix values are
// mapped to the grid ivy.centers(m) x ivx.centers(n) in the natural
// manner. So, if the geometry is to remain undistorted (for
// assumed square pixels) one should have e.g. ivx=Iv(0,ar),
// ivy=Iv(0,1) for ar=((R)matrix.dim2())/matrix.dim1().
// If matrix=img.toVV() for some img from Image<>, one has
// ar=img.getAspRat(). For graphically representing the R2_Func on a
// Graph gr, one should put gr.setAspRatHoldHeight(ar) in order to
// see img undistorted.

};

    R2_Func twoDimInt(const CpmArrays::V<R>& ySeq,
        const CpmArrays::V<R_Matrix>& fVal,
        const IpolControl& c=IpolControl());

// We get a function of two real variables by interpolation from a
// discrete set.
// We employ successive one-dimensional interpolation and assume an
// arrangement of the discrete data which allows to do that.
// The assumption is that there exists a finite set of parallel lines
// to the x-axis on which all the points are located for which a function
// value is prescribed. As long as theses lines are parallel, they could
// form any angle with any axis without creating an essentially new
// situation. A completely scattered arrangement of points is something
// considerably different and is not treated here.
// The interpolated function is the return value.
// The arguments of the function describe the discrete interpolation data
// as follows.
// fVal[i][2] is the function value of f at point
// (x,y)=(fVal[i][1],ySeq[i]).
// Thus fVal[i] describes the behaviour of the function on the line, the
// y-coordinate of which is ySeq[i].

   R2_Func indfuncRectangle(const RR& cornerLL, const RR& cornerUR);
      // Creates the geometrical indicator function of a rectangle
      // with lower-left corner cornerLL and upper-right corner cornerUR
      // Function value is negative in the rectangle area.
      // The absolute value of the function value is the distance from
      // the boundary. So the gradient of the function is the
      // outwards-directed normal vector.

   R2_Func indfuncCircle(const RR& center, R radius);
      // Creates the geometrical indicator function of a circle
      // with given center and radius

   } // namespace

#endif

