//? cpmrfunc.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmrfunc.h>
#include <cpmvectormore.h>
#include <cpmgraph.h>
//#if defined(CPM_USE_EIGEN)
//   #include <boost/math/special_functions/bessel.hpp>
//#endif // defined

   using namespace CpmRoot;
   using namespace CpmArrays;
   using namespace CpmSystem;
   using namespace CpmFunctions;
   using namespace CpmGeo;
   using namespace CpmGraphics;
   //using namespace CpmStd;
#ifndef CPM_Fn
   using std::bind;
   using std::placeholders::_1;
#endif

//////////////////////// class RomCon //////////////////////////////

   bool RomCon::prnOn(ostream& str)const
   {
      cpmwat;
      cpmp(epsRel_);
      cpmp(epsAbs_);
      cpmp(met_);
      cpmp(len_);
      cpmp(cntMax_);
      cpmp(robErrEst_);
      return true;
   }

   bool RomCon::scanFrom(istream& str)
   {
      cpms(epsRel_);
      cpms(epsAbs_);
      cpms(met_);
      cpms(len_);
      cpms(cntMax_);
      cpms(robErrEst_);
      return true;
   }

//////////////////////// class R_Func //////////////////////////////////

namespace{ // anonymous

R fBell(const R& x, const Iv& iv, const R& mean, const R& sigma)
{
   R iva=iv.abs();
   R p=(x-iv.inf())*cpminv(iva);
   R y=(p-mean)*cpminv(sigma);
   return cpmexp(-y*y*0.5);
}

R fWall(const R& x, const R4& a)
{
   Word loc("fWall(R,R4)");
   R a1=a.x1;
   R a2=a.x2;
   R a3=a.x3;
   R a4=a.x4;
   cpmassert(a1<a2,loc);
   cpmassert(a2<=a3,loc);
   cpmassert(a3<a4,loc);
   if (x<=a1 || x>=a4) return 0;
   else if (x>=a2 && x<=a3) return 1;
   else if (x>a1){
      R p=(x-a1)/(a2-a1);
      return R_Func::hill3(p);
   }
   else{
      R p=(x-a3)/(a4-a3);
      return R(1)-R_Func::hill3(p);
   }
}

// a class is the most flexible frame for an algorithm
class LinObj:public FncObj<R,R>{
   const R2 x2;
   const R2 y2;

// dependend quantities, to be set in the constructor
   const R dx; // x-difference
   const R m; // slope

public:
   LinObj(R2 x, R2 y): x2(x), y2(y), dx(x2[2]-x2[1]),
      m(dx==0. ? (y2[1]+y2[2])*0.5 : (y2[2]-y2[1])/dx){}
   // the constructor does all pre-calculations, so that
   // the following function executes as fast as
   // possible

   R operator()(const R& x)const // virtual function in base
   {
      return dx==0. ? m : y2[1]+m*(x-x2[1]);
   }
};

} // anon. namespace

R_Func::R_Func(Iv iv, R mean, R sigma)
{
#ifdef CPM_Fn
   *this=F3<R,Iv,R,R,R>(iv,mean,sigma)(fBell);
#else
   *this=F<R,R>(bind(fBell,_1,iv,mean,sigma));
#endif
}

R_Func::R_Func(R4 a)
{
#ifdef CPM_Fn
   *this=F1<R,R4,R>(a)(fWall);
#else
   *this=F<R,R>(bind(fWall,_1,a));
#endif
}

R_Func::R_Func(R2 x, R2 y):Fa<R,R>(new LinObj(x,y)){}
   // most direct implementation of a R_Func constructor
   // in terms of an algorithm (here encapsulated in LinObj)

R_Func::R_Func(Iv x, Iv y):
   Fa<R,R>(new LinObj(R2(x[1],x[2]),R2(y[1],y[2]))){}

/**************** CONSTRUCTORS FROM INTERPOLATION *********************/

namespace{ // anonymus

R fR_FuncIpol
(const R& x0, const V<R>& x, const V<R>& y, const IpolControl& c )
// This is the only place where R_Func is linked to locIpol().
{
   return locIpol(x0,x,y,c);
}

} // anonymus

R_Func::R_Func(const V<R>& xIn, const V<R>& yIn, const IpolControl& cIn)
// this is the constructor on which all other interpolation constructors
// are built
{
   if (cIn.get_monoTest()){
      V<R> x=xIn;
      V<R> y=yIn;
      IpolControl c=cIn;
      prepIpol(x,y,c.get_eps());
#ifdef CPM_Fn
      *this=F3<R,V<R>,V<R>,IpolControl,R>(x,y,c)(fR_FuncIpol);
#else
      *this=F<R,R>(bind(fR_FuncIpol,_1,x,y,c));
#endif
   }
   else{
#ifdef CPM_Fn
      *this=F3<R,V<R>,V<R>,IpolControl,R>(xIn,yIn,cIn)(fR_FuncIpol);
#else
      *this=F<R,R>(bind(fR_FuncIpol,_1,xIn,yIn,cIn));
#endif
   }
}

R_Func::R_Func(const R_Matrix& data, const IpolControl& c)
{ *this=R_Func(data[1],data[2],c);}

R_Func::R_Func(const R_Vector& data, const IpolControl& c)
{
   Z d=data.dim();
   R_Vector xv(d);
   for (Z i=1;i<=d;i++) xv[i]=i;
   *this=R_Func(xv,data,c);
}

R_Func::R_Func(const Word& filename, const IpolControl& c)
{
   Z mL=2;
   Word loc("R_Func(Word,IpolControl)");
   static Word lastFile="not_a_filename";
   static R_Vector xv;
   static R_Vector yv;
   cpmmessage(mL,loc&" started");
   R x,y;
   if (filename!=lastFile ){
      lastFile=filename;
      IFileStream in(filename);
      const Z iMax=2000;
      R_Vector xx(iMax);
      R_Vector yy(iMax);
      Z i=0;
      while ( CpmRoot::read(x,in())==1 && CpmRoot::read(y,in())==1 ){
         i++;
         xx[i]=x;
         yy[i]=y;
         if (cpmverbose>mL)
            cpmcerr<<endl<<"i="<<i<<" x="<<x<<" y="<<y;
      }
      Z d=i;
      if (cpmverbose>=mL)
            cpmcerr<<endl<<d<<" value pairs read from file "<<-filename;
      xv=R_Vector(d);
      yv=R_Vector(d);
      for (i=1;i<=d;i++){
         xv[i]=xx[i];
         yv[i]=yy[i];
      }
   }
   *this=R_Func(xv,yv,c);
   cpmmessage(mL,loc&" done");
}

namespace{ // anonymus

R poly4Func(const R& x, const R& c0, const R& c1, const R& c2,
   const R& c3,const R& c4)
{
   return (((c4*x+c3)*x+c2)*x+c1)*x+c0;
}

} // anonymus

R_Func::R_Func(R a0, R a1, R a2, R a3, R a4)
{
#ifdef CPM_Fn
   F<R,R> res=F5<R,R,R,R,R,R,R>(a0,a1,a2,a3,a4)(poly4Func);
#else
   F<R,R> res(bind(poly4Func,_1,a0,a1,a2,a3,a4));
#endif
   *this=res;
}

R_Func R_Func::polyFunc(const R_Vector& c)
{
#ifdef CPM_Fn
   return F1<R,R_Vector,R>(c)(evalPoly);
#else
   return F<R,R>(bind(evalPoly,_1,c));
#endif
}

namespace{

   R stretchFunc(const R& x, const R& k, const R& x0, const R_Func& f)
   { return f(x0+(x-x0)/k);}

   R shiftFunc(const R& x, const R& s, const R_Func& f)
   { return f(x-s);}

   R symFunc(R const& x, Z const& mode, R_Func const& f)
   {
      if (x>=0 || mode==0) return f(x);
      else{
         if (mode>0) return f(-x);
         else return -f(-x);
      }
   }

   R mirrorFunc(const R& x, const R& x0, const R& s, const R_Func& f)
   { return (x>x0 ? f(x0-s*(x-x0)): f(x));}

   R appendFunc(const R& x, const R_Func& f, const R& x0,
      const R_Func& g)
   { return x<=x0 ? f(x) : g(x);}

   R morphFunc(const R& x, const R_Func& f, const Iv& iv,
      const R_Func& g)
   {
      R p=iv.ramp(x);
      R x1=iv.inf()+g(p)*iv.abs();
      return f(x1);
   }

   class SqueezeObj: public FncObj<R,R>{

      const R_Func f;
      const Iv iv1;
      const Iv iv2;
// dependent
      const R x0;
      const R2 sL;
      const R2 sL_;
      const R2 sR;
      const R2 sR_;
      const R_Func tL;
      const R_Func tR;
   public:
      SqueezeObj(const R_Func& fI, const Iv& iv1I, const Iv& iv2I);
      R operator()(const R& x)const;
   };

   SqueezeObj::SqueezeObj(const R_Func& fI, const Iv& iv1I,
      const Iv& iv2I):
      f(fI),iv1(iv1I),iv2(iv2I),x0(iv2.mean()),sL(iv1[1],iv2[1]),
      sL_(iv1[1],x0),sR(iv2[2],iv1[2]),sR_(x0,iv1[2]),tL(sL,sL_),
      tR(sR,sR_){}

   R SqueezeObj::operator()(const R& x)const
   {
      if (!iv1.contains(x)) return f(x);
      else if (iv2.contains(x)) return f(x0);
      else if (x<x0) return f(tL(x));
      else return f(tR(x));
   }

   R squeezeLeftFunc(const R& x, const Iv& ivHost, const R& d,
      const R_Func& f)
   {
      Word loc("squeezeLeftFunc(...)");
      R x1=ivHost[1];
      R x2=ivHost[2];
      R x21=x2-x1;
      R x1d=x1+d;
      cpmassert(x21>0,loc);
      cpmassert(d<x21,loc);
      if (x<=x1) return f(x);
      else if (x>=x2) return f(x);
      else if (x<=x1d) return f(x1);
      else{
         R beta=x21/(x21-d);
         R xs=x2-beta*(x2-x);
            // requirement: xs linear function of x
            // x==x1+d ==>  xs==x1
            // x==x2   ==>  xs==x2
            // this is easily seen to hold for xs as defined above
         return f(xs);
      }
   }

   R squeezeRightFunc(const R& x, const Iv& ivHost, const R& d,
      const R_Func& f)
   {
      Word loc("squeezeLeftFunc(...)");
      R x1=ivHost[1];
      R x2=ivHost[2];
      R x21=x2-x1;
      R x2d=x2-d;
      cpmassert(x21>0,loc);
      if (d>=x21){
         cpmdebug(d);
         cpmdebug(x21);
         cpmdebug(x1);
         cpmdebug(x2);
         cpmassert(d<x21,loc);
      }
      if (x<=x1) return f(x);
      else if (x>=x2) return f(x);
      else if (x>=x2d) return f(x2);
      else{
         R beta=x21/(x21-d);
         R xs=x1+beta*(x-x1);
            // requirement: xs linear function of x
            // x==x1 ==>  xs==x1
            // x==x2-d   ==>  xs==x2
            // this is easily seen to hold for xs as defined above
         return f(xs);
      }
   }

   R shrinkFunc(const R& x, const Iv& ivHost, const Iv& ivSub,
      const R_Func& f)
   {
      R x1=ivHost[1];
      R x2=ivHost[2];
      if (x<=x1) return f(x);
      else if (x>=x2) return f(x);
      else{
         if (x1==x2) return f(x);
         else{
            R s1=ivSub[1];
            R s2=ivSub[2];
            if (x<=s1) return f(x1);
            else if (x>=s2) return f(x2);
            else{
               if (s1==s2) return f(x);
               else{
                  R_Func tr(R2(s1,s2),R2(x1,x2));
                  return f(tr(x));
               }
            }
         }
      }
   }

} // a.. namespace

R_Func R_Func::squeezeInConstantLeft(const Iv& ivHost, R d)const
{
#ifdef CPM_Fn
   return F3<R,Iv,R,R_Func,R>(ivHost,d,*this)(squeezeLeftFunc);
#else
   return F<R,R>(bind(squeezeLeftFunc,_1,ivHost,d,*this));
#endif
}

R_Func R_Func::squeezeInConstantRight(const Iv& ivHost, R d)const
{
#ifdef CPM_Fn
   return F3<R,Iv,R,R_Func,R>(ivHost,d,*this)(squeezeRightFunc);
#else
   return F<R,R>(bind(squeezeRightFunc,_1,ivHost,d,*this));
#endif
}

R_Func R_Func::shrink(const Iv& ivHost, const Iv& ivSub)const
{
#ifdef CPM_Fn
   return F3<R,Iv,Iv,R_Func,R>(ivHost,ivSub,*this)(shrinkFunc);
#else
   return F<R,R>(bind(shrinkFunc,_1,ivHost,ivSub,*this));
#endif
}

void R_Func::stretch_(R k, R x0)
{
   if (k==0.)
      cpmerror("void R_Func::scale(R k, R x0): k=0 not allowed");
#ifdef CPM_Fn
   *this=F3<R,R,R,R_Func,R>(k,x0,*this)(stretchFunc);
#else
   *this=F<R,R>(bind(stretchFunc,_1,k,x0,*this));
#endif
}

void R_Func::shift_(R s)
{
#ifdef CPM_Fn
   *this=F2<R,R,R_Func,R>(s,*this)(shiftFunc);
#else
   *this=F<R,R>(bind(shiftFunc,_1,s,*this));
#endif
}

R_Func R_Func::shift(R s)const
{
#ifdef CPM_Fn
   return F2<R,R,R_Func,R>(s,*this)(shiftFunc);
#else
   return F<R,R>(bind(shiftFunc,_1,s,*this));
#endif
}

R_Func R_Func::sym(Z mode)const
{
#ifdef CPM_Fn
   return F2<R,Z,R_Func,R>(mode,*this)(symFunc);
#else
   return F<R,R>(bind(symFunc,_1,mode,*this));
#endif
}

R_Func R_Func::mirror(R x0, R s)const
{
#ifdef CPM_Fn
   return F3<R,R,R,R_Func,R>(x0,s,*this)(mirrorFunc);
#else
   return F<R,R>(bind(mirrorFunc,_1,x0,s,*this));
#endif
}

R_Func R_Func::append(const R_Func& f, R x0)const
{
#ifdef CPM_Fn
   return F3<R,R_Func,R,R_Func,R>(*this,x0,f)(appendFunc);
#else
   return F<R,R>(bind(appendFunc,_1,*this,x0,f));
#endif
}

R_Func R_Func::morph(Iv iv, const R_Func& g)const
{
#ifdef CPM_Fn
   return F3<R,R_Func,Iv,R_Func,R>(*this,iv,g)(morphFunc);
#else
   return F<R,R>(bind(morphFunc,_1,*this,iv,g));
#endif
}

R_Func R_Func::squeezeInConstant(const Iv& ivHost, const Iv& ivIn)const
{
   return F<R,R>(new SqueezeObj(*this,ivHost,ivIn));
}

R_Vector R_Func::eval(const R_Vector& rx)const
{
   return R_Vector(apply<R,R>(*this,rx));
//   R_Vector res(rx);
//   for (Z i=rx.b();i<=rx.e();++i) res[i]=(*this)(rx[i]);
//   return res;
}

/************* DIFFERENTIATION **************************/
// slightly modified from my old calculus.cpp

namespace{
R der(R x, const R_Func& f, R h)
/* der stands for derivative. It returns f'(x). The parameter h
is the finite increment used in the numerical formula.
See Bronstein et al p. 768, Tabelle 7.13
I collect the positive terms and the negative ones to have just one
subtraction.
(Hint of Klein and Godunov: Introductionary Computational Physics)
*/
{
   R x_=x-h-h;
   R yp=f(x_); x_+=h;
   R yn=f(x_)*8; x_+=h*2;
   yp+=f(x_)*8; x_+=h;
   yn+=f(x_);
   return (yp-yn)/(h*12);
}
// Although higher derivatives can be obtained by multiple application
// of the derivation operator, (see below) we give direct formulae for
// the second and the third derivative.

R der_2(R x, const R_Func& f, R h)
/* der_2 stands for second derivative. It returns f"(x). The parameter h
is the finite increment used in the numerical formula.
See Bronstein et al p. 768, Tabelle 7.13
*/
{
   R x_=x-h-h;
   R yn=f(x_); x_+=h;
   R yp=f(x_)*16; x_+=h;
   yn+=f(x_)*30; x_+=h;
   yp+=f(x_)*16; x_+=h;
   yn+=f(x_);
   return (yp-yn)/(h*h*12);
}

R der_3(R x, const R_Func& f, R h)
/* der_3 stands for third derivative. It returns the value of the
third derivative of f at position x. The parameter h
is the finite increment used in the numerical formula.
See Bronstein et al p. 768, Tabelle 7.13
*/
{
   R x_=x-h-h;
   R yn=f(x_); x_+=h;
   R yp=f(x_)*2; x_+=h*2;
   yn+=f(x_)*2; x_+=h;
   yp+=f(x_);
   return (yp-yn)/(h*h*h*2);
}

R der_n(Z n, R x, const R_Func& f, R h)
/*
Recursive calculation of higher derivatives. The logic of recursion
is the following: If function der_n is called for a certain value of
n (and x) then the defining block {} is worked through with this
value of n. The instructions are found to calculate this from (n-3) and
so on, till finally only n=0 or n=1 or n=2 or n=3 is needed and found.
All the memory work is done automatically due to the recursivity of the C
language.
*/
{
   R y;
   if (n<0) n=-n; // security
   if (n==0){
      y=f(x);
   }
   else if (n==1){
      y=der(x,f,h);
   }
   else if (n==2){
      y=der_2(x,f,h);
   }
   else if (n==3){
      y=der_3(x,f,h);
   }
   else{
      y=(-der_n(n-3,x-h*2,f,h)+der_n(n-3,x-h,f,h)*2-der_n(n-3,x+h,f,h)*2
         +der_n(n-3,x+h*2,f,h))/(h*h*h*2);
   }
   return y;
}

R fDer(const R& x, const R_Func& f, const Z& n, const R& h)
{
   cpmassert(h!=0.,"fDer(R,R_Func,Z,R)");
   return der_n(n,x,f,h);
}

} // anon. namespace

R_Func R_Func::der(Z n, R h)const
{
#ifdef CPM_Fn
    return F3<R,R_Func,Z,R,R>(*this,n,h)(fDer);
#else
    return F<R,R>(bind(fDer,_1,*this,n,h));
#endif
}

R_Func R_Func::derR(Z n, R h)const

{
   R_Func res=*this;
   for (Z j=1;j<=n;j++) res=res.der(1,h);
   return res;
}

/*************** INTEGRATION **********/
// slightly modified from my old calculus.cpp

V<Word> RomCon::show()const
{
   V<Word> res;
   res<<"The RomCon object has the following data:";
   res<<"epsAbs_="+cpm(epsAbs_);
   res<<"epsRel_="+cpm(epsRel_);
   res<<"met_="+cpm(met_);
   res<<"len_="+cpm(len_);
   res<<"cntMax_="+cpm(cntMax_);
   res<<"robErrEst_="+cpm(robErrEst_);
   return res;
}

namespace{
R simpson(R ai, R bi, const R_Func& f, Z n)
/*   Integration according to Simpsons rule
   Bronsteyn p.763.
   The function f is integrated from a to b, by subdividing
   this intervall in 2*n parts. Returns
   integral from ai to bi of f with the usual sign convention for
   ai > bi
*/
{
   R a,b,sig;
   if (ai>bi){
      a=ai; b=bi; sig=1;
   }
   else if (ai<bi){
      a=bi; b=ai; sig=-1;
   }
   else return 0.;
   const R norm=1./3.;
   R h,h2,x1,x2,y0,y1=0,y2=0;
   Z j;
   if (n<=0) n=16;  // added 93-3-4 for safety
   h=(b-a)/(n*2); h2=h*2;
   if (h==0) return (b-a)*(f(b)+f(a))*0.5; // b-a!=0 and h==0 possible
   x1=a+h; x2=a+h2;
   y0= f(a)-f(b); // having here "-" allows to have a single
           // for loop
   for (j=1;j<=n;j++){
      y1+=f(x1);
      y2+=f(x2);
      x1+=h2; x2+=h2;
   }
   return h*(y0+y1*4+y2*2)*norm*sig;
}

R fIntegral(R const& x, R_Func const& f, R const& x0, R const& h)
{
   cpmassert(h!=0.,"fIntegral(R,R_Func,R,R)");
   Z n=cpmrnd(cpmabs((x-x0)/(h*2)));
   return simpson(x0, x, f, n);
}

} // anon. namespace

R_Func R_Func::antiDer(R x0, R h)const
{
#ifdef CPM_Fn
   return F3<R,R_Func,R,R,R>(*this,x0,h)(fIntegral);
#else
   return F<R,R>(bind(fIntegral,_1,*this,x0,h));
#endif
}

namespace{

class MidPoint{ // transcript of function midpnt of Press et al.
   // The comment there concerning the static variable s makes clear
   // that we actually have a stateful class.
      R s_,a_,b_;
      Z j_;
      R_Func f_;
   public:
      MidPoint(R_Func const& f, R a, R b):s_(0),a_(a),b_(b),j_(0),f_(f){}
      R next(void);
   };

   R MidPoint::next(void)
      // each call gives the result of a refined integral evaluation
   {
      const R i3=1./3;
      j_++;
      if (j_==1){
         s_=(b_-a_)*f_((a_+b_)*0.5);
      }
      else{
         Z j, it=1;
         for (j=1;j<j_-1;++j) it*=3;
         R fac=1./(3*it);
         R del=(b_-a_)*fac;
         R ddel=del+del;
         R x=a_+del*0.5;
         R sf=0;
         for (j=1;j<=it;++j){
            sf+=f_(x);
            x+=ddel;
            sf+=f_(x);
            x+=del;
         }
         sf*=fac;
         s_=(b_-a_)*sf + s_*i3;
      }
      return s_;
   }

   R Romberg(R_Func const& f, R a, R b,
      RomCon const& ic=RomCon())
   {
      Z mL=2;
      Word loc("Romberg(R_Func,R,R,RomCon)");
      CPM_MA
      const R i9=1./9.;

      Z jMax=ic.cntMax();
      Z jK=ic.len();
      Z met=ic.met();
      R epsRel=ic.epsRel();
      R epsAbs=ic.epsAbs();
      bool robErrEst=ic.robErrEst();

      R_Vector s(jK);
      R_Vector h(jK);
      Z j;
      MidPoint mp(f,a,b);
      R hNext=1;
      R y=mp.next();
      h[1]=hNext; s[1]=y;
      R yOld=y,yNew=y;
      for (j=s.b();j<s.e();++j){
         hNext*=i9;
         h[j+1]=hNext;
         s[j+1]=mp.next();
      }
      R err=0;
      R h_ex=0;
      for (j=1;j<=jMax;++j){
         if (met==1){
            y=polyInte_4(h,s,h_ex,err);
         }
         else if (met==2){
            y=ratiInte_4(h,s,h_ex,err);
         }
         if (met==0){
            y=mp.next();
         }
         if (robErrEst || met==0){
            yOld=yNew;
            yNew=y;
            err=yNew-yOld;
         }
         R tol=CpmRootX::sup<R>(epsAbs,epsRel*cpmabs(y));
         if (cpmabs(err)<=tol){
            return y;
         }
         if (met==1 || met==2){
            hNext*=i9;
            h.in_(hNext,true);
            y=mp.next();
            s.in_(y,true);
         }
      }
      cpmwarning(loc&": too many iterations, last y returned");
      CPM_MZ
      return y;
   }
}

R R_Func::romInt(R x1, R x2, RomCon const& rc)const
{
   R a,b,sig;
   if (x1<x2){
      a=x1;
      b=x2;
      sig=1;
   }
   else if (x2<x1){
      a=x2;
      b=x1;
      sig=-1;
   }
   else{
      return 0.;
   }
   return Romberg(*this,a,b,rc)*sig;
}

R tenPrdFunc(R2 const& xy, R_Func const& fx, R_Func const& fy)
{ return fx(xy.x1)*fy(xy.x2);}

F<R2,R> R_Func::tenPrd(R_Func const& f)const
      //: tensor product
{
#ifdef CPM_Fn
   return F2<R2,R_Func,R_Func,R>(*this,f)(tenPrdFunc);
#else
   return F<R2,R>(bind(tenPrdFunc,_1,*this,f));
#endif
}

///////////////////// special functions /////////////////////////////////

R CpmFunctions::hillFunc(const R& x, const R& p)
{
   if (x<=0) return(1.);
   if (x>=1) return (0.);
   R y=p*(x-0.5)*2/(x*(x-1.));
   if (y<=0) {
      y=cpmexp(y); y/=(y+1);
   }
   else {
      y=cpmexp(-y); y=R(1)/(y+1);
   }
   // so the argument of cpmexp is always negativ
   // hence no overflow errors can result by calling cpmexp.
   // Since the values of cpmexp are never negative, 1+y is never
   // zero.
   return y;
}

namespace{

R terraceFunc(const R& x, const R& k)
// Function proposed by Thomas Dera (02-1-18), slightly simlified
// via Mathematica (tderaramp.nb)
{
   if (x<=0) return 0;
   else if (x>=1) return 1;
   else{
      if (k==1.) return x;
      else if (k==0.) return 0.5;
      else{
         R k1=cpmacos(cpmsqrt(k));
         R k2=cpmsqrt(R(1)/k -1.);
         return (cpmtan((x*2-1)*k1)/k2 + 1)*0.5;
      }
   }
}

R terraceFunc2(const R& x, const R& k)
{
   if (x<=0) return 0;
   else if (x>=1) return 1;
   else{
      if (k==1.) return x;
      else if (k==0.)
         return (R(1)+R(2)*cpmasin(x*2-1.)/cpmpi)*0.5;
      else{
         R k1=cpmsqrt(-k*k+1);
         R k2=cpmasin(k1);
         return (R(1)+cpmasin((x*2-1)*k1)/k2)*0.5;
      }
   }
}

R rampFunc(const R& x, const R& p){ return R(1)-hillFunc(x,p);}

R hillFunc3(const R& x)
{
   if (x<=0) return 1;
   else if (x>=1) return 0;
   else{
      R x_=R(1)-x;
      return x_*x_*(x*2+1);
   }
}

R rampFunc3(const R& x){ return R(1)-hillFunc3(x);}

R idZero(const R& x){ x; return R{0.};}
R id0(const R& x){ x; return R{1.};}
R id(const R& x){ return x;}
R id2(const R& x){ return x*x;}
R id3(const R& x){ return x*x*x;}
R id4(const R& x){ R y=x*x; y*=y; return y;}
R id5(const R& x){ R y=x*x; y*=y; return y*x;}
R id6(const R& x){ R y=x*x*x; y*=y; return y;}
R id7(const R& x){ R y=x*x*x; y*=y; return y*x;}
R id8(const R& x){ R y=x*x; y*=y; y*=y; return y;}
R id9(const R& x){ R y=x*x*x; R z=y; y*=y; return y*z;}
R id10(const R& x){ R y=x*x*x; R z=y; y*=y; return y*z*x;}
R id11(const R& x){ R y=x*x*x; R z=y; y*=y; return y*z*x*x;}
R id12(const R& x){ R y=x*x*x; y*=y; y*=y; return y;}
R absFunc(const R& x){return cpmabs(x); }
R cutNegFunc(const R& x){ return x>0_R ? x : 0_R; }
R log10Func(const R& x){ return cpmlog10(x);}

} //anon. namespace

R_Func R_Func::hill(R p)
{
#ifdef CPM_Fn
   return F1<R,R,R>(p)(hillFunc);
#else
   return F<R,R>(bind(hillFunc,_1,p));
#endif
}

R_Func R_Func::ramp(R p)
{
#ifdef CPM_Fn
   return F1<R,R,R>(p)(rampFunc);
#else
   return F<R,R>(bind(rampFunc,_1,p));
#endif
}

R_Func R_Func::hill3=R_Func(hillFunc3);
R_Func R_Func::ramp3=R_Func(rampFunc3);

R_Func R_Func::terrace(R k)
{
#ifdef CPM_Fn
   return F1<R,R,R>(k)(terraceFunc);
#else
   return F<R,R>(bind(terraceFunc,_1,k));
#endif
}

R_Func R_Func::terrace2(R k)
{
#ifdef CPM_Fn
   return F1<R,R,R>(k)(terraceFunc2);
#else
   return F<R,R>(bind(terraceFunc2,_1,k));
#endif
}

namespace{

   R2 ftBD(R p, R k, R m, Z z)
      // see Thomas Dera: ramp(TD).nb 2002-2-5.
      // Driving smoothly from one position to another
   {
      R mm=m-1;
      R km=k-1;
      R y0=mm/km;
      R x0=(y0*k)/m;
      if (z==-1){
         // freedom at the left end
         R b21=(R(1)-p)*2;
         R b22=p*p;
         R x=R(1)-x0;
         R y=R(1)-y0;
         return R2(x,y)*b21+R2(1,1)*b22;
      }
      else if (z==1){
         // freedom at the right end
         R b21=(R(1)-p)*2;
         R b22=p*p;
         R x=x0;
         R y=y0;
         R2 res=R2(x,y)*b21+R2(1,1)*b22;
         return res;
      }
      else{
         R pp=p*p; // for efficiency
         R b31=(p-1)*(p-1)*p*3; // Bernstein polynomials b30
            // not appearing since first point is (0,0)
         R b32=(p-1)*pp*(-3);
         R b33=pp*p;
         R x=x0;
         R y=y0;
         R x1=R(1)-x0;
         R y1=R(1)-y0;
         return R2(x,y)*b31+R2(x1,y1)*b32+R2(1,1)*b33;
      }
   }
}

R_Func R_Func::terraceBezierDera(R k, R m, bool inv, Z z, Z n,
   const IpolControl& ic)
{
   if (n<2) cpmerror("R_Func::terraceBezierDera: n<2");
   if (m>1) cpmerror("R_Func::terraceBezierDera: m>1");
   if (k>m) cpmerror("R_Func::terraceBezierDera: k>m");
   if (k<0) cpmerror("R_Func::terraceBezierDera: k<0");
   if (k==1 && m==1) return R_Func::x1;
   V<R> x(n);
   V<R> y(n);
   R dp=1./(n-1.);
   R p=0;
   for (Z i=1;i<=n;i++){
      R2 f=ftBD(p,k,m,z);
      x[i]=f.x1;
      y[i]=f.x2;
      p+=dp;
   }
   if (!inv) return R_Func(x,y,ic); else return R_Func(y,x,ic);
}

namespace{

class BerStn: public FncObj<R,R>{
   R_Vector y;
   Z n{0};
   R_Vector by;
   R2 r2;
   R dxInv{0.};
      // a lot of auxiliar quantities that will be initialized
      // via a constructor. These allow to compute the
      // result of operator() very fast
public:
   BerStn(const R_Vector& y, R2 a);
   R operator()(const R&x)const;
};

BerStn::BerStn(const R_Vector& y_, R2 r2_):y(y_),n(y.dim()),
by(binomialCoefficients(n-1)),r2(r2_)
{
   Word loc("BerStn::BerStn(...)");
   if (n>1){
      for (Z i=1;i<=n;i++) by[i]*=y[i];
      R dx=r2.x2-r2.x1;
      cpmassert(dx!=0,loc);
      dxInv=R(1)/dx;
   }
}

R BerStn::operator()(const R&x)const
{
   if (n==0) return 0.;
   else if (n==1) return y[1];
   else{
      R u=(x-r2.x1)*R(dxInv);
      R_Vector u1(n);
      R_Vector u2(n);
      R res=0;
      R v=R(1)-u;
      u1[1]=1;
      u2[n]=1;
      Z i=2,j=n-1;
      while (i<=n){
         u1[i]=u1[i-1]*u;
         u2[j]=u2[j+1]*v;
         i++;
         j--;
      }
      for (i=1;i<=n;i++) res+=by[i]*u1[i]*u2[i];
      return res;
   }
}

} // anon. namespace

R_Func R_Func::Bernstein(const R_Vector& y, R2 a)
// most advanced implementation of a generating mechanism for R_Func's
// Clean separation between precomputations and evaluation
{
   return F<R,R>(new BerStn(y,a));
}

R_Func R_Func::Bezier(R_Vector const& x, R_Vector const& y, Z n)
// pattern taken from the proven function R_Func::terraceBezierDera
{
   Z mL=3;
   Word loc("R_Func::Bezier(...)");
   CPM_MA
   R2 iv(0,1);
   R_Func fx=R_Func::Bernstein(x,iv);
   R_Func fy=R_Func::Bernstein(y,iv);
   V<R> x_(n);
   V<R> y_(n);
   R dp=1./(n-1.);
   R p=0;
   for (Z i=1;i<=n;i++){
      x_[i]=fx(p);
      y_[i]=fy(p);
      p+=dp;
   }
   IpolControl ic("polygon");
      // important to have a monotonity preserving interpolation
      // method here
   ic.set_monoTest(true);
      // important here since Bezier curves are not necessarily
      // function graphs
   CPM_MZ
   return R_Func(x_,y_,ic);
}

namespace{
   R gaussFunc(const R& x, const R& x0, const R& sigmaInv,
      bool const& unitHeight)
   {
      static R fac=R(1)/cpmsqrt(cpmpi*2);
      R xr=(x-x0)*sigmaInv;
      R earg=-xr*xr*0.5;
      return unitHeight ? cpmexp(earg) : fac*sigmaInv*cpmexp(earg);
   }

// The formulas defining bellCompSupport are taken from Chapter 10,
// Theorem 2.1 of 'Advanced Calculus' by Loomis and Sternberg. There they
// are credited to Bonic and Frampton (1966). The constants 1.25 and 4.
// in the present implementations are selected by me in order to let
// the resulting curve be very similar to a true Gaussian. This was not
// an aim declared in the Loomis/Sternberg book.

   R wall(R x){ return x > R(0.) ? cpmexp(R(-1.25)/x) : R(0.);}
   R peakLean(R x, R xL, R xU){ return wall(xU-x)*wall(x-xL);}
   R peak(R x, R xL, R xU)
   {
      R xM=(xL+xU)*R(0.5);
      return peakLean(x,xL,xU)/peakLean(xM,xL,xU);
   }

   R bellCompSupport(R const& x, R const& x0, R const& sigma)
   {
      R sigEff=sigma*R(4.);
      R xL = x0 - sigEff;
      R xU = x0 + sigEff;
      return peak(x,xL,xU);
   }

   R charFuncAux(R const& x, Iv const& iv, R const& soft)
   {
      if (!iv.contains(x)) return R(0);
      if (soft==0) return R(1);// thus for soft==0
            // exactly the old code
      R p=iv.linear(x); // p in [-1,+1]
      if (p==-1 || p==1) return R(0);
      R q=(p+1)*(p-1); // < 0, will be squared
      R r=R(1)/(q*q);
      return cpmexp(soft*soft*(R(1)-r));
   }

   R sigFunc(R const& x)
   {
      if (x>0.) return R(1.);
      else if (x<0.) return R(-1.);
      else return R(0.);
   }

   R thetaFunc(R const& x)
   {
      if (x>=0.) return R(1.);
      else return R(0.);
   }

   R thetacFunc(R const& x)
   {
      if (x>=0.) return R(0.);
      else return R(1.);
   }

   R harOscFunc(R const& x, R const& omega, R const& delta, R const& damp)
   {
      R res=cpmsin(x*omega+delta);
      return (damp==0. ? res : res*cpmexp(-x*damp));
   }

   R transitFunc(R const& s, R2 const& x, R2 const& y, bool const& linear)
    // corrected 2015-01-09
   {
      R dx=x.x2-x.x1;
// first returning values for the cases where s is not
// in a well-defined transition interval
      if (dx>0){ // x[1] < x[2]
         if (s<=x.x1) return y.x1;
         if (s>=x.x2) return y.x2;
      }
      if (dx<0){ // x[2] < x[1]
         if (s<=x.x2) return y.x2;
         if (s>=x.x1) return y.x1;
      }
      R yc=(y.x1+y.x2)*0.5;
      if (dx==0) return yc;
         // notice that in this case there is no order among
         // x.x1 and x.x2 and so one does not know on which side of
         // x.x1 to use y[1] and y[2]
// now s is in the transition interval
      R xc=(x.x1+x.x2)*0.5, dy=(y.x2-y.x1);
      R p=(s-xc)/dx; // p varies from -1/2 ro +1/2
     // if (!linear) p=p*p*p;
     return linear ? yc+dy*p : yc+dy*p*(1.5-2.*p*p);
   }

   R smoothTransitFunc(R const& s, R2 const& x, R2 const& y, R const& p)
   {
      R x1=x.x1, x2=x.x2;
      R y1=y.x1, y2=y.x2;
      cpmassert(x1!=x2,"smoothTransitFunc(R,R2,R2,R)");
      DiIv div(x1,x2);
      R sp=div.p(s);
      if (sp<=0) return y1;
      else if (sp>=1) return y2;
      else return y2+hillFunc(sp,p)*(y1-y2);
    }

   R dTransitFunc(const R& s, const R2& x, const R2& y)
      // corrected 2015-01-09
   {
      R dx=x.x2-x.x1;
// first returning values for the cases where s is not
// in a well-defined transition interval
      if (dx>0){ // x[1] < x[2]
         if (s<=x.x1) return 0;
         if (s>=x.x2) return 0;
      }
      else if (dx<0){ // x[2] < x[1]
         if (s<=x.x2) return 0;
         if (s>=x.x1) return 0;
      }
      else return 0;
      R xc=(x.x1+x.x2)*0.5;
      R dy=y.x2-y.x1;
      R p=(s-xc)/dx; // p varies from -1/2 ro +1/2
      return (dy/dx)*(1.5-p*p*6.);
   }

   R_Func zeroToSingular(R z, R s);
      // A function f which has satisfies f(z)=f'(z)=0 and has a
      // singularity (z-s)/(x-s) at x=s. Actually it is a function
      // of the variable y = (x-s)/(z-s) and as such of the form
      // 1/y - 3y + 2y*y

   R zeroToSingularFunc(R const& x, R const& z, R const& s );
   R dZeroToSingularFunc(R const& x, R const& z, R const& s );

   R potUMfunc(const R& r, const R& U0, const R& r0,
      const R& facU, const R& facR)
   {
      R U1=facU*U0;
      R U2=-U0;
      R r1=0;
      R r2=r0;
      if (r<=0) return U1;
      if (r<=r0) return R_Func::transit(R2(r1,r2),R2(U1,U2))(r);
      R r3=facR*r0;
      if (r<r3) return R_Func::transit(R2(r2,r3),R2(U2,0))(r);
      return 0;
   }

   R dPotUMfunc(const R& r, const R& U0, const R& r0,
      const R& facU, const R& facR)
   {
      R U1=facU*U0;
      R U2=-U0;
      R r1=0;
      R r2=r0;
      if (r<=0) return 0;
      if (r<=r0) return R_Func::dTransit(R2(r1,r2),R2(U1,U2))(r);
      R r3=facR*r0;
      if (r<r3) return R_Func::dTransit(R2(r2,r3),R2(U2,0))(r);
      return 0;
   }

   R potUM2func(const R& r, const R& U0, const R& r0,
      const R& facU, const R& facR)
   {
      R U2=-U0;
     // R r1=0;
      R r2=r0;
      if (r<=0) return 0.;
      if (r<=r0) return facU*zeroToSingularFunc(-r,-r0,0.)+U2;
      R r3=facR*r0;
      if (r<r3) return R_Func::transit(R2(r2,r3),R2(U2,0))(r);
      return 0;
   }

   R dPotUM2func(const R& r, const R& U0, const R& r0,
      const R& facU, const R& facR)
   {
      R U2=-U0;
    //  R r1=0;
      R r2=r0;
      if (r<=0) return 0;
      if (r<=r0) return -facU*dZeroToSingularFunc(-r,-r0,0.);
      R r3=facR*r0;
      if (r<r3) return R_Func::dTransit(R2(r2,r3),R2(U2,0))(r);
      return 0;
   }

   R concatFunc(const R& x, const V<Iv>& v, const V<R_Func>& f)
      // naive implementation which tries not to conect the
      // intervals and values in a natural way
   {
      Z n=v.dim();
      for (Z i=1;i<=n;i++){
         if (v[i].contains(x)) return (f[i])(x);
      }
      return 0;
   }

   R potCoulombFunc(const R& r, const R& gamma, const R& r0)
   {
      return gamma/(r+r0);
   }

   R dPotCoulombFunc(const R& r, const R& gamma, const R& r0)
   {
      R reff=r+r0;
      return -gamma/(reff*reff);
   }

   R potMorseFunc(const R& r, const R& U0, const R& r0, const R& a)
   {
      const R xMin=-100;
      if (a<=0 || r<0) return 0;
      R x=(r0-r)/a;
      if (x<xMin) return 0;
      return U0*(cpmexp(x*2)-cpmexp(x)*2);
   }

   R dPotMorseFunc(const R& r, const R& U0, const R& r0, const R& a)
   {
      const R xMin=-100;
      if (a<=0 || r<0) return 0;
      R x=(r0-r)/a;
      if (x<xMin) return 0;
      return (U0*2/a)*(cpmexp(x)-cpmexp(x*2));
   }

   R potLJfunc(const R& r, const R& U0, const R& r0)
   {
      if (r<=0) return 0;
      R x=r0/r;
      R y=cpmpow(x,6);
      return U0*y*(y-1)*4;
   }

   R dPotLJfunc(const R& r, const R& U0, const R& r0)
   {
      if (r<=0) return 0;
      R x=r0/r;
      R y=cpmpow(x,6);
      return (U0*24/r)*y*(R(1)-y*2);
   }

   R potPolFunc(R const& r, R const& rr)
   {
      if (r>=rr) return R(0.);
      R y=cpmabs(r/rr);
      R y1=y-1.;
      R c1=R(432.);
      R c2=R(864.);
      R c3=R(5.);
      R c4=R(-1296.);
      R c5=R(655.);
      R c6=R(108.);
      return y1*y1*(c1+y*(c2+c3*y*(c4+c5*y)))/c6;
   }

   R dPotPolFunc(R const& r, R const& rr)
   {
      if (r>=rr) return R(0.);
      R y=cpmabs(r/rr);
      R y1=y-1.;
      R c1=R(-3.);
      R c2=R(5.);
      R c3=R(-5184.);
      R c4=R(3275.);
      R c5=R(108.);
      return y1*y*(c1+c2*y)*(c3+c4*y)/(c5*rr);
   }

   R potSimFunc(R const& r, R const& rr)
   {
      if (r>=rr) return 0.;
      R y=cpmabs(r/rr);
      return 1.+(2.*y-3.)*y*y;
   }

   R dPotSimFunc(R const& r, R const& rr)
   {
      if (r>=rr) return 0.;
      R y=cpmabs(r/rr);
      return 6.*y*(y-1.)/rr;
   }

   R potWallFunc(R const& r, R const& rr, R const& rWall, R const& deg)
   {
      if (r<=rWall-rr*2.) return R(0.);
      R x=(r+rr*2.-rWall)/rr;
      return cpmpow(x,deg);
   }

   R dPotWallFunc(R const& r, R const& rr, R const& rWall, R const& deg)
   {
      if (r<=rWall-rr*2.) return R(0.);
      R x=(r+rr*2.-rWall)/rr;
      return cpmpow(x,deg-1.)*(deg/rr);
   }

   R potConfineFunc(R const& x, Iv const& iv, R const& d, R const& deg)
   {
      if (iv.contains(x,false,false)) return 0.; // closed interval
      return cpmpow(iv(x)*cpminv(d),deg);
   }

   R dPotConfineFunc(R const& x, Iv const& iv, R const& d, R const& deg)
   {
     // cout<<"dPotConfineFunc called"<<endl;
      R xL=iv.inf();
      R xU=iv.sup();
      R res=0;
      if (x > xU || x < xL) {
         R di = cpminv(d);
         R xr;
         if (x > xU){
            xr = (x-xU)*di; // positive
            res= -deg*di*cpmpow(xr,deg-1.); // negative
         }
         else if (x < xL ){
            xr = (xL-x)*di; // positive
            res= deg*di*cpmpow(xr,deg-1.); // positive
         }
         else cpmerror("should not happen");
      }
    //  cout<<"xL ="<<xL<<", xU = "<<xU<<", x = "<<x<<" res = "<<res<<endl;
      return res;
   }
}

R_Func R_Func::Gauss(R x0, R sigma, bool unitHeight)
      // Gauss bell curve with mean x0 and standard deviation sigma;
{
   R sigmaInv=cpminv(sigma);
#ifdef CPM_Fn
   return F3<R,R,R,bool,R>(x0,sigmaInv,unitHeight)(gaussFunc);
#else
   return F<R,R>(bind(gaussFunc,_1,x0,sigmaInv,unitHeight));
#endif
}

R_Func R_Func::bell(R x0, R sigma, bool compactSupport)
{
   if (compactSupport){
#ifdef CPM_Fn
      return F2<R,R,R,R>(x0,sigma)(bellCompSupport);
#else
      return F<R,R>(bind(bellCompSupport,_1,x0,sigma));
#endif
   }
   else{
      return R_Func::Gauss(x0,sigma,true);
   }
}

R_Func R_Func::charFunc(Iv const& iv, R soft)
{
#ifdef CPM_Fn
   return F2<R,Iv,R,R>(iv,soft)(charFuncAux);
#else
   return F<R,R>(bind(charFuncAux,_1,iv,soft));
#endif
}

R_Func R_Func::sig = F<R,R>(sigFunc);

R_Func R_Func::theta = F<R,R>(thetaFunc);

R_Func R_Func::thetac =F<R,R>(thetacFunc);

R_Func R_Func::harOsc(R omega, R delta, R damp)
{
#ifdef CPM_Fn
   return F3<R,R,R,R,R>(omega,delta,damp)(harOscFunc);
#else
   return F<R,R>(bind(harOscFunc,_1,omega,delta,damp));
#endif
}

R_Func R_Func::potCoulomb(R gamma, R r0)
{
#ifdef CPM_Fn
   return F2<R,R,R,R>(gamma,r0)(potCoulombFunc);
#else
   return F<R,R>(bind(potCoulombFunc,_1,gamma,r0));
#endif
}

R_Func R_Func::dPotCoulomb(R gamma, R r0)
{
#ifdef CPM_Fn
   return F2<R,R,R,R>(gamma,r0)(dPotCoulombFunc);
#else
   return F<R,R>(bind(dPotCoulombFunc,_1,gamma,r0));
#endif
}

R_Func R_Func::potMorse(R U0, R r0, R a)
{
#ifdef CPM_Fn
   return F3<R,R,R,R,R>(U0,r0,a)(potMorseFunc);
#else
   return F<R,R>(bind(potMorseFunc,_1,U0,r0,a));
#endif
}

R_Func R_Func::dPotMorse(R U0, R r0, R a)
{
#ifdef CPM_Fn
   return F3<R,R,R,R,R>(U0,r0,a)(dPotMorseFunc);
#else
   return F<R,R>(bind(dPotMorseFunc,_1,U0,r0,a));
#endif
}

R_Func R_Func::potLJ(R U0, R r0)
{
#ifdef CPM_Fn
   return F2<R,R,R,R>(U0,r0)(potLJfunc);
#else
   return F<R,R>(bind(potLJfunc,_1,U0,r0));
#endif
}

R_Func R_Func::dPotLJ(R U0, R r0)
{
#ifdef CPM_Fn
   return F2<R,R,R,R>(U0,r0)(dPotLJfunc);
#else
   return F<R,R>(bind(dPotLJfunc,_1,U0,r0));
#endif
}

R_Func R_Func::potUM(R U0, R r0, R facU, R facR)
{
#ifdef CPM_Fn
   return F4<R,R,R,R,R,R>(U0,r0,facU,facR)(potUMfunc);
#else
   return F<R,R>(bind(potUMfunc,_1,U0,r0,facU,facR));
#endif
}

R_Func R_Func::dPotUM(R U0, R r0, R facU, R facR)
{
#ifdef CPM_Fn
   return F4<R,R,R,R,R,R>(U0,r0,facU,facR)(dPotUMfunc);
#else
   return F<R,R>(bind(dPotUMfunc,_1,U0,r0,facU,facR));
#endif
}

R_Func R_Func::potUM2(R U0, R r0, R facU, R facR)
{
#ifdef CPM_Fn
   return F4<R,R,R,R,R,R>(U0,r0,facU,facR)(potUM2func);
#else
   return F<R,R>(bind(potUM2func,_1,U0,r0,facU,facR));
#endif
}

R_Func R_Func::dPotUM2(R U0, R r0, R facU, R facR)
{
#ifdef CPM_Fn
   return F4<R,R,R,R,R,R>(U0,r0,facU,facR)(dPotUM2func);
#else
   return F<R,R>(bind(dPotUM2func,_1,U0,r0,facU,facR));
#endif
}

R_Func R_Func::potPol(R rr)
{
#ifdef CPM_Fn
   return F1<R,R,R>(rr)(potPolFunc);
#else
   return F<R,R>(bind(potPolFunc,_1,rr));
#endif
}

R_Func R_Func::dPotPol(R rr)
{
#ifdef CPM_Fn
   return F1<R,R,R>(rr)(dPotPolFunc);
#else
   return F<R,R>(bind(dPotPolFunc,_1,rr));
#endif
}

R_Func R_Func::potSim(R rr)
{
#ifdef CPM_Fn
   return F1<R,R,R>(rr)(potSimFunc);
#else
   return F<R,R>(bind(potSimFunc,_1,rr));
#endif
}

R_Func R_Func::dPotSim(R rr)
{
#ifdef CPM_Fn
   return F1<R,R,R>(rr)(dPotSimFunc);
#else
   return F<R,R>(bind(dPotSimFunc,_1,rr));
#endif
}

R_Func R_Func::potWall(R rr, R rWall, R deg)
{
#ifdef CPM_Fn
   return F3<R,R,R,R,R>(rr,rWall,deg)(potWallFunc);
#else
   return F<R,R>(bind(potWallFunc,_1,rr,rWall,deg));
#endif
}

R_Func R_Func::dPotWall(R rr, R rWall, R deg)
{
#ifdef CPM_Fn
   return F3<R,R,R,R,R>(rr,rWall,deg)(dPotWallFunc);
#else
   return F<R,R>(bind(dPotWallFunc,_1,rr,rWall,deg));
#endif
}

R_Func R_Func::potConfine(Iv const& iv, R d, R deg)
{
#ifdef CPM_Fn
   return F3<R,Iv,R,R,R>(iv,d,deg)(potConfineFunc);
#else
   return F<R,R>(bind(potConfineFunc,_1,iv,d,deg));
#endif
}

R_Func R_Func::dPotConfine(Iv const& iv, R d, R deg)
{
#ifdef CPM_Fn
   return F3<R,Iv,R,R,R>(iv,d,deg)(dPotConfineFunc);
#else
   return F<R,R>(bind(dPotConfineFunc,_1,iv,d,deg));
#endif
}

R_Func R_Func::transit(const R2& x, const R2& y, bool linear)
{
#ifdef CPM_Fn
   return F3<R,R2,R2,bool,R>(x,y,linear)(transitFunc);
#else
   return F<R,R>(bind(transitFunc,_1,x,y,linear));
#endif
}

R_Func R_Func::smoothTransit(R2 const& x, R2 const& y, R p)
{
#ifdef CPM_Fn
   return F3<R,R2,R2,R,R>(x,y,p)(smoothTransitFunc);
#else
   return F<R,R>(bind(smoothTransitFunc,_1,x,y,p));
#endif
}

R_Func R_Func::dTransit(const R2& x, const R2& y)
{
#ifdef CPM_Fn
   return F2<R,R2,R2,R>(x,y)(dTransitFunc);
#else
   return F<R,R>(bind(dTransitFunc,_1,x,y));
#endif
}

R_Func R_Func::concat(const V<Iv>& v, const V<R_Func>& f)
{
   Z n1=v.dim();
   Z n2=f.dim();
   cpmassert(n1==n2,"R_Func::concat(V<Iv>,V<R_Func>)");
#ifdef CPM_Fn
   return F2<R,V<Iv>,V<R_Func>,R>(v,f)(concatFunc);
#else
   return F<R,R>(bind(concatFunc,_1,v,f));
#endif
}

namespace{

R zeroToSingularFunc(R const& x, R const& z, R const& s )
{
   if (x<=z || x>=s) return 0.;
   R y=(x-s)*cpminv(z-s);
  // return cpminv(y)-y*3.+y*y*2.;
   return cpminv(y)+y*(y*2.-3.);
}

R dZeroToSingularFunc(R const& x, R const& z, R const& s )
{
   if (x<=z || x>=s) return 0.;
   R zi=cpminv(z-s);
   R y=(x-s)*zi;
   R yi=cpminv(y);
   return (y*4.-yi*yi-3.)*zi;
}

}

R_Func R_Func::zeroToSingular(R z, R s)
{
#ifdef CPM_Fn
   return F2<R,R,R,R>(z,s)(zeroToSingularFunc);
#else
   return F<R,R>(bind(zeroToSingularFunc,_1,z,s));
#endif
}
      // A function f which has satisfies f(z)=f'(z)=0 and has a
      // singularity (z-s)/(x-s) at x=s. Actually it is a function
      // of the variable y = (x-s)/(z-s) and as such of the form
      // 1/y - 3y + 2y*y

R_Func R_Func::dZeroToSingular(R z, R s)
{
#ifdef CPM_Fn
   return F2<R,R,R,R>(z,s)(dZeroToSingularFunc);
#else
   return F<R,R>(bind(dZeroToSingularFunc,_1,z,s));
#endif
}

/////////////////// powers and zero-function //////////////////////////////////////

R_Func R_Func::xZero=R_Func(idZero);
R_Func R_Func::x0=R_Func(id0);
R_Func R_Func::x1=R_Func(id);
R_Func R_Func::x2=R_Func(id2);
R_Func R_Func::x3=R_Func(id3);
R_Func R_Func::x4=R_Func(id4);
R_Func R_Func::x5=R_Func(id5);
R_Func R_Func::x6=R_Func(id6);
R_Func R_Func::x7=R_Func(id7);
R_Func R_Func::x8=R_Func(id8);
R_Func R_Func::x9=R_Func(id9);
R_Func R_Func::x10=R_Func(id10);
R_Func R_Func::x11=R_Func(id11);
R_Func R_Func::x12=R_Func(id12);
///////////////////////// other useful static functions ////////////////
R_Func R_Func::abs=R_Func(absFunc);
R_Func R_Func::cutNeg=R_Func(cutNegFunc);
R_Func R_Func::log10=R_Func(log10Func);

namespace{
   R powFunc(R const& x, R const& p){ return cpmpow(x,p);}
}

R_Func R_Func::pow(R p)
{
#ifdef CPM_Fn
   return F1<R,R,R>(p)(powFunc);
#else
   return F<R,R>(bind(powFunc,_1,p));
#endif
}

/**************** search for minimum ***********************************/
// static tools

namespace{

R sxL,sxU,xCenter,halfWidth;

R distToBoundary(R x)
{
   return (x<=xCenter ? x-sxL : sxU-x);
}

inline R parabola(R x_1, R x0, R x1,  R y_1, R y0, R y1,  Z& ind)
// we assume x1-x0 = x0-x_1 > 0 (thus everything can be coded without
// using x_1. We leave it as an argument for symmetry,
// returns the x-value for which a parabola f(x) satisfying
// y_1=f(x_1), y0=f(x0), y1=f(x1), attains a minimum. If it has no
// minimum we return 0 and give ind a indicative value (0 if the
// 'parabola' is linear, -1 if it takes a maximum)
{
   x_1; // not used see above
   R c=y1-y0*2+y_1;
   if (c==0.){
      ind=0; return 0.;
   }
   else if (c<0.){
      ind=-1; return 0.;
   }
   else{
      ind=1; return x0-(x1-x0)*(y1-y_1)*0.5/c;
   }
}

} // anonymous namespace

R R_Func::serMin(R xL, R& xMini, R xU, R tol)
{
   static Z tryMax=3;
   static Z counterMax=10;
   static R reducStart=0.8;
   static R reducIter=0.5;

   static Z counter=0;
   static Z isLinear=0;
   static Z isConstant=0;
   static Z noMinimum=0;
   static Z miniAbove=0;
   static Z miniBelow=0;

   R x_1,x0,x1,y_1,y0,y1,xMini1,xMini2,xOld,xNew,dist,reduc;
   Z ind;
  // CpmRootX::order(xL,xU); release mode linker does not find the
  // template implementation
   // thus we do it explicitely

   if (xU<xL){ R temp=xU; xU=xL;   xL=temp;}

   Word loc("R_Func::mini(R xL, R& xMini, R xU, R tol)");

   cpmassert(xMini<=xU,loc);
   cpmassert(xL<=xMini,loc);
   sxL=xL;
   sxU=xU;
   xCenter=(sxL+sxU)*0.5;
   halfWidth=(sxU-sxL)*0.5;
   x0=xMini;
   xOld=x0;
   reduc=reducStart;
Iteration:
// iteration part dealing with xMini1
   dist=distToBoundary(x0);
   dist*=reduc;
   x_1=x0-dist;
   x1=x0+dist;
   y_1=(*this)(x_1);
   y0=(*this)(x0);
   y1=(*this)(x1);
   xMini1=parabola(x_1,x0,x1,y_1,y0,y1,ind);
   if (ind==-1){ // try to find a new start, we would get a maximum here
      noMinimum++;
      x0=xCenter+ran(halfWidth);
      if (noMinimum<tryMax) goto Iteration; else goto BadEnd;
   }
   else if (ind==0){
      R slope=y1-y_1;
      if (slope!=0){
         isLinear++;
         x0=xCenter+ran(halfWidth);
         if (isLinear<tryMax){
            goto Iteration;
         }
         else{
            xMini=(slope>0 ? xL : xU);
            return (*this)(xMini);
         }
      }
      else{
         isConstant++;
         x0=xCenter+ran(halfWidth);
         if (isConstant<tryMax){
            goto Iteration;
         }
         else{
            xMini=xCenter;
            return (*this)(xMini);
         }
      }

   }
   if (xMini1>=xU){
      miniAbove++;
      x0=xCenter+ran(halfWidth);
      if (miniAbove<tryMax){
            goto Iteration;
      }
      else{
         xMini=xU;
         return (*this)(xMini);
     }
   }
   if (xMini1<=xL){
      miniBelow++;
      x0=xCenter+ran(halfWidth);
      if (miniBelow<tryMax){
            goto Iteration;
      }
      else{
         xMini=xL;
         return (*this)(xMini);
     }
   }
   x0=xMini1;
// iteration part dealing with xMini2, identical to previous block but
// xMini1 replaced by xMini1 (not particularly elegant !!!)
   dist=distToBoundary(x0);
   dist*=reduc;
   x_1=x0-dist;
   x1=x0+dist;
   y_1=(*this)(x_1);
   y0=(*this)(x0);
   y1=(*this)(x1);
   xMini2=parabola(x_1,x0,x1,y_1,y0,y1,ind);
   if (ind==-1){ // try to find a new start, we would get a maximum here
      noMinimum++;
      x0=xCenter+ran(halfWidth);
      if (noMinimum<tryMax) goto Iteration; else goto BadEnd;
   }
   else if (ind==0){
      R slope=y1-y_1;
      if (slope!=0){
         isLinear++;
         x0=xCenter+ran(halfWidth);
         if (isLinear<tryMax){
            goto Iteration;
         }
         else{
            xMini=(slope>0 ? xL : xU);
            return (*this)(xMini);
         }
      }
      else{
         isConstant++;
         x0=xCenter+ran(halfWidth);
         if (isConstant<tryMax){
            goto Iteration;
         }
         else{
            xMini=xCenter;
            return (*this)(xMini);
         }
      }

   }
   if (xMini2>=xU){
      miniAbove++;
      x0=xCenter+ran(halfWidth);
      if (miniAbove<tryMax){
            goto Iteration;
      }
      else{
         xMini=xU;
         return (*this)(xMini);
     }
   }
   if (xMini2<=xL){
      miniBelow++;
      x0=xCenter+ran(halfWidth);
      if (miniBelow<tryMax){
            goto Iteration;
      }
      else{
         xMini=xL;
         return (*this)(xMini);
     }
   }
   xNew=(xMini1+xMini2)*0.5;
   if (  abs(xNew-xOld)>tol )
   {
      counter++;
      if (counter<counterMax){
         xOld=xNew;
         x0=xNew;
         reduc*=reducIter;
         goto Iteration;
      }
      else{
         xMini=-1000;
         cpmerror("R_Func::mini(): no convergence after "&
            cpmwrite(counterMax)&" iterations");
         return 0.;
      }
   }
   else{
      xMini=xNew;
      return (*this)(xMini);
   }

BadEnd:
   xMini=-1000;
   cpmerror("R_Func::mini(): no minimum found");
   return 0;
}

R_Vector CpmFunctions::gradient(const R_Vector& x,
   const CpmFunctions::F<R_Vector,R>& f, R h)
{
   cpmassert(h!=0,"CpmFunctions::gradient(...)");
   R hInv=R(1)/h;
   Z n=x.dim();
   R_Vector res(n);
   R_Vector xval;
   R y=f(x);

   for (Z i=1;i<=n;i++){
      xval=x;
      xval[i]+=h; // shifted on i-axis
      res[i]=(f(xval)-y)*hInv;
   }
   return res;
}

R_Matrix CpmFunctions::gradient(const R_Vector& x,
   const CpmFunctions::F<R_Vector,R_Vector>& f, R h)
{
   cpmassert(h!=0,"R_Matrix CpmFunctions::gradient(...)");
   R hInv=R(1)/h;
   Z n=x.dim();
   R_Matrix res(n,1);
   R_Vector xval;
   R_Vector y=f(x);

   for (Z i=1;i<=n;i++){
      xval=x;
      xval[i]+=h; // shifted on i-axis
      res[i]=(f(xval)-y)*hInv;
         // this says that the first index of the result gives the
         // direction of derivation and the second index gives the
         // component of f for which the derivation is formed. This is as
         // usual d/dx[i] f(x)[j] = grad[i][j]
   }
   return res;
}

R CpmFunctions::besselJ(Z n, R x)
{
   return cpmbessel(n,x);
}

namespace{
   R besselFunc(R const& x, Z const& n)
   { return CpmFunctions::besselJ(n,x);}
}

R_Func R_Func::Bessel(Z n)
{
#ifdef CPM_Fn
   return F1<R,Z,R>(n)(besselFunc);
#else
   return F<R,R>(bind(besselFunc,_1,n));
#endif
}
