//? cpmrfunc.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_R_FUNC_H_
#define CPM_R_FUNC_H_
/*

Description:

   Specializing the 'function template' Fa<X,Y> to X=R and Y=R
   'real functions'. Beyond mere specialization, interpolation is
   treated: There are constructors for class R_Func which define a
   function from discrete lists. In addition, differentiation,
   integration, and some other topics are implemented. Graphical
   representation is implemented by a suitable  member function of class
   Graph.

   Notice that the rather complex template Fa<X,Y> is hidden to the
   user of the present class. He has only to include cpmrfunc.h.
   This, however, includes cpmfa.h, so the program size will be
   influenced by the complexities of this construction.

   Good example for usage of complex templates as a 'production machine'
   for small simple non-template classes.

*/

#include <cpmfa.h>
#include <cpmrmatrix.h>
#include <cpmsmoothlocalinterpol.h>

namespace CpmFunctions{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmArrays::V;
   using CpmArrays::R_Vector;
   using CpmArrays::R_Matrix;
   using CpmArrays::IpolControl;
   using CpmGeo::Iv;
   using CpmLinAlg::R2;
   using CpmLinAlg::R4;

////////////////////////////////// class RomCon ////////////////////

class RomCon{ // control of Romberg integration
   //: Romberg control
   // Method parameters for Romberg integration
   R epsRel_;
      //: epsilon relative
      // bound for relative error
    R epsAbs_;
      //: epsilon absolute
      // bound for absolute error
      // If the extimated error remains below these bounds,
      // the computation is finished
   Z met_;
      //: method
      // 0 for no extrapolation, only sucessive refinement
      // 1 for polynomial extrapolation
      // 2 for rational extrapolation
   Z len_;
      //: length
      // number of values on which extrapolation is based
   Z cntMax_;
      //: counter maximum
      // bound on the number of interval trisections
   bool robErrEst_;
      //: robust error estimate
      // if this is true, the error estimate is based
      // on actual value change seen in the last
      // interval trisection
   typedef RomCon Type;
public:
   Word nameOf()const{ return Word("RomCon");}
   CPM_IO
   RomCon(R epsRel=1e-6, R epsAbs=1e-9, Z met=1, Z len=5, Z cntMax=9,
      bool robErrEst=false): epsRel_(epsRel),epsAbs_(epsAbs),met_(met),
      len_(len),cntMax_(cntMax), robErrEst_(robErrEst){}

    R epsRel()const{ return epsRel_;}
    void setEpsRel_(R eps){ epsRel_=cpmabs(eps);}

    R epsAbs()const{ return epsAbs_;}
    void setEpsAbs_(R eps){ epsAbs_=cpmabs(eps);}

    Z met()const{ return met_;}
    void setMet_(Z m){ met_=cpmpos(m);}

    Z len()const{ return len_;}
    void setLen_(Z n){ len_=cpmpos(n);}

    Z cntMax()const{ return cntMax_;}
    void setCntMax_(Z n){ cntMax_=cpmpos(n);}

    bool robErrEst()const{ return robErrEst_;}
    void robErrEst(bool b){ robErrEst_=b;}
    V<Word> show()const;
      // returns an array of text lines which can be used for
      // screen display using the Frame::setText(V<Word> const&)
      // method
};

///////////////////////////// class R_Func ///////////////////////////////

R hillFunc(const R& x, const R& p=0.810747);
//: hill function
// smooth down-hill connection from (0,1) to (1,0).
// see static R_Func hill(R p=0.810747); for details.
// Of course, 1-hillFunc is a smooth up-hill connection
// from (0,0) to (1,1), so that there is no need for a
// separate function for this case.

class R_Func: public Fa<R,R>{ // functions R to R
   // R-valued functions on R with arithmetics and trivially defined
   // order properties, and construction from discrete data via
   // interpolation.
   // Many functions were added to serve the needs of an industrial
   // project dealing with bending wires by computer controlled
   // motion of bending tools. Here the dependence of tool coordidates
   // on time was represented by instances of R_Func.

public:

// carrying over the constructors from Fa<R,R>

   R_Func(std::function<R(R)> f):Fa<R,R>(f){}

   R_Func(const F<R,R>& g):Fa<R,R>(g){}
      // upcast constructor

   R_Func(const Fa<R,R>& g):Fa<R,R>(g){}
      // upcast constructor

   R_Func(const R_Func& g):Fa<R,R>(g){}
      // copy constructor

//   R_Func(void):Fa<R,R>(){}
      // default constructor

// Construction from function pointers

explicit   R_Func( R (*f)(const R& )):Fa<R,R>(f){}
explicit   R_Func( R (*f)(R)):Fa<R,R>(f){}

// Construtor for constant function and default constructor

explicit R_Func(R const& y1 = R{0.}):Fa<R,R>(y1){};

// new constructor from lists of values, which define a real function
// via interpolation

explicit R_Func(const R_Vector& data,
   const IpolControl& c=IpolControl(4));
// data is a list of y-values
// the corresponding x values are 1,2,...dim(data)=:n

// This applies to the next three functions too.

   R_Func(const V<R>& x, const V<R>& y,
      const IpolControl& c=IpolControl(4));
      // x is a list of x-values
      // y is a list of y-values

explicit R_Func(const R_Matrix& data,
      const IpolControl& c=IpolControl(4));
      // data.dim1() has to be 2
      // data[1] is a list of x-values
      // data[2] is a list of y-values

explicit R_Func(const Word& filename,
   const IpolControl& c=IpolControl(4));
      // filename has to be a name of a file which contains
      // x[1] y[1] x[2] y[2] ... separated by blancs, tabs, or CRs
      // and which may contain comment lines started by *, /, or ;

   R_Func(R a0, R a1, R a2, R a3, R a4=0);
   // generates the 4th order polynomial y=a0+a1*x+a2*x^2+a3*x^3+a4*x^4
   // also a polynomial of third order can be created this way

   R_Func(Iv iv, R mean, R sigma);
   // generates a Gaussian bell curve where mean and sigma are relative
   // to the interval iv

   R_Func(R4 a);
   // generates a smooth wall function function which  transits
   // from 0 to 1 in the interval (a[1],a[2]) and falls back to 0 in the
   // interval (a[3],a[4]). A function call stops with error if the
   // following is not satisfied:
   // a[1]<a[2]<=a[3]<a[4]. Implementation is based on R_Func::hill()

   R_Func(R2 x, R2 y);
   // linear transformation such that x[1] |--> y[1], x[2] |--> y[2]

   R_Func(Iv x, Iv y);
   // linear transformation such that x[1] |--> y[1], x[2] |--> y[2]

   Word nameOf()const{ return Word("R_Func");}

// modifications
   R_Func stretch(R2 const& ri, R2 const& rf)const
   { return circ(R_Func(rf,ri));}
      // returns a function which on the range rf behaves just as
      // the original *this did on the range ri.
      // Thus for R_Func g=f.stretch(ri,rf) we have
      // g(x)=f(tr(x)) where tr: rf-->ri (so, when x varies
      // over rf, the original function gets evaluated over ri)

   R_Func stretch(Iv const& ivi, Iv const& ivf)const
   { return circ(R_Func(ivf.toR2(),ivi.toR2()));}
      // returns a function which on the interrval ivf behaves just as
      // the original *this did on the interval ivi.
      // Thus for R_Func g=f.stretch(ivi,ivf) we have
      // g(x)=f(tr(x)) where tr: ivf-->ivi (so, when x varies
      // over ivf, the original function gets evaluated over ivi)

   R_Vector eval(R_Vector const& rx)const;
      // natural action on R_Vectors. Notice that f(rx) is already
      // defined via the base class. But the result is a V<R> and
      // not directly a R_Vector. Thus f.eval(rx)=R_Vector(f(rx));
      // so the present function isn't needed urgently. Notice also that
      // it is creating desaster to define this operator by overloading
      // operator(), since this would hide the old operator ().
      // I consider this one of the most un-natural aspects of C++.

   void stretch_(R k, R x0=0);
      // changes the function f into x|-->f(x0+(x-x0)/k). Thus for
      // k>1 features of the function (e.g. with of a Gauss curve)
      // get enlarged. If x0 is not in the area of the feature, the
      // feature appears more to get shifted than extended.
      // Anyway, x0 is a fixed point of the operation.

   R_Func morph(Iv iv, const R_Func& g)const;
      // Let g be monotone increasing on [0,1] such that g(0)=0
      // and g(1)=1. For x \in iv we have
      // (f.morph(iv,p))(x)=f(x'), x':=x0+g(p)(x1-x0)
      // p=iv.ramp(x), iv=Iv(x0,x1)

   void shift_(R s);
      // changes *this in a way that the graph gets shifted by s
      // (i.e. to the right on a screen graphics for s>0)

   R_Func shift(R s)const;
      // non-mutating form of thr previous function

   R_Func sym(Z mode=1)const;
      //: symmetrize
      // Defines a function which coincides with *this for x>=0
      // and is symmetric or anti-symmetric relative to x=0
      // depending on the sign of mode:
      // f(-x) := sign(mode)*f(x) for all x>0
      // For mode=0, the returned function equals *this.

   R_Func mirror(R x0, R s=1.)const;
      // Defines a function which is mirror symmetric with respect
      // to x0 and coincides with *this for x<x0. This is a natural
      // requirement for functions that represent 1-dimensional
      // position versus time. Then the "mirrored" function describes
      // a situation where for t=t0 a motion reversal takes  place.
      // First need occured in a stepper motor control problem.
      // (2002-1-21)
      // Let g:=f.mirror(x0).
      // Then g(x)=( x>x0 ?  f(x0-s*(x-x0)) : f(x) )

   R_Func append(const R_Func& f, R x0)const;
      // Let h:=f.append(g,x0).
      // Then h(x)=( x>x0 ?  g(x) : f(x) )

   R_Func constRightOf(R x0)const{ return append(R_Func(evl(x0)),x0);}
      // Let h:=f.constRightOf(x0).
      // Then h(x)=( x>x0 ?  g(x) : g(x0) )

   R_Func squeezeInConstant(Iv const& ivHost, Iv const& ivSqueezeIn)const;
      // Let h:=f.squeezeInConstant(ivHost,ivSqueezeIn)
      // Then f(x)=h(x) for x in the complement of ivHost
      // (thus it is only on ivHost that something happens to f)
      // h(x)=f(x0) for x in ivSqueezeIn, where x0 is the center of
      // ivSqueezeIn. On ivHost\ivSqueeceIn h is defined from f
      // by piecewise linear re-transformation of the argument
      // such that 'space is created for inserting the constant function
      // piece on ivSqueezeIn. Prototype for the most efficient
      // implementation style for R_Func's that evaluate in a way that
      // needs precomputed quantities for efficiency.

   R_Func squeezeInConstantLeft(const Iv& ivHost, R d)const;
      // Let h:=f.squeezeInConstantLeft(ivHost,d)
      // Then f(x)=h(x) for x in the complement of ivHost
      // (thus it is only on ivHost that something happens to f)
      // h(x)=f(x0) for x in [x0, x0+d] where x0=ivHost[1]
      // On ivHost\[x0,x0+d] h is defined from f
      // by linear re-transformation of the argument
      // such that 'space is created for inserting the constant function
      // piece on [x0,x0+d].

   R_Func squeezeInConstantRight(const Iv& ivHost, R d)const;
      // Let h:=f.squeezeInConstantRight(ivHost,d)
      // Then f(x)=h(x) for x in the complement of ivHost
      // (thus it is only on ivHost that something happens to f)
      // h(x)=f(x0) for x in [x0-d, x0] where x0=ivHost[2]
      // On ivHost\[x0-d,x0] h is defined from f
      // by linear re-transformation of the argument
      // such that 'space is created for inserting the constant function
      // piece on [x0-d,x0].

   R_Func shrink(const Iv& ivHost, const Iv& ivSub)const;
      // Consider
      // g=f.shrink(ivHost,ivSub);
      // f=f if ivSub fails to be contained in ivHost.
      // If ivSub is a subinterval of ivHost,
      // g looks on ivSub just as f looked on ivHost.
      // On ivHost\ivSub the g is obtained as constant continuation
      // of g|ivSub.
      // Further: g==f on R\ivHost.

// derivation
   R_Func der(Z n=1, R h=1e-3)const;
      // returns the n-fold derivation of *this taking h as the
      // finite difference. Since higher order formulas are used for
      // the derivatives, h has not to be very small in order to give
      // accurate results. If n is large, h has to be rather large to
      // give anything meaningful.

    R_Func derR(Z n, R h=1e-3)const;
      // iterative definition of by multiple application of der(1)
      // much slower than previous function.

// integration
   R_Func antiDer(R x0, R h=1e-3)const;
      // anti derivative
      // (this->integral(x0))(x) = integral from x0 to x of function
      // *this. Here h is a proposal for the space between the various
      // x-values used in Simpson's formula, actually an adjusted value
      // will be used there.

   R romInt(R x1, R x2, RomCon const& rc = RomCon() )const;
      // Romberg integration
      // More sophisticated method for integrating from x1 to x2.
      // Function evaluation at x1 and x2 is avoided, so integrable
      // singularities are allowed to be there.
      // Convergence then will not be as fast as for non-singular
      // functions.
      // Modified from functions midpnt, and qromo of Press et. al.

// minimization
   R serMin(R xL, R& xMini, R xU, R tol=1e-6);
      // search minimum
      // after call xMini will be a value where the function takes a
      // minimum value in the interval xL <= xMini <= xU
      // The return value is the value which the function gives for
      // xMini. Method is iterative inverse parabolic interpolation so it
      // is fast and simple, but not robust in dealing with
      // non-cooperative functions.
// operation
   F<R2,R> tenPrd(R_Func const& f)const;
      //: tensor product

// static members which are useful
   static R_Func xZero;
      // function x |--> R{0.}
   static R_Func x0;
      // function x |--> R{1.}
   static R_Func x1;
      // function x |--> x
   static R_Func x2;
      // function x |--> x*x
   static R_Func x3;
      // function x |--> x*x*x
   static R_Func x4;
      // function x |--> x*x*x*x
   static R_Func x5;
      // function x |--> x*x*x*x*x
   static R_Func x6;
      // function x |--> x*x*x*x*x*x
   static R_Func x7;
      // function x |--> x*x*x*x*x*x*x
   static R_Func x8;
      // function x |--> x*x*x*x*x*x*x*x
      // x5,x6,x7,x8 added 2011-03-21
   static R_Func x9;
      // x9,x10,x11,x12 added 2015-01-12
   static R_Func x10;
   static R_Func x11;
   static R_Func x12;

   static R_Func log10;
      // function x |--> log10(|x|)

   static R_Func abs;
      // function x |--> |x|

   static R_Func cutNeg;
      // function x |--> Max{0,x}

   static R_Func pow(R p);
      // function x |--> pow(x,p)

// Continuous functions which are constant outside the interval (0,1):

   static R_Func hill(R p=0.810747);
   // smooth down-hill connection from (0,1) to (1,0).
   // The function x |--> hill(x) is a smooth (i.e. derivatives of any
   // order exist) non-increasing function.
   // It has the constant value 1 for x<=0 and the constant value
   // 0 for x>=1.  So, it is a smooth down-hill shape, connecting level
   // 1 with level 0.
   // The whole curve is radially symmetric to the point (x,y)=(0.5,0.5).
   // The idea behind the construction is the following: The negative
   // hyperbolic tangent function is already similar to such a hill.
   // A transformation of the argument brings the x-values 0 and 1 to
   // infinity before they are inserted into the hyperbolic tangent.
   // The normal expression for the latter is transformed in order to
   // call exp only once. Futher, to avoid overflow riscs in exp, this
   // function is called only for negative arguments.
   // Actually there is a whole family of such depending on a parameter p
   // (see Kodak TR 278844A 09/24/92), here the value 0.810747 is
   // taken, which minimizes the L2-norm of the second derivative.

   static R_Func hill3;
   // similar to hill, but a polynomial of third order and function has
   // one continuous derivative only instead of all derivatives.

   static R_Func ramp(R p=0.810747);
      // R_Func(R_Func::x0 - R_Func::hill(p))
      // thus smooth up-hill connection from (0,0) to (1,1).

   static R_Func ramp3;
      // R_Func(R_Func::x0 - R_Func::hill3)
      // thus third order polynomial up-hill connection from
      // (0,0) to (1,1).
      // see my Mathematica file trivialramp.nb for interesting
      // properties of ramps under translation

   static R_Func transit(R2 const& x, R2 const& y, bool linear=false);
      // creates a third-order polynomial f such that
      // f(x[1])=y[1],f(x[2])=y[2], f'(x[1])=0, f'(x[2])=0
      // Outside the x-Range given by argument x, the returned
      // function is defined by constant continuation of f
      // Due to the vanishing derivations of f at x[1] and x[2]
      // the resulting total function is differentiable on R.
      // This description refers to the case that the boolean argument
      // has its default value 'false'. If it has the value 'true'
      // the third-order polynomial is replaced by a first-order
      // (linear) one so that the transition function is still
      // continuous but not differentiable. This turned out to be
      // useful in graphical tone scale transformations.

   static R_Func smoothTransit(R2 const& x, R2 const& y, R p=0.810747);
      // creates a smooth f such that f(x[1])=y[1],f(x[2])=y[2]
      // and that all derivatives of f vanish at x=x[1] and x=x[2].
      // Outside the x-Range given by argument x, the returned
      // function is defined by constant continuation of f.
      // The smooth continuation is closely related to the function
      // 'hill' and the parameter p has the same meaning as there.

   static R_Func dTransit(R2 const& x, R2 const& y);
      // Derivation of transit, evaluates faster than
      // R_Func::transit(x,y).der()
      // Here function transit is always taken with the boolean
      // argument having its default value false.

   static R_Func terrace(R k=0.25);
      // up-hill connection from (0,0) to (1,1). Like ramp,
      // radially symmetric with respect to (0.5,0.5)
      // 0<=k<=1. The derivative at point (0.5,0.5) is
      // by a factor k smaller than in (0,0) and (1,1).
      // Invented by Thomas Dera in the context of stepper
      // motor control (2001-1-18)

   static R_Func terrace2(R k=0.25);
      // similar to previous function; essentially tangens replaced by
      // arcsin (Thomas Dera Mathematica file ramp.nb (2002-01-31)
      // also 0<=k<=1 and the same relation to the derivatives.

   static R_Func terraceBezierDera(R k, R m, bool inv=false, Z z=0,
      Z n=256, const IpolControl& c=IpolControl("polygon"));
      // terrace function as Bezier curve, designed by
      // Thomas Dera 2002-02-06
      // See ramps(TD).pdf
      // If inv==true, the inverse function is returned
      // For z==0 the values k and m are active to shape the curve
      // in a symmetrical way at both ends.
      // For z=-1 the left end is 'free' (there are only 3 Bezier points,
      // the neighbor of (0,0) is missing)
      // For z=1 the right end is 'free' (there are only 3 Bezier points,
      // the neighbor of (1,1) is missing)
      // Note of 2003-07-16: Since the function is defined by
      // interpolation,
      // a small value of n may cause unexpected behavior in detail.
      // This was the case in the CO-former application where a
      // mysterious slope change of a time-position diagram could be
      // removed by changing n from 32 to 64. Now 100 for security.
      // Addition 2003-07-17: For applications it is essential that the
      // function and its derivative are monotonic, this can best be
      // guarantied by "polygon"-interpolation and a large value of n.
      // (since polygon interpolation is simple, a large number of n
      // is OK)

static R_Func polyFunc(const R_Vector& c);
      // returns a suitably coded polynomial function
      // x0:=c[1]
      // 1/r:=c[2]
      // z:=(x-x0)/r
      // f(x)=c[3]+c[4]*z+c[5]*z*z+...+c[n]*z^(n-3).
      // Data vectors c characterizing polynomial functions obtained
      // fit to discrete value lists can be obtained by function
      // CpmArrays::polyFit(...), see cpmvectormore.h
      // Notice that polynomials in x (i.e. x0=0, r=1) that
      // as functions show natural behavior in a
      // short interval of large x (e.g. 100<x<101) have coefficients
      // spanning many orders of magnitude. Therefore allowing for
      // x0 and r helps to get numerically more convenient
      // parametrization of polynomial functions.

   static R_Func Bernstein(R_Vector const& y, R2 a=R2(0,1));
      // Let n:=y.dim(). Then Bernstein(y,a) is the Bernstein
      // approximation for a continuous function f taking
      // values y[i]=f(x[i]), where x[i]=a[1]+((i-1)/(n-1))*(a[2]-a[1])
      // This always gives a smooth naturally looking curve which is
      // a special case of a Bezier spline. It is not exactly
      // interpolating, but the list y is an efficient design tool for
      // 'natural curves' showing a certain desired behavior.
      // See G. Aumann, K. Spitzmueller: Computerorientierte Geometrie,
      // BI 1993, pp. 348-350.
      // The implementation may serve as a pattern for other functions
      // that need extensive pre-calculations for fast evaluation:
      // R_Vector y=....;
      // R2 a(-1,1);
      // R_Func f=R_Func::Bernstein(y,a); // does the pre-calculations
      // R xm=0.5;
      // R min=f.mini(-1,xm,+1); // the evaluations of f needed in
      // computing function mini(...) are fast.

   static R_Func Bezier(const R_Vector& x, const R_Vector& y, Z n=256);
      // Function defined by a Bezier curve with knots (x[i],y[i]).
      // Stops with cpmerror if this curve fails to be a function graph.
      // The final function is generated by interpolating (via polygon
      // method since here one should employ only monotonity preserving
      // interpolation methods) from computed n function values. Since
      // the method is very efficient, there is no problem in having
      // n rather large).
      // For a proper working of the algorithm we don't need x.dim()
      // to equal y.dim(). So the implementation does not test. Not
      // clarified what this means in terms of multiple knots.

   static R_Func Gauss(R x0, R sigma, bool unitHeight=false);
      // Gauss bell curve with mean x0 and standard deviation sigma;
      // If unitHeight==true, the conventional normalization factor
      // is left out so that the maximum function value is 1.
      // This is sometimes useful if the function is used as a building
      // block for curve design as now implemented in class
      // R_FuncRec.

   static R_Func bell(R x0, R sigma, bool compactSupport=true);
      // If the last parameter is false we get the Gauss bell curve with
      // mean x0 and standard deviation sigma and unit height.
      // If compactSupport==true, we get a very similiar curve which,
      // however, vanishes identically for distances farer away than
      // 4 sigma from the peak position x0. In this case the curve has
      // compact support of length 8 sigma and, nevertheless, is a smooth
      // curve.

   static R_Func charFunc(Iv const& iv, R soft=0);
      // For soft=0 (default) this is the
      // 'characteristic function' which takes the value 1 for x in iv
      // and the value 0 else.
      // If soft>0, the function also vanishes outside the interval
      // and takes the value 1 at the center of the interval. Let p
      // be a parameter which varies from -1 to +1 in iv (hence
      // p=iv.linear(x);) then the function value in the interior
      // of the interval is given by
      // y=exp(soft*soft*(1-r)) where r:=1/(q*q), q:=(1-p)*(1+p)
      // soft==1: no indication of a plateau in the middle
      // soft==0.01: good plateau, steep fall near the boundaries of
      // the interval.
      // The smaller soft is, the more rectangular is the function graph.
      // For large values of soft, we approximate the delta-peak (located
      // at the center of iv. However, the hight is 1 and not the
      // integral (as would be needed for a approximation of the Dirac-
      // distribution).

   static R_Func sig;
      //: signum
      // Returns 1. for x>0, 0. for x==0. and -1 for x<0.

   static R_Func theta;
      //:: theta
      // Returns 1. for x>=0  and 0. else.

   static R_Func thetac;//{ return (x0 - theta); }
      //:: theta complement
      // theta + thetac = x0

   static R_Func harOsc(R omega=1., R delta=0., R damping=0.);
      //: harmonic oscillation
      // returns sin(omega*x+delta)*exp(-damping*x)

   static R_Func potCoulomb(R gamma, R r0);
      // Coulomb potential with a bit naive regularization. The
      // regularization changes the function value at any distance.

   static R_Func dPotCoulomb(R gamma, R r0);
      // Derivation of the regularized Coulomb potential.

   static R_Func potMorse(R U0, R r0, R a);
      // Morse potential as described in L.I.Schiff: Quantum Mechanics,
      // third edition,McGraw-Hill, p. 452, (49.16)

   static R_Func dPotMorse(R U0, R r0, R a);
      // Derivation of the Morse potential. Analytical expression,
      // which evaluates faster than MorsePot(U0,r0,a).der().

   static R_Func potLJ(R U0, R r0);
      // Lennard-Jones potential as described in Hoffmann, Schreiber
      // (Eds)
      // Computational Physics, Springer 1996, p. 270 (3)

   static R_Func dPotLJ(R U0, R r0);
      // Derivation of the LJ potential. Analytical expression,
      // which evaluates faster than potLJ(U0,r0).der()

   static R_Func potUM(R U0, R r0, R facU=1e4, R facR=2);
      // potUM(0)=facU*U0,potUM(r0)=-U0,potUM(facR*r0)=0
      // So, unlike the potential functions of Morse and Lennard
      // Jones, the potential is continuous at the origin and
      // the 'repulsive energy' is under control. Also
      // the range is finite and under control too.
      // Note 2015-01-19: That the repulsive energy is finite at the origin
      // is a disadvantage since a pair of particles may be squeezed by
      // surrounding particles so strongly that their mutual repulsion
      // gets overcome. So we have a narrow pair that will violently
      // disintegrate when the surrounding particles move away. This dis-
      // advantage will be overcome by potential potUM2 which has a 1/r
      // singularity at r=0. This avoids the squeezing effect but may
      // create unwantedly high-energy particle pairs in ensembles of
      // randomly placed particles. The most robust method to cobe with
      // this situation is a preparatory run which ramping up the inter-
      // action smoothly and sets the velocities to zero finally. It is
      // advisable to do only a rather limited number of integration
      // steps and repeat the process a few times.

   static R_Func dPotUM(R U0, R r0, R facU=1e4, R facR=2);
      // Derivation of potUM

   static R_Func potUM2(R U0, R r0, R facU=1e4, R facR=2);
      // potUM2(r0)=-U0,potUM2(facR*r0)=0, potUM2(r) ~ 1/r near r=0.
      // The range is finite and under control (range = r0*facR).
      // The parameter facU in this case approximately multiplies 1/r.
      // Actually this is done in a way that the function is differentiable
      // also at r=r0.

   static R_Func dPotUM2(R U0, R r0, R facU=1e4, R facR=2);
      // Derivation of potUM2

// non-singular potentials and forces designed with Mathematica:
// sphericalFluid8.nb
// 'Design of another handsome potential using also odd powers'.
   static R_Func potPol(R rr);
      //: potential polynomial
      // A potential with a repulsive core and a binding well.
      // We denote the function argument by r and assume r>=0.
      // The total range is rr. In the interval [0,rr] the values f(r) are
      // given by a polynomial of degree 5. The derivatives of the
      // polynomial vanish at r=0, r=0.6*rr and r=rr.
      // We have f(0)=4, f(0.6*rr)=-2, f(rr)=0

   static R_Func dPotPol(R rr);
      // differentiated potential polynomial
      // dPolPol=potPol.der()

   static R_Func potSim(R rr);
      //: potential simple

   static R_Func dPotSim(R rr);
      // differentiated simple potential
      // dPolSim=potSim.der()

// simple wall potential and forces
   static R_Func potWall(R rr, R rWall, R deg=6.);
      // We denote the function argument by r and assume r>=0.
      // The potential becomes increasingly repulsive near r=rWall.
      // It vanishes for r <= rWall-2*rr and is x^deg for
      // x:= (r - (rWall -2*rr))/rr else. Notice that the potential is
      // nowhere singular. It has the value 2^deg at r=rWall which is
      // 64 for our default value for deg.

   static R_Func dPotWall(R rr, R rWall, R deg=6.);
      // dPolWall=potWall.der()

   static R_Func potConfine(Iv const& iv, R d, R deg=2.);
      // A simple potential which is constantly zero within iv and grows
      // as pow(x,d), x = (distance to iv)/d, outside of iv.

   static R_Func dPotConfine(Iv const& iv, R d, R deg=2.);
      // dPolConfine=potConfine.der()

   static R_Func concat(const V<Iv>& v, const V<R_Func>& f);
      // We return a function g such that
      // g(x)=f[i](x) for x \in v[i]

   static R_Func Bessel(Z n);
      // Bessel function J_n of integer order

   static R_Func zeroToSingular(R z, R s);
      // A function f which has satisfies f(z)=f'(z)=0 and has a
      // singularity (z-s)/(x-s) at x=s. Actually it is a function
      // of the variable y = (x-s)/(z-s) and as such of the form
      // 1/y - 3y + 2y*y

   static R_Func dZeroToSingular(R z, R s);
      // Derivation of zeroToSingular(z,s)

};

R_Vector gradient(const R_Vector& x, const F<R_Vector,R>& f, R h);
// Gradient of scalar function f at point x using h as a increment for
// x-coordinates. Minimum number of function evaluations for fast
// execution.
// Let gr be the return value. Then gr[i]=df(x)/dx[i].

R_Matrix gradient(const R_Vector& x, const F<R_Vector,R_Vector>& f, R h);
// Gradient of vector-valued function f at point x using h as a increment
// for x-coordinates. Minimum number of function evaluations for fast
// execution.
// Let gr be the return value. Then gr[i][j]=d/dx[i] f(x)[j].

// Bessel function of integer order.
// Here as a simple function on which the implementation of
// R_Func::Bessel(Z) is based.
R besselJ(Z n, R x);
// Bessel function J_n(x) for real x and integer n, n=0,1,2,...
// Implementation from boost/special_functions/bessel.hpp

} // namespace

#endif
