//? cpmrfuncrec.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_R_FUNCREC_H_
#define CPM_R_FUNCREC_H_
/*

Description:
   Defining class R_FuncRec = real function recordable.
   A descendant of class R_Func which contains (unlike R_Func
   itself) a complete set of design data (which are streamable)
   so that the I/O interface CPM_IO, and also CPM_ORDER,
   can be implemented. The tool for this is class Recordable
   from which R_FuncRec is derived.

*/

#include <cpmrfunc.h>
#include <cpmrecord.h>

namespace CpmFunctions{

   using CpmRoot::R;
   using CpmRootX::Record;
   using CpmRootX::Recordable;

/////////////////// class R_FuncRec //////////////////////////////////

class R_FuncRec: public Recordable{ // functions as recordable objects
      // see Description above

   mutable R_Func f_;
      // dependent quantities

   virtual bool vadjust()const;
      // setting f_ if rc_ is given
      // returns true if it was successful

   typedef R_FuncRec Type;
   typedef Recordable Base;

public:
   explicit R_FuncRec(Record const& rc=Record()):Base(rc){}
      // Record is a very flexible input since it can be 
      // obtained from a section of a RecordHandler and thus from
      // a ini-file
      // Example:
/*************************************************************************
Diagr::Diagr(const CpmRootX::RecordHandler& rch):
...
{
  ...

   sec="system data";
   R m,lambda,dx;
   cpmrh(m);
   cpmrh(lambda);
   cpmrh(dx);
   Record rc=rch(); // overladed operator()
// 'readig a function from file'
   Record rcSec=rc("function data","R_FuncRec");
   R_FuncRec frec(rcSec);
   R_Func pot=frec(); // overladed operator()
   act_=Act(m,lambda,dx,pot); // R_Func-typed constructor argument
*************************************************************************/

   R_Func operator()(void)const{ adjust(); return f_;}
   R operator()(R const& x){ adjust(); return f_(x);}
};

/*
// The following is an example of a file content which by reading
// via Record::readFormatted would define a Record that would work as
// an argument of R_FuncRec::R_FuncRec(Record const&) 
// Notice the previous example, where it is shown that a section
// in an arbitrary ini-file also can provide the data.
// Notice that not all data of this maximum set need to be there.
// This defines a elaborate but fixed framework to define a function.
// Details are to be looked up in the definition of
//    bool R_FuncRec::vadjust(void)const
// in file cpmrfuncrec.cpp.

////////////////////////////////////////////////////////////////////////
   R_FuncRec
////////////////////////////////////////////////////////////////////////

R polyFac=1
   c polynomial contribution: enables a polynomial contribution to the
   c potential defined by the polynomial coefficients poly.

R bernsteinFac=0
   c Bernstein contribution: enables a contribution to the potential
   c which is a Bernstein polynomial defined by tentative y-values 
   c bernstein valid at equidistant x-values

R bernsteinX=-0.5
R bernsteinY=0.5

R bezierFac=0
   c Bezier contribution: enables a contribution of a function the graph
   c of which is represented as a Bezier curve given by a list of knots
   c defined by data bezierX and bezierY
 
R ipolFac=0
   c interpolated contribution: enables a contribution to the potential
   c which is given by interpolating the (x,y)-values from lists ipolX
   c and ipolY under control of ipolOrd and ipolMet

R gaussFac=0.1
   c Gauss contribution: enables a contribution to the potential as a
   c list
   c Gaussian bell curves characterized by lists gaussX, gaussSigma,
   c and gaussY

R indFac=0

// Defining the data for the contributions mentioned above
Rs poly=0 0 1
Rs bernstein=0 -1 0 0 0 -1 0
Rs bezierX=-0.5 -0.25 0.20 0.25 0.30  0.5
Rs bezierY=0  -1 0.5 0.5 0.5 0    
Rs ipolX=-0.5   -0.25 -0.125  0     0.125  0.25    0.5
Rs ipolY=0 0.25 0.25 -0.5 0.25 0.25 0
Rs gaussX=0
Rs gaussSigma=0.05
Rs gaussY=0.05
Rs indX=0
Rs indWidth=0
Rs indY=0

Z ipolOrd=2
   c order of interpolation: e.g. 3 for cubic. The order may be much
   c smaller than the number of points; nevertheless the interpolated
   c curve will really pass through all points precisely. 
   
Z ipolMet=2
   c interpolation method: 0,2 smooth polynomial,
   c 1 smooth rational, 3 normal (non-smooth) linear.
   c smooth interpolation is a propriatory method that behaves like
   c splines but guarantees that derivatives of all orders exist
   c (and thus are continuous)

*/
} // namespace

#endif
