//? cpmrmatrix.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_RMATRIX_H_
#define CPM_RMATRIX_H_
/*

   Purpose: class of real matrices together with singular value
      decomposition

*/
#include <cpmrvector.h>
#include <cpmf.h>

namespace CpmArrays{

   using CpmFunctions::F;

//////////////////////// class R_Matrix //////////////////////////////
//
// Matrices with matrix elements of type R. This class implements the
// major algorithms associated with matrices. The most complex (and
// useful) one is forming the pseudo-inverse !a of an arbitrary
// (m,n)-matrix a. Also to be noticed: ~a is the transpose of a.
// Matrix algebra is expressed in terms of *,+,-, and even
// (a/b := a*!b, of course ). All natural ways to construct
// R_Matrices from R_Vectors are implemented. Further R_Matrix objects
// can be appended row-wise using operator & and column-wise using
// operator &&.

class R_Matrix{ // real matrices, rich interface
public:
   Vr<R_Vector> rep;
private:
   R_Matrix(const V<R_Vector>& x):rep(x){}

public:
   static R etaInv;
      // fraction of the largest 'singular value' which is considered
      // 'intentionally different from zero'. If this is negative, a
      // smooth version of the normal pseudo-inverse is chosen.
   typedef R_Matrix Type;
   typedef R ScalarType;
   CPM_IO
   CPM_ORDER
   CPM_TEST
   CPM_DESCRIPTORS
   CPM_CONJUGATION
   CPM_SUM_M
   CPM_PRODUCT_C
   CPM_SCALAR_M
   CPM_DIFFERENCE_PLAIN
   CPM_DIVISION
   CPM_DOT_PRODUCT
      // Inversion is pseudo-inversion in our case. This is always
      // defined and very useful. An internal test on correctness of
      // inversion (actually of singular value decomposition) will be
      // done if cpmverbose>=3
/*
An explanation of the concept 'pseudo inverse' follows:
If m=n and det(v)!=0 the pseudo inverse coincides with the normal
inverse of a matrix.
The pseudo inverse of a number x is in C notation given by
   (x==0 ? 0. : 1./x), (for numerical work the condition x==0 might be
      replaced by a similar one)
The pseudo inverse !diag of a diagonal matrix diag is obtained by forming
the pseudo inverse of the diagonal elements. The pseudo inverse of a
general matrix is defined via singular value decomposition (see next
function svd()).
Let
   v=iso1 * diag * (iso2~)
be the singular value decomposition of matrix v, then
   iso2 * !diag * (iso1~)   (note the interchange of 1 and 2)
is the pseudo inverse of v.
Consider the function
   f(x)=length of vector (y-Ax)
for a m-vector y, a (m,n)-matrix A, and a variable n-vector x.
This function has an absolute minimum at x0=(!A)y. Among all vectors x
satisfying
   f(x)=f(x0)
x0 is the shortest one. Thus the pseudo inverse gives the best and
the most natural resolution of the equation
   y=Ax
even if an exact solution is not existing. The matrix A determines two
projection matrices:
   P_final = projection onto the range of A
   P_initial = projection onto the complement of the null space of A.
The pseudo invers !A of A allows to write these projectors as
   A*!A=P_final,   (!A)*A=P_initial.
*/

// constructors
   R_Matrix():rep(1,R_Vector(1)){}
      // creating a (1,1) zero-matrix

   explicit R_Matrix(Z n):rep(n,R_Vector(n)){}
      // creating a (n,n) zero-matrix

   R_Matrix(Z m, Z n):rep(m,R_Vector(n)){}
       // creating a (m,n) zero-matrix

   R_Matrix(Z m, Z n, R r):rep(m,R_Vector(n,r)){}
       // creating a (m,n) matrix with all matrix elements
       // equal to r

   explicit R_Matrix(const VV<R>&);
      // evident

   explicit R_Matrix(R_Vector const& diag):
      rep(diag.dim(), R_Vector(diag.dim()))
   { for (Z i=1;i<=diag.dim();i++)  rep[i][i]=diag[i];}
      // creating a diagonal matrix from a list diag of diagonal elements

   R_Matrix(Z m, R_Vector const& row):rep(m,row){}
       // creating a (m,row.dim()) matrix made out of m row-vectors row

   R_Matrix(R_Vector const& col, Z n);
       // creating a (col.dim(),n) matrix made out of n column-vectors col
       //

   R_Matrix(R_Vector const& a, R_Vector const& b);
      // dyadic product of the two vectors a and b
      // Let mat be the constructed matrix, then mat[i][j]=a[i]*b[j]

// conversions
   VV<R> toVV()const;
      // evident

   V<R_Vector> toRep()const{ return rep;}
      // evident

// accessors
   Z dim()const{ return rep.dim();}
      //: dimension

   Z dim1()const{ return rep.dim();}
      // number of rows

   Z dim2()const{ return rep[1].dim();}
      // number of columns

   Z b()const{ return rep.b();}
      //. begin

   Z e()const{ return rep.e();}
      //. end

   R_Vector& operator[](Z i){ return rep[i];}

   const R_Vector& operator[](Z i)const{ return rep[i];}

   const R_Vector& cyc(Z i)const{ return rep.cyc(i);}
      //: cyclic
   R_Vector& cyc(Z i){ return rep.cyc(i);}
      //: cyclic

   R_Vector& cui(Z i){ return rep.cui(i);}
      //. component (with) unchecked index
   const R_Vector& cui(Z i)const{ return rep.cui(i);}
      //. component (with) unchecked index


   R_Vector getCol(Z j)const;
      // returns the vector (rep[1][j],...rep[m][j])
      // (m=dim1())

   void setCol(Z j, const R_Vector& v);
      // replaces (rep[1][j],...rep[m][j]) by
      // v[1],...v[k],rep[k+1][j],..rep[m][j], k=v.dim()
      // (in most cases k==m, the the j-th column of rep
      // gets simply replaced by v

   R_Matrix permuteColumns(const V<Z>& per);

   R_Matrix permuteRows(const V<Z>& per)
   { return R_Matrix( rep.permute(per));}

   R_Matrix ins(Z i, const R_Vector& v)const
   { return rep.ins(i,v);}
      // Let b=a.ins(i,v);
      // Then we have
      //    b[j]=a[j]   for j<i
      //    b[j]=v      for j==i
      //    b[j]=a[j-1] for j>i
      // see also Vl<T>::ins(Z, const T&)

   R_Matrix ins(Z i, const R_Matrix& v)const
   {  return rep.ins(i,v.toRep());}
      // Let b=a.ins(i,v);
      // d=v.dim()
      // Then we have
      //    b[j]=a[j]   for j<i
      //    b[j+k]=v[k] for 1<=k<=d
      //    b[j]=a[j-d] for j>j+d
      // see also Vl<T>::ins(Z, const Vl<T>&)

   R_Matrix resize(Z newDim)const
   { return R_Matrix(rep.resize(newDim));}

   R_Matrix resize2(Z newDim2)const
   { return con().resize(newDim2).con();}
      // resizing the range of the second index
      // Could be made twice as fast by a independent implementation.
      // Here I prefer the principle to have each concept in one place.

   R_Matrix eliFirst(Z nDel=1)const
   { return R_Matrix(rep.eliFirst(nDel));}
      // returs a matrix with the first line eliminated from *this

   R_Matrix eliFirst2(Z nDel=1)const
   { return con().eliFirst(nDel).con();}
      // returs a matrix with the first column eliminated from *this

   R_Matrix eli1(IvZ const& iv)const
   { return R_Matrix(rep.eli(iv));}
      // Returns a matrix with the lines indexes by iv eliminated from
      // *this

   R_Matrix eli2(IvZ const& iv)const
   { return con().eli1(iv).con();}
     // Returs a matrix with the columns indexes by iv eliminated from
     // *this

   R_Matrix operator()(F<R,R> const& g)const;
      // returns a matrix obtained by applying function g
      // to each component of *this

   R_Matrix log10()const;
      // previous function with g=log10

   R_Vector operator * (const R_Vector&)const;
      // natural action of a matrix on a vector

   R2 centerOfMass()const;
      // interpreting the matrix components as a mass distribution
      // we return the center of mass as a pair (i,j) of
      // (fractional) pixel coordinates of the center om mass.
      // Since the components of the matrix may be mixed
      // positiv and negativ it is more general to think about
      // a center of electric charge.

   R_Vector linForm()const;
      //: linear form
      // rep[1]&rep[2]&...&rep[n], n=rep.dim(). This allows
      // carrying over from R_Vector many useful descriptors:

   R_Vector dig()const;
      //: diagonal
      // Returns the diagonal as a vector. No restrictions
      // concerning dim1() and dim2().

   R inf(Z i=0)const{ return linForm().inf(i);}
      // returns the 'ith-smallest' component. No ordering
      // for i=0. For i>0 it is the component indexed 1+i
      // of the increasingly ordered version of rep

   R sup(Z i=0)const{ return linForm().sup(i);}
      // returns the 'ith-largest' largest component.
      // For i=0 fast operation, no ordering
      // For i>0 it is the component indexed rep.dim()-i
      // of the increasingly ordered version of rep

   R supAbs()const;
      // returns sup{ rep[i].supAbs()} (i.e. the largest
      // absolute value of a component, sometimes called max norm)

   R sum()const{ return linForm().sum();}
      // returns the sum of all components

   R mean()const { return linForm().mean();}
      // arithmetic mean value of all components, fast

   R med()const { return linForm().med();}
      // median value of all components

   Iv infSup()const{ return linForm().infSup();}
    // returns the smallest and the largest component. Fast operation,
      // no ordering

   R3 infMeanSup()const{ return linForm().infMeanSup();}
   // returns the smallest component, the arithmetic mean value,
   // and the largest component of *this. Fast operation, no ordering

   R3 infMedSup()const{ return linForm().infMedSup();}
   // returns the smallest component, the median value,
   // and the largest component of *this. Achieved by ordering the
   // components. If one should also need the mean value, don't call
   // the previous function in addition, but simply the cheap function
   // mean()

   R_Matrix confine(Z2 const& z2)const{
      Z mL=4;
      Word loc("R_Matrix confine(Z2 const& z2)const");
      CPM_MA
      Z n=dim1();
      R_Matrix res(n);
      for (Z i=1;i<=n;++i){
         res[i]=rep[i].confine(z2);
      }
      CPM_MZ
      return res;
   }

   X2<R_Matrix,Iv> adjustRange(const Iv& ivRange)const;
      // returnd is a matrix, the components of which have been
      // forced into ivRange by a common linear transformation
      // of the components of *this. The original value range
      // is also returned, since it had to be determined anyway.

// concatenation functions as in R_Vector (with the natural replacement
// R_Vector |--> R_Matrix, R |--> R_Vector)

   R_Matrix& operator &=(const R_Matrix& x1)
   { rep&=x1.rep; return *this;}
   // mutating form of concatenation

   R_Matrix& operator &=(R_Vector r)
   { rep&=r; return *this;}

   R_Matrix append2(R_Matrix const& x1)const;
      // concatenation 'with respect to the second index'

   R_Matrix operator && (R_Matrix const& x1)const
   { return append2(x1);}
      // concatenation 'with respect to the second index'

   R_Matrix operator & (const R_Matrix& x1)const
   { return R_Matrix(rep&x1.rep);}
   // concatenation

   R_Matrix operator & (const R_Vector& r)const
   { return R_Matrix(rep&r);}

   R_Matrix prepend(const R_Vector& r)const
   { return R_Matrix(rep.prepend(r));}

// Connection with functions
   F<R2,R> toF(Iv const& iR=Iv(0,1),Iv const& jR=Iv(0,1) )const;
   //: to F
   // Returns a function defined on the rectangle iR x jR which is
   // an area analog to the discrete lattice of pixel centers.
   // Direction of the continuous axes follows the direction of
   // increasing matrix indexes.

// read functions
   bool scanFromFreeFor(istream&);
      // : scan from free format
      // Reads numbers from a stream as long as this can successfully
      // be done and changes *this into the (m,n) R_Matrix having all
      // these numbers, grouped into arrays of m items, as columns.
      // Thus n has not to be known in advance, wheras m=dim() is given
      // by *this.
   bool scanLineWise(istream&);
      //: scan line wise
      // Reads numbers from a 'line-oriented' stream (so far I read always
      // from files the content of which was divided into lines, since the
      // 'newline character' '\n' was used in writing them. The numbers
      // found in the first line define the number of columns of the
      // R_Matrix *this after execution of this function. Futher lines are
      // only accepted until he point where we find a line which has a
      // number of entries wich differs from the number of entries of the
      // first line. Finally *this consists of all these accepted lines
      // (rows) of R-numbers.
// formatted read and write for easy communication with
// textfiles of third parties
   bool wrtFor(ostream& str)const;
      // write formatted
   bool readFor_(istream& str);
      // read formatted
   bool wrtForTrn(ostream& str)const;
      // write formatted transposed
   bool readForTrn_(istream& str);
   // read formatted transposed

   bool toGnuPlot(Word const& fileName)const;
   //: to gnu-plot
   // create an outputfile with name 'fileName' and write into it
   // the matrix elements of *this in a manner that the file is suitable
   // as a data input file for making gnuplot graphics with them.
   // Similar to function
   // CpmArrays::toGnuPlot(Word const& fn, V<V<R2> > const& x ).

   bool toGnuPlot(ostream& str)const;
   //: to gnu-plot
   // Works as previous function, except that it writes
   // to a given stream instead of creating a file.

   static void svdcmp(R_Matrix& a, R_Vector& w, R_Matrix& v);
   // singuar value decomposition
   // Implementation tool that now is made public in order to
   // enable a short program that can directly be compared with
   // a Ruby program.
};

inline bool R_Matrix::prnOn(ostream& out)const
{ return rep.prnOn(out);}

inline bool R_Matrix::scanFrom(istream& in)
{ return rep.scanFrom(in);}

inline Z R_Matrix::com(const R_Matrix& a)const
{ return rep.com(a.rep);}

inline R_Matrix& R_Matrix::operator *=(const R& r)
{ for (Z i=1;i<=dim(); i++) rep[i]*=r;  return *this;}

inline R_Matrix R_Matrix::operator -()const
{ return R_Matrix(-rep);}

inline R_Matrix R_Matrix::ran(Z i)const
{ return R_Matrix(rep.ran(i));}

inline R_Matrix R_Matrix::test(Z i)const
{ return R_Matrix(rep.test(i));}

inline Word R_Matrix::nameOf()const
{ return Word("R_Matrix");}

inline Word R_Matrix::toWord()const
{ return rep.toWord();}

inline R R_Matrix::operator|(R_Matrix const& a)const
{ return (linForm()|a.linForm());}

/////////////////////// class SVD ///////////////////////////////////////
//
// Singular Value Decomposition and functional calculus. Let a be an
// arbitrary matrix and f:R->R an arbitrary function. Then one defines
// f(a) by applying f to the singular values. This is an extension of the
// functional calculus based on a spectral representation of a matrix.
// In our setting this f(a) is obtained in the following way:
// F<R,R> f=...;
// R_Matrix a=...;
// SVD sa(a);  // does the work
// R_Matrix fa=sa(f);  // relatively easy job.
//
// If we have to vary f for fixed a, we have to do the pre-computations
// encapsulated in the SVD constructor only once. We would not get this
// benefit if we would implement the functional calculus by a member
// function of R_Matrix.

/*
Let A be a real (m,n)-matrix. Let d=MIN(m,n). A singular value
decomposition of A is a representation
   A=iso1 * diag * (iso2~)
where
   iso1 is an isometric (m,d)-matrix
   diag is a diagonal   (d,d)-matrix
   iso2 is an isometric (n,d)-matrix.
(A (i,j)-matrix m is called isometric if m~ * m = (j,j)-unit matrix,
where m~ denotes the transposed version of m. If i=j, it also satisfies
m * m~ = unit matrix (not trivial !), then it is called orthogonal.)
The diagonal components of diag will be held as a vector w.
The singular value decomposition is not unique. If all singular
values are different (most natural case) the only degree of freedom is
in the order of the singular values and in 'phasefactors' +/-1.
Here the singular values are ordered in descending order of their
absolute value.
*/

class SVD{
// singular value decomposition of real (m,n)-matrices

   Z m,n,d;
      // as in previous explanation
   R_Matrix A;
      // as in previous explanation
   R_Matrix U;
      // iso1 of previous explanation
   R_Matrix V;
      // iso2
   R_Matrix Vt;
      // iso2~
   R_Vector w;
      // diag=R_Matrix(w)
public:
   explicit SVD(const R_Matrix& a);
      // creates a instance of SVD which provides all quantities which
      // result from a singular value decomposition of a. If such an
      // instance is created once, several 'services' like finding f(a)
      // for arbitrary f:R->R can be provided without much computational
      // burden.
   R_Matrix operator()(const F<R,R>& f)const;
      // returns f(A)
// simple access functions
   R_Matrix getU()const{ return U;}
      // isometric leading factor of singular value decomosition
      // into three factors
   R_Matrix getW()const{ return R_Matrix(w);}
      // singular values as a diagonal matrix central factor of singular
      // value decomosition into three factors
   R_Vector get_w()const{ return w;}
      // list of singular values
   R_Matrix getV()const{ return V;}
      // isometric last 'factor' (actually Vt is the factor, Vt has not
      // to be isometric, however)
   R_Matrix getVt()const{ return Vt;}
      // last factor of singular value decomposition
   Z get_m()const{ return m;}
   Z get_n()const{ return n;}
   Z get_d()const{ return d;}
      // number of singular values
   R getError()const;
      // defined such that the result should deviate from 0 only by
      // numerical inaccuracies. Any deviation from the defining
      // properties of a singular value decomposition will be detected.
};

} // namespace

#endif
