//? cpms.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_S_H_
#define CPM_S_H_
/*

   Description: Defines a template class S<T> of sets
       the elements of which are of type T.
      Intented as a basis for a mathematics style definition of
           associative lists.

*/

#include <cpmv.h>
#include <cpmtypes.h>
#include <set>

////////////////////////// class S<> ///////////////////////////////////

namespace CpmArrays{

   using CpmFunctions::F;
   using CpmRoot::Word;

template <class T>
class S{ // sets of homotypic elements
   // Represents sets, all the elements of which are of type T.
   // It is assumed that T<T and T>T is defined.

public:

   std::set<T> p_;

   typedef S<T> Type;

//public:

   CPM_IO_V
      // CPM_ORDER comes later in this class declaration/definition
   CPM_ORDER

// constructors

   S():p_{}{}
      // constructor for the void set

   S(std::set<T> const& p):p_{p}{}

   void merge(V<Type> const& setList){
      for (Z i=setList.b();i<=setList.e();++i) *this << setList[i];
   }
      // changes *this into the union of *this and the sets in the list

   explicit S(V<T> const& vec){
      S<T> s;
      for (Z i=vec.b();i<=vec.e();++i) s << vec[i];
      *this = s;
   }

      // constructor from an array of T-values (V<T> sufficient, not
      // needing Vo<T>; T, however, has to support ordering (which is
      // clear if S<T> is under consideration). Clearly, multiple
      // appearing of components in vec does not prevent their
      // single appearence in the result (which, otherwise, would not be
      // a set)

    explicit S(std::initializer_list<T> il ){ *this = S<T>(V<T>(il));}

// cardinality access

   Z car()const{ return p_.size();}
      //: cardinality
   Z dim()const{ return p_.size();} // for uniformity with other array types
      //: dimension
   Z size()const{ return p_.size();} // for uniformity with STL code
      //: size

   auto begin(){ return p_.begin();}
      // allows to write code such as
      // S<Z> s{-1,1,-2,2,-3,3};
      // for (auto const& x : s) cout<<x<<endl;
      // for 'looping over the range of s'.
      // also possible
      // for (auto q=s.begin(); q!=s.end();++q){ cout<<*q<<endl;}

   auto end(){ return p_.end();}
   auto cbegin(){ return p_.cbegin();}
   auto cend(){ return p_.cend();}

// set formation operations
   bool addAct_(T const& t){
      auto q=p_.insert(t);
      return q.second;
   }
      // changes *this into the union of (*this) and {t}
      // and returns true if t was not already there
      // so that it had to be added 'actually'

   void operator<<(T const& t){
      p_.insert(t);
   }

   void add_(T const& t){ p_.insert(t);}
      // changes *this into the union of (*this) and {t}
      // return value of addAct not used

   void add_(Type const& s){ for (auto const& x : s.p_){ p_.insert(x);} }
      // changes *this into the union of (*this) and s

   void operator<<(Type const& s){
      for (auto const& x : s.p_){ p_.insert(x);}
   }
   // same as add_(Type const& s)

   bool rem_(T const&  t){
      auto q=p_.find(t);
      if (q!=p_.end()){
         p_.erase(q);
         return true;
      }
      else return false;
   }
      //: remove
      // *this will be changed into *this\{t}
      // If this reduces the cardinality of *this, the return
      // value is 'true' and 'false' else

   bool remove(T const&  t){ return rem_(t);} // heritage

   Type operator |(Type const&  s)const
      // forming the union of *this and s (corresponds to the
      // 'or' operation for indicator functions)
   {
      S<T> res(p_);
      for (auto const& x : s.p_){ res<<x;}
      return res;
   }

   Type operator &(Type const&  s)const
      // forming the section of *this and s (corresponds to the
      // 'and' operation for indicator functions)
   {
      S<T> res;
      for (auto const& x : s.p_){
         if (hasElm(x)){ res<<x;}
      }
      return res;
   }

   Type operator -(Type const&  s)const
      // forming *this\s i.e. the set of all those elements of
      // *this which don't belong to s (relative complement)
   {
      S<T> res(p_);
      for (auto const& x : s.p_){
         if (hasElm(x)){ res.rem_(x);}
      }
      return res;
   }

   Type operator ^(Type const&  s)const{ return (*this-s)|(s-*this);}
      // forming the symmetric difference of s1 and s2
      // (corresponds to 'exclusive or' for indicator functions)

   T const&  operator[](Z i)const{ return toV()[i];}
      // Means of iterating over all elements of the set
      // The specific indexing is gouverned by the < order.
      // If i is 'out of range' i.e. not 1<=i<=card(), this is
      // considered an error which is handled by Vo<T> (finally
      // by V<T>).

   T const&  fir()const{ return *(p_.begin());}
      //: first
      // first element of *this according to the ordering on
      // which the construction of Type is based.
      // causes error if *this is void

   T const&  last()const{ return *(p_.end()--);}
      // last element of *this according to the ordering on
      // which the construction of Type is based.
      // causes error if *this is void

// boolean operations

   bool isVoid(void)const{ return p_.empty();}

   bool isEmpty(void)const{ return p_.empty();}

   bool hasElm(T const& t)const{
      return p_.contains(t);
   }
      //: has element
      // returns true iff t is an element of *this

   bool ni(T const&  t)const{ return hasElm(t);}
      // 'in' reversed, from LATEX symbol \ni for the mirror image of
      // the epsilon-like 'is element'-symbol
      // Notice that the 'in' function would have as first argument a T
      // and as second argument a Type. This cannot be made a member
      // function. Introducing a non-member friend function would offend
      // the C+- coding style.

   void select(V<B> const& sel){
      V<T> temp=toV();
      temp=temp.select(sel);
      *this=S<T>(temp);
   }
      // eliminates from *this all components x[i] for which sel[i]
      // is defined and statisfies sel[i]==false.
      // (see corresonding (non-mutating) function of V<T>)

// relations to other classes:

   V<T> toV(void)const{
      std::vector<T> p;
      for (auto const& x : p_) p.push_back(x);
    //  std::sort(p.begin(),p.end());
      V<T> res(1,p);
      return res;
    }
      // getting the representing vector as a new entity. It may be
      // useful to manipulate this vector by, for instance,
      // the V<T>::condense() function and then form a new set via the
      // constructor from vectors

// Infrastructure functions as declaration macros  known from including
// <cpmvo.h>

      // since T has to allow ordering, we implement the order
      // operations. This allows to form S< Type >.

   virtual S<T>* clone()const{ return new S(*this);}

   Word toWord()const;

   virtual Word nameOf()const{
      Word wi="S< ";
      Word wt=CpmRoot::Name<T>()(T());
      return wi&wt&" >";
   }
};

template <class T>
Z S<T>::com(Type const& s)const{ return toV().com(s.toV());}

template <class T>
bool S<T>::prnOn(ostream& str)const
{
   Z mL=3;
   static Word loc("S<>::prnOn(ostream&)");
   CPM_MA
   cpmwbt;
   bool b;
   for (auto const& x : p_){
      b=CpmRoot::IO<T>().o(x,str);
      cpmassert(b==true,loc);
      if (b==false){
         cpmdebug(x);
         CPM_MZ
         return false;
      }
   }
   cpmwet;
   CPM_MZ
   return true;
}

template <class T>
bool S<T>::scanFrom(istream& str)
{
   Z mL=3;
   static Word loc("S<>::scanFrom(istream&)");
   CPM_MA
   S<T> res;
   T t;
   while (CpmRoot::IO<T>().i(t,str)){ res.add_(t);}
   *this=res;
   CPM_MZ
   return true;
}

} // CpmArrays

#endif
