//? cpmsmoothlocalinterpol.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_SMOOTHLOCALINTERPOL_H_
#define CPM_SMOOTHLOCALINTERPOL_H_
/*
   Description : Smooth interpolation. Gives a smooth interpolating
        curve for any finite list of prescribed real values.
        The list of x-values has to be in ascending order but needs not
        to be aequidistant. (Smooth means that derivatives of any order
        exist.) For details, see the comment to function sli.

*/
#include <cpmtypes.h>
#include <cpmv.h>

namespace CpmArrays{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;

//////////////////// class IpolControl ////////////////////////////

class IpolControl{ // interpolation control
// save arrangement of the parameters which control interpolation
// The meaning of parameters m and method is as in function slint()
// 'smooth local interpolation'(see Kodak-TR 278844A) Here is the
// explanation:
// m is the order of the interpolation method (m=1,2,..., m<=n, n= number
// of data points) i.e. the number of data points entering the basic
// local interpolation function (m=3 for quadratic interpolation).
// method is 0 or 2 for polynomial interpolation and 1 for Bulirsch Stoer
// rational diagonal interpolation.
// The number of points used in interpolation is m if x is
// more than m/2 points away from the boundary. If method=2, things are
// arranged such, that coming closer to the boundary reduces the number
// of points which are used. This is not only useful for interpolation;
// more important is the improvement in extrapolation properties.
// For universal usage there is included normal (not smooth)
// linear interpolation as method=3.
// Notice that such a expensive operation as testing the (potentially
// very long) list of x-values should not be done with each computation
// of an interpolated value. So, the two data monoTest and eps will come
// into action not before interpolation gets embeded in a class (which
// will be R_func) where preparatory activities can be delegated to
// a constructor which has to be called only once for a certain set of
// data.
   Z m{0};
   Z method{0};
// the next two data will be used in R_Func
   Z monoTest{0};
   R eps{0.};
   void adjust();
   typedef IpolControl Type;
public:
   CPM_IO
   Word nameOf()const{ return "IpolControl";}

   IpolControl(){}

   explicit IpolControl(Z mI, Z methodI=0, Z monoTestI=0):
   m(mI),method(methodI),monoTest(monoTestI),eps(1e-6)
   { adjust();}

   explicit IpolControl(const Word& w);
      // More convenient constructor for normal applications.
      // w=polygon,first_order,second_order,third_order,fourth_order,
      // fifth_order, sixth_order.
      // This is broken linear (non-smooth) for w="polygon", and smooth
      // interpolation (method=0) of respective order else.
      // Monotony test of x-list is enabled since reading from file or
      // careless programming may easily create lists that fail to
      // be monotone.

   IpolControl(Word idle, Z code)
      // numerical encoding of the previous argument w for convenience
   {
      switch(code){
         case 1: *this=IpolControl("polygon"); break;
         case 2: *this=IpolControl("second_order"); break;
         case 3: *this=IpolControl("third_order"); break;
         case 4: *this=IpolControl("fourth_order"); break;
         case 5: *this=IpolControl("fifth_order"); break;
         case 6: *this=IpolControl("sixth_order"); break;
         default: cpmerror("IpolControl(Word,Z): unvalid Z-argument");
      }
   }

   Z get_m()const{return m;}
   Z get_method()const{return method;}
   Z get_monoTest()const{return monoTest;}
   void set_monoTest(bool b){ monoTest=(b==true ? 1:0);}
   R get_eps()const{ return eps;}
   void set_eps(R epsIn){ eps=epsIn;}
};

R locIpol(R x, const V<R>& xList, const V<R>& yList,
   const IpolControl& c=IpolControl(4));
   // 'modern form with class for controlling parameters

R locIpol(R ic, V<R> const& yList, IpolControl const& c=IpolControl(4));
   // local interpolation
   // Here ic is a 'continuous index'. The interpolated function f
   // has the property f(x)=yList[(Z)x] if x is integer (by value,
   // not by type).
   // This is here for uniformity with
   // R interpolate(R ic, R_Vector const& x) in cpmvectormore.h.
   // If this function has to be called many times for the same argument
   // yList, one should consider to create xList once ( xList[i]=i)
   // and to call the previous function.

R locIpolPos(R x, const V<R>& xList, const V<R>& yList,
   const IpolControl& c=IpolControl(4));

void prepIpol(V<R>& x, V<R>& y, R eps=1e-6);
// name stands for 'prepare Interpolation'
// Tests whether the list of x's is stricly increasing as expected for
// interpolation purposes. If two x values are very close together they
// get merged and the y-value replaced by the mean

} // namespace

#endif
