//? cpmsystem.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <sys/time.h> // <time.h> not sufficient
#include <iomanip>

#include <cpmgreg.h>
#include <cpmsystem.h>
#include <cpmsystemdependencies.h>

#if !defined(CPM_NOGRAPHICS)
   #include <cpmviewport.h>
#endif

using namespace CpmStd;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::B;
using CpmRoot::Word;
using CpmRoot::toDouble;
using CpmArrays::V;

///////////////////////// class B /////////////////////////////////////////

bool B::prnOn(ostream& out)const
{
   return CpmRoot::write(x1,out);
}

bool B::scanFrom(istream& in)
{
   return CpmRoot::read(x1,in);
}

Z B::com(const B& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   return 0;
}

#if defined(CPM_WRITE_TITLE)
   B CpmRoot::wrtTit{true};
#else
   B CpmRoot::wrtTit{false};
#endif

B CpmRoot::operator""_B(unsigned long long int n)
{ bool b=(bool)n; return B(b);}

//////////////////// class Message ////////////////////////////////////////

namespace{
   const char sep='/';
   const char dot='.';

   bool startDot(Word const& file)
   {
      if (file.dim()==0) return false;
      return file[1]==dot;
   }

   bool absPath(Word const& file)
   // if file start with /home/... it is an absolute path
   // and should not be augmented by a leading dot
   {
      if (file.dim()<6) return false;
      return file[1]==sep && file[2]=='h' &&
      file[3]=='o' && file[4]=='m' && file[5]=='e'
      && file[6]==sep;
   }

   bool startSep(Word const& file)
   {
      if (file.dim()==0) return false;
      return file[1]==sep && !absPath(file);
   }
}


Word CpmSystem::glue(Word const& path, Word const& file)
{
// Do not insert diagnostic messages, like cpmdebug(temp); into
// this function block, since the function is used in Message::ini()
// before the messaging file cpmcerr exists.
   if (file.dim()==0) return path;
   if (path.dim()==0) return file;
   Word res= startSep(file) ? path&file : path&Word(sep)&file;
   if (startSep(res)) res=Word(dot)&res;
   return res;
}

////////////////////// class Message /////////////////////////////////////
// initialization by code
Z CpmSystem::Message::verbose=2;
Z CpmSystem::Message::debug=2;
Z CpmSystem::Message::trigger=0;
Z CpmSystem::Message::maxNumMes=100000;
Word CpmSystem::Message::sttmes1="*";
Word CpmSystem::Message::sttmes2="**";
Word CpmSystem::Message::sttmes3="***";
Word CpmSystem::Message::sttmes4="****";
Word CpmSystem::Message::runId="";
ofstream CpmSystem::Message::out;
ofstream CpmSystem::Message::outData;
B CpmSystem::Message::appRunId{true};

//initialization from cpmsystemdependencies.h
#if defined(CPM_RUN_ID_LENGTH)
   Z CpmSystem::Message::runIdLength=CPM_RUN_ID_LENGTH;
#else
   Z CpmSystem::Message::runIdLength=4;
#endif

#if defined(CPM_OUT_DIR) // for backward compatibility
   Word CpmSystem::Message::outDir1=CPM_OUT_DIR;
         // setting the directory location of cmcerr and cpmdata
   Word CpmSystem::Message::outDir2=CPM_OUT_DIR;
#else
   Word CpmSystem::Message::outDir1=CPM_REL_POS2; // old style
   Word CpmSystem::Message::outDir2=CPM_REL_POS2;
#endif
Word CpmSystem::Message::relPos1=CPM_REL_POS1;
Word CpmSystem::Message::relPos2=CPM_REL_POS2;
Word CpmSystem::Message::relPos3=CPM_REL_POS3;
Word CpmSystem::Message::relPos4=CPM_REL_POS4;
Word CpmSystem::Message::inpDir=CPM_INP_DIR;
Word CpmSystem::Message::fontDir=CPM_FONT_DIR;

#if defined(CPM_NOLOG)
   bool CpmSystem::Message::streamInitialized=true;
   B CpmSystem::Message::silent{true};
#else
   bool CpmSystem::Message::streamInitialized=false;
   B CpmSystem::Message::silent{false};
#endif
Z CpmSystem::Message::sttBarWid1=CPM_PANE1;
Z CpmSystem::Message::sttBarWid2=CPM_PANE2;
Z CpmSystem::Message::sttBarWid3=CPM_PANE3;
Z CpmSystem::Message::sttBarWid4=CPM_PANE4;

Z CpmSystem::Message::vph_=CPM_HEIGHT;
Z CpmSystem::Message::vpw_=CPM_WIDTH;
Z CpmSystem::Message::fsh_=CPM_HEIGHT;
Z CpmSystem::Message::fsw_=CPM_WIDTH;

// functions

void CpmSystem::Message::ini(void)
{
   if (initialized()) return;
   cout<<endl<<"Message::ini(void) started"<<endl;
   std::filesystem::path wd_(std::filesystem::current_path());
   cout<<"Current path (working directory) is "<<wd_<<endl;
   Word loc("CpmSystem::Message::ini(void)");
   Word logFile("cpmcerr");
   Word datFile("cpmdata");
   CpmTime::Greg date;
   date.now();
   runId=date.getCode().tail(runIdLength);
      // runId will no longer be used to create an name appendix for
      // cpmcerr and cpmdata, since 'personalized' versions will generated
      // in the same directory as the other output by function
      // CpmSystem::docRun(bool). runId can be changed by getting a new
      // value for runIdLength from cpmconfig.ini. In the C+- framework
      // the run Id gets used only after the deployment of cpmconfig.ini
      // (if there is no such file, of course the presently generated
      // runId will be employed.
   logFile&=".txt";
   datFile&=".txt";
// the folowing two lines were added 2014-01-19 after a new installation of
// Ubuntu as a conseqence of which the log files according to the unchanged
// logic were created always directly in the user directory
// (/home/mutze in my case).
   logFile=outDir1&"/"&logFile;
   datFile=outDir1&"/"&datFile;
   cout<<endl<<" trying to open "<<-logFile<<endl;
   out.open(-logFile, std::ios_base::out);
   bool streamInitialized1=true;
   if (!out) {
      streamInitialized1=false;
      cout<<endl<<"stream out is not valid"<<endl;
   }
   else cout<<endl<<"stream out is valid"<<endl;
   cout<<endl<<" trying to open "<<-datFile<<endl;
   outData.open(-datFile, std::ios_base::out);
   bool streamInitialized2=true;
   if (!outData){
      streamInitialized2=false;
      cout<<endl<<"stream outData is not valid"<<endl;
   }
   else cout<<endl<<"stream outData is valid"<<endl;
   if (streamInitialized1&&streamInitialized2){
      streamInitialized=true;
      date.writeFormatted(out);
      Word mes=loc&
         ": files "&logFile&" and "&datFile&" successfully created,\n";
      cout<<endl<<-mes<<endl;
      cpmmessage(mes); // works
      // The files get created in the executable's working directory.
      // In MS Visual C++ the executable is automatically put (can be
      // changed in the 'project properties') into a Release or Debug
      // subdirectory of the project directory. cpmcerr.txt and
      // cpmdata.txt then get created just outside of these
      // subdirectories and thus directly in the project directory.
   }
   else{ // then we had trouble in creating the message streams
      if (streamInitialized1){ // since we really need only logFile, this
      // might be a situation not asking for stopping the program
         Word mes1=loc&": not able to create "&datFile;
         cout<<endl<<-mes1<<endl;
         cpmmessage(mes1);

      }
      else{
         Word mes2=loc&": not able to create "&logFile&
            ", stopping program";
         cout<<endl<<mes2<<endl;
         cpmmessage(mes2);
         throw;
            // something went fundamentally wrong, no message can be sent
      }
   }
 //  iniMPREAL(); // only active if R is implemented via mpreal.h
   cout<<endl<<"Message::ini(void) done"<<endl;
}

CpmSystem::Message::Message(void)
{
   ini();
}

CpmSystem::Message::Message(Word const& w)
{
   ini();
   if (verbose>0) out<<endl<<w<<endl;
   std::cout<<endl<<w<<endl;
  //  if (!out) cpmerror("CpmSystem::Message::Message(...): out==0");
}

#if !defined(CPM_NOGRAPHICS)
void CpmSystem::Message::onStatusBar(Word const& w, Z i)
{
   static Z firstrun_=1;
   if (i<=0) return; // was i<0 till 2006-06-30
   //in this case we can't write on the status bar before a graphical
   // window is available
   bool vii=CpmGraphics::Viewport::isInitialized();
   if (firstrun_ && vii){
      firstrun_=0;
      for (Z j=1; j<=4;j++) {
         CpmGraphics::Viewport::onStatusBar(" pane "&cpm(j)&" active", j);
      }
      cpmwait(0.5);
   }
   if (vii) CpmGraphics::Viewport::onStatusBar(w, i);
}
#else
   void CpmSystem::Message::onStatusBar(Word const& w, Z i){}
#endif

void CpmSystem::Message::communication(Word const& classText,
   Word const& commType, Word const& commText, bool fileAlways,
   Z barSeg)
// common building block for messages and warnings, not for errors
// Dependence on preprocessor directives:
// _CONSOLE: everything written on cpmcerr.txt will also be written on
// cout
{
   static R tRetard=0.;
   if (silent) return; // this allows us to use C+- functions which
      // contain cpmmessage or cpmwarning statements in situations
      // in which creation of the log streams did not yet happen.
   static Z messageCounter=0;
   ini(); // esuring that 'out' can be used, no work load if already
   // initialized. For silent==true we never come here by message
   // calling and thus we are sure never to create the log files
   // as long as silent==true. If silent gets switched to false
   // (by code or whatever, next call to the present function will
   // call ini() and thus create the log files.
   Z eowMem=Word::getEndOfWord();
   Word::setEndOfWord(0);
      // then writing Words to a stream has the same effect as writing
      // string literals
   if (verbose){ // only then, something has to be done
      messageCounter++;  // it is useful to count the messages
         // even if they are not restricted in number
      ostringstream ostStatus;
      ostStatus<<classText<<" "<<commType<<" "<<commText<<
      " # "<<messageCounter;
         // status bar should be written also if number of filed messages
         // is limited
      if (barSeg>0){
         onStatusBar(Word(ostStatus.str()),barSeg);
         if (messageCounter>110) cpmwait(tRetard,2);
      }
      ostringstream ostFile;
      if (fileAlways || messageCounter<maxNumMes){
         ostFile<<endl<<"counter = "<<messageCounter<<endl;
         ostFile<<classText<<" "<<commType<<" "<<commText<<endl;
         R t=time(); ostFile<<"  "<<cpmtod(t)<<" s after program start";
         out<<ostFile.str()<<endl<<std::flush;
#if defined(_CONSOLE)
         //if (barSeg<0) std::cout<<ostFile.str()<<endl;
         std::cout<<ostFile.str()<<endl;
#endif
#if defined(CPM_NOGRAPHICS)&&defined(_CONSOLE)
         if (barSeg>0) std::cout<<ostFile.str()<<endl;
#endif
      }
      if (messageCounter==maxNumMes){
         ostFile<<endl<<
         "No more normal messages or warnings will be filed";
         out<<ostFile.str()<<endl;
#if defined(CPM_NOGRAPHICS)&&defined(_CONSOLE)
         std::cout<<ostFile.str()<<endl;
#elif defined(_CONSOLE)
         if (barSeg<=0) std::cout<<ostFile.str()<<endl;
#endif
      }
   }
   Word::setEndOfWord(eowMem);
      // restoring the writing behavior of Word
}

void CpmSystem::Message::warning(Word const& errorText, bool fileAlways)
{
   static R waitTime=0.4;
   waitTime*=0.9;
      // if there are many warnings the run should not get slowed down
      // considerably
   communication("","Warning:",errorText, fileAlways);
   if (cpmverbose>=2) cpmwait(waitTime);
}

void CpmSystem::Message::message(Word const& text, Z barSeg,
   bool fileAlways)
{
   communication("","Message:",text,fileAlways,barSeg);
}

void CpmSystem::Message::message(Z mL, Word const& errorText,
   Z barSeg, bool fileAlways)
{
   if (cpmverbose>=mL)
   communication("","Message:",errorText, fileAlways,barSeg);
}

void CpmSystem::Message::warning(Word const& className,
   Word const& errorText, bool fileAlways)
{
   communication(className,"Warning:",errorText, fileAlways);
}

void CpmSystem::Message::message(Word const& className,
   Word const& errorText, bool fileAlways)
{
   communication(className,"Message:",errorText, fileAlways);
}

void CpmSystem::Message::progress(Word const& taskName, R frac, Z seg,
   Z field)
// new implementation (2005-04-17) asks for graphical action only if
// needed.
// This assumes that a specific segment on which progress display is
// active will not have to insert messages of different character.
// Actually the static memory elements should be associated with the
// segments directly and not only within a specific function such as
// 'progress'.
{
   if (silent) return;
   static Word mem1;
   static Word mem2;
   static Word mem3;
   static Word mem4;
   Word act1,act2,act3,act4;
   R fac2=1000;
  // R_ frac2=cpmtod(frac);
   if (field==3) ;
   else if (field==1) fac2=10;
   else if (field==2) fac2=100;
   else if (field==4) fac2=10000;
   else if (field==5) fac2=100000;
   else field=3;
   bool write=false;
   Word act=cpm(cpmtoz(frac*fac2),field);
   if (seg==1){
      if (act!=mem1){
         mem1=act;
         write=true;
      }
   }
   else if (seg==2){
      if (act!=mem2){
         mem2=act;
         write=true;
      }
   }
   else if (seg==3){
      if (act!=mem3){
         mem3=act;
         write=true;
      }
   }
   else if (seg==4){
      if (act!=mem4){
         mem4=act;
         write=true;
      }
   }
   if (write){ // graphical action only if needed
      write=false;
      Word mes=(verbose && seg==1) ? "Task: "&taskName&": Progress " :
         taskName&": ";
      mes&=act;
      if (verbose>2){
         message(mes,seg); // messages go always to the status bar
      }
      else{
         onStatusBar(mes,seg);
      }
   }
}

// shared code
// prec was 4 now experimental value 8

#define CPM_SC\
   static const Z prec=5;\
   if (verbose){\
      ostringstream ost;\
      ost<<std::setprecision(prec);\
      ost<<text<<" = ";\
      ost.width(8);\
      ost<<valx;\
      if (verbose>2){\
         message(Word(ost.str()),seg);\
      }\
      else{\
        if (verbose==2) cpmcerr<<text<<" = "<<valx<<endl;\
        if (verbose>0) onStatusBar(Word(ost.str()),seg);\
      }\
   }

void CpmSystem::Message::value(Word const& text, R val, Z seg)
{
   double valx=cpmtod(val);
   CPM_SC
}

void CpmSystem::Message::value(Word const& text, Z val, Z seg)
{
   Z valx=val;
   CPM_SC
}

#undef CPM_SC

void CpmSystem::Message::values2(Word const& text, R v1, R v2, Z seg)
{
   static const Z prec=4;
   if (verbose){
      ostringstream ost;
      ost.precision(prec);
      ost<<text<<": ";
      ost.width(10);
      double v1x=cpmtod(v1);
      double v2x=cpmtod(v2);
      ost<<v1x<<", "<<v2x<<endl;
      if (verbose>2){
         message(Word(ost.str()),seg);
      }
      else{
         onStatusBar(Word(ost.str()),seg);
      }
   }
}

void CpmSystem::Message::values3(Word const& text, R v1, R v2,
   R v3, Z seg)
{
   static const Z prec=3;
   if (verbose){
      ostringstream ost;
      ost.precision(prec);
      ost<<text<<": ";
      ost.width(7);
      double v1x=cpmtod(v1);
      double v2x=cpmtod(v2);
      double v3x=cpmtod(v3);
      ost<<v1x<<", "<<v2x<<", "<<v3x<<endl;
      if (verbose>2){
         message(Word(ost.str()),seg);
      }
      else{
         onStatusBar(Word(ost.str()),seg);
      }
   }
}

void CpmSystem::Message::values4(Word const& text, R v1, R v2,
   R v3, R v4, Z seg)
{
   static const Z prec=3;
   if (verbose){
      ostringstream ost;
      ost.precision(prec);
      ost.width(7);
      double v1x=cpmtod(v1);
      double v2x=cpmtod(v2);
      double v3x=cpmtod(v3);
      double v4x=cpmtod(v4);
      ost<<text<<": "<<v1x<<", "<<v2x<<", "<<v3x<<", "<<v4x;
      if (verbose>2){
         message(Word(ost.str()),seg);
      }
      else{
         onStatusBar(Word(ost.str()),seg);
      }
   }
}

#define CPM_SC\
   if (verbose){\
      ostringstream ost;\
      ost<<text<<"="<<valx;\
      message(Word(ost.str()));\
   }

void CpmSystem::Message::report(Word const& text, Z val)
{
   Z valx=val;
   CPM_SC
}

void CpmSystem::Message::report(Word const& text, R val)
{
   double valx=cpmtod(val);
   CPM_SC
}

#undef CPM_SC
//////////////////////////// class Error ////////////////////////////////

// notice: base class constructor Message() called anyway

CpmSystem::Error::Error(Word const& messageText)
{
#if !defined(CPM_NOLOG)
   silent=false; // errors have to be reported
   verbose=1;
   communication("C+-",messageText,"CpmError",true,-1);
      // -1 lets this write also to the console
#elif defined(_CONSOLE)
   cout<<endl<<"C+- "<<messageText<<" CpmError"<<endl;
#endif
   throw;  // errors should terminate
}

CpmSystem::Error::Error(Word const& className, Word const& messageText)
{
#if !defined(CPM_NOLOG)
   silent=false;
   verbose=1;
   communication(className,messageText,"CpmError",true,-1);
      // -1 lets this write also to the console
#elif defined(_CONSOLE)
   cout<<endl<<"C+- "<<className<<" "<<messageText<<" CpmError"<<endl;
#endif
   throw;  // errors should terminate
}

////////////////// class Timer /////////////////////////////////////////

R CpmSystem::Timer::getSecondsLeft()const
   {
      static const R future=137*3600;
      if (!act) return future;
      R tAct=CpmSystem::time();
      return tEnd-tAct;
   }

R CpmSystem::Timer::getHoursLeft()const
{
   const R inv_h=1./3600.;
   return inv_h*getSecondsLeft();
}

///////////////////////// class-less functions //////////////////////////

void CpmSystem::lethalError(Word const& w)
{
   cout << "CpmSystem::letalError: " << -w << endl;
      // no service from Message needed. In particular,
      // cpmcerr.txt needs not to exist.
   exit(137);
}

void CpmSystem::error(Word const& w)
{
#if !defined(CPM_NOLOG)
   Message::ini();  // unclear whether needed
   Message::onStatusBar("Error",1);
   Message::onStatusBar(w,2);
#endif
   if (Message::debug>0) throw Error(w);
}

void CpmSystem::error(Word const& w1, Word const& w2)
{
#if !defined(CPM_NOLOG)
   Message::ini();
   Message::onStatusBar(w1,1);
   Message::onStatusBar(w2,2);
#endif
   throw Error(w1,w2);
}

R CpmSystem::time(void)
{
   static bool first=true;
   static timeval t1;
   if (first){
      gettimeofday(&t1, NULL);
      first=false;
   }
   timeval t2;
   gettimeofday(&t2, NULL);
   long sec  = t2.tv_sec  - t1.tv_sec;
   long usec = t2.tv_usec - t1.tv_usec;
   double t = sec + usec*1.e-6;
   return R(t);
}

void CpmSystem::wait(R t, Z pane)
{
   if (t<=0.) return;
   R tStart=time();
   R tEnd=tStart+t;
   R tLeft=tEnd-tStart;
   Z sLeft=cpmrnd(tLeft);
   while (tLeft>0)
   {
      tLeft=tEnd-time();
      if (pane>=1 && pane<=4){
         Z s0Left=cpmrnd(tLeft);
         if (s0Left<sLeft){
            sLeft=s0Left;
            cpmvalue(" tLeft",s0Left,pane);
               // visible only if cpmverbose>0
         }
      }
   }
   return;
}

////////////////// class IFileStream //////////////////////////
CpmSystem::IFileStream::IFileStream(Word const& fileName,
   bool safe, bool binary):name_(fileName)
{
   fs_=(binary? new ifstream(-fileName,std::ios_base::binary) :
      new ifstream(-fileName));
   if (!*fs_||fs_->bad()){  // till 2010-09-06 the test was fs==0
      Word mes="IFileStream(Word,bool,bool): unable to open "&fileName;
      if (safe) cpmerror(mes);
      else cpmwarning(mes);
   }
   else{
      cpmmessage("IFileStream(Word,bool,bool) did open "&fileName);
   }
}

bool CpmSystem::IFileStream::skipComments()
{
   return CpmRoot::eatComments(*fs_);
}

////////////////// class OFileStream //////////////////////////
CpmSystem::OFileStream::OFileStream(Word const& fileName,
   bool safe, bool binary):name_(fileName)
{
   fs_=(binary? new ofstream(-fileName,std::ios_base::binary) :
      new ofstream(-fileName));
   if (!*fs_||fs_->bad()){
      Word mes="OFileStream(Word,bool,bool): unable to open "&fileName;
      if (safe) cpmerror(mes);
      else cpmwarning(mes);
   }
   else{
      cpmmessage("OFileStream(Word,bool,bool) did open "&fileName);
   }
}

void CpmSystem::copyFile(Word const& fileNameIn, Word const& fileNameOut)
{
   cout<<"try to copy "<<-fileNameIn<<" to "<<-fileNameOut<<endl;
   std::filesystem::copy(-fileNameIn,-fileNameOut);
   cout<<" copy done"<<endl;
}

void CpmSystem::docRun(bool all)
{
#if !defined(CPM_MPI)
   cout<<"CpmSystem::docRun(bool) started"<<endl;
   if (Message::getAppRunId()==false) return;
   Word appendix = cpmrunid ;
   Word pre1=Message::getOutDir1();
   Word pre2=Message::getOutDir2();
   Word s1=pre1&"/cpmcerr.txt";
   Word t1=pre2&"/cpmcerr_"&appendix&".txt";
   Word s2=pre1&"/cpmdata.txt";
   Word t2=pre2&"/cpmdata_"&appendix&".txt";
   CpmSystem::copyFile(s1,t1);
   if (all){
      CpmSystem::copyFile(s2,t2);
   }
   cout<<"CpmSystem::docRun(bool) done"<<endl;
#endif
}
