//? cpmtypes.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_TYPES_H_
#define CPM_TYPES_H_
/*

   Description: Declares basic infrastructure functions, macros,
      and ---most importantly--- the class versions B, Z1, and R1
      of bool, Z, and R.
*/
#include <cpmsystem.h> // includes cpmwords.h and
   // thus also cpmnumbers.h
#include <cpmmacros.h>
#include <cpmx.h>

#include <cmath>

namespace CpmRootX{ // CpmRoot extended
      // A statement
      //    using namespace CpmRoot;
      // makes life much easier when working with C+-. It is therefore
      // desirable that this does not inject to many names in the users
      // scope. Thus we separate the 'true root' from an 'extended root'
      // that hosts also names such as 'order', 'inf', and 'sup' which
      // are more likely to interfere with a user's names.

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::N;
   using CpmRoot::R;
   using CpmRoot::L;
   using CpmRoot::Word;

   const R hugeNumber=1e64;
   const R tinyNumber=R(1.)/hugeNumber;

#define cpmcheck(X)\
 if (!CpmRoot::isVal(X)){\
   ostringstream ost;\
   ost<<" bad value: "<<#X "="<< X;\
   cpmmessage(Word(ost.str()));}
   // checks whether a quantity of type Z,N,R,Rh is valid
   // (and stops the program) and gives a message

#define cpmdebug(X) cpmcerr<<endl<<"C+- debug: "<<#X "="<< X <<endl
   // useful for placing temporary debugging statements into one's code.
   // From Bruce Eckel: Thinking in C++, Prentice Hall 1995, p. 325
   // typical usage:
   //
   // R_Matrix y=....;
   // cpmdebug(y);          // ';' needed !
   // The part 'debug:' helps to find the debug info in the potentially
   // very large file cpmcerr.txt

#define cpmnote(X) cpmdata<<endl<<"C+- note: "<<#X "="<< X <<endl
   // Works as cpmdebug, but but writes on cpmdata. More suitable
   // for data which are of temporary interest, so that it would
   // not pay to provide a more elaborate documentation for them.
   // Notice that cpmdata is more easily surveyed than cpmcerr.

#define cpmis(X) std::string(#X)
   // 'is' stands for 'identifier to string'.
   // This can't be written as a C or C++ function. Such a function
   // always looks on the value of a variable, wheras the present
   // construct looks on the name of the variable and returns it as a
   // quantity of type string.
   // Consider:
   //    R x=3.14;
   //    cout<<cpmis(x);
   // This will write 'x' and not '3.14'

#define cpmiw(X) Word(#X)
   // much alike previous macro

typedef void (*fpWordToVoid)(const Word&);
   // function pointer for function Word to void as a type

typedef void (*fpVoidToVoid)(void);
   // function pointer for function void to void as a type

Z sig(Z i);
R sig(R i);
   // added 03-11-05, 0 for i==0, -1 for i<0, +1 for i>0

R krn(Z i, Z j);
   // Kronecker; defined as 1. for i==j, 0. else
   // added 2004-08-26

Z parity(Z k);
/*{
   if (k%2) return -1; else return 1;
}*/

Z fac(Z n);
//: factorial
// fac(n) = 1x2x3x...x(n-1)xn , n>0 (no other values allowed here!)
// Most basic implementation.

Z pow(Z const& x, Z p);

bool isPowOfTwo(Z const& x);

Z nextPowOfTwo(Z const& x);
// We return the smallest y \in P2 such that *this <= y
// Here P2 := {0,2,4,8,16,32,....}

Z nextLog2(Z const& x);
// We return the conventional dual logarithm of
// nextPowOfTwo() ( which then is always \in N)

// Auxiliar functions.

void copyTextFile(ifstream& from, ofstream& to);
   // here, the streams from and to have to exist already

void copyTextFile(const Word& nameOfSource, const Word& nameOfCopy );
   // a file of the name nameOfSource has to exist and the copy with the
   // name indicated by the second argument will exist after the function
   // body has been executed

string toString(ifstream& from);
   // copies the content of a text file into a standard library string
   // and thus opens up rich manipulation capabilities mainly by writing
   // these to stringstreams which finally can easily be filed back. In
   // this manner e.g. concatenation of text files becomes nearly trivial

// addition 97-6-5, notice that names min and max clash with
// macro names of Windows. Thus we choose presumably save naming.
// Version with const T& arguments is not always working. Sometimes one
// gets unability to resolve function overload

///////////////////////// sup //////////////////////////////////////////
   template <class T>
   T sup(T t1, T t2)
      // sup=supremum=maximum
   {
      if (t1<t2) return t2; else return t1;
   }

   template <class T>
   T sup(T t1, T t2, T t3)
   {
      T temp;
      if (t1<t2) temp=t2; else temp=t1;
      if (temp<t3) return t3; else return temp;
   }

   template <class T>
   T sup(T t1, T t2, T t3, T t4)
   {
      T temp1=sup<T>(t1,t2);
      T temp2=sup<T>(t3,t4);
      return sup<T>(temp1,temp2);
   }

///////////////////////// inf //////////////////////////////////////////

   template <class T>
   T inf(T t1, T t2)
      // inf=infimum=minimum
   {
      if (t1<t2) return t1; else return t2;
   }

   template <class T>
   T inf(T t1, T t2, T t3)
      // inf=infimum=minimum
   {
      T temp;
      if (t1<t2) temp=t1; else temp=t2;
      if (temp<t3) return temp; else return t3;
   }

   template <class T>
   T inf(T t1, T t2, T t3, T t4)
   {
      T temp1=inf<T>(t1,t2);
      T temp2=inf<T>(t3,t4);
      return inf<T>(temp1,temp2);
   }

   //////////////////////// swap /////////////////////////////////////////

   template <class T>
   void swap(T& t1, T& t2)
   // After function call t1 has the value that t2 had before
   // and t2 has the value that t1 had before.
   // If T is a class (not built-in type) we assume that T defines
   // operator = and a copy constructor.
   {
      T temp=t2;
      t2=t1;
      t1=temp;
   }

///////////////////////// order //////////////////////////////////////////

   template <class T>
   void order(T& t1, T& t2)
      // after function call we have t1<=t2
      // If T is a class (not built-in type) we assume that T defines
      // a '<=' operation such that always at least one of t1<=t2 ,
      // t2<=t1 is valid. Also operator = and a copy constructor have to
      // be defined.
   {

      if (t1<=t2){
         return;
      }
      else{
         T temp=t2;
         t2=t1;
         t1=temp;
      }
   }

// Class equivalents of R , Z, bool.
// These can conveniently be used for class data members
// since they get initialized properly.
using CpmArrays::X2;

#if !defined(CPM_MP)
// Once I introduced an implementalion of floatingpoint reals and of
// integers as classes. This was the only way to provide for real and for
// integer data members a automatic initialization by means of the
// defined default constructor. With C++11 one can data members
// define as e.g. R x{0.}; for what was R1 x; before.
// We can now avoid the ugly mixture of R and R1 (and Z and Z1).
// Particularly inconvenient was the maintenance of R1, Z1 for the
// the multiprecision options. So the intent is to avoid any use of
// R1 and Z1 in regular C+- code. Nevertheless I conserve the definition
// of R1 and Z1 to enable experimentation, especially with the discrete
// lattice underlying the floating point numbers.

/////////////////////////////// class R1 //////////////////////////////

class R1{  // real numbers as a class
   // since there is no need for 1-tupels, we
   // define this a class endowed with infratructure similar than R2,
   // R3, ... but with arithmetics gained from the non-class type R via
   // automatic conversion (in one direction only to avoid ambiguities)
   // Most functions use R-arguments and not R1-ones since R1's have to
   // be converted to R anyway in most situations.
public:
   static const R1 min;
   static const R1 max;
   static const R1 eps0;
   static const R1 eps1;
   static const R1 logmax;
   static const R1 log10max;
   static const R1 pi;
   static const R1 piInv;

   R x1;
   typedef R1 Type;
// constructors
   R1():x1(0){}
   explicit R1(R a):x1(a){} // no automatic conversion wanted
   operator R()const{ return x1;}
      // auto-conversion to R is wanted here
   R operator()(void)const{ return x1;}
      // natural output of the intrinsic value
   R1& operator=(R a){ x1=a; return *this;}
   R1& operator+=(R a){ x1+=a; return *this;}
   R1& operator-=(R a){ x1-=a; return *this;}
   R1& operator*=(R a){ x1*=a; return *this;}
   R1& operator/=(R a){ x1*=R(R1(a).inv()); return *this;}
   R1 operator - ()const{ return R1(-x1);}
 //  R operator*(R const& a)const{ return x1*a;}
 //  friend R operator*(R const& a, R1 const& b){ return a*b.x1;}

   R abs()const{ if (x1<0) return -x1; return x1;}
   R dis(const R1&)const;
   X2<R,R1> polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // is sign(x1)
   CPM_IO
 //  CPM_ORDER
   friend bool operator == (R1 r1, R1 r2){ return r1.x1 == r2.x1;}
   friend bool operator != (R1 r1, R1 r2){ return r1.x1 != r2.x1;}
   friend bool operator >= (R1 r1, R1 r2){ return r1.x1 >= r2.x1;}
   friend bool operator <= (R1 r1, R1 r2){ return r1.x1 <= r2.x1;}
   friend bool operator > (R1 r1, R1 r2){ return r1.x1 > r2.x1;}
   friend bool operator < (R1 r1, R1 r2){ return r1.x1 < r2.x1;}
   Z com(R1 r)const; // Compare function as tool, since R itself
      // does not work with com.
   Z dim()const{ return 1;}
   R1 inf(R1 a)const{ return x1<=a.x1 ? R1(x1) : R1(a.x1);}
   R1 sup(R1 a)const{ return x1>=a.x1 ? R1(x1) : R1(a.x1);}
   R& operator[](Z i){ i; return x1;}  // defining components in analogy
      // to what is done for R2, R3, R4
   const R& operator[](Z i)const{ i; return x1;}
   void set(R t){ x1=t;}
   void makePos_(){ if (x1<0) x1=-x1;}
      //: make positive
   Word nameOf()const{ return "R1";}
   Word toWord()const{ return Word::write(x1);}
   bool isVal()const{ return CpmRoot::isVal(x1);}
// functions referring to R and R1 as a discrete sets ('lattice')
   R1& next_();
      // next (in lattice)
   R1& prv_();
      // previous (in lattice)
   Z dz(R a)const;
      // distance on the lattice of represented numbers
      // counts the points exactly and thus is very slow for
      // typical pairs of numbers. If two numbers arrose from
      // getting the same 'mathematical number' by different
      // numerical expressions, the values should be close enough
      // so that dz should be small (approximately the number of
      // operations in the algorithm ???). Function dr is a much
      // faster approximate representation of this discrete distance.
   R eps(Z dir=1)const;
      // distance to the next lattice point in positive direction
      // for dir>=0 and in negative direction for dir<0.
      // In this latter case also the result is returned as negative
      // Thus x+x.eps(1) is the next lattice neighbor of x in
      // positive direction and x+x.eps(-1) is that in negative
      // direction
   R dr(R a)const; // const?
      // approximate number of lattice intervals between a and *this
      // (always >=0)
   Z run(R xFinal, ostream& str, Z writePeriod=1000)const; //const?
      // runs from *this to xFinal in steps over all represented
      // numbers and prints out ( to stream o) each n'th of
      // them with n given by writePeriod
      // Returns the number of printed numbers.
// safe functions
   R1 inv()const;
      // inverse, returns R1::max.x1 for x1==0
   R exp()const;
      // Returns R1::max.x1 for all x1 that would normally give infinity.
      //
      // Example of usage: instead of exp(x) write R1(x).exp() if it is
      // important to get a valid, non-exceptionl, result
   R log()const;
   R log10()const;
      // these four functions are defined on whole R1.
      // logs are log of absolute value, a combination which
      // is common in mathematics.

   R sign()const{ return (x1> 0. ? 1. : (x1==0. ? 0. : -1.));}
         //: sign (English),
         // akin to Latin signum

   R sign(R b)const{ return x1>0 ? x1*R1(b).sign() : -x1*R1(b).sign();}
        // 'takes the modulus from *this and the sign from b'
};

/////////////////////////////// class Z1 ////////////////////////////////

class Z1{ // integer numbers as a class
   // tuples of 1 elements, for formal
   // symmetry with others
public:
   Z x1;
   static const Z1 min;
   static const Z1 max;
   typedef Z1 Type;
// constructors
   Z1():x1(0){}

   explicit Z1(long int a):x1(a){} // no automatic conversion wanted
   explicit Z1(int a):x1(a){} // no automatic conversion wanted
   explicit Z1(R r):x1(cpmrnd(r)){} // no automatic conversion wanted
   operator Z()const{ return x1;}
   R toR()const{ return R(x1);}
   Z1& operator=(Z a){ x1=a; return *this;}
  // Z1& operator=(bool b){ if (b) x1=1; else x1=0; return *this;}
   Z1& operator+=(Z a){ x1+=a; return *this;}
   Z1& operator-=(Z a){ x1-=a; return *this;}
   Z1& operator*=(Z a){ x1*=a; return *this;}

   Z1 pow(Z p)const;
      // Z1(x1^p), cpmassert p>=0

   Z abs()const{ if (x1<0) return -x1; return x1;}
   R dis(const Z1&)const;
   X2<Z,Z1> polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // is sign(x1)
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 1;}
   Z1 inf(Z1 a)const{ return x1<=a.x1 ? Z1(x1) : Z1(a.x1);}
   Z1 sup(Z1 a)const{ return x1>=a.x1 ? Z1(x1) : Z1(a.x1);}
   Z& operator[](Z i){ i; return x1;}  // defining components in analogy
   // to what is done for Z2
   const Z& operator[](Z i)const{i; return x1;}
   Word nameOf()const{ return "Z1";}
   Word toWord()const{ return Word::write(x1);}
   bool isVal()const{ return CpmRoot::isVal(x1);}
   Z parity()const{ return CpmRootX::parity(x1);}

   bool isPowOfTwo()const;
      // Let N:={1,2,3,4,...} (no zero!) and
      // P2:={ 2,4,8,16,...}
      // Then we return true iff *this belongs to P2

   Z1 nextPowOfTwo()const;
      // We return the smallest y \in P2 such that *this <= y

   Z1 nextLog2()const;
      // We return the conventional dual logarithm of
      // nextPowOfTwo() ( which then is always \in N)
};
#endif

} //namespace CpmRootX

namespace CpmRoot{
// All similar declarations are in cpmword.h There, the present
// type is not yet defined. So we have to defer the declaration till
// here. That this works, is a triumph of C++'s templates.

   template<>
   class Name<CpmRootX::fpVoidToVoid>{ // also this type needs a name
   public:
      Name(){}
      Word operator()(CpmRootX::fpVoidToVoid const& t)const
      { t; return Word("fpVoidToVoid");}
   };

   template<>
   class Name<CpmRootX::fpWordToVoid>{ // also this type needs a name
   public:
      Name(){}
      Word operator()(CpmRootX::fpWordToVoid const& t)const
      { t; return Word("fpWordToVoid");}
   };

}// namespace

// new names
#define cpminf       CpmRootX::inf
#define cpmsup       CpmRootX::sup

#define cpmord      CpmRootX::order
#define cpmswp      CpmRootX::swap
#define cpmtin      CpmRootX::tinyNumber
#define cpmhug      CpmRootX::hugeNumber

#endif
