//? cpmuc.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_UC_H_
#define CPM_UC_H_
/*
Purpose: Defining class UseCount and P<>.
Class UseCount is a tool for implementing reference
counting.

Credit: Modified from Andrew Koenig: Ruminations on C++,
AT&T Ch. 7. Authoritative and enlightening treatment. However,
compared to the present version, some essentials are missing
in Koenig's book. Desastrous misprints in the code there!

misprints in Koenig:
(p. 72 reattach():
replace --p by --*p,
p. 73 makeonly(): eliminate p=new ...; )

Recent history:
    2012-01-11 non-constant functions renamed to ending in '_'
    2017-02-18 p renamed to p_, u renamed to u_
    2017-05-23 macro const& used instead of $, '$' needs blanks to\
     separate it
      from variable names or type names. However, e. g.
      bool reattach_(UseCount $);
      is OK
*/


#include <cpmbasicinterfaces.h>

////////////////////////// class UseCount //////////////////////////////

namespace CpmArrays{

   using CpmRoot::Z;

class UseCount { // support for reference counting and 'copy on write'
   // class UseCount of Koenig. See explanations there.
   // Instances of this class are used as data members in client classes
   // which may describe the state of several instances by a common piece
   // of 'free store'. Typical example is class P<> within the present
   // file. Classes derived from P<> are defined in cpmp.h, cpmfl.h,
   // and in cpmfile.h
public:
   UseCount():p_(new Z(1)){}
   UseCount(UseCount const& u):p_(u.p_){++*p_;}
   ~UseCount(){ if (--*p_==0) delete p_;}
      // not virtual since no derivations will be defined
   bool only()const{return *p_==1;}
      //: only
      // to be called in the destructor of the client class
   bool makeOnly()const{if (*p_==1) return false; --*p_; return true;}
      //: make only
      // The name 'makeOnly' needs explanation: If *this is 'only' already
      // no 'making' is involved and 'makeOnly' returns 'false'.
      // If it is not 'only' we do something and thus return 'true'.
      // What is achieved in this doing is, however, not necessarily the
      // status 'only' one step in the direction of becoming 'only'.
      // Only if we had *p_==2 prior to calling makeOnly() we have
      // *p_==1, i.e. the status 'only' after the call.
      // To be called in the assignment operator of the client class.
      // Named unalias() by Bruce Eckel (p. 437). To me both names
      // don't provide a clear suggestion.
      // Consider a typical usage of this function as it appears in
      // the copy on write function (to be called in the assignment
      // operator) of the array class V<T> which has data
      // T* p_, UseCount u_, ... :
      // template <class T>
      // void V<T>::cow_(void){
      //    if (u_.makeonly()){ p_= (sz_==0 ? 0 : copy()); u_.startNew();}
      // }
      // This function thus calls u_.makeonly() once. Only if its return
      // value is true any action happens. This is the case only if
      // *(u_.p_) > 1 so that the piece of free store which u_'s client is
      // using (and to which p_ is pointing) is also being used by a
      // further client. Then action is needed: (*this) connects to a copy
      // of the original piece of free store so that this original piece
      // will have one user less. This is what u_.makeonly does in its
      // instruction --*p. The situation thus is that the old users of the
      // 'land' still use it and the individual *this who tries to change
      // things has to move to a new piece of 'land' which is made into
      // a copy of his old 'land' and on which the intended changes can be
      // made without interfering with the needs of the users of the
      // 'old land'. This is now a clear procedure. Earlier my
      // understanding was that the old users have move and it was not
      // clear how *this would enforce such a cooperation of the old
      // users.

   bool reattach_(UseCount const&);
      //:: reattach
      //'::' means that the name is longer than the one formed according
      // to the CPM standard abbreviation scheme. In most cases no
      // abbreviation at all.
      // To be called in the assignment operator of the client class
      // See P<> and V<>.
      // Recall that the names of non-constant functions end in '_'.

   void startNew_(){ p_=new Z(1);}
      //:: start new
      // my addition
   Z getCount(void)const{return *p_;}
      //:: get count

private:
   UseCount& operator=(UseCount const&);
      // not implemented,to make sure that this will never be used
   Z* p_;
};

///////////////////////  class P< >  ///////////////////////////////////
// Class of constant smart pointers, no arrays (thus no indexing),
// reference counting, but no copy on write. Tool for implementation
// of function class F<X,Y>.
// See cpmp.h for a derived class which implements copy on write.

template <class T>
   // T is a any non-abstract class or any built-in type.
   // If T is immutable, i.e. an instance of T cannot be modified after
   // it has been created, (see Andrew Koenig: Ruminations on C++, AT&T
   // 1997, p. 178, 190 for the relevance of immutable types) P<T>
   // implements the 'strict value interface' (see file cpmv.h).
   // In the notation of predicate logic:
   // non-abstract T & immutable T ==> value class P<T>
   //
   // Let T(A const& a) be a constructor of type T, then
   // the normal syntax for creating an instance of P<T> is
   // A a=...;
   // P<T> pt=new T(a)
   // or
   // P<T> pt;
   // pt=P<T>(new T(a));
   // or
   // P<T> pt(new T(a));

class P { // 'P' as in 'pointer', constant smart pointer.
   // Handle class the instances of which have to be initialized by
   // 'new'-generated pointers. The class takes the full responsibility
   // for deleting these pointers properly.
   typedef P<T> Type; // needed for declaration macro CPM_ORDER
public:
   CPM_ORDER
      // Will be implemented by comparison of adresses.
      // No order-related operators of T used.

   P(T *p=0):p_(p){}
      // Constructor, including default constructor.
      // important !!!: use this constructor only
      // for pointers initialized with a 'new' statement (on free store,
      // not static!) Important also: This allows to create P<T>'s from
      // pointers X* x where X is derived from T.
      // P<T>(new X(..arguments..)) will hold all information, that
      // the object X x(..arguments..) contains, although T does not know
      // about the additional members of X.
      // Also u_ gets initialized by its default constructor which sets
      // the use count to 1.
      // The piece of free store that thus comes under the control of P
      // is of course not associated with a use-counter already. The case
      // that he is, will be handled by the next function:

   P(P<T> const& h):u_(h.u_),p_(h.p_){}
      // copy constructor
      // Here the pointer h.p_ points to a piece of storage which is
      // already in use by h.u_.getCount() clients.
      // The copy constructor of UseCount is made such that it increases
      // the usecount by 1.

   virtual ~P(){ if (u_.only()) delete p_; }
      // destructor

   P<T>& operator=(P<T> const& h);
      // assignment

// access operator, which does not allow to change p
   T const& operator()(void)const{ return *p_;}
      // Let pt an instance of P<T>; then it is conventional to denote
      // the value as *pt. According to my aim to avoid pointers in
      // public interfaces, I also tend to avoid pointer notation.
      // So I denote the value of pt by pt().
      // This looks very natural since 'getting to the data content' of a
      // container-like quantity is much like evaluating a function.
      // Since no argument is needed (as for standard rand(), for example)
      // a void pair of brackets gives the right association.
      // Assume that T = V<X>, then we may form pt()[1].
      // In dereferencing notation we had to write (*pt)[1].

   bool isVal()const{ return !(p_==0);}
      //: is valid
protected:
   UseCount u_;
   T *p_;
};

///////////////////////////// Implementation ///////////////////////////

template <class T>
Z P<T>::com(P<T> const& obj)const
{
   if (p_<obj.p_) return 1;
   if (p_>obj.p_) return -1;
   return 0;
}

template <class T>
P<T>& P<T>::operator=(P<T> const& h)
{
   if (u_.reattach_(h.u_)) delete p_;
   p_=h.p_;
   return *this;
}

} // namespace

#endif
