//? cpmv.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_V_H_
#define CPM_V_H_
/*

Purpose: Building on std::vector

obsolete: Basic array class with valid indexing ranging in a contiguous
   subset of Z. Two cases are of particular interest: That the lowest
   valid index is 0 (as for std::vector<>) or 1 (as e.g. in the functions
   in Press et al.). When dealing with FFT index ranges
   {-n,...,-1,0,1,...,n} are adequate (not yet implemented).
   Most constructors of V create instances of V<T> with valid indexes
   starting at 1. The two functions b_(Z) and e_(Z) provide means
   to arbitrarily shift the range of valid indexes. Access to components
   by means of indexing is range checked.

History: Till October 2010 there was a class template Vl<T> with
indexing starting at 0 and - based on that (not derived from that)
a class template V<T> with indexing starting at 1. This caused an
uncomfortable situation: Whenever dealing with a topic where I felt
that efficiency would be of utmost importance (e.g. font representation
and quantum dynamics) I was tempted to use Vl<> not only as an
internal device but also in the interface of public member functions.
This made the implementation code and even the classes dependent
on a decision that with a slight twist of emphasis could also have
been made differently. The new state of affairs is that Vl has been
eliminated (actually replaced by V) in all C+- code. If it should still
shine up in some comment, ignore it.

Recent history:
    2012-01-11 function cow() renamed to cow_()
    2012-01-18 function X2<Z,bool> findAsc(T const& t)const added
    2012-01-19 function prnOn modified so that indexes are printed
      (in commentarized form) together with the components.
    2017-02-18 Short report on a project which looked promising
      but was finally abandoned:
      The rather mature state that C++ has reached with C++11 triggered the
      desire to make more systematic use of the standard library
      facilities. Especially that now all standard containers are said to
      implement moving as a replacement of copying where appropriate
      promised to make it feasible to replace T* p_ by std::vector<T> p_
      and thus free the implementation of V<T> from defining copy, move,
      assignment operators. The first disappointing problem was that then
      V<bool>, since based on std::vector<bool>, did no longer work in my
      code which used operator[](Z) for V<bool> as in all other V<T>'s.
      Of course replacing V<bool> by V<B> would help. But re-writing all
      the many functions of V<T> in the new style looks disappointingly
      tedious to me. Before making a second attempt I need to understand
      whether the new move functionality is in fact a full replacement for
      the reference counting and copy on write functionality which is
      implemented in V<T> as of today. There is a project
      'codingexperiments' in ~/e/cpm/codingexperiments which illustrates
      the malfunction of vector<bool>, which also is well known to the
      WWW. Further there is ~/e/cpm0ExperimentBasedOnvector which
      contains the experimental version of cpmv.h. In this version by far
      not all uses of T* are replaced by vector<T>.


Credit: Modified and extended from Andrew Koenig: Ruminations on C++,
AT&T Ch. 7. Authoritative and enlightening treatment. However,
compared to the present version, some essentials are missing in
Koenig's book. Also two misprints caused trouble.

This defines a template V<T> of T-valued lists or 'vectors with
T-valued components'. It is similar to std::vector<T> of the STL
but it differs from this in some respect. In describing
V<T> and these differences, I'll introduce some notions and notations
that will be used over and over in defining and stating properties of
other CPM classes
( CPM = 'C+-' or 'Classes for Physics and Mathematics').

1. V<> instances (called simply 'Vectors' here) never request more
   memory than necessary. Thus appending elements always needs new
   allocation of memory.
   V<> is more adapted to the role of vectors as compound
   (structured) quantities than to the role as a sequential container for
   its components. Therefore, appending components at the end is
   not basic to the concept and is not worth special optimizing effords.
   Notice that V<> is not a 'polymorphic container'
   i.e. the components can hold only T typed objects and not the
   additional information content which instances of classes derived
   from T may carry. We may, however, form V<T*> for any type or class
   to support polymorphism in the old-fashioned pointer-based way.
   A safer way is to use the polymorphic vector template Vp mentioned
   in item 5.

2. V<> implements 'reference counting' and 'copy on write'. Thus
   there is no need to work with references to vectors to avoid
   superfluous copying. Assume that we nevertheless observe the
   rule to declare function arguments always as references
   or constant references. Then, the main effect of reference
   counting is that large objects built within the body of a
   function (such as matrix multiplication) will not be copied
   to the outside but simply referenced. If they are no longer
   needed, they nevertheless get deleted automatically.
   If such a referenced large object gets changed by program
   action and some old client still needs the unchanged version
   for his reference, a copy of the large object is made automatically
   and made available to its user.

3. Requirements on T in order to allow the formation of V<T>:
   To illustrate the point up front: std::vector<T> will only be built
   (at least with the MS Visual C++ 5.0 compiler) if T<T and T==T are
   defined. For V<T> there is no such restriction. To be sure: T<T is an
   obvious abbreviation of the statement:
   'for all objects t1 and t2 of type T, the expression t1<t2 is defined'
   (compare David R. Musser, Atul Saini: STL Tutorial and Reference
   Guide, Addison-Wesley 1996, p. 246). Now the pertinent statement:
   V<T>  is well defined if (not iff!) T is a built-in type or a class
   which 'implements the value interface'. Then, V<T> also implements the
   value interface. If, moreover, T 'implements the strict value
   interface', then V<T> also implements the strict value interface.
   Here, a class T is said to implement the value interface if it
   publicly defines T(), T(const T&), and T& operator=(const T&).
   T is said to implement the strict value interface if, in addition, it
   has value semantics (as opposed to pointer semantics, see
   Andrew Koenig: Ruminations on C++, AT&T 1997, p. 62. and
   Bjarne Stroustrup: The C++ Programming Language, 3. edition,
   Addison-Wesley 1997, p. 294). Classes that implement the strict value
   are most convenient to use. Code only using such quantities is
   normally easy to understand. To have an even shorter denotation for
   this favorable species we call such a class a value class or a bit
   more general:

      T is a value class : <==> T is a built-in type or a class
      that implements the strict value interface.

   As was stated already, we have:

      value class T ==> value class V<T>

   The predicates 'T satisfies the value interface' and
   'T satisfies the strict value interface' can be evaluated for
   classes providing random generators by means of the test classes
   Test_v<T> and Test_sv<T> defined in file cpmtests.h. Although V<>
   provides no random generator (Vr<>, to be mentioned soon, does)
   the code of Test_sv<T> can be read as an operational definition
   of the concept 'strict value interface'. An interesting (or common-
   place ?) observation: syntactic aspects of a program are characterized
   by the program's interaction with the compiler; semantic aspects
   ('pointer sematics',...) are characterized by the data created during
   execution.

   Requiring order operators or even arithmetic operators in T, allows to
   define T-'valued' Vectors which themselves carry natural
   order/arithmetic operators.
   Therefore V is only the starting point in a series of vector
   templates with increasing functionality, accompanied by increasing
   assumptions on the structure of T. Presently this hierarchy looks as
   follows:

      V<T> , Vo<T> , Va<T> , Vr<T>

   Every such class is derived from its predecessor in this series
   by derivation without adding new data members. Thus there are
   unambiguous casts between all these classes.

      Vo : More order related methods added to V, basic order functions
         which are declared in CPM_ORDER now already here.
      Va : arithmetic operations added to Vo
      Vr : rich infrastructure supporting automated testing of
         internal consistency.

   This approach of integrating the functions (algorithms) into an
   inheritance tree of template classes is radically different from the
   STL approach. STL keeps algorithms as separate entities outside the
   classes and uses iterators and adaptors for making them work together.
   Both methods have their advantages and disadvantages. It is probably
   fair to say that the C+- approach is less universal but more
   convenient to use within the framework it fits.

4. On polymorphism: V<T*> may be formed, but doesn't implement the value
   interface for T and derived classes. However, with the smart pointer
   templates P<T>, Pp<T>, and Po<T> defined in file cpmp.h we can form
   e.g. V<Pp<T> > and get the functionality (together with some 'extras')
   which one would expect from  V<T*>. See file cpmp.h for details and
   the polymorphic version Vp of the vector templates mentioned so far.
   See test class PolymorphicMulti<*,*,*> in cpmtests.h for a operational
   definition of polymorphism of container classes.
*/

#include <cpmfl.h> // Includes <cpmuc.h> for std::size_t
   // (and std::ptrdiff_t, which is not being used so far).
   // Small and efficient function template with minimum
   // infra-structure requirements.
#include <cpmzinterval.h>
   // includes cpmsystem.h and cpmx.h
   // With using arrays something may go wrong and so messaging
   // capability is indispensible.
#include <cpmtypes.h>

#include <cpmmacros.h>
   // Provides help to write debugging-friendly function blocks.
#include <vector>
#include <set>
/////////////////////////////// class V<> ///////////////////////////////
// Array with index check, reference counting, and copy on write, and
// 'value semantics' (as opposed to 'pointer semantics')
// Generalized from Koenig's class Handle p. 72-73.
// Index range is now defined as an instance of class IvZ. So each
// 'vector' has its individual index range which may start with 1
// (following the convention of the Numerical Receipes) or with 0 as
// for into the 'vector' of the STL.
//  V< V<ColRef> > is the type of bitmap data in my graphical workhorse
// class Img24. This can be considered a proof for good performance of
// the class. Efficient conversion functions from and to std::vector<>
// are now provided.
// There is a nice way to iterate over the whole index range without
// mentioning the bounds as 0 and dim-1, or as 1 and dim:
//    V<T> v=....;
//    for (Z i=v.b();i<=v.e();++i) v[i]=...;

namespace CpmArrays{

   using namespace CpmStd;

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmRoot::N;
   using CpmRoot::toN;
   using CpmRoot::R;
   using CpmRoot::B;
   using CpmRoot::Root;
   using CpmSystem::Error;
   using CpmFunctions::F;
   using std::vector;

#ifdef CPM_Fn
   using CpmFunctions::F1;
#endif

// some infrastructure

   enum Begin { LEAN };
      //: begin
      // Never form V<Begin>

   enum Outside { DEFAULT, CYCLIC, CONSTANT/*, ZERO */};
      //: outside
      // Controls the meaning of 'out of range indexes'.
      // DEFAULT: definition as the default value associated with
      // the type under consideration
      // CYCLIC: setting the meaning of out of range indexes by
      // cyclic repetition of value
      // CONSTANT: continuation as constant from the nearest value
      // ZERO: ????? not used ?

   extern Z dimMax;
      // If the dimension of an array was either the result of a
      // calculation or of reading from a file, then the result may be
      // off the programmer's intent by orders of magnitude if something
      // went wrong. So it is helpful to exclude unnaturally large arrays
      // from becoming allocated.

   extern Z firInd;
      // first index of arrays. Means to let C+-
      // cooperate with std containers.

   extern bool ranChc;
      //: range check
      // Initialized as true.

   extern bool ranChcAlw;
      //: range check always
      // If this is true, all access operators even those the name
      // of which suggests the opposite get checked --- with poor
      // diagnostics, though.
      // Initialized as true, since access to non-allocated memory,
      // as a rule, causes disaster. I had to discover in 2008-03-03
      // that such a case happened in the workhorse function
      // CpmGraphics::Graph::mark(V< V<C> >,...)
      // on a regular basis, due to an seemingly safe but actually
      // un-safe usage of V<>::cui().

   extern bool signal;


   void setDimMax(Z n);
      // sets dimMax=n unless n<0. In this case error with message

   Z safeDim(Z n);
      // returns n for 0<=n<=dimMax, otherwise error with message

   Z makeIndValNotMember(Z i, IvZ iv, Outside mode);
       //: make index valid

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.
   // If the index operator [] is to be used and the variable ranChc
   // (which is initialized as 'true') was not set to 'false' an out
   // of range error will result in a runtime error which will be
   // documented on cpmcerr.txt. See ranChcAlw for an additional control.
   // The error message is particularly explicit (indicating the type of
   // T) when also the macro CPM_NAMEOF is defined. If this is the case,
   // the type T needs to define the member function
   // CpmRoot::Word nameOf()const. For all Cpm-classes this is the case
   // and for user classes, there should be no difficulty in adding such
   // a function. If one wants to make use of the function declared
   // by the declaration macros CPM_ORDER type T needs to define the
   // member function  CpmRoot::Z com(T const&)const;
   // If one wants to make use of the functions declared by the
   // declaration macro CPM_IO, type T needs to define the member
   // functions bool prnOn(ostream&)const and bool scanFrom(istream&).
   // These requirements do not apply to basic types:
   // For T = N, Z, R, Rh, L, bool, string one  may use all functions of
   // V<T> without further requirements.

class V{ // vector template, indexing starts with 1 by default.
   // The index range can however be shifted or even initially set
   // arbitrarily.
   // Although implementation details are inspired from handle classes,
   // the V class is a value array and not a handle class.
   // All allocations made with new T[] all de-allocations are  delete[]
   // All data are private, so the only access to data in
   // derived classes is over the public functions of the class. All
   // these incorporate reference counting internally where needed (only
   // if the preprocessing directive CPM_USECOUNT is active). The user of
   // the class can't see this and has not to be aware of this.

   typedef V<T> Type;

public:
   CPM_IO
   CPM_ORDER
   R dis(V<T> const& h)const;

 //  std::vector<T> rep()const{ return p_;}

   V():iv_(),sz_(0),n_(0),p_(){}
      // default constructor

explicit V(Z n):iv_(n), sz_(n), n_(toN(sz_)), p_(n_)
   {iv_.b_(firInd);}
      // Has n components. Gives an error for n<0 .
      // The components are initialized by the default constructor
      // of T if T is a class for which such a constructor is defined
      // (explicitly or implicitly).
      // If T is a built-in type, initialization is done as 0.
      // See BS3, p. 131 for initialization of built-in types via
      // formal constructor calls.
      // If n>0, the first valid index is 1, which reflects the
      // normal behavior of class V.

   V(Z n, Begin bg);
      // Defined as the previous function. But the first valid index
      // is 0 (if n>0 so that there is at least one valid index).
      // This is a somewhat contrived construction: One has to make sure
      // that this does not interfere with the definition
      // V(Z n, T const& t, Z first=1) for some choice of T. Since we
      // agree on using V<T> only for C+- types T, we will never be
      // tempted to consider V<Begin>.
      // Typical usage:
      //    V<R> v(4,LEAN); // recall: enum Begin { LEAN }
      // lets v have valid indexes 0,1,2,3. v[0]=...=v[3]=0.
      //    V<R> w(4);
      // lets w have valid indexes 1,2,3,4. w[1]=...=w[4]=0.

explicit V(IvZ const& iv);
      // Has a component for each element of iv.
      // The comFponents are initialized by the default constructor
      // of T if T is a class for which such a constructor is defined
      // (explicitely or implicitely).
      // If T is a built-in type, initialization is done as 0.
      // See BS3, p. 131 for initialization of built-in types via
      // formal constructor calls.

   V(Z first, vector<T> const& v);
      // Construction from a standard library vector.

   V(Z n, T const& t, Z first=1) ;
      // Has n components all initialized as t.
      // The third argument gives the first valid index.

   V(Z n, T const& t, Begin bg);
      // Has n components all initialized as t.
      // The third argument (that can have only one value, namely LEAN)
      // says that the first valid index is 0.

   V(IvZ const& iv, T const& t);
      // Has iv.car() components all initialized as t.

   V(IvZ const& iv, vector<T> const& p):iv_{iv}, sz_{iv_.car()},
      n_{toN(sz_)}, p_{p}
      { cpmassert(n_ == p_.size(), "size mismatch in V(IvZ,vector)")}
      // Has iv.car() components all initialized with the components of p.

   V(Z n, F<Z,T> const& f);
      // construction from a function. Of course,
      // V<T> v(n,f);
      // implies v[i]==f(i) for all valid indexes i of v.

   V(IvZ const& iv, F<Z,T> const& f);
      // construction from a function. Of course,
      // V<T> v(iv,f);
      // implies v[i]==f(i) for all valid indexes i of v.

 // constructors from explicit lists
   explicit V(std::initializer_list<T> il );
      // requires C++11
      // constructors from explicit lists such as
      // V<Z> v{0,1,2,4,8};
      // The first valid index always is 1. Therefore
      // v[1]=0,...v[5]=8

   virtual V<T>* clone(void)const{ return new V(*this);}

   V<T> toClnBase()const{ return *this;}
      //: to clone base

   Z dim()const { return sz_;}
      //: dimension
      // Returns the number of components of the vector *this

// Returns the number of components of the vector represented by
// the built-in integer type std::siz_t (alias N). Thus the term
// vector<X>(nc()) is perfectly right without any type conversion.
   N nc()const{ return p_.size();}
      //. number (of) components

   Z size()const { return sz_;}
      //: size
      // for uniformity with STL

   IvZ dom()const { return iv_;}
      //: domain
      // Notice that with the array *this there is the
      // function f: {iv_b(),...,iv_.e()}-->T, i|-->(*this)[i]
      // associated in a natural manner.
      // For this function, dom() is just the domain.
      // The understanding of arrays as functions with domains of type
      // IvZ seems to be a good guide for defining some of the member
      // functions in a more natural manner by using arguments of type
      // IvZ. Present examples are the functions fa_ and valOn.

   bool isVoid()const { return iv_.isVoid();}
      //: is void
      //  short answer on whether the dimension is zero

   bool valInd(Z i)const{ return iv_.hasElm(i);}
      //: valid index
      // returns the validity of i as an index of *this

   Z makeIndVal(Z i, Outside mode=CYCLIC )const
   {
      if (iv_.hasElm(i)) return i;
      if (mode==CYCLIC) return iv_.cyc(i);
      else return iv_.con(i);
   }
      //: make index valid
      // If i is a valid index we return i. If not, the result is
      // iv_.cyc(i) for mode=CYCLIC and iv_.con(i) else. Notice that the
      // normal treatment of mode==DEFAULT works not by replacing one
      // value of the index by another. It works on the value of vector
      // components and makes use of the default constructor T()

   bool sameDom(V<T> const& v)const{ return iv_==v.iv_;}
      //: same domain

   vector<T> std()const{ return p_;}
      //: standard
      // Returns a STL-vector which holds all components of *this.

   virtual Word nameOf()const;
      //: name of
      // returns a name of the type

// constant access functions

   const T& operator[](Z i)const;
      // Getting to the value of a component of a const instant of
      // V<T>. For instance
      // const V<T> v = ... ;
      // T t = v[3];
      // If CPM_USECOUNT is enabled (the normal case) the actual
      // process of going from v to v[3] depends on whether v is of type
      // V<T> or const V<T>. In the non-constant case the evaluation of []
      // may involve a copy action on v (and returning the component of
      // the copy). If we intend only to read the component, such an
      // copy action is not needded and should be avoided by casting v
      // to type const V<T>.
      // Casting v from V<T> to const V<T> works this way:
      //    v = static_cast<const V<T>>(v);
      // Casting back to mutable seems to work only by using a new name:
      //    auto vMutable = const_cast<V<T>&>(v);
      //    vMutable[1] = T(); //( for instance)
      // or by applying mutating operations to an anonymous object as in:
      //    const_cast<V<T>&>(v)[1]=T();
      // Notice the '&' here which the compiler requires. By the way, the
      // effect of the 'anonymus action' actually is to change the state
      // of v: v[1]==T().
      // My analysis of the situation is in ~/codingexperiments/main.cpp
      // In the following there are many pairs of functions
      // T const& f(...)const;
      // T& f(...);
      // for which the above considerations apply mutandis mutatis.

   T const& cui(Z i)const;
      //. component (with) unchecked index
      // getting to the value of a component for 'read', e.g.
      // T t=cui(i);
      // It helps to write efficient code in classes which use V<>-typed
      // data members.
      // Notice that writing loops by using b() and e() to define
      // the range is much safer than using limits like 1 and dim().
      // Index range check is missing only if variable ranChcAlw was
      // changed to 'false'.

   T const& pi(N i)const{ return p_[i];}
      //. plain index
      // index range not checked.
      // It helps to write efficient code in classes which use V<>-typed
      // data members.

   T const& pr(Z i, Outside mode)const{ //. plain read, see below for more
      if (0 <= i && i < sz_) return p_[i];
         // Notice that for i of type CpmRoot::Z the expression p_[i]
         // is perfectly right. Of course, the same is true for i of
         // type std::size_t (for which CpmRoot::N is an alias).
      if (i == -1){
         if (mode==DEFAULT) return def_;
         if (mode==CYCLIC) return p_[n_-1];
         if (mode==CONSTANT) return p_[0];
      }
      if (i == sz_){
         if (mode==DEFAULT) return def_;
         if (mode==CYCLIC) return p_[0];
         if (mode==CONSTANT) return p_[n_-1];
      }
      return def_; // should never happen
   }
      //. plain read
      // from an index which may be out of range by one, in which case
      // the return value is determined in an self-evident way by the
      // Outside-typed second argument.

   T& pi(N i){ return p_[i];}
      //. plain index
      // index range not checked.
      // It helps to write efficient code in classes which use V<>-typed
      // data members.

   T const& cyc(Z i)const;
      //: cyclic
      // getting to the value of a component for 'read', e.g.
      // T t=cyc(i);
      // Here i is understood as cyclic (i.e. i modulo sz_). So no value
      // of the index has to be considered 'out of range' and for i
      // 'in range' cyc(i)==(*this)[i]

   T const& li(Z i)const{ return p_[i];}
      //. lean index
      // The valid range is for (Z i=0;i<sz_;++i) li(i)

   T& li(Z i)
      //. lean index
   {
      return p_[i];
   }

   T const& con(Z i)const;
      //: constant
      // Same logic as cyc() but with constant continuation
      // instead of cyclic.

   T operator()(Z i, Outside mode=DEFAULT)const;
      //: ()
      // Read access defined for all i.
      // By this definition, a vector becomes a mapping from Z to T.
      // For i's outside the proper range,
      // the default T is returned for mode==DEFAULT or if
      // sz_==0. If mode==CYCLIC cyc(i) gets returned.
      // For mode==CONSTANT we return p_[0] for i<=0 and p_[sz_-1]
      // for i>=sz_.
      // Notice that the return value is not a reference in
      // accordance with function behavior.

   T const& read(Z i, Outside mode=DEFAULT)const;
      //: read
      // Same as previous function but returning a reference instead of a
      // T.
      // Reading components in an efficient (as &'s), safe and flexible
      // manner.
      // See T operator()(Z i, Outside mode=DEFAULT)const; for the
      // meaning of the second argument.

   T const& r(Z i)const{ return p_[iv_.ri(i)];}
      //. read
      // Unchecked and fast version of T const& oparator[](Z i)const

   T& w(Z i)const{ return p_[iv_.ri(i)];}
      //. write
      // Unchecked and fast version of T& oparator[](Z i)

   F<Z,T> fnc(Outside meth=DEFAULT)const;
      // function
      // returns the function Z --> T, i |--> (*this)(i,meth)

 // non-constant access functions
   T& operator[](Z i);
      //: []
      // See T const& operator[](Z i)const for discussion of details
      // that are relevant in the case that CPM_USECOUNT is defined.

   T& cui(Z i);
      //. component (with) unchecked index
      // setting to the value of a component
      // T t=...;
      // V<T> x=...;
      // x.cui(i)=t;

    T& cyc(Z i);
      //: cyclic
      // setting to the value of a component
      // T t=...;
      // V<T> x=...;
      // x.cyc(i)=t; meaning of i is modulo sz_. So i is never out of
      // range

   T& con(Z i);
      //: constant
      // Same logic as cyc() but with constant continuation
      // instead of cyclic.

   V<T>& b_(Z i){ iv_.b_(i); return *this;}
      //. set b(), i.e. the first valid index.
      // For instance
      // V<R> v("",sqrt(2),sqrt(3),sqrt(4),sqrt(5));
      // for (Z i=1;i<=v.dim();++i) cout<<v[i]<<endl;
      // v.b_(0);
      // for (Z i=0;i<v.dim();++i) cout<<v[i]<<endl;
      // shows all components of the vector in both cases.

   V<T>& e_(Z i){ iv_.e_(i); return *this;}
      //. set e(),  i.e. the last valid index.

// accessing the first and the last element for reading
   T const& fir()const;
      //: first
   T const& last()const;
      //: last

// accessing the first and the last element for writing
   T& fir();
      //: first
   T& last();
      //: last

// getting the first and the last valid index
   Z b()const { return iv_.b();}
      //. begin
      // Returns the first valid index.

   Z e()const { return iv_.e();}
      //. end
      // Returns the last valid index.
      // Allows to write loops over all components as
      // for (Z i=v.b();i<=v.e();i++) ... v[i] ...;
      // Note that this is safe also for v.dim()==0, since then
      // v[e()] will never be called. Here one could replace
      // v[i] by v.cui(i) without danger.

   Z n()const { return iv_.n();}
      //. next
      // Returns the index next to the last valid one.
      // This allows to write loops over all components as
      // for (Z i=v.b();i<v.n();i++) ... v[i] ...;
      // Note that this is safe also for v.dim()==0, since then
      // v[e()] will never be called. Here one could replace
      // v[i] by v.cui(i) without danger.

   V<T> meet(IvZ const& iv)const;
      //: meet
      // Returns a vector which, when considered as a function is the
      // restriction of function *this to the domain iv_ & iv.

   V<T> join(IvZ const& iv)const;
      //: join
      // Returns a vector which, when considered as a function is the
      // extension of function *this to the domain iv_ | iv, where
      // all function values on iv\iv_ are T().

   V<T> operator +(Z i)const{ return V<T>(iv_+i,p_,"");}
      //: operator +
      // Returns a vector res such that res.dom() is the shifted
      // domain dom()+i of *this and has the same components as
      // *this.

   V<T> operator -(Z i)const{ return V<T>(iv_-i,p_,"");}
      //: operator -
      // Returns a vector res such that res.dom() is the shifted
      // domain dom()-i of *this and has the same components as
      // *this.

   void set_(T const& t);
      //: set
      // non-constant function which sets all components of (*this)
      // equal to t

   V<T> set(Z i, T const& t)const;
      //: set
      // Returns a vector that originates from *this by setting the
      // value of component i.
      // Regular behaviour:
      // if we say
      // Z i=...;
      // T t= ...;
      // V<T> v=...;
      // v=v.set(i,t);
      // then the i-th component (see eli() ) of v will get the value t
      // unless i<1.
      // If i is a value for which (*this)[i-1] is not yet defined,
      // the vector becomes enlarged to the necessary size and all
      // not specified components initialized with the default
      // constructor of T

   T in_(T const& t, bool reversed=false);
      //. insert
      // This operations treats *this as a shift register:
      // All components get shifted by one position 'to the right' and
      // the total length (dim) of the vector remains the same. So what
      // was the last component prior to the operation has to be removed
      // from the vector, in order to not wasting information, this
      // removed component will shine up as the return value of the
      // function. After the operation, the first component of the vector
      // will be t.
      // If reversed==true, t gets inputted at the end, and all shift
      // operations go 'to the left'.


   Z locAsc(T const& t)const;
      //: locate ascending
      // Here it is assumed that *this is ascendingly strictly ordered:
      // If sz_>=2 we have p_[i]<p_[i+1] for all i \in {0,sz_-2}.
      // For sz_<2 there is no condition.
      // For sz_==0 we stop with error.
      // For t<fir() we return b()-1
      // For t>=last() we return e()
      // If none of the previous conditions was met we necessarily have
      // sz_>=2 and we return the uniquely determined j such that
      //    p_[j]<=t<p_[j+1].
      // The possible results from this regular part of the functionality
      // thus are b(),...,e()-1.
      // Notice that the very similar function locate of Press et al.
      // only guarantees p_[j]<=t<=p_[j+1] for its return value j.

   Z find(T const& t, bool ordered=true)const;
      //: find
      // If the return value is b()-1, there is no valid index i
      // such that (*this)[i]==t. Otherwise, the return a value i
      // such that (*this)[i]==t.
      // If ordered==true, this assumes that (*this) is in increasing
      // order. The method is by bisection and logarithically fast.
      // If no order is assumed the components are compared against t
      // starting from index b(). The first occurrence of t gets reported.

// building new objects from given ones (generative methods)

// concatenating vectors and appending components. These operations have
// always O(sz_) complexity. The corresponding combined assignements are
// less direct to implement are not considered useful in the present
// context (they would not be more efficient than the friend versions,
// since new memory has to be allocated anyway).

   V<T> app(T const& t)const;
      // returns a vector which is obtained from *this by appending t
      // as the last component
      // notice also the prepend functions to be introduced
      // after the insert-functions (in order to have the inline
      // definition available).

   V<T> app(V<T> const& h)const;
      // returns a vector which is obtained from *this by appending h
      // at the back end

   void push_back(T const& t) { *this=app(t);}
      // for uniformity with STL

   V<T>& operator<<(T const& t) { return *this&=t;}
      // appending an element in Ruby-style
      // Allows successive application as in
      // V<Word> w;
      // w<<"many"<<"words"<<"get"<<"easily"<<"stored";

   V<T>& operator<<(V<T> const& h) { return *this=app(h);}
      // appending an array in Ruby-style

   V<T>& operator&=(T const& t){
      iv_.n_();
      sz_++;
      p_.push_back(t);
      return *this;
   }

   V<T>& operator&=(V<T> const& h) { return *this=app(h);}
   V<T> operator&(T const& t)const { return app(t);}
   V<T> operator&(V<T> const& h)const { return app(h);}
      // appending in my favorite style

   V<IvZ> valOn(F<T,bool> const& f)const;
      //: valid on
      // retuns the array of those sub-intervals of the
      // whole indexing interval dom() on which the function
      // dom()-->bool, i|-->f((*this)[i]) yields true.
      // Notice that the result res \in V<IvZ> is a convenient
      // representation of a subset of dom(). Considering
      // *this as a function g: dom()-->T, then the function
      // h:=g&f is of type dom()-->bool and res, as a subset
      // of dom(), is h^-1({true}).

// resizing

   V<T> resize(Z newDim)const;
      // Returns a vector res such that res.dim()=newDim. If newDim is
      // smaller than dim(), the end of *this will be cut away. If newDim
      // is larger, T()'s will be added.
      // For newDim<0 the action is as if newDim==0.

   V<T> cut(Z i)const{ return resize(sz_-i);}
      // returned is a V<T> which results from *this by removing i
      // components from the end. For exotic values of i, see code
      // and explanation of resize.

// elimination and insertion

   V<T> eli(IvZ const& iv)const;
      // eli stands for eliminate
      // Eliminating all components which belong to iv. No exceptions
      // can happen! Universal and convenient of elimination. All other
      // forms are superfluous. Moreover, they are to be considered as
      // obsolete.

   V<T> eli(Z i, Z nEli=1)const{ return eli(IvZ(nEli,i,0));}
      // eli stands for eliminate
      // returned is a vector which is obtained from *this by
      // eliminating the i-th component and the nEli-1 following ones
      // (thus nEli components are removed) and shifting all later
      // components (if there are such components left) 'to the left' to
      // close the gap.

  // V<T> eliFirst(Z nEli=1)const { return eli(1,nEli);}
   V<T> eliFirst(Z nEli=1)const { return eli(IvZ(nEli,b(),0));}
      // returned is a vector which is obtained from *this by
      // eliminating the nEli first components. If no
      // argument is provided, actually the first component
      // gets eliminated. Notice that in the three-argument constructor
      // IvZ(i,j,k) the first argument is the cardinality, the second
      // argument is the first element, and the third argument is dummy.

  // V<T> eliLast()const { return eli(sz_,1);}
   V<T> eliLast()const { return eli(IvZ(e(),e()));}
      // returned is a vector which is obtained from *this by
      // eliminating the last component.

   V<T> eli1(V<T>& h, Z i)const;
      // We return a vector which results from *this by eliminating
      // h.dim() components starting at the i'th component (for i<1,
      // i==1 is understood). After the call h will hold the
      // eliminated components in due order (and no more - even if h was
      // longer before).
      // The function name ends in '1' to indicate that the first
      // argument is a non-constant reference, used for communication
      // of a part of the result.

   // condensation (contracting equivalent components into one)

   X2< V<T>, V<Z> > condense( bool (*equi)(const T&, const T&))const;
      // given an equivalence relation equi on T (t1~t2 <==>
      // equi(t1,t2)==true )
      // we divide the components of *this into equivalence classes.
      // Let the resturned pair be written as (res1,res2) . Then
      // (i) (*this)[i]~res1[res2[i]]
      // (ii) there is a j such that  (*this)[j]==res1[res2[i]]
      // Actually, the construction is made such that this j is the
      // smallest j for which (*this)[j]~res1[res2[i]].

   V<T> select(V<B> const& s)const;
      // returned is a list which is obtained from *this by eliminating
      // all components (*this)[i] for which s[i] is defined and
      // statisfies s[i]==false. Beside of this removal, the order of the
      // components in *this is retained. So if s is defined
      // by a condition s[i]=Condition((*this)[i]), for the vector
      // component, this condition has to express a property we like to
      // have fulfilled for the result-vector of the select-operation.
      // 's expresses the favorable condition' and  n o t  the one to be
      // eliminated. Notice, that the select operation makes sense for
      // all values of s.dim().

   V<T> ins(Z i, T const& t)const;
      //: insert
      // returned is a vector which is obtained from *this by
      // inserting t as the i-th component and shifting all later
      // components 'to the right' to give room for t.
      // The phrase 'i-th component' refers to natural counting (thus
      // p_[i-1] is the i-th component of *this).

   V<T>& ins_(Z i, T const& t);
      //: insert
      // returned is a vector which is obtained from *this by
      // inserting t as the i-th component and shifting all later
      // components 'to the right' to give room for t.
      // The phrase 'i-th component' refers to natural counting (thus
      // p_[i-1] is the i-th component of *this).


   V<T> ins(Z i, V<T> const& h)const ;
      // returned is a vector which is obtained from *this by
      // inserting h as the i-th and following component,
      // and shifting all later components of *this
      // 'to the right' to give room for h.
      // The phrase 'i-th component' refers to natural counting (thus
      // p_[i-1] is the i-th component of *this).

   V<T> prepend(T const& t)const
      // returns a vector which is obtained from *this by appending t
      // as the first component
      { return ins(1,t);}

   V<T>& prepend_(T const& t);
      // changes *this by appending t  as the first component

   V<T> prepend(V<T> const& h)const
      // returns a vector which is obtained from *this by appending h
      // at the front end
      { return h.app(*this);}

   V<T> rev()const;
      //: reversed
      // returned is the reversed vector (indexing in the opposite
      // direction)

   V<T>& rev_(){ return *this = rev();}
      //: reversed
      // changes *this into a reversed version of itndexing in the
      // opposite direction

   // re-indexing

   V<T> rot(Z s, Outside mode=CYCLIC)const;
      //: rotate
      // Name as the list transformation function Rotate of Mathematica.
      // v.rot(s)[i] == v(i-s,mode)
      // This means that we shift the component i of the original vector
      // into the new position i+s

   void rot_(Z s, Outside mode=CYCLIC){ *this=rot(s,mode);}
      //: rotate
      // Same as rot, but as a mutating operation.


   V<T> compose(V<Z> const& w)const;
      //:: compose
      // v.compose(w)[i] = v[w[i]]

   V<T> permute(V<Z> const& w)const{ return compose(w);}
      //:: permute

   // transforming components

   V<T> operator+(V<T> const& h)const{
      cout<<"operator + called"<<endl;
      V<T> res(sz_);
      for (Z i=0; i<sz_; ++i){ (res.p_)[i]=p_[i]+(h.p_)[i];}
      return res;
   }

   template <class Y>
   V<Y> operator()(F<T,Y> const& f)const { return V<Y>(iv_,fnc()&f);}
      // generating arrays of different type by a type changing function

   V<T> operator()(T (*f)(T const&))const{ return fa(f);}
      // generating arrays of the same type with components f(c) instead of
      // c.

   V<T> operator()(T (*f)(T))const{ return faa(f);}
      // generating arrays of the same type with components f(c) instead of
      // c.

   V<T> operator()(std::function<T(T)>(f))const{ return fsf(f);}
      // generating arrays of the same type with components f(c) instead of
      // c.

// defining generative laws and mutating laws for various expressions by
// function pointers. The idea behind is, that for T which allows
// particular operations, we can define those without using iteration via
// a operator[] since these all are defined directly in terms of
// pointers.
// See implementation of +=, -=, ... in Va<T> how this works

   T fAcc1(T (*f)(T const&), void (*acc)(T&, T const&))const;
      // returns an accumulated value of all values f(c), where c
      // are the components of *this. Accumulation is defined by
      // the argument acc:
      // T res=T(); 'for all components c' acc(res,c)

   T fAcc2(T (*f)(T const&, T const&), void (*acc)(T&, T const&),
      V<T> const& h )const ;
      // returns an accumulated value of all values f(c,ch), where c and
      // ch are the components of *this and h respectively. Accumulation
      // is defined by the argument acc:
      // T res; 'for all components c' acc(res,c)
      // useful in defining scalar products

   template <class Y>
   Y fAcc3(Y (*f)(T const&, T const&), void (*acc)(Y&, Y const&),
      V<T> const& h )const ;
      // returns an accumulated value of all values f(c,ch), where c and
      // ch are the components of *this and h respectively. Accumulation
      // is defined by the argument acc

   V<T> fAcc4(F<T2<T>,T> const& f)const;
      // The kind of accumulation needed for computing the forces in
      // particle systems with pair interaction. Here we assume that
      // forces and positions are of the same type, e.g. R2 or R3.
      // Let x[1],...x[n] be the components of *this. We first compute
      // the incomplete matrix f(x[i],x[j]) i=1,...n, j<i and complete it
      // assuming f(x[i],x[j]) = - f(x[j],x[i]). In application mentioned
      // earlier this is the collection of mutually forces. The total force
      // on particle i is sum over j of f(x[i],x[j]). It is the i-th
      // component of the V<T>-typed return value of the function.
      // The force type T thus needs to provide operations unary - and +=.

   V<T> fAcc5(F<R,R> const& dPot)const;
      // The kind of accumulation needed for computing the forces in
      // particle systems with pair interaction derived from a distance-
      // dependent (radial symmetric) potential. The argument dPot is
      // the derivative with respect to r of the r-dependent radially
      // symmetric pair potential.

   V<T> fAcc6(F<R,R> const& dPot)const;
      // forces from a space-dependent potential

   V<T> fsf(std::function<T(T)> f)const;
      // transforming the components with std::function

   V<T> fa(T (*f)(T const&))const ;
      // returns a vector defined by replacing the components c of *this
      // by f(c)

   V<T> faa(T (*f)(T))const ;
      // returns a vector defined by replacing the components c of *this
      // by f(c)

   V<T> fb(T (*f)(T const&, T const&), T const& t)const;
      // returns a vector defined by replacing the components c of *this
      // by f(c,t)

   template <class Y>
   V<T> fb2(T (*f)(T const&, Y const&), Y const& t)const;
      // returns a vector defined by replacing the components c of *this
      // by f(c,t)

   template <class Y>
   void fb2_(T (*f)(T const&, Y const&), Y const& y);
      // replaces *this by a vector defined by replacing the components
      // c of *this by f(c,y)

   void fc_(T (*f)(T const&, T const&), T const& t) ;
      // replaces *this by a vector defined by replacing the components c
      // of *this by f(c,t)

   void fd(T (*f)(T const&, T const&), T& t)const;
      // replaces t by f(c,t) for each component c
      // If, for instance, f(c,t)=t+c*c we replace t by t+sum of c*c

   V<T> fe(T (*f)(T const&, T const&), V<T> const& h)const;
      // returns a vector defined by replacing the components c of *this
      // by f(c,c') where c' are the components of h.
      // Error if dim()!=h.dim()

   template <class Y>
   V<T> fet(T (*f)(T const&, Y const&), V<Y> const& h)const;
      // returns a vector defined by replacing the components c of *this
      // by f(c,c') where c' are the Y-typed components of h.
      // Error if dim()!=h.dim(). Template version of fe. Unfortunately
      // h.p_ is not accessible in the implementation of this function.
      // This enforced introducing the non-canonical access function rep()
      // in 2014-01-23.

   V<T> fe2(T (*f)(T const&, T const&), V<T> const& h)const;
      // Very similar to fe. However, the dimension of the result
      // is the maximum of dim() and h.dim(). Non-existing components
      // of any of the operands are replaced by T().

   void ff_(T (*f)(T const&, T const&), V<T> const& h, Z i=0);
      // Replaces the components of p_ starting from p_[i]
      // by f(c,c') where c' are the components of h.p_. Thus for i=0
      // and h.dim()>=sz_ the whole array p_ is affected.

   void ffl_(T (*f)(T const&, T const&), V<T> const& h);
      // lean form of ff_
      // Replaces the components of p_
      // by f(c,c') where c' are the components of h.p_
      // p_ and h.p_ are assumed to be of the same size


   void fg_(T (*f)(T const&, T const&, T const&),
      V<T> const& h, T const& t);
      // replaces the components c of *this
      // by f(c,c',t) where c' are the components of h

   void fa_(F<T,T> const& f, IvZ iv);
      // replaces the components c of *this with index in iv
      // by f(c); modern form of fa.

   void fh_(T const& w1, T const& w2, T const& w3, Outside mode);
      // replaces the component c[i] of *this by
      // w1*read(i-1,mode)+w2*read(i,mode)+w3*read(i+1,mode)
      // Thus makes use of multiplication in T.

   template <class Y>
   void fht_(Y const& w1, Y const& w2, Y const& w3, Outside mode);
      // replaces the component c[i] of *this by
      // read(i-1,mode)*w1+read(i,mode)*w2+read(i+1,mode)*w3
      // Thus makes use of multiplication in T.

   template <class Y>
   V<T> fh(T (*f)(T const&, T const&, T const&, Y const&, Z),
      Y const&, Outside mode)const;
      // Returns a vector in which the component c[i] is replaced
      // by f(c[i-1],c[i],c[i+1],y,i),where the i-/+1 are understood
      // according to mode. The quantity y helps to provide parameters
      // required by a concrete situation. Worked for implementing
      // the Hamiltonian corresponding to Dirac's relativistic wave
      // equation.

   template <class Y>
   V<Y> fi(Y (*f)(T const&, T const&), void (*acc)(Y&, Y const&))const;
      // Returns a vector with Y-valued components y[i] which are obtained
      // by accumulating the terms f((*this)[i],(*this)[j]).
      // Accumulation is defined by the argument acc.

   template <class Y>
   V<Y> fj(Y (*f)(T const&, T const&), void (*acc)(Y&, Y const&))const;
      // Returns a vector with Y-valued components y[i] which are obtained
      // by accumulating the terms f((*this)[i],(*this)[j]).
      // Accumulation is defined by the argument acc.
      // We assume f((*this)[i],(*this)[j]) == - f((*this)[j],(*this)[i])
      // and use this for doing only one evaluation of f
      // for the two pairs (i,j) and (j,i) and no evaluation
      // of f for any pair (i,i).
      // We assume that for any Y-object y, -y is defined.

     vector<T> rep()const{ return p_;} // 20014-01-24
      // This is needed in the implementation of V<T>::fet<Y>. Here we need
      // direct access to the data member p_ of a function argument of type
      // V<Y>. Unfortunately (and unexpectedly) what is OK for V<T> does
      // not work for V<Y>. This function conflicts with my ambition to
      // ban pointers from the public interface of C+- classes.

private:
// static data
   static const T def_;
      // allows read function to return references
// static functions
   static T fCyc(Z const& i, V<T> const& v) { return v(i,CYCLIC);}
   static T fCon(Z const& i, V<T> const& v) { return v(i,CONSTANT);}
   static T fDef(Z const& i, V<T> const& v) { return v(i,DEFAULT);}
 //  static N toN(Z i) { return static_cast<N>(i);}
      // size type
      // Instead of new T[i], I now write new T[toN(i)] so that
      // allocation is always fed with a parameter which fits the system
      // (in 64 bit systems, size_t may be 64 bit wide). size_t is known
      // due to inclusion of <cpmfl.h> and it is assumed to take the
      // bit-width of the machine properly into account.
   void ini_(){sz_=iv_.car();n_=toN(sz_);p_=vector<T>(n_);}
 //  void ini_(T *p){sz_=iv_.car();p_=p;}

protected:
// data members
   // These should be accessible to derived classes for enabling fast
   // component operations.
   IvZ iv_;
      // Set of valid indexes. Since (*this) can be considered a
      // function iv_ --> T, this object is also called the domain
      // of *this.

   Z sz_{0};
      // number of valid indexes (depends on iv_, equals iv_.car())

   N n_{0};

   std::vector<T> p_;
      // data member that holds the components as a container.
};

//////// using V for some special template arguments///////////////////

V<Z> IvZtoVofZ(IvZ const& iv);
   //: IvZ to V of Z
   // returns the Z's that make up the interval iv
   // as the components of an ordered array (increasing
   // order, of course)

V<Z> VofIvZtoVofZ(V<IvZ> const& viv);
   //: V of IvZ to V of Z
   // appends the results from applying the previous functions
   // to the viv[i] according to the obvious code
   //  { V<Z> res(0); Z n=viv.dim();
   //  for (Z i=0;i<n;++i) res&=IvZtoVofZ(viv[i]); return res;}

V<Word> comLine(int argc, char* argv[]);
   //: command line
   // converts the C-traditional command-line argument into a C+- array.

////////////////////// Implementation ///////////////////////////////
template <class T>
const T V<T>::def_=T();

template <class T>
Word V<T>::nameOf()const{
   Word nt=Root<T>(T()).nameOf();
   Word wi="V<";
   return wi&nt&">";
}

template <class T>
R V<T>::dis(V<T> const& h)const
{
   R res=0.;
   if (iv_!=h.iv_) return R(1);
   for (Z i=b();i<=e();++i){
      res+=Root<T>(cui(i)).dis(h.cui(i));
   }
   return res;
}

template <class T>
inline T const& V<T>::operator[](Z i)const
{
   if (ranChc){
      if (!iv_.hasElm(i)){
         cout<<"index access error at i="<<i<<endl;
         cpmerror(nameOf()&
         "::operator[]const: read-index out of  range: i= "&
         cpm(i)&" iMin= "&cpm(iv_.b())&" iMax= "&cpm(iv_.e()));
      }
   }
   if (signal) cout<<" const [] called"<<endl;
   return p_[iv_.ri(i)];
}


template <class T>
inline T& V<T>::operator[](Z i)
{
   if (ranChc){
      if (!iv_.hasElm(i)){
         cout<<"index access error at i="<<i<<endl;
         cpmerror(nameOf()&"::operator[]: write-index out of  range: i= "&
         cpm(i)&" iMin= "&cpm(iv_.b())&" iMax= "&cpm(iv_.e()));
      }
   }
   if (signal) cout<<" mutating [] called"<<endl;
   return p_[iv_.ri(i)];
}

// using intentionally previously defined []-indexing
// for common messaging and safeness.
template <class T>
T const& V<T>::fir()const{ return (*this)[iv_.b()];}

template <class T>
T& V<T>::fir(){ return (*this)[iv_.b()];}

template <class T>
T const& V<T>::last()const{ return (*this)[iv_.e()];}

template <class T>
T& V<T>::last(){ return (*this)[iv_.e()];}

template <class T>
inline T const& V<T>::cui(Z i)const
{
   if (ranChcAlw){
      if (!iv_.hasElm(i)) cpmerror("index out of range");
   }
   return p_[iv_.ri(i)];
}

template <class T>
inline T& V<T>::cui(Z i)
{
   if (ranChcAlw){
      if (!iv_.hasElm(i)) cpmerror("index out of range");
   }
   return p_[iv_.ri(i)];
}

template <class T>
T const& V<T>::cyc(Z i)const
{
   return (*this)[iv_.cyc(i)];
}

template <class T>
T& V<T>::cyc(Z i)
{
   return (*this)[iv_.cyc(i)];
}

template <class T>
T const& V<T>::con(Z i)const
{
   if (sz_==0) return def_;
   return (*this)[iv_.con(i)];
}

template <class T>
T& V<T>::con(Z i)
{
   if (sz_==0){
      cpmerror("V<T>::con(Z i): i="&cpm(i)&
         " is no valid index in void array");
      return p_[0]; // never happens
   }
   return (*this)[iv_.con(i)];
}

template <class T>
T const& V<T>::read(Z i, Outside mode)const
{
   if (iv_.hasElm(i)) return p_[iv_.ri(i)];
   else{
      if (mode==DEFAULT) return def_; // reference to it can be returned
      else if (mode==CYCLIC) return cyc(i);
      else return con(i);
   }
}

template <class T>
T V<T>::operator()(Z i, Outside mode)const
{
   if (sz_== 0){
       return T();
   }
   else if (iv_.hasElm(i)){
      return p_[iv_.ri(i)];
   }
   else if (mode==DEFAULT){
      return T();
   }
   else if (mode==CYCLIC){
      return cyc(i);
   }
   else  return con(i);
}

template <class T>
F<Z,T> V<T>::fnc(Outside mode)const
{
#ifdef CPM_Fn
   if (mode==DEFAULT) return F1<Z,V<T>,T>(*this)(fDef);
   else if (mode==CYCLIC) return F1<Z,V<T>,T>(*this)(fCyc);
   else return F1<Z,V<T>,T>(*this)(fCon);
#else
   if (mode==DEFAULT) return F<Z,T>(bind(fDef,_1,*this));
   else if (mode==CYCLIC) return F<Z,T>(bind(fCyc,_1,*this));
   else return F<Z,T>(bind(fCon,_1,*this));
#endif
}

// constructors
// 137 is a dummy argument

/*
template <class T>
V<T>::V(Z n):iv_(safeDim(n),1,137)
{
   ini_();
}
*/

template <class T>
V<T>::V(Z n, Begin bg):iv_(safeDim(n),0,137)
{
   ini_();
}

template <class T>
V<T>::V(Z n, T const& t, Begin bg):iv_(safeDim(n),0,137),
   sz_(iv_.car()),n_(toN(sz_)),p_(n_,t){}


template <class T>
V<T>::V(Z n, T const& t, Z first):iv_(safeDim(n),first,137),
   sz_(iv_.car()),n_(toN(sz_)),p_(n_,t){}

template <class T>
V<T>::V(Z n, F<Z,T> const& f):iv_(safeDim(n),1,137),sz_(iv_.car()),
n_(toN(sz_)),p_(n_)
{
   N i=0;
   for (Z j=iv_.b();i<n_;++i,++j) p_[i] = f(j);
}

template <class T>
V<T>::V(IvZ const& iv):iv_(iv), sz_(iv_.car()), n_(toN(sz_)),p_(n_){}


template <class T>
V<T>::V(IvZ const& iv, T const& t):iv_(iv), sz_(iv_.car()),
   n_(toN(sz_)), p_(n_,t){}

template <class T>
V<T>::V(IvZ const& iv, F<Z,T> const& f):iv_(iv),sz_(iv_.car()),
   n_(toN(sz_)), p_(n_)
{
   N i=0;
   for (Z j=iv_.b();i<n_;++i,++j) p_[i] = f(j);
}

template <class T>
V<T>::V(std::initializer_list<T> il ):
iv_(il.size()), sz_(iv_.car()), n_(toN(sz_)), p_{il}{}
   // notice that, in view of the construction of iv_, the first
   // valid index of the constructed vector is 1

template <class T>
V<T>::V(Z first, std::vector<T> const& v):
iv_(v.size(),first,137), sz_(iv_.car()), n_(toN(sz_)), p_(v){}

template <class T>
V<T> V<T>::meet(IvZ const& iv)const
{
   IvZ ivRes=iv_.meet(iv);
   V<T> res(ivRes); // all components are T()
   for (Z i=res.b();i<=res.e();++i){
       if (iv.ni(i)) res[i]=(*this)[i];
   }
   return res;
}

template <class T>
V<T> V<T>::join(IvZ const& iv)const
{
   IvZ ivRes=iv_.join(iv);
   V<T> res(ivRes); // all components are T()
   for (Z i=res.b();i<=res.e();++i){
       if (iv_.ni(i)) res[i]=(*this)[i];
   }
   return res;
}

template <class T >
Z V<T>::find(T const& t, bool ordered)const
{
   Z n=dim();
   Z i0=b();
   Z iEx=i0-1;
   Z res{137};
   if (n==0){
      res = iEx;
   }
   else{
      if (ordered){
         auto q=std::lower_bound(p_.begin(),p_.end(),t);
         if (q==p_.end() || *q!=t){
            res = iEx;
         }
         else{
            res = i0+q-p_.begin();
         }
      }
      else{
         auto  q=std::find(p_.begin(),p_.end(),t);
         if (q==p_.end() || *q!=t){
            res = iEx;
         }
      }
   }
   return res;
}

// modified from function locate of Press et al.

template <class T>
Z V<T>::locAsc(T const& t)const
{
   if (sz_==0){
      cpmerror("V<T>::locAsc(T): array is void");
      return -137; // never happens
   }
   if (t<fir()) return b()-1;
   if (t>=last()) return e();
      // if sz_==1 we have fir()==last and then the two previous
      // conditions are an alternative: one of them holds and we
      // are ready. Thus now sz_>=2
   Z jl=0,ju=sz_;
   while (ju-jl>1){
      Z jm=(jl+ju)/2;
      if (t >= p_[jm]) jl=jm; else ju=jm; // Press et al. have > here
   }
   return jl+b();
}

template <class T>
V<T> V<T>::app(V<T> const& h)const
{
   vector<T> pf;
   pf.reserve(sz_+h.size());
   pf.insert(pf.end(),p_.begin(),p_.end());
   pf.insert(pf.end(),h.p_.begin(),h.p_.end());
   return V<T>(iv_.b(),pf);
}

template <class T>
V<T> V<T>::app(T const& t)const
{
   vector<T> pf(p_);
   pf.push_back(t);
   return V<T>(iv_.b(),pf);
}

template <class T>
V<T>& V<T>::prepend_(T const& t)
{
   p_.push_back(t);
   std::rotate(p_.rbegin(), p_.rbegin() + 1, p_.rend());
   iv_.n_();
   sz_++;
   return *this;
}

template <class T>
V<T> V<T>::rev()const
{
   if (sz_<2){
      return *this;
      // nothing to do if we have no or one component
   }
   else{
      vector<T> p{p_};
      for (Z i=0,j=sz_-1; i<sz_; ++i,--j) p[i]=p_[j];
      return V<T>(b(),p);
   }
}

template <class T>
V<T> V<T>::resize(Z newDim)const
{
   if (newDim<=0) return V<T>();
   vector<T> p(newDim);
   for (Z i=0;i<newDim;++i){
      p[i]= i<sz_ ? p_[i] : T();
   }
   return V<T>(b(),p);
}

template <class T>
V<T> V<T>::eli(IvZ const& iv)const
{
   IvZ ie=iv&dom();
   if (ie.isVoid()) return *this ;
   else{
      V<B> vb(dom());
      for (Z i=vb.b();i<=vb.e();++i){
         vb.cui(i)=!ie.hasElm(i);
      }
      return select(vb);
   }
}

template <class T>
V<T> V<T>::eli1(V<T>& h, Z i)const
{
// Here we use natural counting of components so that
// the j-th component of h is h[j-1].
   Z nEli=h.dim();
   if (i<1) i=1;
   if (sz_==0){ // nothing eliminated, nothing left
      h=V<T>(0);
      return V<T>(0);
   }
   else if (i>sz_){ // nothing eliminated
      h=V<T>(0);
      return *this;
   }
   else{ // now i>=1 and i<=sz_ and sz_>=1
      // if true, we have sz_>=1
   // the i-th component is the first to be eliminated
   // so we have i-1 components of *this which are on the left-hand
   // side of the elimination area. These have to shine up in
   // the result vector to be returned
      Z nRes1=i-1;
   // So the maximum number of
   // components of *this that run the risk to get eliminated is
   // sz_-nRes1
      Z nEliRisk=sz_-nRes1;
      if (nEli>nEliRisk) nEli=nEliRisk;
      Z nRes2=sz_-nRes1-nEli;
      Z nRes=nRes1+nRes2;
      T* pRes=new T[toN(nRes)];
      T* pEli=new T[toN(nEli)];
      T* itRes=pRes;
      T* itEli=pEli;
      T* itp=p_;
      Z j=nRes1;
      while (j--) *itRes++=*itp++;
      j=nEli;
      while (j--) *itEli++=*itp++;
      j=nRes2;
      while (j--) *itRes++=*itp++; // for j==0 nothing done
      h=V<T>(nEli,pEli);
      return V<T>(nRes,pRes);
   }
}

template <class T>
V<T> V<T>::ins(Z i, T const& t)const
{
   Word loc("V<T> V<T>::ins(Z i, T const& t)const");
   vector<T> p{p_};
   Z iLocal=i-b();
   auto it=p.begin()+iLocal;
   p.insert(it,t);
   return V(b(),p);
}

template <class T>
V<T>& V<T>::ins_(Z i, T const& t)
{
   Word loc("V<T> V<T>::ins(Z i, T const& t)const");
 //  vector<T> p{p_};
   Z iLocal=i-b();
   auto it=p_.begin()+iLocal;
   p_.insert(it,t);
   sz_++;
   iv_.n_();
   return *this;
}

template <class T>
V<T> V<T>::ins(Z ia, V<T> const& h)const
{
   Z mL=3;
   Word loc("V<T> V<T>::ins(Z i, V<T> const& h)const");
   CPM_MA
   vector<T> p{p_};
   auto it=p.begin()+ia;
   p.insert(it,h.p_.begin(),h.p_.end());
   return V(b(),p);
}

template <class T>
T V<T>::in_(T const& t, bool reversed)
// safe logic by indexing, propably not utmost performance
{
   if (sz_==0) return T();
   Z i;
   T res;
   if (!reversed){
      res=p_[sz_-1]; // last component of array
      for (i=sz_-1;i>0;i--){
         p_[i]=p_[i-1]; // causal processing: on the right-hand side we
         // evaluate only expressions which were not yet changed during
         // the loop
      }
      p_[0]=t;
   }
   else{
      res=p_[0]; // first component of array
      for (i=0;i<sz_-1;i++){
         p_[i]=p_[i+1]; // causal processing: on the right-hand side we
         // evaluate only expressions which were not yet changed during
         // the loop
      }
      p_[sz_-1]=t;
   }
   return res;
}

// re-indexing

template <class T>
V<T> V<T>::compose(V<Z> const& w)const
{
   V<T> res(iv_);
   for (Z i=w.b();i<=w.e();i++){
     res[i]=operator()(w[i]);
   }
   return res;
}

template <class T>
V<T> V<T>::rot(Z s, Outside mode)const
{
   V<T> res(iv_);
   for (Z i=b();i<=e();++i) res[i]=operator()(i-s,mode);
   return res;
}

template <class T>
X2< V<T>, V<Z> > V<T>::condense(
   bool (*equiv)(const T&, const T&))const
// implementation based on function eclazz of the Numerical Receipes of
// Press et al.
{
   const Z mL=3;
   static Word loc("V<T>::condense()");
   CPM_MA
   Z n=dim();
   if (n<1){ // addition 2002-02-23
      CPM_MZ
      return X2< V<T>, V<Z> >(V<T>(0),V<Z>(0));
   }
   V<Z> res2(n);
   Z k,j;
   res2[1]=1;
   for (j=2;j<=n;j++) {
      res2[j]=j;
      for (k=1;k<=(j-1);k++) {
         res2[k]=res2[res2[k]];
         if ((*equiv)((*this)[j],(*this)[k])) res2[res2[res2[k]]]=j;
      }
   }
   for (j=1;j<=n;j++) res2[j]=res2[res2[j]];
   Z m=-1;
   Z rj;
   for (j=1;j<=n;j++){
      rj=res2[j];
      if (rj>m) m=rj;
   }
   V<Z> aux(m,0); // unfortunately the NR algorithm does not guarantee
      // that there are no gaps between the valid values of rj. Therfore
      // we find out the valid ones by an additional loop
   V<Z> found(m,0);
   for (j=1;j<=n;j++){
      rj=res2[j];
      if(found[rj]==0){
         found[rj]=1;
         aux[rj]=j;
      }
   }
   Z mAct=0;
   for (j=1;j<=m;j++) mAct+=found[j];
   V<T> res1(mAct);
   Z jAct=1;
   for (j=1;j<=m;j++){ // notice that aux[j] was initialized as 0
      if (aux[j]>0) res1[jAct++]=(*this)[aux[j]];
   }
   CPM_MZ
   return X2< V<T>,V<Z> >(res1,res2);
}


template <class T>
void V<T>::set_(T const& t)
{
   for (Z i=0;i<sz_;++i) p_[i]=t;
}

template <class T>
V<T> V<T>::set(Z i, T const& t)const
{
   Z iC=i-1;
   // now iC is a 'C-pointer index'

   if (iC<0) return *this;   // nothing changed
   else if (iC<sz_){
      // vector can already hold the new value
      V<T> res(*this);
      res[iC]=t;
      return res;
   }
   else{
      // now we have to return an enlarged vector
      Z j, sz2=iC+1;
      V<T> res(sz2);
      res[iC]=t;
      T* it=res.p_;
      T* itp=p_;
      Z i=sz_;
      while (i--) *it++ = *itp++;
      return res;
   }
}

template <class T>
T V<T>::fAcc1(T (*f)(T const&), void (*acc)(T&, T const&))const
{
   T res=T();
   for (Z i=0;i<sz_;i++) acc(res,f(p_[i]));
   return res;
}

template <class T>
T V<T>::fAcc2(T (*f)(T const&, T const&), void (*acc)(T&, T const&),
   V<T> const& h)const
{
   T res=T();
   for (Z i=0;i<sz_;i++) acc(res,f(p_[i],h.p_[i]));
   return res;
}

template <class T>
template <class Y>
Y V<T>::fAcc3(Y (*f)(T const&, T const&), void (*acc)(Y&, Y const&),
      V<T> const& h )const
{
   Y res=Y();
   for (Z i=0;i<sz_;i++) acc(res,f(p_[i],h.p_[i]));
   return res;
}

template <class T>
V<T> V<T>::fAcc4(F<T2<T>,T> const& f)const
{
   Z d=dim();
   V< V<T> > aux(d,V<T>(d));
   for (Z i=0;i<d;++i){
      for (Z j=0;j<i;++j){
         T fij=f(T2<T>(li(i),li(j)));
         aux.li(i).li(j)=fij;
         aux.li(j).li(i)=-fij;
      }
   }
   T* p1=new T[toN(sz_)];
   for (Z i=0;i<sz_;i++){
      T res;
      for (Z j=0;j<sz_;j++) res+=aux.li(i).li(j);
      p1[i]=res;
   }
   return V<T>(dom(),p1,"");
}

namespace{

template <class T>
T fFunc5(T2<T> const& xij, F<R,R> const& dPot)
// This is intended to represent the force which particle i feels due to
// particle j being present.
{
   T eij=xij[1]-xij[2]; // eij points from j to i. Thus a positive
   // multiple of eij corresponds to a force on i which drives it away
   // from particle j. This such is the case of a repulsive potential.
   // This has dPot(r)/dr < 0 so that with the sign built into the formula
   // for the return value we in fact have the case of the positive
   // multiple we started with.
   R r=eij.nor_();
   return eij*(-dPot(r));
}

template <class T>
T fFunc6(T const& xi, F<R,R> const& dPot)
{
   T ei=xi; // ei points from the origin to i. Same situation as in fFunc5.
   R r=ei.nor_();
   return ei*(-dPot(r));
}

}

template <class T>
V<T> V<T>::fAcc5(F<R,R> const& dPot)const
{
#ifdef CPM_Fn
   F< T2<T>, T > f = F1< T2<T>, F<R,R>, T >(dPot)(fFunc5<T>);
#else
   F<T2<T>,T> f((std::bind(fFunc5<T>,_1,dPot)));
#endif

   Z d=dim(); // one could call fAcc4 here at the cost of an additional
      // function call. Here we ask for 'utmost efficiency'.
   V< V<T> > aux(d,V<T>(d));
   for (Z i=0;i<d;++i){
      for (Z j=0;j<i;++j){
         T fij=f(T2<T>(li(i),li(j)));
         aux.li(i).li(j)=fij;
         aux.li(j).li(i)=-fij;
      }
   }
   vector<T> p1(toN(sz_));
   for (Z i=0;i<sz_;i++){
      T res;
      for (Z j=0;j<sz_;j++) res+=aux.li(i).li(j);
      p1[i]=res;
   }
   return V<T>(dom(),p1);
}

template <class T>
V<T> V<T>::fAcc6(F<R,R> const& dPot)const
{
#ifdef CPM_Fn
   F<T,T> f = F1<T,F<R,R>,T>(dPot)(fFunc6<T>);
#else
   F<T,T> f(std::bind(fFunc6<T>,_1,dPot));
#endif
   vector<T> p1(toN(sz_));
   for (Z i=0;i<sz_;++i){
      p1[i]=f(li(i));
   }
   return V<T>(dom(),p1);
}

template <class T>
V<T> V<T>::fa(T (*f)(T const&))const
{
   vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i]);
   return V<T>(iv_,p);
}

//template <class T>
//V<T> V<T>::faa(T (*f)(T))const
//{
//   T* p1=new T[toN(sz_)];
//   T* it=p1;
//   T* itp=p_;
//   Z i=sz_;
//   while(i--) *it++ = f(*itp++);
//   return V<T>(dom(),p1,"");
//}

template <class T>
V<T> V<T>::fsf(std::function<T(T)> f)const
{
   vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i]);
   return V<T>(iv_,p);
}

template <class T>
V<T> V<T>::faa(T (*f)(T))const
{
   vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i]);
   return V<T>(iv_,p);
}

template <class T>
V<T> V<T>::fb(T (*f)(T const&, T const&), T const& t)const
{
  vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i],t);
   return V<T>(dom(),p);
}

template <class T>
template <class Y>
V<T> V<T>::fb2(T (*f)(T const&, Y const&), Y const& y)const
{
   vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i],y);
   return V<T>(dom(),p);
}

template <class T>
template <class Y>
void V<T>::fb2_(T (*f)(T const&, Y const&), Y const& y)
{
   for (N i=0;i<n_;++i) p_[i]=f(p_[i],y);
}

template <class T>
void V<T>::fc_(T (*f)(T const&, T const&), T const& t)
{
   for (N i=0;i<n_;++i) p_[i]=f(p_[i],t);
}

template <class T>
void V<T>::fd(T (*f)(T const&, T const&), T& t)const
{

   for (N i=0;i<n_;++i) t=f(p_[i],t);
}

template <class T>
V<T> V<T>::fe(T (*f)(T const&, T const&), V<T> const& h)const
{
   if (!sameDom(h)) throw Error("V<T>::fe(): domain mismatch");
   vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i],h.p_[i]);
   return V<T>(dom(),p);
}

template <class T>
template <class Y>
V<T> V<T>::fet(T (*f)(T const&, Y const&), V<Y> const& h)const
{
   if (dom()!=h.dom()) throw Error("V<T>::fet(): domain mismatch");
   vector<T> p=p_;
   for (N i=0;i<n_;++i) p[i]=f(p_[i],h.pi(i));
   return V<T>(dom(),p);
}

template <class T>
V<T> V<T>::fe2(T (*f)(T const&, T const&), V<T> const& h)const
{
   IvZ iv1{iv_};
   IvZ iv2{h.iv_};
   IvZ iv = iv1|iv2;
   V<T> res(iv);
   for (Z i=iv.b();i<=iv.e();++i){
      if (iv1.hasElm(i)&&iv2.hasElm(i)) res[i]=f((*this)[i],h[i]);
   }
   return res;
}

template <class T>
void V<T>::ff_(T (*f)(T const&, T const&), V<T> const& h, Z is)
{
   if (is<0) is=0;
   for (Z i=0; i+is<sz_; ++i){
      if (h.valInd(i)) p_[i+is]=f(p_[i+is],h.p_[i]);
   }
}

template <class T>
void V<T>::ffl_(T (*f)(T const&, T const&), V<T> const& h)
{
   for (N i=0; i<n_; ++i){
      p_[i]=f(p_[i],h.p_[i]);
   }
}

template <class T>
void V<T>::fa_(F<T,T> const& f, IvZ iv)
{
   for (Z i=b();i<=e();++i){
      if (iv.hasElm(i)) (*this)[i]=f((*this)[i]);
   }
}

template <class T>
void V<T>::fg_(T (*f)(T const&, T const&, T const&),
      V<T> const& h, T const& t)
{
   cpmassert(sameDom(h),"dimension mismatch in V::fg_(f,V,T)");
   for (Z i=0;i<sz_;++i) p_[i]=f(p_[i],h.p_[i],t);
}

template <class T >
void V<T>::fh_(T const& w_1, T const& w0, T const& w1, Outside mode)
{
   if (sz_<2) return;
   T v_1=read(b()-1,mode);
   T v0=p_[0];
   T v1=p_[1];
   Z k=0;
   while (k<sz_){
      p_[k]=v_1*w_1+v0*w0+v1*w1;
      k++;
      v_1=v0;
      v0=v1;
      v1=(k < sz_-1 ? p_[k+1] : read(e()+1,mode));
   }
}

template <class T >
template <class Y>
void V<T>::fht_(Y const& w_1, Y const& w0, Y const& w1, Outside mode)
{
   if (sz_<2) return;
   T v_1=read(b()-1,mode);
   T v0=p_[0];
   T v1=p_[1];
   Z k=0;
   while (k<sz_){
      p_[k]=v_1*w_1+v0*w0+v1*w1;
      k++;
      v_1=v0;
      v0=v1;
      v1=(k < sz_-1 ? p_[k+1] : read(e()+1,mode));
   }
}

template <class T >
template <class Y>
V<T> V<T>::fh(T (*f)(T const&, T const&, T const&, Y const&, Z),
      Y const& y, Outside mode)const
{
   N n=nc();
   if (n<2) return *this;
   V<T> q=*this;
   V<T> p=*this;
   for (N i=0; i<n;++i){
      p.pi(i)=f(q.pr(i-1,mode),q.pr(i,mode),q.pr(i+1,mode),y,i);
   }
   return p;
}

template <class T >
template <class Y>
V<Y> V<T>::fi(Y (*f)(T const&, T const&), void (*acc)(Y&, Y const&))const
{
   Z i,j;
   Y* q=new Y[toN(sz_)];
   for (i=0;i<sz_;i++){
      Y yi=Y();
      for (j=0;j<sz_;j++){
         acc(yi,f(p_[i],p_[j]));
      }
      q[i]=yi;
   }
   return V<Y>(dom(),q);
}

template <class T >
template <class Y>
V<Y> V<T>::fj(Y (*f)(T const&, T const&), void (*acc)(Y&, Y const&))const
{
   Z i,j;
   Y* q=new Y[toN(sz_)];
   for (i=0;i<sz_;i++) q[i]=Y();
   for (i=0;i<sz_;i++){
      for (j=0;j<i;j++){
         Y fij=f(p_[i],p_[j]);
         acc(q[i],fij);
         acc(q[j],-fij);
      }
   }
   return V<Y>(dom(),q);
}

template <class T >
V<T> V<T>::select(V<B> const& s)const
{
   vector<T> pRes;
   for (Z i=b();i<=e();++i){
      bool keep{true};
      if (s.valInd(i)&&s[i]==false) keep=false;
      if (keep) pRes.push_back((*this)[i]);
   }
   return V<T>(b(),pRes);
}

template <class T >
V<IvZ> V<T>::valOn(F<T,bool> const& f)const
{
   Z mL=3;
   static Word loc("V<T>::valOn(F<T,bool>)");
   CPM_MA
   V<IvZ> res;
   bool yetFoundTrue=false;
   Z firstTrue=0, firstFalseAfterTrue=0;
   for (Z i=b();i<=e();++i){
      bool val=f((*this)[i]);
      if (val){ // we found true
         if (!yetFoundTrue){ // then start a interval of validity
            yetFoundTrue=true;
            firstTrue=i;
         }
         else{ // normally nothing to do
            // but if we are at the end of the array the
            // last pending truth interval has to be
            // considered as finished and has to be added
            if (i==e()){
               firstFalseAfterTrue=n();
               IvZ ivAct(firstTrue,firstFalseAfterTrue-1);
               res&=ivAct; // appending
               // nothing else to do since we are finished
               CPM_MZ
               return res;
            }
         }
      }
      else{ // we found false
         if (yetFoundTrue){ // then i is the terminator
            // of a truth interval
            firstFalseAfterTrue=i;
            IvZ ivAct(firstTrue,firstFalseAfterTrue-1);
            res&=ivAct;
            yetFoundTrue=false;
         }
      }
   }
   CPM_MZ
   return res;
}

template <class T>
bool V<T>::prnOn(ostream& str)const
{
   Z mL=3;
   Word loc=nameOf()&"::prnOn(...)";
   CPM_MA
   cpmwt((nameOf()&" begin").str());
   Root<IvZ>(iv_).prnOn(str);
   if (iv_.car()==0){
      Word w(" no components exist");
      cpmcerr<<w<<endl;
      cpmcerr<<" iv_.b() = "<<iv_.b()<<endl;
      goto label;
   }
   for (Z i=b();i<=e();i++){
      if (CpmRoot::wrtTit){
         Word wi("// i="); // added 2012-01-19
         // same as in S<>
         wi&=cpm(i);
         bool bi=wi.prnOn(str);
         cpmassert(bi==true,loc);
      }
      if (!Root<T>((*this)[i]).prnOn(str)){
         cpmwarning("failed to write component indexed "&cpm(i));
         cpmwarning("index range is from "&cpm(b())&" to "&cpm(e()));
         CPM_MZ
         return false;
      }
   }
label:
   cpmwt((nameOf()&" end").str());
      // for large sz_ it would be difficult to find the
      // end of the vector data if these are written to a file
      // and a human reader wants to inspect them
   CPM_MZ
   return true;
}

template <class T>
bool V<T>::scanFrom(istream& str)
{
   Z mL=3;
   Word loc=nameOf()&"::scanFrom(...)";
   CPM_MA
   Root<IvZ> ivIn;
   bool suc = ivIn.scanFrom(str);
   if (!suc){
      cpmwarning(loc&": can't read IvZ");
      CPM_MZ
      return false;
   }
   IvZ iv=ivIn();
   Z n=iv.car();
   cpmmessage(mL,"dimension read as "&cpm(n));
   if (n>dimMax) cpmwarning(loc&": n>dimMax");
   V<T> res(iv);
   Root<T> riIn;
   for (Z i=res.b();i<=res.e();i++){
      suc = riIn.scanFrom(str);
      if (!suc){
         Word mes="failed to read component indexed "&cpm(i)&
             " index range is from "&cpm(res.b())&" to "&cpm(res.e());
         cpmwarning(mes);
         CPM_MZ
         return false;
      }
      res.cui(i) = riIn();
   }
   *this=res;
   CPM_MZ
   return true;
}

template <class T >
Z V<T>::com(V<T> const& s)const
// short vectors < longer vectors
{
   Z d1=dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
       for (Z i=0;i<d1;i++){
         Z ci=Root<T>(p_[i]).com(s.p_[i]);
         if (ci!=0) return ci;
      }
      return 0;
   }
}
/*************************** V<bool> ***********************************/
// since std::vector<bool> has a quite irregular special definition we get
// a lot of compilation errors if we use existing C+- code the
// instantiation V<bool> while V<B> works fine.
/*
template<>
class V<bool>{
   IvZ iv_;
   std::vector<CpmRoot::B> p_;
public:
   B const& cui(Z i)const{return p_[iv_.ri(i)];}
   B& cui(Z i){return p_[iv_.ri(i)];}
   B const& operator[](Z i)const{return p_[iv_.ri(i)];};
   B& operator[](Z i){return p_[iv_.ri(i)];};
};
*/

/*
B const& CpmArrays::V<bool>::operator[](Z i)const
{
   if (ranChcAlw){
      if (!iv_.hasElm(i)) cpmerror("index out of range");
   }
   return p_[iv_.ri(i)];
}


B& CpmArrays::V<bool>::operator[](Z i)
{
   if (ranChcAlw){
      if (!iv_.hasElm(i)) cpmerror("index out of range");
   }
   return p_[iv_.ri(i)];
}
*/


/************************ iterated templates **************************/
// describing multi-indexed quantities
// Notice that these all have automatically the member functions of V
// defined!

#define CPM_V1 V<T>
#define CPM_V2 V<V<T> >
#define CPM_V3 V<V<V<T> > >
#define CPM_V4 V<V<V<V<T> > > >

/************************** class VV ***********************************/
template <class T>
class VV: public CPM_V2{ // matrices

   typedef CPM_V2 Base;

public:

// constructors

   VV(Z d1, Z d2):CPM_V2(d1,CPM_V1(d2)){}
   VV(Z d1, Z d2, T const& t):CPM_V2(d1, CPM_V1(d2,t)){}
      // all components are equal to t
   VV(void):CPM_V2(){}
   VV(CPM_V2 const& x):CPM_V2(x){}

   V<T> lin()const;
      // 'linear version of matrix'
      // by appending all rows

// dimension access
   Z dim1()const{ return Base::dim();}
   Z dim2()const{ return (*this)[1].dim();}

   virtual Z size()const // virtual in base V<...>
      { return dim1()==0 ? 0 : dim1()*dim2();}

// component access
   const T& operator()(Z i, Z j)const
      // returns T() for out of range indexes
   { return Base::read(i).read(j);}

   T& operator()(Z i, Z j)
   { return (*this)[i][j];}

   VV<T> trn()const
      //: transpose
      // Returns the transposed matrix
   {
      VV<T> res(dim2(),dim1());
      for (Z i=1;i<=dim1();i++){
         for (Z j=1;j<=dim2();++j){
            res[j][i] = (*this)[i][j];
         }
      }
      return res;
   }

   template <class Y>
   V<Y> each(void (*f)(T const&, Y&))const;
      //: each
      // collecting the application of f on every pixel in a linear array

   template <class Y>
   VV<Y> operator()(Y (*f)(T const&))const;
      //: operator()
      // creating a matrix with a transformed value range
};

template <class T>
template <class Y>
V<Y> VV<T>::each(void (*f)(T const&, Y&))const
{
   Z m1=dim1(),m2=dim2(),k=1,i1,i2;
   V<Y> res(m1*m2);
   for (i1=1;i1<=m1;i1++){
      for (i2=1;i2<=m2;i2++){
         f((*this)[i1][i2],res[k++]);
      }
   }
   return res;
}

template <class T>
template <class Y>
VV<Y> VV<T>::operator()(Y (*f)(T const&))const
{
   Z m=dim1(),n=dim2(),i,j;
   VV<Y> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         res[i][j]=f((*this)[i][j]);
      }
   }
   return res;
}

template <class T>
V<T> VV<T>::lin()const
{
   if (dim1()==0) return V<T>(0);
   else{
      V<T> res=(*this)[1];
      for (Z i=2;i<=dim1();i++) res&=(*this)[i];
      return res;
   }
}

/************************** class VVV***********************************/
// tensors of rank 3

template <class T>
class VVV: public CPM_V3{ // tensors of rank 3

      typedef CPM_V3 Base;

public:

// constructors

   VVV(Z d1, Z d2, Z d3, T  const& t):CPM_V3(d1,CPM_V2(d2,CPM_V1(d3,t))){}
   VVV(Z d1, Z d2, Z d3):CPM_V3(d1,CPM_V2(d2,CPM_V1(d3))){}
   VVV(void):CPM_V3(){}
   VVV(CPM_V3 const& x):CPM_V3(x){}

// dimension access
   Z dim1()const{ return Base::dim();}
   Z dim2()const{ return (*this)[1].dim();}
   Z dim3()const{ return (*this)[1][1].dim();}

   virtual Z size()const; // virtual in base V<...>

// component access

   const T & operator()(Z i, Z j, Z k)const
   // returns T() for out of range indexes
   { return Base::read(i).read(j).read(k);}

   T & operator()(Z i, Z j, Z k)
   { return (*this)[i][j][k];}
};

template <class T>
Z VVV<T>::size()const
{
   Z d1=dim1();
   if (d1==0) return d1;
   else{
      Z d2=dim2();
      if (d2==0) return d2;
      else{
         Z d3=dim3();
         if (d3==0) return d3;
         else return d1*d2*d3;
      }
   }
}

/************************ class VVVV ***********************************/
template <class T>
class VVVV: public CPM_V4{ // tensors of rank 4

      typedef CPM_V4 Base;

public:

// constructors

   VVVV(Z d1, Z d2, Z d3, Z d4, T const& t):
   CPM_V4(d1,CPM_V3(d2,CPM_V2(d3,CPM_V1(d4,t)))){}
   VVVV(Z d1, Z d2, Z d3, Z d4):
   CPM_V4(d1,CPM_V3(d2,CPM_V2(d3,CPM_V1(d4)))){}
   VVVV(void):CPM_V4(){}
   VVVV(CPM_V4 const& x):CPM_V4(x){}

// dimension access
   Z dim1()const{ return Base::dim();}
   Z dim2()const{ return (*this)[1].dim();}
   Z dim3()const{ return (*this)[1][1].dim();}
   Z dim4()const{ return (*this)[1][1][1].dim();}

   virtual Z size()const; // virtual in base V<...>

// component access

   T const& operator()(Z i, Z j, Z k, Z l)const
  // returns T() for out of range indexes
   { return Base::read(i).read(j).read(k).read(l);}

   T& operator()(Z i, Z j, Z k, Z l)
   { return (*this)[i][j][k][l];}
};

template <class T>
Z VVVV<T>::size()const
{
   Z d1=dim1();
   if (d1==0) return d1;
   else{
      Z d2=dim2();
      if (d2==0) return d2;
      else{
         Z d3=dim3();
         if (d3==0) return d3;
         else{
            Z d4=dim4();
            if (d4==0) return d4;
            else return d1*d2*d3*d4;
         }
      }
   }
}
#undef CPM_V4
#undef CPM_V3
#undef CPM_V2
#undef CPM_V1

}  // namespace CpmArrays

namespace CpmRoot{
// This is a pattern for partial specializations --- to which
// the specializations to class templates belong --- of the
// IO class template started in cpmnumbers.h and the
// Name class template started in cpmword.h.
// A corresponding definition is needed for all non-C+- classes which
// shall be used as template arguments of C+- containers and
// are expected to provide then the same functionality as
// corresponding C+- classes.

   template<class T>
   class IO< std::vector<T> >{ // partial specialization
    public:
      IO(){}
      bool o(std::vector<T> const& v, ostream& str)const
      { return CpmArrays::V<T>(v).prnOn(str);}
      bool i(std::vector<T>& v, istream& str)const
      {
         CpmArrays::V<T> vl;
         bool res=vl.scanFrom(str);
         v=vl.std();
         return res;
      }
   };

#if defined(CPM_NAMEOF)
   template<class T>
   class Name< std::vector<T> >{ // partial specialization
   public:
      Name(){}
      Word operator()(std::vector<T> const& vt )const
      { return Word("std::vector<"&CpmRoot::Name<T>()(T())&">");}
   };
   template<class T>
   class Name< std::set<T> >{ // partial specialization
   public:
      Name(){}
      Word operator()(std::set<T> const& vt )const
      { return Word("std::set<"&CpmRoot::Name<T>()(T())&">");}
   };
#endif
}
#endif
