//? cpmvectormore.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmvectormore.h>
#include <cpmmini.h>
#include <cpmalgorithms.h>
#include <cpmfile.h>

using namespace CpmRoot;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmStd;
using CpmAlgorithms::Percentilizer;
using CpmGeo::Iv;
using std::vector;

R2 CpmArrays::lineFit(R_Vector const& x, R_Vector const& y,
   R_Vector const& w)
// Modified, and considerably stripped in functionality, from function
// fit of Press et al.
{
   Z mL=3;
   static Word loc("CpmArrays::fitLine(...)");
   CPM_MA
   Z i;
   Z n=CpmRootX::inf<Z>(x.dim(),y.dim(),w.dim());
   R ss=0,sx=0,sy=0;
   for (i=1;i<=n;i++) {
      R wi=cpmabs(w[i]);
      ss += wi;
      sx += x[i]*wi;
      sy += y[i]*wi;
   }
   R sxoss=sx/ss;
   R st2=0,b=0;
   for (i=1;i<=n;i++) {
      R vi=cpmsqrt(cpmabs(w[i]));
      R t=(x[i]-sxoss)*vi;
      st2 += t*t;
      b += t*y[i]*vi;
   }
   b/=st2;
   R a=(sy-sx*b)/ss;
   CPM_MZ
   return R2(a,b);
}

R_Vector CpmArrays::polyFit(R_Vector const& x, R_Vector const& y,
   Z deg, R x0, R r, bool autoAdjust)
{
   using CpmGeo::Iv;
   using CpmRoot::inv;
     Z n=x.dim(),i;
   if (y.dim()<n)
      cpmerror(
      "polyFit(R_Vector const&, R_Vector const&, Z): dimension mismatch");
   if (autoAdjust){
      Iv iv(x[1],x[n]);
      x0=iv.mean();
      r=iv.radius();
   }
   R rInv=inv(r);
   R_Vector xc(n);
   for (i=1;i<=n;i++) xc[i]=(x[i]-x0)*rInv;
   Z d=deg+1;
   R_Matrix a(d,n);
   a[1]=R_Vector(n,1.);
   for ( i=2;i<=d;i++) a[i]=a[i-1]*xc;
   R_Vector res(2);
   res[1]=x0;
   res[2]=rInv;
   res=res&(!~a*y);
   return res;
}

R CpmArrays::evalPoly(const R& x, R_Vector const& a)
{
   Z n=a.dim();
   R x0=a[1];
   R rInv=a[2];
   R y=a[3];
   R xi=(x-x0)*rInv;
   R xip=xi;
      // power of xi
   for (Z i=4;i<=n;i++){
      y+=a[i]*xip;
      xip*=xi;
   }
   return y;
}

R_Vector CpmArrays::linFit(const V< F<R,R> >& f, R_Vector const& x,
   const R_Vector& y)
{
     Z n=x.dim();
   if (y.dim()<n)
      cpmerror("linFit(const V<F<R,R> >& ,...): dimension mismatch");
   Z m=f.dim(),i,j;
   R_Matrix a(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         a[i][j]=f[i](x[j]);
      }
   }
   return (!~a*y);
}

namespace{

   R auxFunc(R_Vector const& p, const F<X2<R,R_Vector>,R>& f,
      R_Vector const& xList, R_Vector const& yList)
   {
      R sum=0;
      Z n=xList.dim(),i;
      for (i=1;i<=n;i++){
         R val=f(X2<R,R_Vector>(xList[i],p))-yList[i];
         sum+=(val*val);
      }
      return sum;
   }
}

R_Vector CpmArrays::nonLinFit(const F<X2<R,R_Vector>,R>& f,
   R_Vector const& x, R_Vector const& y, R_Vector const& pStart,
   R fracTol, bool verbose)
{
#ifdef CPM_Fn
   using CpmFunctions::F3;
#endif
   using CpmAlgorithms::Mini;
     Z n=x.dim();
   if (y.dim()<n)
      cpmerror("nonLinFit(...): dimension mismatch");
#ifdef CPM_Fn
   Mini mini(F3<R_Vector, F<X2<R,R_Vector>,R>,
      R_Vector, R_Vector,R>(f,x,y)(auxFunc),fracTol);
#else
   Mini mini(F<R_Vector,R>(bind(auxFunc,_1,f,x,y)),fracTol);
#endif
   mini.verbose=(verbose ? 1 : 4);
   return mini.getPoint(pStart);
}

// Interpolation functions

R CpmArrays::polyInte_4(R_Vector const& xa, R_Vector const& ya, R x,
   R& intError)
{
   Z n=xa.dim();
   cpmassert(n<=ya.dim(),"CpmArrays::polyInte(...)");
   Z i,m,ns=1;
   R den,dif,dift,ho,hp,w;
   dif=cpmabs(x-xa[1]);
   R_Vector c(n);
   R_Vector d(n);
   R y,dy=0;
   for (i=1;i<=n;i++) {
      if ( (dift=cpmabs(x-xa[i])) < dif) {
         ns=i;
         dif=dift;
      }
      c[i]=ya[i];
      d[i]=ya[i];
   }
   y=ya[ns--];
   for (m=1;m<n;m++) {
      for (i=1;i<=n-m;i++) {
         ho=xa[i]-x;
         hp=xa[i+m]-x;
         w=c[i+1]-d[i];
         if ( (den=ho-hp) == 0.0){
            cpmdebug(xa);
            cpmerror(
               "Error in function polyInt(): two x-values coincide");
         }
         den=w/den;
         d[i]=hp*den;
         c[i]=ho*den;
      }
      y += (dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
   }
   intError=dy;
   return y;
}

R CpmArrays::polyInte(R_Vector const& xa, R_Vector const& ya, R x)
{
   R intError=1;
   return polyInte_4(xa,ya,x,intError);
}

R CpmArrays::ratiInte_4(R_Vector const& xa, R_Vector const& ya, R x,
   R& intError)
{
   const R hugeUM=1e30;
   const R tinyUM=1e-30;
   Z n=xa.dim();
   Word loc("ratiInte(...)");
   cpmassert(n<=ya.dim(),loc);
   Z m,i,ns=1;
   R w,t,hh,h,dd;
   R_Vector c(n);
   R_Vector d(n);
   hh=cpmabs(x-xa[1]);
   R y,dy=0;
   for (i=1;i<=n;i++) {
      h=cpmabs(x-xa[i]);
      if (h == 0.0) {
         y=ya[i];
         dy=0.0;
            intError=dy;
         return y;
      } else if (h < hh) {
         ns=i;
         hh=h;
      }
      c[i]=ya[i];
      d[i]=ya[i]+tinyUM;
   }
   y=ya[ns--];
   for (m=1;m<n;m++) {
      for (i=1;i<=n-m;i++) {
         w=c[i+1]-d[i];
         h=xa[i+m]-x;
         cpmassert(h!=0.,loc);
         t=(xa[i]-x)*d[i]/h;
         dd=t-c[i+1];
         if (dd == 0.){
            cpmwarning("ratiInte(...): pole found at argument position");
            y=hugeUM;
            dy=hugeUM;
            intError=dy;
            return y;
         }
         dd=w/dd;
         d[i]=c[i+1]*dd;
         c[i]=t*dd;
      }
      y += (dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
   }
   intError=dy;
   return y;
}

R CpmArrays::ratiInte(R_Vector const& xa, R_Vector const& ya, R x)
{
   R intError=1;
   return ratiInte_4(xa,ya,x,intError);
}

Z  CpmArrays::locate(R_Vector const& x, R x_)
{
   Z n=x.dim();
   R xb=x[1];
   R xe=x[n]; // for shortness
   cpmassert(xb<xe,"CpmArrays::locate(R_Vector,R)");

   if (x_>=xe) return n;
   if (x_<xb)   return 0;

   Z k=cpmtoz(((x_-xb)/(xe-xb))*(n-1) + 1.);
      // varies from 1 to n for x_ from x[1] to x[n];
      // k increases by one if x_ increases by (xe-xb)/(n-1)
   Z j=k;

   if ( x_==x[k] ){    // for equidistant lists this is the case
      ;
   }
   else if ( x_ > x[k] ){         // then there is a chance that
                           // j can be increased
      while ( x_ >= x[j+1] ) j++;
   }                        // x_>=x[j] has to hold and j has to be
                           // the largest value for this to hold
   else { // then x_ < x[j]      thus j has to be decreased
      while ( x_ < x[j] ) j--;        // Since x_ has to be >= x[j],
   }
   return j;
}

namespace{
   void powFast(R& w, R p)
      // changes w into w^p
   {
      if (p==1.);
      else if (p==2.) w*=w;
      else if(p==3.) w=w*w*w;
      else w=cpmpow(w,p);
   }
}

R CpmArrays::locatePeak(R_Vector const& x, R p)
{
   R a=0, b=0, w;
   Z n=x.dim();
   for (Z i=1;i<=n;i++){
      w=cpmabs(x[i]);
      powFast(w,p);
      a+=w*i;
      b+=w;
   }
   if (b==0.) return 0; else return a/b;
}


R2 CpmArrays::peakAndSigma(R_Vector const& x, R p)
{
   Z mL=4;
   static Word loc("CpmArrays::peakAndSigma(R_Vector,R)");
   cpmmessage(mL, loc&" started");
   Z i,n=x.dim();
   V<R> distr(n);
   cpmassert(n>0,loc);
   R sum=0;
   for (i=1;i<=n;i++){
      R w=cpmabs(x[i]);
      powFast(w,p);
      sum+=w;
      distr[i]=w;
   }
   cpmassert(sum>0,loc);
   R sumInv=R(1)/sum;
   for (i=1;i<=n;i++) distr[i]*=sumInv;
   R iPeak=0;
   for (i=1;i<=n;i++) iPeak+=R(i)*distr[i];
   R sigma=0;
   for (i=1;i<=n;i++){
      R di=R(i)-iPeak;
      di*=di;
      sigma+=di*distr[i];
   }
   cpmassert(sigma>0,loc);
   sigma=cpmsqrt(sigma);
   cpmmessage(mL, loc&" done");
   return R2(iPeak,sigma);
}

R_Vector CpmArrays::smoothen(R_Vector const& x, bool periodic)
{
   Z i,n=x.dim();
   R_Vector res(n);
   if (!periodic){
      for (i=2;i<n;i++){
         res[i]=x[i-1]*0.25+x[i]*0.5+x[i+1]*0.25;
      }
      res[1]=x[1]*0.75+x[2]*0.25;
      res[n]=x[n]*0.75+x[n-1]*0.25;
   }
   else{
      for (i=1;i<=n;i++){
         res[i]=x.cyc(i-1)*0.25+x[i]*0.5+x.cyc(i+1)*0.25;
      }
   }
   return res;
}

R_Vector CpmArrays::smoothen(R_Vector const& x, Z p, bool periodic)
{
   if (p<1) return x;
   R_Vector res=smoothen(x);
   Z pDone=1;
   while (p>pDone++) res=smoothen(res,periodic);
   return res;
}

X2< V<Z>, V<Z> >  CpmArrays::locExt(R_Vector const& x)
{
   Z n=x.dim();
   V<Z> minima(0),maxima(0);
   for (Z i=2;i<n;++i){ // active only for n > 2
      R xm=x[i-1];
      R x0=x[i];
      R xp=x[i+1];
      if (x0<xm && x0<xp) minima&=i;
      else if (x0>xm && x0>xp) maxima&=i;
      else ;
   }
   return X2< V<Z>,V<Z> >(minima,maxima);
}

R_Vector CpmArrays::derive(R_Vector const& x)
{
   Z n=x.dim();
   R_Vector res(n);
   for (Z i=1;i<n;i++){
      res[i]=x[i+1]-x[i];
   }
   res[n]=0;
   return res;
}

T2<R_Vector> CpmArrays::derive(R_Vector const& x, R_Vector const& y)
{
   Z n=x.dim();
   cpmassert(n==y.dim(),"CpmArrays::derive(R_Vector, R_Vector)");
   Z m=n-1;
   R_Vector xc(m), dy(m);
   for (Z i=1;i<=m;i++){
      xc[i]=(x[i]+x[i+1])*0.5;
      R dx=x[i+1]-x[i];
      cpmassert(dx>0.,"CpmArrays::derive(R_Vector, R_Vector)");
      dy[i]=(y[i+1]-y[i])/dx;
   }
   return T2<R_Vector>(xc,dy);
}

R_Vector CpmArrays::integrate(R x1, R_Vector const& x)
{
   Z n=x.dim();
   R_Vector res(n);
   R s=x1;
   for (Z i=1;i<=n;i++){
      res[i]=s;
      s+=x[i]; // x[n] is not used, is 0 if x was obtained by
   }
   return res;
}


R_Vector CpmArrays::accumulate(R_Vector const& x, bool average)
{
   Z n=x.dim();
   if (n<2) return x;
   R_Vector res(n);
   res[1]=x[1];
   for (Z i=2;i<=n;i++){
      res[i]=res[i-1]+x[i];
   }
   if (average){
      for (Z i=2;i<=n;i++){
         R iInv=1./i;
         res[i]*=iInv;
      }
   }
   return res;
}

R CpmArrays::extMean(R_Vector const& x)
{
   Word loc("CpmArrays::extMean(R_Vector)");
   Z sm=3;
   R_Vector y=smoothen(x,sm);
   y=accumulate(y,true);
   X2< V<Z>,V<Z> > xvv=locExt(y);
   V<Z> vMin=xvv.c1();
   V<Z> vMax=xvv.c2();
   Z dMin=vMin.dim(), dMax=vMax.dim();
   if (dMin==0 || dMax==0){
      cpmwarning(loc&": no extrapolation possible, plain mean taken");
      return y.last();
   }
   else{
      R lastMin=y[vMin.last()];
      R lastMax=y[vMax.last()];
      return (lastMin+lastMax)*0.5;
   }
}

void CpmArrays::makePositive(R_Vector& x, R tiny)
{
   Z n=x.dim();
   for (Z i=1;i<=n;i++){
      if (x[i]<tiny) x[i]=tiny;
   }
}

R_Vector CpmArrays::reverse(R_Vector const& x)
{
   Z n=x.dim();
   R_Vector res(n);
   Z n1=n+1;
   for (Z i=1;i<=n;i++){
      res[i]=x[n1-i];
   }
   return res;
}

R_Vector CpmArrays::monotonize(R_Vector const& x, Z pSmooth)
{
   R_Vector temp=smoothen(x,pSmooth);
      // Raw data might not yet be suitable for further processing,
      // thus they will be regularized.
      // Smoothen and derive commute, so it makes no sense to
      // smoothen the derivative got in a later step
   R peak=temp[1];
   Z n=x.dim();
   for (Z i=1;i<=n;i++){
      peak=(peak>temp[i] ? peak : temp[i]);
      temp[i]=peak;
   }
   return temp;
}

R CpmArrays::interpolate(R ic, R_Vector const& x)
{
   Z n=x.dim();
   if (ic<=1.){
      return x[1];
   }
   else if (ic>=n){
      return x[n];
   }
   else{
      R xiL=cpmfloor(ic);
      Z iL=cpmtoz(xiL);
      cpmassert(iL<n,"CpmArrays::interpolate()");
      R p=ic-xiL;
      return x[iL]+p*(x[iL+1]-x[iL]);
   }
}

R CpmArrays::interpolate(R xp, R_Vector const& x, R_Vector const& y)
{
   Word loc("CpmArrays::interpolate(...)");
   Z n=x.dim();
   Z n2=y.dim();
   cpmassert(n==n2,loc);
   if (xp<=x[1]){
      return y[1];
   }
   else if (xp>=x[n]){
      return y[n];
   }
   else{
      Z iL=locate(x,xp);
      R xL=x[iL];
      R xU=x[iL+1];
      R yL=y[iL];
      R yU=y[iL+1];
      R dx=xU-xL;
      cpmassert(dx!=0,loc);
      R p=(xp-xL)/dx;
      return yL+p*(yU-yL);
   }
}

R CpmArrays::interpolateInverse(R xc, R_Vector const& x)
{
   Word loc("CpmArrays::interpolateInverse(...)");
   Z n=x.dim();
   if (xc<=x[1]) return 1.;
   if (xc>=x[n]) return n;
   Z i=locate(x,xc);
   cpmassert(i<n && i>0,loc);
   R nom=(x[i+1]-x[i]);
   cpmassert(nom>0.,loc);
   R p=(xc-x[i])/nom;
   cpmassert(p<1. && p>=0.,loc);
   return R(i)+p;
}

R_Vector CpmArrays::moments(R_Vector const& data)
{
   Z mL=3;
   static Word loc("R_Vector CpmArrays::moments(R_Vector const& data)");
   cpmmessage(mL,loc&" started");
   Z j,n=data.dim();
   cpmassert(n>0,loc);
   if (n==1){
      R_Vector res(6,0);
      res[1]=data[1];
      return res;
   }
   R s=0.,p;
   for (j=1;j<=n;j++) s += data[j];
   R ave=s/n;
   R adev=0,svar=0,skew=0,curt=0;
   for (j=1;j<=n;j++) {
      adev += cpmabs(s=data[j]-ave);
      svar += (p=s*s);
      skew += (p*=s);
      curt += (p*=s);
   }
   adev /= n;
   svar /= (n-1);
   R sdev=cpmsqrt(svar);
   if (svar!=0.) {
      skew /= (R(n)*svar*sdev);
      curt=curt/(R(n)*svar*svar)-3.0;
   }
   else{
      cpmwarning(
         "moment(): variance = 0 found, skew and curt set to 0 ",1);
           // number of warnings should be limited
      skew=0;
      curt=0;
   }
   R_Vector res(6);
   res[1]=ave;
   res[2]=adev;
   res[3]=sdev;
   res[4]=svar;
   res[5]=skew/cpmsqrt(R(15)/n);
   res[6]=curt/cpmsqrt(R(96)/n);
   cpmmessage(mL,loc&" done");
   return res;
}

/////////////////////////////// module statistics ////////////////////////

namespace{ // anonymous
R probks(R alam)
// Auxiliar function to ksone.
{
   const R EPS1=0.001,EPS2=1.0e-8;
   R fac=2.,sum=0.,term,termbf=0.;
   R a2 = alam*alam*(-2);
   for (Z j=1;j<=100;j++) {
      term=fac*cpmexp(a2*R(j)*R(j));
      sum += term;
      if (cpmabs(term) <= EPS1*termbf || cpmabs(term) <= EPS2*sum)
         return sum;
      fac = -fac;
      termbf=cpmabs(term);
   }
   return 1.0;
}

void ksone(R_Vector& data, R (*func)(R x), R& d, R& prob)
/*
Given an array data[1..n], and given a user-supplied function of a single
variable func which is a cumulative distribution function ranging from 0
(for smallest values of its argument) to 1 (for largest values of its
argument), this routine returns the 'Komogorov-Smirnov statistic' d, and
the significance level prob. Small values of prob show that the
cumulative function of data is significantly different from func.
The array data is modified by being sorted into ascending order.
*/

{
   Z n=data.dim();
   R fo=0.,fn,ff,en,dt;
   data.order_();
   en=n;
   d=0.;
   for (Z j=1;j<=n;j++) {
      fn=R(j)/en;
      ff=(*func)(data[j]);
      dt = CpmRootX::sup(cpmabs(fo-ff),cpmabs(fn-ff));
      if (dt > d) d=dt;
      fo=fn;
   }
   prob=probks(cpmsqrt(en)*d);
}

R erfcc(R x)
{
   R t,z,ans;

   z=cpmabs(x);
   t=R(1)/(z*0.5 +1);
   ans=t*cpmexp(-z*z-R(1.26551223)+t*(R(1.00002368)+t*(R(0.37409196)+
      t*(R(0.09678418)+
      t*(R(-0.18628806)+t*(R(0.27886807)+t*(R(-1.13520398)+
      t*(R(1.48851587)+t*(R(-0.82215223)+t*0.17087277)))))))));
   return (x >= 0.0 ? ans : R(2)-ans);
}

R mean_stat=0., sigma_stat=0.;
R isr2_stat=R(1)/cpmsqrt(2.);
R cum_stat(R x)
// cumulative distribution of the best fitting normal distribution.
// Used for Kolmogorov-Smirnov test
{
   R x_n=(x-mean_stat)/sigma_stat;
   x_n*=isr2_stat;
   return erfcc(x_n)*(-0.5)+1;
}

B neg(B const& b){ return !b;}

} // anonymous namespace

R_Vector CpmArrays::statistics(R_Vector const& dat)

{
   Z n=dat.dim();
   R_Vector res1=moments(dat);
   R norm = cpmsqrt(n/24.);
   mean_stat=res1[1];
   sigma_stat=res1[3];
   res1[6]*=norm;
   res1[5]*=(norm*2);
   R d=0,prob=0;
   R_Vector datN=dat;
   ksone(datN,cum_stat,d,prob);
      // now datN are in ascending order
   R_Vector res2(4);
   res2[1]=prob;
   res2[2]=datN[1];
      // minimum
   res2[3]=( n==2*(n/2) ? (datN[n/2]+datN[1+n/2])*0.5 : datN[(1+n)/2] );
      // median
   res2[4]=datN[n];
      // maximum

   return res1&res2;
      // appending the results (efficient return)
}

R_Vector CpmArrays::bimodalAnalysis(R_Vector const& dat)
{
   Word loc("R_Vector CpmArrays::bimodalAnalysis(R_Vector)");
   Z i,n=dat.dim();
   R3 ims=dat.infMeanSup();
   R mean0=(ims[1]+ims[3])*0.5; // between min and max
   R mean1=(ims[2]+mean0)*0.5; // taking mean also into account
   V<B> sel(n);
   for (i=1;i<=n;i++) sel[i]=(dat[i]>=mean1);
   R_Vector upper=dat.select(sel); // the valuea above mean
   sel=sel(F<B,B>(neg)); // elegant modification of sel
   R_Vector lower=dat.select(sel);
      // the values below limit
   R_Vector res(2);
   R rL=lower.mean();
   R rU=upper.mean();
   cpmassert(rL<=rU,loc);
   res[1]=rL;
   res[2]=rU;
   return res;
}

//////////////////////////// end of module statistics() //////////////////

R_Vector CpmArrays::gradient( const R_Matrix& x, R_Vector const& y)
{
   cpmmessage("R_Vector gradient() started");
   try{
   Z n=y.dim();
   Z d=x.dim2();
   cpmassert( n==x.dim1()&&n>d,"CpmArrays::gradient(R_Matrix,R_Vector)");
      // since d>=1 as for each dimension in R_Matrix, this implies n>1
   R_Vector dy(n-1);
   R_Matrix dx(n-1,1);
   for (Z i=1;i<n;i++){
      dy[i]=y[i+1]-y[1];
      dx[i]=x[i+1]-x[1];
   }
   return (!dx)*dy;
   }
   catch(...){
      Error("Problem in gradient( const R_Matrix&, R_Vector const&)");
      return R_Vector();
   }
}

R_Matrix CpmArrays::gradient( const R_Matrix& x, const R_Matrix& y)
{
   cpmmessage("R_Matrix gradient() started");
   try{
   Z n=y.dim();
   Z d=x.dim2();
   cpmassert( n==x.dim1() && n>d,"gradient(R_Matrix,R_Matrix)");
      // since d>=1 as for each dimension in R_Matrix, this implies n>1
   R_Matrix dy(n-1,1);
   R_Matrix dx(n-1,1);
   for (Z i=1;i<n;i++){
      dy[i]=y[i+1]-y[1];   // now the y[i] are vectors of dimension p, dy
      // is of matrix type (n-1,p)
      dx[i]=x[i+1]-x[1];   // dx is of matrix type (n-1,d)
   }
   return (!dx)*dy;     // !dx is of type (d,n-1), thus result of type
   // (d,p), OK
   }
   catch(...){
      Error("Problem in gradient( const R_Matrix&, const R_Matrix&)");
      return R_Matrix();
   }
}

////////////////////////////////// geometry //////////////////////////////

void CpmArrays::rotation(R_Vector& x, R_Vector const& a,
   R_Vector const& b)
{
// preliminaries: form two orthonormal vectors out of a,b. x is not used

   Z n1=x.dim(), n2=a.dim(), n3=b.dim();
   if (n1!=n2 || n2!=n3)
   cpmerror("rotation(R_Vector,R_Vector,R_Vector): dimension mismatch");
   R aa=a.abs(), bb=b.abs();
   if (aa==0. || bb==0.)
   cpmerror(
      "rotation(R_Vector,R_Vector,R_Vector): zero vector not allowed");
   aa=R(1)/aa; bb=R(1)/bb;
   R_Vector a1=a*aa;
   R_Vector b1=b*bb; // now a1 and b1 are unit vectors
   if (a1==b1) return; // nothing to do
   R c=(b1|a1);
   R s=-c*c+1;
   if (s<0.)
   cpmerror("rotation(R_Vector,R_Vector,R_Vector): unexpected error");
   s=cpmsqrt(s);
   R_Vector a2=b1-c*a1;
   aa=a2.abs();
   if (aa==0.)
   cpmerror("rotation(R_Vector,R_Vector,R_Vector): unexpected error");
   aa=R(1)/aa;
   a2*=aa; // now a1 and a2 are orthonormal vectors

// action on x

   R x1=(x|a1);
   R x2=(x|a2);
   R y1=c*x1-s*x2;
   R y2=s*x1+c*x2;
      // The vector x1*a1+x2*a2 will be rotated to y1*a1+y2*a2
      // and the orthogonal part x-x1*a1+x2*a2 of x will be left
        // unchanged
   x+=(y1-x1)*a1+(y2-x2)*a2;
}

void CpmArrays::simplex(R_Matrix& s, Z d)
{
   Z i,j,d1=d+1;
   R d1Inv=1./d1;
   R_Matrix s_(d1,d);
   s=s_;
   R_Vector e0(d1); // zero vector
   R_Matrix e(d1); // unit matrix
      // The d1 corners of the simplex in d-dimensional space are
      // formed out of the d1 canonical basis vectors in d1-dimensional
      // space. After 'removing the mean value' they all belong to a
      // hyper plane which finally is rotated into a natural
      // d-dimensional subspace
   for (i=1;i<=d1;i++) e0+=e[i];
   e0*=d1Inv; // mean value of the e[i]
   R_Vector b=e[d1];
   R_Vector a=e0;
   for (i=1;i<=d1;i++) e[i]-=e0; // now the d+1 vectors e[i] have
      // zero sum
   R r=e[1].abs();
   if (r==0.)
   cpmerror("simplex: unexpected condition r=0");
    r=R(1)/r;
   for (i=1;i<=d1;i++) rotation(e[i],a,b);
   for (i=1;i<=d1;i++){
      for (j=1;j<=d;j++){
         s[i][j]=r*e[i][j];
      }
   }
}

Z2 CpmArrays::linearToMatrix(Z k, Z m, Z n)
{
   if (m<1 || n<1)
      cpmerror("CpmArrays::linearToMatrix(...) : m<1 || n<1");
   Z mn=m*n;
   k--;
   while (k<0) k+=mn;
   while (k>=mn) k-=mn;
   Z p=k/n; // result varies from 0 to m-1
   Z q=k%n; // result varies from 0 to n-1
   return Z2(p+1,q+1);
}

Z CpmArrays::matrixToLinear(Z i, Z j, Z m, Z n)
{
   if (m<1 || n<1)
      cpmerror("CpmArrays::matrixToLinear(...) : m<1 || n<1");
   i--;
   while (i< 0) i+=m;
   while (i>=m) i-=m; // now i in {0,...,m-1}
   j--;
   while (j< 0) j+=n;
   while (j>=n) i-=n; // now j in {0,...,n-1}
   return i*n+j+1; // runs from 1 to m*n OK !
}

R_Vector CpmArrays::binomialCoefficients(Z n)
{
   Word loc("CpmArrays::binomialCoefficients(Z)");
   cpmassert(n>=-1,loc);
   Z np=n+1;
   if (np==0) return R_Vector(0);
   R_Vector res(np,1.); // initialized with 1 is OK
   if (np<=2){
      return res;
   }
   else if (np==3){
      res[2]=2;
      return res;
   }
   else if (np==4){
      res[2]=3;
      res[3]=3;
      return res;
   }
   else{
      R zaehler=n;
      R nenner=1;
      Z j=2;
      Z k=n;
      while (k>=j){
         res[j]=res[j-1]*zaehler/nenner;
         if (k>j) res[k]=res[j];
         zaehler-=1;
         nenner+=1;
         k--;
         j++;
      }
      return res;
   }
}

V< X2<Z,R> > CpmArrays::levelCrossing(R_Vector const& y, R frac,
   bool autoLevel, R levelSpezifier)
{
   Z mL=3;
   static Word loc("CpmArrays::levelCrossing(...)");
   cpmmessage(mL,loc&" started");
   Z i,n=y.dim();
   if (n<1) return V< X2<Z,R> >(0);
   R threshold;
   if (!autoLevel) threshold=levelSpezifier;
   else{
      Percentilizer per(y);
      threshold=per(levelSpezifier);
   }
   threshold*=frac;
   vector<Z> vz;
   vector<R> vr;
   R yPrev=y[1]; // we start below threshold
   for (i=1;i<=n;i++){
      R yi=y[i];
      bool b1=(yPrev<threshold);
      bool b2=(yi<threshold);
      if (b1==b2){
         ;
      } // from now on b1!=b2; thus (b1 && !b2) or (!b1 && b2)
      //  yi and yPrev lie on different sides of threshold
      // thus yi!=yPrev ==> (yi-yPrev)!=0
      else if (b1){ // ascending passage of threshold
         vz.push_back(1);
         // (threshold-yPrev):p=(yi-yPrev):1
         // p=(threshold-yPrev)/(yi-yPrev)
         R p=(threshold-yPrev)/(yi-yPrev);
         vr.push_back(R(i)+p);
      }
      else if (!b1){ // descending passage of threshold
         vz.push_back(-1);
         // (yPrev-yi):1=(yPrev-threshold):p ==>
         // p*(yPrev-yi)=(yPrev-threshold) ==>
         // p=(yPrev-threshold)/(yPrev-yi) ==>
         // p=(threshold-yPrev)/(yi-yPrev) (same formula as in
         // previous case
         R p=(threshold-yPrev)/(yi-yPrev);
         vr.push_back(R(i)+p);
      }
      yPrev=yi;
   }
   Z nres=(Z)vz.size();
   V< X2<Z,R> > res(nres);
   for (i=1;i<=nres;i++) res[i]=X2<Z,R>(vz[i-1],vr[i-1]);
   cpmmessage(mL,loc&" done");
   return res;
}

Iv CpmArrays::pairedLevelCrossing(R_Vector const& y, R frac,
   bool autoLevel, R levelSpezifier)
{
   V< X2<Z,R> > arr=levelCrossing(y,frac,autoLevel,levelSpezifier);
   Z i,n=arr.dim();
   if (n<1) return Iv();
   Z z1=arr[1].c1();
   R x1=arr[1].c2();
   R x2=0;
   bool pairingFound=false;
   for (i=1;i<=n;i++){
      Z zi=arr[i].c1();
      if (zi!=z1){
         x2=arr[i].c2();
         pairingFound=true;
      }
   }
   if (!pairingFound) return Iv();
   else return Iv(x1,x2);
}

R CpmArrays::area(R_Vector const& x, R_Vector const& y)
{
    Z nx=x.dim(), ny=y.dim();
    cpmassert(nx==ny,"CpmArrays::area: dimension mismatch");
    R sum = x[1]*y[nx] - x[nx]*y[1]; // contribution of the ends
    for (Z i=1;i<nx;++i){ // 'regular' part
        sum += x[i+1]*y[i] - x[i]*y[i+1];
    }
    return sum*0.5;
}

///////////////////////////////// function ... ///////////////////////////

   V< V<R2> > CpmArrays::homDataRan(V< V<R2> > const& x)
   {
      Word loc("CpmArrays::homDataRan(V< V<R2> >");
      Z d=x.dim();
      Z par=d%2;
      cpmassert(par==0,loc);
      V< V<R2> > y=x;
      Z i,j,dh=d/2;
      for (i=1;i<=dh;++i){
         Z di=x[i].dim();
         Z is=i+dh;
         Z dis=x[is].dim();
         R_Vector vi(di);
         for (j=1;j<=di;++j){
            vi[j]=x[i][j][2];
         }
         R_Vector vis(dis);
         for (j=1;j<=dis;++j){
            vis[j]=x[is][j][2];
         }
         R mi=vi.supAbs();
         R mis=vis.supAbs();
         mi=CpmRootX::sup<R>(mi,mis);
         if (mi>0.){
            mi=R(1)/mi;
            for (j=1;j<=di;++j){
               y[i][j][2]*=mi;
            }
            Z is=i+dh;
            for (j=x[is].b();j<=x[is].e();++j){
               y[is][j][2]*=mi;
            }
         }
      }
      return y;
   }

   V< V<R2> > CpmArrays::difOfCorData(V< V<R2> > const& xa, bool homFirst)
   {
      Word loc("CpmArrays::homDataRan(V< V<R2> >");
      V< V<R2> > x= homFirst ? homDataRan(xa): xa;
      Z d=x.dim();
      Z par=d%2;
      cpmassert(par==0,loc);
      R tol=1e-6;

      Z i,j,dh=d/2;
      V< V<R2> > res(dh);
      for (i=1;i<=dh;++i){
         Z is=i+dh;
         Z di=x[i].dim();
         Z dis=x[is].dim();
         cpmassert(di==dis,loc);
         V<R2> resi(di);
         for (j=1;j<=di;++j){
            R ti=x[i][j][1];
            R tis=x[is][j][1];
            R di=CpmRoot::dis(ti,tis);
            cpmassert(di<=tol,loc);
            R tim=(ti+tis)*0.5;
            R yi=x[is][j][2]-x[i][j][2];
            resi[j]=R2(tim,yi);
         }
         res[i]=resi;
      }
      return res;
   }

   bool CpmArrays::toGnuPlot(ostream& str, V< V<R2> > const& x )
   {
      Z mL=3;
      Word loc("CpmArrays::toGnuPlot(ostream,V< V<R2> >)");
      CPM_MA
      Z n=x.dim(),i,j;
      Z n1=x[1].dim();
      Z err=0;
      for (i=2;i<=n;++i) err+=cpmpos(n1-x[i].dim());
      cpmassert(err==0,loc);
      for (j=1;j<=n1;++j){ // each j writes a line of data to stream
         for (i=1;i<=n;++i){ // each i generates a pair (t,y) in a
            // data line
            R2 val=x[i][j];
            R t=val[1];
            R y=val[2];
            str<<t<<" "<<y<<" ";
         }
         str<<endl;
      }
      CPM_MZ
      return str ? true : false;
   }

   bool CpmArrays::toGnuPlot(Word const& fileName, V< V<R2> > const& x )
   {
      Z mL=3;
      Word loc("CpmArrays::toGnuPlot(Word,V< V<R2> >)");
      CPM_MA
      CpmSystem::OFile ofi(fileName);
      bool res=CpmArrays::toGnuPlot(ofi(),x );
      CPM_MZ
      return res;
   }


