//? cpmword.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_WORD_H_
#define CPM_WORD_H_
/*
   Description: Declares class Word which is now essentially identical
      to string of the standard library. It is formed from it according
      to the advice in BS3, end of Chapter 20.3.

      Word is a minimal string class to which Cpm classes refer for
         type identification and for messaging
*/
#include <cpminterfaces.h>
#if !defined(CPM_NAMEOF)
   #include <typeinfo>
#endif

namespace CpmRoot{

   using namespace CpmStd;

   class Word;

   Word toWord(R x, Z prc);

   Word toWord(unsigned int i, const char* ff);
      // ff standard format string
   Word toWord(unsigned long int i, const char* ff);
   Word toWord(int i, const char* ff);
   Word toWord(long int i, const char* ff);
   Word toWord(R r, const char* ff);

   ofstream& to_ofstream(Word const& w);
   ifstream& to_ifstream(Word const& w);
   string toString(Word const& w);

// concatenation of words described by operator &
   Word operator &(Word const&, Word const&);
   Word operator &(const char* cp, Word const& w);
   Word operator &(Word const& w, const char* cp);
   Word operator &(char cp[], Word const& w);
   Word operator &(Word const& w, char cp[]);

// for convenience same is allowed to be denoted +
   Word operator +(Word const&, Word const&);
   Word operator +(const char* cp, Word const& w);
   Word operator +(Word const& w, const char* cp);
   Word operator +(char cp[], Word const& w);
   Word operator +(Word const& w, char cp[]);

/////////////////////// class Word /////////////////////////////////////

   class Word{ //  wrapping charcter strings, rich functionaliy
      // Instances of Word are dynamically allocated strings of
      // characters like those of string (on which class the present
      // implementation is based). Input and Output works as expected only
      // if we have a word in the narrower sense that the character
      // string does not contain 'whitespace'.
      // Due to functions toBool(), toZ(), toR() a Word is a universal
      // carrier of numerical information. For instance the content
      // of a command line may be represented as a list of Words (an
      // instance of V<Word> in C+- )

   string rep_;

   static bool scnLin;
      // modifies the function scanFrom() in a way that also character
      // strings containing blanks can be read in to yield a single Word.
      // Such a facility is needed if Word is to be used as the data type
      // of names of more complex entities like organizations, theories,
      // etc.: Certainly 'set theory' and 'Eastman Kodak Company' should
      // be allowed names.

   static Z endOfWord; // end of Word
      // 0 nothing, 1 blank, else newline after w.prnOn()

   typedef Word Type;
   typedef Word CloneBase;

public:

   static void setScanLine(bool sl){ scnLin=sl;}
      //: set scan line
      // see private static data member scnLin
   static bool getScanLine(){ return scnLin;}
      //: get scan line
      // see private static data member scnLin

   static void setEndOfWord(Z i){ endOfWord=i;}
      //: set end of word
      // see private static data member endOfWord
   static Z getEndOfWord(){ return endOfWord;}
      //: get end of word
      // see private static data member endOfWord

   explicit Word(char c):rep_(1,c){}
   // a Word consisting of just one character c

   explicit Word(Z i, char c='a'):rep_(i,c){}
   // a Word consisting of i copies of character c
   // useful for getting new words by changing letters at given positions

   Word(const char* charPtr):rep_(charPtr){}
      // construction of a Word from char*, and thus from literals
      // such as Word w("convenient programming")
      // Although this defines a conversion from const char* to
      // Word, in a statement  cout<<"This is ..."<<endl;  the
      // interpretation of "This is ..." as const char* applies,
      // since exact matches are preferred over matches which employ
      // user-defined conversions. This is important, since
      // cout<<Word("This is ...")<<endl; may append some separator
      // to the input string depending on the value of the
      // static data member endOfWord.

   explicit Word(string const& s):rep_(s){}
      // construction of a Word from string

   explicit Word(ostringstream const& ost):rep_(ost.str()){}
      // construction of a Word from a string stream
      // This is the most convenient way to create detailed texts held as
      // a Word, as in the following code fragment:
      // ostringstream ost;
      // ost<<...<<...endl<<...;
      // Word text(ost);

   Word():rep_(){}
      // default constructor, creates the void word which consists of
      // 0 characters

   const char* operator-(void)const{ return rep_.c_str();}
      // explicit conversion from Word to const char*
      // The result is owned by the Word and the user should not try to
      // delete it.

   const char* c_str(void)const{ return rep_.c_str();}
      // to make Word and string as much alike as possible

   string str()const{ return rep_;}
      //: string

   string toStr()const{ return rep_;}
      //: to string

   string std()const{ return rep_;}
      //. standard

   Word skipChr(char c)const;
      //: skip character
      // Returns a Word which results from *this by eliminating c
      // everywhere.

   void skipSpc_();
      //: skip space, mutating function
      // eliminates all characters which by the function isspace(char)
      // of <cctype> is classified as space

   void skipSurSpc_();
      //: skip surrounding space
      // eliminates all space characters (whitespace) as above which
      // surround a chain of space separated traditional words.
      // So the result is a robust representation for a textual content
      // of a line of text, since unvisible parts at the beginning and
      // the end get eliminated
      // added 2004-10-12

   Word firstWord()const;
      // returns the first whitespace separated part of the string rep_
      // (which may contain whitespace). It would be nice to
      // return a complete list of Words here, but we prefer not to
      // build on any array class in the present basic class.
      // The result may be the void Word.

   Word firstWord_();
      // returns the first whitespace separated part of the string rep_
      // (which may contain whitespace) and changes *this into
      // the part of the word which remains.
      // The ending '_' is reminescent of Ruby's '!' as a indicator
      // for mutating functions. Calling this function in succession
      // allows to create a complete list of words in a line without
      // having a list- or array-class available at this point.

// we have to define later a general template function
// template <class T1, class T2>
// operator & (const T1&, const T2&)
// This implies that in (non-template !) functions operator & only exact
// matches will be accepted (see Stroustrup p. 589), certainly not the
// ones that can be achieved by userdefined conversions. So, we have to
// take care of char* and char[] in this place

// concatenation of words described by operator &
   Word& operator &=(Word const& w);
   Word& operator &=(const char* cp);
   friend Word operator &(Word const&, Word const&);
   friend Word operator &(const char* cp, Word const& w);
   friend Word operator &(Word const& w, const char* cp);
   friend Word operator &(char cp[], Word const& w);
   friend Word operator &(Word const& w, char cp[]);

// for convenience same is allowed to be denoted +. This is convenient
// since + binds stronger than & and especially stronger than <<
   Word& operator +=(Word const& w);
   Word& operator +=(const char* cp);
   friend Word operator +(Word const&, Word const&);
   friend Word operator +(const char* cp, Word const& w);
   friend Word operator +(Word const& w, const char* cp);
   friend Word operator +(char cp[], Word const& w);
   friend Word operator +(Word const& w, char cp[]);

// access functions
   Z dim(void)const{ return Z(rep_.length());}
      //: dimension
      // is 0 for the void Word
   Z b(void)const{ return 1;}
      // b for begin, first index of a letter
   Z e(void)const{ return dim();}
      // e for end, last index of a letter
      // A loop over the letters of Word w takes the form
      // for (Z i=w.b();i<=w.e();++i){ ... = w[i];}
      // This is correct also for the void word.
      // This looping style is also valid for all C+- array classes.
   Z size(void)const{ return dim();}
      //: size
   Z length(void)const{ return dim();}
      // for uniformity with std library types

   bool isVoid()const { return dim()==0;}
      //: is void
      //  short answer on whether the dimension is zero

   friend Z length(Word const& w){ return w.dim();}
      // number of characters (no addition for a terminator)

   char operator[](Z i)const;
      // save reading of characters. Valid indexes are 1...dim()

   char& operator[](Z i);
      // save overwriting of characters in a word.

   bool valInd(Z i)const{ return i>0 && i<=dim();}
      //: valid index
      // returns the validity of i as an index of *this

   Word cut(Z i)const;
      //: cut
      // returned is a Word which results from *this by removing i
      // components from the end. If i<0 we remove -i components from
      // the beginning.

// functions dealing Words as filenames
   Word resize(Z newDim)const;
      // returns a Word res such that res.dim()=newDim. If newDim is
      // smaller than dim(), the end of *this will be cut away.
      // If newDim is larger, blanks will be added.

   Word remExt(bool extended=false)const;
      //: remove extension
      // Intention of the function: returns what is left if any
      // file name extension (if there is one) is removed from
      // *this. So for *this=="myprog.ini" we return "myprog".
      // Actually we read and copy *this from the beginning until we
      // find a character which signifies the beginning of the extension.
      // In nomal mode this is a dot, in extended mode this is any capital
      // letter. The extended mode arrose from the need to distinguish
      // executables made by different compilers by a suffix such as
      // palaMS.exe, palaMinGW.exe, pala_ms.exe. Since the application
      // framework in cpmapplication.cpp deduces the program name,
      // and thus the name of the ini-file to search, from
      // the name of the executable, the above-mentioned executables would
      // try to read ini-files palaMS.ini, palaMinGW.ini, and
      // pala_ms.ini. They should, however, read simply pala.ini.
      // This is achieved by removing the extension from the executables
      // name for a true value of the argument.
      // The presently implemented notion of extension in 'extended mode'
      // is, that it is everything whatis not a lower-case letter or a
      // digit. This reflects .... obsolete

   Word remApp()const;
      //: remove appendix
      // Here, we remove the maximum contiguous string formed out of
      // capital letters and underscores at the end of *this. So name
      // appendices reflecting compilers or compiler options should be
      // chosen to consist of capital letters and/or underscores. So
      // not all of the functions in the previous function comment work:
      // Instead of palaMinGW, pala_ms one should use palaMINGW and
      // pala_MS

   Word appBefExt(Word const& app)const;
      //: append before extension
      // Word("myprog.ini").appBefExt("2") yields "myprog2.ini"

   Word baseName()const;
      //: base name
      // Only the characters right to the right-most "\" or "/"
      // are returned as a Word. If *this is a file name including
      // a directory path, then the actual file name gets extracted and
      // returned. (corresponds do File:basename of Ruby)

   bool readable()const;
      // returns true if *this is the name of a file that can be
      // opened and gives a valid istream to read from.
      // The stream will be closed after the test and not returned.

   Word makeFileName()const;
      //: make file name
      // returns a version of *this in which all characters which may not
      // be acceptable in file names are replaced by "_"
// end of functions dealing Words as filenames

   Word rev()const;
      //: reverse
      // returns the reversed version of *this, i.e. the one read from
      // the end

   Word& rev_(){ return *this=rev();}
      //: reverse
      // mutating form of reversal

   Word slc_(char c, bool firstSep=true);
      //: slice
      // If c does not occur in *this, we return the original *this and
      // change *this to Word("").
      // One could do it the other way round, then the implementation
      // of remExt in terms of the present function would need to handle
      // the case that no dot is present (as for executables in Linux)
      // would need special treatment.
      // Now the interesting case:
      // If c appears in *this, we fix the position where it
      // appears in the leftmost position for firstSep==true
      // and in the rightmost position else.
      // we return the part of the word which is left of c
      // and change *this to the part right of c. Then we have for
      // Word w=...;
      // Word wMem=w;
      // char c=...; // such that c occurs in w !
      // Word wl=w.slc_(c);
      // the property
      // wl&Word(c)&w==wMem;

   Word tail(Z i)const;
      //: tail
      // if i>=dim() we return *this. Else we remove as many components
      // from the beginning that only i remain (' the last i components of
      // the original Word') and return the result.

   Z find(char c)const;
      //: find
      // returns the first position (1.....dim()) of a c in the
      // Word *this. If the character c isn't there we get 0

    Z find(Word const& w)const;
      //: find
      // returns the first position (1.....dim()) of a
      // subword w in the Word *this. If the Word w isn't there
      // we get 0

   static Word write(R n, Z fieldLength=0);
   static Word write(Z n, Z fieldLength=0);
   static Word write(N n, Z fieldLength=0);
        // If fieldLength=0 the natural length of the number is fully
      // outputted. Else the length is casted or extended to fieldLength.
   static Word write(string h);
   static Word write(bool b);

   static Word ascii();
      // All ASCII characters implemented in in the fonts
      // used by CpmGraphics::Frame. These characters are concatenated
      // in a Word. Helpful for font inspection and development.

   R toR()const;
      // outputs a R obtained by converting *this
   Z toZ()const;
      // outputs a Z obtained by converting *this
   N toN()const;
      // outputs a N obtained by converting *this
   bool toBool()const;
      // outputs a bool obtained by converting *this

// these declarations are repeated outside the class
// and have only influence on the implementation

   friend Word toWord(unsigned int i, const char* ff);
   friend Word toWord(unsigned long int i, const char* ff);
   friend Word toWord(int i, const char* ff);
   friend Word toWord(long int i, const char* ff);
   friend Word toWord(R r, const char* ff);

   friend ofstream& to_ofstream(Word const& w);
   friend ifstream& to_ifstream(Word const& w);
   friend string toString(Word const& w);

   bool readLine(istream& in){return CpmRoot::readLine(rep_,in);}
      // making *this equal to the content of the next non-comment
      // line in stream in

   CPM_IO
      // scanFrom(in);  makes *this equal to the next noncomment
      // whitespace separated sequence of characters in stream in.
      // Thus the code
      //
      //    Word w("This is a word");
      //    w.prnOn(aStream);
      //    Word w2;
      //    w2.scanFrom(aStream)
      //
      // will yield w2=="This"
      // and not w2=="This is a word"! This is sometimes what one wants,
      // sometimes not.
   CPM_ORDER
   CPM_TEST_ALL // implementation borrowed from string
// Declarations of CPM_DESCRIPTORS together with implementation
   Word toWord()const{ return *this;}
   Word nameOf()const{ return "Word";}
};

inline string toString(Word const& w){ return w.rep_;}

#define cpmwrite CpmRoot::Word::write
#define cpm CpmRoot::Word::write

// conversion of selected basic types to string
// as a basis for doing the conversion to Word in a consistent
// manner.
// The parameter prc (precision) makes no sense for integer types
// Most conversions from built-in types to Word, which have not
// a char*-typed formatting string as one argument, are based on these
// toStr-functions. An exception are the functions write which have
// precisely defined writing field length which is needed for writing
// e.g. to tables or quietly on the status bar.
// prc<=0 means that no precision gets set in the code, so that we then
// rely on the default precision set by the compiler.
   string toStr(Z x);
   string toStr(N x);
   string toStr(L x);
   string toStr(bool x);
   string toStr(R const& x, Z prc=0);
/*
In cpmnumbers.h the following list of interface service class
templates was defined
    IO<T>, Comp<T>, Inv<T>, AbsSqr<T>, Abs<T>, Dis<T>,
    Test<T>, Ran<T>, Hash<T>, Conj<T>, Neutrals<T>
and the following two completing members were annonced
    Name<T>, ToWord<T>.
These will be defined now:
*/
///////////////// class Name<> ///////////////////////////////////////////
// We need a template class istead of a template function since we need
// specializations also for templates such as std::vector<> (see cpmv.h).
// Whether one has CPM_NAMEOF defined or not --- the C+- basic
// built-in types always have the C+- name (i.e 'R', not 'double')
// For the more complex types, the names created by typeinfo
// (which look acceptable) are used for CPM_NAMEOF not defined.

   template<class T>
   class Name{ // tool class template for defining the nameOf function
   public:
      Name(){}
      Word operator()(T const& t)const
      {
#if defined(CPM_NAMEOF)
         return t.nameOf();
#else
         return Word(typeid(t).name());
#endif
      }
   };

   template<>
   class Name<Z>{ // Z: integers
   public:
      Name(){}
      Word operator()(Z const& t)const
         { t; return Word("Z");}
   };

   template<>
   class Name<N>{ // N: natural numbers
   public:
      Name(){}
      Word operator()(N const& t)const
         { t; return Word("N");}
   };

   template<>
   class Name<R>{ // multiple precision real numbers
   public:
      Name(){}
      Word operator()(R const& t)const
         { t; return Word("R");}
   };

   template<>
   class Name<L>{ // L: letter
   public:
      Name(){}
      Word operator()(L const& t)const
         { t; return Word("L");}
   };

   template<>
   class Name<bool>{ // specialization
   public:
      Name(){}
      Word operator()(bool const& t)const
         { t; return Word("bool");}
   };

   template<>
   class Name<char>{ // specialization
   public:
      Name(){}
      Word operator()(char const& t)const
         { t; return Word("char");}
   };

   template<>
   class Name<string>{ // specialization
   public:
      Name(){}
      Word operator()(string const& t)const
         { t; return Word("string");}
   };

   template<class T>
   Word nameOf(T const& t){ return Name<T>()(t); }
      //: name of
   // convenient syntax for getting a name of the type of
   // a quantity. Definition is such that one may also give
   // Cpm-names to class templates. See cpmv.h
   // for treating template <class T> std::vector<T> in this way.

/////////////////////// class ToWord<> ///////////////////////////////////

   template<class T>
   class ToWord{
   public:
      ToWord(){}
      Word operator()(T const& t)const{ return t.toWord();}
   };

   template<>
   class ToWord<R>{
   public:
      ToWord(){}
      Word operator()(R const& t)const{ return Word(toStr(t));}
   };

   template<>
   class ToWord<Z>{
   public:
      ToWord(){}
      Word operator()(Z const& t)const{ return Word(toStr(t));}
   };

   template<>
   class ToWord<N>{
   public:
      ToWord(){}
      Word operator()(N const& t)const{ return Word(toStr(t));}
   };

   template<>
   class ToWord<L>{
   public:
      ToWord(){}
      Word operator()(L const& t)const{ return Word(toStr(t));}
   };

   template<>
   class ToWord<bool>{
   public:
      ToWord(){}
      Word operator()(bool const& t)const{ return Word(toStr(t));}
   };

   template<>
   class ToWord<string>{
   public:
      ToWord(){}
      Word operator()(string const& t)const{ return Word(t);}
   };

   template<class T>
   Word toWord(T const& t){ return ToWord<T>()(t); }
      //: to word
      // Presently not important, MPI-functionalty relies on
      // prnOn and scanFrom

   //////////////////////////// class Root<> /////////////////////////////

   // A final step towards unification of infrastructure functions:
   // The Root class template provides member functions which the
   // argument type T may not define. Even if type T is only accessible
   // through its declaration, we may specialize the 13 basic
   // Root - related interface service class templates
   //
   //    IO<T>, Comp<T>, Inv<T>, AbsSqr<T>, Abs<T>, Dis<T>, Test<T>,
   //    Ran<T>, Hash<T>, Conj<T>, Neutrals<T>, ToWord<T>, Name<T>
   //
   // for this type T and thus make Root<T> equipped with these member
   // functions.
   // How this class simplifies the implementation of template classes
   // can be seen from the implementation of the interfaces
   // CPM_IO and CPM_ORDER in cpmv.h
   // Notice that in addition to the template argument type T, there are
   // the following types involved in the definition of Root<>:
   //
   // bool
   // std::istream
   // std::ostream
   // CpmRoot::Z
   // CpmRoot::R
   // CpmRoot::Word
   //
   // Notice also, that the true nature of CpmRoot::Z and CpmRoot::R
   // depends on the macros CPM_LONG and CPM_MP in cpmdefinitions.h.
   // This is an overview of all 26 member functions of Root<T>:
/*
      Root(T const& t = T());
      T operator()(void)const;
      bool prnOn(ostream& str)const;
      bool scanFrom(istream& str);
      Z com(T const& t)const;
      bool operator == ( T const& t)const;
      bool operator != (T const& t)const;
      bool operator < (T const& t)const;
      bool operator > (T const& t)const;
      bool operator <= (T const& t)const;
      bool operator >= (T const& t)const;
      T inf(T const& t)const;
      T sup(T const& t)const;
      Z hash()const;
      R absSqr()const;
      R abs()const;
      R dis(T const& t)const;
      T inv()const;
      T operator!()const;
      T net(Z i=0)const;
      T con()const;
      T operator~()const;
      T test(Z tvs)const;
      T ran(Z j=0)const;
      Word nameOf()const
      Word toWord()const
*/

   template<class T>
   class Root{ // Provides the basic functions for dealing with CpmRoot's
     // basic types as member functions (or operators) of a class
     // template.
      T a_;
   public:
      explicit Root(T const& t = T()):a_(t){}
      //Root(T& t , Word const& ):a_(t){}
         // should allow to change t e.g. in
         // Root<T>(t,"").scanFrom(in);
         // does not work however!
      T operator()(void)const{ return a_;}
      bool prnOn(ostream& str)const{ return IO<T>().o(a_,str);}
         //: print on
         // we don't redefine ostream& << T
         // Root<T>(t).prnOn(out);
         // is the general idiom for writing a T-typed object to stream.
         // Notice that the I/O-macros cpmp(X) and cpms(X) are implemented
         // based on functions prnOnT and scanFromT and that class Root
         // can hardly be used for this purpose (macros are always
         // special).
      bool scanFrom(istream& str){ return IO<T>().i(a_,str);}
         //: scan from
         // we don't redefine istream& >> T
         // The general idiom for reading a T-typed object t from
         // stream is:
         // Root<T> tIn;
         // bool suc=tIn.scanFrom(str);
         // t=tIn();
         // A more compact way (avoiding Root) is
         // bool suc=scanT(t,str);
         // or
         // bool CpmRoot::scanFrom(t,str);

      Z com(T const& t)const{ return Comp<T>()(a_,t);}
         //: compare
      bool operator == ( T const& t)const{ return 0==com(t);}
      bool operator != (T const& t)const{ return 0!=com(t);}
      bool operator < (T const& t)const{ return 0 < com(t);}
      bool operator > (T const& t)const{ return 0 > com(t);}
      bool operator <= (T const& t)const{ return 0 <= com(t);}
      bool operator >= (T const& t)const{ return 0 >= com(t);}
      T inf(T const& t)const{ return 0 < com(t) ? a_ : t;}
         //: infimum
      T sup(T const& t)const{ return 0 > com(t) ? a_ : t;}
         //: supremum
      Z hash()const{ return Hash<T>()(a_);}
         //: hash
      R absSqr()const{ return AbsSqr<T>()(a_);}
         //: absolute (value) squared
      R abs()const{ return Abs<T>()(a_);}
         //: absolute (value)
      R dis(T const& t)const{ return Dis<T>()(a_,t);}
         //: distance
      T inv()const{return Inv<T>()(a_);}
         //: inverse
      T operator!()const{return Inv<T>()(a_);}
      T net(Z i=0)const{ return Neutrals<T>()(a_,i);}
         //: neutrals
      T con()const{ return Conj<T>()(a_);}
         //: conjugate
      T operator~()const{ return Conj<T>()(a_);}
      T test(Z tvs)const{ return Test<T>()(a_,tvs);}
         //: test, //. tvs: test value size ~ complexity
      T ran(Z j=0)const{ return Ran<T>()(a_,j);}
         //: random
      Word nameOf()const{ return Name<T>()(a_);}
         //: name of
      Word toWord()const{ return ToWord<T>()(a_);}
         //: to word
   };

   template<typename T> // not clear. Seems not to work with T = R
   bool scanFrom(T& t, istream& str)
   {
      //Root<T> rt(t);
      Root<T> tIn;
      bool b=tIn.scanFrom(str); // reading the 'value' of tIn from str
      if (b) t = tIn();
      return b;
   }

} // namespace

#define cpmnam CpmRoot::nameOf
#define cpmtow CpmRoot::toWord

#endif
