//? cpmx.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_X_H_
#define CPM_X_H_
/*

Purpose: Defining lean cartesian products of 2-8 factors
   to be used similar to pair<> of the standard library.
   Needed for enabling functions to return longer value
   lists than just pairs.
   Non-constant reference arguments as a work-around are considered
   bad style here.
   If one needs cartesian products very often as function arguments or as
   function return values, this might be an indication that it would
   natural to define one or more classes which aggregate these components
   into a meaningful object.

   All these cartesian producs define IO and order operations in the
   same manner as Vl and V do this. This is a considerable simplification
   compared to my old approach to have templates Xno, Xnio for those
   template arguments which implement order operations and IO operations.

   Notice that access to components can be done via references and
   values. Notice also that, other than for pair, first,second,...
   are functions and not data. So the following is correct:
   X2<R,Z> a(4,6);
   Z i=a.second(); // not i=a.second;
      // i.e. i=6
   However,
   Z i=a.c2();
   looks nicer to me

   Addition February 2009: For pairs, triplets, and quartets there are now
   also homotypic versions. Component access is given here through public
   data members x1, x2, x3, x4.
   Also here, IO and order relations are implemented.
*/
#include <cpmword.h>
#include <vector>
#include <tuple>

namespace CpmArrays{

   using CpmRoot::Word;
   using CpmRoot::Z;
   using namespace CpmStd;
   using std::tuple; // 2015-06-11 conversion to std::tupel added.
   using std::get;

// Classes introduced by this file with fixed names:

//       X2, X3, X4, X5, X6, X7, X8
//       T2, T3, T4

// All template arguments are assumed to have (not necessarily to define
// explicitely) default constructor, copy constructor, and operator =

// Lists of 2,3,4,5,6,7 or 8 components (hetero-typic lists)
//   and
// homo-typic lists with 2,3, or 4 components.
// Components of Xi accessible by member functions c1(),c2(),...
// Components of Ti accessible by member functions c1(),c2(),... and by
// conventional indexing (first valid index is 1)
// (indexing added 2011-03-22)
///////////////////////////// class X2<> ////////////////////////////

template <class Y1, class Y2>
class X2{ // heterotypic pairs
   // Lean Cartesian product of Y1 with Y2
protected:
   Y1 x1;
   Y2 x2;
   typedef X2<Y1,Y2> Type;

public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 2;}
 // constructors
   X2(Y1 const& x1_, Y2 const& x2_):x1(x1_),x2(x2_){}
   X2(void):x1(),x2(){}
   X2(X2<Y1,Y2> const& z): x1(z.x1), x2(z.x2){}
   X2(tuple<Y1,Y2> const& t): x1(get<0>(t)),x2(get<1>(t)){}

// 'modern' access functions. Similar to operator[]
// getting the second component of object a is done like
//
// ...=a.c2();
//
// setting is done like
//
// a.c2()=...;
//
   Y1& c1(){ return x1;}
   Y2& c2(){ return x2;}
   Y1 const& c1()const{ return x1;}
   Y2 const& c2()const{ return x2;}
   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}

   Word nameOf()const
   {
       return Word("X2< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          " >");
   }

   tuple<Y1,Y2> toTup()const{ return tuple<Y1,Y2>{x1,x2};}

};

template <class Y1, class Y2>
Z X2<Y1,Y2>::com(X2<Y1,Y2> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   return 0;
}

template <class Y1, class Y2>
bool X2<Y1,Y2>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   return true;
}

template <class Y1, class Y2>
bool X2<Y1,Y2>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   return true;
}

///////////////////////////// class T2<> ////////////////////////////

template <class X>
class T2{ // homotypic pairs
   // Lean Cartesian product of X with itself.
public:
   X x1, x2;
   typedef T2<X> Type;
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 2;}
 // constructors
   T2(X const& x1_, X const& x2_):x1(x1_),x2(x2_){}
   T2(void):x1(),x2(){}
   T2(T2<X> const& z): x1(z.x1), x2(z.x2){}
   X& operator[](Z i){ if (i>=2) return x2; else return x1;}
   X const& operator[](Z i)const{ if (i>=2) return x2; else return x1;}
   X& fir(){ return x1;}
      //: first
   X& last(){ return x2;}
      //: last
   Z b()const { return 1;}
      //. begin
      // Returns the first valid index.
   Z e()const { return 2;}
      //. end
      // Returns the last valid index.
   X& c1(void){ return x1;}
   X& c2(void){ return x2;}
   X const& c1(void)const{ return x1;}
   X const& c2(void)const{ return x2;}
   Word nameOf()const
   {
      return Word("T2< "&CpmRoot::Name<X>()(x1)&" >");
   }
};

template <class Y>
Z T2<Y>::com(T2<Y> const& s)const
{
   if (x1<s.x1) return 1;
   if (x1>s.x1) return -1;
   if (x2<s.x2) return 1;
   if (x2>s.x2) return -1;
   return 0;
}

template <class Y>
bool T2<Y>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   return true;
}

template <class Y>
bool T2<Y>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   return true;
}

///////////////////////////// class X3<> ////////////////////////////

// for more factors:

template <class Y1, class Y2, class Y3>
class  X3{ // heterotypic triplets
protected:
   Y1 x1;
   Y2 x2;
   Y3 x3;
   typedef X3<Y1,Y2,Y3> Type;
public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 3;}
   X3(Y1 const& x1_, Y2 const& x2_, Y3 const& x3_):x1(x1_),x2(x2_),
      x3(x3_){}
   X3(X2<Y1,Y2> const& y_, Y3 const& x3_):x1(y_.c1()),x2(y_.c2()),
      x3(x3_){}
   X3(void):x1(),x2(),x3(){}
   X3(X3<Y1,Y2,Y3> const& z): x1(z.x1), x2(z.x2), x3(z.x3){}
   X3(tuple<Y1,Y2,Y3> const& t): x1(get<0>(t)),x2(get<1>(t)),
      x3(get<2>(t)){}

   Y1& c1(void){ return x1;}
   Y2& c2(void){ return x2;}
   Y3& c3(void){ return x3;}

   Y1 const& c1(void)const{ return x1;}
   Y2 const& c2(void)const{ return x2;}
   Y3 const& c3(void)const{ return x3;}

   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}
   Y3 third()const{ return x3;}

   Word nameOf()const
   {
       return Word("X3< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          ","&
          CpmRoot::Name<Y3>()(Y3())&
          " >");
   }

   tuple<Y1,Y2,Y3> toTup()const{ return tuple<Y1,Y2,Y3>{x1,x2,x3};}

};

template <class Y1, class Y2, class Y3>
Z X3<Y1,Y2,Y3>::com(X3<Y1,Y2,Y3> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   if (c3()<s.c3()) return 1;
   if (c3()>s.c3()) return -1;
   return 0;
}

template <class Y1, class Y2, class Y3>
bool X3<Y1,Y2,Y3>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   return true;
}

template <class Y1, class Y2, class Y3>
bool X3<Y1,Y2,Y3>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   return true;
}

///////////////////////////// class T3<> ////////////////////////////

template <class X>
class T3{ // homotypic triplets
   // Lean Cartesian product of X with itself.
public:
   X x1, x2, x3;
   typedef T3<X> Type;
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 3;}
 // constructors
   T3(X const& x1_, X const& x2_, X const& x3_):x1(x1_),x2(x2_),x3(x3_){}
   T3(void):x1(),x2(),x3(){}
   T3(T3<X> const& z): x1(z.x1), x2(z.x2),x3(z.x3){}

   X& operator[](Z i){ if (i>=3) return x3; else if (i==2) return x2;
    else return x1;}
   X const& operator[](Z i)const{ if (i>=3) return x3; else if (i==2)
    return x2; else return x1;}
   X& fir(){ return x1;}
      //: first
   X& last(){ return x3;}
      //: last
   Z b()const { return 1;}
      //. begin
      // Returns the first valid index.
   Z e()const { return 3;}
      //. end
      // Returns the last valid index.

   X& c1(void){ return x1;}
   X& c2(void){ return x2;}
   X& c3(void){ return x3;}

   X const& c1(void)const{ return x1;}
   X const& c2(void)const{ return x2;}
   X const& c3(void)const{ return x3;}

   Word nameOf()const
   {
      return Word("T3< "&CpmRoot::Name<X>()(x1)&" >");
   }
};

template <class Y>
Z T3<Y>::com(T3<Y> const& s)const
{
   if (x1<s.x1) return 1;
   if (x1>s.x1) return -1;
   if (x2<s.x2) return 1;
   if (x2>s.x2) return -1;
   if (x3<s.x3) return 1;
   if (x3>s.x3) return -1;
   return 0;
}

template <class Y>
bool T3<Y>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   return true;
}

template <class Y>
bool T3<Y>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   return true;
}

///////////////////////////// class X4<> ////////////////////////////

template <class Y1, class Y2, class Y3, class Y4>
class  X4{ // heterotypic 4-tuples
protected:
   Y1 x1;
   Y2 x2;
   Y3 x3;
   Y4 x4;
   typedef X4<Y1,Y2,Y3,Y4> Type;

public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 4;}
   X4(Y1 const& x1_, Y2 const& x2_, Y3 const& x3_, Y4 const& x4_):
      x1(x1_),x2(x2_),x3(x3_),x4(x4_){}
   X4(X3<Y1,Y2,Y3> const& y_, Y4 const& x4_):
      x1(y_.c1()),x2(y_.c2()),x3(y_.c3()),x4(x4_){}
   X4(void):x1(),x2(),x3(),x4(){}
   X4(X4<Y1,Y2,Y3,Y4> const& z): x1(z.x1), x2(z.x2), x3(z.x3),x4(z.x4){}
   X4(tuple<Y1,Y2,Y3,Y4> const& t): x1(get<0>(t)),x2(get<1>(t)),
      x3(get<2>(t)),x4(get<3>(t)){}

   Y1& c1(void){ return x1;}
   Y2& c2(void){ return x2;}
   Y3& c3(void){ return x3;}
   Y4& c4(void){ return x4;}

   Y1 const& c1(void)const{ return x1;}
   Y2 const& c2(void)const{ return x2;}
   Y3 const& c3(void)const{ return x3;}
   Y4 const& c4(void)const{ return x4;}

   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}
   Y3 third()const{ return x3;}
   Y4 fourth()const{ return x4;}

   Word nameOf()const
   {
       return Word("X4< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          ","&
          CpmRoot::Name<Y3>()(Y3())&
          ","&
          CpmRoot::Name<Y4>()(Y4())&
          " >");
   }

   tuple<Y1,Y2,Y3,Y4> toTup()const
   {
      return tuple<Y1,Y2,Y3,Y4>{x1,x2,x3,x4};
   }

};

template <class Y1, class Y2, class Y3, class Y4>
Z X4<Y1,Y2,Y3,Y4>::com(X4<Y1,Y2,Y3,Y4> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   if (c3()<s.c3()) return 1;
   if (c3()>s.c3()) return -1;
   if (c4()<s.c4()) return 1;
   if (c4()>s.c4()) return -1;
   return 0;
}

template <class Y1, class Y2, class Y3, class Y4>
bool X4<Y1,Y2,Y3,Y4>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   cpmp(x4);
   return true;
}

template <class Y1, class Y2, class Y3, class Y4>
bool X4<Y1,Y2,Y3,Y4>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   cpms(x4);
   return true;
}

///////////////////////////// class T4<> /////////////////////////////////

template <class X>
class T4{ // homotypic quartets
   // Lean Cartesian product of X with itself.
public:
   X x1, x2, x3, x4;
   typedef T4<X> Type;
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 4;}
 // constructors
   T4(X const& x1_, X const& x2_, X const& x3_, X const& x4_):x1(x1_),
      x2(x2_),x3(x3_),x4(x4_){}
   T4(void):x1(),x2(),x3(),x4(){}
   T4(T4<X> const& z): x1(z.x1), x2(z.x2),x3(z.x3),x4(z.x4){}

   X& operator[](Z i)
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}

   X const& operator[](Z i)const
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}

   X& fir(){ return x1;}
      //: first
   X& last(){ return x4;}
      //: last
   Z b()const { return 1;}
      //. begin
      // Returns the first valid index.
   Z e()const { return 4;}
      //. end
      // Returns the last valid index.

   X& c1(void){ return x1;}
   X& c2(void){ return x2;}
   X& c3(void){ return x3;}
   X& c4(void){ return x4;}

   X const& c1(void)const{ return x1;}
   X const& c2(void)const{ return x2;}
   X const& c3(void)const{ return x3;}
   X const& c4(void)const{ return x4;}

   Word nameOf()const
   {
      return Word("T4< "&CpmRoot::Name<X>()(x1)&" >");
   }
};

template <class Y>
Z T4<Y>::com(T4<Y> const& s)const
{
   if (x1<s.x1) return 1;
   if (x1>s.x1) return -1;
   if (x2<s.x2) return 1;
   if (x2>s.x2) return -1;
   if (x3<s.x3) return 1;
   if (x3>s.x3) return -1;
   if (x4<s.x4) return 1;
   if (x4>s.x4) return -1;
   return 0;
}

template <class Y>
bool T4<Y>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   cpmp(x4);
   return true;
}

template <class Y>
bool T4<Y>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   cpms(x4);
   return true;
}

///////////////////////////// class X5<> ////////////////////////////

template <class Y1, class Y2, class Y3, class Y4, class Y5>
class  X5{ // heterotypic 5-tuples
protected:
   Y1 x1;
   Y2 x2;
   Y3 x3;
   Y4 x4;
   Y5 x5;
   typedef X5<Y1,Y2,Y3,Y4,Y5> Type;
public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 5;}
   X5(Y1 const& x1_, Y2 const& x2_, Y3 const& x3_,
      Y4 const& x4_, Y5 const& x5_):x1(x1_),x2(x2_),x3(x3_),
      x4(x4_),x5(x5_){}
   X5(X4<Y1,Y2,Y3,Y4> const& y_, Y5 const& x5_):
      x1(y_.c1()),x2(y_.c2()),x3(y_.c3()),
      x4(y_.c4()),x5(x5_){}
   X5(void):x1(),x2(),x3(),x4(),x5(){}
   X5(X5<Y1,Y2,Y3,Y4,Y5> const& z): x1(z.x1), x2(z.x2),
      x3(z.x3),x4(z.x4),x5(z.x5){}

   X5(tuple<Y1,Y2,Y3,Y4,Y5> const& t): x1(get<0>(t)),x2(get<1>(t)),
      x3(get<2>(t)),x4(get<3>(t)),x5(get<4>(t)){}

   Y1& c1(void){ return x1;}
   Y2& c2(void){ return x2;}
   Y3& c3(void){ return x3;}
   Y4& c4(void){ return x4;}
   Y5& c5(void){ return x5;}

   Y1 const& c1(void)const{ return x1;}
   Y2 const& c2(void)const{ return x2;}
   Y3 const& c3(void)const{ return x3;}
   Y4 const& c4(void)const{ return x4;}
   Y5 const& c5(void)const{ return x5;}

   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}
   Y3 third()const{ return x3;}
   Y4 fourth()const{ return x4;}
   Y5 fifth()const{ return x5;}

   Word nameOf()const
   {
       return Word("X5< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          ","&
          CpmRoot::Name<Y3>()(Y3())&
          ","&
          CpmRoot::Name<Y4>()(Y4())&
          ","&
          CpmRoot::Name<Y5>()(Y5())&
          " >");
   }

   tuple<Y1,Y2,Y3,Y4,Y5> toTup()const
   {
      return tuple<Y1,Y2,Y3,Y4,Y5>{x1,x2,x3,x4,x5};
   }
};

template <class Y1, class Y2, class Y3, class Y4, class Y5>
Z X5<Y1,Y2,Y3,Y4,Y5>::com(X5<Y1,Y2,Y3,Y4,Y5> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   if (c3()<s.c3()) return 1;
   if (c3()>s.c3()) return -1;
   if (c4()<s.c4()) return 1;
   if (c4()>s.c4()) return -1;
   if (c5()<s.c5()) return 1;
   if (c5()>s.c5()) return -1;
   return 0;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5>
bool X5<Y1,Y2,Y3,Y4,Y5>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   cpmp(x4);
   cpmp(x5);
   return true;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5>
bool X5<Y1,Y2,Y3,Y4,Y5>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   cpms(x4);
   cpms(x5);
   return true;
}

///////////////////////////// class X6<> ////////////////////////////

template <class Y1, class Y2, class Y3, class Y4, class Y5, class Y6>
class  X6{ // heterotypic 6-tuples
protected:
   Y1 x1;
   Y2 x2;
   Y3 x3;
   Y4 x4;
   Y5 x5;
   Y6 x6;
   typedef X6<Y1,Y2,Y3,Y4,Y5,Y6> Type;
public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 6;}

   X6(Y1 const& x1_, Y2 const& x2_, Y3 const& x3_,
      Y4 const& x4_, Y5 const& x5_, Y6 const& x6_):x1(x1_),x2(x2_),
      x3(x3_), x4(x4_),x5(x5_),x6(x6_){}

   X6(X5<Y1,Y2,Y3,Y4,Y5> const& y_, Y6 const& x6_):
      x1(y_.c1()),x2(y_.c2()),x3(y_.c3()),
      x4(y_.c4()),x5(y_.c5()),x6(x6_){}

   X6(void):x1(),x2(),x3(),x4(),x5(),x6(){}

   X6(X6<Y1,Y2,Y3,Y4,Y5,Y6> const& z): x1(z.x1), x2(z.x2),
      x3(z.x3),x4(z.x4),x5(z.x5),x6(z.x6){}

   X6(tuple<Y1,Y2,Y3,Y4,Y5,Y6> const& t): x1(get<0>(t)),x2(get<1>(t)),
      x3(get<2>(t)),x4(get<3>(t)),x5(get<4>(t)),x6(get<5>(t)){}

   Y1& c1(void){ return x1;}
   Y2& c2(void){ return x2;}
   Y3& c3(void){ return x3;}
   Y4& c4(void){ return x4;}
   Y5& c5(void){ return x5;}
   Y6& c6(void){ return x6;}

   Y1 const& c1(void)const{ return x1;}
   Y2 const& c2(void)const{ return x2;}
   Y3 const& c3(void)const{ return x3;}
   Y4 const& c4(void)const{ return x4;}
   Y5 const& c5(void)const{ return x5;}
   Y6 const& c6(void)const{ return x6;}

   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}
   Y3 third()const{ return x3;}
   Y4 fourth()const{ return x4;}
   Y5 fifth()const{ return x5;}
   Y6 sixth()const{ return x6;}

   Word nameOf()const
   {
       return Word("X6< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          ","&
          CpmRoot::Name<Y3>()(Y3())&
          ","&
          CpmRoot::Name<Y4>()(Y4())&
          ","&
          CpmRoot::Name<Y5>()(Y5())&
          ","&
          CpmRoot::Name<Y6>()(Y6())&
          " >");
   }

   tuple<Y1,Y2,Y3,Y4,Y5,Y6> toTup()const
   {
      return tuple<Y1,Y2,Y3,Y4,Y5,Y6>{x1,x2,x3,x4,x5,x6};
   }
};

template <class Y1, class Y2, class Y3, class Y4, class Y5, class Y6>
Z X6<Y1,Y2,Y3,Y4,Y5,Y6>::com(X6<Y1,Y2,Y3,Y4,Y5,Y6> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   if (c3()<s.c3()) return 1;
   if (c3()>s.c3()) return -1;
   if (c4()<s.c4()) return 1;
   if (c4()>s.c4()) return -1;
   if (c5()<s.c5()) return 1;
   if (c5()>s.c5()) return -1;
   if (c6()<s.c6()) return 1;
   if (c6()>s.c6()) return -1;
   return 0;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5, class Y6>
bool X6<Y1,Y2,Y3,Y4,Y5,Y6>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   cpmp(x4);
   cpmp(x5);
   cpmp(x6);
   return true;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5, class Y6>
bool X6<Y1,Y2,Y3,Y4,Y5,Y6>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   cpms(x4);
   cpms(x5);
   cpms(x6);
   return true;
}

///////////////////////////// class X7<> ////////////////////////////

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7>
class  X7{ // heterotypic 7-tuples
protected:
   Y1 x1;
   Y2 x2;
   Y3 x3;
   Y4 x4;
   Y5 x5;
   Y6 x6;
   Y7 x7;
   typedef X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7> Type;
public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 7;}
   X7(Y1 const& x1_, Y2 const& x2_, Y3 const& x3_,
      Y4 const& x4_, Y5 const& x5_, Y6 const& x6_, Y7 const& x7_):
      x1(x1_),x2(x2_),x3(x3_),x4(x4_),x5(x5_),x6(x6_),x7(x7_){}

   X7(X6<Y1,Y2,Y3,Y4,Y5,Y6> const& y_, Y7 const& x7_):
      x1(y_.c1()),x2(y_.c2()),x3(y_.c3()),
      x4(y_.c4()),x5(y_.c5()),x6(y_.c6()),x7(x7_){}

   X7(void):x1(),x2(),x3(),x4(),x5(),x6(),x7(){}
   X7(X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7> const& z): x1(z.x1), x2(z.x2),
      x3(z.x3),x4(z.x4),x5(z.x5),x6(z.x6),x7(z.x7){}

   X7(tuple<Y1,Y2,Y3,Y4,Y5,Y6,Y7> const& t): x1(get<0>(t)),x2(get<1>(t)),
      x3(get<2>(t)),x4(get<3>(t)),x5(get<4>(t)),x6(get<5>(t)),
      x7(get<6>(t)){}


   Y1& c1(void){ return x1;}
   Y2& c2(void){ return x2;}
   Y3& c3(void){ return x3;}
   Y4& c4(void){ return x4;}
   Y5& c5(void){ return x5;}
   Y6& c6(void){ return x6;}
   Y7& c7(void){ return x7;}

   Y1 const& c1(void)const{ return x1;}
   Y2 const& c2(void)const{ return x2;}
   Y3 const& c3(void)const{ return x3;}
   Y4 const& c4(void)const{ return x4;}
   Y5 const& c5(void)const{ return x5;}
   Y6 const& c6(void)const{ return x6;}
   Y7 const& c7(void)const{ return x7;}

   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}
   Y3 third()const{ return x3;}
   Y4 fourth()const{ return x4;}
   Y5 fifth()const{ return x5;}
   Y6 sixth()const{ return x6;}
   Y7 seventh()const{ return x7;}

   Word nameOf()const
   {
       return Word("X7< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          ","&
          CpmRoot::Name<Y3>()(Y3())&
          ","&
          CpmRoot::Name<Y4>()(Y4())&
          ","&
          CpmRoot::Name<Y5>()(Y5())&
          ","&
          CpmRoot::Name<Y6>()(Y6())&
          ","&
          CpmRoot::Name<Y7>()(Y7())&
          " >");
   }

   tuple<Y1,Y2,Y3,Y4,Y5,Y6,Y7> toTup()const
   {
      return tuple<Y1,Y2,Y3,Y4,Y5,Y6,Y7>{x1,x2,x3,x4,x5,x6,x7};
   }

};

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7>
Z X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7>::com(X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   if (c3()<s.c3()) return 1;
   if (c3()>s.c3()) return -1;
   if (c4()<s.c4()) return 1;
   if (c4()>s.c4()) return -1;
   if (c5()<s.c5()) return 1;
   if (c5()>s.c5()) return -1;
   if (c6()<s.c6()) return 1;
   if (c6()>s.c6()) return -1;
   if (c7()<s.c7()) return 1;
   if (c7()>s.c7()) return -1;
   return 0;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7>
bool X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   cpmp(x4);
   cpmp(x5);
   cpmp(x6);
   cpmp(x7);
   return true;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7>
bool X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   cpms(x4);
   cpms(x5);
   cpms(x6);
   cpms(x7);
   return true;
}

///////////////////////////// class X8<> ////////////////////////////

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7, class Y8>
class  X8{ // heterotypic 8-tuples
protected:
   Y1 x1;
   Y2 x2;
   Y3 x3;
   Y4 x4;
   Y5 x5;
   Y6 x6;
   Y7 x7;
   Y8 x8;
   typedef X8<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8> Type;
public:
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 8;}
   X8(Y1 const& x1_, Y2 const& x2_, Y3 const& x3_,
      Y4 const& x4_, Y5 const& x5_, Y6 const& x6_, Y7 const& x7_,
      Y8 const& x8_):
      x1(x1_),x2(x2_),x3(x3_),x4(x4_),x5(x5_),x6(x6_),x7(x7_),x8(x8_){}

   X8(X7<Y1,Y2,Y3,Y4,Y5,Y6,Y7> const& y_, Y8 const& x8_):
      x1(y_.c1()),x2(y_.c2()),x3(y_.c3()),
      x4(y_.c4()),x5(y_.c5()),x6(y_.c6()),x7(y_.c7()),x8(x8_){}

   X8(void):x1(),x2(),x3(),x4(),x5(),x6(),x7(),x8(){}
   X8(X8<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8> const& z): x1(z.x1), x2(z.x2),
      x3(z.x3),x4(z.x4),x5(z.x5),x6(z.x6),x7(z.x7),x8(z.x8){}

   X8(tuple<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8> const& t): x1(get<0>(t)),
      x2(get<1>(t)),x3(get<2>(t)),x4(get<3>(t)),x5(get<4>(t)),
      x6(get<5>(t)),x7(get<6>(t)),x8(get<7>(t)){}

   Y1& c1(void){ return x1;}
   Y2& c2(void){ return x2;}
   Y3& c3(void){ return x3;}
   Y4& c4(void){ return x4;}
   Y5& c5(void){ return x5;}
   Y6& c6(void){ return x6;}
   Y7& c7(void){ return x7;}
   Y8& c8(void){ return x8;}

   Y1 const& c1(void)const{ return x1;}
   Y2 const& c2(void)const{ return x2;}
   Y3 const& c3(void)const{ return x3;}
   Y4 const& c4(void)const{ return x4;}
   Y5 const& c5(void)const{ return x5;}
   Y6 const& c6(void)const{ return x6;}
   Y7 const& c7(void)const{ return x7;}
   Y8 const& c8(void)const{ return x8;}

   Y1 first()const{ return x1;}
   Y2 second()const{ return x2;}
   Y3 third()const{ return x3;}
   Y4 fourth()const{ return x4;}
   Y5 fifth()const{ return x5;}
   Y6 sixth()const{ return x6;}
   Y7 seventh()const{ return x7;}
   Y8 eighth()const{ return x8;}

   Word nameOf()const
   {
       return Word("X8< "&
          CpmRoot::Name<Y1>()(Y1())&
          ","&
          CpmRoot::Name<Y2>()(Y2())&
          ","&
          CpmRoot::Name<Y3>()(Y3())&
          ","&
          CpmRoot::Name<Y4>()(Y4())&
          ","&
          CpmRoot::Name<Y5>()(Y5())&
          ","&
          CpmRoot::Name<Y6>()(Y6())&
          ","&
          CpmRoot::Name<Y7>()(Y7())&
          ","&
          CpmRoot::Name<Y8>()(Y8())&
          " >");
   }

   tuple<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8> toTup()const
   {
      return tuple<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8>{x1,x2,x3,x4,x5,x6,x7,x8};
   }

};

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7, class Y8>
Z X8<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8>::com(
   X8<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8> const& s)const
{
   if (c1()<s.c1()) return 1;
   if (c1()>s.c1()) return -1;
   if (c2()<s.c2()) return 1;
   if (c2()>s.c2()) return -1;
   if (c3()<s.c3()) return 1;
   if (c3()>s.c3()) return -1;
   if (c4()<s.c4()) return 1;
   if (c4()>s.c4()) return -1;
   if (c5()<s.c5()) return 1;
   if (c5()>s.c5()) return -1;
   if (c6()<s.c6()) return 1;
   if (c6()>s.c6()) return -1;
   if (c7()<s.c7()) return 1;
   if (c7()>s.c7()) return -1;
   if (c8()<s.c8()) return 1;
   if (c8()>s.c8()) return -1;
   return 0;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7, class Y8>
bool X8<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8>::prnOn(ostream& str)const
{
   cpmp(x1);
   cpmp(x2);
   cpmp(x3);
   cpmp(x4);
   cpmp(x5);
   cpmp(x6);
   cpmp(x7);
   cpmp(x8);
   return true;
}

template <class Y1, class Y2, class Y3, class Y4, class Y5,
   class Y6, class Y7, class Y8>
bool X8<Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8>::scanFrom(istream& str)
{
   cpms(x1);
   cpms(x2);
   cpms(x3);
   cpms(x4);
   cpms(x5);
   cpms(x6);
   cpms(x7);
   cpms(x8);
   return true;
}

// In a context where X1,X2,X3,X4,X5,X6,X7,X8 are used as
// names for template arguments one could get in trouble
// by a statement
//
// using CpmArrays::X2;
//
// thus one should avoid this. Then the full name CpmArrays::X2 would
// be necessary  which is too long a name for this innocent construct.
// By the following definition we may use cpmX2, which follows the same
// notational scheme as cpmerror, cpmwrite, ...

#define cpmX2 CpmArrays::X2
#define cpmX3 CpmArrays::X3
#define cpmX4 CpmArrays::X4
#define cpmX5 CpmArrays::X5
#define cpmX6 CpmArrays::X6
#define cpmX7 CpmArrays::X7
#define cpmX8 CpmArrays::X8

#define cpmT2 CpmArrays::T2
#define cpmT3 CpmArrays::T3
#define cpmT4 CpmArrays::T4

} // CpmArrays

#endif
