//? cpmzinterval.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_ZINTERVAL_H_
#define CPM_ZINTERVAL_H_
/*
    Description:
      class for 'intervals' in Z

    Recent history:
    2012-01-03 functions setUnn, setSec, setDif, leftOf, rigOf added,
      #include <cpmx.h> added

*/

#include <cpmsystem.h>
#include <cpmx.h>
   // needed for functions that return a pair of IvZ objects

namespace CpmArrays{

   using CpmRoot::Z;
  // using CpmRoot::div;
   using CpmRoot::modWL;
   using CpmRoot::R;
   using CpmRoot::rnd;
   using CpmRoot::Word;

////////////////////// class IvZ ///////////////////////////////////////
// class IvZ will be used in the definition of the basic array template
// V<> which, in turn, will be used to define the set template S<>.
// So it is not possible at this point to relate IvZ with S<Z>
// although clearly each instance of IvZ determines uniquely
// an instance of S<Z>. The C++ classes IvZ and S<Z> are very
// different in implementation and efficiency. For instance
//    IvZ iv(1,10000);
// defines a 'set' of 10000 integers by memorizing just two
// integers (1 and 10000 in this case). Whereas
//    S<Z> s(10000);
// (which calls a private constructor and thus may appear only in
// member function definition code) allocates memory for
// 10000 integers. When V<> will have been defined, we can give
// as V<IvZ> a storage-economic representation of arbitrary
// finite subsets of Z.
// That IvZ 'represents' subsets of Z expresses itself in
// the member function
//    bool hasElm(Z const& i)const
// which S<Z> defines with just the same signature.

class IvZ{
   // 'intervals of integers', i.e. contiguous finite subsets of Z
   // The void set is included; this is important since T-valued arrays
   // indexed by i's ranging over some instance of IvZ will be used
   // for defining our basic array template class V<T>. V<T> has to have
   // void arrays, which correspond to a void index set.
   // No virtual functions for efficency.
// independent data
   Z c_;
      // c for 'cardinality' arbitrary non-negative integer,
      // 0 for the void set.
   Z l_;
      // l for 'left', arbitrary integer

// dependent quantity
   Z r_;
      // r for 'right'
      // r_= l_ + c_ - 1
      // the right-most element of *this if *this is not void.
      // it is the left neigbor of l_ if the set is void.

// initializing the dependent data
   void ini_(){
      if (c_<0){ c_=0; l_; r_=l_-1;}
      else r_=l_+c_-1;
   }

// enabling declaration macros
   typedef IvZ Type;
public:
// infra-structure by declaration macros
   CPM_IO
      // stream interaction
   CPM_ORDER
      // order-related functions
      // The void interval preceeds every non-void interval.
      // All void intervals are equal.
      // For non-void intervals we use lexicographic order of the pairs
      // (l_,r_).
   CPM_TEST_X
      // functions ran, test, hash, dis, abs, absSqr

   Word nameOf()const{ return "IvZ";}
      //: name of
      // class name of *this

   Word toWord()const;
      //: to word
      // text form of *this

// constructors

   IvZ():c_(0),l_(1){ini_();}
      // default constructor. The l_(1) expresses our preference of
      // vectors for which 1 is the lowes index value.

   explicit IvZ(Z n):c_(n),l_(1){ini_();}
      // The 'standard set' with n elements.
      // For n=0 it is the default constructor, which makes the void set.
      // Here, by definition, the smallest element of any non-void
      // standard set is 1 (not 0!).

   IvZ(Z i, Z j)
   {
      if (i<=j){ c_=j-i+1; l_=i; }
      else { c_=i-j+1; l_=j; }
      r_=l_+c_-1;
   }
      // the minimum contiguous subset of Z which contains both
      // i and j; no order restrictions to i,j. Notice that the
      // void set can't be obtained that way.

   IvZ(Z c, Z l, Z dummy):c_(c),l_(l){dummy; ini_();}
      // constructor which gives the cardinality and the first
      // element.

   IvZ(R a, R b);
      // construction of the smallest IvZ which contains a and b.

// basic properties
   bool isVoid()const{ return c_==0;}
      //: is void
      //':' means that the name is formed according to the CPM standard
      // scheme.

   Z car()const{ return c_;}
      //: cardinality
      // cardinality of the set given by *this

   Z size()const{ return c_;}
      //: size
      // equals the cardinality of the set given by *this

   Z width()const{ return c_==0 ? 0 : c_-1;}
      //:: width
      //'::' means that the name is longer than the one formed according
      // to the CPM standard abbreviation scheme. In most cases
      // no abbreviation at all.
      // distance between the last point and the first set.
      // Thus diameter of the set. Obviously the diameter
      // of a singlet set is zero, as is the diameter of a void set.

// The intention behind inf() and sup() is that
// iv==IvZ(iv.inf(),iv.sup()) for every non-void iv \in IvZ
// Thus inf() and sup() have to warn if called for a void set.
   Z inf()const;
      //: infimum

   Z sup()const;
      //: supremum

   Z mean()const{ return l_ + c_/2; }
      //: mean
      // IvZ(1,5).mean() = 3 = median of (1,2,3,4,5)
      // Iv(1,4).mean() = 3 lowerst integer which is >= median of
      // (1,2,3,4)

   Z b()const{ return l_;}
      //. begin
      //'.' means that the name is shorter than the one formed according
      // to the CPM standard abbreviation scheme. In most cases
      // abbreviation to the first letter of the full name.

   Z n()const{ return r_+1;}
      //. next
      // relative to the numbers belonging to *this
      // this is the next number with respect to standard order

   IvZ& n_(){ c_++; ini_(); return *this; }
      //. next
      // iv.n_() results from iv if the number iv.n() is appended

   IvZ& p_(){ c_++; l_--; ini_(); return *this; }
      //. previous
      // iv.n_() results from iv if the number iv.n() is appended

   Z e()const{ return r_;}
      //. end
      // actually the last number belonging to *this
      // if *this is non-void as a set.
      // Looping over an IvZ ivz can allways be done as
      // for (Z i=ivz.b();i<=ivz.e();++i) ... ivz[i]....
      // or, equally well
      // for (Z i=ivz.b();i<ivz.n();++i) ... ivz[i]....

   Z fir()const{ return l_;}
      //: first

   Z last()const{ return r_;}
      //: last

   Z operator[](Z i)const{ if (i<=1) return l_; else return r_;}
      //: []

   Z gap(IvZ const& iv);
      //: gap
      // Distance between the sets *this and iv.
      // Here the understanding is that the distance from a void
      // set to any set is 0.

   Z relPos(Z i)const
      //: relative position
      // gives a code for the location of i relative to *this.
   {
      if (i<l_) return -1;
      else if (i>r_) return 1;
      else return 0;
   }

   Z relInd(Z i)const{ return i-l_;}
      //: relative index
      // gives the difference of i with respect to the left end
      // ('the origin') of *this.
      // If *this is void the result is meaningless. Nevertheless
      // for efficiency reasons we do not test voidness.

   Z fromRelInd(Z i)const{ return l_+i;}
      //: from relative index
      // iv.fromRelInd(i) returns an index j such that i = iv.relInd(j)

   Z ri(Z i)const{ return i-l_;}
      //. relative index
      // short form of relInd

   Z cyc(Z const& i)const
      //: cyclic
      // Z-->Z, i|-->iv.cyc(i) is a periodic function with period
      // iv.car(), which coincides with the idendity function id on
      // the set iv.
      // Test: MathematicaMisc/cyc.nb. This did not work:
      // Mathematica has a different reminder definition from % of C++
      // Now I have made a function which follows the Mathematica
      // convention.
      {
         return l_+modWL(i-l_,c_);
      }

   Z con(Z i)const
      //: constant
      // Z-->Z, i|-->iv.con(i) is the identity function id on
      // iv and defined as the constant continuation of id
      // outside of iv.
   {
      if (i<=l_) return l_;
      else if (i>=r_) return r_;
      else return i;
   }

   IvZ leftOf(Z i)const
      //: left of
      // Returns the part of *this the elements j of which
      // satisfy j < i (and the void interval else).
   {
      if (l_ >= i) return IvZ();
      if (r_ < i) return *this;
      return IvZ(l_,i-1);
   }

   IvZ rigOf(Z i)const
      //: right of
      // Returns the part of *this the elements j of which
      // satisfy  i < j (and the void interval else).
   {
      if (r_ <= i) return IvZ();
      if (l_ > i) return *this;
      return IvZ(i+1,r_);
   }

   IvZ operator|(IvZ const& iv)const;
      // minimal interval that contains the union
      // join or l.u.b. (lowest upper bound) in lattice terminology

   IvZ join(IvZ const& iv)const{ return (*this)|iv;}
      //:join

   IvZ operator&(IvZ const& iv)const;
      // section, intersection
      // meet or g.l.b (greates lower bound) in lattice terminology.

   IvZ meet(IvZ const& iv)const{ return (*this)&iv;}
      //: meet

// set operations may yield a pair of IvZs as result
   T2<IvZ> setUnn(IvZ const& iv)const;
      //: set union
      // Consider T2<IvZ> res = iv1.setUnn(iv2);
      // If the set union of iv1 and iv2 is contiguous, the corresponding
      // IvZ-object will be res.e(), and res.b() is set to be void.
      // This order is chosen to have res.b() < res.e() in all cases.
      // Notice that the void IvZ preceeds all IvZ's.
      // If the set union is not contiguous it determines two IvZ-objects
      // which will be stored in res such that res.b() < res.e().

   T2<IvZ> setSec(IvZ const& iv)const{return T2<IvZ>(IvZ(),meet(iv));}
      //: set section
      // Coincides with lattice meet, which it returns as the second
      // component of the result. The first component is set to be void.
      // (Then the components are ordered).

   T2<IvZ> setDif(IvZ const& iv)const;
      //: set difference
      // Consider T2<IvZ> res = iv1.setDif(iv2);
      // If the set difference iv1\iv2 is contiguous, the corresponding
      // IvZ-object will be res.e(), and res.b() is set to be void.
      // See setUnn for reasons.
      // If the set union is not contiguous it determines two IvZ-objects
      // which will be stored in res such that res.b() < res.e().

// further functions
   IvZ app(IvZ const& iv)const
   {
      if (c_==0) return iv;
      else return IvZ(c_+iv.c_,l_);
   }
      //: append
      // We return a IvZ which is obtained from *this (if it is not void)
      // by appending iv.car() elements at the right end. If this is void
      // we simply return iv.

   bool hasElm(Z i)const{ return l_<=i && i<=r_;}
      //: has element
      // answers the question whether point i is an element of the
      // interval *this. S<Z> has a member function of the same
      // signature.

   bool ni(Z i)const{ return hasElm(i);}
      //: ni
      // 'in' reversed, from LATEX symbol \ni for the mirror image of
      // the epsilon-like 'is element'-symbol

   bool contains(IvZ const& iv)const;
      //:: contains
      // says whether iv is a subset (true subset 'or equal') of *this
      // or not.

   bool isSubSetOf(IvZ const& iv)const{  return iv.contains(*this);}
      //: is sub-set of
      // returns true iff *this is a subset of iv

   bool isSupSetOf(IvZ const& iv)const{ return (*this).contains(iv);}
      //: is super-set of
      // returns true iff s is a subset of *this

   IvZ operator+(Z s)const{return IvZ(l_+s,r_+s);}
      //: +
      // Returns a copy of *this with both ends shifted by s.

   IvZ operator-(Z s)const{return IvZ(l_-s,r_-s);}
      //: -
      // Returns a copy of *this with both ends shifted by -s.

   IvZ& operator+=(Z s){ l_+=s; r_+=s; return *this;}
      //: +=

   IvZ& operator-=(Z s){ l_-=s; r_-=s; return *this;}
      //: -=

   IvZ operator+(IvZ const& iv)const
      //: +
      // Range of i+j for i \in *this and j \in iv.
   { return IvZ(b()+iv.b(),e()+iv.e());}

   IvZ operator-(IvZ const& iv)const
      //: -
      // Range of i-j for i \in *this and j \in iv.
   { return IvZ(b()-iv.b(),e()-iv.e());}

   IvZ& b_(Z i){Z ir=i-l_; return operator+=(ir);}
      //. set begin
      // Changes b() into i and returns a reference to the
      // modified object.
      // Recall that the names of non-constant functions end in '_'.

   IvZ& e_(Z i){Z ir=i-r_; return operator+=(ir);}
      //. set end
      // Changes e() into i and returns a reference to the
      // modified object.

   Z ranSel(Z j=0)const;
      //: random selection
      // Returns a 'randomly' selected element in *this
      // Here the argument j plays the same role as in
      //    R CpmRoot::randomR(Z j=0);
      // (see cpmnumbers.h)
      // Notice that there is also a function ran(Z) which
      // returns a IvZ, so that the present function can not
      // be named simply ran.
      // If *this is void, we return 0 and issue a warning.

   static Z testLatIdn(Z complexity=512);
      //: test lattice identities
      // Tests the complete set of lattice identities.
      // See e.g. S. Mac Lane, G. Birkhoff: Algebra MacMillan 1968
      // p. 487 Theorem 4.
 };

//////////////////////////// Implementation ////////////////////////////
// see also cpmzinterval.cpp
inline Z IvZ::inf()const
{
   if (c_==0)
      cpmwarning("Iv::inf() undefined for void instance, b() returned");
   return l_;
}
inline Z IvZ::sup()const
{
   if (c_==0)
      cpmwarning("Iv::sup() undefined for void instance, e() returned");
   return r_;
}

} // namespace

#endif
