//? cpmvo.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_VO_H_
#define CPM_VO_H_
/*
   Purpose: Defining a array class Vo<T> which exploits the assumption\
    that
   T supports ordering. See cpmv.h

*/
#include <cpmv.h>
#include <utility>
#include <algorithm>
#include <unordered_set>

namespace CpmArrays{

///////////////////////// class Vo ///////////////////////////////////

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.
   // T < T and T > T assumed.

class Vo : public V<T>{ // version of V with order-related operations
   // derivation which implements the order interface

   typedef Vo<T> Type;
   typedef V<T> Base;
public:

Vo():V<T>(){}

explicit Vo(Z n):V<T>(n){}

explicit Vo(IvZ ivz):V<T>(ivz){}

   Vo(IvZ ivz, T const& t):V<T>(ivz,t){}

// constructors from explicit lists
explicit  Vo(std::initializer_list<T> il ):V<T>(il){}
   // requires C++11
      // constructors from explicit lists such as
      // Vo<Z> v{1,2,4,8};

   Vo(Z n, T const& t):V<T>(n,t){}

   Vo(const V<T>& h):V<T>(h){}
         // 'down-cast'-constructor

   Vo(const Vo<T>& h):V<T>(h){}
         // 'down-cast'-constructor

   virtual V<T>* clone(void)const{ return new Vo(*this);}

   virtual Word nameOf()const
      //: name of
   {
      Word wi="Vo<";
      return wi&CpmRoot::Name<T>()(T())&">";
   }

   V<Z> permutationForIncreasingOrder(void)const;
      // Returned is an array per such that i<j enforces
      //  (*this)[per[i]]<(*this)[per[j]]
      //       or
      // (*this)[per[i]]==(*this)[per[j]]
      // Assumes that valid indexing starts with 1 (i.e. b()==1)

   V<Z> permutationForDecreasingOrder(void)const
      // analogous to previous function
   { return permutationForIncreasingOrder().rev();}

   void order_(void);
      // Permutes the indexes such that
      // (*this)[i]<(*this)[j] || (*this)[i]==(*this)[j]
      // whenever i<j. The components (instances of T) are allowed to
      // be large objects since excessive copying is avoided.

   Vo<T>& sort_(void){ order_(); return *this;}
      //: sort

   void sortInc_();

   void unique_(bool ord=true);
      // Calling this function changes *this (need not to be ordered)
      // into an array in which no two components have the same vale.
      // Thus 'elimination of doublets'. Nothing about potential
      // reordering is known. If the argument is 'true' the array, after
      // the doublets have been eliminated gets ascendingly ordered.

   V<Z> perForDecOrd(void)const
      //: permutation for decreasing order
   { return permutationForDecreasingOrder();}

   V<Z> perForIncOrd(void)const
      //: permutation for increasing order
   { return permutationForIncreasingOrder();}

   Vo<T> increasingCopy(void)const;
      // returns a increasingly ordered version of (*this) without
      // changing *this

   Vo<T> incCopy(void)const{ return increasingCopy();}
      //: increasing copy

   Vo<T> strictlyIncreasingVersion(void)const;
      // returns a stricly increasingly ordered version of (*this)
      // which leaves out multiply appearing components
      // (again, without changing *this)

   Vo<T> strIncVer(void)const{ return strictlyIncreasingVersion();}
      //: strictly increasing version

   Vo<T> decreasingCopy(void)const;
      // returns a decreasingly ordered version of (*this) without
      // changing *this

   Vo<T> decCopy(void)const{ return decreasingCopy();}
      //: decreasing copy

   Vo<T> purge(const V<T>& s)const;
      // returned is a list which is obtained from *this by eliminating
      // all components which equal to one of the components of s. Beside
      // of this removal, the order of the components in *this is
      // retained.

   Vo<T> pur(const V<T>& s)const{ return purge(s);}
      //: purge

   Z locate(T const& t)const{ return loc3(t).c1();}
      // locate
      // returned is a value j such that t satisfies
      //    (*this)[j] <= x < (*this)[j+1].
      // (*this) must be monotonic increasing
      // j=b()-1 or j=e()+1 is returned to indicate that t is out of
      // range.

    X3<Z,Z,bool> loc3(T const& t)const;
      //: locate with a 'three-fold return value' res of type X3<Z,Z,bool>
      // which describes how the object t is placed relativ to the
      // components of the vector *this. Let us explain:
      // We write
      //    i for res.c1() ( an integer)
      //    j for res.c2() ( an integer)
      //    found for res.c3() ( an boolean).
      // The meaning of j depends on the quantities j and found.
      // Let us begin with discussing the variable 'found':
      // found==true implies j==0 and (*this)[i]==t.
      // found==false implies j in {-1, 0, 1}.
      //    j==0 implies (*this)[i] < t < (*this)[j+1]
      //    j==-1 implies i == iL-1 where iL is the first valid index
      //       of the vector *this
      //    j==1 implies i == iU+1 where iU is the last valid index
      //       of the vector *this.

  // Vo<T> enq(T const&)const;
      //: enqueue
      // assumes that (*this) is monotonic. Then t will be inserted in
      // the orderly correct position into of a copy of (*this). This copy
      // remains monotonic and gets returned.
      // The second argument comes only in action for dim()==1, since
      // then no direction is provided by *this and a direction is needed
      // for the result (for dim()==0, the result has dim()==1 and no
      // direction).

   Z  enq_(T const&);
      // assumes that (*this) is monotonic. Then t will be inserted in
      // the orderly correct position into of a copy of (*this). This copy
      // remains monotonic and gets returned.
      // The second argument comes only in action for dim()==1, since
      // then no direction is provided by *this and a direction is needed
      // for the result (for dim()==0, the result has dim()==1 and no
      // direction).

   Z enqX_(T const& t){ enq_(t); return find(t);}
      // enplaces t into *this and returns the index i such that
      // (*this)[i]=t

   T sup()const;
      //: supremum
      // returns the largest component of *this; faster than getting the
      // largest element by ordering

   T inf()const;
      //: infimum
      // returns the smallest component of *this; faster than getting the
      // largest element by ordering

   Z indSup()const;
      //: index (of) supremum
      // returns the smallest i for which (*this)[i]=(*this).sup()

   Z indInf()const;
      //: index (of) infimum
      // returns the smallest i for which (*this)[i]=(*this).inf()

   T2<Z> indInfSup()const;
      //: returns indInf(), indSup() as a pair in a single pass

   T2<T> infSup()const;
      //: infimum supremum
      // Let res be the return value, then res.first==inf(),
      // res.second==sup(). Fast single-pass operation
};


// Function indexx from Press et al. changed into a function template
// p. 339  (role of M: see p. 333; algorithm is based on Quicksort, see
// sort(), anyway the choice of M may influence performance but never
// puts a limitation on the size of the vector to be ordered.

namespace{
   inline void swapZ(Z& i, Z& j){ Z ii=i; i=j; j=ii;}
}

template <class T>
V<Z> Vo<T>::permutationForIncreasingOrder(void)const
// Indexes an array v[1..n], i.e. outputs the array indx[1..n] such that
// v[indx[j]] is in ascending order for j=1,2,...,n. The input
// quantities n and v are not changed. Especially, v-elements are
// not copied back and forth which is essential if T is a large class
{
   const Z mL=4;
   cpmmessage(mL,"Vo<T>::permutation...(): started");
   const Z M=7;
   const Z NSTACK=50;

   Z nVec=Base::dim();
   Vo<Z> indx(nVec);
   Z i,indxt,ir=nVec,j,k,l=1,jstack=0;
   T a;

   Vo<Z> istack(NSTACK);

   for (j=1;j<=nVec;j++) indx[j]=j;
   for (;;) {
      if (ir-l < M) {
         for (j=l+1;j<=ir;j++) {
            indxt=indx[j];
            a=(*this)[indxt];
            for (i=j-1;i>=1;i--) {
               if ((*this)[indx[i]] < a) break;
               if ((*this)[indx[i]] == a) break;
               indx[i+1]=indx[i];
            }
            indx[i+1]=indxt;
         }
         if (jstack == 0) break;
         ir=istack[jstack--];
         l=istack[jstack--];
      } else {
         k=(l+ir)/2; // was >> 1 instead of /2 before
         swapZ(indx[k],indx[l+1]);
         if ((*this)[indx[l+1]] > (*this)[indx[ir]]) {
            swapZ(indx[l+1],indx[ir]);
         }
         if ((*this)[indx[l]] > (*this)[indx[ir]]) {
            swapZ(indx[l],indx[ir]);
         }
         if ((*this)[indx[l+1]] > (*this)[indx[l]]) {
            swapZ(indx[l+1],indx[l]);
         }
         i=l+1;
         j=ir;
         indxt=indx[l];
         a=(*this)[indxt];
         for (;;) {  // the i<nVec and j>1 test added by UM, otherwise in
               // patological cases of < and > relations we get a range
               // error
            do i++; while (i<nVec && (*this)[indx[i]] < a);
            do j--; while (j>1 && (*this)[indx[j]] > a);
            if (j < i) break;
            swapZ(indx[i],indx[j]);
         }
         indx[l]=indx[j];
         indx[j]=indxt;
         jstack += 2;
         if (jstack > NSTACK) cpmerror("NSTACK too small in ordering()");
         if (ir-i+1 >= j-l) {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
         } else {
            istack[jstack]=j-1;
            istack[jstack-1]=l;
            l=i;
         }
      }
   }
   cpmmessage(mL,"Vo<T>::permutation...(): done");
   return indx;
}

template <class T>
void Vo<T>::unique_(bool ord){
   std::unordered_set<T> s;
   for (Z i: Base::p_) s.insert(i);
   Base::p_.assign(s.begin(),s.end());
   if (ord) std::sort( Base::p_.begin(), Base::p_.end() );
   Base::sz_=Base::p_.size();
   Base::iv_=IvZ(Base::sz_,Base::b());
}

/*********************************************
template <class T>
Z Vo<T>::locate(T const& x)const
{
   Word loc("Z Vo<T>::locate(T&)const");

   Z sz=Base::sz_;
   cpmassert(sz>0,loc);

   Z jl=Base::b()-1;
   Z ju=Base::e()+1;

   T xl=Base::fir();
   T xu=Base::last();
   cpmassert(xl<xu,loc); // ascending order needed. Of course this is not
      // sufficient condition for this.
   Z res{jl-1};
   if (x<xl){
      res=jl;
      return res;
   }
   if (x>xu){
      res=ju;
      return res;
   }
   auto q=std::upper_bound(Base::p_.begin(),Base::p_.end(),x);
   if (q==Base::p_.end()){
      res = ju;
      return res;
   }
   else{
      Z locDis=q-Base::p_.begin();
      res=jl+locDis;
      cpmassert(Base::valInd(res),loc);
      cpmassert(Base::valInd(res+1),loc);
      cpmassert((*this)[res]<=x && x<(*this)[res+1],loc);
      return res;
   }
}
*******************************/

template <class T>
X3<Z,Z,bool> Vo<T>::loc3(T const& x)const
{
   Word loc("X3<Z,Z,bool> Vo<T>::loc3(T&)const");

   Z sz=Base::sz_;
   cpmassert(sz>0,loc);

   Z jl=Base::b()-1;
   Z ju=Base::e()+1;

   T xl=Base::fir();
   T xu=Base::last();
   cpmassert(xl<xu,loc); // ascending order needed. Of course this is not
      // sufficient condition for this.
   X3<Z,Z,bool> res(jl-1,-2,false); // never a valid return value
   if (x<xl){
      res=X3<Z,Z,bool>(jl,-1,false);
      return res;
   }
   if (x>xu){
      res=X3<Z,Z,bool>(ju,1,false);
      return res;
   }
   auto q=std::upper_bound(Base::p_.begin(),Base::p_.end(),x);
   if (q==Base::p_.end()){
      res = X3<Z,Z,bool>(ju,1,false);
      return res;
   }
   else{
      Z locDis=q-Base::p_.begin();
      Z j0=jl+locDis;
      Z j1=j0+1;
      cpmassert(Base::valInd(j0),loc);
      cpmassert(Base::valInd(j1),loc);
      cpmassert((*this)[j0]<=x && x<(*this)[j1],loc);
      bool found = (*this)[j0]==x;
      return X3<Z,Z,bool>(j0,0,found);
   }
}

// functions based on order relations in T

template <class T>
Vo<T> Vo<T>::increasingCopy(void)const
{
   Vo<Z> per=permutationForIncreasingOrder();
   Z i, n=Base::dim();
   Vo<T> y(n);
   for (i=1;i<=n;i++) y[i]=(*this)[per[i]];
   return y;
}

namespace{
   template <class T>
   bool fEqual(T const& t1, T const& t2){ return t1==t2;}
}

template <class T>
Vo<T> Vo<T>::strictlyIncreasingVersion(void)const
{
   Vo<T> res=increasingCopy();
   return res.condense(fEqual).c1();
}

template <class T>
Vo<T> Vo<T>::decreasingCopy(void)const
{
   Vo<Z> per=permutationForIncreasingOrder();
   Z i, n=Base::dim();
   Vo<T> y(n);
   for (i=1;i<=n;i++) y[1+n-i]=(*this)[per[i]];
   return y;
}

template <class T>
void Vo<T>::sortInc_()
{
   std::sort(Base::p_.begin(), Base::p_.end());
}

template <class T>
void Vo<T>::order_(void){ sortInc_();}

template <class T >
Vo<T> Vo<T>::purge(const V<T>& s)const
{
   Z i,j,n=Base::dim(),ns=s.dim();
   V<B> vs(n,B(true));
   for (i=1;i<=n;i++){
      T t=(*this)[i];
      for (j=1;j<=ns;j++){
         if( t==s[j] ){ vs[i]=false; break; }
      }
   }
   return Base::select(vs);
}

template <class T >
T Vo<T>::sup()const
{
   Z n=Base::dim();
   cpmassert(n>0,"T Vo<T>::sup()const");
   T t,res=Base::fir();
   for (Z i=Base::b()+1;i<=Base::e();i++){
      t=Base::cui(i);
      if (t>res) res=t;
   }
   return res;
}

template <class T >
Z Vo<T>::indSup()const
{
   Z n=Base::dim();
   cpmassert(n>0,"Vo<T>::indSup()const");
   T t,valMax=(*this)[1];
   Z res=1;
   for (Z i=2;i<=n;i++){
      t=(*this)[i];
      if (t>valMax){ valMax=t; res=i;}
   }
   return res;
}

template <class T >
T Vo<T>::inf()const
{
   Z n=Base::dim();
   cpmassert(n>0,"T Vo<T>::inf()const");
   T t,res=Base::fir();
   for (Z i=Base::b()+1;i<=Base::e();i++){
      t=Base::cui(i);
      if (t<res) res=t;
   }
   return res;
}

template <class T >
Z Vo<T>::indInf()const
// needed as fast as possible
{
   Z n=Base::dim();
   cpmassert(n>0,"Z Vo<T>::indInf()const");
   Z res=0;
   T t,valMin=Base::cui(res);

   for (Z i=1;i<n;i++){
      t=Base::cui(i);
      if (t<valMin){ valMin=t; res=i;}
   }
   return ++res;
}

template <class T >
T2<Z> Vo<T>::indInfSup()const
{
   Z n=Base::dim();
   if (n<=0) cpmerror("Vo<T>::indInfSup(): dim()<=0");
   T t,valMin=(*this)[1],valMax=valMin;
   Z resMin=1;
   Z resMax=1;
   for (Z i=2;i<=n;i++){
      t=(*this)[i];
      if (t<valMin){ valMin=t; resMin=i;}
      if (t>valMax){ valMax=t; resMax=i;}
   }
   return T2<Z>(resMin,resMax);
}

template <class T >
T2<T> Vo<T>::infSup()const
{
   Z n=Base::dim();
   if (n<=0) cpmerror("Vo<T>::infSup(): dim()<=0");
   T t,valMin=Base::fir(),valMax=valMin;
   for (Z i=Base::b()+1;i<=Base::e();i++){
      t=Base::cui(i);
      if (t<valMin){ valMin=t; }
      if (t>valMax){ valMax=t; }
   }
   return T2<T>(valMin,valMax);
}

/**********
template <class T >
Vo<T> Vo<T>::enq(T const& t)const
{
   Z mL=3;
   Word loc("Vo<T> Vo<T>::enqueue(T const& t, bool asc)const");
   CPM_MA
   Z d=Base::dim();
   if (d==0) return Vo<T>(1,t);
   if (d==1){
      Vo<T> res(2);
      T t1=(*this)[1];
      if (asc){
         if (t1<t){
            res[1]=t1;
            res[2]=t;
         }
         else{
            res[1]=t;
            res[2]=t1;
         }
      }
      else{
         if (t1<t){
            res[1]=t;
            res[2]=t1;
         }
         else{
            res[1]=t1;
            res[2]=t;
         }
      }
      return res;
   }
   Z i=locate(t);
   CPM_MZ
   return Base::ins(i+1,t);
}
***************/

template <class T >
Z Vo<T>::enq_(T const& t)
{
   Z mL=3;
   Word loc("Vo<T> Vo<T>::enqueue(T const& t, bool asc)const");
   CPM_MA
   Z d=Base::size();
   Z dn=0;
   if (d==0){
      Base::p_.push_back(t);
      Base::iv_.n_();
      Base::sz_++;
      return 1;
   }
   if (d==1){
      T c1=Base::fir();
      if (t==c1) return 0;
      else if (t>c1){
         Base::p_.push_back(t);
         Base::iv_.n_();
         Base::sz_++;
         return 1;
      }
      else{
         Base::p_.push_back(t);
         std::rotate(Base::p_.rbegin(), Base::p_.rbegin() + 1,
         Base::p_.rend());
         Base::iv_.n_();
         Base::sz_++;
         return 1;
      }
   }
   // d >= 2
   auto y=loc3(t); // the more informative version of locate
   Z i=y.c1();
   Z j=y.c2();
   bool found=y.c3();
   if (j==1){
      Base::p_.push_back(t);
      Base::sz_++;
      Base::iv_.n_();
      dn=1;
   }
   else if(j==-1){
      Base::p_.push_back(t);
      std::rotate(Base::p_.rbegin(), Base::p_.rbegin() + 1,
         Base::p_.rend());
         Base::sz_++;
         Base::iv_.n_();
         dn=1;
   }
   else{
      if (found){
         ; // since t is already in the array, we have not to insert it.
      }
      else{
         Base::p_.insert(Base::p_.begin()+(i-Base::b()+1),t);
         Base::sz_++;
         Base::iv_.n_();
         dn = 1;
      }
   }
   CPM_MZ
   return dn;
}

///////////////////////// iterated templates ///////////////////////////
// describing multi-indexed quantities
//////////////////////////// class VVo /////////////////////////////////
// matrices

template <class T>
class VVo: public VV<T>{// version of VV with order-related operations

   typedef VV<T> Base;
   typedef VVo<T> Type;

public:
   CPM_ORDER
// constructors
   VVo(Z d1, Z d2):Base(d1,d2){}
   VVo( Z d1, Z d2, T const& t):Base(d1,d2,t){}
      // all components are equal to t
   VVo(void):Base(){}
   VVo(VV<T> const& x):Base(x){}
   VVo( V< V<T> > const& x):Base(x){}
};

template <class T >
Z VVo<T>::com(VVo<T> const& s)const
{
   Z d1=Base::dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
      for (Z i=1;i<=d1;i++){
         Vo<T> ti=(*this)[i];
         Vo<T> si=s[i];
         Z ci=CpmRoot::comT<T>(ti,si);
         if (ci!=0) return ci;
      }
      return 0;
   }
}

///////////////////////// class VVVo ///////////////////////////////////
// tensors of rank 3

template <class T>
class VVVo: public VVV<T>{// version of VVV with order-related operations

   typedef VVV<T> Base;
   typedef VVVo<T> Type;

public:
   CPM_ORDER
// constructors
   VVVo(Z d1,Z d2,Z d3):Base(d1,d2,d3){}
   VVVo(Z d1,Z d2,Z d3,T const& t):Base(d1,d2,d3,t){}
      // all components are equal to t
   VVVo(void):Base(){}
   VVVo(VVV<T> const& x):Base(x){}
   VVVo( V< V< V<T> > > const& x):Base(x){}
};

template <class T >
Z VVVo<T>::com(VVVo<T> const& s)const
{
   Z d1=Base::dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
      for (Z i=1;i<=d1;i++){
         VVo<T> ti=(*this)[i];
         VVo<T> si=s[i];
         Z ci=CpmRoot::comT<T>(ti,si);
         if (ci!=0) return ci;
      }
      return 0;
   }
}

///////////////////////// class VVVVo //////////////////////////////////
// tensors of rank 4

template <class T>
class VVVVo:public VVVV<T>{//version of VVVV with order-related operations

   typedef VVVV<T> Base;
   typedef VVVVo<T> Type;

public:
   CPM_ORDER
// constructors
   VVVVo(Z d1,Z d2,Z d3,Z d4):Base(d1,d2,d3,d4){}
   VVVVo(Z d1,Z d2,Z d3,Z d4,T const& t):Base(d1,d2,d3,d4,t){}
      // all components are equal to t
   VVVVo(void):Base(){}
   VVVVo(VVVV<T> const& x):Base(x){}
   VVVVo( V< V< V< V<T> > > > const& x):Base(x){}
};

template <class T >
Z VVVVo<T>::com(VVVVo<T> const& s)const
{
   Z d1=Base::dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
      for (Z i=1;i<=d1;i++){
         VVVo<T> ti=(*this)[i];
         VVVo<T> si=s[i];
         Z ci=CpmRoot::comT<T>(ti,si);
         if (ci!=0) return ci;
      }
      return 0;
   }
}

} // namespace

#endif
